//	AmbisonicData.h - Simple struct representing an Ambisonic-ally encoded
//					  sample.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef AMBISONICDATA_H_
#define AMBISONICDATA_H_

#include "ThreeFloats.h"
#include <cmath>
#include <iostream>


///	Struct representing an Ambisonic-ally encoded sample.
struct dllStuff AmbisonicData
{
	float W;
	float X;
	float Y;
	float Z;

	///	Default constructor.
	AmbisonicData():
	W(0.0f),
	X(0.0f),
	Y(0.0f),
	Z(0.0f)
	{
		
	};

	///	Simple constructor.
	/*!
		All values are set to val.
	 */
	AmbisonicData(float val):
	W(val),
	X(val),
	Y(val),
	Z(val)
	{
		
	};

	///	Constructor.
	AmbisonicData(float w, float x, float y, float z):
	W(w),
	X(x),
	Y(y),
	Z(z)
	{
		
	};

	///	Copy Constructor.
	AmbisonicData(const AmbisonicData& other):
	W(other.W),
	X(other.X),
	Y(other.Y),
	Z(other.Z)
	{
		
	};

	///	Assignment operator.
	AmbisonicData& operator=(const AmbisonicData& op2)
	{
		W = op2.W;
		X = op2.X;
		Y = op2.Y;
		Z = op2.Z;
		return *this;
	};

	///	Assignment operator (assigns all members the value of val).
	AmbisonicData& operator=(float val)
	{
		W = val;
		X = val;
		Y = val;
		Z = val;
		return *this;
	};

	///	Assignment operator (ThreeFloats).
	/*!
		This version converts a 3d position into the appropriate Ambisonic
		components necessary to place a monophonic sound at that position in
		the 3d audio space (to do this you multiply the monophonic sound by
		the Ambisonic result of this operation, to get the Ambisonic output).

		This is a relatively processor-heavy operation, so be careful about
		using it in critical parts of your code.

		\todo Can I make this more efficient?
	 */
	AmbisonicData& operator=(const ThreeFloats& val)
	{
		float distance, distanceSquared, invDistance;
		//float theta, phi, h;
		ThreeFloats tempf(val.x, val.y, val.z);

		//First get the x, y, z positions on the unit sphere.
		distanceSquared = (val.x*val.x) + (val.y*val.y) + (val.z*val.z);
		distance = sqrtf(distanceSquared);
		if(distance > 0.3f)
			invDistance = 1.0f/distance;
		else
			invDistance = 0.3f;

		/*A different approach, with more calculations.  All we really need to
		do is scale the position vector to the unit sphere.

		theta = static_cast<float>(atan2(val.z, val.x)); //azimuth
		phi = asinf(val.y/distance); //elevation

		tempf.y = sinf(phi);
		h = sqrtf(1 - (tempf.y*tempf.y));
		tempf.z = h * sinf(theta);
		tempf.x = h * cosf(theta);*/

		tempf.x = val.x * invDistance;
		tempf.y = val.y * invDistance;
		tempf.z = val.z * invDistance;

		//These lines unnecessary, since distance is 1.
		/*distanceSquared = (tempf.x*tempf.x) + (tempf.y*tempf.y) + (tempf.z*tempf.z);
		distance = sqrtf(distanceSquared);*/

		//The 0.3f is to avoid divide by zero errors - apparently the
		//0.3 value was suggested by Richard Furse.
		//invDistance = 1.0f/(distanceSquared+0.3f);
		//0.76923076923076923076923076923077f = 1/1.3
		X = -tempf.z * 0.76923076923076923076923076923077f;
		Y = tempf.x * 0.76923076923076923076923076923077f;
		Z = tempf.y * 0.76923076923076923076923076923077f;
		if(distance < 0.7f)
			W = 0.707f/(distance+0.3f);
		else
			W = 0.707f;

		return *this;
	};

	///	Addition operator.
	AmbisonicData operator+(const AmbisonicData& op2)
	{
		AmbisonicData retval;

		retval.W = W + op2.W;
		retval.X = X + op2.X;
		retval.Y = Y + op2.Y;
		retval.Z = Z + op2.Z;

		return retval;
	};

	///	Addition operator (adds val to all members).
	AmbisonicData operator+(float val)
	{
		AmbisonicData retval;

		retval.W = W + val;
		retval.X = X + val;
		retval.Y = Y + val;
		retval.Z = Z + val;

		return retval;
	};

	///	Subtraction operator.
	AmbisonicData operator-(const AmbisonicData& op2)
	{
		AmbisonicData retval;

		retval.W = W - op2.W;
		retval.X = X - op2.X;
		retval.Y = Y - op2.Y;
		retval.Z = Z - op2.Z;

		return retval;
	};

	///	Subtraction operator (subrtacts val from all members).
	AmbisonicData operator-(float val)
	{
		AmbisonicData retval;

		retval.W = W - val;
		retval.X = X - val;
		retval.Y = Y - val;
		retval.Z = Z - val;

		return retval;
	};

	///	Multiplication operator.
	AmbisonicData operator*(const AmbisonicData& op2)
	{
		AmbisonicData retval;

		retval.W = W * op2.W;
		retval.X = X * op2.X;
		retval.Y = Y * op2.Y;
		retval.Z = Z * op2.Z;

		return retval;
	};

	///	Multiplication operator (multiplies all members by val).
	AmbisonicData operator*(float val)
	{
		AmbisonicData retval;

		retval.W = W * val;
		retval.X = X * val;
		retval.Y = Y * val;
		retval.Z = Z * val;

		return retval;
	};

	///	Division operator.
	AmbisonicData operator/(const AmbisonicData& op2)
	{
		AmbisonicData retval;

		retval.W = W / op2.W;
		retval.X = X / op2.X;
		retval.Y = Y / op2.Y;
		retval.Z = Z / op2.Z;

		return retval;
	};

	///	Division operator (divides all members by val).
	AmbisonicData operator/(float val)
	{
		AmbisonicData retval;

		retval.W = W / val;
		retval.X = X / val;
		retval.Y = Y / val;
		retval.Z = Z / val;

		return retval;
	};

	///	Addition + assignment operator.
	AmbisonicData& operator+=(const AmbisonicData& op2)
	{
		W += op2.W;
		X += op2.X;
		Y += op2.Y;
		Z += op2.Z;

		return *this;
	};

	///	Addition + assignment operator (adds val to all members).
	AmbisonicData operator+=(float val)
	{
		W += val;
		X += val;
		Y += val;
		Z += val;

		return *this;
	};

	///	Subtraction + assignment operator.
	AmbisonicData& operator-=(const AmbisonicData& op2)
	{
		W -= op2.W;
		X -= op2.X;
		Y -= op2.Y;
		Z -= op2.Z;

		return *this;
	};

	///	Subtraction + assignment operator (adds val to all members).
	AmbisonicData operator-=(float val)
	{
		W -= val;
		X -= val;
		Y -= val;
		Z -= val;

		return *this;
	};

	///	Multiplication + assignment operator.
	AmbisonicData& operator*=(const AmbisonicData& op2)
	{
		W *= op2.W;
		X *= op2.X;
		Y *= op2.Y;
		Z *= op2.Z;

		return *this;
	};

	///	Multiplication + assignment operator (adds val to all members).
	AmbisonicData operator*=(float val)
	{
		W *= val;
		X *= val;
		Y *= val;
		Z *= val;

		return *this;
	};

	///	Division + assignment operator.
	AmbisonicData& operator/=(const AmbisonicData& op2)
	{
		W /= op2.W;
		X /= op2.X;
		Y /= op2.Y;
		Z /= op2.Z;

		return *this;
	};

	///	Division + assignment operator (adds val to all members).
	AmbisonicData operator/=(float val)
	{
		W /= val;
		X /= val;
		Y /= val;
		Z /= val;

		return *this;
	};

	///	Equality operator.
	bool operator==(const AmbisonicData& op2)
	{
		return (W==op2.W)&&(X==op2.X)&&(Y==op2.Y)&&(Z==op2.Z);
	};

	///	Inequality operator.
	bool operator!=(const AmbisonicData& op2)
	{
		return (W!=op2.W)||(X!=op2.X)||(Y!=op2.Y)||(Z!=op2.Z);
	};

	friend AmbisonicData dllStuff operator+(float op1, const AmbisonicData& op2);
	friend AmbisonicData dllStuff operator-(float op1, const AmbisonicData& op2);
	friend AmbisonicData dllStuff operator*(float op1, const AmbisonicData& op2);
	friend AmbisonicData dllStuff operator/(float op1, const AmbisonicData& op2);
};

#endif
