//	AttributeHelper.h - Classes used to convert single string values into the
//						correct multiple value attribute.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef ATTRIBUTEHELPER_H_
#define ATTRIBUTEHELPER_H_

#include "DllStuff.h"
#include "TwoFloats.h"
#include "ThreeFloats.h"
#include "FourFloats.h"
#include "Quaternion.h"

#include <string>
#include <vector>

//------------------------------------------------------------------------------
///	Base class for AttributeHelpers.
class dllStuff AttributeHelperBase
{
  public:
	///	Constructor.
	AttributeHelperBase() {};
	///	Destructor.
	virtual ~AttributeHelperBase() {};

	///	Returns true if val has a '.' in it (& is therefore valid for AttributeHelper).
	bool isValid(const std::string& val);

	///	Method to return the address minus any indices.
	static std::string getAddress(const std::string& address);
  protected:
	///	Returns the index at the end of a string.
	/*!
		\param address String to be parsed to find an index (will be in the
		format: "<string text>.<index>".

		\return The index, if found.  -1 otherwise ('otherwise' could mean that
		there was no index in the string, or that we couldn't parse the index
		we found).
	 */
	int getIndex(const std::string& address);
	///	Returns the secondary index in a string.
	/*!
		\param address The string to be parsed. (should be in the format:
		"<string text>.<secondary index>.<index>".

		\return The index, if found.  -1 otherwise.

		This method is specifically for attributes consisting of arrays of types
		like ThreeFloats, FourFloats etc. where there would be two indices if
		a single parameter is to be indexed.
	 */
	int getSecondaryIndex(const std::string& address);
};

//------------------------------------------------------------------------------
///	Template class for the (specialised) AttributeHelpers.
template <class T>
class dllStuff AttributeHelper : public AttributeHelperBase
{
  public:
	///	Constructor.
	/*!
		\param attr This is the address of the attribute whose value we want to
  		change.
	 */
	AttributeHelper(T *attr);
	///	Destructor.
	~AttributeHelper();

	///	Converts val into the correct value, writes to attribute.
	/*!
		\param address The OSC address used (we use this to obtain the index of
		the value to be changed.
		\param val The value to which we'll set the named attribute.
	 */
	void convert(const std::string& address, const std::string& val);
};

//------------------------------------------------------------------------------
///	Specialised version of AttributeHelper for TwoFloats.
template <>
class dllStuff AttributeHelper<TwoFloats> : public AttributeHelperBase
{
  public:
	///	Constructor.
	AttributeHelper(TwoFloats *attr);

	///	Converts val into the correct value, writes to attribute.
	void convert(const std::string& address, const std::string& val);
  private:
	///	The pointer to the attribute we're updating.
	TwoFloats *attribute;
};

//------------------------------------------------------------------------------
///	Specialised version of AttributeHelper for ThreeFloats.
template <>
class dllStuff AttributeHelper<ThreeFloats> : public AttributeHelperBase
{
  public:
	///	Constructor.
	AttributeHelper(ThreeFloats *attr);

	///	Converts val into the correct value, writes to attribute.
	void convert(const std::string& address, const std::string& val);
  private:
	///	The pointer to the attribute we're updating.
	ThreeFloats *attribute;
};

//------------------------------------------------------------------------------
///	Specialised version of AttributeHelper for FourFloats.
template <>
class dllStuff AttributeHelper<FourFloats> : public AttributeHelperBase
{
  public:
	///	Constructor.
	AttributeHelper(FourFloats *attr);

	///	Converts val into the correct value, writes to attribute.
	void convert(const std::string& address, const std::string& val);
  private:
	///	The pointer to the attribute we're updating.
	FourFloats *attribute;
};

//------------------------------------------------------------------------------
///	Specialised version of AttributeHelper for Quaternions.
template <>
class dllStuff AttributeHelper<Quaternion> : public AttributeHelperBase
{
  public:
	///	Constructor.
	AttributeHelper(Quaternion *attr);

	///	Converts val into the correct value, writes to attribute.
	void convert(const std::string& address, const std::string& val);
  private:
	///	The pointer to the attribute we're updating.
	Quaternion *attribute;
};

//------------------------------------------------------------------------------
///	Specialised version of AttributeHelper for vector<double>.
template <>
class dllStuff AttributeHelper< std::vector<double> > : public AttributeHelperBase
{
  public:
	///	Constructor.
	AttributeHelper(std::vector<double> *attr);

	///	Converts val into the correct value, writes to attribute.
	void convert(const std::string& address, const std::string& val);
  private:
	///	The pointer to the attribute we're updating.
	std::vector<double> *attribute;
};

//------------------------------------------------------------------------------
///	Specialised version of AttributeHelper for vector<float>.
template <>
class dllStuff AttributeHelper< std::vector<float> > : public AttributeHelperBase
{
  public:
	///	Constructor.
	AttributeHelper(std::vector<float> *attr);

	///	Converts val into the correct value, writes to attribute.
	void convert(const std::string& address, const std::string& val);
  private:
	///	The pointer to the attribute we're updating.
	std::vector<float> *attribute;
};

//------------------------------------------------------------------------------
///	Specialised version of AttributeHelper for vector<int>.
template <>
class dllStuff AttributeHelper< std::vector<int> > : public AttributeHelperBase
{
  public:
	///	Constructor.
	AttributeHelper(std::vector<int> *attr);

	///	Converts val into the correct value, writes to attribute.
	void convert(const std::string& address, const std::string& val);
  private:
	///	The pointer to the attribute we're updating.
	std::vector<int> *attribute;
};

//------------------------------------------------------------------------------
///	Specialised version of AttributeHelper for vector<ThreeFloats>.
template <>
class dllStuff AttributeHelper< std::vector<ThreeFloats> > : public AttributeHelperBase
{
  public:
	///	Constructor.
	AttributeHelper(std::vector<ThreeFloats> *attr);

	///	Converts val into the correct value, writes to attribute.
	void convert(const std::string& address, const std::string& val);
  private:
	///	The pointer to the attribute we're updating.
	std::vector<ThreeFloats> *attribute;
};

//------------------------------------------------------------------------------
///	Specialised version of AttributeHelper for vector<FourFloats>.
template <>
class dllStuff AttributeHelper< std::vector<FourFloats> > : public AttributeHelperBase
{
  public:
	///	Constructor.
	AttributeHelper(std::vector<FourFloats> *attr);

	///	Converts val into the correct value, writes to attribute.
	void convert(const std::string& address, const std::string& val);
  private:
	///	The pointer to the attribute we're updating.
	std::vector<FourFloats> *attribute;
};

//------------------------------------------------------------------------------
///	Specialised version of AttributeHelper for vector<TwoFloats>.
template <>
class dllStuff AttributeHelper< std::vector<TwoFloats> > : public AttributeHelperBase
{
  public:
	///	Constructor.
	AttributeHelper(std::vector<TwoFloats> *attr);

	///	Converts val into the correct value, writes to attribute.
	void convert(const std::string& address, const std::string& val);
  private:
	///	The pointer to the attribute we're updating.
	std::vector<TwoFloats> *attribute;
};

#endif
