//	BindableNode.h - Base class for all bindable nodes.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef BINDABLENODE_H_
#define BINDABLENODE_H_

#include "DllStuff.h"
#include "AbstractNode.h"

#include <string>
#include <map>
#include <vector>

namespace HeilanX3D
{

//------------------------------------------------------------------------------
///	Base class for all bindable nodes.
/*!
	Note: All subclasses <em>MUST</em> take into account the fact that Heilan is
	a multi-threaded program - nodes which are not accessed solely by the
	graphics thread must make sure all set/get variable operations are enclosed
	within critical sections.

	Note: All subclasses <em>MUST</em> call
	BindableNodeSingleton::registerNode() in initialise(), or there will
	be no way to access them.  They <em>MUST</em> also call
	BindableNodeSingleton::unregisterNode() in their destructors, or bad things
	will happen (like, monsters will jump out of your computer and chew on your
	toes).
 */
class dllStuff BindableNode : public ChildNode
{
  public:
	///	Constructor.
	BindableNode();
	///	Destructor.
	virtual ~BindableNode();
};

//------------------------------------------------------------------------------
typedef std::vector<BindableNode *> Nodes;
typedef std::map<std::string, BindableNode *> ActiveNodes;
typedef std::map<std::string, Nodes *> AllNodes;

///	Singleton class used to provide access to all bindable nodes.
/*!
	This keeps track of all the BindableNodes in the scene - it provides methods
	to access the currently active node of each class, as well as methods to
	set which node is currently active.
 */
class dllStuff BindableNodeSingleton
{
  public:
	///	Returns the sole instance of BindableNodeSingleton.
	static BindableNodeSingleton *getInstance();

	///	Registers a BindableNode with the singleton.
	/*!
		The first node registered for a particular type will be set as the
		currently active node for that type.
	 */
	void registerNode(BindableNode *node);
	///	Unregisters a BindableNode.
	/*!
		This is necessary because NodeConstructor has to construct a node if
		it wants to determine it's type, resulting in BindableNodes registering
		themselves with BindableNodeSingleton that aren't necessarily in the
		scene.
	 */
	void unregisterNode(BindableNode *node);

	///	Returns the currently active node of type 'type'.
	/*!
		Returns 0 if it doesn't exist.
	 */
	BindableNode *getActiveNode(const std::string& type) const;
	///	Selects the next node of type 'type' to become the currently active node.
	void nextNode(const std::string& type);
	///	Selects the previous node of type 'type' to become the currently active node.
	void previousNode(const std::string& type);
	///	Returns the number of nodes in the scene of type 'type'.
	int getNumNodes(const std::string& type) const;
	///	Returns the indexed node in the scene of type 'type'.
	/*!
		Returns 0 if the node does not exist.
	 */
	BindableNode *getIndexedNode(const std::string& type, int index) const;
	///	Selects the indexed node of type 'type' to become the currently active node.
	void setIndexedNode(const std::string& type, int index);

//Stupid fix for MSVC v6.
#ifdef WIN32
	///	Destructor.
	~BindableNodeSingleton();
#endif
  private:
	///	Constructor.
	BindableNodeSingleton();
#ifndef WIN32
	///	Destructor.
	~BindableNodeSingleton();
#endif

	///	A map containing the currently active nodes for each type.
	ActiveNodes activeNodes;
	///	A map containing all the BindableNodes in the scene.
	AllNodes allNodes;
};

}

#endif
