//	Camera.h - Singleton class representing the camera/listening point in the
//			   scene.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef CAMERA_H_
#define CAMERA_H_

//Stupid MSVC v6 warning about the STL...
#ifdef WIN32
#pragma warning(disable:4786)
#endif

#include "ThreeFloats.h"
//#include "FourFloats.h"
#include "CriticalSection.h"
#include "Quaternion.h"

#include <iostream>
#include <string>

namespace HeilanX3D
{

class Navigation;

//------------------------------------------------------------------------------
///	Singleton class representing the camera/listening point in the scene.
/*!
	Camera holds the current position of the camera/listening point, as well as
	being responsible for the current Navigation subclass being used.

	The position is only ever set from the graphics thread - at the start of
	every audio block, the current 'graphic' position is copied to a separate
	variable, to be used by the audio thread.  This is to avoid threading issues
	and minimise the number of critical sections needed.

	\todo Handle rotation correctly (using arcball?).
 */
class dllStuff Camera
{
  public:
	///	Returns the sole instance of Camera.
	static Camera *getInstance();

	/// To ensure we get our position etc. from the current Viewpoint on startup.
	void initialise();

	///	Returns the current Navigation class.
	Navigation *getNavigation() const {return currentNav;};
	///	Sets the available Navigation types.
	/*!
		\param text Space-separated list of available Navigation types for this
		scene.
	 */
	void setNavigationTypes(const std::string& text);
	///	Sets the currently active Navigation type.
	/*!
		\param text Name of the type (e.g. EXAMINE, ANY, etc.).

		If the requested type is not present in navTypes, no action is taken.
		The default navigation type (i.e. present on construction of Camera) is
		examine.
	 */
	void setNavigation(const std::string& text);

	///	Updates the current position from the graphics thread.
	void setGraphicPosition(const ThreeFloats& val, bool noOSC = false);
	///	Updates the current audio position to match the graphic position.
	/*!
		This MUST be called at the start of every audio block, to ensure the
		camera's position gets updated.
	 */
	void updateAudioPosition();
	///	Returns the current 'graphic' position.
	/*!
		Note that Camera stores both a graphic position and an audio position -
		the graphic position must only ever be read from the graphics thread,
		and the audio position only ever from the audio thread.

		\sa getAudioPosition()
	 */
	const ThreeFloats& getGraphicPosition() const {return gPosition;};
	///	Returns the current 'audio' position.
	/*!
		\sa getGraphicPosition()
	 */
	const ThreeFloats& getAudioPosition() const {return aPosition;};

	///	Updates the current rotation from the graphics thread.
	void setGraphicRotation(const Quaternion& val);
	///	Updates the current audio rotation to match the graphic position.
	/*!
		This MUST be called at the start of every audio block, to ensure the
		camera's rotation gets updated.
	 */
	void updateAudioRotation();
	///	Returns the current 'graphic' rotation.
	/*!
		Note that Camera stores both a graphic rotation and an audio rotation -
		the graphic rotation must only ever be read from the graphics thread,
		and the audio rotation only ever from the audio thread.

		\sa getAudioRotation()
	 */
	const Quaternion& getGraphicRotation() const {return gRotation;};
	///	Returns the current 'audio' rotation.
	/*!
		\sa getGraphicRotation()
	 */
	const Quaternion& getAudioRotation() const {return aRotation;};

	///	Called from Browser::paint() to draw/setup the camera.
	void draw();

	///	Sets whether we're rendering offline or not.
	void setOfflineRendering(bool val) {offline = val;};
	///	Returns whether we're rendering offline or not.
	bool getOfflineRendering() const {return offline;};
	///	Returns our current offline ticks (in ms).
	unsigned int getOfflineTicks() const {return offlineTicks;};
	///	Sets our current offline ticks (in ms).
	void setOfflineTicks(unsigned int val) {offlineTicks = val;};

	///	Sets the width of the window used to render the graphics.
	void setViewWidth(int val) {viewWidth = val;};
	///	Sets the height of the window used to render the graphics.
	void setViewHeight(int val) {viewHeight = val;};
	///	Returns the width of the window used to render the graphics.
	int getViewWidth() const {return viewWidth;};
	///	Returns the height of the window used to render the graphics.
	int getViewHeight() const {return viewHeight;};

	//Stupid fix for MSVC v6.
#ifdef WIN32
	///	Destructor.
	~Camera();
#endif
  private:
	///	Constructor.
	Camera();
#ifndef WIN32
	///	Destructor.
	~Camera();
#endif

	///	Graphic position.
	ThreeFloats gPosition;
	///	Audio position.
	ThreeFloats aPosition;

	///	Graphic rotation.
	Quaternion gRotation;
	///	Audio rotation.
	Quaternion aRotation;

	///	Space-separated list of available Navigation types for this scene.
	std::string navTypes;
	///	The current navigation class used to control how the camera moves.
	Navigation *currentNav;

	///	Critical section used to avoid threading issues with aPosition.
	CriticalSection posCritSec;
	///	Critical section used to avoid threading issues with aRotation.
	CriticalSection rotCritSec;

	///	Whether we're rendering offline or not.
	bool offline;
	///	Our current offline ticks value.
	unsigned int offlineTicks;

	///	The width of the viewport used to display the graphics.
	int viewWidth;
	///	The height of the viewport used to display the graphics.
	int viewHeight;
};

}

#endif
