//	Lighting.h - Singleton class to handle lighting.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef LIGHTING_H_
#define LIGHTING_H_

//Stupid MSVC v6 warning about the STL...
#ifdef WIN32
#pragma warning(disable:4786)
#endif

#include "ThreeFloats.h"

#include "GLHeaders.h"
#include <map>

namespace HeilanX3D
{

///	Singleton class to handle lighting.
/*!
	We need this because we just use OpenGL's lighting model, and it only
	provides a limited number of lights(8).  Also, we need to be able to
	specify that lights only apply in certain scopes, so we need a way to push/
	pop lights.
 */
class dllStuff Lighting
{
  public:
	///	Returns the sole Lighting instance.
	static Lighting& getInstance();

	///	Attempts to add a local directional light (X3D spec) to the scene.
	/*!
		/param direction The direction of the light.
		/param colour The colour of the light.
		/param ambient The ambient intensity of the light.

		This method calls the OpenGL functions needed for the light to be made
		part of the scene.

		This will only add a light to the scene if there is a light available.
		Note that most OpenGL implementations only allow for 8 lights in total.
	 */
	void addDirectionalLight(const ThreeFloats& direction,
							 const ThreeFloats& colour,
							 float ambient);

	///	Attempts to add a local point light (X3D spec) to the scene.
	/*!
		/param position The position of the light.
		/param attenuation The attenuation factor of the light.
		/param colour The colour of the light.
		/param ambient The ambient intensity of the light.

		This method calls the OpenGL functions needed for the light to be made
		part of the scene.

		This will only add a light to the scene if there is a light available.
		Note that most OpenGL implementations only allow for 8 lights in total.
	 */
	void addPointLight(const ThreeFloats& position,
					   const ThreeFloats& attenuation,
					   const ThreeFloats& colour,
					   float ambient);

	///	Attempts to add a local spot light (X3D spec) to the scene.
	/*!
		/param position The position of the light.
		/param attenuation The attenuation factor of the light.
		/param colour The colour of the light.
		/param ambient The ambient intensity of the light.
		/param beamWidth The beamWidth of the light (the middle brightness bit).
		/param cutOffAngle The cut off angle of the light (the extent it reaches
		to the sides).
		/param direction The direction of the light.

		This method calls the OpenGL functions needed for the light to be made
		part of the scene.

		This will only add a light to the scene if there is a light available.
		Note that most OpenGL implementations only allow for 8 lights in total.
	 */
	void addSpotLight(const ThreeFloats& position,
					  const ThreeFloats& attenuation,
					  const ThreeFloats& colour,
					  float ambient,
					  float beamWidth,
					  float cutOffAngle,
					  const ThreeFloats& direction);

	///	Attempts to add a global directional light (X3D spec) to the scene.
	/*!
		/param direction The direction of the light.
		/param colour The colour of the light.
		/param ambient The ambient intensity of the light.

		This method calls the OpenGL functions needed for the light to be made
		part of the scene.

		This will only add a light to the scene if there is a light available.
		Note that most OpenGL implementations only allow for 8 lights in total.
	 */
	void addGlobalDirectionalLight(const ThreeFloats& direction,
							 	   const ThreeFloats& colour,
							 	   float ambient);

	///	Attempts to add a global point light (X3D spec) to the scene.
	/*!
		/param position The position of the light.
		/param attenuation The attenuation factor of the light.
		/param colour The colour of the light.
		/param ambient The ambient intensity of the light.

		This method calls the OpenGL functions needed for the light to be made
		part of the scene.

		This will only add a light to the scene if there is a light available.
		Note that most OpenGL implementations only allow for 8 lights in total.
	 */
	void addGlobalPointLight(const ThreeFloats& position,
							 const ThreeFloats& attenuation,
							 const ThreeFloats& colour,
							 float ambient);

	///	Attempts to add a global spot light (X3D spec) to the scene.
	/*!
		/param position The position of the light.
		/param attenuation The attenuation factor of the light.
		/param colour The colour of the light.
		/param ambient The ambient intensity of the light.
		/param beamWidth The beamWidth of the light (the middle brightness bit).
		/param cutOffAngle The cut off angle of the light (the extent it reaches
		to the sides).
		/param direction The direction of the light.

		This method calls the OpenGL functions needed for the light to be made
		part of the scene.

		This will only add a light to the scene if there is a light available.
		Note that most OpenGL implementations only allow for 8 lights in total.
	 */
	void addGlobalSpotLight(const ThreeFloats& position,
						    const ThreeFloats& attenuation,
						    const ThreeFloats& colour,
						    float ambient,
						    float beamWidth,
						    float cutOffAngle,
						    const ThreeFloats& direction);

	///	Saves the current lighting setup.
	/*!
		Called by ParentNode before any child nodes are drawn.
	 */
	void pushLighting();
	///	Disables any lighting done since the last pushLighting().
	/*!
		Called by ParentNode after any child nodes are drawn.
	 */
	void popLighting();

	///	Turn the headlight on and off.
	/*!
		\param val True == on, false == off.
	 */
	void switchHeadlight(bool val);
	///	Called from Browser::paint() to draw the headlight (or not).
	void drawHeadlight();

//Stupid fix for MSVC v6.
#ifdef WIN32
	///	Destructor.
	~Lighting();
#endif
  private:
	///	Constructor.
	Lighting();
#ifndef WIN32
	///	Destructor.
	~Lighting();
#endif

	///	Simple struct representing a light.
	struct Light
	{
		///	The light's position.
		ThreeFloats position;
		///	The light's colour.
		ThreeFloats colour;
		///	The light's OpenGL identifier.
		GLenum id;
		///	Whether the light's global or not.
		bool global;
	};

	typedef std::multimap<int, Light *> LightMap;

	///	Our currently active lights.
	LightMap activeLights;
	///	Our current 'state' (i.e. number of pushes we've had).
	int currentState;

	///	Array of the possibly active lights.
	/*!
		We assume that there are only 7 lights available (8 in total - we use
		light 0 as a (global) headlight).
	 */
	Light lights[7];
	///	Index of the next light which can be made active.
	/*!
		7 == no lights left.
	 */
	int nextActive;

	///	True if the headlight should be drawn.
	bool headlight;
};

}

#endif
