//	MiscHelper.h - Various miscellaneous helper bits and pieces.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef MISCHELPER_H_
#define MISCHELPER_H_

#include "DllStuff.h"

#include <ctime>

//------------------------------------------------------------------------------
///	Rounding function, since standard c++ doesn't seem to have one.
inline int heilanRound(float val)
{
	return (int)(val + 0.5f);
};

//------------------------------------------------------------------------------
///	Simple seed generator for heilanRand, based on time().
inline int heilanRandSeed()
{
	static int val = time(0);

	return ++val;
};

//------------------------------------------------------------------------------
///	Simple random number generator.
/*!
	\return Random value between -1 and +1.
 */
inline float heilanRand(int seed = 0)
{
	seed = (seed<<13) ^ seed;
   	return (1.0f - ((seed*(seed*seed*15731 + 789221)+1376312589)&0x7fffffff)/1073741824.0f);
};

//------------------------------------------------------------------------------
///	Helper function to calculate a normal given 3 vertices.
inline ThreeFloats heilanGetNormal(const ThreeFloats& v1,
								   const ThreeFloats& v2,
								   const ThreeFloats& v3)
{
	float tempf;
	ThreeFloats a, b;
	ThreeFloats retval;
	
	//Calculate a & b.
    a.x = v1.x - v2.x;
    a.y = v1.y - v2.y;
    a.z = v1.z - v2.z;

    b.x = v2.x - v3.x;
    b.y = v2.y - v3.y;
    b.z = v2.z - v3.z;

    //Calculate cross product.
    retval.x = (a.y * b.z)-(a.z * b.y);
    retval.y = (a.z * b.x)-(a.x * b.z);
    retval.z = (a.x * b.y)-(a.y * b.x);

    //Normalise.
	tempf = retval.getDistance(ThreeFloats(0.0f));
	if(tempf != 0.0f)
		retval /= tempf;

	return retval;
};

//------------------------------------------------------------------------------
///	Cheap fake log function.
/*!
	Meant for envelope generators etc.  Don't expect it to provide an actual
	logarithm.
 */
inline float heilanLog(float val)
{
	return val/(val + (-0.111111111f*(val-1.0f)));
};

//------------------------------------------------------------------------------
///	Cheap fake exp function.
/*!
	Meant for envelope generators etc.  Don't expect it to provide an actual
	exponential.
 */
inline float heilanExp(float val)
{
	return val/(val + (-9.0f*(val-1.0f)));
};

//------------------------------------------------------------------------------
///	Cheap tanh function.
//From http://www.musicdsp.org/archive.php?classid=5#238
inline float heilanTanh(float val)
{
	if(val < -3.0f)
        return -1.0f;
    else if(val > 3.0f)
        return 1.0f;
    else
        return val * (27.0f + val * val)/(27.0f + 9.0f * val * val);
};

#endif
