//	ParentNode.h - Base class for all nodes which hold child nodes.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PARENTNODE_H_
#define PARENTNODE_H_

#include "AbstractNode.h"
#include "ConditionalWait.h"
#include <vector>

namespace HeilanX3D
{

typedef std::vector<AbstractNode *> NodeArray;
class AudioThread;

///	Base class for all nodes which can hold child nodes.
class dllStuff ParentNode : public AbstractNode
{
  public:
	///	Constructor.
	ParentNode();
	///	Destructor.
	virtual ~ParentNode();

	///	Called to initialise things.
	/*!
		Note: subclasses of ParentNode should <em>NOT</em> override this - 
		use the parentInitialise() method instead.
	 */
	void initialise();

	///	Called to draw the node (& it's children) in the scene.
	/*!
		This calls preChildDraw(), then graphicDoStuff() for all it's children,
		then postChildDraw().  Don't override this unless you've got a
		<em>very</em> good reason!  (e.g. MultiTexture)  Use pre- and
		postChildDraw() instead.
	 */
	virtual void graphicDoStuff();
	///	Called to get the audio output for this node and all it's children.
	/*!
		\param numSamples The number of samples of audio to generate.
		\param pos The position of the node in the 3d scene.  This is necessary
		because most nodes will be moved in the scene by Transform ParentNodes,
		and therefore won't have direct access to their current position.  This
		variable lets the position be propagated through the scene (and it will
		work when the audio gets multi-threaded).

		This calls preChildAudio(), the audioDoStuff() for all it's children,
		then postChildAudio(), so use those methods, don't override this one.
		The audio from the children (and this node, if necessary) is summed
		before the method returns.
	 */
	//AmbisonicData *audioDoStuff(int numSamples, const ThreeFloats& pos);
	///	Called by any ParentNodes (or the Browser) to get this node's encoded audio.
	/*!
		\param numSamples The number of samples to calculate for this audio
		block.
		\param pos The current 3d position of the node (used for the Ambisonic
		encoding.
		\return 0 if the subclass does not produce audio.

		This is overridden so that we can sum our children's audio.  The process
		is: sum all the childrens' audio, then call getMonoAudio() on ourself,
		sum it to the childrens' audio, and return.

		\sa AbstractNode::getMonoAudio, AmbisonicData
	 */
	virtual AmbisonicData *getAmbisonicAudio(int numSamples,
											 const ThreeFloats& pos);

	///	Called prior to calling audioDoStuff(), for nodes which need to reset things.
	/*!
		Don't try to override this in subclasses - ParentNode needs to pass on
		the message to it's children nodes - use parentAudioBlockStart()
		instead.
	 */
	void audioBlockStart();
	///	Called to set the current audio block size.
	/*!
		This will call setBlockSizeCallback() so that subclasses may be informed
		when the block size changes.  We do it this way because ParentNode needs
		it's own block when it's summing audio from it's children, and we don't
		want to risk people accidentally overriding setBlockSize(), and breaking
		that (which is why setBlockSize() isn't virtual here).
	 */
	void setBlockSize(int newSize);
	///	Called to set the current audio samplerate.
	void setSamplerate(float val);

	///	Call this to add a child node to this node's list of children.
	/*
		\param newChild The newly-constructed child node to add.
		\param addToStart Whether the node should be added to the start of our
		vector of children (true), or the end (false, the default).

		The ParentNode will take care of deleting all it's child nodes, so you
		don't have to worry about that.
	 */
	void addNode(AbstractNode *newChild, bool addToStart = false);
	///	Returns the number of child nodes this ParentNode has.
	int getNumChildren() const {return children.size();};
	///	Removes all the children of this node, and deletes them.
	/*!
		\param dontDelete If true, don't delete the children, otherwise do.
	 */
	void clear(bool dontDelete=false);

	///	When it's in parallel mode, this tells us a child node's audio is ready.
	void parallelAudioReady(AmbisonicData *buffer);

	///	Overridden to support per-node parallelism.
	void setMetadata(const std::string& name, const std::string& value);

	///	Called to inform SceneConstructor which ParentNode is currently on top.
	void sceneConstructionEnter(SceneConstructor *constructor);
	///	Called to remove this node from the SceneConstructor stack.
	void sceneConstructionLeave(SceneConstructor *constructor);

	///	Method used to search for a particular DEF-named node in a scene.
	AbstractNode *findNode(const std::string& name);
	///	Provides access to the indexed child node.
	/*!
		Returns 0 if the index is invalid.
	 */
	AbstractNode *getChild(int index) const;

	///	Used to make sure unROUTEd nodes get their output queues emptied periodically.
	/*!
		\sa AbstractNode::clearEvents
	 */
	void clearEvents();
  protected:
	///	Called immediately after the scene has been constructed.
	/*!
		This gets called before the node's childrens' initialise() methods
		are called.
	 */
	  virtual void parentInitialise() {};

	///	Called before the node starts drawing it's children.
	/*!
		Draws nothing by default.
	 */
	virtual void preChildDraw();
	///	Called after the node has finished drawing it's children.
	/*!
		Draws nothing by default.
	 */
	virtual void postChildDraw();
	/*virtual AmbisonicData *preChildAudio(int numSamples,
										 const ThreeFloats& newPos);*/
	/*virtual AmbisonicData *postChildAudio(int numSamples,
										  const ThreeFloats& newPos);*/
	///	Called immediately prior to audioDoStuff().
	virtual void parentAudioBlockStart() {};
	///	Helper method to encode a block of mono AmbisonicData audio to Ambisonic B-format.
	/*!
		Could this not be replaced by simply calling
		AbstractNode::getAmbisonicAudio()?  It's doing the same thing anyway,
		right?
	 */
	AmbisonicData *encodeBlock(int numSamples,
							   const ThreeFloats& pos,
							   AmbisonicData *block);

	///	Called when the audio block size changes via setBlockSize().
	/*!
		This is pure virtual to remind subclasses to implement their own block
		allocation when generating audio.
	 */
	virtual void audioBlockSizeChanged(int newSize) = 0;
	///	Called when the audio samplerate changes via setSamplerate().
	virtual void samplerateChanged(float newRate) {};

	///	Called by subclasses to update the node's position.
	/*!
		The node's position is passed on to all it's children in audioDoStuff(),
		because Transform nodes would change a child node's position (requiring
		the audio to be calculated from that position), yet the child itself
		wouldn't have direct access to that position data.  Subclasses such as
		Transform therefore call setPosition() to update their position, to be
		passed on to child nodes in audioDoStuff().
	 */
	void setPosition(const ThreeFloats& newPos);

	///	Called in sceneConstructionLeave() so ParentNodes can interrogate their children.
	virtual void childrenLoaded() {};

	///	Called when setMetadata() is called, providing the same function.
	/*!
		\sa AbstractNode::setMetadata()
	 */
	virtual void setParentMetadata(const std::string& name,
								   const std::string& value) {};
  private:
	///	vector of all this node's children.
	NodeArray children;

	///	Temporary buffer used when summing children's audio outputs.
	AmbisonicData *summingBuffer;
	///	The current audio block size.
	int blockSize;
	///	Current position within the scene.
	/*!
		Used by audioDoStuff() to pass the current position onto our child
		nodes.

		\sa setPosition()
	 */
	ThreeFloats position;

	///	True if the audio from child nodes should be parallelised.
	bool paralleliseChildren;
	///	Used to determine when we've collected audio from all our children in parallel mode.
	ConditionalWait waiter;
	///	How many child nodes we've received audio from in parallel mode.
	unsigned int pNumChildren;
	///	Current number of samples to process in parallel mode.
	int pNumSamples;
	///	Vector of the AudioThreads assigned to child nodes.
	std::vector<AudioThread *> audioThreads;

	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;
};

}

#endif
