//	Quaternion.h - A struct to represent a Quaternion.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef QUATERNION_H_
#define QUATERNION_H_

#include "ThreeFloats.h"
#include "DllStuff.h"

#include <cmath>

///	A struct to represent a Quaternion.
/*!
	Most assignment type operations will automatically normalise the Quaternion.
 */
struct dllStuff  Quaternion
{
	double w;
	double x;
	double y;
	double z;

	///	Default constructor.
	Quaternion():
	w(1.0f),
	x(0.0f),
	y(0.0f),
	z(0.0f)
	{
		
	};

	///	Constructor.
	Quaternion(double wVal, double xVal, double yVal, double zVal):
	w(wVal),
	x(xVal),
	y(yVal),
	z(zVal)
	{
		if(*this != 0.0)
			normalise();
	};

	///	Copy constructor.
	Quaternion(const Quaternion& other)
	{
		w = other.w;
		x = other.x;
		y = other.y;
		z = other.z;
	};

	///	Assignment operator.
	Quaternion& operator=(const Quaternion& other)
	{
		w = other.w;
		x = other.x;
		y = other.y;
		z = other.z;
		if(*this != 0.0)
			normalise();

		return *this;
	};

	///	Multiplication operator (Quaternion).
	Quaternion operator*(const Quaternion& other) const
	{
		return Quaternion((w*other.w) - (x*other.x) - (y*other.y) - (z*other.z),
						  (w*other.x) + (x*other.w) + (y*other.z) - (z*other.y),
						  (w*other.y) + (y*other.w) + (z*other.x) - (x*other.z),
						  (w*other.z) + (z*other.w) + (x*other.y) - (y*other.x));

	};

	///	Multiplication operator (ThreeFloats).
	ThreeFloats operator*(const ThreeFloats& other) const
	{
		ThreeFloats retval(other);
		Quaternion tempQuat;
		const ThreeFloats origin(0.0f);
		float distance = other.getDistance(origin);

		tempQuat.w = 0.0f;
		tempQuat.x = other.x;
		tempQuat.y = other.y;
		tempQuat.z = other.z;
		if(retval != 0.0f)
			tempQuat.normalise();

		//tempQuat *= conjugate();
		//tempQuat *= *this;
		tempQuat = (*this) * tempQuat * conjugate();

		retval.x = (float)tempQuat.x;
		retval.y = (float)tempQuat.y;
		retval.z = (float)tempQuat.z;

		//'Un-normalise' the vector.
		retval *= distance;

		return retval;
	};

	///	Multiplication assignment operator (Quaternion).
	Quaternion operator*=(const Quaternion& other)
	{
		Quaternion tempQuat((w*other.w) - (x*other.x) - (y*other.y) - (z*other.z),
						    (w*other.x) + (x*other.w) + (y*other.z) - (z*other.y),
						    (w*other.y) + (y*other.w) + (z*other.x) - (x*other.z),
						    (w*other.z) + (z*other.w) + (x*other.y) - (y*other.x));

		*this = tempQuat;

		return *this;
	};

	///	Inequality operator (Quaternion).
	bool operator!=(const Quaternion& other) const
	{
		return (w!=other.w)||(x!=other.x)||(y!=other.y)||(z!=other.z);
	};

	///	Inequality operator (float).
	bool operator!=(float other)
	{
		return (w!=other)||(x!=other)||(y!=other)||(z!=other);
	};

	///	Normalises the Quaternion.
	void normalise()
	{
		double tempd;

		tempd = (w*w) + (x*x) + (y*y) + (z*z);
		if(fabs(tempd - 1.0) > 0.00001)
		{
			tempd = 1.0/sqrt(tempd);
			w *= tempd;
			x *= tempd;
			y *= tempd;
			z *= tempd;
		}
	};

	///	Returns the conjugate of the Quaternion.
	Quaternion conjugate() const
	{
		return Quaternion(w, -x, -y, -z);
	};

	///	Sets the Quaternion from axis/angle values.
	void setFromAxisAngle(double angle, ThreeFloats &axis)
	{
		double sinAngle;
		ThreeFloats tempVector(axis);

		if(axis == 0.0f)
		{
			w = 1.0;
			x = 0.0;
			y = 0.0;
			z = 0.0;
		}
		else
		{
			tempVector.normalise();
			angle *= 0.5f;
			sinAngle = sin(angle);

			tempVector *= (float)sinAngle;
			x = tempVector.x;
			y = tempVector.y;
			z = tempVector.z;
			w = cos(angle);
		}
	};

	///	Sets the Quaternion from Euler angles.
	/*!
		This method will suffer from Gimbal Lock.
	 */
	void setFromEulerAngles(double pitch, double yaw, double roll)
	{
		//Create 3 Quaternions, one for each angle, then multiply them
		//together.

		//Magic value here = (pi/180)/2.
		double r = pitch * 0.0087266462599716478846184538424431;
		double p = yaw * 0.0087266462599716478846184538424431;
		double y = roll * 0.0087266462599716478846184538424431;

		double sinP = sin(p);
		double sinY = sin(y);
		double sinR = sin(r);
		double cosP = cos(p);
		double cosY = cos(y);
		double cosR = cos(r);

		w = (cosR*cosP*cosY) + (sinR*sinP*sinY);
		x = (sinR*cosP*cosY) - (cosR*sinP*sinY);
		this->y = (cosR*sinP*cosY) + (sinR*cosP*sinY);
		z = (cosR*cosP*sinY) - (sinR*sinP*cosY);
		normalise();
	};

	///	Returns a matrix which can be fed to OpenGL.
	double *getMatrix()
	{
		double x2 = x*x;
		double y2 = y*y;
		double z2 = z*z;
		double xy = x*y;
		double xz = x*z;
		double yz = y*z;
		double wx = w*x;
		double wy = w*y;
		double wz = w*z;

		/*matrix[0] = 1.0 - (2.0*(y2+z2));
		matrix[1] = 2.0 * (xy+wz);
		matrix[2] = 2.0 * (xz-wy);
		matrix[3] = 0.0;

		matrix[4] = 2.0 * (xy-wz);
		matrix[5] = 1.0 - (2.0*(x2+z2));
		matrix[6] = 2.0 * (yz+wx);
		matrix[7] = 0.0;

		matrix[8] = 2.0 * (xz+wy);
		matrix[9] = 2.0 * (yz-wx);
		matrix[10] = 1.0 - (2.0*(x2+y2));
		matrix[11] = 0.0;

		matrix[12] = 0.0;
		matrix[13] = 0.0;
		matrix[14] = 0.0;
		matrix[15] = 1.0;*/

		matrix[0] = 1.0 - (2.0*(y2+z2));
		matrix[1] = 2.0 * (xy-wz);
		matrix[2] = 2.0 * (xz+wy);
		matrix[3] = 0.0;

		matrix[4] = 2.0 * (xy+wz);
		matrix[5] = 1.0 - (2.0*(x2+z2));
		matrix[6] = 2.0 * (yz-wx);
		matrix[7] = 0.0;

		matrix[8] = 2.0 * (xz-wy);
		matrix[9] = 2.0 * (yz+wx);
		matrix[10] = 1.0 - (2.0*(x2+y2));
		matrix[11] = 0.0;

		matrix[12] = 0.0;
		matrix[13] = 0.0;
		matrix[14] = 0.0;
		matrix[15] = 1.0;

		return matrix;
	};

	///	Returns the Quaternion's axis/angle values.
	void getAxisAngle(double& angle, ThreeFloats& axis)
	{
		double scale = sqrt((x*x) + (y*y) + (z*z));

		if(scale == 0.0f)
		{
			axis.x = 0.0f;
			axis.y = 0.0f;
			axis.z = 0.0f;
		}
		else
		{
			scale = 1.0/scale;
			axis.x = (float)(x*scale);
			axis.y = (float)(y*scale);
			axis.z = (float)(z*scale);
		}
		angle = acos(w) * 2.0;
	};
  private:
	///	Used for the getMatrix() method's return value.
	double matrix[16];
};

#endif
