//	SoundFileLoader.h - Base class for sound file loaders.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SOUNDFILELOADER_H_
#define SOUNDFILELOADER_H_

//Stupid MSVC v6 warning about the STL...
#ifdef WIN32
#pragma warning(disable:4786)
#endif

#include "DllStuff.h"

#include <string>
#include <map>

namespace HeilanX3D
{

//------------------------------------------------------------------------------
///	Base class for sound file loaders.
/*!
	We have a base class because we want to be able to load ogg files as well as
	the various libsndfile formats, so we need to have a standard interface for
	accessing both libsndfile, and vorbisfile functions.

	The class is designed to work in the same way as c++' iostream classes -
	i.e. to read 64 samples from the file, do:
	\code
	float buffer[64];
	SoundFileLoader phil("filepath.wav");

	phil.width(64);
	phil >> buffer;
	\endcode

	Note that the file loader will only ever load a mono stream - for stereo
	files etc. it will just return the 1st channel.  Also, for simplicity, the
	loader will loop back to the beginning when it reaches the end of the file,
	ensuring that you always get the number of samples you requested with
	width().

	\todo Add samplerate conversion.
 */
class dllStuff SoundFileLoader
{
  public:
	///	Constructor.
	/*!
		\param filePath Path to the sound file to be loaded.

		The file should be opened here, and closed in the destructor.
	 */
	SoundFileLoader(const std::string& filePath);
	///	Destructor.
	/*!
		The file should be closed here.
	 */
	virtual ~SoundFileLoader();

	///	Sets the number of samples to read from the file with operator>>.
	/*!
		\param newWidth If this is -1, the method will just return the current
		width.  Otherwise it will set the current width to newWidth, and return
		that width.
	 */
	int width(int newWidth = -1);

	///	Returns the current read position within the file.
	virtual int tellg() const = 0;
	///	Sets the current read position within the file.
	virtual void seekg(int position) = 0;

	///	Operator used to read from the file.
	/*!
		\param buffer Assumed to be a pre-allocated buffer of floats, of at
		least size currentWidth.
	 */
	virtual SoundFileLoader& operator>>(float *buffer) = 0;

	///	Returns the path to this file.
	std::string getPath() const {return path;};
	///	Returns the number of samples in the 1st channel of the file.
	virtual int getFileSize() const = 0;
  protected:
	///	Called whenever width() is called, indicating a change in width.
	virtual void widthChanged() {};
  private:
	///	The path to the current file.
	std::string path;

	///	The current width.
	int currentWidth;
};

//------------------------------------------------------------------------------
///	Base class for SFLoaderConstructor.
/*!
	\sa SFLoaderConstructor
 */
class dllStuff SFLoaderConstructorBase
{
  public:
	///	Constructor (does nothing).
	SFLoaderConstructorBase() {};
	///	Destructor (does nothing).
	virtual ~SFLoaderConstructorBase() {};

	///	Returns a new instance of an AbstractNode subclass.
	virtual SoundFileLoader *createInstance(const std::string& path) const = 0;
};

//------------------------------------------------------------------------------
///	A simple template class used to construct SoundFileLoaders from a factory.
/*!
	All subclasses of SoundFileLoader should have one of these.  You would do
	something like the following, and then register thingConstructor with
	SFLoaderFactory...
	\code
	class thing : public SoundFileLoader
	{
	  public:
		...
		static SFLoaderConstructor<thing> thingConstructor;
	};
	\endcode
 */
template <class T>
class dllStuff SFLoaderConstructor : public SFLoaderConstructorBase
{
  public:
	///	Constructor (does nothing).
	SFLoaderConstructor() {};
	///	Destructor (does nothing).
	~SFLoaderConstructor() {};

	///	Returns a new instance of an AbstractNode subclass.
	SoundFileLoader *createInstance(const std::string& path) const {return new T(path);};
};

//------------------------------------------------------------------------------
///	Factory which returns the appropriate SoundFileLoader for a given file.
/*!
	To use it, you'd do something like:
	\code
	float buffer[1024];
	//From "soundfile.wav", the factory searches for the first loader it has
	//that can handle wav files, constructs an instance and returns it.
	auto_ptr<SoundFileLoader *> file(SFLoaderFactory::getInstance().getFileLoader("soundfile.wav"));

	file->width(1024);
	(*file) >> buffer;
	\endcode
	Note that you are responsible for deleting the loader once you've got it
	(hence the auto_ptr in the example above).
 */
class dllStuff SFLoaderFactory
{
  public:
	///	Returns the sole instance of this factory.
	static SFLoaderFactory& getInstance();

	///	Registers a new SFLoaderConstructorBase with the factory.
	/*!
		\param constructor Pointer to the SoundFileLoader's constructor.
		\param fileTypes Space separated list of all the file types this
		SoundFileLoader can handle.
	 */
	void registerSFLoader(SFLoaderConstructorBase *constructor,
						  const std::string& fileTypes);

	///	Returns a pointer to a newly-constructed SoundFileLoader.
	/*!
		\param path Path to the file to be loaded.

		Returns 0 if no file loader could be found which matches the file type
		of path.
	 */
	SoundFileLoader *getFileLoader(const std::string& path) const;

	///	Used to convert a string to lowercase.
	/*!
		Not entirely sure why I just can't pass in tolower to transform()...
	 */
	static char toLower(char in);

#ifdef WIN32
	///	Stupid fix for MSVC v6.
	~SFLoaderFactory();
#endif
  private:
	///	Constructor.
	SFLoaderFactory();
#ifndef WIN32
	///	Destructor.
	~SFLoaderFactory();
#endif

	///	All the SoundFileLoaders the factory knows about.
	/*!
		When registerSFLoader() is called, it separates all the file types it
		was passed, and adds a separate entry for each one.  This is to make it
		simpler to search for the correct file loader - you just have to get the
		extension of the file, then do loaders.find(extension) to get it.
	 */
	std::map<std::string, SFLoaderConstructorBase *> loaders;
};

}

#endif
