//	StringParser.h - Simple parser class to read from a string and output the
//					 digits from it.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef STRINGPARSER_H_
#define STRINGPARSER_H_

#include "DllStuff.h"
#include "TwoFloats.h"
#include "ThreeFloats.h"
#include "FourFloats.h"
#include "Quaternion.h"

#include <string>
#include <vector>

///	Simple parser class to read from a string and output the digits from it.
class dllStuff StringParser
{
  public:
	///	Constructor.
	/*!
		\param input The string to parse.
	 */
	StringParser(const std::string& input);
	///	Destructor.
	~StringParser();

	///	Returns the number of numbers in the string.
	int getNumNumbers() const {return numbers.size();};
	///	Returns the number of bools in the string.
	int getNumBools() const {return bools.size();};
	///	Returns the number of TwoFloats structures in the string.
	/*!
		This first checks if the number of numbers in the string is a multiple
		of 2.  If not, it will return 0, even if there are numbers in the
		string.
	 */
	int getNumTwoFloats() const;
	///	Returns the number of ThreeFloats structures in the string.
	/*!
		This first checks if the number of numbers in the string is a multiple
		of 3.  If not, it will return 0, even if there are numbers in the
		string.
	 */
	int getNumThreeFloats() const;
	///	Returns the number of FourFloats structures in the string.
	/*!
		This first checks if the number of numbers in the string is a multiple
		of 4.  If not, it will return 0, even if there are numbers in the
		string.
	 */
	int getNumFourFloats() const;
	///	Returns the number of Quaternion structures in the string.
	/*!
		This first checks if the number of numbers in the string is a multiple
		of 4.  If not, it will return 0, even if there are numbers in the
		string.
	 */
	int getNumQuaternions() const;

	///	Returns the indexed number as a double.
	/*!
		\param index The index of the required number.
		\return Returns 0.0 if the index is outwith the numbers vector's limits.
	 */
	double getDouble(int index = 0) const;
	///	Returns the indexed number as a float.
	/*!
		\param index The index of the required number.
		\return Returns 0.0f if the index is outwith the numbers vector's
		limits.

		Note: all numbers are stored as doubles, so there *may* be some
		information loss with this method.
	 */
	float getFloat(int index = 0) const;
	///	Returns the indexed number as an int.
	/*!
		\param index The index of the required number.
		\return Returns 0 if the index is outwith the numbers vector's limits.

		Note: all numbers are stored as doubles, so there may be some
		information loss with this method.
	 */
	int getInt(int index = 0) const;
	///	Returns the indexed numbers as a TwoFloats structure.
	/*!
		\param index The index of the required TwoFloats.  This method assumes
		that numbers contains a multiple of 2 number of numbers.  index is
		the index to the TwoFloats, not the index to the individual numbers
		stored (i.e. if you passed in an index of 1, that would be calculated
		internally as 2 - pointing to the third number stored).
		\return Returns 0 if the index is outwith the numbers vector's limits.

		Note: all numbers are stored as doubles, so there may be some
		information loss with this method.
	 */
	TwoFloats getTwoFloats(int index = 0) const;
	///	Returns the indexed numbers as a ThreeFloats structure.
	/*!
		\param index The index of the required ThreeFloats.  This method assumes
		that numbers contains a multiple of 3 number of numbers.  index is
		the index to the ThreeFloats, not the index to the individual numbers
		stored (i.e. if you passed in an index of 1, that would be calculated
		internally as 3 - pointing to the fourth number stored).
		\return Returns 0 if the index is outwith the numbers vector's limits.

		Note: all numbers are stored as doubles, so there may be some
		information loss with this method.
	 */
	ThreeFloats getThreeFloats(int index = 0) const;
	///	Returns the indexed numbers as a FourFloats structure.
	/*!
		\param index The index of the required FourFloats.  This method assumes
		that numbers contains a multiple of 4 number of numbers.  index is
		the index to the ThreeFloats, not the index to the individual numbers
		stored (i.e. if you passed in an index of 1, that would be calculated
		internally as 4 - pointing to the fifth number stored).
		\return Returns 0 if the index is outwith the numbers vector's limits.

		Note: all numbers are stored as doubles, so there may be some
		information loss with this method.
	 */
	FourFloats getFourFloats(int index = 0) const;
	///	Returns the indexed numbers as a Quaternion structure.
	/*!
		\param index The index of the required Quaternion.  This method assumes
		that numbers contains a multiple of 4 number of numbers.  index is
		the index to the Quaternion, not the index to the individual numbers
		stored (i.e. if you passed in an index of 1, that would be calculated
		internally as 4 - pointing to the fifth number stored).
		\return Returns 0 if the index is outwith the numbers vector's limits.

		Note: all numbers are stored as doubles, so there may be some
		information loss with this method.
	 */
	Quaternion getQuaternion(int index = 0) const;
	///	Returns the indexed bool.
	bool getBool(int index = 0) const;
  private:
	///	Helper method to fill in numbers.
	void fillNumbers(const std::string& input);
	///	Helper method to fill in bools.
	void fillBools(const std::string& input);
	///	Helper method to detect a hex number and return it.
	/*!
		\return Returns -1 if it's not a hex number.
	 */
	int getHex(const std::string& val);

	///	vector containing all the numbers read from the input string.
	std::vector<double> numbers;
	///	vector containing all the bools read from the input string.
	std::vector<bool> bools;
};

#endif
