//	ThreeFloats.h - Simple struct containing three floats.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef THREEFLOATS_H_
#define THREEFLOATS_H_

#include "DllStuff.h"

#include <cmath>

//------------------------------------------------------------------------------
///	Cheap sqrt function.
#ifdef __BIG_ENDIAN__
#include <ppc_intrinsics.h> //Possibly only on OSX?

//From http://www.idevgames.com/forum/showthread.php?t=2596
inline float heilanSqrt(register float x)
{
	register float res;

	if(x == 0.0f)
		return 0.0f;

	res = __frsqrtes(x);
	res += 0.5f * res * (1.0f - x * res * res);
	res += 0.5f * res * (1.0f - x * res * res);

	return res * x;
};
#elif defined(__GNUC__) //I assume this is x86-only?
//From http://betterexplained.com/articles/understanding-quakes-fast-inverse-square-root/
inline float heilanSqrt(float x)
{
	/*Doesn't work.  Bloody John Carmack, thinks he's so great...
	
	float retval;
	float xhalf = 0.5f*x;
    int i = *(int*)(void *)&x;	//The (void *) stops GCC warning about
								//dereferencing type-punned pointers.
    i = 0x5f3759df - (i>>1);
    retval = *(float*)(void *)&i;
    retval = retval*(1.5f - xhalf*retval*retval);

    return 1.0f/retval;*/
	return sqrtf(x);
};
#else //Assumes x86 and MSVC...
//From http://www.musicdsp.org/showArchiveComment.php?ArchiveID=206
// ~6 clocks on Pentium M vs. ~24 for single precision sqrtf
inline float heilanSqrt(float x)
{
	float z;

	_asm
	{
		rsqrtss xmm0, x
		rcpss xmm0, xmm0
		movss z, xmm0 // z ~= sqrt(x) to 0.038%
	}

	return z;
};
#endif

//------------------------------------------------------------------------------
///	Simple struct containing three floats.
struct dllStuff ThreeFloats
{
	float x;
	float y;
	float z;

	///	Default constructor.
	ThreeFloats()
	{
		x = y = z = 0.0f;
	};

	///	Constructor, all members set to val.
	ThreeFloats(float val)
	{
		x = y = z = val;
	};

	///	Constructor.
	ThreeFloats(float xVal, float yVal, float zVal)
	{
		x = xVal;
		y = yVal;
		z = zVal;
	};

	///	Assignment operator.
	ThreeFloats& operator=(const ThreeFloats& op2)
	{
		x = op2.x;
		y = op2.y;
		z = op2.z;
		return *this;
	};

	///	Assignment operator (assigns all members the value of val).
	ThreeFloats& operator=(float val)
	{
		x = val;
		y = val;
		z = val;
		return *this;
	};

	///	Addition operator.
	ThreeFloats operator+(const ThreeFloats& op2) const
	{
		ThreeFloats retval;

		retval.x = x + op2.x;
		retval.y = y + op2.y;
		retval.z = z + op2.z;

		return retval;
	};

	///	Subtraction operator.
	ThreeFloats operator-(const ThreeFloats& op2) const
	{
		ThreeFloats retval;

		retval.x = x - op2.x;
		retval.y = y - op2.y;
		retval.z = z - op2.z;

		return retval;
	};

	///	Multiplication operator.
	ThreeFloats operator*(const ThreeFloats& op2) const
	{
		ThreeFloats retval;

		retval.x = x * op2.x;
		retval.y = y * op2.y;
		retval.z = z * op2.z;

		return retval;
	};

	///	Multiplication operator (float).
	ThreeFloats operator*(float val) const
	{
		ThreeFloats retval;

		retval.x = x * val;
		retval.y = y * val;
		retval.z = z * val;

		return retval;
	};

	///	Division operator (ThreeFloats).
	ThreeFloats operator/(const ThreeFloats& op2) const
	{
		ThreeFloats retval;

		retval.x = x * op2.x;
		retval.y = y * op2.y;
		retval.z = z * op2.z;

		return retval;
	};

	///	Division operator (float).
	ThreeFloats operator/(float val) const
	{
		ThreeFloats retval;

		retval.x = x / val;
		retval.y = y / val;
		retval.z = z / val;

		return retval;
	};

	///	Addition assignment operator.
	ThreeFloats& operator+=(const ThreeFloats& op2)
	{
		x += op2.x;
		y += op2.y;
		z += op2.z;

		return *this;
	};

	///	Subtraction assignment operator.
	ThreeFloats& operator-=(const ThreeFloats& op2)
	{
		x -= op2.x;
		y -= op2.y;
		z -= op2.z;

		return *this;
	};

	///	Multiplication assignment operator.
	ThreeFloats& operator*=(const ThreeFloats& op2)
	{
		x *= op2.x;
		y *= op2.y;
		z *= op2.z;

		return *this;
	};

	///	Multiplication assignment operator.
	ThreeFloats& operator*=(float op2)
	{
		x *= op2;
		y *= op2;
		z *= op2;

		return *this;
	};

	///	Division assignment operator.
	ThreeFloats& operator/=(const ThreeFloats& op2)
	{
		x /= op2.x;
		y /= op2.y;
		z /= op2.z;

		return *this;
	};

	///	Division assignment operator.
	ThreeFloats& operator/=(float op2)
	{
		x /= op2;
		y /= op2;
		z /= op2;

		return *this;
	};

	///	Equality operator.
	bool operator==(const ThreeFloats& op2) const
	{
		return (x==op2.x)&&(y==op2.y)&&(z==op2.z);
	};

	///	Inequality operator.
	bool operator!=(const ThreeFloats& op2) const
	{
		return (x!=op2.x)||(y!=op2.y)||(z!=op2.z);
	};

	///	Returns the distance between two ThreeFloats.
	float getDistance(const ThreeFloats& op2) const
	{
		ThreeFloats temp3f;

		temp3f.x = op2.x - x;
		temp3f.y = op2.y - y;
		temp3f.z = op2.z - z;
		return heilanSqrt((temp3f.x*temp3f.x) +
					 (temp3f.y*temp3f.y) +
					 (temp3f.z*temp3f.z));
	};

	///	Normalises the ThreeFloats as a vector.
	void normalise()
	{
		float tempf;

		tempf = (x*x) + (y*y) + (z*z);
		tempf = 1.0f/sqrtf(tempf);

		x *= tempf;
		y *= tempf;
		z *= tempf;
	};

	friend ThreeFloats dllStuff operator*(float op1, const ThreeFloats& op2);
};

#endif
