//	Webcam.cpp - Class representing a webcam (or other video input), with
//			   methods to retrieve its frames.
//  ----------------------------------------------------------------------------
//	This file is part of 'Niall's Cam Lib', a simple library for getting video
//	from webcams (or other video inputs).
//	Copyright (C) 2007-2008  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Webcam.h"

#include <iostream>

#ifdef LINUX
#include <linux/videodev.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#endif

using namespace std;

//------------------------------------------------------------------------------
Webcam::Webcam():
width(320),
height(240),
running(false)
{
	int i;

	//Setup to use 24-bit RGB format.
	frameData = new unsigned char[width*height*3];
	for(i=0;i<(width*height*3);++i)
		frameData[i] = 0;

#ifdef WIN32
	WNDCLASSEX wndClass;

	bitmapInfo = 0;

	//Register the window class we use in startCapture().
	wndClass.cbSize = sizeof(WNDCLASSEX);
	wndClass.style = 0;
	wndClass.lpfnWndProc = customProc;
	wndClass.cbClsExtra = 0;
	wndClass.cbWndExtra = 0;
	wndClass.hInstance = GetModuleHandle(NULL);
	wndClass.hIcon = NULL;
	wndClass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndClass.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wndClass.lpszMenuName = NULL;
	wndClass.lpszClassName = "NiallsCamLibWindow";
	wndClass.hIconSm = NULL;
	RegisterClassEx(&wndClass);
#endif
}

//------------------------------------------------------------------------------
Webcam::~Webcam()
{
	if(running)
		stopCapture();

#ifdef WIN32
	//Unregister the window class.
	UnregisterClass("NiallsCamLibWindow", GetModuleHandle(NULL));
#endif

	delete [] frameData;
}

//------------------------------------------------------------------------------
bool Webcam::startCapture()
{
	if(running)
		return false;

#ifdef WIN32
	if(bitmapInfo)
		delete [] bitmapInfo;

	//Create invisible window to hold the capture window.
	hwnd = CreateWindowEx(0,
						  "NiallsCamLibWindow",
						  "",
						  0,
						  0,
						  0,
						  1,
						  1,
						  HWND_DESKTOP,
						  NULL,
						  GetModuleHandle(NULL),
						  NULL);
	//Create capture window.
	capHwnd = capCreateCaptureWindow("NiallsCamLibWindow",
									 WS_CHILD,
									 0,
									 0,
									 width,
									 height,
									 hwnd,
									 1);

	if(capHwnd)
	{
		CAPDRIVERCAPS caps;
		BITMAPINFO *actualInfo;
		DWORD bitmapSize;

		//Connect the 1st webcam to the capture window.
		if(!capDriverConnect(capHwnd, 0))
			cout << "Could not connect driver." << endl;
		//Attach our callback to the capture window.
		capSetCallbackOnVideoStream(capHwnd, videoCallback);
		//Make sure we can access this object from the callback.
		if(!capSetUserData(capHwnd, this))
			cout << "Could not associate user data with capture." << endl;

		//Check everything's initialised.
		caps.fCaptureInitialized = false;
		capDriverGetCaps(capHwnd, &caps, sizeof(caps));
		if(caps.fCaptureInitialized)
		{
			//Allocate originalInfo and newInfo.
			bitmapSize = capGetVideoFormatSize(capHwnd);
			bitmapInfo = new unsigned char[bitmapSize];

			//Get the current video format.
			actualInfo = (BITMAPINFO *)bitmapInfo;
			capGetVideoFormat(capHwnd, actualInfo, bitmapSize);

			//Set video format (width*height), 24-bit RGB.
			actualInfo->bmiHeader.biWidth = width;
			actualInfo->bmiHeader.biHeight = height;
			actualInfo->bmiHeader.biBitCount = 24;
			actualInfo->bmiHeader.biSizeImage = 0;
			actualInfo->bmiHeader.biCompression = BI_RGB;
			actualInfo->bmiHeader.biClrUsed = 0;
			actualInfo->bmiHeader.biClrImportant = 0;
			actualInfo->bmiHeader.biPlanes = 1;

			if(capSetVideoFormat(capHwnd, actualInfo, bitmapSize))
			{
				CAPTUREPARMS captureParams;

				//Set capture parameters.
				capCaptureGetSetup(capHwnd,
								   &captureParams,
								   sizeof(CAPTUREPARMS));
				captureParams.dwRequestMicroSecPerFrame = 1000000/60; //60fps.
				captureParams.fMakeUserHitOKToCapture = false;
				captureParams.fYield = true;
				captureParams.fCaptureAudio = false;
				captureParams.vKeyAbort = 0;
				captureParams.fAbortLeftMouse = false;
				captureParams.fAbortRightMouse = false;
				captureParams.fLimitEnabled = false;
				captureParams.fMCIControl = false;
				captureParams.fStepCaptureAt2x = false;
				captureParams.wStepCaptureAverageFrames = 5;
				capCaptureSetSetup(capHwnd,
								   &captureParams,
								   sizeof(CAPTUREPARMS));

				//Start capture.
				capCaptureSequenceNoFile(capHwnd);
				running = true;
			}
			else
			{
				capDriverDisconnect(capHwnd);
				DestroyWindow(capHwnd);
				DestroyWindow(hwnd);
				delete [] bitmapInfo;
			}
		}
		else
		{
			capDriverDisconnect(capHwnd);
			DestroyWindow(capHwnd);
			DestroyWindow(hwnd);
		}
	}
	else if(hwnd)
		DestroyWindow(hwnd);
#elif defined(LINUX)
	struct video_capability cap;
	struct video_window win;
	struct video_picture vpic;

	//Open camera.
	fd = open("/dev/video0", O_RDONLY);
	if(fd != -1)
	{
		//Check if it's a Video4Linux device.
		if(ioctl(fd, VIDIOCGCAP, &cap) != -1)
		{
			cout << "Video4Linux Camera:" << endl;
			cout << "-------------------" << endl;
			cout << "Camera name: " << cap.name << endl;
			cout << "Camera type: ";
			if(cap.type & VID_TYPE_CAPTURE)
				cout << "Capture ";
			else if(cap.type & VID_TYPE_TUNER)
				cout << "Tuner ";
			else if(cap.type & VID_TYPE_TELETEXT)
				cout << "Teletext ";
			else if(cap.type & VID_TYPE_OVERLAY)
				cout << "Overlay ";
			else if(cap.type & VID_TYPE_CHROMAKEY)
				cout << "Chromakey ";
			else if(cap.type & VID_TYPE_CLIPPING)
				cout << "Clipping ";
			else if(cap.type & VID_TYPE_FRAMERAM)
				cout << "Frame RAM ";
			else if(cap.type & VID_TYPE_SCALES)
				cout << "Scaling ";
			else if(cap.type & VID_TYPE_MONOCHROME)
				cout << "Monochrome ";
			else if(cap.type & VID_TYPE_SUBCAPTURE)
				cout << "Subcapture " << endl;
			cout << endl;
			cout << "Camera num channels: " << cap.channels << endl;
			cout << "Camera num audio devices: " << cap.audios << endl;
			cout << "Camera Max Width: " << cap.maxwidth << endl;
			cout << "Camera Max Height: " << cap.maxheight << endl;
			cout << "Camera Min Width: " << cap.minwidth << endl;
			cout << "Camera Min Height: " << cap.minheight << endl;

			//Get the video window.
			if(ioctl(fd, VIDIOCGWIN, &win) != -1)
			{
				//Set the window's size.
				win.width = width;
				win.height = height;
				ioctl(fd, VIDIOCSWIN, &win);

				//Get the video picture data.
				if(ioctl(fd, VIDIOCGPICT, &vpic) != -1)
				{
					//Set colours.
					vpic.depth = 24;
					vpic.palette = VIDEO_PALETTE_RGB24;
					vpic.contrast = 49151;
					vpic.brightness = 10496;
					if(ioctl(fd, VIDIOCSPICT, &vpic) != -1)
					{
						//Start capture thread.
						if(!pthread_create(&thread,
										   NULL,
										   Webcam::captureThread,
										   this))
						{
							running = true;
						}
						else
						{
							cout << "Camera error: could not create capture thread." << endl;
							close(fd);
						}
					}
					else
					{
						cout << "Camera error: could not use 24-bit RGB colour." << endl;
						close(fd);
					}
				}
				else
				{
					cout << "Camera error: could not get picture data." << endl;
					close(fd);
				}
			}
			else
			{
				cout << "Camera error: could not get video window." << endl;
				close(fd);
			}
		}
		else
		{
			cout << "Camera error: could not get device capabilities." << endl;
			close(fd);
		}
	}
	else
		cout << "Camera error: could not open camera." << endl;
#endif

	return running;
}

//------------------------------------------------------------------------------
void Webcam::stopCapture()
{
	if(!running)
		return;

	running = false;

#ifdef WIN32
	//Stop capture.
	capCaptureStop(capHwnd);

	BITMAPINFO *actualInfo = (BITMAPINFO *)bitmapInfo;
	delete [] bitmapInfo;
	bitmapInfo = 0;

	//Disconnect driver.
	capDriverDisconnect(capHwnd);

	//Destroy windows.
	DestroyWindow(capHwnd);
	DestroyWindow(hwnd);
#elif defined(LINUX)
	//Wait for thread to quit.
	pthread_join(thread, NULL);

	//Close camera.
	close(fd);
#endif
}

//------------------------------------------------------------------------------
void Webcam::addCallback(WebcamCallback *callback)
{
	callbacks.push_back(callback);
}

//------------------------------------------------------------------------------
void Webcam::callCallbacks(unsigned char *data, long bufferSize)
{
	vector<WebcamCallback *>::iterator it;

#ifdef LINUX
	if(read(fd, frameData, (width*height*3)) < (width*height*3))
		cout << "Camera error: could not read frameData." << endl;
#elif defined(WIN32)
	if(bufferSize == (width*height*3))
		memcpy(frameData, data, bufferSize);
#endif

	for(it=callbacks.begin();it!=callbacks.end();++it)
		(*it)->callback(frameData);
}

#ifdef WIN32
//------------------------------------------------------------------------------
LRESULT CALLBACK Webcam::customProc(HWND hwnd,
									UINT uMsg,
									WPARAM wParam,
									LPARAM lParam)
{
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

//------------------------------------------------------------------------------
LRESULT CALLBACK Webcam::videoCallback(HWND hwnd, LPVIDEOHDR lpVHdr)
{
	Webcam *cam = (Webcam *)capGetUserData(hwnd);

	if(cam)
	{
		cam->callCallbacks((unsigned char *)lpVHdr->lpData,
						   lpVHdr->dwBufferLength);
	}

	return true;
}
#elif defined(LINUX)
//------------------------------------------------------------------------------
void *Webcam::captureThread(void *data)
{
	Webcam *webcam = (Webcam *)data;
	timeval timeToWait;
	fd_set setToCheck;
	int fd;

	if(webcam)
	{
		fd = webcam->getFileDescriptor();
		timeToWait.tv_sec = 0;
		timeToWait.tv_usec = 2500; // 1/4 of a second.

		while(webcam->isRunning())
		{
			FD_ZERO(&setToCheck);
			FD_SET(fd, &setToCheck);

			if(select((fd+1), &setToCheck, NULL, NULL, &timeToWait) == -1)
				cout << "Camera: select() error." << endl;
			else if(FD_ISSET(fd, &setToCheck))
				webcam->callCallbacks(0, 0);
		}
	}

	return 0;
}
#endif
