//	Webcam.h - Class representing a webcam (or other video input), with
//			   methods to retrieve its frames.
//  ----------------------------------------------------------------------------
//	This file is part of 'Niall's Cam Lib', a simple library for getting video
//	from webcams (or other video inputs).
//	Copyright (C) 2007-2008  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef WEBCAM_H_
#define WEBCAM_H_

#ifdef WIN32
#include <windows.h>
#include <vfw.h>
#elif defined(LINUX)
#include <pthread.h>
#endif

#include <vector>

//------------------------------------------------------------------------------
///	Class representing a Webcam callback.
/*!
	Assigning a subclass of this to a Webcam object will result in the
	callback() method being called whenever the video input updates its current
	frame.
 */
class WebcamCallback
{
  public:
	///	Virtual destructor (only needed to kill some gcc warnings, imo).
	virtual ~WebcamCallback() {};

	///	To be called when a Webcam object updates its curernt frame.
	/*!
		\param data The pixel data, as an array of RGB bytes.
	 */
	virtual void callback(unsigned char *data) = 0;
};

//------------------------------------------------------------------------------
///	Class representing a webcam (or other video input), with methods to retrieve its frames.
/*!
	The default size of the images taken is 320x240.

	The linux port currently only supports Video4Linux 1, as that's all my camera
	supports.
 */
class Webcam
{
  public:
	///	Constructor.
	Webcam();
	///	Destructor.
	~Webcam();

	///	Starts capture.
	/*!
		\return True on success, false otherwise.
	 */
	bool startCapture();
	///	Stops capture.
	void stopCapture();

	///	Sets the size of the images taken.
	/*!
		Call this before you call startCapture().
	 */
	void setSize(int w, int h);

	///	Adds a callback to this Webcam.
	void addCallback(WebcamCallback *callback);
	///	Calls the Webcam's callbacks.
	/*!
		This should only be called from the platform-specific callbacks.
	 */
	void callCallbacks(unsigned char *data, long bufferSize);

	///	Returns whether or not the Webcam is currently capturing.
	bool isRunning() const {return running;};

#ifdef WIN32
	///	Window procedure for the invisible parent window.
	static LRESULT CALLBACK customProc(HWND hwnd,
									   UINT uMsg,
									   WPARAM wParam,
									   LPARAM lParam);
	///	Callback for the video capture.
	static LRESULT CALLBACK videoCallback(HWND hwnd, LPVIDEOHDR lpVHdr);
#elif defined(LINUX)
	///	Helper method to return the camera's file descriptor.
	int getFileDescriptor() {return fd;};
	

	///	Callback for the capture thread.
	static void *captureThread(void *data);
#endif
  private:
	///	Width of the capture window.
	int width;
	///	Height of the capture window.
	int height;
	///	The current frame.
	unsigned char *frameData;
	///	All the callbacks registered with this Webcam.
	std::vector<WebcamCallback *> callbacks;

	///	Whether we're currently capturing or not.
	bool running;

#ifdef WIN32
	///	Handle to the invisible parent window we need to start capture.
	HWND hwnd;
	///	Handle to the capture window.
	HWND capHwnd;
	///	Information about the bitmap the webcam uses.
	unsigned char *bitmapInfo;
#elif defined(LINUX)
	///	File descriptor to the camera.
	int fd;
	///	Pthread where capturing takes place.
	pthread_t thread;
#endif
};

#endif
