//	PerlinTexture.cpp - Perlin noise texture.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "PerlinTexture.h"

#include <cmath>

namespace HeilanX3D
{

//------------------------------------------------------------------------------
PerlinTexture::PerlinTexture(int components, int size):
numComponents(components),
texSize(size),
persistence(0.99f)
{
	generatePerlin();
}

//------------------------------------------------------------------------------
PerlinTexture::~PerlinTexture()
{
	
}

//------------------------------------------------------------------------------
void PerlinTexture::setPersistence(float val)
{
	persistence = val;

	generatePerlin();
}

//------------------------------------------------------------------------------
void PerlinTexture::generatePerlin()
{
	int i, j, k, l;
	int x, y, c;
	int tempint;
	float tempX, tempY;
	int tempX2, tempY2;
	float corners[4][4];
	float temp1, temp2, temp3, temp4;
	float tempP;
	const float * const noiseTexture = noiseTex.getPixels();

	//Clear main texture.
	for(y=0;y<texSize;++y)
	{
		for(x=0;x<texSize;++x)
		{
			for(c=0;c<numComponents;++c)
				pixels[(((y*texSize)+x)*numComponents)+c] = 0.5f;
		}
	}

	//Loop through noise octaves.
	for(i=4, j=0;i<texSize;i*=2, ++j)
	{
		tempP = powf(persistence, (float)i);

		//Go through the texture's pixels, add to perlinTexture.
		for(y=0;y<texSize;++y)
		{
			tempY = (float)y/(float)texSize;
			tempY *= (float)i;

			for(x=0;x<texSize;++x)
			{
				tempX = (float)x/(float)texSize;
				tempX *= (float)i;

				tempint = ((y*texSize)+x) * numComponents;

				//Get corner pixel values.
				for(k=-1;k<3;++k)
				{
					for(l=-1;l<3;++l)
					{
						tempX2 = (int)tempX + l;
						if(tempX2 < 0)
							tempX2 += i;
						else if(tempX2 >= i)
							tempX2 -= i;

						tempY2 = (int)tempY + k;
						if(tempY2 < 0)
							tempY2 += i;
						else if(tempY2 >= i)
							tempY2 -= i;

						corners[k+1][l+1] = noiseTexture[tempX2 +
														 (tempY2*texSize)];
					}
				}

				//Interpolate.
				temp1 = bSplineInterp(corners[0][0],
									  corners[0][1],
									  corners[0][2],
									  corners[0][3],
									  (tempX-(float)((int)tempX)));
				temp2 = bSplineInterp(corners[1][0],
									  corners[1][1],
									  corners[1][2],
									  corners[1][3],
									  (tempX-(float)((int)tempX)));
				temp3 = bSplineInterp(corners[2][0],
									  corners[2][1],
									  corners[2][2],
									  corners[2][3],
									  (tempX-(float)((int)tempX)));
				temp4 = bSplineInterp(corners[3][0],
									  corners[3][1],
									  corners[3][2],
									  corners[3][3],
									  (tempX-(float)((int)tempX)));
				for(c=0;c<numComponents;++c)
				{
					pixels[tempint+c] += (bSplineInterp(temp1,
														temp2,
														temp3,
														temp4,
														(tempY-(float)((int)tempY)))-0.5f) *
									   	 tempP;
				}
			}
		}
	}

	//Add final noise texture.
	tempP = powf(persistence, (float)i);
	for(y=0;y<texSize;++y)
	{
		for(x=0;x<texSize;++x)
		{
			tempint = ((y*texSize)+x) * numComponents;
			for(c=0;c<numComponents;++c)
				pixels[tempint+c] += (noiseTexture[tempint/numComponents]-0.5f) * tempP;
		}
	}
}

}
