//	AccelLine.h - Line segments drawn according to accelerometer data.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef ACCELLINE_H_
#define ACCELLINE_H_

#include "AbstractNode.h"
#include "CriticalSection.h"

#include "GLHeaders.h"

/*+
	Name: AccelLine
	Type: ChildNode.
	Description: Draws a line from 3d accelerometer data.
	Attribute:	xAccel
	AttType:	SFFloat (input)
	AttDesc:	The x-axis acceleration.

	Attribute:	yAccel
	AttType:	SFFloat (input)
	AttDesc:	The y-axis acceleration.

	Attribute:	zAccel
	AttType:	SFFloat (input)
	AttDesc:	The z-axis acceleration.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0.0 0.0 -96.0"&gt;
			&lt;AccelLine DEF="bob"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

//------------------------------------------------------------------------------
///	Line segments drawn according to accelerometer data.
/*!
	\ingroup libheilanextras
 */
class AccelLine : public ChildNode
{
  public:
	///	Constructor.
	AccelLine();
	///	Destructor.
	~AccelLine();

	///	Initialises the texture.
	void initialise();

	///	Draws the AccelLine.
	void graphicDoStuff();
	///	Overridden to do nothing.
	//AmbisonicData *audioDoStuff(int numSamples, const ThreeFloats& pos) {return 0;};

	///	Overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set this node's attributes.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<AccelLine> constructor;
  private:
	///	Helper method to calculate the colour of a segment.
	ThreeFloats getColour(const ThreeFloats& pos);

	///	Simple struct to represent a single line segment.
	struct Line
	{
		///	Colour of the segment.
		ThreeFloats colour;
		///	Velocity of the accelerometer at this point in time.
		ThreeFloats velocity;
		///	End position of the segment.
		ThreeFloats endPos;
	};

	///	Some constants.
	enum
	{
		NumSegments = 2048,
		LineLength = 1024
	};

	///	Our array of potential line segments.
	Line lines[NumSegments];
	///	Current write position within lines.
	int writePos;
	///	If the x-axis pos at the current writePos needs written.
	bool needX;
	///	If the y-axis pos at the current writePos needs written.
	bool needY;
	///	If the z-axis pos at the current writePos needs written.
	bool needZ;

	///	Used to perform the integration.
	float deltaTime;

	///	The number of points in the line to draw.
	int lineLength;

	///	The texture we use.
	GLuint texture;
	///	Our pixel data.
	unsigned char pixels[256*256*2];
	///	Storage for our atom quad.
	GLuint atom;

	///	Whether to draw the line as blobs or connected quads.
	bool drawAsQuads;

	///	Used to ignore too-small values.
	float calibVal;

	///	Test hi-pass filter.
	ThreeFloats filtLast;
	///	Test hi-pass filter.
	ThreeFloats filtCoeff;

	///	Velocity filter.
	ThreeFloats yLast;
	///	Velocity filter.
	ThreeFloats xLast;
	///	Velocity filter.
	ThreeFloats aLast;

	///	Previous unfixed velocity value.
	ThreeFloats lastVel;
	///	Previous unfixed position value.
	ThreeFloats lastPos2;

	///	Acceleration average.
	ThreeFloats accelAv[16];
	///	Velocity dc filter coefficient.
	float dcFiltCoeff;
	///	Position dc filter coefficient.
	float dcFiltCoeff2;

	///	Critical section to protect the line segments.
	CriticalSection critSec;
};

}

#endif
