//	AccelTransform.cpp - Simple Transform node controlled by 3d acceleration
//						 data.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "AccelTransform.h"
#include "StringParser.h"
#include "AttributeHelper.h"
#include "Camera.h"

#include "GLHeaders.h"
#include <iostream>
#include <sstream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<AccelTransform> AccelTransform::constructor;

//------------------------------------------------------------------------------
AccelTransform::AccelTransform():
ParentNode(),
center(0.0f),
rotation(0.0f, 0.0f, 1.0f, 0.0f),
scale(1.0f),
scaleOrientation(0.0f, 0.0f, 1.0f, 0.0f),
translation(0.0f),
bboxCenter(0.0f),
bboxSize(-1.0f),
moveCamera(false),
cameraNeedsMoving(false)
{
	setType("AccelTransform");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
AccelTransform::~AccelTransform()
{
	
}

//------------------------------------------------------------------------------
void AccelTransform::setAttribute(const string& attribute, EventValue& val)
{
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "xAccel")
	{
		translation.x += (float)val;
		cameraNeedsMoving = true;
	}
	else if(realAtt == "yAccel")
	{
		translation.y += (float)val;
		cameraNeedsMoving = true;
	}
	else if(realAtt == "zAccel")
	{
		translation.z += (float)val;
		cameraNeedsMoving = true;
	}
	else if(realAtt == "center")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&center);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			center = parser.getThreeFloats();
	}
	else if(realAtt == "rotation")
	{
		StringParser parser(val);
		AttributeHelper<FourFloats> helper(&rotation);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumFourFloats() > 0)
			rotation = parser.getFourFloats();
	}
	else if(realAtt == "scale")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&scale);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			scale = parser.getThreeFloats();
	}
	else if(realAtt == "scaleOrientation")
	{
		StringParser parser(val);
		AttributeHelper<FourFloats> helper(&scaleOrientation);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumFourFloats() > 0)
			scaleOrientation = parser.getFourFloats();
	}
	else if(realAtt == "translation")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&translation);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			translation = parser.getThreeFloats();
	}
	else if(realAtt == "bboxCenter")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&bboxCenter);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			bboxCenter = parser.getThreeFloats();
	}
	else if(realAtt == "bboxSize")
	{
		StringParser parser(val);

		if(parser.getNumThreeFloats() > 0)
			bboxSize = parser.getThreeFloats();
	}
	else if(realAtt == "moveCamera")
	{
		StringParser parse(val);

		if(parse.getNumBools() > 0)
		{
			if(parse.getBool())
			{
				moveCamera = true;
				cameraNeedsMoving = true;
			}
			else
				moveCamera = false;
		}
	}
}

//------------------------------------------------------------------------------
void AccelTransform::preChildDraw()
{
	//Setup matrix.
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	//Move camera, if necessary.
	if(moveCamera && cameraNeedsMoving)
	{
		ThreeFloats tempCam = translation;

		tempCam.y = -tempCam.y;

		Camera::getInstance()->setGraphicPosition(tempCam);
	}
	if(cameraNeedsMoving)
	{
		stringstream tempConverter;

		//Send event.
		tempConverter << translation.x << " ";
		tempConverter << -translation.y << " ";
		tempConverter << translation.z;
		sendEvent("translation_changed", tempConverter.str());
	}

	//Now perform various translations etc.

	//Apply the rotation.
	glRotatef(rotation.r, rotation.x, rotation.y, rotation.z);

	//Apply the translation, offset from center.
	glTranslatef(center.x+translation.x,
				 center.y+translation.y,
				 center.z+translation.z);

	//First sort out the scale orientation.
	glRotatef(scaleOrientation.r,
			  scaleOrientation.x,
			  scaleOrientation.y,
			  scaleOrientation.z);
	//Now Apply our scale operation.
	glScalef(scale.x, scale.y, scale.z);
	//Revert to previous orientation.
	glRotatef(-scaleOrientation.r,
			  -scaleOrientation.x,
			  -scaleOrientation.y,
			  -scaleOrientation.z);
}

//------------------------------------------------------------------------------
void AccelTransform::postChildDraw()
{
	GLint tempint;

	//Reset matrix.
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	//Check if we should pop the texture matrix.
	//This is far from ideal, but I can't think of anywhere else to put it,
	//since TextureAccelTransform is a child node...
	glGetIntegerv(GL_TEXTURE_STACK_DEPTH, &tempint);
	if(tempint > 1)
	{
		glMatrixMode(GL_TEXTURE);
		glPopMatrix();

		//Just in case...
		glMatrixMode(GL_MODELVIEW);
	}
}

//------------------------------------------------------------------------------
AmbisonicData *AccelTransform::preChildAudio(int numSamples,
										const ThreeFloats& newPos)
{
	setPosition(translation + newPos);

	return 0;
}

}
