//	Transform.h - Simple Transform node controlled by 3d acceleration data.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef ACCELTRANSFORM_H_
#define ACCELTRANSFORM_H_

#include "ParentNode.h"
#include "FourFloats.h"

/*+
	Name: AccelTransform
	Type: ParentNode.
	Description: A modified Transform node to change its position according to
				 3d accelerometer data.  Can also move the camera accordingly,
				 and it outputs the event <i>'translation_changed'</i>, to be
				 routed to other nodes.  Shares the same attributes as
				 Transform, plus the following:
	Attribute:	xAccel
	AttType:	SFFloat (input)
	AttDesc:	The x-axis acceleration, used to modify the node's translation.

	Attribute:	yAccel
	AttType:	SFFloat (input)
	AttDesc:	The y-axis acceleration, used to modify the node's translation.

	Attribute:	zAccel
	AttType:	SFFloat (input)
	AttDesc:	The z-axis acceleration, used to modify the node's translation.

	Attribute:	moveCamera
	AttType:	SFBoolean (input)
	AttDesc:	If true, the camera is moved when the node's translation is
				changed.

	Attribute:	translation_changed
	AttType:	SFVec3f (output)
	AttDesc:	The current (updated) translation value.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;
		&lt;Background skyColor="1 1 1 0.66 0.66 1" skyAngle="0 1" groundColor="0.125 0.5 0.125 0.5 0.75 0.4 1 1 1" groundAngle="0 1.3 1.570796327"/&gt;

		&lt;AccelTransform DEF="bob" moveCamera="true"&gt;
			&lt;Transform translation="0.0 0.0 -8.0 "&gt;
				&lt;Shape&gt;
					&lt;Appearance&gt;
						&lt;Material DEF="boxColour" diffuseColor="1.0 0.0 0.0 "/&gt;
					&lt;/Appearance&gt;
					&lt;Box/&gt;
				&lt;/Shape&gt;
			&lt;/Transform&gt;
		&lt;/AccelTransform&gt;

		&lt;Transform translation="-2 0 -10"&gt;
			&lt;Shape&gt;
				&lt;Appearance&gt;
					&lt;Material DEF="boxColour" diffuseColor="0 0 1"/&gt;
				&lt;/Appearance&gt;
				&lt;Box/&gt;
			&lt;/Shape&gt;
		&lt;/Transform&gt;

		&lt;Transform translation="2 1 -10"&gt;
			&lt;Shape&gt;
				&lt;Appearance&gt;
					&lt;Material DEF="boxColour" diffuseColor="0.1 0.8 0.1"/&gt;
				&lt;/Appearance&gt;
				&lt;Box/&gt;
			&lt;/Shape&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Simple Transform node controlled by 3d acceleration data.
/*!
	\ingroup libheilanextras

	Can also move the camera to point at the node's centre.
 */
class AccelTransform : public ParentNode
{
  public:
	///	Constructor.
	AccelTransform();
	///	Destructor.
	~AccelTransform();

	///	Used to set the transform's attributes..
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<AccelTransform> constructor;
  private:
	///	We override this so we can save and restore the current OpenGL matrix.
	void preChildDraw();
	///	We override this so we can save and restore the current OpenGL matrix.
	void postChildDraw();
	///	We override this so we can update our position for our children.
	AmbisonicData *preChildAudio(int numSamples, const ThreeFloats& newPos);

	///	Implemented to do nothing.
	void audioBlockSizeChanged(int newSize) {};

	///	The centre of the transform operation.
	ThreeFloats center;
	///	The rotation of the transform.
	FourFloats rotation;
	///	The scale of the transform.
	ThreeFloats scale;
	///	The orientation of the scale operation.
	FourFloats scaleOrientation;
	///	The translation of the transform.
	ThreeFloats translation;
	///	The centre of the bounding box.
	ThreeFloats bboxCenter;
	///	The size of the bounding box.
	ThreeFloats bboxSize;

	///	Whether or not the camera should move when translation changes.
	bool moveCamera;

	///	Whether or not the camera needs moving.
	bool cameraNeedsMoving;
};

}

#endif
