//	Ashitaka.cpp - The Ashitaka audiovisual instrument.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Ashitaka.h"
#include "StringParser.h"
#include "Camera.h"
#include "TwoFloats.h"

#include <iostream>
#include <cassert>
#include <sstream>

using namespace std;
using namespace AshitakaClasses;

namespace HeilanX3D
{

NodeConstructor<Ashitaka> Ashitaka::constructor;

//------------------------------------------------------------------------------
Ashitaka::Ashitaka():
ChildNode(),
audioBuffer(0),
bufferSize(0),
paintCount(0)
{
	setType("Ashitaka");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
Ashitaka::~Ashitaka()
{
	if(audioBuffer)
		delete [] audioBuffer;
}

//------------------------------------------------------------------------------
void Ashitaka::initialise()
{
	quadManager.setAudioSynth(&audioSynth);

	//Audio settings.
	setMaxFrontAudioRadius(50.0f);
	setMaxBackAudioRadius(50.0f);
	setMinFrontAudioRadius(25.0f);
	setMinBackAudioRadius(25.0f);

	//setupMappingsMIDI();
	setupMappingsInterface();
	/*interface.slider1 >> quadManager.mainSize;
	interface.slider2 >> quadManager.mainXPos;
	interface.slider3 >> quadManager.mainYPos;
	interface.slider4 >> quadManager.mainZPos;
	interface.slider5 >> quadManager.mainTwist;
	interface.slider6 >> quadManager.mainExplode;
	//interface.slider7 >> quadManager.mainStream1;
	//interface.slider8 >> quadManager.mainStream2;
	//interface.slider9 >> quadManager.mainStream3;
	interface.slider7 >> quadManager.splitOff;*/
}

//------------------------------------------------------------------------------
void Ashitaka::graphicDoStuff()
{
	TwoFloats wobble;
	stringstream tempstr;

	wobble = quadManager.update();
	quadManager.draw();

	//Send wobble parameters.
	tempstr << wobble.x;
	sendEvent("wobble1", tempstr.str());
	tempstr.str("");
	tempstr << wobble.y;
	sendEvent("wobble2", tempstr.str());

	++paintCount;
	if(paintCount > 5)
	{
		int i;
		Quad *quads = quadManager.getQuads();
		stringstream tempOut;
		ThreeFloats temp3f;

		paintCount = 0;

		for(i=0;i<QuadManager::NumQuads;++i)
		{
			temp3f = quads[i].getCentre();
			tempOut << temp3f.x << " " << temp3f.y << " " << temp3f.z << " ";
			temp3f = quads[i].getColour();
			tempOut << temp3f.x << " " << temp3f.y << " " << temp3f.z << " ";
		}
		sendEvent("quadsPosCol", tempOut.str());
	}
}

//------------------------------------------------------------------------------
AmbisonicData *Ashitaka::getAmbisonicAudio(int numSamples,
										   const ThreeFloats& pos)
{
	int i;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;
	const ThreeFloats camPos = Camera::getInstance()->getAudioPosition();
	const Quaternion& camRot = Camera::getInstance()->getAudioRotation();
	//const float rotMult = 6.283185306f/360.0f;
	AmbisonicData *goAudio = 0;
	ThreeFloats tempPos(audioSynth.positionX,
						audioSynth.positionY,
						audioSynth.positionZ);
	tempPos += pos;

	//Handle main GO.
	audioSynth.getAudioBlock(audioBuffer, numSamples);
	goAudio = quadManager.getAudio(numSamples);
	if(audioBuffer)
	{
		//Get current Ambisonic position (see AmbisonicData::operator=()),
		//relative to the camera.

		//So what we do here is:
		//1.) Ensure relativePos is equal to this node's spatial position in the
		//    scene.
		relativePos = pos;
		//2.) Rotate relativePos by the rotation of the camera, by multiplying
		//    it by the camera's orientation Quaternion.
		relativePos = camRot * relativePos;
		//3.) Translate relativePos by the position of the camera.
		relativePos -= camPos;

		lastAtten = calcRelativeAttenuation(lastPos);
		if(lastPos != relativePos)
		{
			ambiStart = lastPos;
			ambiEnd = relativePos;
			//We increment the position by a constant value, so we don't get
			//clicks or zipper noise.  Of course, there are disadvantages to
			//this approach...
			ambiInc = 0.005f;
			currentAmbi = ambiStart;
			newAtten = calcRelativeAttenuation(relativePos);
		}
		else
		{
			currentAmbi = relativePos;
			newAtten = lastAtten;
		}
		currentAtten = lastAtten;

		//Encode audio.
		for(i=0;i<numSamples;++i)
		{
			//Update currentAmbi and currentAtten if necessary.
			if(ambiInc > 0.0f)
			{
				//Calc currentAmbi.
				currentAmbi = ((1.0f-ambiIndex) * ambiStart) +
							  (ambiIndex * ambiEnd);
				//Calc currentAtten.
				currentAtten = ((1.0f-ambiIndex) * lastAtten) +
							   (ambiIndex * newAtten);

				ambiIndex += ambiInc;

				//Stop updating index if we've reached our destination.
				//I get weird rounding errors here, hence the .0000001f etc...
				if(ambiIndex > 1.0000001f)
				{
					ambiInc = 0.0f;
					currentAtten = newAtten;
				}
			}

			//Convert float sound file to Ambisonic, put it in the buffer.
			audioBuffer[i] = currentAmbi * (audioBuffer[i].W * currentAtten);

			//Add extra GOs' audio.
			if(goAudio)
				audioBuffer[i] += goAudio[i];
		}
		//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
		if(ambiIndex >= 1.0f)
		{
			lastPos = relativePos;
			lastAtten = currentAtten;
		}
		else
		{
			lastPos = ((1.0f-ambiIndex) * lastPos) + (ambiIndex * relativePos);
			lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
		}
	}

	return audioBuffer;
}

//------------------------------------------------------------------------------
void Ashitaka::setBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
	quadManager.setBlockSize(newSize);
}

//------------------------------------------------------------------------------
void Ashitaka::setSamplerate(float val)
{
	audioSynth.setSamplerate(val);
}

//------------------------------------------------------------------------------
void Ashitaka::setAttribute(const string& attribute, EventValue& val)
{
	if(attribute == "length")
		interface.length = val;
	else if(attribute == "twist")
		interface.twist = val;
	else if(attribute == "force1")
		interface.force1 = val;
	else if(attribute == "force2")
		interface.force2 = val;
	else if(attribute == "force3")
		interface.force3 = val;
	else if(attribute == "force4")
		interface.force4 = val;
	else if(attribute == "accelX")
		interface.accelX = val;
	else if(attribute == "accelY")
		interface.accelY = val;
	else if(attribute == "accelZ")
		interface.accelZ = val;
	else if(attribute == "slider1")
		interface.slider1 = val;
	else if(attribute == "slider2")
		interface.slider2 = val;
	else if(attribute == "slider3")
		interface.slider3 = val;
	else if(attribute == "slider4")
		interface.slider4 = val;
	else if(attribute == "slider5")
		interface.slider5 = val;
	else if(attribute == "slider6")
		interface.slider6 = val;
	else if(attribute == "slider7")
		interface.slider7 = val;
	else if(attribute == "slider8")
		interface.slider8 = val;
	else if(attribute == "slider9")
		interface.slider9 = val;
}

//------------------------------------------------------------------------------
void Ashitaka::setupMappingsMIDI()
{
	//Position.
	{
		interface.slider1 >> quadManager.mainXPos;
		interface.slider1 >> audioSynth.positionX;
		interface.slider2 >> quadManager.mainYPos;
		interface.slider2 >> audioSynth.positionY;
		interface.slider3 >> quadManager.mainZPos;
		interface.slider3 >> audioSynth.positionZ;
	}

	//Excitation.
	{
		pluckTriggerScale.setConstant(1.0f);
		exciteScale.setConstant(0.125f);
		bowDispExciteSum.setInputs(&(interface.slider4),
								   &(audioSynth.bowDisplacement));
		bowDispExciteSum.setConstant(0.5f);

		interface.slider4 >> bowDispExciteSum;
		audioSynth.bowDisplacement >> bowDispExciteSum >> quadManager.mainTwist;
		interface.slider4 >> exciteScale >> audioSynth.bowForce;
		exciteScale >> audioSynth.bowVelocity;
		exciteScale >> pluckTriggerScale >> audioSynth.pluckTrigger;
	}

	//Num Facets
	{
		quadManager.numMainQuads >> numQuadsInvert >> audioSynth.hiFilterCutoff;
	}

	//Internal Movement
	{
		audioSynth.rmsLo >> quadManager.mainStream1;
		audioSynth.rmsMid >> quadManager.mainStream2;
		audioSynth.rmsHi >> quadManager.mainStream3;
	}

	//Connectedness/Sustain
	{
		explodeSum.setInputs(&explodeGtThan,
							 &(audioSynth.transientDetector));
		connectSum.setInputs(&dampGtThan, &dampGtThan2);
		connectSum2.setInputs(&fbGtThan, &fbGtThan2);
		connectSum.setConstant(0.5f);
		connectSum2.setConstant(0.5f);
		dampMult.setConstant(0.9f);
		dampGtThan.setConstant(0.5f);
		dampGtThan2.setConstant(0.5f);
		fbGtThan.setConstant(0.5f);
		fbGtThan2.setConstant(0.5f);
		explodeGtThan.setConstant(0.5f);

		interface.slider5 >> explodeInvert >> explodeGtThan >> explodeSum;
		audioSynth.transientDetector >> explodeSum >> quadManager.mainExplode;

		interface.slider5 >> dampInvert >> dampGtThan >> connectSum;
		quadManager.mainConsistency >> dampInvert2 >> dampGtThan2 >> connectSum;
		connectSum >> dampMult >> audioSynth.dampingValue;

		interface.slider5 >> fbGtThan >> connectSum2;
		quadManager.mainConsistency >> fbGtThan2 >> connectSum2;
		connectSum2 >> audioSynth.feedbackAmp;
	}

	//Pitch/Size
	{
		sizeSum.setConstant(0.1f);
		sizeMult.setConstant(4.0f);
		pitchSum1.setInputs(&(quadManager.mainPerceivedSize),
							&(quadManager.mainAvgVelocity));
		pitchSum1.setConstant(0.5f);
		pitchSum2.setInputs(&(interface.slider6), &pitchSum1);
		pitchSum2.setConstant(1.0f/1.5f);

		interface.slider6 >> sizeInvert >> sizeSum >> sizeMult >> quadManager.mainSize;

		quadManager.mainPerceivedSize >> pitchSum1;
		quadManager.mainAvgVelocity >> pitchSum1;
		interface.slider6 >> pitchSum2;
		pitchSum1 >> pitchSum2 >> audioSynth.pitch;
	}
}

//------------------------------------------------------------------------------
void Ashitaka::setupMappingsInterface()
{
	//Position.
	{
		/*interface.slider1 >> quadManager.mainXPos;
		interface.slider1 >> audioSynth.positionX;
		interface.slider2 >> quadManager.mainYPos;
		interface.slider2 >> audioSynth.positionY;
		interface.slider3 >> quadManager.mainZPos;
		interface.slider3 >> audioSynth.positionZ;*/

		interface.accelX >> xAxisConvert >> quadManager.mainXPos;
		xAxisConvert >> audioSynth.positionX;
		interface.accelY >> yAxisConvert >> quadManager.mainYPos;
		yAxisConvert >> audioSynth.positionY;
		interface.accelZ >> zAxisConvert >> quadManager.mainZPos;
		zAxisConvert >> audioSynth.positionZ;
	}

	//Excitation.
	{
		pluckTriggerScale.setConstant(8.0f);
		exciteScale.setConstant(0.125f);
		bowDispExciteSum.setInputs(&(interface.force1),
								   &(audioSynth.bowDisplacement));
		bowDispExciteSum.setConstant(0.5f);

		interface.force1 >> bowDispExciteSum;
		audioSynth.bowDisplacement >> bowDispExciteSum >> quadManager.mainTwist;
		interface.force1 >> exciteScale >> audioSynth.bowForce;
		exciteScale >> audioSynth.bowVelocity;
		exciteScale >> pluckTriggerScale >> audioSynth.pluckTrigger;
	}

	//Num Facets
	{
		quadManager.numMainQuads >> numQuadsInvert >> audioSynth.hiFilterCutoff;
	}

	//Internal Movement
	{
		audioSynth.rmsLo >> quadManager.mainStream1;
		audioSynth.rmsMid >> quadManager.mainStream2;
		audioSynth.rmsHi >> quadManager.mainStream3;
	}

	//Connectedness/Sustain
	{
		explodeSum.setInputs(&explodeGtThan,
							 &(audioSynth.transientDetector));
		connectSum.setInputs(&dampGtThan, &dampGtThan2);
		connectSum2.setInputs(&fbGtThan, &fbGtThan2);
		connectSum.setConstant(0.5f);
		connectSum2.setConstant(0.5f);
		dampMult.setConstant(0.9f);
		dampGtThan.setConstant(0.5f);
		dampGtThan2.setConstant(0.5f);
		fbGtThan.setConstant(0.6f);
		fbGtThan2.setConstant(0.6f);
		explodeGtThan.setConstant(0.5f);

		interface.length >> explodeInvert >> explodeGtThan >> explodeSum;
		audioSynth.transientDetector >> explodeSum >> quadManager.mainExplode;

		interface.length >> dampInvert >> dampGtThan >> connectSum;
		quadManager.mainConsistency >> dampInvert2 >> dampGtThan2 >> connectSum;
		connectSum >> dampMult >> audioSynth.dampingValue;

		interface.length >> fbGtThan >> connectSum2;
		quadManager.mainConsistency >> fbGtThan2 >> connectSum2;
		connectSum2 >> audioSynth.feedbackAmp;
	}

	//Pitch/Size
	{
		sizeSum.setConstant(0.1f);
		sizeMult.setConstant(4.0f);
		pitchSum1.setInputs(&(quadManager.mainPerceivedSize),
							&(quadManager.mainAvgVelocity));
		pitchSum1.setConstant(0.5f);
		pitchSum2.setInputs(&(interface.twist), &pitchSum1);
		pitchSum2.setConstant(1.0f/1.5f);

		interface.twist >> sizeInvert >> sizeSum >> sizeMult >> quadManager.mainSize;

		quadManager.mainPerceivedSize >> pitchSum1;
		quadManager.mainAvgVelocity >> pitchSum1;
		interface.twist >> pitchSum2;
		pitchSum1 >> pitchSum2 >> audioSynth.pitch;
	}

	//Split Off.
	{
		interface.force2 >> quadManager.splitOff2;
		interface.force3 >> quadManager.splitOff3;
		interface.force4 >> quadManager.splitOff4;
	}
}

}
