//	Ashitaka.h - The Ashitaka audiovisual instrument.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef ASHITAKA_H_
#define ASHITAKA_H_

#include "AbstractNode.h"
#include "Ashitaka/Interface.h"
#include "Ashitaka/AudioSynth.h"
#include "Ashitaka/VisualSynth.h"
#include "Ashitaka/Map1InSumConst.h"
#include "Ashitaka/Map1InSubConst.h"
#include "Ashitaka/Map1InMultConst.h"
#include "Ashitaka/Map1InSumConstNormal.h"
#include "Ashitaka/Map1InInvert.h"
#include "Ashitaka/Map1InExp.h"
#include "Ashitaka/Map1InParamFilter.h"
#include "Ashitaka/Map1InGtThanNormal.h"
#include "Ashitaka/Map2InSum.h"
#include "Ashitaka/Map2InMult.h"
#include "Ashitaka/Map2InSumMultConst.h"
#include "Ashitaka/Accel2Pos.h"
#include "Ashitaka/QuadManager.h"

/*+
	Name: Ashitaka
	Type: ChildNode.
	Description: The first attempt at my Ashitaka audiovisual instrument, which
				 I'm currently developing for my PhD.  I don't know how easy it
				 is to play it without the custom interface I've built for it...
	Attribute:	length
	AttType:	SFFloat (input)
	AttDesc:	The lengthwise potentiometer on the interface.

	Attribute:	twist
	AttType:	SFFloat (input)
	AttDesc:	The rotational potentiometer on the interface.

	Attribute:	force1
	AttType:	SFFloat (input)
	AttDesc:	The 1st force sensor on the interface.

	Attribute:	force2
	AttType:	SFFloat (input)
	AttDesc:	The 2nd force sensor on the interface.

	Attribute:	force3
	AttType:	SFFloat (input)
	AttDesc:	The 3th force sensor on the interface.

	Attribute:	force4
	AttType:	SFFloat (input)
	AttDesc:	The 4th force sensor on the interface.

	Attribute:	accelX
	AttType:	SFFloat (input)
	AttDesc:	The x-axis accelerometer on the interface.

	Attribute:	accelY
	AttType:	SFFloat (input)
	AttDesc:	The y-axis accelerometer on the interface.

	Attribute:	accelZ
	AttType:	SFFloat (input)
	AttDesc:	The z-axis accelerometer on the interface.

	Attribute:	slider1
	AttType:	SFFloat (input)
	AttDesc:	For controlling the instrument from a MIDI fader box.

	Attribute:	slider2
	AttType:	SFFloat (input)
	AttDesc:	For controlling the instrument from a MIDI fader box.

	Attribute:	slider3
	AttType:	SFFloat (input)
	AttDesc:	For controlling the instrument from a MIDI fader box.

	Attribute:	slider4
	AttType:	SFFloat (input)
	AttDesc:	For controlling the instrument from a MIDI fader box.

	Attribute:	slider5
	AttType:	SFFloat (input)
	AttDesc:	For controlling the instrument from a MIDI fader box.

	Attribute:	slider6
	AttType:	SFFloat (input)
	AttDesc:	For controlling the instrument from a MIDI fader box.

	Attribute:	slider7
	AttType:	SFFloat (input)
	AttDesc:	For controlling the instrument from a MIDI fader box.

	Attribute:	slider8
	AttType:	SFFloat (input)
	AttDesc:	For controlling the instrument from a MIDI fader box.

	Attribute:	slider9
	AttType:	SFFloat (input)
	AttDesc:	For controlling the instrument from a MIDI fader box.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY" headlight="FALSE"/&gt;

		&lt;PaintingBackground DEF="backg"/&gt;

		&lt;Transform DEF="temp" translation="0.0 0.0 0.0"&gt;
			&lt;DirectionalLight DEF="lightDir" color="1.0 1.0 1.0" direction="-0.25 0.5 1.0"/&gt;
			&lt;DirectionalLight DEF="lightDir2" color="0.25 0.4 0.25" direction="0.0 -1.0 1.0"/&gt;

			&lt;Shape&gt;
				&lt;Ashitaka DEF="bob"/&gt;
			&lt;/Shape&gt;
		&lt;/Transform&gt;

		&lt;ROUTE fromNode="bob" fromField="quadsPosCol" toNode="backg" toField="splodgePosCol"/&gt;
		&lt;ROUTE fromNode="bob" fromField="wobble1" toNode="backg" toField="bDStretch"/&gt;
		&lt;ROUTE fromNode="bob" fromField="wobble2" toNode="backg" toField="fDStretch"/&gt;

	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	The Ashitaka audiovisual instrument.
/*!
	\ingroup libheilanextras
 */
class Ashitaka : public ChildNode
{
  public:
	///	Constructor.
	Ashitaka();
	///	Destructor.
	~Ashitaka();

	///	To initialise some stuff.
	void initialise();

	///	Draws the instrument's visuals.
	void graphicDoStuff();
	///	Calculates the audio from the various parts of the instrument,
	AmbisonicData *getAmbisonicAudio(int numSamples, const ThreeFloats& pos);

	///	Overridden to do nothing (for now).
	void setBlockSize(int newSize);
	///	Called to update the samplerate.
	void setSamplerate(float val);

	///	Used to set this node's attributes.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<Ashitaka> constructor;
  private:
	///	Helper method to set up the mappings (fader box).
	void setupMappingsMIDI();
	///	Helper method to set up the mappings (interface).
	void setupMappingsInterface();

	///	The (physical) interface for the instrument.
	AshitakaClasses::Interface interface;
	///	The audio synth for the instrument.
	AshitakaClasses::AudioSynth audioSynth;
	///	The visual synth for the instrument.
	//AshitakaClasses::VisualSynth visualSynth;

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;

	///	Handles the Quads and GravityObjects.
	AshitakaClasses::QuadManager quadManager;

	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;

	///	Counter used to determine when to update the PaintingBackground.
	int paintCount;

	///	Used to convert x-axis accelerometer data to x position.
	Accel2Pos xAxisConvert;
	///	Used to convert y-axis accelerometer data to y position.
	Accel2Pos yAxisConvert;
	///	Used to convert z-axis accelerometer data to z position.
	Accel2Pos zAxisConvert;

	///	Used to scale Excitation before it's sent to audioSynth.pluckTrigger.
	Map1InMultConst pluckTriggerScale;
	///	Used to scale Excitation so the bowing doesn't get too loud.
	Map1InMultConst exciteScale;
	///	Used to sum and scale audioSynth.bowDisplacement and Excitation.
	Map2InSumMultConst bowDispExciteSum;
	///	Used to invert quadManager.numMainQuads.
	Map1InInvert numQuadsInvert;
	///	Used to invert signal going to explode.
	Map1InGtThanNormal explodeGtThan;
	///	Used to invert signal going to explode.
	Map1InInvert explodeInvert;
	///	Used to sum transientDetector and Connectedness.
	Map2InSum explodeSum;
	///	Used to sum Connectedness and quadManager.mainConsistency.
	Map2InSumMultConst connectSum;
	///	Used to sum Connectedness and quadManager.mainConsistency.
	Map2InSumMultConst connectSum2;
	///	Used to invert signal going to damping.
	Map1InInvert dampInvert;
	///	Used to invert signal going to damping.
	Map1InInvert dampInvert2;
	///	Used to multiply signal going to damping.
	Map1InMultConst dampMult;
	///	Used to make sure the value goes 0->0->1.
	Map1InGtThanNormal dampGtThan;
	///	Used to make sure the value goes 0->0->1.
	Map1InGtThanNormal dampGtThan2;
	///	Used to make sure the value goes 0->0->1.
	Map1InGtThanNormal fbGtThan;
	///	Used to make sure the value goes 0->0->1.
	Map1InGtThanNormal fbGtThan2;
	///	Used to invert the signal going to quadManager.mainSize.
	Map1InInvert sizeInvert;
	///	Used to sum the signal going to quadManager.mainSize.
	Map1InSumConstNormal sizeSum;
	///	Used to scale the signal going to quadManager.mainSize.
	Map1InMultConst sizeMult;
	///	Used to sum and scale the 2 visual inputs to audioSynth.pitch.
	Map2InSumMultConst pitchSum1;
	///	Used to sum and scale pitchSum1 and the input to audioSynth.pitch.
	Map2InSumMultConst pitchSum2;
};
	
}

#endif
