//	AudioSynth.cpp - The audio synthesizer part of Ashitaka.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "AudioSynth.h"

#include <iostream>

using namespace std;

namespace AshitakaClasses
{

//------------------------------------------------------------------------------
AudioSynth::AudioSynth():
stopPos(0.0f),
feedbackPos(0.25f),
feedbackAmp(0.0f),
dampingValue(0.0f),
dampingLeft(0.0f),
dampingRight(0.25f),
bowForce(0.0f),
bowVelocity(0.0f),
bowPosition(7.0f*(1.0f/8.0f)),
samplerate(44100.0f),
lastPluck(1.0f),
bowMode(false),
bowPointPosition(0.0f),
bowForceExerted(0.0f),
lastStringFreq(110.0f),
pluckAmp(0.0f),
canPluck(true),
pluckTrans(false),
pluckAmount(0.0f),
pluckAvoidance(0),
dcSpeed(0.0f),
dcPos(0.0f),
transEnv(0.0f),
rmsWrite(0),
plucked(0.0f),
adWrite(0),
triggerBufCount(0)
{
	int i;

	for(i=0;i<88200;++i)
		audioData[i] = 0.0f;

	for(i=0;i<ADInputSize;++i)
		adInputs[i] = 0.0f;

	pluckTrigger.registerCallback(*this);

	//Setup delay parameters.
	/*delayTime >> delay.time;
	delayFeedback >> delay.feedback;
	delayMix >> delay.mix;
	delay.scrubDir >> delayScrub;*/

	//Calculate audio stuff.
	for(i=0;i<StringLength;++i)
	{
		positions[i] = 0.0f;
		forces[i] = 0.0f;
		velocities[i] = 0.0f;
	}

	//Calculate cells' mass.
	float defaultVelocityMultiplier;
	float decay = 5.0f; //5s?
	float intendedFreq = 110.0f;
	float actualFreq;
	float compensationFactor;
	defaultVelocityMultiplier = (1.0f-((.00012f*44100.0f)/samplerate)/decay);
	actualFreq = (24000.0f*samplerate/44100.0f)/(StringLength+1);
	compensationFactor=powf(4.0f, log10f(actualFreq/intendedFreq)/log10f(2.0f));
	mass = 3.5f * compensationFactor;
	invMass = 1.0f/mass;

	velocityMultiplier = defaultVelocityMultiplier;
	stopMultiplier = 1.0f-(powf(10000.0f, 0.95f)/10000.0f);
	dampMultiplier = 1.0f-(powf(10000.0f, dampingValue)/10000.0f);

	//Clear rmsBuffer.
	for(i=0;i<RmsSize;++i)
	{
		rmsBuffer[i] = 0.0f;
		loBuffer[i] = 0.0f;
		midBuffer[i] = 0.0f;
		hiBuffer[i] = 0.0f;
	}

	loFilter.f = 2.0f * sinf(3.141592653f * 0.25f * 0.33f);
	midFilter.f = 2.0f * sinf(3.141592653f * 0.25f * 0.5f);
	midFilter.q = 0.5f;
	hiFilter.f = 2.0f * sinf(3.141592653f * 0.25f * 0.66f);
}

//------------------------------------------------------------------------------
AudioSynth::~AudioSynth()
{
	
}

//------------------------------------------------------------------------------
void AudioSynth::valueChanged(float val)
{
	int i, first, last;
	float min = 1.0f;
	float max = 0.0f;

	for(i=7;i>0;--i)
		pluckVals[i] = pluckVals[i-1];
	pluckVals[0] = val;

	first = 0;
	last = 0;
	for(i=0;i<8;++i)
	{
		if(pluckVals[i] < min)
		{
			min = pluckVals[i];
			last = i;
		}
		if(pluckVals[i] > max)
		{
			max = pluckVals[i];
			first = i;
		}
	}

	if(((max-min) > 0.5f) && (last > first))
	{
		pluckTrans = true;
		pluckAmount = ((max-min)-0.5f) * 2.0f;
	}
}

//------------------------------------------------------------------------------
void AudioSynth::getAudioBlock(AmbisonicData *buffer, int bufferSize)
{
	int i, j;
	float tempf = 0.0f;
	float tempL, tempM, tempH;
	float tempPitchf;
	int tempPitchi;
	float stringFreq = (pitch * 165.0f)+55.0f;
	float tempTrans = transientDetector;
	float tempInstant = instantAmplitude;
	map<unsigned int, float>::iterator triggerIt;
	//float dampExtra = 0.0f;
	int overDampCount = 0;

	if(stringFreq > 220.0f)
		stringFreq = 220.0f;

	//hiPassOutput.f = 2.0f * sinf(3.141592653f * hiFilterCutoff);
	dampMultiplier = 1.0f-(powf(10000.0f, dampingValue)/10000.0f);

	//Calculate audio, put in buffer.
	for(i=0;i<bufferSize;++i)
	{
		//Update string frequency if necessary.
		if(stringFreq != lastStringFreq)
		{
			float actualFreq;
			float compensationFactor;

			actualFreq = (24000.0f*samplerate/44100.0f)/(StringLength+1);
			compensationFactor=powf(4.0f, log10f(actualFreq/stringFreq)/log10f(2.0f));
			mass = 3.5f * compensationFactor;
			invMass = 1.0f/mass;

			lastStringFreq = stringFreq;
		}

		//Calculate monophonic audio.
		tempf = 0.0f;

		//Calculate forces.
		{
			for(j=1;j<(StringLength-1);++j)
			{
				forces[j] = (positions[j-1] + positions[j+1]) -
							(2.0f * positions[j]);
			}
			forces[0] = forces[1] - forces[0];
			forces[StringLength-1] = forces[StringLength-2] -
									 forces[StringLength-1];
		}

		//Update pitch stuff.
		{
			tempPitchf = stopPos * (float)(StringLength-1);
			tempPitchi = (int)tempPitchf;
		}
		//Add feedback.
		if(feedbackAmp > 0.0f)
		{
			int feedPos = (int)(feedbackPos * (float)(StringLength-1));
			positions[7*(StringLength/8)] += (float)tanh((positions[feedPos]/100.0f) *
														 feedbackAmp * 3.1415f) *
											 0.5f;
		}

		//Calculate plucking.
		triggerIt = triggerBuffer.find(triggerBufCount);
		if(triggerIt != triggerBuffer.end())
			triggerBuffer.erase(triggerIt);
		if(!pluckAvoidance && pluckTrans && canPluck)
		{
			pluckAmp = pluckAmount + 1.0f;
			canPluck = false;
			plucked = 1.0f;
			pluckTrans = false;
			triggerBuffer.insert(make_pair(triggerBufCount, pluckAmount));
			pluckAvoidance = 1200;
		}
		else if(/*!pluckAvoidance && */!pluckTrans && !canPluck)
			canPluck = true;

		if(pluckAvoidance > 0)
		{
			--pluckAvoidance;
			pluckTrans = false;
		}

		if(pluckAmp > 0.01f)
		{
			int pPos = (int)(pluckPos * (float)(StringLength-1));
			for(j=0;j<pPos;++j)
				forces[j] += (pluckAmp * ((float)j/(float)pPos));
			for(j=pPos;j<(StringLength-1);++j)
				forces[j] += (pluckAmp * (1.0f-((float)(j-pPos)/(127.0f-(float)pPos))));
			//forces[pluckPos] += pluckAmp;
			pluckAmp = pluckAmp - 0.5f;
		}
		else
			pluckAmp = 0.0f;

		++triggerBufCount;
		if(triggerBufCount > TriggerBufferSize)
			triggerBufCount = 0;

		//Calculate bow stuff.
		if((bowForce > 0.01f) && (bowVelocity > 0.01f))
		{
			float relativeVelocity;
			int bPos = (int)(bowPosition * (float)(StringLength-1));

			bowPointPosition += bowVelocity;
			relativeVelocity = bowVelocity - velocities[bPos];

			if(!bowMode)
			{
				float stickingForce = bowPointPosition - positions[bPos];

				if(stickingForce > bowForce)
					bowMode = true;
				else
					bowForceExerted = stickingForce;
			}
			else
			{
				float slippingForce = bowForce/(1.0f+fabsf(relativeVelocity));

				if(velocities[bPos] >= 0.0f)
				{
					bowMode = false;
					bowPointPosition = positions[bPos];
				}
				else
					bowForceExerted = slippingForce;
			}

			forces[bPos] += bowForceExerted;
			if(bPos > 1)
				forces[bPos-1] += bowForceExerted * 0.5f;
			if(bPos < (StringLength-2))
				forces[bPos+1] += bowForceExerted * 0.5f;
		}

		//Calculate positions.
		int dLeft = (int)(dampingLeft * (float)(StringLength-1));
		int dRight = (int)(dampingRight * (float)(StringLength-1));
		for(j=1;j<(StringLength-1);++j)
		{
			velocities[j] += forces[j] * invMass;
			if((j >= dLeft) && (j < dRight))
				velocities[j] *= dampMultiplier;
			else if(j > tempPitchi)
				velocities[j] *= velocityMultiplier;
			else
				velocities[j] *= stopMultiplier;
			positions[j] += velocities[j];
			positions[j] += DenormalFix;
			positions[j] -= DenormalFix;

			if(!((j+overDampCount) % 4))
			//if(!(j%4))
			{
				//This causes major headaches on Windows if we do it every cell.
				if((fabs(positions[j]) > 10.0f) && (feedbackAmp > 0.01f))
					velocities[j] = 0.0f;
			}
			++overDampCount;
			if(overDampCount > 3)
				overDampCount = 0;
		}

		//Calculate audio.
		tempf = positions[3*(StringLength/4)];


		buffer[i].W = (tempf * 0.5f)/100.0f;

		//Apply hi-pass filter to output.
		/*hiPassOutput = buffer[i].W;
		buffer[i].W = hiPassOutput.hi;*/

		//Apply dc filter.
		dcSpeed += (buffer[i].W - dcPos) * 0.000004567f;
		dcPos += dcSpeed;
		dcSpeed *= 0.96f;
		buffer[i].W = buffer[i].W - dcPos;

		//Update audioData.
		//(the adInputs buffer is an attempt to avoid bad clicks when we read
		// past the write position)
		/*audioData[adWrite] = adInputs[ADInputSize-1];
		for(j=(ADInputSize-1);j>0;--j)
			adInputs[j] = adInputs[j-1];
		adInputs[0] = buffer[i].W;*/
		audioData[adWrite] = buffer[i].W;
		++adWrite;
		if(adWrite >= 88200)
			adWrite = 0;
	}

	//Apply delay to calculated audio.
	//delay.processAudio(buffer, bufferSize);

	//Calculate transients etc.
	for(i=0;i<bufferSize;++i)
	{
		tempf = fabs(buffer[i].W);

		//Calculate transients and instantaneous amplitude.
		if(tempf > transEnv)
		{
			if((tempf-transEnv) > 0.15f)
				tempTrans = 1.0f;

			transEnv = tempf;
		}
		else
		{
			if(transEnv > 0.0f)
				transEnv -= 0.005f;
			if(tempTrans > 0.0f)
				tempTrans -= 0.0001f;
		}
		tempInstant = transEnv;

		//Update rms buffer.
		rmsBuffer[rmsWrite] = transEnv;

		//Update loBuffer.
		loFilter = buffer[i].W;
		loBuffer[rmsWrite] = loFilter.lo;
		//Update midBuffer.
		midFilter = buffer[i].W;
		midBuffer[rmsWrite] = midFilter.band;
		//Update hiBuffer.
		hiFilter = buffer[i].W;
		hiBuffer[rmsWrite] = hiFilter.hi;

		++rmsWrite;
		if(rmsWrite >= RmsSize)
			rmsWrite = 0;
	}

	//transientDetector = fabs(transEnv-lastTrans) > 0.5f;
	if(plucked > 0.0f)
	{
		plucked -= 0.03f;
	}
	transientDetector = plucked;
	lastTrans = transEnv;
	instantAmplitude = tempInstant;

	tempf = 0.0f;
	tempL = tempM = tempH = 0.0f;
	for(i=0;i<RmsSize;++i)
	{
		tempf += (rmsBuffer[i] * rmsBuffer[i]);
		tempL += (loBuffer[i] * loBuffer[i]);
		tempM += (midBuffer[i] * midBuffer[i]);
		tempH += (hiBuffer[i] * hiBuffer[i]);
	}
	tempf /= (float)RmsSize;
	rmsAmplitude = sqrtf(tempf);
	tempL /= (float)RmsSize;
	rmsLo = sqrtf(tempL);
	tempM /= (float)RmsSize;
	rmsMid = sqrtf(tempM);
	tempH /= (float)RmsSize;
	rmsHi = sqrtf(tempH);

	bowDisplacement = 0.0f;
	bowDisplacement = bowPointPosition * 0.005f;
}

//------------------------------------------------------------------------------
void AudioSynth::setSamplerate(float val)
{
	samplerate = val;
}

//------------------------------------------------------------------------------
float *AudioSynth::getAudioData()
{
	int i/*, j*/;
	//float tempf;
	float *retval = new float[88200];

	//I'm ignoring threading issues.  Does it matter?
	for(i=0;i<88200;++i)
		retval[i] = audioData[i];

	//To try and avoid clicks at the write position.
	/*for(i=(ADInputSize-2),j=adWrite;i>=0;--i)
	{
		tempf = (float)i/(float)(ADInputSize-2);
		retval[j] *= (1.0f-tempf);
		retval[j] += adInputs[i] * tempf;
		++j;
		if(j >= 88200)
			j = 0;
	}*/

	return retval;
}

}
