//	AudioSynth.h - The audio synthesizer part of Ashitaka.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef AUDIOSYNTH_H_
#define AUDIOSYNTH_H_

#include "Parameter.h"
#include "AmbisonicData.h"
#include "AudioSynthDelay.h"

#include <map>

namespace AshitakaClasses
{

//------------------------------------------------------------------------------
///	The audio synthesizer part of Ashitaka.
class AudioSynth : public ParameterCallback
{
  public:
	///	Constructor.
	AudioSynth();
	///	Destructor.
	~AudioSynth();

	///	Used to fill pluckVals from pluckTrigger.
	void valueChanged(float val);

	///	Returns the current block of (monophonic) audio from the synth.
	/*!
		\param buffer The buffer to write audio to (only the W component is
		used).
		\param bufferSize The number of samples to write to buffer.
	 */
	void getAudioBlock(AmbisonicData *buffer, int bufferSize);

	///	Updates us with the current samplerate.
	void setSamplerate(float val);

	///	Copies audioData to a new buffer and returns it.
	float *getAudioData();
	///	Returns the triggerBuffer.
	const std::map<unsigned int, float>& getTriggerData() const {return triggerBuffer;};
	///	Returns the current write point within triggerBuffer.
	unsigned int getTriggerWritePoint() const {return triggerBufCount;};

	//--Input--
	///	Stop position parameter.
	Parameter stopPos;
	///	Pluck amplitude parameter.
	//Parameter pluckAmp;
	///	Pluck position parameter.
	Parameter pluckPos;
	///	Pluck trigger parameter.
	Parameter pluckTrigger;
	///	Feedback position parameter.
	Parameter feedbackPos;
	///	Feedback amplitude parameter.
	Parameter feedbackAmp;
	///	Damping amount parameter.
	Parameter dampingValue;
	///	Left damping position parameter.
	Parameter dampingLeft;
	///	Right damping position parameter.
	Parameter dampingRight;
	///	Bow force parameter.
	Parameter bowForce;
	///	Bow velocity parameter.
	Parameter bowVelocity;
	///	Bow position parameter.
	Parameter bowPosition;
	///	Frequency of the string.
	Parameter pitch;

	///	Hi pass filter cutoff.
	Parameter hiFilterCutoff;

	//X-axis position.
	Parameter positionX;
	//Y-axis position.
	Parameter positionY;
	//Z-axis position.
	Parameter positionZ;

	///	Delay time parameter.
	Parameter delayTime;
	///	Delay feedback parameter.
	Parameter delayFeedback;
	///	Delay mix parameter.
	Parameter delayMix;

	//--Output--
	///	Transient detector output.
	Parameter transientDetector;
	///	Instantaneous amplitude output.
	Parameter instantAmplitude;
	///	RMS amplitude output.
	Parameter rmsAmplitude;
	///	How much the bow is displaced.
	Parameter bowDisplacement;
	///	Low freq RMS output.
	Parameter rmsLo;
	///	Mid freq RMS output.
	Parameter rmsMid;
	///	High freq RMS output.
	Parameter rmsHi;
	///	Delay scrub direction output.
	//Parameter delayScrub;
  private:
	///	Simple state variable filter struct.
	struct Filter
	{
		float f;
		float q;
		float lo;
		float band;
		float hi;
		float notch;

		///	Constructor.
		Filter():
		f(0.0f),
		q(1.0f),
		lo(0.0f),
		band(0.0f),
		hi(0.0f),
		notch(0.0f)
		{
			
		};

		///	Updates the filter with a new input.
		Filter& operator=(float val)
		{
			lo = lo + f * band;
			hi = q * val - lo - q * band;
			band = f * hi + band;
			notch = hi + lo;

			return *this;
		};
	};

	///	Various constants.
	enum
	{
		StringLength = 128,
		RmsSize = 2205,
		ADInputSize = 16,
		TriggerBufferSize = 176400
	};

	///	The current samplerate.
	float samplerate;

	///	Delay line for the string (i.e. the positions of the cells).
	float positions[StringLength];
	///	Forces for the string cells.
	float forces[StringLength];
	///	Velocities for the string cells.
	float velocities[StringLength];
	///	Mass of the individual cells.
	float mass;
	///	Inverse mass of the individual cells.
	float invMass;
	///	How much to multiply the cells' velocities by (default).
	float velocityMultiplier;
	///	How much to multiply the cells' velocities by (stopped).
	float stopMultiplier;
	///	How much to multiply the cells' velocities by (damped).
	float dampMultiplier;

	///	Position of the stop.
	//float stopPos;
	///	Current pluck amplitude.
	//float pluckAmp;
	///	Last pluck value (only pluck when value initially goes past 50%).
	float lastPluck;
	///	Position on the string to get feedback from.
	//int feedbackPos;
	///	Amplitude of the feedback.
	//float feedbackAmp;
	///	Damping value.
	//float dampingValue;
	///	Damping left position.
	//int dampingLeft;
	///	Damping right position.
	//int dampingRight;
	///	Pluck position.
	//int pluckPos;

	///	Bow force.
	//float bowForce;
	///	Bow velocity.
	//float bowVelocity;
	///	Bow position.
	//int bowPosition;
	///	Bow mode.
	bool bowMode;
	///	Position (y-axis) of the bow.
	float bowPointPosition;
	///	The actual force exerted by the bow.
	float bowForceExerted;

	///	The current string frequency.
	float lastStringFreq;
	///	The new string frequency.
	//float stringFreq;

	///	Plucking amplitude.
	float pluckAmp;
	///	Used to determine when to pluck the string.
	bool canPluck;
	///	Used to detect transients in pluckTrigger.
	float pluckVals[8];
	///	True if we need to pluck the string.
	bool pluckTrans;
	///	How hard to pluck.
	float pluckAmount;
	///	Used to avoid multiple plucks.
	int pluckAvoidance;

	///	The hi-pass filter on the output.
	Filter hiPassOutput;

	///	For the dc filter on the output.
	float dcSpeed;
	///	For the dc filter on the output.
	float dcPos;

	///	Envelope for the transient detector and instantaneous amplitude.
	float transEnv;
	///	Buffer for rmsAmplitude.
	float rmsBuffer[RmsSize];
	///	Write pointer to rmsBuffer.
	int rmsWrite;

	///	Filter for the loRMS output.
	Filter loFilter;
	///	Buffer for loRMS.
	float loBuffer[RmsSize];
	///	Filter for the midRMS output.
	Filter midFilter;
	///	Buffer for midRMS.
	float midBuffer[RmsSize];
	///	Filter for the hiRMS output.
	Filter hiFilter;
	///	Buffer for hiRMS.
	float hiBuffer[RmsSize];

	///	Used to detect transients.
	float lastTrans;
	///	Used to detect plucks.
	float plucked;

	///	Delay effect applied to the string's output.
	AudioSynthDelay delay;

	///	Stores the last 2-seconds worth of audio from the instrument.
	float audioData[88200];
	///	Current write position within audioData.
	int adWrite;
	///	Stores the latest inputs to audioData (not used - didn't work anyway).
	float adInputs[ADInputSize];

	///	Stores the last 4-seconds worth of percussive triggers from the instrument.
	std::map<unsigned int, float> triggerBuffer;
	///	Counter for triggerBuffer.
	unsigned int triggerBufCount;
};

}

#endif
