//	CircleGO.cpp - Circle (multiple) GravityObject2.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "CircleGO.h"
#include "MiscHelper.h"

using namespace std;

namespace AshitakaClasses
{

//------------------------------------------------------------------------------
CircleGO::CircleGO(int circleIndex):
GravityObject2(),
index(circleIndex),
quadsSize(0),
radius(0.5f),
radBase(0.5f),
rotVal(0.0f),
firstRun(true),
pitchIndex(0),
pitchUpdate(true)
{
	float tempf;
	CircleExciter *excite = new CircleExciter(this, circleIndex);

	//tempf = 128.0f * (0.75f + ((float)index/24.0f));
	//tempf = 128.0f;
	if(index%2)
		tempf = 128.0f;
	else
		tempf = 64.0f;
	setExciter(excite);
	setStringLength(128);
	setInfluenceRadius(1.5f);
	setMainColour(ThreeFloats(0.0f, 0.25f, 0.0f));
	decay = 0.45f + (((float)index/5.0f) * 0.2f);
}

//------------------------------------------------------------------------------
CircleGO::~CircleGO()
{
	
}

//------------------------------------------------------------------------------
void CircleGO::updatePrimaryQuads()
{
	float angle;
	unsigned int i;
	ThreeFloats temp3f;
	vector<Quad2 *>& quads = getPrimaryQuads();

	if(firstRun)
	{
		centrePos = getPosition();

		angle = (float)index/6.0f;
		angle *= 6.283185306f;
		temp3f.x = sinf(angle) * 2.75f;
		temp3f.z = cosf(angle) * 2.75f;
		temp3f += centrePos;
		setPosition(temp3f);

		firstRun = false;
	}

	//Rotate GO and move with the noise envelope.
	angle = (float)index/6.0f;
	angle += rotVal;
	if(angle > 1.0f)
		angle -= (float)((int)angle);
	angle *= 6.283185306f;
	temp3f.x = sinf(angle) * 2.75f;
	temp3f.z = cosf(angle) * 2.75f;
	temp3f *= 1.0f+(radius-radBase);
	temp3f += centrePos;
	setPosition(temp3f);

	if(index%2)
	{
		if(pitchUpdate && ((radius-radBase) > 0.0f))
			pitchUpdate = false;
		else if(!pitchUpdate && ((radius-radBase) == 0.0f))
		{
			++pitchIndex;
			if(pitchIndex > 9)
				pitchIndex = 0;
			setPitch(getPitchContour()[pitchIndex]);

			pitchUpdate = true;
		}
	}
	else
		setPitch(1.0f);

	rotVal += 0.0005f;
	if(rotVal > 1.0f)
		rotVal -= 1.0f;

	//Move quads.
	for(i=0;i<quadsSize;++i)
		updateQuad(quads[i], 1.0f);
}

//------------------------------------------------------------------------------
void CircleGO::updateSecondaryQuad(Quad2 *quad)
{
	updateQuad(quad, 0.25f);
}

//------------------------------------------------------------------------------
void CircleGO::updateLifeForce()
{
	float life = getLifeForce();

	if(life > 0.0f)
		setLifeForce(life-0.0005f);
}

//------------------------------------------------------------------------------
void CircleGO::updateRadius(float val)
{
	radius = radBase + val;
}

//------------------------------------------------------------------------------
void CircleGO::numQuadsChanged()
{
	quadsSize = getPrimaryQuads().size();
}

//------------------------------------------------------------------------------
void CircleGO::updateQuad(Quad2 *quad, float weight)
{
	float tempf;
	float distance;
	ThreeFloats temp3f;
	ThreeFloats tempCol;
	ThreeFloats quadNewPos;
	const ThreeFloats origin;
	ThreeFloats quadPos = quad->getPosition();
	ThreeFloats goPos = getPosition();

	//Move quad along direction.
	temp3f = quad->getPosition()+(quad->getDirection() * (1.0f/32.0f));
	quad->moveTo(temp3f, weight);

	//Ensure we remain on the surface.
	temp3f = quad->getNewPosition();
	distance = temp3f.getDistance(goPos);
	if((distance < (radius-0.0005f)) || (distance > (radius+0.0005f)) ||
	   (temp3f.y < (goPos.y-0.0005f)) || (temp3f.y > (goPos.y+0.0005f)))
	{
		temp3f -= goPos;
		temp3f.y = 0.0f;
		temp3f *= (radius/distance);
		temp3f += goPos;
		quad->moveTo(temp3f, weight);
	}

	//Update direction.
	quadNewPos = quad->getNewPosition();
	temp3f = quadNewPos - goPos;
	distance = temp3f.getDistance(goPos);
	if(distance != 0.0f)
		temp3f /= distance;
	tempf = temp3f.x;
	temp3f.x = (tempf * cosf(1.570796327f)) - (temp3f.z * sinf(1.570796327f));
	temp3f.y = 0.0f;
	temp3f.z = (tempf * sinf(1.570796327f)) + (temp3f.z * cosf(1.570796327f));
	temp3f /= temp3f.getDistance(origin);
	quad->setDirection(temp3f);

	quad->setColour(ThreeFloats(0.0f, 0.5f, 0.0f), weight);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
CircleGO::CircleExciter::CircleExciter(CircleGO *go, int startingPoint):
Exciter(),
parentGO(go),
index(startingPoint),
counter(0.0f),
env(0.0f)
{
	addBuffer();
}

//------------------------------------------------------------------------------
CircleGO::CircleExciter::~CircleExciter()
{
	
}

//------------------------------------------------------------------------------
vector<GravityObject2::Exciter::Buffer> *CircleGO::CircleExciter::getAudio(int numSamples)
{
	int i;
	float tempf;
	float resoclip;
	unsigned int j;
	vector<Buffer>& buffers = getBuffers();

	if(parentGO)
		parentGO->updateRadius(env);

	for(j=0;j<buffers.size();++j)
		buffers[j].position = 0.5f;

	for(i=0;i<numSamples;++i)
	{
		if(index == 0)
		{
			if(counter < 0.5f)
				env = heilanExp(counter*2.0f);
			else
				env = heilanExp(1.0f-((counter-0.5f)*2.0f));

			filter.setCoeff((env * 950.0f)+50.0f, 15.0f);
			//cout << filter.b[0] << " : " << filter.b[1] << " : " << filter.b[2] << " : " << filter.b[3] << " : " << filter.b[4] << endl;
			tempf = heilanRand(heilanRandSeed());

			//Apply filter.
			resoclip = filter.b[3];
			if(resoclip > 1)
				resoclip = 1;
			else if(resoclip < -1)
				resoclip = -1;
			tempf = tempf - (resoclip * filter.res) - 1.0e-24f;
			filter.b[0] = ((tempf - filter.b[0]) * filter.cut) + filter.b[0];
			filter.b[1] = ((filter.b[0] - filter.b[1]) * filter.cut) + filter.b[1];
			filter.b[2] = ((filter.b[1] - filter.b[2]) * filter.cut) + filter.b[2];
			filter.b[3] = ((filter.b[2] - filter.b[3]) * filter.cut) + filter.b[3];
			tempf = filter.b[3] * env;
		}
		else
		{
			env = 0.0f;
			tempf = env;
		}

		counter += 0.00001f;
		if(counter > 1.0f)
		{
			index = (index+1)%6;
			counter = 0.0f;
		}

		tempf *= 0.125f;
		for(j=0;j<buffers.size();++j)
			buffers[j].audioData[i] = tempf + 1.0e-24f; //Denormal fix.
	}

	return &buffers;
}

}
