//	GravityObject.cpp - Virtual base class for all GravityObjects.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "GravityObject.h"
#include "Quad.h"
#include "Camera.h"

#include <algorithm>

using namespace std;

namespace AshitakaClasses
{

//------------------------------------------------------------------------------
GravityObject::GravityObject():
radius(10.0f),
lifeForce(1.0f),
lastAtten(0.0f)
{
	
}

//------------------------------------------------------------------------------
GravityObject::~GravityObject()
{
	
}

//------------------------------------------------------------------------------
void GravityObject::getAmbiAudio(AmbisonicData *buffer, int numSamples)
{
	int i;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;

	getMonoAudio(buffer, numSamples);

	if(buffer)
	{
		//Get current Ambisonic position (see AmbisonicData::operator=()),
		//relative to the camera.
		relativePos = position;
		relativePos -= HeilanX3D::Camera::getInstance()->getAudioPosition();
		lastAtten = calcRelativeAttenuation(lastPos);
		if(lastPos != relativePos)
		{
			ambiStart = lastPos;
			ambiEnd = relativePos;
			//We increment the position by a constant value, so we don't get
			//clicks or zipper noise.  Of course, there are disadvantages to
			//this approach...
			ambiInc = 0.005f;
			currentAmbi = ambiStart;
			newAtten = calcRelativeAttenuation(relativePos);
		}
		else
		{
			currentAmbi = relativePos;
			newAtten = lastAtten;
		}
		currentAtten = lastAtten;

		//Encode audio.
		for(i=0;i<numSamples;++i)
		{
			//Update currentAmbi and currentAtten if necessary.
			if(ambiInc > 0.0f)
			{
				//Calc currentAmbi.
				currentAmbi = ((1.0f-ambiIndex) * ambiStart) +
							  (ambiIndex * ambiEnd);
				//Calc currentAtten.
				currentAtten = ((1.0f-ambiIndex) * lastAtten) +
							   (ambiIndex * newAtten);

				ambiIndex += ambiInc;

				//Stop updating index if we've reached our destination.
				//I get weird rounding errors here, hence the .0000001f etc...
				if(ambiIndex > 1.0000001f)
				{
					ambiInc = 0.0f;
					currentAtten = newAtten;
				}
			}

			//Convert float sound file to Ambisonic, put it in the buffer.
			buffer[i] = currentAmbi * (buffer[i].W * currentAtten);
		}
		//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
		if(ambiIndex >= 1.0f)
		{
			lastPos = relativePos;
			lastAtten = currentAtten;
		}
		else
		{
			lastPos = ((1.0f-ambiIndex) * lastPos) + (ambiIndex * relativePos);
			lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
		}
	}
}

//------------------------------------------------------------------------------
void GravityObject::addPrimaryQuad(Quad *quad)
{
	primaryQuads.push_back(quad);
	quad->setPrimaryGravityObject(this);
	numQuadsChanged();
}

//------------------------------------------------------------------------------
void GravityObject::removePrimaryQuad(Quad *quad)
{
	vector<Quad *>::iterator it;

	it = find(primaryQuads.begin(), primaryQuads.end(), quad);

	if(it != primaryQuads.end())
	{
		primaryQuads.erase(it);
		numQuadsChanged();
	}
}

//------------------------------------------------------------------------------
void GravityObject::setLifeForce(float life)
{
	lifeForce = life;
}

//------------------------------------------------------------------------------
void GravityObject::setPosition(const ThreeFloats& pos)
{
	position = pos;
}

//------------------------------------------------------------------------------
void GravityObject::setInfluenceRadius(float val)
{
	radius = val;
}



//------------------------------------------------------------------------------
float GravityObject::calcRelativeAttenuation(const ThreeFloats& relPos)
{
	float mindist, camdist, attdist, dropoffdist;
	float retval;

	//min distance == 10
	//max distance == 50

	mindist = 10.0f;
	camdist = sqrtf((relPos.x * relPos.x)+
				    (relPos.y * relPos.y)+
				    (relPos.z * relPos.z));
	attdist = camdist - 10.0f;
	dropoffdist = 50.0f - 10.0f;

	if(camdist < mindist)
		retval = 1.0f;
	else
	{
		retval = -20.0f * (attdist/dropoffdist);
		retval = powf(10.0f, (retval*0.1f));
	}

	return retval;
}

}
