//	GravityObject.h - Virtual base class for all GravityObjects.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef GRAVITYOBJECT_H_
#define GRAVITYOBJECT_H_

#include "ThreeFloats.h"
#include "AmbisonicData.h"

#include <vector>
#include <map>

namespace AshitakaClasses
{

class Quad;

///	Virtual base class for all GravityObjects.
/*!
	A GravityObject is responsible for the positioning and motion of a number
	of Quads.  A GravityObject's <em>primary</em> Quads are those Quads which
	are primarily controlled by that GravityObject.  A Quad may only be assigned
	one primary GravityObject.  Secondary Quads are those Quads which have a
	different primary GravityObject, but which fall within this GravityObject's
	sphere of influence.  A Quad may have any number of secondary
	GravityObjects.  A Quad's motion is affected most by its primary
	GravityObject, with any secondary GravityObjects having a much lesser
	degree of influence.

	A GravityObject may also have a life force, which determines how long it
	will 'live' before it dies and its Quads are assigned a different primary
	GravityObject.

	\sa Quad
 */
class GravityObject
{
  public:
	///	Constructor.
	GravityObject();
	///	Destructor.
	virtual ~GravityObject();

	///	Called at the start of an update cycle.
	/*!
		This is used to reset the numSecondary variable.
	 */
	virtual void updateStart() {};
	///	Updates the positions of all this GO's primary quads.
	virtual void updatePrimaryQuads() = 0;
	///	Updates the position of the (passed-in) secondary quad.
	virtual void updateSecondaryQuad(Quad *quad) = 0;
	///	Updates this GO's lifeForce.
	virtual void updateLifeForce() = 0;

	///	Returns an ambisonically-encoded audio block for this GO.
	void getAmbiAudio(AmbisonicData *buffer, int numSamples);
	///	Returns a mono audio block for this GO.
	/*!
		This requires a mono audio stream, so only fill out the W component of
		the passed-in AmbisonicData block.
	 */
	virtual void getMonoAudio(AmbisonicData *buffer, int numSamples) = 0;
	///	Called when a GO is first split off.
	/*!
		\param data A 88200-sample buffer of audio data from the main
		instrument/AudioSynth class.  <em>It is our responsibility to delete
		it.</em>
	 */
	virtual void setAudioData(float *data) {delete [] data;};
	///	Called when a GO is first split off.
	/*!
		\param data A map that contains the sample at which a percussive
		trigger occurred, and the intensity of it.  This data is relative to
		the previous 2-seconds of performance.
		\param writePoint The position currently being written to in data.
	 */
	virtual void setTriggerData(const std::map<unsigned int, float>& data,
								unsigned int writePoint) {};

	///	Assigns this GravityObject to be the primary GravityObject of the Quad passed in.
	void addPrimaryQuad(Quad *quad);
	///	Removes the Quad from our list of primary Quads.
	void removePrimaryQuad(Quad *quad);

	///	Sets this GravityObject's life force.
	void setLifeForce(float life);
	///	Sets this GravityObject's position.
	void setPosition(const ThreeFloats& pos);

	///	Returns this GravityObject's life force.
	float getLifeForce() const {return lifeForce;};
	///	Returns this GravityObject's position.
	const ThreeFloats& getPosition() const {return position;};
	///	Returns this GravityObject's radius of influence.
	float getInfluenceRadius() const {return radius;};

	///	Returns the primary Quads for this GravityObject.
	std::vector<Quad *>& getPrimaryQuads() {return primaryQuads;};
  protected:
	///	Called when the number of primary Quads changes.
	virtual void numQuadsChanged() {};
	///	Sets the GO's influence radius.
	void setInfluenceRadius(float val);
  private:
	///	Helper method to return the audio attenuation relative to a point.
	/*!
		\param relPos This will usually be the position of the node relative to
		the camera.  This method will calculate the correct audio attenuation
		according to the node's distance from the camera, and the
		minFrontRadius, minBackRadius, maxFrontRadius and maxBackRadius
		variables.
	 */
	float calcRelativeAttenuation(const ThreeFloats& relPos);

	///	Vector of this GravityObject's primary Quads.
	std::vector<Quad *> primaryQuads;

	///	The position of this GravityObject.
	ThreeFloats position;
	///	The radius of influence of this GravityObject.
	float radius;
	///	This GravityObject's life force.
	float lifeForce;

	///	Used in getAmbiAudio().
	float lastAtten;
	///	Used in getAmbiAudio().
	ThreeFloats lastPos;
};
	
}

#endif
