//	GravityObject2.cpp - Second version of the GravityObject base class.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "GravityObject2.h"
#include "AmbisonicData.h"
#include "MiscHelper.h"
#include "Camera.h"

#include <algorithm>

using namespace std;
using namespace HeilanX3D;

namespace AshitakaClasses
{

//------------------------------------------------------------------------------
GravityObject2::GravityObject2():
exciter(0),
radius(0.5f),
lifeForce(1.0f),
lastAtten(0.0f),
forward(0),
backward(0),
size(128),
currentSize(128.0f),
desiredSize(128.0f),
backwardAllpass(0.0f),
forwardAllpass(0.0f),
stringOutput(0),
outputPos(96),
audioOutputPos(48),
pluckPos(24),
inputPos(32),
pluckIndex(0.0f),
blCoeff(0.25f),
blLast(0.0f),
clipping(false)
{
	int i;

	decay = 0.8f;
	setStringLength(size);

	for(i=0;i<10;++i)
		pitchContour[i] = 0.0f;
}

//------------------------------------------------------------------------------
GravityObject2::~GravityObject2()
{
	if(exciter)
		delete exciter;
	if(stringOutput)
		delete [] stringOutput;
	if(forward)
		delete [] forward;
	if(backward)
		delete [] backward;
}

//------------------------------------------------------------------------------
void GravityObject2::getAmbiAudio(AmbisonicData *buffer,
								  int numSamples,
								  const ThreeFloats& pos)
{
	int i;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;
	ThreeFloats position(xPos, yPos, zPos);
	const ThreeFloats camPos = Camera::getInstance()->getAudioPosition();
	const Quaternion& camRot = Camera::getInstance()->getAudioRotation();
	const float rotMult = 6.283185306f/360.0f;

	getMonoAudio(buffer, numSamples);

	if(buffer)
	{
		//Get current Ambisonic position (see AmbisonicData::operator=()),
		//relative to the camera.

		//So what we do here is:
		//1.) Ensure relativePos is equal to this node's spatial position in the
		//    scene.
		relativePos = pos + position;
		//2.) Rotate relativePos by the rotation of the camera, by multiplying
		//    it by the camera's orientation Quaternion.
		relativePos = camRot * relativePos;
		//3.) Translate relativePos by the position of the camera.
		relativePos -= camPos;

		lastAtten = calcRelativeAttenuation(lastPos);
		if(lastPos != relativePos)
		{
			ambiStart = lastPos;
			ambiEnd = relativePos;
			//We increment the position by a constant value, so we don't get
			//clicks or zipper noise.  Of course, there are disadvantages to
			//this approach...
			ambiInc = 0.005f;
			currentAmbi = ambiStart;
			newAtten = calcRelativeAttenuation(relativePos);
		}
		else
		{
			currentAmbi = relativePos;
			newAtten = lastAtten;
		}
		currentAtten = lastAtten;

		//Encode audio.
		for(i=0;i<numSamples;++i)
		{
			//Update currentAmbi and currentAtten if necessary.
			if(ambiInc > 0.0f)
			{
				//Calc currentAmbi.
				currentAmbi = (ambiStart * (1.0f-ambiIndex)) +
							  (ambiEnd * ambiIndex);
				//Calc currentAtten.
				currentAtten = (lastAtten * (1.0f-ambiIndex)) +
							   (newAtten * ambiIndex);

				ambiIndex += ambiInc;

				//Stop updating index if we've reached our destination.
				//I get weird rounding errors here, hence the .0000001f etc...
				if(ambiIndex > 1.0000001f)
				{
					ambiInc = 0.0f;
					currentAtten = newAtten;
				}
			}

			//Convert float sound file to Ambisonic, put it in the buffer.
			buffer[i] = currentAmbi * (buffer[i].W * currentAtten);
		}
		//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
		if(ambiIndex >= 1.0f)
		{
			lastPos = relativePos;
			lastAtten = currentAtten;
		}
		else
		{
			lastPos = (lastPos * (1.0f-ambiIndex)) + (relativePos * ambiIndex);
			lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
		}
	}
}

//------------------------------------------------------------------------------
void GravityObject2::getMonoAudio(AmbisonicData *buffer, int numSamples)
{
	int i;
	float tempf;
	int tempPos;
	float coeff;
	unsigned int j;
	int actualSize;
	float interpAmount;
	unsigned int tempSize;
	vector<Exciter::Buffer> *exciterBuffers = 0;

	critSec.lock();
	{
		//Get Exciter audio.
		if(exciter)
			exciterBuffers = exciter->getAudio(numSamples);

		//Fill sIAudio.
		sIAudio.clear();
		//This bit's causing problems...
		//(why was it causing problems?  have I fixed it?)
		tempSize = stringInputs.size();
		for(j=0;j<tempSize;++j)
			sIAudio.push_back(stringInputs[j]->getStringOutput());

		//Can you hear the resultant stepping from taking this out of the loop?
		/*if(currentSize < desiredSize)
		{
			currentSize += 0.01f;
			if(currentSize > desiredSize)
				currentSize = desiredSize;
		}
		else if(currentSize > desiredSize)
		{
			currentSize -= 0.01f;
			if(currentSize < desiredSize)
				currentSize = desiredSize;
		}
		tempf = currentSize;
		actualSize = static_cast<int>(tempf);
		interpAmount = tempf - static_cast<float>(actualSize);

		outputPos = (int)(currentSize * 0.75f);
		audioOutputPos = (int)(currentSize * 0.375f);
		pluckPos = (int)(currentSize * 0.1875f);
		inputPos = (int)(currentSize * 0.25f);

		if(actualSize < size)
			++actualSize;*/

		//Write audio to buffer.
		for(i=0;i<numSamples;++i)
		{
			//Handle pitch changes.
			//****Make this a non-linear interpolation?****
			if(currentSize < desiredSize)
			{
				currentSize += 0.01f;
				if(currentSize > desiredSize)
					currentSize = desiredSize;
			}
			else if(currentSize > desiredSize)
			{
				currentSize -= 0.01f;
				if(currentSize < desiredSize)
					currentSize = desiredSize;
			}
			tempf = currentSize;
			actualSize = static_cast<int>(tempf);
			interpAmount = tempf - static_cast<float>(actualSize);

			outputPos = (int)(currentSize * 0.75f);
			audioOutputPos = (int)(currentSize * 0.375f);
			pluckPos = (int)(currentSize * 0.1875f);
			inputPos = (int)(currentSize * 0.25f);

			if(actualSize < size)
				++actualSize;

			//Handle plucking.
			if(pluckIndex > 0.0f)
			{
				for(j=0;j<(unsigned)pluckPos;++j)
				{
					tempf = (float)j/(float)pluckPos;
					tempf *= pluckIndex * pluckIndex * 0.25f;
					forward[j] += tempf;
					backward[j] -= tempf;
				}
				for(j=pluckPos;j<(unsigned)actualSize;++j)
				{
					tempf = (float)(actualSize - j)/(float)(actualSize - pluckPos);
					tempf *= pluckIndex * pluckIndex * 0.25f;
					forward[j] += tempf;
					backward[j] -= tempf;
				}

				pluckIndex += 0.01f;
				if(pluckIndex >= 0.5f)
				{
					pluckIndex = 0.0f;
					//Go for a (hopefully) more interesting sound by randomising
					//the plucking position each time.
					pluckPos = (int)((heilanRand(heilanRandSeed()) + 1.0f)*(float)actualSize/2);
				}
			}

			//Handle Exciter excitation.
			if(exciterBuffers)
			{
				tempSize = exciterBuffers->size();
				for(j=0;j<tempSize;++j)
				{
					Exciter::Buffer& buf = (*exciterBuffers)[j];

					tempPos = (int)(buf.position * currentSize);
					forward[tempPos] += buf.audioData[i];
					backward[tempPos] -= buf.audioData[i];
				}
			}

			//Get input from other strings.
			tempSize = sIAudio.size();
			for(j=0;j<tempSize;++j)
			{
				tempf = sIAudio[j][i] * 0.8f;

				forward[inputPos] += tempf;
				backward[inputPos] -= tempf;
			}

			//Update string.
			//****This is seriously slowing things down!****
			for(j=(actualSize-1);j>0;--j)
			{
				forward[j] = forward[j-1];
				backward[j] = backward[j-1];
			}
			//Necessary?
			/*for(j=(size-1);j>=(unsigned)actualSize;--j)
			{
				forward[j] = 0.0f;
				backward[j] = 0.0f;
			}*/

			coeff = interpAmount;
			if(coeff < 0.5f)
				coeff += 1.0f;
			coeff = (1.0f-coeff)/(1.0f+coeff);

			//Nut wrap-around.
			//forward[0] = -backward[size-1];
			if(actualSize < (size-1))
			{
				tempf = (coeff * (backward[actualSize-1] - backwardAllpass));
				tempf += backward[actualSize-2];
				backwardAllpass = tempf;
			}
			else
				tempf = backward[actualSize-1];
			forward[0] = -tempf;

			//Bridge wrap-around.
			if(actualSize < (size-1))
			{
				tempf = (coeff * (forward[actualSize-1] - forwardAllpass));
				tempf += forward[actualSize-2];
				forwardAllpass = tempf;
			}
			else
				tempf = forward[actualSize-1];
			tempf = blLast + (1.0f-decay)*(tempf - blLast);
			backward[0] = -tempf;
			blLast = tempf;

			//Write outputs, tanh to avoid hard clipping.
			buffer[i].W = forward[audioOutputPos] + backward[audioOutputPos];
			if(fabs(buffer[i].W) > 0.66f)
				clipping = true;
			else
				clipping = false;
			buffer[i].W = heilanTanh(buffer[i].W * 3.0f) * 0.75f; //Not Pi because we want to attenuate it a bit.
			stringOutput[i] = buffer[i].W * 0.5f;

			stringOutput[i] = forward[outputPos] + backward[outputPos];
			stringOutput[i] = heilanTanh(stringOutput[i] * 3.0f) * 0.25f;
		}
	}
	critSec.unlock();
}

//------------------------------------------------------------------------------
void GravityObject2::setBlockSize(int newSize)
{
	int i;

	if(stringOutput)
		delete [] stringOutput;

	stringOutput = new float[newSize];

	for(i=0;i<newSize;++i)
		stringOutput[i] = 0.0f;

	if(exciter)
		exciter->setBlockSize(newSize);
}

//------------------------------------------------------------------------------
void GravityObject2::collide()
{
	pluckIndex = 0.01f;
}

//------------------------------------------------------------------------------
void GravityObject2::connectStrings(GravityObject2 *obj)
{
	critSec.lock();
	{
		stringInputs.push_back(obj);
	}
	critSec.unlock();
}

//------------------------------------------------------------------------------
void GravityObject2::removeString(GravityObject2 *obj)
{
	vector<GravityObject2 *>::iterator it;

	critSec.lock();
	{
		it = find(stringInputs.begin(), stringInputs.end(), obj);
		if(it != stringInputs.end())
			stringInputs.erase(it);
	}
	critSec.unlock();
}

//------------------------------------------------------------------------------
bool GravityObject2::stringIsConnected(GravityObject2 *obj) const
{
	vector<GravityObject2 *>::const_iterator it;

	it = find(stringInputs.begin(), stringInputs.end(), obj);
	return (it != stringInputs.end());
}

//------------------------------------------------------------------------------
void GravityObject2::drawStringConnections(bool lineColour) const
{
	unsigned int i;
	ThreeFloats temp3f;

	for(i=0;i<stringInputs.size();++i)
	{
		temp3f = stringInputs[i]->getPosition();

		if(lineColour)
		{
			glColor4f(1.0f, 1.0f, 1.0f, 0.25f);

			glLineWidth(4.0f);
			glBegin(GL_LINES);
			{
				glVertex3f(xPos, yPos, zPos);
				glVertex3f(temp3f.x, temp3f.y, temp3f.z);
			}
			glEnd();

			glLineWidth(2.0f);
			glBegin(GL_LINES);
			{
				glVertex3f(xPos, yPos, zPos);
				glVertex3f(temp3f.x, temp3f.y, temp3f.z);
			}
			glEnd();
		}
		else
		{
			glColor4f(0.0f, 0.0f, 0.0f, 0.25f);

			glLineWidth(4.0f);
			glBegin(GL_LINES);
			{
				glVertex3f(xPos, yPos, zPos);
				glVertex3f(temp3f.x, temp3f.y, temp3f.z);
			}
			glEnd();

			glLineWidth(2.0f);
			glBegin(GL_LINES);
			{
				glVertex3f(xPos, yPos, zPos);
				glVertex3f(temp3f.x, temp3f.y, temp3f.z);
			}
			glEnd();
		}

		glLineWidth(1.0f);
	}
}

//------------------------------------------------------------------------------
void GravityObject2::setLifeForce(float life)
{
	lifeForce = life;
}

//------------------------------------------------------------------------------
void GravityObject2::setPosition(const ThreeFloats& pos)
{
	//position = pos;
	xPos = pos.x;
	yPos = pos.y;
	zPos = pos.z;
}

//------------------------------------------------------------------------------
void GravityObject2::setGesture(const ThreeFloats * const gestureData)
{
	int i;

	for(i=0;i<10;++i)
		gesture[i] = gestureData[i];

	gestureUpdated();
}

//------------------------------------------------------------------------------
void GravityObject2::setPitches(const std::vector<float>& val)
{
	pitches = val;
}

//------------------------------------------------------------------------------
void GravityObject2::setPitchContour(const float * const val)
{
	int i;

	for(i=0;i<10;++i)
		pitchContour[i] = val[i];
}

//------------------------------------------------------------------------------
void GravityObject2::addPrimaryQuad(Quad2 *quad)
{
	quads.push_back(quad);
	numQuadsChanged();
}

//------------------------------------------------------------------------------
void GravityObject2::removePrimaryQuad(Quad2 *quad)
{
	vector<Quad2 *>::iterator it;

	it = find(quads.begin(), quads.end(), quad);

	if(it != quads.end())
	{
		quads.erase(it);
		numQuadsChanged();
	}
}

//------------------------------------------------------------------------------
void GravityObject2::setExciter(Exciter *excite)
{
	if(exciter)
		delete exciter;
	exciter = excite;
}

//------------------------------------------------------------------------------
void GravityObject2::setInfluenceRadius(float val)
{
	radius = val;
}

//------------------------------------------------------------------------------
void GravityObject2::setStringLength(int val)
{
	int i;

	if(forward)
		delete [] forward;
	if(backward)
		delete [] backward;

	desiredSize /= size;

	size = val;
	desiredSize *= size;
	if(currentSize > ((float)size-1.0f))
		currentSize = (float)size-1.0f;
	forward = new float[size];
	backward = new float[size];

	for(i=0;i<size;++i)
	{
		forward[i] = 0.0f;
		backward[i] = 0.0f;
	}

	outputPos = (int)((float)size * 0.75f);
	audioOutputPos = (int)((float)size * 0.375f);
	pluckPos = (int)((float)size * 0.1875f);
	inputPos = (int)((float)size * 0.25f);
}

//------------------------------------------------------------------------------
void GravityObject2::setPitch(float val)
{
	//const float lo = (44100.0f/440.0f);
	const float lo = (44100.0f/349.2282314330f);
	//const float hi = (44100.0f/880.0f);
	const float hi = (44100.0f/698.4564628660f);

	desiredSize = (val * (lo-hi))+hi;
}

//------------------------------------------------------------------------------
void GravityObject2::setMainColour(const ThreeFloats& col)
{
	mainColour = col;
}

//------------------------------------------------------------------------------
void GravityObject2::lockCritSec()
{
	critSec.lock();
}

//------------------------------------------------------------------------------
void GravityObject2::unlockCritSec()
{
	critSec.unlock();
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
GravityObject2::Exciter::Exciter():
blockSize(0)
{
	
}

//------------------------------------------------------------------------------
GravityObject2::Exciter::~Exciter()
{
	unsigned int i;

	//if(blockSize > 0)
	{
		for(i=0;i<buffers.size();++i)
			delete [] buffers[i].audioData;
	}
}

//------------------------------------------------------------------------------
void GravityObject2::Exciter::setBlockSize(int newSize)
{
	unsigned int i;

	if(blockSize)
	{
		for(i=0;i<buffers.size();++i)
			delete [] buffers[i].audioData;
	}

	blockSize = newSize;
	for(i=0;i<buffers.size();++i)
		buffers[i].audioData = new float[newSize];
}

//------------------------------------------------------------------------------
void GravityObject2::Exciter::addBuffer()
{
	buffers.push_back(Buffer());

	if(blockSize > 0)
		buffers[buffers.size()-1].audioData = new float[blockSize];

	numBuffersChanged(buffers.size());
}

//------------------------------------------------------------------------------
void GravityObject2::Exciter::removeBuffer(int index)
{
	if((unsigned)index >= buffers.size())
		return;

	if(blockSize)
		delete [] buffers[index].audioData;
	buffers.erase(buffers.begin()+index);

	numBuffersChanged(buffers.size());
}

}
