//	GravityObject2.h - Second version of the GravityObject base class.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef GRAVITYOBJECT2_H_
#define GRAVITYOBJECT2_H_

#include "Quad2.h"
#include "Parameter.h"
#include "CriticalSection.h"

#include <vector>

struct AmbisonicData;

namespace AshitakaClasses
{

///	Second version of the GravityObject base class.
class GravityObject2
{
  public:
	///	Constructor.
	GravityObject2();
	///	Destructor.
	virtual ~GravityObject2();

	///	Returns an ambisonically-encoded audio block for this GO.
	void getAmbiAudio(AmbisonicData *buffer,
					  int numSamples,
					  const ThreeFloats& pos);
	///	Returns a mono audio block for this GO.
	/*!
		This requires a mono audio stream, so only fill out the W component of
		the passed-in AmbisonicData block.
	 */
	void getMonoAudio(AmbisonicData *buffer, int numSamples);
	///	Sets the maximum number of samples processed in a block.
	void setBlockSize(int newSize);
	///	Returns a block of audio taken from the string model.
	/*!
		This is used when strings are connected to each other.
	 */
	const float * const getStringOutput() const {return stringOutput;};
	///	Called when 2 Quads collide, to pluck the string.
	void collide();

	///	Called to connect another GravityObject2's string model to this one.
	void connectStrings(GravityObject2 *obj);
	///	Called to remove a GravityObject2 from our list of connected ones.
	void removeString(GravityObject2 *obj);
	///	Returns true if we're already connected to obj.
	bool stringIsConnected(GravityObject2 *obj) const;

	///	Used to visualise connected strings.
	void drawStringConnections(bool lineColour) const;

	///	Updates the positions of all this GO's primary quads.
	virtual void updatePrimaryQuads() = 0;
	///	Updates the position of the (passed-in) secondary quad.
	virtual void updateSecondaryQuad(Quad2 *quad) = 0;
	///	Updates this GO's lifeForce.
	virtual void updateLifeForce() = 0;
	///	Can be used to update a subclass's Exciter.
	/*!
		This gets called after updatePrimaryQuads() and updateSecondaryQuad(),
		so if your Exciter relies on the position of your quads relative to
		your GO's position, you can tell it about them here.
	 */
	virtual void updateExciter() {};

	///	Sets this GravityObject's life force.
	void setLifeForce(float life);
	///	Sets this GravityObject's position.
	void setPosition(const ThreeFloats& pos);
	///	Sets this GravityObject's copy of the main GravityObject's previous path.
	void setGesture(const ThreeFloats * const gestureData);
	///	Sets the pitches this GravityObject should work with.
	void setPitches(const std::vector<float>& val);
	///	Sets this GravityObject's copy of the main GravityObject's previous pitches.
	void setPitchContour(const float * const val);

	///	Returns this GravityObject2's position in the space.
	const ThreeFloats getPosition() {return ThreeFloats(xPos, yPos, zPos);};
	//const ThreeFloats& getPosition() const {return position;};
	///	Returns this GravityObject2's radius of influence.
	float getInfluenceRadius() const {return radius;};
	///	Returns this GravityObject2's life force.
	float getLifeForce() const {return lifeForce;};
	///	Returns this GravityObject2's main colour (for the sparks...).
	const ThreeFloats& getMainColour() const {return mainColour;};
	///	Returns the main GravityObject's pitch contour, if we want to do something with it.
	float *getPitchContour() {return pitchContour;};

	///	Adds a Quad2 to our list of primary Quad2s.
	void addPrimaryQuad(Quad2 *quad);
	///	Removes the Quad from our list of primary Quads.
	void removePrimaryQuad(Quad2 *quad);

	///	Returns the primary Quads for this GravityObject.
	std::vector<Quad2 *>& getPrimaryQuads() {return quads;};
	///	Returns the number of primary Quads for this GravityObject.
	int getNumQuads() const {return quads.size();};

	//--Parameters--------------------------------------------------------------
	///	X-axis position.
	Parameter xPos;
	///	Y-axis position.
	Parameter yPos;
	///	Z-axis position.
	Parameter zPos;

	///	The decay value of the string.
	Parameter decay;
	//--------------------------------------------------------------------------

	//--------------------------------------------------------------------------
	///	Nested class used to excite GravityObject2's string model.
	/*!
		The idea is that subclasses of GravityObject2 will also subclass
		Exciter, so that each subclass excites the string in a different way.

		Exciter is nested inside GravityObject2 because I don't see any need to
		have it exposed to the outside world.  Presumably it won't be a problem
		inheriting from a nested class?
	 */
	class Exciter
	{
	  public:
		///	Constructor.
		Exciter();
		///	Destructor.
		virtual ~Exciter();

		//----------------------------------------------------------------------
		///	Struct representing a buffer, and it's position on the string.
		struct Buffer
		{
			///	The position on the string to apply this audio to.
			float position;
			///	The audio data in this buffer.
			float *audioData;
		};
		//----------------------------------------------------------------------

		///	Generates the excitation audio, returns it.
		virtual std::vector<Buffer> *getAudio(int numSamples) = 0;
		///	Updates the buffers' sizes.
		void setBlockSize(int newSize);

		///	Adds a buffer to buffers.
		void addBuffer();
		///	Removes the indexed buffer from buffers.
		void removeBuffer(int index);
		///	Returns the number of buffers.
		unsigned int getNumBuffers() const {return buffers.size();};
	  protected:
		///	Returns this class's buffers, for subclasses to use.
		std::vector<Buffer>& getBuffers() {return buffers;};

		///	Used to inform subclasses when the number of buffers changes.
		virtual void numBuffersChanged(int newSize) {};
	  private:
		///	Our excitation buffers.
		std::vector<Buffer> buffers;
		///	The current size of audioData in our Buffers.
		int blockSize;
	};
	//--------------------------------------------------------------------------
  protected:
	///	Called when the number of primary Quads changes.
	virtual void numQuadsChanged() {};
	///	Sets the GO's influence radius.
	void setInfluenceRadius(float val);

	///	Sets this object's Exciter.
	void setExciter(Exciter *excite);
	///	Returns this object's Exciter.
	Exciter *getExciter() const {return exciter;};
	///	Sets the length of the string (in cells).
	void setStringLength(int val);
	///	Sets the pitch of the string (0->1, w/1 being the string length).
	void setPitch(float val);

	///	Sets this GO2's main colour (for the sparks...).
	void setMainColour(const ThreeFloats& col);

	///	Returns whether or not we're clipping.
	bool getClipping() const {return clipping;};

	///	Called when setGesture is called.
	virtual void gestureUpdated() {};
	///	Returns the main GravityObject's path, if we want to do something with it.
	const ThreeFloats * const getGesture() const {return gesture;};
	///	Returns this GravityObject's pitches.
	const std::vector<float>& getPitches() const {return pitches;};

	///	Locks the CriticalSection surrounding the audio processing code.
	/*!
		This is necessary because subclasses may update Exciter data from the
		main (non-audio) thread.
	 */
	void lockCritSec();
	///	Unocks the CriticalSection surrounding the audio processing code.
	/*!
		\sa lockCritSec()
	 */
	void unlockCritSec();
  private:
	///	Helper method to return the audio attenuation relative to a point.
	/*!
		\param relPos This will usually be the position of the node relative to
		the camera.  This method will calculate the correct audio attenuation
		according to the node's distance from the camera, and the
		minFrontRadius, minBackRadius, maxFrontRadius and maxBackRadius
		variables.
	 */
	inline float calcRelativeAttenuation(const ThreeFloats& relPos)
	{
		float mindist, camdist, attdist, dropoffdist;
		float retval;

		//min distance == 10
		//max distance == 50

		mindist = 10.0f;
		camdist = sqrtf((relPos.x * relPos.x)+
						(relPos.y * relPos.y)+
						(relPos.z * relPos.z));
		attdist = camdist - 10.0f;
		dropoffdist = 50.0f - 10.0f;

		if(camdist < mindist)
			retval = 1.0f;
		else
		{
			retval = -20.0f * (attdist/dropoffdist);
			retval = powf(10.0f, (retval*0.1f));
		}

		return retval;
	};

	///	Helper method used in getAmbisonicAudio().
	inline void rotateAroundY(ThreeFloats& pos, float angle)
	{
		if(angle != 0.0f)
		{
			float tempX = pos.x;
			float sinAngle = sinf(angle);
			float cosAngle = cosf(angle);

			pos.x = (pos.x * cosAngle) + (pos.z * sinAngle);
			pos.z = -(tempX * sinAngle) + (pos.z * cosAngle);
		}
	};
	///	Helper method used in getAmbisonicAudio().
	inline void rotateAroundX(ThreeFloats& pos, float angle)
	{
		if(angle != 0.0f)
		{
			float tempY = pos.y;
			float sinAngle = sinf(angle);
			float cosAngle = cosf(angle);

			pos.y = (pos.y * cosAngle) - (pos.z * sinAngle);
			pos.z = (tempY * sinAngle) + (pos.z * cosAngle);
		}
	};

	///	Our primary Quad2s.
	std::vector<Quad2 *> quads;

	///	Our Exciter.
	Exciter *exciter;

	///	Our position in the space.
	//ThreeFloats position;
	///	The radius of influence of this GravityObject.
	float radius;
	///	This GravityObject's life force.
	float lifeForce;

	///	Used in getAmbiAudio().
	float lastAtten;
	///	Used in getAmbiAudio().
	ThreeFloats lastPos;

	//----String model parameters----
	///	Our forward delay line.
	float *forward;
	///	Our backward delay line.
	float *backward;
	///	The (total) size of our delay lines.
	int size;

	///	The current size of our delay lines.
	float currentSize;
	///	The delay line size we're moving towards (if we're not there already).
	float desiredSize;
	///	Allpass delay element for the nut wrap-around.
	float backwardAllpass;
	///	Allpass delay element for the bridge wrap-around.
	float forwardAllpass;

	///	Other GravityObject2s for us to receive input from.
	std::vector<GravityObject2 *> stringInputs;
	///	Temporary storage for the audio from stringInputs.
	std::vector<const float *> sIAudio;
	///	Our saved block of audio to pass on to connected GravityObject2s.
	float *stringOutput;

	///	The output position for other GravityObject2s.
	int outputPos;
	///	The output position for the audio.
	int audioOutputPos;
	///	The position for the plucking.
	int pluckPos;
	///	The inputs position from other StringModels.
	int inputPos;

	///	Used to calculate the plucking.
	float pluckIndex;

	///	Basic lowpass filter.
	float blCoeff;
	///	Basic lowpass filter.
	float blLast;

	///	True if the sound is getting close to hard-clipping levels.
	bool clipping;

	///	This GO2's main colour (for the sparks...).
	ThreeFloats mainColour;

	///	Our copy of the main GravityObject's previous path.
	ThreeFloats gesture[10];
	///	Pitches to work with (different GOs will use them differently).
	std::vector<float> pitches;
	///	Our copy of the main GravityObject's previous pitches.
	float pitchContour[10];

	///	CriticalSection protecting the audio processing code.
	CriticalSection critSec;
};

}

#endif
