//	Parameter.h - Represents a parameter in Ashitaka, which may be mapped to
//				  other Parameters.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PARAMETER_H_
#define PARAMETER_H_

#include "ParameterCallback.h"

#include <vector>
#include <iostream>

#define DenormalFix 1e-18f

/*!
	\defgroup AshitakaParameters Ashitaka Parameters and Mappings

	In order to try and simplify the process of mapping the various parameters
	of the Ashitaka instrument together, I've come up with the following scheme:

	Any parameter of a larger component (i.e. the interface, the visual
	synthesizer, the audio synthesizer or a mapping layer) which needs to be
	available to other components should be a Parameter (and a public member of
	the owning class).  Parameter can essentially be considered as a single
	float, with extra functionality.

	For example, when the value of a Parameter is changed, it will call a
	ParameterCallback (or multiple ParameterCallbacks), allowing other objects
	to be informed that the Parameter has changed.  Parameter itself is actually
	a subclass of ParameterCallback, allowing one Parameter to be mapped to
	another with the following code:
\code
Parameter param1;
Parameter param2;

param1 >> param2;
\endcode
	Now when param1's value is changed, it will also set param2's value, i.e.
\code
param1 = 1.0f;

cout << (float)param2 << endl;
//Console output: 1.0
\endcode

	To allow for more complex mappings, specific subclasses of Parameter can be
	used which transform their inputs before passing the result on as an output.
	e.g.
\code
Parameter in1;
Parameter in2;
Map2InSum mapping;
Parameter out;

//Used to inform mapping what its inputs are.
mapping.setInputs(in1, in2);

//The following lines set up the actual mappings.
in1 >> mapping;
in2 >> mapping >> out;

in1 = 1.0f;
cout << (float)out << endl;
//Console output: 1.0

in2 = 2.0f;
cout << (float)out << endl;
//Consolse output: 3.0
\endcode

For mapping classes which require multiple inputs, a setInputs method must be
used, as ParameterCallback::valueChanged() only provides a single float
parameter.  When a multiple input mapping class receives a valueChanged()
message then, it ignores the value passed in, and instead directly retrieves the
values of the Parameters set in setInputs in order to calculate its output.
 */

///	Represents a parameter in Ashitaka, which may be mapped to other Parameters.
/*!
	\ingroup AshitakaParameters
	\ingroup libheilanextras

	All parameters are floats, for simplicity.


	Any mapping classes (i.e. classes which transform their input in some way
	before outputting it) should derive from this class and override
	valueChanged() to perform the transformation in.

	e.g. you'd probably do something like: \code
//Assuming this mapping takes one input...
void TanhMapping::valueChanged(float val)
{
	//Map val to a tanh curve, use Parameter's valueChanged() to pass it on to
	//any ParameterCallbacks.
	Parameter::valueChanged(heilanTanh(val));
}\endcode
 */
class Parameter : public ParameterCallback
{
  public:
	///	Constructor.
 	Parameter(float initialVal = 0.0f);
	///	Destructor.
	virtual ~Parameter();

	///	Verbose method to change the Parameter's value.
	/*!
		\param val The new value of the Parameter.
	 */
	inline void changeValue(float val)
	{
		value = val;

		if(callbacks.size() > 0)
		{
			for(unsigned int i=0;i<callbacks.size();++i)
				callbacks[i]->valueChanged(value);
		}
	};
	///	Concise method to change the Parameter's value.
	/*!
		\param val The new value of the Parameter.
	 */
	inline Parameter& operator=(float val)
	{
		value = val;

		if(callbacks.size() > 0)
		{
			for(unsigned int i=0;i<callbacks.size();++i)
				callbacks[i]->valueChanged(value);
		}

		return *this;
	};
	///	Verbose method to return the value of Parameter.
	inline float getValue() const {return value;};
	///	Concise method to return the value of Parameter.
	inline operator float() const {return value;};

	///	Verbose method to register a callback with the Parameter.
	/*!
		\param callback The ParameterCallback to add to Parameter's vector of
		callbacks.
	 */
	void registerCallback(ParameterCallback &callback);
	///	Concise method to register a callback with the Parameter.
	/*!
		\param callback The ParameterCallback to add to Parameter's vector of
		callbacks.
	 */
	Parameter &operator>>(ParameterCallback& callback);
	///	Concise method to register a callback with the Parameter.
	/*!
		\param callback The ParameterCallback to add to Parameter's vector of
		callbacks.

		This version is so that you can string Parameters together, e.g.
\code
param1 >> param2 >> param3;
\endcode
	 */
	Parameter &operator>>(Parameter& p2);

	///	Gets called if another Parameter's registered to output to this Parameter.
	virtual void valueChanged(float val);
  private:
	///	The value of Parameter.
	float value;
	///	Parameter's vector of ParameterCallbacks.
	std::vector<ParameterCallback *> callbacks;
};

#endif
