//	PlanetGO.cpp - GravityOBject2-based planetary cloud-type gravity object.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "PlanetGO.h"
#include "MiscHelper.h"

#include <iostream>

using namespace std;

namespace AshitakaClasses
{

//------------------------------------------------------------------------------
PlanetGO::PlanetGO():
GravityObject2(),
ParameterCallback(),
planetPos(0.0f),
clipEnv(0.0f),
whichPitch(0),
holdPitch(false)
{
	PlanetExciter *excite = new PlanetExciter(this, quadsPos, planetPos);
	setExciter(excite);
	setStringLength(128);
	setInfluenceRadius(1.0f);
	setMainColour(ThreeFloats(1.0f, 1.0f, 0.25f));

	//We only use 4 buffers to save on the cpu consumption.
	excite->addBuffer();
	excite->addBuffer();
	excite->addBuffer();
	excite->addBuffer();
	quadsPos.push_back(ThreeFloats(0.0f));
	quadsPos.push_back(ThreeFloats(0.0f));
	quadsPos.push_back(ThreeFloats(0.0f));
	quadsPos.push_back(ThreeFloats(0.0f));

	//So we get informed (via our valueChanged() method) when they change.
	topLeftQuads >> *this;
	topRightQuads >> *this;
	bottomLeftQuads >> *this;
	bottomRightQuads >> *this;

	//Testing.
	topLeftQuads = 0.0f;
	topRightQuads = 0.0f;
	bottomLeftQuads = 0.0f;
	bottomRightQuads = 0.0f;
}

//------------------------------------------------------------------------------
PlanetGO::~PlanetGO()
{
	
}

//------------------------------------------------------------------------------
void PlanetGO::updatePrimaryQuads()
{
	unsigned int i;
	vector<Quad2 *>& quads = getPrimaryQuads();
	unsigned int quadsSize = quads.size();
	float forces[4];
	ThreeFloats tempCol;

	forces[0] = topLeftQuads;
	forces[1] = topRightQuads;
	forces[2] = bottomLeftQuads;
	forces[3] = bottomRightQuads;

	if(getClipping())
		clipEnv = 1.0f;
	else if(clipEnv > 0.0f)
		clipEnv -= 0.01f;

	for(i=0;i<quadsSize;++i)
	{
		updateQuad(quads[i], forces[i&3]);

		//Update the colour if we're clipping.
		//(because the use of forces in the line above screws things up in
		// updateQuad())
		if(clipEnv >= 0.0f)
		{
			tempCol = ThreeFloats(1.0f, 1.0f, 0.0f) * clipEnv;
			tempCol += ThreeFloats(0.0f) * (1.0f-clipEnv);
	
			quads[i]->setColour(tempCol, 1.0f);
		}
		else
		{
			tempCol = 0.0f;
			quads[i]->setColour(tempCol, 1.0f);
		}
	}
}

//------------------------------------------------------------------------------
void PlanetGO::updateSecondaryQuad(Quad2 *quad)
{
	updateQuad(quad, 0.25f);
}

//------------------------------------------------------------------------------
void PlanetGO::updateExciter()
{
	unsigned int i;
	vector<Quad2 *>& quads = getPrimaryQuads();

	lockCritSec();
	{
		planetPos = getPosition();

		//for(i=0;i<quads.size();++i)
		for(i=0;i<4;++i)
			quadsPos[i] = quads[i]->getPosition();
	}
	unlockCritSec();
}

//------------------------------------------------------------------------------
void PlanetGO::valueChanged(float val)
{
	int i;
	int tempint;
	float tempf = /*1.0f-*/pitch;
	const vector<float>& pitches = getPitches();
	float *contour = getPitchContour();

	if(pitches.size() == 0)
		return;

	if((tempf > 0.4f) && (tempf < 0.6f))
	{
		holdPitch = false;
		setPitch(pitches[whichPitch]/12.0f);
	}
	else if((tempf <= 0.45f) && (!holdPitch))
	{
		tempf *= 2.5f; //To get it w/in 0->1.
		tempint = whichPitch-1;
		if(tempint < 0)
			tempint = pitches.size()-1;
		tempf = (tempf * pitches[whichPitch]) + ((1.0f-tempf) * pitches[tempint]);
		setPitch(tempf/12.0f);
		if(pitch < 0.1f)
		{
			whichPitch = tempint;
			holdPitch = true;
			cout << "d: " << whichPitch << " : " << pitch << " : " << tempf << endl;

			for(i=0;i<9;++i)
				contour[i] = contour[i+1];
			contour[9] = tempf/12.0f;
		}
	}
	else if((tempf >= 0.6f) && (!holdPitch))
	{
		tempf -= 0.6f; //To get it w/in 0->0.4.
		tempf *= 2.5f; //To get it w/in 0->1.
		tempint = (whichPitch+1) % pitches.size();
		tempf = ((1.0f-tempf) * pitches[whichPitch]) + (tempf * pitches[tempint]);
		setPitch(tempf/12.0f);
		if(pitch > 0.95f)
		{
			whichPitch = tempint;
			holdPitch = true;
			cout << "u: " << whichPitch << " : " << pitch << " : " << tempf << endl;

			for(i=0;i<9;++i)
				contour[i] = contour[i+1];
			contour[9] = tempf/12.0f;
		}
	}

	/*float tempf;

	//Check if pitch should be locked.
	if((topLeftQuads+topRightQuads+bottomLeftQuads+bottomRightQuads) > 2.0f)
	{
		if(!pitchLocked)
		{
			atPitch = powf(2.0f, ((float)((int)(pitch*12.0f)) * (1.0f/12.0f)));
			atPitch *= 0.5f;
			atPitch = 1.0f-atPitch;
			pitchLocked = true;
		}
	}
	else
		pitchLocked = false;

	//Set the pitch.
	//tempf = ((1.0f-pitch)*0.5f)+0.5f;
	if(pitchLocked)
		tempf = atPitch;
	else
		tempf = 1.0f-pitch;
	setPitch(tempf);*/
}

//------------------------------------------------------------------------------
void PlanetGO::numQuadsChanged()
{
	/*unsigned int i;
	unsigned int numQuads = getPrimaryQuads().size();
	Exciter *excite = getExciter();

	if(numQuads < excite->getNumBuffers())
	{
		for(i=(excite->getNumBuffers()-1);i>=numQuads;--i)
			excite->removeBuffer(i);
	}
	else if(numQuads > excite->getNumBuffers())
	{
		for(i=excite->getNumBuffers();i<numQuads;++i)
			excite->addBuffer();
	}

	if(numQuads < quadsPos.size())
	{
		for(i=(quadsPos.size()-1);i>=numQuads;--i)
			quadsPos.pop_back();
	}
	else if(numQuads > quadsPos.size())
	{
		for(i=quadsPos.size();i<numQuads;++i)
			quadsPos.push_back(ThreeFloats(0.0f));
	}*/
}

//------------------------------------------------------------------------------
void PlanetGO::updateQuad(Quad2 *quad, float weight)
{
	float distance;
	ThreeFloats temp3f;
	ThreeFloats tempCol;
	const ThreeFloats origin;
	ThreeFloats goPos = getPosition();
	float radius = getInfluenceRadius();

	//Update the colour if we're clipping.
	if(clipEnv >= 0.0f)
	{
		tempCol = ThreeFloats(1.0f, 1.0f, 0.0f) * clipEnv;
		tempCol += ThreeFloats(0.0f) * (1.0f-clipEnv);

		quad->setColour(tempCol, weight);
	}
	else
	{
		tempCol = 0.0f;
		quad->setColour(tempCol, weight);
	}

	//So that the quads move slower when there's a long decay, and faster when
	//there's a very short one.
	weight *= ((decay-0.5f) * 2.0f);

	if(weight < 0.05f)
	{
		temp3f = quad->getPosition();
		distance = temp3f.getDistance(goPos);

		if(weight < 0.005f)
			weight = 0.005f;

		//Move quad along direction.
		if(distance > 0.0f)
		{
			quad->moveTo(quad->getPosition()+(quad->getDirection() *
											  (1.0f/distance)),
						 weight);
		}

		return;
	}

	//Move quad along direction.
	//quad->moveTo(quad->getPosition()+(quad->getDirection() * (1.0f/16.0f)),
	quad->moveTo(quad->getPosition()+(quad->getDirection() * (1.0f/64.0f)),
				 weight);

	//Ensure we remain on the surface.
	temp3f = quad->getNewPosition();
	distance = temp3f.getDistance(goPos);
	if((distance < (radius-0.0005f)) || (distance > (radius+0.0005f)))
	{
		temp3f -= goPos;
		temp3f *= (radius/distance);
		temp3f += goPos;
		quad->moveTo(temp3f, weight);
	}

	//Bring direction onto sphere's surface.
    //temp3f = quad->getNewPosition() + (quad->getDirection() * (1.0f/16.0f));
	temp3f = quad->getNewPosition() + quad->getDirection();
    distance = temp3f.getDistance(goPos);
    if((distance < (radius-0.0005f)) || (distance > (radius+0.0005f)))
    {
		temp3f -= goPos;
		//Bring it onto the surface.
		temp3f *= (radius/distance);
		temp3f += goPos;
		//Return it to within -1 -> 1.
		temp3f -= quad->getNewPosition();
		//temp3f.y = 0.0f; //Uncomment to use only horizontal motion.
		//Normalise it.
		//distance = temp3f.getDistance(goPos);
		distance = temp3f.getDistance(origin);
		temp3f /= distance;

 		quad->setDirection(temp3f);
    }
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
PlanetGO::PlanetExciter::PlanetExciter(PlanetGO *go,
									   const vector<ThreeFloats>& pos,
									   const ThreeFloats& goPos):
Exciter(),
quadsPos(pos),
planetPos(goPos),
parentGO(go)
{
	int i;
	float freq = 440.0f;

	sinesCoeff = 2.0f*sinf(3.141592653f*freq/44100.0f);

	for(i=0;i<4;++i)
	{
		freq += heilanRand(heilanRandSeed()) * 16.0f;
		prevFreqs[i] = freq;
		coeffs[i] = 2.0f*sinf(3.141592653f*freq/44100.0f);

		granAmp[i] = 0.0f;
		granDirection[i] = false;
		granCount[i] = (int)((heilanRand(heilanRandSeed())+1.0f)*2000.0f);
	}
}

//------------------------------------------------------------------------------
PlanetGO::PlanetExciter::~PlanetExciter()
{
	
}

//------------------------------------------------------------------------------
vector<GravityObject2::Exciter::Buffer> *PlanetGO::PlanetExciter::getAudio(int numSamples)
{
	int j;
	float amp;
	float x, y;
	float tempf;
	int tempint;
	unsigned int i;
	ThreeFloats temp3f;
	vector<Buffer>& buffers = getBuffers();
	//unsigned int bufSize = buffers.size();
	float invNumSines = 1.0f/(float)sines.size();
	float forces[4];
	float pitch = parentGO->pitch;
	float newFreq;

	forces[0] = parentGO->topLeftQuads;
	forces[1] = parentGO->topRightQuads;
	forces[2] = parentGO->bottomLeftQuads;
	forces[3] = parentGO->bottomRightQuads;

	invNumSines *= 4.0f;
	for(i=0;i<4;++i)
	{
		//Update pitch if necessary.
		newFreq = pitch * 12.0f;
		tempint = (int)newFreq;
		newFreq = (forces[i] * (float)tempint) + ((1.0f-forces[i]) * newFreq);
		//newFreq = 220.0f * powf(2.0f, (newFreq * (1.0f/12.0f)));
		newFreq = 174.6141157165f * powf(2.0f, (newFreq * (1.0f/12.0f)));
		if(newFreq != prevFreqs[i])
		{
			prevFreqs[i] = newFreq;
			coeffs[i] = 2.0f*sinf(3.141592653f*newFreq/44100.0f);

			//To keep the volume constant.
			tempf = fabsf(sines[i].x0)+fabsf(sines[i].x1);
			if(tempf > 1.0f)
			{
				tempf = 1.0f/tempf;
				sines[i].x0 *= tempf;
				sines[i].x1 *= tempf;
			}
		}

		//Calculate amplitude based on the indexed Quad's rotation around the
		//PlanetGO.
		//This turns out to be a bad idea as you get discontinuities when Quads
		//collide and change direction.
		//At the moment this just controls the sines' positions on the string.
		temp3f = quadsPos[i];
		x = temp3f.x;
		x *= x;
		y = temp3f.y;
		y *= y;
		tempf = sqrtf(x+y);
		if(tempf != 0.0f)
		{
			x = temp3f.x/tempf;
			y = temp3f.y/tempf;
		}
		else
		{
			x = 0.0f;
			y = 0.0f;
		}
		amp = cosf(y * 6.283185307179586476925286766559f)*sinf(x * 6.283185307179586476925286766559f);
		amp += 1.0f;
		amp *= 0.5f;

		//Calculate the position on the string to apply this excitation to.
		//buffers[i].position = amp;
		buffers[i].position = (float)i * 0.25f;

		//Calculate the sine tone, write to buffers[i].
		for(j=0;j<numSamples;++j)
		{
			//Calculate amp, using a simple granulator-style approach.
			if(granCount[i] == 0)
			{
				if(!granDirection[i])
				{
					granAmp[i] += 0.00025f;
					if(granAmp[i] >= 1.0f)
						granDirection[i] = true;
				}
				else
				{
					granAmp[i] -= 0.00025f;
					if(granAmp[i] <= 0.0f)
					{
						granAmp[i] = 0.0f;
						granDirection[i] = false;
						granCount[i] = (int)((heilanRand(heilanRandSeed())+1.0f)*3000.0f);
					}
				}
			}
			else
				--granCount[i];

			//Calculate further amp/attenuation according to force sensors.
			tempf = forces[i] * granAmp[i] * invNumSines;

			sines[i].x0 = sines[i].x0 - (coeffs[i] * sines[i].x1);
			sines[i].x1 = sines[i].x1 + (coeffs[i] * sines[i].x0);
			buffers[i].audioData[j] = sines[i].x1 * tempf;
		}
	}

	return &buffers;
}

//------------------------------------------------------------------------------
void PlanetGO::PlanetExciter::numBuffersChanged(int newSize)
{
	int i;
	int tempint;
	Sine tempSin;

	tempSin.x0 = 1.0f;
	tempSin.x1 = 0.0f;

	while((unsigned)newSize < sines.size())
		sines.pop_back();

	while((unsigned)newSize > sines.size())
	{
		tempint = sines.size() & 3;
		//So that the sines don't start with the same phase.
		for(i=0;i<(int)((heilanRand(heilanRandSeed())+1.0f)*50.0f);++i)
		{
			tempSin.x0 = tempSin.x0 - (coeffs[tempint] * tempSin.x1);
			tempSin.x1 = tempSin.x1 + (coeffs[tempint] * tempSin.x0);
		}
		sines.push_back(tempSin);
	}
}

}
