//	PlanetGO.h - GravityOBject2-based planetary cloud-type gravity object.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PLANETGO_H_
#define PLANETGO_H_

#include "GravityObject2.h"
#include "Parameter.h"

namespace AshitakaClasses
{

///	GravityOBject2-based planetary cloud-type gravity object.
class PlanetGO : public GravityObject2,
				 public ParameterCallback
{
  public:
	///	Constructor.
	PlanetGO();
	///	Destructor.
	~PlanetGO();

	///	Used to update the positions of the primary Quad2s.
	void updatePrimaryQuads();
	///	Used to update the position of a secondary Quad2.
	void updateSecondaryQuad(Quad2 *quad);
	///	Overridden to do nothing.
	/*!
		(there will only be one PlanetGO in Ashitaka, and it will never die)
	 */
	void updateLifeForce() {};
	///	Used to update quadsPos for PlanetExciter.
	void updateExciter();

	///	Used so we know when the ...Quads Parameters change.
	void valueChanged(float val);

	//--Parameters--------------------------------------------------------------
	///	Top-left Quads control.
	Parameter topLeftQuads;
	///	Top-right Quads control.
	Parameter topRightQuads;
	///	Bottom-left Quads control.
	Parameter bottomLeftQuads;
	///	Bottom-right Quads control.
	Parameter bottomRightQuads;

	///	Pitch control.
	Parameter pitch;
	//--------------------------------------------------------------------------
  private:
	//--------------------------------------------------------------------------
	///	Class representing how PlanetGO excites its string model.
	class PlanetExciter : public GravityObject2::Exciter
	{
	  public:
		///	Constructor.
		PlanetExciter(PlanetGO *go,
	  				  const std::vector<ThreeFloats>& pos,
					  const ThreeFloats& goPos);
		///	Destructor.
		~PlanetExciter();

		///	Where the excitation is generated.
		std::vector<Buffer> *getAudio(int numSamples);
	  private:
		///	So we can add or remove Sines when the number of buffers changes.
		void numBuffersChanged(int newSize);

		///	Simple struct holding the data we need to generate a (low-cost) sine wave.
		struct Sine
		{
			float x0;
			float x1;
		};

		///	The sine tones we use to excite the string.
		std::vector<Sine> sines;
		///	The coefficient used in generating the sine tones from sines.
		float sinesCoeff;
		///	Test.
		float coeffs[4];
		///	The previous frequencies.
		float prevFreqs[4];

		///	Granulator amplitude.
		float granAmp[4];
		///	Whether the granulator's in the attack or decay stage of the envelope.
		bool granDirection[4];
		///	Used to determine when to generate the next grain.
		int granCount[4];

		///	The quads' positions.
		const std::vector<ThreeFloats>& quadsPos;
		///	The gravity object's position.
		const ThreeFloats& planetPos;

		///	Our parent PlanetGO.
		PlanetGO *parentGO;
	};
	//--------------------------------------------------------------------------

	///	Used to update our Exciter to use the correct number of Buffers.
	void numQuadsChanged();
	///	Helper method to update the position of the Quad2 in question.
	void updateQuad(Quad2 *quad, float weight);

	///	Copies of all primary quads' positions for PlanetExciter.
	std::vector<ThreeFloats> quadsPos;
	///	Copy of our position for PlanetExciter.
	ThreeFloats planetPos;

	///	Used to visualise the fact that we're clipping.
	float clipEnv;

	///	Used to determine when to lock the pitch, and when to let it roam free.
	/*bool pitchLocked;
	///	Which pitch we're locked at.
	float atPitch;*/

	///	Which pitch we're currently at.
	int whichPitch;
	///	Whether to hold the current pitch till we've returned to the centre.
	bool holdPitch;
};

}

#endif
