//	PolyhedronGO.h - Polyhedron-based GravityObject.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef POLYHEDRONGO_H_
#define POLYHEDRONGO_H_

#include "GravityObject.h"
#include "Quad.h"
#include "Parameter.h"
#include "MiscHelper.h"

#include <cmath>
#include <iostream>

namespace AshitakaClasses
{

///	Polyhedron-based GravityObject.
class PolyhedronGO : public GravityObject
{
  public:
	///	Constructor.
	PolyhedronGO();
	///	Destructor.
	~PolyhedronGO();

	///	Used to reset the numSecondary variable.
	void updateStart();
	///	Updates all this GO's primary Quads.
	void updatePrimaryQuads();
	///	Updates the position of the passed-in secondary Quad.
	void updateSecondaryQuad(Quad *quad);
	///	Updates this GO's lifeForce.
	void updateLifeForce();

	///	Returns the audio for this GO.
	void getMonoAudio(AmbisonicData *buffer, int numSamples);

	//--Inputs--
	///	Size parameter.
	Parameter size;
	///	X position parameter.
	Parameter xPos;
	///	Y position parameter.
	Parameter yPos;
	///	Z position parameter.
	Parameter zPos;
	///	Twist parameter.
	Parameter twist;
	///	Explode parameter.
	Parameter explode;
	///	Stream 1 parameter.
	Parameter stream1;
	///	Stream 2 parameter.
	Parameter stream2;
	///	Stream 3 parameter.
	Parameter stream3;

	//--Outputs--
	///	Number of quads.
	Parameter numQuads;
	///	How many gaps there are between Quads.
	Parameter consistency;
	///	The perceived size of the whole shape.
	Parameter perceivedSize;
	///	The average velocity of the Quads.
	Parameter avgVelocity;
  private:
	///	Gets called when the number of primary Quads changes.
	void numQuadsChanged();

	///	Helper method to set a Quad's vertices based on its position in the polyhedron.
	/*!
		\param quad The Quad to act on.
		\param x The row within the sphere to place the Quad at.
		\param y The column within the sphere to place the Quad at.
		\param interpVal If the Quad is not currently at the correct position,
		this value determines how much of its current position is interpolated
		into its final position (i.e. it will move towards the correct position
		over time, rather than just jumping straight there).
	 */
	inline void setQuad(Quad *quad, float x, float y, float interpVal = 0.0625f)
	{
		int i;
		float tempf, tempf2;
		float distance;
		const float xInc = (1.0f/(float)numRows);
		const float yInc = (0.5f/(float)numColumns);
		const ThreeFloats * const vertices = quad->getVertices();
		const ThreeFloats& pos = getPosition();
		ThreeFloats tempVerts[8];
		int intX = heilanRound(x * (float)numRows);
		int intY = heilanRound((y-0.25f) * 2.0f * (float)numColumns);
		ThreeFloats direction;
		ThreeFloats centre;
		ThreeFloats oldCentre;
		int intX1, intX2;
		float intY1, intY2;
		float streamValue[4];

		//Calculate stream visualisation.
		intX1 = intX;
		if(intX1 > 3)
			intX1 = 7-intX1;
		intX2 = intX + 1;
		if(intX2 > 7)
			intX2 -= 7;
		if(intX2 > 3)
			intX2 = 7-intX2;
		intY1 = (y-0.25f) * 4.0f;
		intY2 = intY1 + 1.0f;
		tempf = intY1 - (float)((int)intY1);
		if(intY2 >= 2.0f)
			intY2 = 2.0f;
		tempf2 = tempf + (2.0f/(float)numColumns);
		streamValue[0] = (streamVals[(int)intY1][intX1] * (1.0f-tempf)) +
						 (streamVals[(int)intY2][intX1] * tempf);
		streamValue[0] *= 2.0f;
		streamValue[0] += 1.0f;
		streamValue[1] = (streamVals[(int)intY1][intX2] * (1.0f-tempf)) +
						 (streamVals[(int)intY2][intX2] * tempf);
		streamValue[1] *= 2.0f;
		streamValue[1] += 1.0f;
		streamValue[2] = (streamVals[(int)intY1][intX1] * (1.0f-tempf2)) +
						 (streamVals[(int)intY2][intX1] * tempf2);
		streamValue[2] *= 2.0f;
		streamValue[2] += 1.0f;
		streamValue[3] = (streamVals[(int)intY1][intX2] * (1.0f-tempf2)) +
						 (streamVals[(int)intY2][intX2] * tempf2);
		streamValue[3] *= 2.0f;
		streamValue[3] += 1.0f;

		//Handle twisting.
		if(twist > 0.0f)
		{
			if(intY % 2)
				x += twist; //i.e. twist == 1: rotation by 2pi.
			else
				x -= twist;
		}

		tempVerts[0].x = sinf(x*6.283185306f) * cosf(y*6.283185306f) * streamValue[0];
		tempVerts[0].y = (y*4.0f)-2.0f;
		tempVerts[0].z = cosf(x*6.283185306f) * cosf(y*6.283185306f) * streamValue[0];

		tempVerts[1].x = sinf((x+xInc)*6.283185306f) * cosf(y*6.283185306f) * streamValue[1];
		tempVerts[1].y = (y*4.0f)-2.0f;
		tempVerts[1].z = cosf((x+xInc)*6.283185306f) * cosf(y*6.283185306f) * streamValue[1];

		tempVerts[5].x = sinf((x+xInc)*6.283185306f) * cosf((y+yInc)*6.283185306f) * streamValue[3];
		tempVerts[5].y = ((y+yInc)*4.0f)-2.0f;
		tempVerts[5].z = cosf((x+xInc)*6.283185306f) * cosf((y+yInc)*6.283185306f) * streamValue[3];

		tempVerts[4].x = sinf(x*6.283185306f) * cosf((y+yInc)*6.283185306f) * streamValue[2];
		tempVerts[4].y = ((y+yInc)*4.0f)-2.0f;
		tempVerts[4].z = cosf(x*6.283185306f) * cosf((y+yInc)*6.283185306f) * streamValue[2];

		tempVerts[3].x = sinf(x*6.283185306f) * cosf(y*6.283185306f) * 0.6f;
		tempVerts[3].y = (y*4.0f)-2.0f;
		tempVerts[3].z = cosf(x*6.283185306f) * cosf(y*6.283185306f) * 0.6f;

		tempVerts[2].x = sinf((x+xInc)*6.283185306f) * cosf(y*6.283185306f) * 0.6f;
		tempVerts[2].y = (y*4.0f)-2.0f;
		tempVerts[2].z = cosf((x+xInc)*6.283185306f) * cosf(y*6.283185306f) * 0.6f;

		tempVerts[6].x = sinf((x+xInc)*6.283185306f) * cosf((y+yInc)*6.283185306f) * 0.6f;
		tempVerts[6].y = ((y+yInc)*4.0f)-2.0f;
		tempVerts[6].z = cosf((x+xInc)*6.283185306f) * cosf((y+yInc)*6.283185306f) * 0.6f;

		tempVerts[7].x = sinf(x*6.283185306f) * cosf((y+yInc)*6.283185306f) * 0.6f;
		tempVerts[7].y = ((y+yInc)*4.0f)-2.0f;
		tempVerts[7].z = cosf(x*6.283185306f) * cosf((y+yInc)*6.283185306f) * 0.6f;

		if(intY % 2)
			++intX;
		for(i=0;i<8;++i)
		{
			tempVerts[i] *= size * 0.25f;
			tempVerts[i] += pos;

			/*if((explode > 0.0f) && (intX % 2))
			{
				direction = tempVerts[i] - pos;
				distance = sqrtf((direction.x * direction.x) +
								 (direction.y * direction.y) +
								 (direction.z * direction.z));
				direction /= distance;

				tempVerts[i] += (direction * explode);
			}*/
			centre += tempVerts[i];
			oldCentre += vertices[i];
		}
		centre /= 8.0f;
		oldCentre /= 8.0f;

		if((explode > 0.0f) && (intX % 2))
		{
			direction = centre - pos;
			distance = sqrtf((direction.x * direction.x) +
							 (direction.y * direction.y) +
							 (direction.z * direction.z));
			direction /= distance;

			for(i=0;i<8;++i)
				tempVerts[i] += direction * explode;
		}

		tempf = centre.x - oldCentre.x;
		distance = tempf * tempf;
		tempf = centre.y - oldCentre.y;
		distance += tempf * tempf;
		tempf = centre.z - oldCentre.z;
		distance += tempf * tempf;
		distance = sqrtf(distance);

		if(distance > 0.001f)
		{
			for(i=0;i<8;++i)
			{
				tempVerts[i] = (tempVerts[i] * interpVal) +
							   ((ThreeFloats)vertices[i] * (1.0f-interpVal));
			}
		}

		quad->setVertices(tempVerts);
	};
	///	Helper method to position vertices around a centre point (not used?).
	inline void positionVertices(ThreeFloats *verts,
								 const ThreeFloats& centre,
								 const ThreeFloats& direction)
	{
		verts[0].x = centre.x - 0.5f + direction.x;
		verts[0].y = centre.y - 0.5f + direction.y;
		verts[0].z = centre.z + 0.5f + direction.z;

		verts[1].x = centre.x + 0.5f + direction.x;
		verts[1].y = centre.y - 0.5f + direction.y;
		verts[1].z = centre.z + 0.5f + direction.z;

		verts[2].x = centre.x + 0.5f + direction.x;
		verts[2].y = centre.y - 0.5f + direction.y;
		verts[2].z = centre.z - 0.5f + direction.z;

		verts[3].x = centre.x - 0.5f + direction.x;
		verts[3].y = centre.y - 0.5f + direction.y;
		verts[3].z = centre.z - 0.5f + direction.z;

		verts[4].x = centre.x - 0.5f - direction.x;
		verts[4].y = centre.y + 0.5f - direction.y;
		verts[4].z = centre.z + 0.5f - direction.z;

		verts[5].x = centre.x + 0.5f - direction.x;
		verts[5].y = centre.y + 0.5f - direction.y;
		verts[5].z = centre.z + 0.5f - direction.z;

		verts[6].x = centre.x + 0.5f - direction.x;
		verts[6].y = centre.y + 0.5f - direction.y;
		verts[6].z = centre.z - 0.5f - direction.z;

		verts[7].x = centre.x - 0.5f - direction.x;
		verts[7].y = centre.y + 0.5f - direction.y;
		verts[7].z = centre.z - 0.5f - direction.z;
	};
	///	Helper method to determine the number of rows and columns.
	/*!
		This does a square root of the value passed in, and if it doesn't
		return an integer, reduces the value until it gets an integer.

		...so it could be <em>slow</em>...
	 */
	inline int getSquareRoot(int numQuads)
	{
		float tempf;
		int retval = 0;

		if(!numQuads)
			return retval;

		tempf = sqrtf((float)numQuads);
		retval = (int)tempf;
		while((tempf-(float)retval) > 0.0f)
		{
			--numQuads;
			if(!numQuads)
			{
				retval = 0;
				break;
			}
			tempf = sqrtf((float)numQuads);
			retval = (int)tempf;
		}

		return retval;
	};

	///	Current number of rows in the polyhedron.
	int numRows;
	///	Current number of columns in the polyhedron.
	int numColumns;

	///	The number of secondary Quads we currently have.
	int numSecondary;

	///	The 3 previous values for the streams.
	float streamVals[3][4];
	///	Counter for when to update streamVals.
	int streamCounter;
};

}

#endif
