//	Quad.h - The element which is controlled by GravityObjects.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef QUAD_H_
#define QUAD_H_

#include "ThreeFloats.h"
#include "GLHeaders.h"

//#define INVERT_COLOURS

namespace AshitakaClasses
{

class GravityObject;

///	The element which is controlled by GravityObjects.
/*!
	A Quad is essentially a kind of particle, whose position and motion is
	determined by a primary GravityObject (and to a lesser degree, by any number
	of secondary GravityObjects).

	Visually it consists of 8 vertices, intially connected together in a cube
	shape, though this may be altered by the Quad's interactions with
	GravityObjects.

	\sa GravityObject
 */
class Quad
{
  public:
	///	Constructor.
	Quad();
	///	Destructor.
	~Quad();

	///	Draws the Quad.
	void draw();

	///	Sets the positions of the Quad's 8 vertices.
	/*!
		\param verts Pointer to an array of 8 ThreeFloats structures.  <em>Don't
		send fewer than 8!</em>

		This also updates the centre, direction, velocity and acceleration
		variables.
	 */
	void setVertices(ThreeFloats *verts);
	///	Sets the (main) colour of the Quad.
	/*!
		Note that each Quad is given an individual brightness when it is
		constructed, which will alter the actual colour you pass in.  As such,
		it probably makes more sense to consider this method as a way of setting
		the Quad's <em>hue</em>.

		The Quad is also arbitrarily shaded, so only the 'front' face will be
		the full brightness colour.
	 */
	void setColour(const ThreeFloats& col);
	///	Sets the Quad's primary GravityObject.
	void setPrimaryGravityObject(GravityObject *obj);
	///	Sets the Quad's primary GO's size.
	void setSize(float val);

	///	Returns the positions of the Quad's 8 vertices.
	const ThreeFloats * const getVertices() const {return vertices;};
	///	Returns the Quad's velocity.
	float getVelocity() const {return velocity;};
	///	Returns the Quad's velocity.
	const ThreeFloats& getVelocity3d() const {return velocity2;};
	///	Returns the Quad's acceleration.
	float getAcceleration() const {return acceleration;};
	///	Returns the Quad's direction.
	const ThreeFloats& getDirection() const {return direction;};
	///	Returns the Quad's centre.
	const ThreeFloats& getCentre() const {return centre;};
	///	Returns the Quad's primary GravityObject.
	GravityObject *getPrimaryGravityObject() const {return primary;};
	///	Returns the Quad's current colour.
	const ThreeFloats& getColour() const {return baseColour;};
  private:
	///	Helper method to draw the Quad's geometry.
	/*!
		\param alpha How much alpha to render with.
		\param trail Whether or not to draw the trail.
	 */
	void drawQuad(float alpha, bool trail = false);
	///	Helper method to draw the Quad as a simple texture.
	void drawBlob(float alpha, bool trail = false);
	///	Helper method to load a texture.
	void loadTexture();

	///	How many previous positions to store for the trail.
	enum
	{
		TrailLength = 32
	};

	///	The positions of the 8 vertices.
	ThreeFloats vertices[8];
	///	The brightness of the Quad.
	float brightness;
	///	The actual colour of the Quad.
	ThreeFloats colour;
	///	The base colour of the Quad.
	ThreeFloats baseColour;
	///	The centre of the Quad.
	ThreeFloats centre;
	///	The direction of the Quad.
	ThreeFloats direction;
	///	The velocity of the Quad.
	float velocity;
	///	The velocity of the Quad.
	ThreeFloats velocity2;
	///	The acceleration of the Quad.
	float acceleration;

	///	The size of the GravityObject (determines the size of the blob).
	float goSize;

	///	The Quad's primary GravityObject.
	GravityObject *primary;

	///	For when we're currently drawing a trail.
	float drawTrail;
	///	To determine when to draw a trail.
	int trailCount;
	///	Start position of the trail.
	ThreeFloats trailStart;
	///	End position of the trail.
	ThreeFloats trailEnd;
	///	The last TrailLength positions.
	ThreeFloats trailPos[TrailLength];
	///	Current write position within trailPos.
	int trailWritePos;

	///	Texture to use instead of the polygons.
	GLuint texture;
};

}

#endif
