//	QuadManager.cpp - Object to control GravityObjects and Quads.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "QuadManager.h"
#include "SimpleAttractorGO.h"
#include "PolyhedronGO.h"
#include "SnakeGO.h"
#include "JellyfishGO.h"
#include "RandomLineGO.h"
#include "DiamondPercGO.h"
#include "Camera.h"
#include "AudioSynth.h"

#include <algorithm>

using namespace std;

namespace AshitakaClasses
{

//------------------------------------------------------------------------------
QuadManager::QuadManager():
splitCounter(0),
wobble(0.0f),
mainBuffer(0),
tempBuffer(0),
bufferSize(0),
audioSynth(0)
{
	int i;

	//Setup main GO, and its mappings.
	mainGravityObject = new PolyhedronGO();
	mainSize >> ((PolyhedronGO *)mainGravityObject)->size;
	mainXPos >> ((PolyhedronGO *)mainGravityObject)->xPos;
	mainYPos >> ((PolyhedronGO *)mainGravityObject)->yPos;
	mainZPos >> ((PolyhedronGO *)mainGravityObject)->zPos;
	mainTwist >> ((PolyhedronGO *)mainGravityObject)->twist;
	mainExplode >> ((PolyhedronGO *)mainGravityObject)->explode;
	mainStream1 >> ((PolyhedronGO *)mainGravityObject)->stream1;
	mainStream2 >> ((PolyhedronGO *)mainGravityObject)->stream2;
	mainStream3 >> ((PolyhedronGO *)mainGravityObject)->stream3;

	((PolyhedronGO *)mainGravityObject)->numQuads >> numMainQuads;
	((PolyhedronGO *)mainGravityObject)->consistency >> mainConsistency;
	((PolyhedronGO *)mainGravityObject)->perceivedSize >> mainPerceivedSize;
	((PolyhedronGO *)mainGravityObject)->avgVelocity >> mainAvgVelocity;

	for(i=0;i<NumQuads;++i)
		mainGravityObject->addPrimaryQuad(&(quads[i]));

	mainSize = 1.0f;
}

//------------------------------------------------------------------------------
QuadManager::~QuadManager()
{
	unsigned int i;

	for(i=0;i<gravityObjects.size();++i)
		delete gravityObjects[i];

	delete mainGravityObject;

	if(mainBuffer)
		delete [] mainBuffer;
	if(tempBuffer)
		delete [] tempBuffer;
}

//------------------------------------------------------------------------------
TwoFloats QuadManager::update()
{
	int i;
	int count;
	int newGO;
	float tempf;
	Quad *tempQuad;
	TwoFloats retval;
	vector<GravityObject *>::iterator it;
	float splitOff = 0.0f;
	string goType;

	//Update GOs' lifeForces.
	if(gravityObjects.size() > 0)
	{
		for(it=(gravityObjects.end()-1);it>=gravityObjects.begin();--it)
		{
			(*it)->updateLifeForce();
			if((*it)->getLifeForce() == 0.0f)
			{
				vector<Quad *>& tempQuads = (*it)->getPrimaryQuads();

				for(i=(tempQuads.size()-1);i>=0;--i)
				{
					tempQuad = tempQuads[i];
					reparentQuad(tempQuad);
					(*it)->removePrimaryQuad(tempQuad);
				}

				delete (*it);
				gravityObjects.erase(it);
			}
		}
	}

	if(splitOff1 > 0.5f)
	{
		goType = "SimpleAttractor";
		splitOff = splitOff1;
	}
	else if(splitOff2 > 0.5f)
	{
		goType = "Snake";
		splitOff = splitOff2;
	}
	else if(splitOff3 > 0.5f)
	{
		goType = "RandomLine";
		splitOff = splitOff3;
	}
	else if(splitOff4 > 0.5f)
	{
		goType = "DiamondPerc";
		splitOff = splitOff4;
	}

	//Split off Quads if necessary.
	if((splitOff > 0.5f) && (!splitCounter))
	{
		vector<Quad *>& mainQuads = mainGravityObject->getPrimaryQuads();

		addGravityObject(goType, 1.0f);
		newGO = gravityObjects.size()-1;

		for(i=(mainQuads.size()-1), count=4;
			((i>=0) && (count>0));
			--i, --count)
		{
			tempQuad = mainQuads[i];
			mainGravityObject->removePrimaryQuad(tempQuad);
			gravityObjects[newGO]->addPrimaryQuad(tempQuad);
			gravityObjects[newGO]->setPosition(mainGravityObject->getPosition());
		}

		splitOff = 0.0f;
		splitCounter = 30;
		cout << "Split Off." << endl;
	}
	else if(splitCounter)
	{
		--splitCounter;
		splitOff = 0.0f;
	}

	//Inform GOs that we're at the start of a new frame.
	mainGravityObject->updateStart();
	for(it=gravityObjects.begin();it!=gravityObjects.end();++it)
		(*it)->updateStart();

	//Inform Quads about their size.
	for(i=0;i<NumQuads;++i)
		quads[i].setSize(mainSize);

	//Update secondary Quads.
	if(gravityObjects.size() > 0)
	{
		updateSecondary(mainGravityObject);
		for(it=gravityObjects.begin();it!=gravityObjects.end();++it)
			updateSecondary(*it);
	}

	//Update primary Quads.
	mainGravityObject->updatePrimaryQuads();
	for(it=gravityObjects.begin();it!=gravityObjects.end();++it)
		(*it)->updatePrimaryQuads();

	tempf = ((float)gravityObjects.size()/100.0f);
	wobble += 0.1f * tempf;
	while(wobble > 1.0f)
		wobble -= 1.0f;
	retval.x = sinf(wobble * 6.283185306f);
	retval.x += 1.0f;
	retval.x *= 0.5f;
	retval.x *= tempf;
	retval.y = cosf(wobble * 6.283185306f);
	retval.y += 1.0f;
	retval.y *= 0.5f;
	retval.y *= tempf;

	return retval;
}

//------------------------------------------------------------------------------
void QuadManager::draw()
{
	int i;

	for(i=0;i<NumQuads;++i)
		quads[i].draw();
}

//------------------------------------------------------------------------------
AmbisonicData *QuadManager::getAudio(int numSamples)
{
	unsigned int i;
	int j;
	ThreeFloats tempPos;
	AmbisonicData *retval = 0;

	if(bufferSize && mainBuffer && tempBuffer)
	{
		//Clear mainBuffer.
		for(j=0;j<numSamples;++j)
			mainBuffer[j] = 0.0f;

		for(i=0;i<gravityObjects.size();++i)
		{
			//Get GO audio.
			gravityObjects[i]->getAmbiAudio(tempBuffer, numSamples);

			//Add to existing audio.
			for(j=0;j<numSamples;++j)
				mainBuffer[j] += tempBuffer[j];
		}
		retval = mainBuffer;
	}

	return retval;
}

//------------------------------------------------------------------------------
void QuadManager::setBlockSize(int newSize)
{
	if(mainBuffer)
		delete [] mainBuffer;
	if(tempBuffer)
		delete [] tempBuffer;

	bufferSize = newSize;
	mainBuffer = new AmbisonicData[bufferSize];
	tempBuffer = new AmbisonicData[bufferSize];
}

//------------------------------------------------------------------------------
void QuadManager::setAudioSynth(AudioSynth *synth)
{
	audioSynth = synth;
}

//------------------------------------------------------------------------------
void QuadManager::addGravityObject(string type, float life)
{
	GravityObject *tempObj = 0;

	if(type == "Polyhedron")
		tempObj = new PolyhedronGO();
	else if(type == "SimpleAttractor")
		tempObj = new SimpleAttractorGO();
	else if(type == "Snake")
		tempObj = new SnakeGO();
	else if(type == "Jellyfish")
		tempObj = new JellyfishGO();
	else if(type == "RandomLine")
		tempObj = new RandomLineGO();
	else if(type == "DiamondPerc")
		tempObj = new DiamondPercGO();
	else
		return;

	if(tempObj)
	{
		tempObj->setAudioData(audioSynth->getAudioData());
		tempObj->setTriggerData(audioSynth->getTriggerData(),
								audioSynth->getTriggerWritePoint());
		tempObj->setLifeForce(life);
	}

	gravityObjects.push_back(tempObj);
}

//------------------------------------------------------------------------------
void QuadManager::reparentQuad(Quad *quad)
{
	float dist;
	float tempf;
	ThreeFloats temp3f;
	vector<GravityObject *>::iterator it;
	GravityObject *obj = mainGravityObject;

	temp3f = quad->getCentre();
	temp3f -= mainGravityObject->getPosition();
	dist = sqrtf((temp3f.x*temp3f.x) +
				 (temp3f.y*temp3f.y) +
				 (temp3f.z*temp3f.z));

	for(it=gravityObjects.begin();it!=gravityObjects.end();++it)
	{
		if(quad->getPrimaryGravityObject() != *it)
		{
			temp3f = quad->getCentre();
			temp3f -= (*it)->getPosition();
			tempf = sqrtf((temp3f.x*temp3f.x) +
						  (temp3f.y*temp3f.y) +
						  (temp3f.z*temp3f.z));

			if(tempf < dist)
			{
				dist = tempf;
				obj = *it;
			}
		}
	}

	obj->addPrimaryQuad(quad);
}

//------------------------------------------------------------------------------
bool QuadManager::compareZOrder(Quad *left, Quad *right)
{
	float distance1, distance2;
	ThreeFloats direction;
	ThreeFloats camPos = HeilanX3D::Camera::getInstance()->getGraphicPosition();
	ThreeFloats tempPos = camPos - left->getCentre();

	distance1 = sqrtf((tempPos.x*tempPos.x) +
					  (tempPos.y*tempPos.y) +
					  (tempPos.z*tempPos.z));

	tempPos = camPos - right->getCentre();
	distance2 = sqrtf((tempPos.x*tempPos.x) +
					  (tempPos.y*tempPos.y) +
					  (tempPos.z*tempPos.z));

	return (distance1 > distance2);
}

}
