//	QuadManager.h - Object to control GravityObjects and Quads.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef QUADMANAGER_H_
#define QUADMANAGER_H_

#include "Quad.h"
#include "GravityObject.h"
#include "Parameter.h"
#include "TwoFloats.h"

#include <string>
#include <cmath>
#include <iostream>

namespace AshitakaClasses
{

class AudioSynth;

///	Object to control GravityObjects and Quads.
/*!
	This class takes care of issues like instantiating new GravityObjects,
	bounds-checking for assigning secondary GravityObjects etc.
 */
class QuadManager
{
  public:
	///	Constructor.
	QuadManager();
	///	Destructor.
	~QuadManager();

	///	Updates the various positions etc.
	/*!
		\return The two background wobble parameters.
	 */
	TwoFloats update();
	///	Draws all the Quads.
	void draw();

	///	Returns the audio from all the extra GOs.
	AmbisonicData *getAudio(int numSamples);
	///	Used to set the size of the buffer returned by getAudio().
	void setBlockSize(int newSize);

	///	Tells us where to get audio data from for the main instrument.
	void setAudioSynth(AudioSynth *synth);

	///	Adds a new GravityObject to the scene.
	/*!
		\param type What type of GravityObject to create.
		\param life How much life force to assign the GravityObject.
	 */
	void addGravityObject(std::string type, float life);
	///	Called to re-assign a Quad to a new primary GravityObject.
	/*!
		This should be called by GravityObjects when they die for all their
		primary Quads.
	 */
	void reparentQuad(Quad *quad);

	///	Returns a pointer to all the Quads.
	Quad *getQuads() {return quads;};

	///	Helper method used to sort the Quads according to their z-order.
	static bool compareZOrder(Quad *left, Quad *right);

	//--Inputs--
	///	Main GravityObject size parameter.
	Parameter mainSize;
	///	Main GravityObject x position.
	Parameter mainXPos;
	///	Main GravityObject y position.
	Parameter mainYPos;
	///	Main GravityObject z position.
	Parameter mainZPos;
	///	Main GravityObject twist parameter.
	Parameter mainTwist;
	///	Main GravityObject explode parameter.
	Parameter mainExplode;
	///	Main GravityObject stream1 parameter.
	Parameter mainStream1;
	///	Main GravityObject stream2 parameter.
	Parameter mainStream2;
	///	Main GravityObject stream3 parameter.
	Parameter mainStream3;

	///	Parameter used to split off Quads from the main GO.
	Parameter splitOff1;
	///	Parameter used to split off Quads from the main GO.
	Parameter splitOff2;
	///	Parameter used to split off Quads from the main GO.
	Parameter splitOff3;
	///	Parameter used to split off Quads from the main GO.
	Parameter splitOff4;
	///	Parameter used to split off Quads from the main GO.
	Parameter splitOff5;
	///	Parameter used to split off Quads from the main GO.
	Parameter splitOff6;
	///	Parameter used to split off Quads from the main GO.
	Parameter splitOff7;

	//--Outputs--
	///	Number of main quads.
	Parameter numMainQuads;
	///	How many gaps there are between Quads.
	Parameter mainConsistency;
	///	The perceived size of the whole shape.
	Parameter mainPerceivedSize;
	///	The average velocity of the Quads.
	Parameter mainAvgVelocity;

	///	The number of Quads in the space.
	enum
	{
		NumQuads = 8 * 8 * 2 //i.e. 8 slices, 8 rows in a polyhedron.
	};
  private:
	///	Helper method to update secondary Quads.
	inline void updateSecondary(GravityObject *object)
	{
		int i;
		ThreeFloats temp3f;
		float tempf;

		for(i=0;i<NumQuads;++i)
		{
			if(quads[i].getPrimaryGravityObject() != object)
			{
				temp3f = quads[i].getCentre();
				temp3f -= object->getPosition();
				tempf = sqrtf((temp3f.x*temp3f.x) +
							  (temp3f.y*temp3f.y) +
							  (temp3f.z*temp3f.z));

				if(tempf < object->getInfluenceRadius())
				{
					object->updateSecondaryQuad(&(quads[i]));
					//std::cout << "Within radius." << std::endl;
				}
			}
		}
	};
	///	Helper method to convert GO audio to Ambi B format.
	inline void convertAudio(AmbisonicData *buffer,
					 		 int numSamples,
							 const ThreeFloats& pos)
	{
		
	};

	///	The main performer-controlled GravityObject.
	GravityObject *mainGravityObject;
	///	Vector of any other GravityObjects.
	std::vector<GravityObject *> gravityObjects;
	///	All the Quads in the space.
	Quad quads[NumQuads];

	///	Used to make sure we don't trigger multiple split offs accidentally.
	int splitCounter;

	///	Used to wobble the background when there are a lot of GOs active.
	float wobble;

	///	The audio buffer returned by getAudio().
	AmbisonicData *mainBuffer;
	///	Temporary audio buffer.
	AmbisonicData *tempBuffer;
	///	The size of mainBuffer and tempBuffer.
	int bufferSize;

	///	Pointer to our AudioSynth.
	AudioSynth *audioSynth;
};
	
}

#endif
