//	QuadManager2.cpp - Second version of the QuadManager class.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "QuadManager2.h"
#include "AmbisonicData.h"
#include "PlanetGO.h"
#include "PulsatingGO.h"
#include "FlockingGO.h"
#include "CircleGO.h"
#include "MiscHelper.h"
#include "Camera.h"

#include <string>
#include <cassert>

using namespace std;
using namespace HeilanX3D;

namespace AshitakaClasses
{

//------------------------------------------------------------------------------
QuadManager2::QuadManager2():
mainGravityObject(0),
mainBuffer(0),
tempBuffer(0),
bufferSize(0),
collisionCount(0),
split(-1),
splitCount(0),
gestureCount(0),
sceneRot(0.0f),
rotDirection(false),
rotIndex(0.0f),
lineColour(true)
{
	int i;

	for(i=0;i<NumSparks;++i)
		sparks[i] = 0;
}

//------------------------------------------------------------------------------
QuadManager2::~QuadManager2()
{
	unsigned int i;

	for(i=0;i<gravityObjects.size();++i)
		delete gravityObjects[i];

	if(mainBuffer)
		delete [] mainBuffer;
	if(tempBuffer)
		delete [] tempBuffer;

	for(i=0;i<NumSparks;++i)
	{
		if(sparks[i])
			delete sparks[i];
	}
}

//------------------------------------------------------------------------------
void QuadManager2::initialise()
{
	int i;
	int x, y;
	float tempf;
	unsigned char pixels[256*256*2];
	unsigned char *pixelPointer = pixels;

	mainGravityObject = new PlanetGO();
	assert(mainGravityObject != 0); //Should probably be an exception?
	gravityObjects.push_back(mainGravityObject);

	//Add all our Quads to mainGravityObject.
	for(i=0;i<NumQuads;++i)
	{
		mainGravityObject->addPrimaryQuad(&(quads[i]));
		quads[i].setPrimaryGravityObject(mainGravityObject);
	}

	//Attach our Parameters to mainGravityObject's.
	mainXPos >> mainGravityObject->xPos;
	mainYPos >> mainGravityObject->yPos;
	mainZPos >> mainGravityObject->zPos;
	length >> mainGravityObject->decay;
	twist >> ((PlanetGO *)mainGravityObject)->pitch;
	topLeftForce >> ((PlanetGO *)mainGravityObject)->topLeftQuads;
	topRightForce >> ((PlanetGO *)mainGravityObject)->topRightQuads;
	bottomLeftForce >> ((PlanetGO *)mainGravityObject)->bottomLeftQuads;
	bottomRightForce >> ((PlanetGO *)mainGravityObject)->bottomRightQuads;

	//Generate texture.
	for(y=0;y<256;++y)
	{
		for(x=0;x<256;++x)
		{
			tempf = sinf(((float)x/256.0f)*3.1415926535897932384626433832795f);
			tempf *= sinf(((float)y/256.0f)*3.1415926535897932384626433832795f);
			tempf = (exp(tempf)-1.0f)/(exp(1.0f)-1.0f)*1.5f-0.1f;
			if(tempf < 0.0f)
				tempf = 0.0f;
			else if(tempf > 1.0f)
				tempf = 1.0f;
			*pixelPointer = 255;
			*(pixelPointer+1) = (unsigned char)(tempf * 255.0f);
			pixelPointer += 2;
		}
	}

	//Setup texture.
	glGenTextures(1, &sparkTexture);
	glBindTexture(GL_TEXTURE_2D, sparkTexture);
	glTexImage2D(GL_TEXTURE_2D,
				 0,
				 2,
				 256,
				 256,
				 0,
				 GL_LUMINANCE_ALPHA,
				 GL_UNSIGNED_BYTE,
				 pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);

	//Setup sparks.
	for(i=0;i<NumSparks;++i)
		sparks[i] = new AshitakaSparks(sparkTexture);
}

//------------------------------------------------------------------------------
void QuadManager2::update()
{
	vector<GravityObject2 *>::iterator it;
	unsigned int i, j, k, l;
	GravityObject2 *tempGO;
	Quad2 *tempQuad;
	float distance;
	float tempf;
	int m;
	unsigned int numGOs = gravityObjects.size();

	//Update gesture.
	if(gestureCount <= 0)
	{
		for(i=9;i>0;--i)
			gesture[i] = gesture[i-1];
		gesture[0] = ThreeFloats(mainXPos, mainYPos, mainZPos);
		gestureCount = 5;
	}
	else
		--gestureCount;

	//Update GOs' lifeForces.
	if(numGOs > 1)
	{
		critSec.lock();
		{
			for(it=(gravityObjects.end()-1);it>gravityObjects.begin();--it)
			{
				tempGO = *it;
				tempGO->updateLifeForce();

				//if(tempGO->getLifeForce() <= 0.0f)
				if(!(tempGO->getLifeForce() > 0.0f))
				{
					vector<Quad2 *>& tempQuads = tempGO->getPrimaryQuads();

					for(m=(tempQuads.size()-1);m>=0;--m)
					{
						tempQuad = tempQuads[m];
						reparentQuad(tempQuad);
						tempGO->removePrimaryQuad(tempQuad);
					}

					gravityObjects.erase(it);
					//Make sure we disconnect any connected strings.
					numGOs = gravityObjects.size();
					for(i=0;i<numGOs;++i)
					{
						if(gravityObjects[i]->stringIsConnected(tempGO))
							gravityObjects[i]->removeString(tempGO);
						if(tempGO->stringIsConnected(gravityObjects[i]))
							tempGO->removeString(gravityObjects[i]);
					}
					delete tempGO;
				}
			}
		}
		critSec.unlock();
	}

	//Split off new GravityObject2s if necessary.
	if((splitCount == 0) && (split != -1))
	{
		//int newGO;
		//int count;
		//vector<Quad2 *>& mainQuads = mainGravityObject->getPrimaryQuads();

		if(split == 0)
			addGravityObject("Pulsating", 1.0f);
		else if(split == 1)
			addGravityObject("Flocking", 1.0f);
		else if(split == 2)
		{
			for(i=0;i<6;++i)
				addGravityObject("Circle", 1.0f, i);
		}
		/*newGO = gravityObjects.size()-1;

		for(m=(mainQuads.size()-1), count=8;
			((m>=0) && (count>0));
			--m, --count)
		{
			tempQuad = mainQuads[m];
			mainGravityObject->removePrimaryQuad(tempQuad);
			gravityObjects[newGO]->addPrimaryQuad(tempQuad);
			tempQuad->setPrimaryGravityObject(gravityObjects[newGO]);
		}
		gravityObjects[newGO]->setPosition(mainGravityObject->getPosition());
		gravityObjects[newGO]->setGesture(gesture);*/

		split = -1;
		splitCount = 5;
	}
	else if(splitCount)
		--splitCount;

	//Update primary quads.
	mainGravityObject->updatePrimaryQuads();

	numGOs = gravityObjects.size();
	for(i=1;i<numGOs;++i)
		gravityObjects[i]->updatePrimaryQuads();

	//Update secondary quads.
	if(numGOs > 1)
	{
		for(i=0;i<NumQuads;++i)
		{
			//Test w/mainGravityObject.
			if(quads[i].getPrimaryGravityObject() != mainGravityObject)
			{
				distance = quads[i].getPosition().getDistance(mainGravityObject->getPosition());
				if(distance < (mainGravityObject->getInfluenceRadius()))
					mainGravityObject->updateSecondaryQuad(&(quads[i]));
			}

			//Test w/any other GravityObject2s.
			for(j=0;j<numGOs;++j)
			{
				if(quads[i].getPrimaryGravityObject() != gravityObjects[j])
				{
					distance = quads[i].getPosition().getDistance(gravityObjects[j]->getPosition());
					if(distance < (gravityObjects[j]->getInfluenceRadius()))
						gravityObjects[j]->updateSecondaryQuad(&(quads[i]));
				}
			}
		}
	}

	//Update Exciters.
	for(i=0;i<numGOs;++i)
		gravityObjects[i]->updateExciter();

	//Connect/disconnect strings if necessary.

	//This bit's a bit complicated.  The following ascii diagram hopefully
	//explains why it's done this way:
	//   2
	// 1   3
	//  5 4
	//The above represents 5 GravityObject2s, all within each others' radius of
	//influence.  We want to connect their string models in the order
	//1->2->3->4->5->1, so that we get a complete ring, with the vibrations
	//travelling through all strings.  In order to do this though, we need to
	//parse(?) the scene in a very particular way (remember that two-way
	//connections are not allowed, because the string models get unstable then).
	//The naive approach (of just iterating through gravityObjects and
	//connecting any ones that are close enough) would result in something like:
	//1->2->3->4->5
	//1->5
	//...which is not what we want, as we won't get the circular connections we
	//want.
	//
	//Instead, we iterate through each GravityObject2, and connect it to the
	//first GravityObject2 we find within its radius, then break, and move onto
	//the next GravityObject2 (i-based loop).  This means we get the
	//1->2->3->4->5->1 pattern we want, but doesn't allow for multiple
	//connections, so I added in the l-based for loop.  It's not ideal, but I
	//don't want to spend too much cpu time connecting strings, so a maximum of
	//2 outward connections will have to do for now.
	for(l=0;l<2;++l)
	{
		for(i=0;i<numGOs;++i)
		{
			for(j=1;j<numGOs;++j)
			{
				//Make sure we check against all GOs that aren't
				//gravityObjects[i].
				k = (i+j) % numGOs;

				distance = gravityObjects[i]->getPosition().getDistance(gravityObjects[k]->getPosition());
				tempf = gravityObjects[i]->getInfluenceRadius();
				tempf += gravityObjects[k]->getInfluenceRadius();
				if(distance < tempf)
				{
					//Only connect if they're not already connected.
					if((!gravityObjects[i]->stringIsConnected(gravityObjects[k])) &&
					   (!gravityObjects[k]->stringIsConnected(gravityObjects[i])))
					{
						gravityObjects[i]->connectStrings(gravityObjects[k]);
						break;
					}
				}
				else
				{
					if(gravityObjects[i]->stringIsConnected(gravityObjects[k]))
						gravityObjects[i]->removeString(gravityObjects[k]);
					if(gravityObjects[k]->stringIsConnected(gravityObjects[i]))
						gravityObjects[k]->removeString(gravityObjects[i]);
				}
			}
		}
	}

	//Move all the Quads to their new positions.
	for(i=0;i<NumQuads;++i)
	{
		quads[i].interpolate();
		++i;
		quads[i].interpolate();
		++i;
		quads[i].interpolate();
		++i;
		quads[i].interpolate();
	}

	//Collision detection (this comment courtesy of the bureau of redundant
	//commenting).
	collisionDetection();
}

//------------------------------------------------------------------------------
void QuadManager2::draw()
{
	unsigned int i;
	ThreeFloats temp3f;
	ThreeFloats mainPos;
	ThreeFloats perp1, perp2;
	const ThreeFloats origin;
	float tempf;
	float tempRot;
	ThreeFloats tempPos;
	//FourFloats tempRot2;
	Quaternion tempRot3(0.0, 0.0, 1.0, 0.0);

	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	//Update scene's rotation.
	if(rotIndex != 0.0f)
	{
		if(rotDirection == 0)
		{
			tempRot = sceneRot * rotIndex;
			tempPos.x = scenePos.x * rotIndex;
			tempPos.z = (8 * (1.0f-rotIndex)) + (scenePos.z * rotIndex);

			/*tempRot2.x = -tempRot;
			Camera::getInstance()->setGraphicRotation(tempRot2);
			Camera::getInstance()->setGraphicPosition(tempPos);*/
			//tempRot3.w = -tempRot;
			temp3f = ThreeFloats(0.0f, 1.0f, 0.0f);
			tempRot3.setFromAxisAngle(-tempRot*0.017453292519943295769236907684886,
									  temp3f);
			Camera::getInstance()->setGraphicRotation(tempRot3);
			//Camera::getInstance()->setGraphicPosition(tempPos);

			rotIndex += 0.025f;
			if(rotIndex >= 1.0f)
			{
				rotIndex = 1.0f;
				rotDirection = 1;
			}
		}
		else if(rotDirection == 1)
		{
			tempRot = sceneRot;
			tempPos.x = scenePos.x;
			tempPos.z = scenePos.z;

			/*tempRot2.x = -tempRot;
			Camera::getInstance()->setGraphicRotation(tempRot2);
			Camera::getInstance()->setGraphicPosition(tempPos);*/
			//tempRot3.w = -tempRot;
			temp3f = ThreeFloats(0.0f, 1.0f, 0.0f);
			tempRot3.setFromAxisAngle(-tempRot*0.017453292519943295769236907684886,
									  temp3f);
			Camera::getInstance()->setGraphicRotation(tempRot3);
			//Camera::getInstance()->setGraphicPosition(tempPos);

			rotIndex -= 0.05f;
			if(rotIndex <= 0.0f)
			{
				rotIndex = 1.0f;
				rotDirection = 2;
			}
		}
		else
		{
			tempf = heilanLog(rotIndex);
			//tempf = rotIndex;
			tempRot = sceneRot * tempf;
			tempPos.x = scenePos.x * tempf;
			tempPos.z = (8 * (1.0f-tempf)) + (scenePos.z * tempf);

			rotIndex -= 0.01f;
			if(rotIndex <= 0.0f)
			{
				rotIndex = 0.0f;
				rotDirection = 0;

				//tempRot2.x = 0.0f;
				tempPos.x = 0.0f;
				tempPos.z = 8.0f;
				/*Camera::getInstance()->setGraphicRotation(tempRot2);
				Camera::getInstance()->setGraphicPosition(tempPos);*/
				//tempRot3.w = 0.0;
				temp3f = ThreeFloats(0.0f, 1.0f, 0.0f);
				tempRot3.setFromAxisAngle(0.0, temp3f);
				Camera::getInstance()->setGraphicRotation(tempRot3);
				//Camera::getInstance()->setGraphicPosition(tempPos);
			}
			else
			{
				/*tempRot2.x = -tempRot;
				Camera::getInstance()->setGraphicRotation(tempRot2);
				Camera::getInstance()->setGraphicPosition(tempPos);*/
				//tempRot3.w = -tempRot;
				temp3f = ThreeFloats(0.0f, 1.0f, 0.0f);
				tempRot3.setFromAxisAngle(-tempRot*0.017453292519943295769236907684886,
										  temp3f);
				Camera::getInstance()->setGraphicRotation(tempRot3);
				//Camera::getInstance()->setGraphicPosition(tempPos);
			}
		}
	}
	/*else
	{
		tempRot = 0.0f;
		tempPos.x = 0.0f;
		tempPos.z = 8.0f;
	}*/

	//Apply scene's rotation.
	/*glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glRotatef(-tempRot, 0.0f, 1.0f, 0.0f);
	glTranslatef(-tempPos.x, 0.0f, -tempPos.z);*/

	//Draw Quads.
	for(i=0;i<NumQuads;++i)
		quads[i].draw();

	//Add in sparks flying when Quads collide.
	for(i=0;i<NumSparks;++i)
	{
		if(sparks[i]->isActive())
			sparks[i]->draw();
	}

	//Draw any connected strings.
	for(i=0;i<gravityObjects.size();++i)
		gravityObjects[i]->drawStringConnections(lineColour);

	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
}

//------------------------------------------------------------------------------
AmbisonicData *QuadManager2::getAudio(int numSamples, const ThreeFloats& pos)
{
	unsigned int i;
	int j;
	AmbisonicData *retval = 0;

	if(bufferSize && mainBuffer && tempBuffer)
	{
		//Clear mainBuffer.
		for(j=0;j<numSamples;++j)
			mainBuffer[j] = 0.0f;

		critSec.lock();
		{
			for(i=0;i<gravityObjects.size();++i)
			{
				//Get GO audio.
				gravityObjects[i]->getAmbiAudio(tempBuffer, numSamples, pos);

				//Add to existing audio.
				for(j=0;j<numSamples;++j)
					mainBuffer[j] += tempBuffer[j];
			}
		}
		critSec.unlock();
		retval = mainBuffer;
	}

	return retval;
}

//------------------------------------------------------------------------------
void QuadManager2::setBlockSize(int newSize)
{
	unsigned int i;

	if(mainBuffer)
		delete [] mainBuffer;
	if(tempBuffer)
		delete [] tempBuffer;

	bufferSize = newSize;
	mainBuffer = new AmbisonicData[bufferSize];
	tempBuffer = new AmbisonicData[bufferSize];

	for(i=0;i<gravityObjects.size();++i)
		gravityObjects[i]->setBlockSize(newSize);
}

//------------------------------------------------------------------------------
void QuadManager2::splitOff(int type)
{
	/*if(mainGravityObject->getNumQuads() > 8)
		split = type;*/
	if(type == 2)
	{
		if(gravityObjects.size() < 6)
			split = type;
	}
	else
	{
		if(gravityObjects.size() < 11)
			split = type;
	}
}

//------------------------------------------------------------------------------
void QuadManager2::addGravityObject(string type, float life, int index)
{
	int count;
	unsigned int m;
	Quad2 *tempQuad;
	GravityObject2 *tempObj = 0;
	vector<Quad2 *>& mainQuads = mainGravityObject->getPrimaryQuads();

	if(type == "Planet")
		tempObj = new PlanetGO(); //Should this really be here?
	else if(type == "Pulsating")
		tempObj = new PulsatingGO();
	else if(type == "Flocking")
		tempObj = new FlockingGO();
	else if(type == "Circle")
		tempObj = new CircleGO(index);
	else
		return;

	if(tempObj)
	{
		tempObj->setLifeForce(life);
		tempObj->setBlockSize(bufferSize);
	}

	gravityObjects.push_back(tempObj);

	for(m=(mainQuads.size()-1), count=8;
		((m>=0) && (count>0));
		--m, --count)
	{
		tempQuad = mainQuads[m];
		mainGravityObject->removePrimaryQuad(tempQuad);
		tempObj->addPrimaryQuad(tempQuad);
		tempQuad->setPrimaryGravityObject(tempObj);
	}
	tempObj->setPosition(mainGravityObject->getPosition());
	tempObj->setGesture(gesture);
	tempObj->setPitchContour(mainGravityObject->getPitchContour());
}

//------------------------------------------------------------------------------
void QuadManager2::reparentQuad(Quad2 *quad)
{
	float dist;
	float tempf;
	ThreeFloats temp3f;
	vector<GravityObject2 *>::iterator it;
	GravityObject2 *obj = mainGravityObject;

	dist = quad->getPosition().getDistance(mainGravityObject->getPosition());

	for(it=gravityObjects.begin();it!=gravityObjects.end();++it)
	{
		if(quad->getPrimaryGravityObject() != *it)
		{
			tempf = quad->getPosition().getDistance((*it)->getPosition());

			if(tempf < dist)
			{
				dist = tempf;
				obj = *it;
				break;
			}
		}
	}

	obj->addPrimaryQuad(quad);
	quad->setPrimaryGravityObject(obj);
}

//------------------------------------------------------------------------------
void QuadManager2::rotateScene(bool dir)
{
	float tempf;

	if(rotIndex == 0.0f)
	{
		if(!dir)
		{
			sceneRot = -90.0f;
			tempf = -90.0f * (6.283185306f/360.0f);
		}
		else
		{
			sceneRot = 90.0f;
			tempf = 90.0f * (6.283185306f/360.0f);
		}
		scenePos.x = 8.0f * sinf(tempf);
		scenePos.z = 8.0f * cosf(tempf);
		rotIndex = 0.025f;
		rotDirection = 0;
	}
}

//------------------------------------------------------------------------------
void QuadManager2::setPitches(const vector<float>& val)
{
	mainGravityObject->setPitches(val);
}

//------------------------------------------------------------------------------
void QuadManager2::setLineColour(bool val)
{
	lineColour = val;
}

}
