//	QuadManager2.h - Second version of the QuadManager class.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef QUADMANAGER2_H_
#define QUADMANAGER2_H_

#include "Quad2.h"
#include "GravityObject2.h"
#include "Parameter.h"
#include "AshitakaSparks.h"
#include "CriticalSection.h"

#include <iostream>

struct AmbisonicData;

namespace AshitakaClasses
{

///	Second version of the QuadManager class.
class QuadManager2
{
  public:
	///	Constructor.
	QuadManager2();
	///	Destructor.
	~QuadManager2();

	///	Used to initialise some bits and pieces.
	void initialise();

	///	Updates the various positions etc.
	void update();
	///	Draws all the Quads.
	void draw();

	///	Returns the audio from all the extra GOs.
	AmbisonicData *getAudio(int numSamples, const ThreeFloats& pos);
	///	Used to set the size of the buffer returned by getAudio().
	void setBlockSize(int newSize);

	///	Temporary?
	void splitOff(int type);

	///	Adds a new GravityObject to the scene.
	/*!
		\param type What type of GravityObject to create.
		\param life How much life force to assign the GravityObject.
		\param index Only used with CircleGO, I think?
	 */
	void addGravityObject(std::string type, float life, int index = 0);
	///	Called to re-assign a Quad to a new primary GravityObject.
	/*!
		This should be called by GravityObjects when they die for all their
		primary Quads.
	 */
	void reparentQuad(Quad2 *quad);

	///	Tells us to rotate the scene (not used?).
	/*!
		\param dir False == left, true == right.
	 */
	void rotateScene(bool dir);

	///	Sets the pitches to use.
	void setPitches(const std::vector<float>& val);

	///	Sets the colour of any connecting lines (black or white).
	void setLineColour(bool val);

	//--Parameters--------------------------------------------------------------
	///	X-axis position of the primary GO.
	Parameter mainXPos;
	///	Y-axis position of the primary GO.
	Parameter mainYPos;
	///	Z-axis position of the primary GO.
	Parameter mainZPos;

	///	The Lengthwise Potentiometer.
	Parameter length;
	///	The rotational Potentiometer.
	Parameter twist;
	///	Top-left force sensor.
	Parameter topLeftForce;
	///	Top-right force sensor.
	Parameter topRightForce;
	///	Bottom-left force sensor.
	Parameter bottomLeftForce;
	///	Bottom-right force sensor.
	Parameter bottomRightForce;
	//--------------------------------------------------------------------------

	///	The number of Quads in the space.
	enum
	{
		NumQuads = 128
	};
  private:
	///	Helper method to Handle the collision detection.
	/*!
		We don't do anything special, but hopefully it's enough to provide a
		reasonable simulation of collisions between Quads.

		We treat the Quads as spheres for now, so the detection part is really
		easy - will maybe need to improve on this in the future (detect spheres
		initially; if they overlap, do more detailed detection).

		When we've detected a collision, we just swap the Quads' directions and
		velocities, which should be enough to make things look relatively
		realistic(?).
	 */
	inline void collisionDetection()
	{
		float distance;
		float vel1, vel2;
		unsigned int i, j, k;
		ThreeFloats temp3f;
		const ThreeFloats origin;

		for(i=0;i<NumQuads;++i)
		{
			for(j=(i+1);j<NumQuads;++j)
			{
				distance = quads[i].getPosition().getDistance(quads[j].getPosition());

				if(distance < (quads[i].getRadius() + quads[j].getRadius()))
				{
					//Swap direction.
					//(Check if directions are roughly the same)
					temp3f = quads[i].getDirection();
					distance = temp3f.getDistance(quads[j].getDirection());
					if(distance > 0.05f)
					{
						quads[i].setDirection(quads[j].getDirection());
						quads[j].setDirection(temp3f);
					}
					else
					{
						//Flip quads[i].getDirection() so it's definitely going
						//in (roughly) the opposite direction to
						//quads[i].getDirection().

						//Rotate temp3f by 90 degrees.
						/*float tempf = temp3f.x;
						temp3f.x = (temp3f.z * cosf(1.570796327f)) -
								   (tempf * sinf(1.570796327f));
						temp3f.z = (temp3f.z * sinf(1.570796327f)) -
								   (tempf * cosf(1.570796327f));*/
						temp3f.x = -temp3f.x;
						temp3f.y = -temp3f.y;
						temp3f.z = -temp3f.z;
						quads[i].setDirection(quads[j].getDirection());
						quads[j].setDirection(temp3f);
					}

					//'Swap' velocity.
					vel1 = quads[i].getVelocity();
					if(vel1 < 0.1f)
						vel1 += 0.1f;
					//*2 because it seems to stop Quads getting caught on each
					//other (not a very scientific solution, admittedly).
					vel2 = quads[j].getVelocity() /* 2.0f*/;
					if(vel2 < 0.1f)
						vel2 += 0.1f;

					//Hopefully this will be enough to separate them?
					temp3f = quads[i].getDirection() * (vel2 * 2.0f);
					temp3f += quads[i].getPosition();
					quads[i].moveTo(temp3f, 1.0f);
					temp3f = quads[j].getDirection() * (vel1 * 2.0f);
					temp3f += quads[j].getPosition();
					quads[j].moveTo(temp3f, 1.0f);

					//Make sure they actually are apart (?).
					/*distance = quads[i].getNewPosition().getDistance(quads[j].getNewPosition());
					while(distance < (quads[i].getRadius() + quads[j].getRadius()))
					{
						temp3f = quads[i].getDirection() * (vel2 * 2.0f);
						temp3f += quads[i].getNewPosition();
						quads[i].moveTo(temp3f, 1.0f);
						temp3f = quads[j].getDirection() * (vel1 * 2.0f);
						temp3f += quads[j].getNewPosition();
						quads[j].moveTo(temp3f, 1.0f);
						distance = quads[i].getNewPosition().getDistance(quads[j].getNewPosition());
					}*/

					quads[i].interpolate();
					quads[j].interpolate();

					//Pluck appropriate strings.
					quads[i].getPrimaryGravityObject()->collide();
					if(quads[i].getPrimaryGravityObject() != quads[j].getPrimaryGravityObject())
						quads[j].getPrimaryGravityObject()->collide();

					//Generate some sparks.
					temp3f = quads[j].getPosition() - quads[i].getPosition();
					temp3f *= 0.5f;
					temp3f += quads[i].getPosition();
					for(k=0;k<NumSparks;++k)
					{
						if(!sparks[k]->isActive())
						{
							sparks[k]->setColour(quads[i].getPrimaryGravityObject()->getMainColour());
							sparks[k]->go(temp3f);
							break;
						}
					}
				}
			}
		}
	};

	///	The number of sparks we can use (must be a power of two).
	enum
	{
		NumSparks = 64
	};

	///	The main performer-controlled GravityObject.
	/*!
		mainGravityObject is just a copy of gravityObjects[0], which is never
		deleted.
	 */
	GravityObject2 *mainGravityObject;
	///	Vector of any other GravityObjects.
	std::vector<GravityObject2 *> gravityObjects;
	///	All the Quads in the space.
	Quad2 quads[NumQuads];

	///	The audio buffer returned by getAudio().
	AmbisonicData *mainBuffer;
	///	Temporary audio buffer.
	AmbisonicData *tempBuffer;
	///	The size of mainBuffer and tempBuffer.
	int bufferSize;

	///	temp.
	int collisionCount;

	///	Used to split off new GOs.
	int split;
	///	Used to make sure we don't split off too many GOs.
	int splitCount;

	///	Our SparkParticles.
	AshitakaSparks *sparks[NumSparks];
	///	Texture for SparkParticles.
	GLuint sparkTexture;

	///	The previous 10 positions of the main GO (used to detect gestures).
	ThreeFloats gesture[10];
	///	Counter used to determine when to update gesture.
	int gestureCount;

	///	Used to rotate the scene.
	float sceneRot;
	///	Used to rotate the scene.
	ThreeFloats scenePos;
	///	If we're moving to sceneRot(0) or back to the centre(2).
	int rotDirection;
	///	How much we're rotated at the moment.
	float rotIndex;

	///	Used to store the current set of pitches to use.
	std::vector<float> pitches;

	///	The colour to draw any connecting lines.
	bool lineColour;

	///	Critical section protecting the GOs in the audio thread.
	CriticalSection critSec;
};

}

#endif
