//	VisualSynth.h - The visual synthesizer part of Ashitaka.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef VISUALSYNTH_H_
#define VISUALSYNTH_H_

#include "Parameter.h"
#include "ThreeFloats.h"
#include "CriticalSection.h"

#include "GLHeaders.h"
#include <cmath>

namespace AshitakaClasses
{

///	The visual synthesizer part of Ashitaka.
/*!
	\ingroup libheilanextras
 */
class VisualSynth
{
  public:
	///	Constructor.
	VisualSynth();
	///	Destructor.
	~VisualSynth();

	///	Draws the output to the screen via OpenGL.
	void draw();

	//--Inputs--
	///	X-axis position of the main object.
	Parameter posX;
	///	Y-axis position of the main object.
	Parameter posY;
	///	Z-axis position of the main object.
	Parameter posZ;
	///	X-axis size of the main object.
	Parameter sizeX;
	///	Y-axis size of the main object.
	Parameter sizeY;
	///	Z-axis size of the main object.
	Parameter sizeZ;

	///	Visibility of the (AudioSynth) physically-modelled string.
	Parameter stringVisibility;

	///	How much the main object is twisted.
	Parameter twist;

	///	Force applied at the 1st point on the Tao surface.
	Parameter force1;
	///	Force applied at the 2nd point on the Tao surface.
	Parameter force2;
	///	Force applied at the 3rd point on the Tao surface.
	Parameter force3;
	///	Force applied at the 4th point on the Tao surface.
	Parameter force4;

	//--Outputs--
	///	Distance between max and min points on the Tao surface.
	Parameter maxMinDistance;
	///	*3D* size of the object.
	Parameter size3D;
  private:
	///	Helper method to calculate the normal of a polygon.
	inline void calcNormal(const ThreeFloats& v1,
						   const ThreeFloats& v2,
						   const ThreeFloats& v3)
	{
		ThreeFloats a;
		ThreeFloats b;
		ThreeFloats normal;
		float length;

		a.x = v1.x - v2.x;
		a.y = v1.y - v2.y;
		a.z = v1.z - v2.z;

		b.x = v2.x - v3.x;
		b.y = v2.y - v3.y;
		b.z = v2.z - v3.z;

		normal.x = (a.y * b.z) - (a.z * b.y);
		normal.y = (a.z * b.x) - (a.x * b.z);
		normal.z = (a.x * b.y) - (a.y * b.x);

		length = sqrtf((normal.x*normal.x)+(normal.y*normal.y)+(normal.z*normal.z));
		if(length == 0.0f)
			length = 1.0f;
		else
			length = 1.0f/length;

		glNormal3f((normal.x*length), (normal.y*length), (normal.z*length));
	};
	///	Second version, to return the normal instead of drawing it.
	inline ThreeFloats calcNormal2(const ThreeFloats& v1,
								   const ThreeFloats& v2,
								   const ThreeFloats& v3)
	{
		ThreeFloats a;
		ThreeFloats b;
		ThreeFloats normal;
		float length;

		a.x = v1.x - v2.x;
		a.y = v1.y - v2.y;
		a.z = v1.z - v2.z;

		b.x = v2.x - v3.x;
		b.y = v2.y - v3.y;
		b.z = v2.z - v3.z;

		normal.x = (a.y * b.z) - (a.z * b.y);
		normal.y = (a.z * b.x) - (a.x * b.z);
		normal.z = (a.x * b.y) - (a.y * b.x);

		length = sqrtf((normal.x*normal.x)+(normal.y*normal.y)+(normal.z*normal.z));
		if(length == 0.0f)
			length = 1.0f;
		else
			length = 1.0f/length;

		normal.x *= length;
		normal.y *= length;
		normal.z *= length;

		return normal;
	};
	///	To calculate smooth normals.
	inline ThreeFloats calcSmooth(ThreeFloats& n1,
								  ThreeFloats& n2,
								  ThreeFloats& n3,
								  ThreeFloats& n4)
	{
		ThreeFloats retval;

		retval.x = (n1.x + n2.x + n3.x + n4.x)*0.25f;
		retval.y = (n1.y + n2.y + n3.y + n4.y)*0.25f;
		retval.z = (n1.z + n2.z + n3.z + n4.z)*0.25f;

		return retval;
	};

	///	Helper method to calculate a vertex on the sphere.
	void calcVertex(int x, int y);
	///	Helper method for calcVertex().
	inline ThreeFloats& getVertex(int x, int y)
	{
		if(x >= NumSections)
			x -= NumSections;
		if(y >= NumSections)
			y -= NumSections;

		return vertices[(y*NumSections)+x].sculptPos;
	}

	///	Helper method to update vertices according to the Tao surface.
	void updateVertices2();

	///	Linear interpolation helper method.
	/*!
		Assumes val is of the range 0->1.
	 */
	inline float linearInterp(float left, float right, float val)
	{
		float retval;

		retval = left + (val*(right-left));

		return retval;
	};
	///	B-spline interpolation helper method.
	inline float bSplineInterp(float x0,
							   float x1,
							   float x2,
							   float x3,
							   float val)
	{
		// 4-point, 3rd-order B-spline (x-form)
		float ym1py1 = x0+x2;
		float c0 = (1.0f/6.0f) * ym1py1 + (2.0f/3.0f) * x1;
		float c1 = (1.0f/2.0f) * (x2-x0);
		float c2 = (1.0f/2.0f) * ym1py1 - x1;
		float c3 = (1.0f/2.0f) * (x1-x2) + (1.0f/6.0f) * (x3-x0);
		return ((c3*val+c2)*val+c1)*val+c0;
	};
	///	Helper method to subdivide and draw a quad.
	/*!
		\param x The x position within positions of the lower left vertex.
		\param y The y position within positions of the lower left vertex.
		\param vertexCount The vertex in vertices that we're focussing on.
	 */
	inline void bezierSubdivide(int x, int y, int vertexCount)
	{
		int i, j;
		float tempi, tempj;
		float temp1, temp2, temp3, temp4;
		int tempX, tempY;
		float tempVertices[NumSubdivisions+1][NumSubdivisions+1];
		float corners[4][4];
		int tempint;

		//Calculate corner positions.
		for(i=-1;i<3;++i)
		{
			for(j=-1;j<3;++j)
			{
				tempX = x + j;
				if(tempX < 0)
					tempX += SurfaceWidth;
				else if(tempX >= SurfaceWidth)
					tempX -= SurfaceWidth;

				tempY = y + i;
				if(tempY < 0)
					tempY += SurfaceHeight;
				else if(tempY >= SurfaceHeight)
					tempY -= SurfaceHeight;

				corners[i+1][j+1] = positions[tempX + (tempY*SurfaceWidth)] * 2.0f;
			}
		}

		//y-axis
		for(i=0;i<NumSubdivisions;++i)
		{
			tempi = (float)i/(float)(NumSubdivisions-1);

			//x-axis
			for(j=0;j<NumSubdivisions;++j)
			{
				tempj = (float)j/(float)(NumSubdivisions-1);

				if(spikyMode)
				{
					temp1 = linearInterp(corners[1][1], corners[1][2], tempj);
					temp2 = linearInterp(corners[2][1], corners[2][2], tempj);
					tempVertices[j][i] = linearInterp(temp1, temp2, tempi);
				}
				else
				{
					temp1 = bSplineInterp(corners[0][0],
										  corners[0][1],
										  corners[0][2],
										  corners[0][3],
										  tempj);
					temp2 = bSplineInterp(corners[1][0],
										  corners[1][1],
										  corners[1][2],
										  corners[1][3],
										  tempj);
					temp3 = bSplineInterp(corners[2][0],
										  corners[2][1],
										  corners[2][2],
										  corners[2][3],
										  tempj);
					temp4 = bSplineInterp(corners[3][0],
										  corners[3][1],
										  corners[3][2],
										  corners[3][3],
										  tempj);
					tempVertices[j][i] = bSplineInterp(temp1,
													   temp2,
													   temp3,
													   temp4,
													   tempi);
				}
			}
		}

		tempint = vertexCount;
		if(spikyMode)
		{
			float thing1 = 0.0f;
			float thing2 = 0.0f;
			for(i=0;i<NumSubdivisions;++i)
			{
				for(j=0;j<NumSubdivisions;++j)
				{
					vertices[tempint].sculptPos.x = vertices[tempint].defaultPos.x *
													(1.0f + tempVertices[j][i]) *
													(thing1*thing2);
					vertices[tempint].sculptPos.z = vertices[tempint].defaultPos.z *
													(1.0f + tempVertices[j][i]) *
													(thing1*thing2);
					vertices[tempint].sculptPos.y = vertices[tempint].defaultPos.y *
													//(thing1*thing2);
													thing1;
					++tempint;
					thing1 = 1.0f-thing1;
				}
				tempint -= NumSubdivisions;
				tempint += NumSections;
				thing2 = 1.0f-thing2;
			}

		}
		else
		{
			for(i=0;i<NumSubdivisions;++i)
			{
				for(j=0;j<NumSubdivisions;++j)
				{
					vertices[tempint].sculptPos.x = vertices[tempint].defaultPos.x *
													(1.0f + tempVertices[j][i]);
					vertices[tempint].sculptPos.z = vertices[tempint].defaultPos.z *
													(1.0f + tempVertices[j][i]);
					vertices[tempint].sculptPos.y = vertices[tempint].defaultPos.y;
					//++vertexCount;
					++tempint;
				}
				tempint -= NumSubdivisions;
				tempint += NumSections;
			}
		}
	};

	///	Struct representing a vertex in the sphere.
	struct Vertex
	{
		///	The Vertex' current position.
		ThreeFloats pos;
		///	The Vertex' default position.
		ThreeFloats defaultPos;
		///	The Vertex' sculpted position.
		ThreeFloats sculptPos;
	};

	///	Enum containing the number of sides we use to draw the sphere.
	enum
	{
		SurfaceWidth = 5,
		SurfaceHeight = 5,
		NumSubdivisions = 6,
		NumSections = (NumSubdivisions*SurfaceWidth),
		TextureSize = 256, //Width and height both equal TextureSize.
		TrailLength = 256
	};

	///	Used to protect the vertices while they're being drawn.
	CriticalSection critSec;

	///	Whether the sphere's in spiky mode (linear interp), or smooth mode (bezier).
	bool spikyMode;

	///	All the Vertices in the sphere.
	Vertex vertices[NumSections*NumSections];
	///	Current left sculpt length position.
	int leftSculptL;
	///	Current left sculpt twist position.
	int leftSculptT;
	///	Current right sculpt length position.
	int rightSculptL;
	///	Current right sculpt twist position.
	int rightSculptT;

	///	Used to update the vertices in graphicDoStuff() when necessary.
	bool verticesNeedUpdated;

	///	Delay line for the surface (i.e. the positions of the cells).
	float positions[SurfaceWidth*SurfaceHeight];
	///	Forces for the string cells.
	float forces[SurfaceWidth*SurfaceHeight];
	///	Velocities for the string cells.
	float velocities[SurfaceWidth*SurfaceHeight];
	///	Mass of the individual cells.
	float mass;
	///	Inverse mass of the individual cells.
	float invMass;
	///	How much to multiply the cells' velocities by.
	float velocityMultiplier;
	///	The default velocity multiplier.
	float defaultVelocityMultiplier;
	///	The subdivided vertex positions from the surface.
	float vertexPositions[SurfaceWidth*NumSubdivisions*SurfaceHeight*NumSubdivisions];

	///	Our array of trail segments.
	ThreeFloats trail[TrailLength];
	///	Current write position within trail.
	int writePos;

	/// Test.
	ThreeFloats vel;
};

}

#endif
