//	Ashitaka2.cpp - The Ashitaka2 audiovisual instrument.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Ashitaka2.h"
#include "StringParser.h"
#include "Camera.h"
#include "TwoFloats.h"
#include "AmbisonicData.h"
#include "AttributeHelper.h"

#include <iostream>
#include <cassert>
#include <sstream>

using namespace std;
using namespace AshitakaClasses;

namespace HeilanX3D
{

NodeConstructor<Ashitaka2> Ashitaka2::constructor;

//------------------------------------------------------------------------------
Ashitaka2::Ashitaka2():
ChildNode(),
audioBuffer(0),
bufferSize(0),
pause(false)
{
	setType("Ashitaka2");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
Ashitaka2::~Ashitaka2()
{
	if(audioBuffer)
		delete [] audioBuffer;
}

//------------------------------------------------------------------------------
void Ashitaka2::initialise()
{
	quadManager.initialise();

	//Audio settings.
	setMaxFrontAudioRadius(50.0f);
	setMaxBackAudioRadius(50.0f);
	setMinFrontAudioRadius(25.0f);
	setMinBackAudioRadius(25.0f);

	//setupMappingsMIDI();
	setupMappingsInterface();
}

//------------------------------------------------------------------------------
void Ashitaka2::graphicDoStuff()
{
	if(!pause)
		quadManager.update();
	quadManager.draw();
}

//------------------------------------------------------------------------------
AmbisonicData *Ashitaka2::getAmbisonicAudio(int numSamples,
										    const ThreeFloats& pos)
{
	AmbisonicData *goAudio = 0;

	//Handle main GO.
	goAudio = quadManager.getAudio(numSamples, pos);

	return goAudio;
}

//------------------------------------------------------------------------------
void Ashitaka2::setBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
	quadManager.setBlockSize(newSize);
}

//------------------------------------------------------------------------------
void Ashitaka2::setSamplerate(float val)
{
	//audioSynth.setSamplerate(val);
}

//------------------------------------------------------------------------------
void Ashitaka2::setAttribute(const string& attribute, EventValue& val)
{
	int i;
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "pitches")
	{
		StringParser parser(val);
		AttributeHelper< vector<float> > helper(&pitches);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else
		{
			if(pitches.size() > 0)
				pitches.clear();

			for(i=0;i<parser.getNumNumbers();++i)
				pitches.push_back(parser.getFloat(i));
		}

		quadManager.setPitches(pitches);
		quadManager.twist = quadManager.twist; //So our current pitch gets updated.
	}
	else if(realAtt == "accelX")
		xAccel = val;
	else if(realAtt == "accelY")
		yAccel = val;
	else if(realAtt == "accelZ")
		zAccel = val;
	else if(realAtt == "length")
		quadManager.length = ((1.0f-(float)val)*0.499f)+0.5f;
	else if(realAtt == "twist")
		quadManager.twist = val;
	else if(realAtt == "force1")
		quadManager.topLeftForce = val;
	else if(realAtt == "force2")
		quadManager.topRightForce = val;
	else if(realAtt == "force3")
		quadManager.bottomLeftForce = val;
	else if(realAtt == "force4")
		quadManager.bottomRightForce = val;
	else if(realAtt == "split1")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
		{
			if(parser.getBool())
				quadManager.splitOff(0);
		}
	}
	else if(realAtt == "split2")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
		{
			if(parser.getBool())
				quadManager.splitOff(1);
		}
	}
	else if(realAtt == "split3")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
		{
			if(parser.getBool())
				quadManager.splitOff(2);
		}
	}
	else if(realAtt == "rotateLeft")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
		{
			if(parser.getBool())
				quadManager.rotateScene(false);
		}
	}
	else if(realAtt == "rotateRight")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
		{
			if(parser.getBool())
				quadManager.rotateScene(true);
		}
	}
	else if(attribute == "pause")
	{
		StringParser parse(val);

		if(parse.getNumBools() > 0)
			pause = parse.getBool();
	}
	else if(attribute == "backColour")
	{
		StringParser parse(val);

		if(parse.getNumThreeFloats() > 0)
		{
			ThreeFloats temp3f = parse.getThreeFloats();

			quadManager.setLineColour((temp3f.x + temp3f.y + temp3f.z) < 1.66f);
		}
	}
}

//------------------------------------------------------------------------------
void Ashitaka2::setupMappingsMIDI()
{
	
}

//------------------------------------------------------------------------------
void Ashitaka2::setupMappingsInterface()
{
	xAccel >> xAxisConvert >> quadManager.mainXPos;
	yAccel >> yAxisConvert >> quadManager.mainYPos;
	zAccel >> zAxisConvert >> quadManager.mainZPos;
}

}
