//	Ashitaka2.h - The Ashitaka2 audiovisual instrument.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef ASHITAKA2_H_
#define ASHITAKA2_H_

#include "AbstractNode.h"
#include "Ashitaka/QuadManager2.h"
#include "Ashitaka/Accel2Pos.h"

/*+
	Name: Ashitaka2
	Type: ChildNode.
	Description: The second (and current) attempt at my Ashitaka audiovisual
				 instrument, which I'm currently developing for my PhD.  I don't
				 know how easy it is to play it without the custom interface
				 I've built for it...
	Attribute:	length
	AttType:	SFFloat (input)
	AttDesc:	The lengthwise potentiometer on the interface.

	Attribute:	twist
	AttType:	SFFloat (input)
	AttDesc:	The rotational potentiometer on the interface.

	Attribute:	force1
	AttType:	SFFloat (input)
	AttDesc:	The 1st force sensor on the interface.

	Attribute:	force2
	AttType:	SFFloat (input)
	AttDesc:	The 2nd force sensor on the interface.

	Attribute:	force3
	AttType:	SFFloat (input)
	AttDesc:	The 3th force sensor on the interface.

	Attribute:	force4
	AttType:	SFFloat (input)
	AttDesc:	The 4th force sensor on the interface.

	Attribute:	accelX
	AttType:	SFFloat (input)
	AttDesc:	The x-axis accelerometer on the interface.

	Attribute:	accelY
	AttType:	SFFloat (input)
	AttDesc:	The y-axis accelerometer on the interface.

	Attribute:	accelZ
	AttType:	SFFloat (input)
	AttDesc:	The z-axis accelerometer on the interface.

	Attribute:	split1
	AttType:	SFBool (input)
	AttDesc:	Splits a PulsatingGO off the main PlanetGO when a TRUE value is
				received.

	Attribute:	split2
	AttType:	SFBool (input)
	AttDesc:	Splits a FlockingGO off the main PlanetGO when a TRUE value is
				received.

	Attribute:	split3
	AttType:	SFBool (input)
	AttDesc:	Splits a CircleGO off the main PlanetGO when a TRUE value is
				received.

	Attribute:	rotateLeft
	AttType:	SFBool (input)
	AttDesc:	Rotates the instrument round by 90 degrees when a TRUE value is
				received.

	Attribute:	rotateRight
	AttType:	SFBool (input)
	AttDesc:	Rotates the instrument round by 90 degrees when a TRUE value is
				received.

	Example:
&lt;?xml version='1.0' encoding='utf-8'?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile='Full'&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type='ANY' headlight='FALSE'/&gt;

		&lt;Ashitaka2 DEF='bob'/&gt;

		&lt;!--This stuff lets you split off new GOs using the number keys.--&gt;
		&lt;KeySensor DEF='keyS'/&gt;
		&lt;KeyTrigger DEF='keyT'/&gt;
		&lt;ROUTE fromNode='keyS' fromField='isActive' toNode='keyT' toField='set_activeVal'/&gt;
		&lt;ROUTE fromNode='keyS' fromField='keyPress' toNode='keyT' toField='set_key'/&gt;
		&lt;ROUTE fromNode='keyS' fromField='keyRelease' toNode='keyT' toField='set_key'/&gt;
		&lt;ROUTE fromNode='keyT' fromField='1_changed' toNode='bob' toField='split1'/&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	The Ashitaka2 audiovisual instrument.
/*!
	\ingroup libheilanextras
 */
class Ashitaka2 : public ChildNode
{
  public:
	///	Constructor.
	Ashitaka2();
	///	Destructor.
	~Ashitaka2();

	///	To initialise some stuff.
	void initialise();

	///	Draws the instrument's visuals.
	void graphicDoStuff();
	///	Calculates the audio from the various parts of the instrument,
	AmbisonicData *getAmbisonicAudio(int numSamples, const ThreeFloats& pos);

	///	Overridden to do nothing (for now).
	void setBlockSize(int newSize);
	///	Called to update the samplerate.
	void setSamplerate(float val);

	///	Used to set this node's attributes.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<Ashitaka2> constructor;
  private:
	///	Helper method to set up the mappings (fader box).
	void setupMappingsMIDI();
	///	Helper method to set up the mappings (interface).
	void setupMappingsInterface();

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;

	///	Handles the Quads and GravityObjects.
	AshitakaClasses::QuadManager2 quadManager;

	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;

	///	Used to store the current set of pitches to use.
	std::vector<float> pitches;

	///	I don't understand why this is necessary...
	Parameter xAccel;
	///	I don't understand why this is necessary...
	Parameter yAccel;
	///	I don't understand why this is necessary...
	Parameter zAccel;
	///	Used to convert x-axis accelerometer data to x position.
	Accel2Pos xAxisConvert;
	///	Used to convert y-axis accelerometer data to y position.
	Accel2Pos yAxisConvert;
	///	Used to convert z-axis accelerometer data to z position.
	Accel2Pos zAxisConvert;

	///	Whether or not to pause the instrument's operation.
	bool pause;
};
	
}

#endif
