//	Atom.cpp - A fuzzy 'atom'.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Atom.h"
#include "Camera.h"
#include "StringParser.h"
#include "AttributeHelper.h"
#include "AmbisonicData.h"

#include <cassert>
#include <iostream>
#include <map>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<Atom> Atom::constructor;

//------------------------------------------------------------------------------
Atom::Atom():
ChildNode(),
orbit(0.0f),
orbitSpeed(0.01f),
orbitMagnitude(1.0f),
energy(1.0f),
numNeighbours(0),
xSection(0),
ySection(0),
audioBuffer(0),
bufferSize(0),
lastAtten(0.0f)
{
	setType("Atom");
	setVisual(true);
}

//------------------------------------------------------------------------------
Atom::~Atom()
{
	if(audioBuffer)
		delete [] audioBuffer;
}

//------------------------------------------------------------------------------
void Atom::initialise()
{
	int x, y;
	float tempf;
	unsigned char *pixelPointer = pixels;

	rand();
	orbitVector.x = (float)rand()/(float)RAND_MAX;
	orbitVector.y = 1.0f-orbitVector.x;
	orbitVector.z = 0.0f;
	orbitSpeed = 0.1f;

	//Generate texture.
	for(y=0;y<256;++y)
	{
		for(x=0;x<256;++x)
		{
			tempf = sinf(((float)x/256.0f)*3.1415926535897932384626433832795f);
			tempf *= sinf(((float)y/256.0f)*3.1415926535897932384626433832795f);
			tempf = (exp(tempf)-1.0f)/(exp(1.0f)-1.0f)*1.5f-0.1f;
			if(tempf < 0.0f)
				tempf = 0.0f;
			else if(tempf > 1.0f)
				tempf = 1.0f;
			*pixelPointer = 255;
			*(pixelPointer+1) = (unsigned char)(tempf * 255.0f);
			pixelPointer += 2;
		}
	}

	//Setup texture.
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexImage2D(GL_TEXTURE_2D,
				 0,
				 2,
				 256,
				 256,
				 0,
				 GL_LUMINANCE_ALPHA,
				 GL_UNSIGNED_BYTE,
				 pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);
}

//------------------------------------------------------------------------------
void Atom::graphicDoStuff()
{
	float tempf;
	ThreeFloats newPos;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	//Update energy.
	if(!numNeighbours)
	{
		if(energy > 0.25f)
			energy -= 0.0005f;
		else
			energy = 0.25f;
	}
	else
		energy += (float)numNeighbours * 0.25f;

	tempf = sinf(orbit) * orbitMagnitude * (1.0f/((float)numNeighbours+1.0f));
	newPos = orbitVector * tempf;

	orbit += orbitSpeed * (1.0f/((float)numNeighbours+1.0f));
	if(orbit > 6.283185307179586476925286766559f)
		orbit = fmod(orbit, 6.283185307179586476925286766559f);

	glTranslatef(centre.x, centre.y, centre.z);
	glTranslatef(newPos.x, newPos.y, newPos.z);

	//Update xSection and ySection.
	xSection = (int)(newPos.x + centre.x);
	ySection = (int)(newPos.y + centre.y);

	//glBlendFunc(GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);

	glBindTexture(GL_TEXTURE_2D, texture);

	glDisable(GL_LIGHTING);

	glBegin(GL_QUADS);
	{
		glColor3f((0.5f * (1.0f-(energy * 0.25f)) + (0.2f * (energy * 0.25f))),
				  (0.5f * (1.0f-(energy * 0.25f)) + (0.2f * (energy * 0.25f))),
				  (0.5f * (1.0f-(energy * 0.25f)) + (0.2f * (energy * 0.25f))));
		glTexCoord2f(0.0f, 1.0f);
		glVertex3f(-0.5f, 0.5f, 0.0f);
		glTexCoord2f(1.0f, 1.0f);
		glVertex3f(0.5f, 0.5f, 0.0f);
		glTexCoord2f(1.0f, 0.0f);
		glVertex3f(0.5f, -0.5f, 0.0f);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(-0.5f, -0.5f, 0.0f);
	}
	glEnd();

	glBindTexture(GL_TEXTURE_2D, 0);
	glEnable(GL_LIGHTING);
	glDisable(GL_BLEND);

	glPopMatrix();
}

//------------------------------------------------------------------------------
/*AmbisonicData *Atom::audioDoStuff(int numSamples, const ThreeFloats& pos)
{
	return 0;
}*/

//------------------------------------------------------------------------------
void Atom::setBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}

//------------------------------------------------------------------------------
void Atom::setAttribute(const std::string& attribute, EventValue& val)
{
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "centre")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&centre);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			centre = parser.getThreeFloats();

		xSection = (int)centre.x;
		ySection = (int)centre.y;
	}
}

//------------------------------------------------------------------------------
void Atom::incrementEnergy(float val)
{
	energy = val;
}

//------------------------------------------------------------------------------
void Atom::setNumNeighbours(int val)
{
	numNeighbours = val;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
NodeConstructor<AtomController> AtomController::constructor;

AtomController::AtomController():
ParentNode()
{
	setType("AtomController");
	setVisual(true);
}

//------------------------------------------------------------------------------
AtomController::~AtomController()
{
	
}

//------------------------------------------------------------------------------
void AtomController::setAttribute(const string& attribute, EventValue& val)
{
	
}

//------------------------------------------------------------------------------
void AtomController::preChildDraw()
{
	unsigned int i;
	vector<Atom *>::iterator it;
	vector<int> xSections;
	vector<int> ySections;
	multimap<int, Atom *> sortedAtoms;
	multimap<int, Atom *>::iterator it2;

	//First fill out x & ySections.
	for(it=atoms.begin();it!=atoms.end();++it)
	{
		bool exists = false;
		int tempX, tempY;
		int existingIndex = 0;

		tempX = (*it)->getXSection();
		tempY = (*it)->getYSection();

		for(i=0;i<xSections.size();++i)
		{
			if((tempX == xSections[i])&&(tempY == ySections[i]))
			{
				exists = true;
				existingIndex = i;
				break;
			}
		}

		if(!exists)
		{
			xSections.push_back(tempX);
			ySections.push_back(tempY);
			sortedAtoms.insert(make_pair(xSections.size()-1, (*it)));
		}
		else
			sortedAtoms.insert(make_pair(existingIndex, (*it)));
	}

	//For every section index in sortedAtoms, set their energy etc.
	for(i=0;i<xSections.size();++i)
	{
		//Get number of Atoms in this section.
		int numAtoms = sortedAtoms.count(i);

		//Go through each Atom in this section, update it.
		for(it2=sortedAtoms.lower_bound(i);
			it2!=sortedAtoms.upper_bound(i);
			++it2)
		{
			it2->second->setNumNeighbours(numAtoms-1);
			it2->second->incrementEnergy((float)(numAtoms-1) * 0.25f);
		}
	}
}

//------------------------------------------------------------------------------
void AtomController::postChildDraw()
{
	
}

//------------------------------------------------------------------------------
void AtomController::childrenLoaded()
{
	unsigned int i;
	Atom *tempAtom = 0;

	for(i=0;i<(unsigned)getNumChildren();++i)
	{
		if(getChild(i)->getType() == "Atom")
		{
			tempAtom = dynamic_cast<Atom *>(getChild(i));

			if(tempAtom)
				atoms.push_back(tempAtom);
		}
	}
}

}
