//	Atom.h - A fuzzy 'atom'.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef ATOM_H_
#define ATOM_H_

#include "ParentNode.h"

#include "GLHeaders.h"

/*+
	Name: Atom
	Type: ChildNode.
	Description: A fuzzy visual 'atom'.  Must be used as a child of
				 AtomController.  The atom modulates its motion according to
				 the number of neighbouring atoms.
	Attribute:	centre
	AttType:	SFVec3F (input)
	AttDesc:	The 3d position of the Atom.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;AtomController&gt;
			&lt;Atom centre="0.0 -1.0 -7.2"/&gt;
			&lt;Atom centre="-2.0 0.0 -7.1"/&gt;
			&lt;Atom centre="2.0 0.5 -7.0"/&gt;
			&lt;Atom DEF="bob" centre="0.0 1.0 -7.3"/&gt;
		&lt;/AtomController&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

//How do I handle two nodes in one file?
/*
	Name: AtomController
	Type: ParentNode, to be used with a number of child Atom nodes.
	Description: This node handles the various inter-Atom issues, such as
			 	 calculating how many nodes are in a particular area, and then
			 	 informing the individual nodes how many neighbours they have.
 */

namespace HeilanX3D
{

//------------------------------------------------------------------------------
///	A fuzzy 'atom'.
/*!
	\ingroup libheilanextras
 */
class Atom : public ChildNode
{
  public:
	///	Constructor.
	Atom();
	///	Destructor.
	~Atom();

	///	Initialises the texture.
	void initialise();

	///	Draws the Atom.
	void graphicDoStuff();
	///	So we can hear the Atom.
	//AmbisonicData *audioDoStuff(int numSamples, const ThreeFloats& pos);

	///	Sets the current block size to use for the audio.
	void setBlockSize(int newSize);

	///	Used to set this node's attributes.
	/*!
		Also used by child AudioClip nodes to communicate that they've loaded
		their sound file, and it's ready for streaming.
	 */
	void setAttribute(const std::string& attribute, EventValue& val);

	///	Called by AtomController to increment this Atom's energy level.
	/*!
		\param val The amount to increment our energy level by.
	 */
	void incrementEnergy(float val);
	///	Called by AtomController to set the number of neighbours in our section.
	void setNumNeighbours(int val);

	///	Returns the X section this Atom's in.
	int getXSection() const {return xSection;};
	///	Returns the Y section thisAtom's in.
	int getYSection() const {return ySection;};

	static NodeConstructor<Atom> constructor;
  private:
	///	The texture we use.
	GLuint texture;
	///	Our pixel date.
	unsigned char pixels[256*256*2];

	///	Current position in our orbit.
	float orbit;
	///	The speed of our orbit.
	float orbitSpeed;
	///	Vector along which we're orbiting.
	ThreeFloats orbitVector;
	///	The magnitude of our orbit.
	float orbitMagnitude;
	///	Our current energy level.
	float energy;
	///	The number of neighbours in our section.
	int numNeighbours;
	///	The current X section we're in.
	int xSection;
	///	The current Y section we're in.
	int ySection;

	///	The Atom's current centre position.
	ThreeFloats centre;

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;
	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;
};

//------------------------------------------------------------------------------
///	ParentNode used to deal with inter-Atom issues.
/*!
	\ingroup libheilanextras
 */
class AtomController : public ParentNode
{
  public:
	///	Constructor.
	AtomController();
	///	Destructor.
	~AtomController();

	///	Used to set this node's attributes.
	/*!
		Also used by child AudioClip nodes to communicate that they've loaded
		their sound file, and it's ready for streaming.
	 */
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<AtomController> constructor;
  private:
	///	Called before the node starts drawing its children.
	void preChildDraw();
	///	Called after the node has finished drawing its children.
	void postChildDraw();

	///	Overridden to do nothing.
	void audioBlockSizeChanged(int newSize) {};

	///	So we get informed when children Atoms are added.
	void childrenLoaded();

	///	Vector of all our children Atoms.
	std::vector<Atom *> atoms;
};

}

#endif
