//	Atom.cpp - A sphere made up of fuzzy atoms.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "AtomSphere.h"
#include "Camera.h"
#include "StringParser.h"
#include "AttributeHelper.h"
#include "AmbisonicData.h"

#include <cassert>
#include <iostream>
#include <map>
#include <algorithm>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<AtomSphere> AtomSphere::constructor;

bool zOrderSort(ThreeFloats a, ThreeFloats b)
{
	return a.z < b.z;
}

//------------------------------------------------------------------------------
AtomSphere::AtomSphere():
ChildNode(),
audioBuffer(0),
bufferSize(0),
lastAtten(0.0f)
{
	setType("AtomSphere");
	setVisual(true);
}

//------------------------------------------------------------------------------
AtomSphere::~AtomSphere()
{
	if(audioBuffer)
		delete [] audioBuffer;
}

//------------------------------------------------------------------------------
void AtomSphere::initialise()
{
	int x, y;
	float tempf;
	float tempX, tempY, tempZ;
	float normX, normY, normZ;
	unsigned char *pixelPointer = pixels;
	vector<ThreeFloats> tempPoints;

	//Generate texture.
	for(y=0;y<256;++y)
	{
		for(x=0;x<256;++x)
		{
			tempf = sinf(((float)x/256.0f)*3.1415926535897932384626433832795f);
			tempf *= sinf(((float)y/256.0f)*3.1415926535897932384626433832795f);
			tempf = (exp(tempf)-1.0f)/(exp(1.0f)-1.0f)*1.5f-0.1f;
			if(tempf < 0.0f)
				tempf = 0.0f;
			else if(tempf > 1.0f)
				tempf = 1.0f;
			*pixelPointer = 255;
			*(pixelPointer+1) = (unsigned char)(tempf * 255.0f);
			pixelPointer += 2;
		}
	}

	//Generate circle.
	for(x=0;x<NumSections;++x)
	{
		sinTable[x] = sinf(((float)x/(float)NumSections) * 6.283185306f);
		cosTable[x] = cosf(((float)x/(float)NumSections) * 6.283185306f);
	}

	//Generate tempPoints.
	for(y=0;y<NumSections;++y)
	{
		for(x=0;x<NumSections;++x)
		{
			tempX = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) * 
					cosf(((float)y/(float)NumSections) * 6.283185306f);
			tempY = sinf((((float)x/(float)NumSections)-0.5f) * 3.141592653f);
			tempZ = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) *
					sinf(((float)y/(float)NumSections) * 6.283185306f);
			tempPoints.push_back(ThreeFloats(tempX, tempY, tempZ));
		}
	}
	sort(tempPoints.begin(), tempPoints.end(), zOrderSort);

	//Setup texture.
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexImage2D(GL_TEXTURE_2D,
				 0,
				 2,
				 256,
				 256,
				 0,
				 GL_LUMINANCE_ALPHA,
				 GL_UNSIGNED_BYTE,
				 pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);

	//Generate the display list.
	atom = glGenLists(1);
	glNewList(atom, GL_COMPILE);
	{
		glBegin(GL_QUADS);
		{
			for(y=0;y<NumSections;++y)
			{
				for(x=0;x<NumSections;++x)
				{
					tempX = tempPoints[x+(y*NumSections)].x;
					tempY = tempPoints[x+(y*NumSections)].y;
					tempZ = tempPoints[x+(y*NumSections)].z;
					tempf = sqrtf((tempX*tempX)+(tempY*tempY)+(tempZ*tempZ));
					normX = tempX/tempf;
					normY = tempY/tempf;
					normZ = tempZ/tempf;
	
					glNormal3f(normX, normY, normZ);
					glTexCoord2f(0.0f, 0.0f);
					glVertex3f(tempX-0.25f, tempY-0.25f, tempZ);
					glTexCoord2f(1.0f, 0.0f);
					glVertex3f(tempX+0.25f, tempY-0.25f, tempZ);
					glTexCoord2f(1.0f, 1.0f);
					glVertex3f(tempX+0.25f, tempY+0.25f, tempZ);
					glTexCoord2f(0.0f, 1.0f);
					glVertex3f(tempX-0.25f, tempY+0.25f, tempZ);
				}
			}
		}
		glEnd();
	}
	glEndList();
}

//------------------------------------------------------------------------------
void AtomSphere::graphicDoStuff()
{
	float tempf;
	ThreeFloats newPos;

	float tempX, tempY, tempZ;
	float normX, normY, normZ;
	vector<ThreeFloats> tempPoints;

	//Generate tempPoints.
	for(int y=0;y<NumSections;++y)
	{
		for(int x=0;x<NumSections;++x)
		{
			tempX = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) * 
					cosf(((float)y/(float)NumSections) * 6.283185306f);
			tempY = sinf((((float)x/(float)NumSections)-0.5f) * 3.141592653f);
			tempZ = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) *
					sinf(((float)y/(float)NumSections) * 6.283185306f);
			tempPoints.push_back(ThreeFloats(tempX, tempY, tempZ));
		}
	}
	sort(tempPoints.begin(), tempPoints.end(), zOrderSort);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glEnable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);

	glBindTexture(GL_TEXTURE_2D, texture);

	//glCallList(atom);
		glBegin(GL_QUADS);
		{
			for(int y=0;y<NumSections;++y)
			{
				for(int x=0;x<NumSections;++x)
				{
					tempX = tempPoints[x+(y*NumSections)].x;
					tempY = tempPoints[x+(y*NumSections)].y;
					tempZ = tempPoints[x+(y*NumSections)].z;
					tempf = sqrtf((tempX*tempX)+(tempY*tempY)+(tempZ*tempZ));
					normX = tempX/tempf;
					normY = tempY/tempf;
					normZ = tempZ/tempf;
	
					glNormal3f(normX, normY, normZ);
					glTexCoord2f(0.0f, 0.0f);
					glVertex3f(tempX-0.25f, tempY-0.25f, tempZ);
					glTexCoord2f(1.0f, 0.0f);
					glVertex3f(tempX+0.25f, tempY-0.25f, tempZ);
					glTexCoord2f(1.0f, 1.0f);
					glVertex3f(tempX+0.25f, tempY+0.25f, tempZ);
					glTexCoord2f(0.0f, 1.0f);
					glVertex3f(tempX-0.25f, tempY+0.25f, tempZ);
				}
			}
		}
		glEnd();

	glBindTexture(GL_TEXTURE_2D, 0);
	glEnable(GL_DEPTH_TEST);

	glPopMatrix();
}

//------------------------------------------------------------------------------
/*AmbisonicData *AtomSphere::audioDoStuff(int numSamples, const ThreeFloats& pos)
{
	return 0;
}*/

//------------------------------------------------------------------------------
void AtomSphere::setBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}

//------------------------------------------------------------------------------
void AtomSphere::setAttribute(const std::string& attribute, EventValue& val)
{
	/*string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "centre")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&centre);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			centre = parser.getThreeFloats();

		xSection = (int)centre.x;
		ySection = (int)centre.y;
	}*/
}

}
