//	AtomSphere.h - A sphere made up of fuzzy atoms.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef ATOMSPHERE_H_
#define ATOMSPHERE_H_

#include "ParentNode.h"

#include "GLHeaders.h"

/*+
	Name: AtomSphere
	Type: ChildNode.
	Description: This node draws a sphere from a number of fuzzy 'atoms' (like
				 in Atom), instead of the usual polygons.  Tends to produce high
				 cpu usage if the camera gets close to it.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0.0 0.0 -8.0"&gt;
			&lt;AtomSphere/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;

 */

namespace HeilanX3D
{

//------------------------------------------------------------------------------
///	A sphere made up of fuzzy atoms.
/*!
	\ingroup libheilanextras
 */
class AtomSphere : public ChildNode
{
  public:
	///	Constructor.
	AtomSphere();
	///	Destructor.
	~AtomSphere();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the AtomSphere.
	void graphicDoStuff();
	///	So we can hear the AtomSphere.
	//AmbisonicData *audioDoStuff(int numSamples, const ThreeFloats& pos);

	///	Sets the current block size to use for the audio.
	void setBlockSize(int newSize);

	///	Used to set this node's attributes.
	/*!
		Also used by child AudioClip nodes to communicate that they've loaded
		their sound file, and it's ready for streaming.
	 */
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<AtomSphere> constructor;
  private:
	///	The dimensions of the sphere.
	enum
	{
		NumSections = 32
	};

	///	The texture we use.
	GLuint texture;
	///	Our pixel date.
	unsigned char pixels[256*256*2];
	///	Storage for our atom quad.
	GLuint atom;

	///	The points on a circle that we use to create the sphere.
	float sinTable[NumSections];
	///	The points on a circle that we use to create the sphere.
	float cosTable[NumSections];

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;
	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;
};

}

#endif
