//	AudioInput.h - Object allowing you to stream audio into a scene from a
//				   soundcard.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef AUDIOINPUT_H_
#define AUDIOINPUT_H_

#include "AbstractNode.h"

//#include "AvailableExtensions.h"

/*+
	Name: AudioInput
	Type: ChildNode.
	Description: Node allowing you to stream audio into a scene from a
				 soundcard.
	Attribute:	channel
	AttType:	SFInt32 (input)
	AttDesc:	Which channel of the audio input to use (use the
	--enumerateSoundcards command line flag to determine which values are
	valid).

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;Transform translation='-2 0 0'&gt;
			&lt;Shape&gt;
				&lt;Appearance&gt;
					&lt;Material diffuseColor='1 0 0'/&gt;
				&lt;/Appearance&gt;
				&lt;Box/&gt;
			&lt;/Shape&gt;
			&lt;AudioInput channel='0'/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Object allowing you to stream audio into a scene from a soundcard.
class AudioInput : public ChildNode
{
  public:
	///	Constructor.
	AudioInput();
	///	Destructor.
	~AudioInput();

	///	Overridden to do nothing.
	void graphicDoStuff() {};

	///	Outputs the audio.
	AmbisonicData *getMonoAudio(int numSamples);
	///	Sets the size of the node's temporary audio buffer.
	void setBlockSize(int newSize);

	///	Used to set the size and solidity of the box.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<AudioInput> constructor;
  private:
	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;

	///	The input audio channel to use.
	int channel;
};

}

#endif
