//	CylinderBackground.cpp - Special background node.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "CylinderBackground.h"
#include "StringParser.h"
#include "AttributeHelper.h"
#include "Camera.h"
#include "FileFinder.h"
#include "GLHeaders.h"

#include "SDL_image.h"
#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<CylinderBackground> CylinderBackground::constructor;

//------------------------------------------------------------------------------
CylinderBackground::CylinderBackground():
BindableNode(),
topColour(1.0f, 1.0f, 1.0f),
middleColour(-1.0f, -1.0f, -1.0f),
bottomColour(1.0f, 1.0f, 1.0f)
{
	setType("CylinderBackground");
	setVisual(true);
}

//------------------------------------------------------------------------------
CylinderBackground::~CylinderBackground()
{
	BindableNodeSingleton::getInstance()->unregisterNode(this);
}

//------------------------------------------------------------------------------
void CylinderBackground::initialise()
{
	int i;

	//Generate coefficients.
	for(i=0;i<NumSegments;++i)
	{
		xVerts[i] = sinf(((float)i/(float)NumSegments) * 6.283185306f);
		yVerts[i] = cosf(((float)i/(float)NumSegments) * 6.283185306f);
	}

	//Register with the BindableNodeSingleton.
	BindableNodeSingleton::getInstance()->registerNode(this);
}

//------------------------------------------------------------------------------
void CylinderBackground::graphicDoStuff()
{
	int i;
	float tempMatrix[16];
	ThreeFloats middle;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	//Move so we're centred around the camera.
	//I got this idea from http://www.levp.de/3d/index.html
	glGetFloatv(GL_MODELVIEW_MATRIX, tempMatrix);
	tempMatrix[12] = 0.0f;
	tempMatrix[13] = 0.0f;
	tempMatrix[14] = 0.0f;
	glLoadMatrixf(tempMatrix);

	//Disable writing to the z-buffer, so everything appears behind the actual
	//scene.
    glDepthMask(GL_FALSE);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);

	//Calculate middle colour if we're interpolating.
	if(middleColour.x == -1)
		middle = (topColour + bottomColour)*0.5f;
	else
		middle = middleColour;

	//Draw cylinder.
	glScalef(1.0f, 0.425f, 1.0f);
	for(i=0;i<(NumSegments-1);++i)
	{
		glBegin(GL_QUADS);
		{
			//Top quad.
			glColor3f(topColour.x, topColour.y, topColour.z);
			glVertex3f(xVerts[i+1], 1.0f, yVerts[i+1]);
			glVertex3f(xVerts[i], 1.0f, yVerts[i]);
			glColor3f(middle.x, middle.y, middle.z);
			glVertex3f(xVerts[i], 0.0f, yVerts[i]);
			glVertex3f(xVerts[i+1], 0.0f, yVerts[i+1]);

			//Middle quad.
			glColor3f(middle.x, middle.y, middle.z);
			glVertex3f(xVerts[i+1], 0.0f, yVerts[i+1]);
			glVertex3f(xVerts[i], 0.0f, yVerts[i]);
			glColor3f(bottomColour.x, bottomColour.y, bottomColour.z);
			glVertex3f(xVerts[i], -1.0f, yVerts[i]);
			glVertex3f(xVerts[i+1], -1.0f, yVerts[i+1]);
		}
		glEnd();
	}

	glBegin(GL_QUADS);
	{
		//Top quad.
		glColor3f(topColour.x, topColour.y, topColour.z);
		glVertex3f(xVerts[0], 1.0f, yVerts[0]);
		glVertex3f(xVerts[i], 1.0f, yVerts[i]);
		glColor3f(middle.x, middle.y, middle.z);
		glVertex3f(xVerts[i], 0.0f, yVerts[i]);
		glVertex3f(xVerts[0], 0.0f, yVerts[0]);

		//Middle quad.
		glColor3f(middle.x, middle.y, middle.z);
		glVertex3f(xVerts[0], 0.0f, yVerts[0]);
		glVertex3f(xVerts[i], 0.0f, yVerts[i]);
		glColor3f(bottomColour.x, bottomColour.y, bottomColour.z);
		glVertex3f(xVerts[i], -1.0f, yVerts[i]);
		glVertex3f(xVerts[0], -1.0f, yVerts[0]);
	}
	glEnd();

	glEnable(GL_LIGHTING);

	//Re-enable writing to the z-buffer.
    glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);

	glPopMatrix();
}

//------------------------------------------------------------------------------
void CylinderBackground::setAttribute(const string& attribute, EventValue& val)
{
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "topColour")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&topColour);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			topColour = parser.getThreeFloats();
	}
	else if(realAtt == "middleColour")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&middleColour);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			middleColour = parser.getThreeFloats();
	}
	else if(realAtt == "bottomColour")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&bottomColour);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			bottomColour = parser.getThreeFloats();
	}
}

}
