//	DeformableNurbs.cpp - A NURBS sphere which can be deformed in various ways.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "DeformableNurbs.h"
#include "Camera.h"
#include "StringParser.h"

using namespace std;

namespace HeilanX3D
{

NodeConstructor<DeformableNurbs> DeformableNurbs::constructor;

//------------------------------------------------------------------------------
DeformableNurbs::DeformableNurbs():
stretch(1.0f),
twist(0.0f),
leftSculpt(0.0f),
rightSculpt(0.0f),
leftGrab(0.0f),
rightGrab(0.0f),
leftStretch(0.5f),
rightStretch(0.5f),
leftSculptL(24),
leftSculptT(8),
rightSculptL(8),
rightSculptT(8),
nurbsObject(0),
verticesNeedUpdating(false)
{
	setType("DeformableNurbs");
	setVisual(true);
}

//------------------------------------------------------------------------------
DeformableNurbs::~DeformableNurbs()
{
	if(nurbsObject)
		gluDeleteNurbsRenderer(nurbsObject);
}

//------------------------------------------------------------------------------
void DeformableNurbs::initialise()
{
	int i = 0;

	//Initialise nurbsObject.
	if(!nurbsObject)
		nurbsObject = gluNewNurbsRenderer();
	if(nurbsObject)
	{
		gluNurbsProperty(nurbsObject, GLU_SAMPLING_TOLERANCE, 50.0);
		gluNurbsProperty(nurbsObject, GLU_DISPLAY_MODE, GLU_FILL);
	}

	//Initialise cos and sin tables.
	for(i=0;i<360;++i)
	{
		cosTable[i] = cosf(((float)i/360.0f) * 6.283185306f);
		sinTable[i] = cosf(((float)i/360.0f) * 6.283185306f);
	}

	//Initialise knots.
	knots[0] = 0;
	knots[1] = 0;
	knots[2] = 0;
	knots[3] = 1;
	knots[4] = 2;
	knots[5] = 2;
	knots[6] = 2;

	//Initialise left control points.
	i = 0;
	leftPoints[i] = 0.0f;
	leftPoints[i+1] = 0.0f;
	leftPoints[i+2] = -1.0f;
	i += 3;

	leftPoints[i] = 0.0f;
	leftPoints[i+1] = 0.0f;
	leftPoints[i+2] = -1.0f;
	i += 3;

	leftPoints[i] = 0.0f;
	leftPoints[i+1] = 0.0f;
	leftPoints[i+2] = -1.0f;
	i += 3;

	leftPoints[i] = 0.0f;
	leftPoints[i+1] = 0.0f;
	leftPoints[i+2] = -1.0f;
	i += 3;

	leftPoints[i] = 0.0f;
	leftPoints[i+1] = 1.0f;
	leftPoints[i+2] = -0.75f;
	i += 3;

	leftPoints[i] = -0.5f;
	leftPoints[i+1] = 1.0f;
	leftPoints[i+2] = -0.75f;
	i += 3;

	leftPoints[i] = -1.0f;
	leftPoints[i+1] = 0.5f;
	leftPoints[i+2] = -0.75f;
	i += 3;

	leftPoints[i] = -1.0f;
	leftPoints[i+1] = 0.0f;
	leftPoints[i+2] = -0.75f;
	i += 3;

	leftPoints[i] = 0.0f;
	leftPoints[i+1] = 1.0f;
	leftPoints[i+2] = 0.75f;
	i += 3;

	leftPoints[i] = -0.5f;
	leftPoints[i+1] = 1.0f;
	leftPoints[i+2] = 0.75f;
	i += 3;

	leftPoints[i] = -1.0f;
	leftPoints[i+1] = 0.5f;
	leftPoints[i+2] = 0.75f;
	i += 3;

	leftPoints[i] = -1.0f;
	leftPoints[i+1] = 0.0f;
	leftPoints[i+2] = 0.75f;
	i += 3;

	leftPoints[i] = 0.0f;
	leftPoints[i+1] = 0.0f;
	leftPoints[i+2] = 1.0f;
	i += 3;

	leftPoints[i] = 0.0f;
	leftPoints[i+1] = 0.0f;
	leftPoints[i+2] = 1.0f;
	i += 3;

	leftPoints[i] = 0.0f;
	leftPoints[i+1] = 0.0f;
	leftPoints[i+2] = 1.0f;
	i += 3;

	leftPoints[i] = 0.0f;
	leftPoints[i+1] = 0.0f;
	leftPoints[i+2] = 1.0f;
	i += 3;

	//Initialise right control points.
	i = 0;
	rightPoints[i] = 0.0f;
	rightPoints[i+1] = 0.0f;
	rightPoints[i+2] = 1.0f;
	i += 3;

	rightPoints[i] = 0.0f;
	rightPoints[i+1] = 0.0f;
	rightPoints[i+2] = 1.0f;
	i += 3;

	rightPoints[i] = 0.0f;
	rightPoints[i+1] = 0.0f;
	rightPoints[i+2] = 1.0f;
	i += 3;

	rightPoints[i] = 0.0f;
	rightPoints[i+1] = 0.0f;
	rightPoints[i+2] = 1.0f;
	i += 3;

	rightPoints[i] = 0.0f;
	rightPoints[i+1] = 1.0f;
	rightPoints[i+2] = 0.75f;
	i += 3;

	rightPoints[i] = 0.5f;
	rightPoints[i+1] = 1.0f;
	rightPoints[i+2] = 0.75f;
	i += 3;

	rightPoints[i] = 1.0f;
	rightPoints[i+1] = 0.5f;
	rightPoints[i+2] = 0.75f;
	i += 3;

	rightPoints[i] = 1.0f;
	rightPoints[i+1] = 0.0f;
	rightPoints[i+2] = 0.75f;
	i += 3;

	rightPoints[i] = 0.0f;
	rightPoints[i+1] = 1.0f;
	rightPoints[i+2] = -0.75f;
	i += 3;

	rightPoints[i] = 0.5f;
	rightPoints[i+1] = 1.0f;
	rightPoints[i+2] = -0.75f;
	i += 3;

	rightPoints[i] = 1.0f;
	rightPoints[i+1] = 0.5f;
	rightPoints[i+2] = -0.75f;
	i += 3;

	rightPoints[i] = 1.0f;
	rightPoints[i+1] = 0.0f;
	rightPoints[i+2] = -0.75f;
	i += 3;

	rightPoints[i] = 0.0f;
	rightPoints[i+1] = 0.0f;
	rightPoints[i+2] = -1.0f;
	i += 3;

	rightPoints[i] = 0.0f;
	rightPoints[i+1] = 0.0f;
	rightPoints[i+2] = -1.0f;
	i += 3;

	rightPoints[i] = 0.0f;
	rightPoints[i+1] = 0.0f;
	rightPoints[i+2] = -1.0f;
	i += 3;

	rightPoints[i] = 0.0f;
	rightPoints[i+1] = 0.0f;
	rightPoints[i+2] = -1.0f;
	i += 3;
}

//------------------------------------------------------------------------------
void DeformableNurbs::graphicDoStuff()
{
	if(verticesNeedUpdating)
	{
		verticesNeedUpdating = false;
		updateVertices();
	}

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	critSec.lock();
	{
		//Left patch.
		gluBeginSurface(nurbsObject);
		{
			gluNurbsSurface(nurbsObject,
						    7,
						    knots,
						    7,
						    knots,
						    4*3,
						    3,
						    leftPoints,
						    3,
						    3,
						    GL_MAP2_VERTEX_3);
		}
		gluEndSurface(nurbsObject);
		//Right patch.
		gluBeginSurface(nurbsObject);
		{
			gluNurbsSurface(nurbsObject,
						    7,
						    knots,
						    7,
						    knots,
						    4*3,
						    3,
						    rightPoints,
						    3,
						    3,
						    GL_MAP2_VERTEX_3);
		}
		gluEndSurface(nurbsObject);

		glScalef(1.0f, -1.0f, 1.0f);
		//Left patch (inverted y).
		gluBeginSurface(nurbsObject);
		{
			gluNurbsSurface(nurbsObject,
						    7,
						    knots,
						    7,
						    knots,
						    4*3,
						    3,
						    leftPoints,
						    3,
						    3,
						    GL_MAP2_VERTEX_3);
		}
		gluEndSurface(nurbsObject);
		//Right patch (inverted y).
		gluBeginSurface(nurbsObject);
		{
			gluNurbsSurface(nurbsObject,
						    7,
						    knots,
						    7,
						    knots,
						    4*3,
						    3,
						    rightPoints,
						    3,
						    3,
						    GL_MAP2_VERTEX_3);
		}
		gluEndSurface(nurbsObject);
	}
	critSec.unlock();

	glPopMatrix();
}

//------------------------------------------------------------------------------
void DeformableNurbs::setAttribute(const string& attribute, EventValue& val)
{
	//int i;
	StringParser parser(val);

	if(attribute == "stretch")
	{
		stretch = val;
		verticesNeedUpdating = true;
	}
	else if(attribute == "twist")
	{
		twist = val;
		verticesNeedUpdating = true;
	}
	else if(attribute == "leftSculpt")
	{
		leftSculpt = val;
		verticesNeedUpdating = true;
	}
	else if(attribute == "rightSculpt")
	{
		rightSculpt = val;
		verticesNeedUpdating = true;
	}
	else if(attribute == "leftGrab")
		leftGrab = val;
	else if(attribute == "rightGrab")
		rightGrab = val;
}

}
