//	DeformableNurbs.h - A NURBS sphere which can be deformed in various ways.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef DEFORMABLENURBS_H_
#define DEFORMABLENURBS_H_

#include "AbstractNode.h"
#include "CriticalSection.h"

#include "GLHeaders.h"
#include <cmath>

/*+
	Name: DeformableNurbs
	Type: ChildNode.
	Description: A sphere which may be deformed by manipulating it's attributes.
			 	 This is a test node for my final audiovisual instrument, hence
				 the slightly counterintuitive attributes (they make more sense
				 when considering the physical interface I'm designing for it).
	Attribute:	leftGrab
	AttType:	SFFloat (input)
	AttDesc:	'Grabs' the left side of the sphere, an operation which
				modulates the other attributes.

	Attribute:	rightGrab
	AttType:	SFFloat (input)
	AttDesc:	'Grabs' the right side of the sphere, an operation which
				modulates the other attributes.

	Attribute:	stretch
	AttType:	SFFloat (input)
	AttDesc:	When the grabs are 0, stretches the sphere along the x
				dimension.  Otherwise changes the sculpting position (x-axis) of
				the left and/or right sides.

	Attribute:	twist
	AttType:	SFFloat (input)
	AttDesc:	When the grabs are 0, twists the sphere around the x axis.
				Otherwise changes the sculpting position (rotating around the
				object) of the left and/or right sides.

	Attribute:	leftSculpt
	AttType:	SFFloat (input)
	AttDesc:	When leftGrab is non-zero, controls the radius (from the
				sphere's centre) of the vertex at the current left sculpting
				position.

	Attribute:	rightSculpt
	AttType:	SFFloat (input)
	AttDesc:	When rightGrab is non-zero, controls the radius (from the
				sphere's centre) of the vertex at the current right sculpting
				position.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY" headlight="FALSE"/&gt;

		&lt;Transform DEF="temp" translation="0.0 0.0 -8.0"&gt;
			&lt;DeformableNurbs DEF="bob"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	A sphere which can be deformed in various ways.
/*!
	\ingroup libheilanextras
 */
class DeformableNurbs : public ChildNode
{
  public:
	///	Constructor.
	DeformableNurbs();
	///	Destructor.
	~DeformableNurbs();

	///	Sets up various bits and pieces.
	void initialise();

	///	Draws the sphere.
	void graphicDoStuff();
	///	Not used, overridden to do nothing.
	/*AmbisonicData *audioDoStuff(int numSamples,
								const ThreeFloats& pos) {return 0;};*/

	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set the sphere's parameters.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<DeformableNurbs> constructor;
  private:
	inline void updateVertices()
	{
		int i = 0;
		float tempSin, tempCos;

		tempSin = sinf(twist * 6.283185306f);
		tempCos = cosf(twist * 6.283185306f);

		/*
		leftPoints[i] = leftPoints[i];
		leftPoints[i+1] = (leftPoints[i+1] * tempCos) + (leftPoints[i+2] * tempSin);
		leftPoints[i+2] = (leftPoints[i+1] * -tempSin) + (leftPoints[i+2] * tempCos);
		*/

		//Left control points.
		leftPoints[i] = 0.0f;
		leftPoints[i+1] = 0.0f;
		leftPoints[i+2] = -1.0f;
		i += 3;
	
		leftPoints[i] = 0.0f;
		leftPoints[i+1] = 0.0f;
		leftPoints[i+2] = -1.0f;
		i += 3;
	
		leftPoints[i] = 0.0f;
		leftPoints[i+1] = 0.0f;
		leftPoints[i+2] = -1.0f;
		i += 3;
	
		leftPoints[i] = 0.0f;
		leftPoints[i+1] = 0.0f;
		leftPoints[i+2] = -1.0f;
		i += 3;
	
		/*leftPoints[i] = 0.0f;
		leftPoints[i+1] = 1.0f;
		leftPoints[i+2] = -0.75f;*/
		leftPoints[i] = 0.0f;
		leftPoints[i+1] = (1.0f * tempCos) + (-0.75f * tempSin);
		leftPoints[i+2] = (1.0f * -tempSin) + (-0.75f * tempCos);
		i += 3;
	
		/*leftPoints[i] = -0.5f * stretch;
		leftPoints[i+1] = 1.0f;
		leftPoints[i+2] = -0.75f;*/
		leftPoints[i] = -0.5f * stretch;
		leftPoints[i+1] = (1.0f * tempCos) + (-0.75f * tempSin);
		leftPoints[i+2] = (1.0f * -tempSin) + (-0.75f * tempCos);
		i += 3;
	
		/*leftPoints[i] = -stretch;
		leftPoints[i+1] = 0.5f;
		leftPoints[i+2] = -0.75f;*/
		leftPoints[i] = -stretch;
		leftPoints[i+1] = (0.5f * tempCos) + (-0.75f * tempSin);
		leftPoints[i+2] = (0.5f * -tempSin) + (-0.75f * tempCos);
		i += 3;
	
		/*leftPoints[i] = -stretch;
		leftPoints[i+1] = 0.0f;
		leftPoints[i+2] = -0.75f;*/
		leftPoints[i] = -stretch;
		leftPoints[i+1] = (0.0f * tempCos) + (-0.75f * tempSin);
		leftPoints[i+2] = (0.0f * -tempSin) + (-0.75f * tempCos);
		i += 3;
	
		/*leftPoints[i] = 0.0f;
		leftPoints[i+1] = 1.0f;
		leftPoints[i+2] = 0.75f;*/
		leftPoints[i] = 0.0f;
		leftPoints[i+1] = (1.0f * tempCos) + (0.75f * tempSin);
		leftPoints[i+2] = (1.0f * -tempSin) + (0.75f * tempCos);
		i += 3;
	
		/*leftPoints[i] = -0.5f * stretch;
		leftPoints[i+1] = 1.0f;
		leftPoints[i+2] = 0.75f;*/
		leftPoints[i] = -0.5f * stretch;
		leftPoints[i+1] = (1.0f * tempCos) + (0.75f * tempSin);
		leftPoints[i+2] = (1.0f * -tempSin) + (0.75f * tempCos);
		i += 3;
	
		/*leftPoints[i] = -stretch;
		leftPoints[i+1] = 0.5f;
		leftPoints[i+2] = 0.75f;*/
		leftPoints[i] = -stretch;
		leftPoints[i+1] = (0.5f * tempCos) + (0.75f * tempSin);
		leftPoints[i+2] = (0.5f * -tempSin) + (0.75f * tempCos);
		i += 3;
	
		/*leftPoints[i] = -stretch;
		leftPoints[i+1] = 0.0f;
		leftPoints[i+2] = 0.75f;*/
		leftPoints[i] = -stretch;
		leftPoints[i+1] = (0.0f * tempCos) + (0.75f * tempSin);
		leftPoints[i+2] = (0.0f * -tempSin) + (0.75f * tempCos);
		i += 3;
	
		leftPoints[i] = 0.0f;
		leftPoints[i+1] = 0.0f;
		leftPoints[i+2] = 1.0f;
		i += 3;
	
		leftPoints[i] = 0.0f;
		leftPoints[i+1] = 0.0f;
		leftPoints[i+2] = 1.0f;
		i += 3;
	
		leftPoints[i] = 0.0f;
		leftPoints[i+1] = 0.0f;
		leftPoints[i+2] = 1.0f;
		i += 3;
	
		leftPoints[i] = 0.0f;
		leftPoints[i+1] = 0.0f;
		leftPoints[i+2] = 1.0f;
		i += 3;

		/*
		rightPoints[i] = leftPoints[i];
		rightPoints[i+1] = (leftPoints[i+1] * tempCos) + (leftPoints[i+2] * tempSin);
		rightPoints[i+2] = (leftPoints[i+1] * -tempSin) + (leftPoints[i+2] * tempCos);
		*/
	
		//Right control points.
		i = 0;
		rightPoints[i] = 0.0f;
		rightPoints[i+1] = 0.0f;
		rightPoints[i+2] = 1.0f;
		i += 3;
	
		rightPoints[i] = 0.0f;
		rightPoints[i+1] = 0.0f;
		rightPoints[i+2] = 1.0f;
		i += 3;
	
		rightPoints[i] = 0.0f;
		rightPoints[i+1] = 0.0f;
		rightPoints[i+2] = 1.0f;
		i += 3;
	
		rightPoints[i] = 0.0f;
		rightPoints[i+1] = 0.0f;
		rightPoints[i+2] = 1.0f;
		i += 3;
	
		rightPoints[i] = 0.0f;
		rightPoints[i+1] = 1.0f;
		rightPoints[i+2] = 0.75f;
		i += 3;
	
		/*rightPoints[i] = 0.5f * stretch;
		rightPoints[i+1] = 1.0f;
		rightPoints[i+2] = 0.75f;*/
		rightPoints[i] = 0.5f * stretch;
		rightPoints[i+1] = (1.0f * tempCos) + (0.75f * tempSin);
		rightPoints[i+2] = (1.0f * -tempSin) + (0.75f * tempCos);
		i += 3;
	
		/*rightPoints[i] = stretch;
		rightPoints[i+1] = 0.5f;
		rightPoints[i+2] = 0.75f;*/
		rightPoints[i] = stretch;
		rightPoints[i+1] = (0.5f * tempCos) + (0.75f * tempSin);
		rightPoints[i+2] = (0.5f * -tempSin) + (0.75f * tempCos);
		i += 3;
	
		/*rightPoints[i] = stretch;
		rightPoints[i+1] = 0.0f;
		rightPoints[i+2] = 0.75f;*/
		rightPoints[i] = stretch;
		rightPoints[i+1] = (0.0f * tempCos) + (0.75f * tempSin);
		rightPoints[i+2] = (0.0f * -tempSin) + (0.75f * tempCos);
		i += 3;
	
		/*rightPoints[i] = 0.0f;
		rightPoints[i+1] = 1.0f;
		rightPoints[i+2] = -0.75f;*/
		rightPoints[i] = 0.0f;
		rightPoints[i+1] = (1.0f * tempCos) + (-0.75f * tempSin);
		rightPoints[i+2] = (1.0f * -tempSin) + (-0.75f * tempCos);
		i += 3;
	
		/*rightPoints[i] = 0.5f * stretch;
		rightPoints[i+1] = 1.0f;
		rightPoints[i+2] = -0.75f;*/
		rightPoints[i] = 0.5f * stretch;
		rightPoints[i+1] = (1.0f * tempCos) + (-0.75f * tempSin);
		rightPoints[i+2] = (1.0f * -tempSin) + (-0.75f * tempCos);
		i += 3;
	
		/*rightPoints[i] = stretch;
		rightPoints[i+1] = 0.5f;
		rightPoints[i+2] = -0.75f;*/
		rightPoints[i] = stretch;
		rightPoints[i+1] = (0.5f * tempCos) + (-0.75f * tempSin);
		rightPoints[i+2] = (0.5f * -tempSin) + (-0.75f * tempCos);
		i += 3;
	
		/*rightPoints[i] = stretch;
		rightPoints[i+1] = 0.0f;
		rightPoints[i+2] = -0.75f;*/
		rightPoints[i] = stretch;
		rightPoints[i+1] = (0.0f * tempCos) + (-0.75f * tempSin);
		rightPoints[i+2] = (0.0f * -tempSin) + (-0.75f * tempCos);
		i += 3;
	
		rightPoints[i] = 0.0f;
		rightPoints[i+1] = 0.0f;
		rightPoints[i+2] = -1.0f;
		i += 3;
	
		rightPoints[i] = 0.0f;
		rightPoints[i+1] = 0.0f;
		rightPoints[i+2] = -1.0f;
		i += 3;
	
		rightPoints[i] = 0.0f;
		rightPoints[i+1] = 0.0f;
		rightPoints[i+2] = -1.0f;
		i += 3;
	
		rightPoints[i] = 0.0f;
		rightPoints[i+1] = 0.0f;
		rightPoints[i+2] = -1.0f;
		i += 3;
	};
	///	Helper method to interpolate cosTable.
	inline float getCos(float val)
	{
		int tempint = static_cast<int>(val) ;
		float tempf = val - static_cast<float>(tempint);

		tempint %= 360;

		return (cosTable[tempint] * (1.0f-tempf)) + (cosTable[tempint+1] * tempf);
	}
	///	Helper method to interpolate sinTable.
	inline float getSin(float val)
	{
		int tempint = static_cast<int>(val) ;
		float tempf = val - static_cast<float>(tempint);

		tempint %= 360;

		return (sinTable[tempint] * (1.0f-tempf)) + (sinTable[tempint+1] * tempf);
	}

	///	Used to protect the vertices while they're being drawn.
	CriticalSection critSec;

	///	How much the sphere is being stretched by.
	float stretch;
	///	How much the sphere is being twisted by.
	float twist;
	///	Left sculptor control.
	float leftSculpt;
	///	Right sculptor control.
	float rightSculpt;
	///	Left grabber control.
	float leftGrab;
	///	Right grabber control.
	float rightGrab;
	///	How much the left side is stretched by.
	float leftStretch;
	///	How much the right side is stretched by.
	float rightStretch;

	///	Current left sculpt length position.
	int leftSculptL;
	///	Current left sculpt twist position.
	int leftSculptT;
	///	Current right sculpt length position.
	int rightSculptL;
	///	Current right sculpt twist position.
	int rightSculptT;

	///	The knot values for the patches.
	float knots[7];
	///	The control points for the left patches.
	float leftPoints[16 * 3];
	///	The control points for the right patches.
	float rightPoints[16 * 3];
	///	The GLU nurbs object we use to draw the patches.
	GLUnurbsObj *nurbsObject;

	///	Used to indicate to graphicDoStuff() that the vertices need updated.
	bool verticesNeedUpdating;

	///	Cos table for optimising the twist operation.
	float cosTable[360];
	///	Sin table for optimising the twist operation.
	float sinTable[360];
};

}

#endif
