//	DeformableSphere.cpp - A sphere which can be deformed in various ways.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "DeformableSphere.h"
#include "Camera.h"
#include "StringParser.h"

#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<DeformableSphere> DeformableSphere::constructor;

//------------------------------------------------------------------------------
DeformableSphere::DeformableSphere():
stretch(1.0f),
twist(0.0f),
leftSculpt(0.0f),
rightSculpt(0.0f),
leftGrab(0.0f),
rightGrab(0.0f),
leftStretch(0.5f),
rightStretch(0.5f),
leftSculptL(24),
leftSculptT(8),
rightSculptL(8),
rightSculptT(8),
verticesNeedUpdated(false)
{
	setType("DeformableSphere");
	setVisual(true);
}

//------------------------------------------------------------------------------
DeformableSphere::~DeformableSphere()
{
	
}

//------------------------------------------------------------------------------
void DeformableSphere::initialise()
{
	int x, y;

	//Generate default positions.
	for(y=0;y<NumSections;++y)
	{
		for(x=0;x<NumSections;++x)
		{
			vertices[(y*NumSections)+x].defaultPos.x = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) * 
				   									   cosf(((float)y/(float)NumSections) * 6.283185306f);
			vertices[(y*NumSections)+x].defaultPos.y = sinf((((float)x/(float)NumSections)-0.5f) * 3.141592653f);
			vertices[(y*NumSections)+x].defaultPos.z = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) *
				   									   sinf(((float)y/(float)NumSections) * 6.283185306f);
			vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].defaultPos;
			vertices[(y*NumSections)+x].sculptPos = vertices[(y*NumSections)+x].defaultPos;
		}
	}
}

//------------------------------------------------------------------------------
void DeformableSphere::graphicDoStuff()
{
	int x, y;
	//int i;

	if(verticesNeedUpdated)
	{
		updateVertices();
		verticesNeedUpdated = false;
	}

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glRotatef(90.0f, 0.0f, 0.0f, 1.0f);

	critSec.lock();
	{
		glBegin(GL_QUADS);
		{
			//Main sphere.
			for(y=0;y<NumSections;++y)
			{
				for(x=1;x<(NumSections-2);++x)
				{
					calcVertex(x, y);
					calcVertex(x+1, y);
					calcVertex(x+1, y+1);
					if(x < (NumSections-2))
						calcVertex(x, y+1);
					else
						calcVertex(x+1, y+1);
				}
			}
		}
		glEnd();

		glBegin(GL_TRIANGLES);
		{
			ThreeFloats verts1[3];

			//Top of the sphere.
			for(x=0;x<NumSections;++x)
			{
				verts1[0].x = 0.0f;
				//verts1[0].y = 1.0f;
				verts1[0].y = leftStretch;
				verts1[0].z = 0.0f;
				verts1[1] = getVertex((NumSections-1), x);
				if(x < (NumSections-1))
					verts1[2] = getVertex((NumSections-1), x+1);
				else
					verts1[2] = getVertex((NumSections-1), 0);

				glNormal3f(0.0f, 1.0f, 0.0f);
				glTexCoord2f((((float)x+0.5f)/(float)NumSections),
							 1.0f);
				glVertex3f(verts1[0].x, verts1[0].y, verts1[0].z);

				glTexCoord2f(((float)x/(float)NumSections),
							 ((float)(NumSections-1)/(float)NumSections));
				glVertex3f(verts1[1].x, verts1[1].y, verts1[1].z);

				glTexCoord2f(((float)(x+1)/(float)NumSections),
							 ((float)(NumSections-1)/(float)NumSections));
				glVertex3f(verts1[2].x, verts1[2].y, verts1[2].z);
			}

			//Bottom of the sphere.
			for(x=0;x<NumSections;++x)
			{
				verts1[0].x = 0.0f;
				//verts1[0].y = -1.0f;
				verts1[0].y = -rightStretch;
				verts1[0].z = 0.0f;
				verts1[1] = getVertex(2, x);
				if(x < (NumSections-1))
					verts1[2] = getVertex(2, x+1);
				else
					verts1[2] = getVertex(2, 0);

				glNormal3f(0.0f, -1.0f, 0.0f);
				glTexCoord2f((((float)x+0.5f)/(float)NumSections),
							 1.0f);
				glVertex3f(verts1[0].x, verts1[0].y, verts1[0].z);

				glTexCoord2f(((float)x/(float)NumSections),
							 ((float)(NumSections-1)/(float)NumSections));
				glVertex3f(verts1[1].x, verts1[1].y, verts1[1].z);

				glTexCoord2f(((float)(x+1)/(float)NumSections),
							 ((float)(NumSections-1)/(float)NumSections));
				glVertex3f(verts1[2].x, verts1[2].y, verts1[2].z);
			}
		}
		glEnd();
	}
	critSec.unlock();

	glPopMatrix();
}

//------------------------------------------------------------------------------
void DeformableSphere::setAttribute(const string& attribute, EventValue& val)
{
	//int i;
	StringParser parser(val);

	if(attribute == "stretch")
	{
		stretch = val;
		verticesNeedUpdated = true;
	}
	else if(attribute == "twist")
	{
		twist = val;
		verticesNeedUpdated = true;
	}
	else if(attribute == "leftSculpt")
	{
		leftSculpt = val;
		verticesNeedUpdated = true;
	}
	else if(attribute == "rightSculpt")
	{
		rightSculpt = val;
		verticesNeedUpdated = true;
	}
	else if(attribute == "leftGrab")
		leftGrab = val;
	else if(attribute == "rightGrab")
		rightGrab = val;
}

//------------------------------------------------------------------------------
void DeformableSphere::calcVertex(int x, int y)
{
	ThreeFloats verts1[4];
	ThreeFloats verts2[4];
	ThreeFloats verts3[4];
	ThreeFloats verts4[4];
	ThreeFloats n1;
	ThreeFloats n2;
	ThreeFloats n3;
	ThreeFloats n4;
	ThreeFloats n5;

	//Current quad.
	verts1[0] = getVertex(x+1, y+1);
	verts1[1] = getVertex(x, y+1);
	verts1[2] = getVertex(x, y);
	verts1[3] = getVertex(x+1, y);

	//Next quad.
	verts2[0] = getVertex(x+2, y+1);
	verts2[1] = getVertex(x+1, y+1);
	verts2[2] = getVertex(x+1, y);
	verts2[3] = getVertex(x+2, y);

	//Quad above.
	verts3[0] = getVertex(x+1, y+2);
	verts3[1] = getVertex(x, y+2);
	verts3[2] = getVertex(x, y+1);
	verts3[3] = getVertex(x+1, y+1);

	//Quad above and along.
	verts3[0] = getVertex(x+2, y+2);
	verts3[1] = getVertex(x+1, y+2);
	verts3[2] = getVertex(x+1, y+1);
	verts3[3] = getVertex(x+2, y+1);

	n1 = calcNormal2(verts1[0], verts1[1], verts1[2]);
	n2 = calcNormal2(verts2[0], verts2[1], verts2[2]);
	n3 = calcNormal2(verts3[0], verts3[1], verts3[2]);
	n4 = calcNormal2(verts4[0], verts4[1], verts4[2]);
	n5 = calcSmooth(n1, n2, n3, n4);

	glNormal3f(n5.x, n5.y, n5.z);
	glTexCoord2f(((float)(y+1)/(float)NumSections),
				 ((float)x/(float)NumSections));
	glVertex3f(verts1[0].x, verts1[0].y, verts1[0].z);
}

//------------------------------------------------------------------------------
void DeformableSphere::updateVertices()
{
	int i, x, y;
	ThreeFloats rotMatrix[3];
	ThreeFloats tempThree;

	//Update sculpting positions.
	if(leftGrab < 0.5f)
	{
		leftSculptL = static_cast<int>((stretch * (1.0f-leftGrab)) * 13) + 17;
		leftSculptL %= 29;
		leftSculptL += 1;
		leftSculptT = static_cast<int>((twist * (1.0f-leftGrab)) * 30)+1;
		leftSculptT %= 29;
		leftSculptT += 1;
	}
	else
		leftStretch = stretch;
	if(rightGrab < 0.5f)
	{
		rightSculptL = static_cast<int>((stretch * (1.0f-rightGrab)) * 13) + 1;
		rightSculptL %= 30;
		rightSculptL += 1;
		rightSculptT = static_cast<int>((twist * (1.0f-rightGrab)) * 30)+1;
		rightSculptT %= 30;
		rightSculptT += 1;
	}
	else
		rightStretch = stretch;

	//Calculate leftSculpt.
	//i = (8*NumSections)+24;
	i = (leftSculptT * NumSections) + leftSculptL;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-leftSculpt);
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-leftSculpt);
	++i;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(leftSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(leftSculpt*0.5f));
	i -= 2;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(leftSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(leftSculpt*0.5f));
	++i;
	i -= NumSections;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(leftSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(leftSculpt*0.5f));
	//++i;
	i += NumSections * 2;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(leftSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(leftSculpt*0.5f));

	//Calculate rightSculpt.
	//i = (8*NumSections)+8;
	i = (rightSculptT * NumSections) + rightSculptL;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-rightSculpt);
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-rightSculpt);
	++i;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(rightSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(rightSculpt*0.5f));
	i -= 2;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(rightSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(rightSculpt*0.5f));
	++i;
	i -= NumSections;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(rightSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(rightSculpt*0.5f));
	//++i;
	i += NumSections * 2;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(rightSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(rightSculpt*0.5f));

	//Calculate stretch.
	/*for(i=0;i<(NumSections*NumSections);++i)
	{
		vertices[i].pos = vertices[i].defaultPos;
		vertices[i].pos.y *= stretch;
	}*/
	/*for(y=0;y<(NumSections/2);++y)
	{
		for(x=0;x<NumSections;++x)
		{
			vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].defaultPos;
			vertices[(y*NumSections)+x].pos.y *= leftStretch;
		}
	}
	for(y=(NumSections/2);y<NumSections;++y)*/
	for(y=0;y<NumSections;++y)
	{
		for(x=0;x<(NumSections/2);++x)
		{
			//vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].defaultPos;
			vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].sculptPos;
			vertices[(y*NumSections)+x].pos.y *= rightStretch;
		}
		for(x=(NumSections/2);x<NumSections;++x)
		{
			//vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].defaultPos;
			vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].sculptPos;
			vertices[(y*NumSections)+x].pos.y *= leftStretch;
		}
	}

	//Calculate rotation.
	for(y=0;y<NumSections;++y)
	{
		for(x=0;x<(NumSections/2);++x)
		{
			float tempf = (1.0f-((float)x/((float)NumSections/2.0f)));

			tempf *= (leftGrab + rightGrab) * 0.5f;
			rotMatrix[0].x = cosf(twist * tempf * 2.0f * 3.141592653f);
			rotMatrix[0].z = sinf(twist * tempf * 2.0f * 3.141592653f);
			/*rotMatrix[2].x = -sinf(twist * tempf * 2.0f * 3.141592653f);
			rotMatrix[2].z = cosf(twist * tempf * 2.0f * 3.141592653f);*/

			tempThree.x = (vertices[(y*NumSections)+x].pos.x * rotMatrix[0].x) +
						  (vertices[(y*NumSections)+x].pos.z * rotMatrix[0].z);
			tempThree.y = vertices[(y*NumSections)+x].pos.y;
			tempThree.z = (vertices[(y*NumSections)+x].pos.x * -rotMatrix[0].z) +
						  (vertices[(y*NumSections)+x].pos.z * rotMatrix[0].x);

			vertices[(y*NumSections)+x].pos = tempThree;
		}
		for(x=(NumSections/2)+1;x<NumSections;++x)
		{
			float tempf = (((float)x-(NumSections/2))/((float)NumSections/2.0f));

			tempf *= (leftGrab + rightGrab) * 0.5f;
			rotMatrix[0].x = cosf(-twist * tempf * 2.0f * 3.141592653f);
			rotMatrix[0].z = sinf(-twist * tempf * 2.0f * 3.141592653f);
			//rotMatrix[2].x = -sinf(-twist * tempf * 2.0f * 3.141592653f);
			//rotMatrix[2].z = cosf(-twist * tempf * 2.0f * 3.141592653f);

			tempThree.x = (vertices[(y*NumSections)+x].pos.x * rotMatrix[0].x) +
						  (vertices[(y*NumSections)+x].pos.z * rotMatrix[0].z);
			tempThree.y = vertices[(y*NumSections)+x].pos.y;
			tempThree.z = (vertices[(y*NumSections)+x].pos.x * -rotMatrix[0].z) +
						  (vertices[(y*NumSections)+x].pos.z * rotMatrix[0].x);

			vertices[(y*NumSections)+x].pos = tempThree;
		}
	}
}

}
