//	DeformableSphere.h - A sphere which can be deformed in various ways.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef DEFORMABLESPHERE_H_
#define DEFORMABLESPHERE_H_

#include "AbstractNode.h"
#include "CriticalSection.h"

#include "GLHeaders.h"
#include <cmath>

/*+
	Name: DeformableSphere
	Type: ChildNode.
	Description: A sphere which may be deformed by manipulating it's attributes.
			 	 This is a test node for my final audiovisual instrument, hence
				 the slightly counterintuitive attributes (they make more sense
				 when considering the physical interface I'm designing for it).
	Attribute:	leftGrab
	AttType:	SFFloat (input)
	AttDesc:	'Grabs' the left side of the sphere, an operation which
				modulates the other attributes.

	Attribute:	rightGrab
	AttType:	SFFloat (input)
	AttDesc:	'Grabs' the right side of the sphere, an operation which
				modulates the other attributes.

	Attribute:	stretch
	AttType:	SFFloat (input)
	AttDesc:	When the grabs are 0, stretches the sphere along the x
				dimension.  Otherwise changes the sculpting position (x-axis) of
				the left and/or right sides.

	Attribute:	twist
	AttType:	SFFloat (input)
	AttDesc:	When the grabs are 0, twists the sphere around the x axis.
				Otherwise changes the sculpting position (rotating around the
				object) of the left and/or right sides.

	Attribute:	leftSculpt
	AttType:	SFFloat (input)
	AttDesc:	When leftGrab is non-zero, controls the radius (from the
				sphere's centre) of the vertex at the current left sculpting
				position.

	Attribute:	rightSculpt
	AttType:	SFFloat (input)
	AttDesc:	When rightGrab is non-zero, controls the radius (from the
				sphere's centre) of the vertex at the current right sculpting
				position.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY" headlight="FALSE"/&gt;

		&lt;Transform DEF="temp" translation="0.0 0.0 -8.0"&gt;
			&lt;DeformableSphere DEF="bob"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	A sphere which can be deformed in various ways.
/*!
	\ingroup libheilanextras
 */
class DeformableSphere : public ChildNode
{
  public:
	///	Constructor.
	DeformableSphere();
	///	Destructor.
	~DeformableSphere();

	///	Sets up the vertices.
	void initialise();

	///	Draws the sphere.
	void graphicDoStuff();
	///	Not used, overridden to do nothing.
	/*AmbisonicData *audioDoStuff(int numSamples,
								const ThreeFloats& pos) {return 0;};*/

	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set the sphere's parameters.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<DeformableSphere> constructor;
  private:
	///	Helper method to calculate the normal of a polygon.
	inline void calcNormal(const ThreeFloats& v1,
						   const ThreeFloats& v2,
						   const ThreeFloats& v3)
	{
		ThreeFloats a;
		ThreeFloats b;
		ThreeFloats normal;
		float length;

		a.x = v1.x - v2.x;
		a.y = v1.y - v2.y;
		a.z = v1.z - v2.z;

		b.x = v2.x - v3.x;
		b.y = v2.y - v3.y;
		b.z = v2.z - v3.z;

		normal.x = (a.y * b.z) - (a.z * b.y);
		normal.y = (a.z * b.x) - (a.x * b.z);
		normal.z = (a.x * b.y) - (a.y * b.x);

		length = sqrtf((normal.x*normal.x)+(normal.y*normal.y)+(normal.z*normal.z));
		if(length == 0.0f)
			length = 1.0f;
		else
			length = 1.0f/length;

		glNormal3f((normal.x*length), (normal.y*length), (normal.z*length));
	};
	///	Second version, to return the normal instead of drawing it.
	inline ThreeFloats calcNormal2(const ThreeFloats& v1,
								   const ThreeFloats& v2,
								   const ThreeFloats& v3)
	{
		ThreeFloats a;
		ThreeFloats b;
		ThreeFloats normal;
		float length;

		a.x = v1.x - v2.x;
		a.y = v1.y - v2.y;
		a.z = v1.z - v2.z;

		b.x = v2.x - v3.x;
		b.y = v2.y - v3.y;
		b.z = v2.z - v3.z;

		normal.x = (a.y * b.z) - (a.z * b.y);
		normal.y = (a.z * b.x) - (a.x * b.z);
		normal.z = (a.x * b.y) - (a.y * b.x);

		length = sqrtf((normal.x*normal.x)+(normal.y*normal.y)+(normal.z*normal.z));
		if(length == 0.0f)
			length = 1.0f;
		else
			length = 1.0f/length;

		normal.x *= length;
		normal.y *= length;
		normal.z *= length;

		return normal;
	};
	///	To calculate smooth normals.
	inline ThreeFloats calcSmooth(ThreeFloats& n1,
								  ThreeFloats& n2,
								  ThreeFloats& n3,
								  ThreeFloats& n4)
	{
		ThreeFloats retval;

		retval.x = (n1.x + n2.x + n3.x + n4.x)*0.25f;
		retval.y = (n1.y + n2.y + n3.y + n4.y)*0.25f;
		retval.z = (n1.z + n2.z + n3.z + n4.z)*0.25f;

		return retval;
	};

	///	Helper method to calculate a vertex on the sphere.
	void calcVertex(int x, int y);
	///	Helper method for calcVertex().
	inline ThreeFloats& getVertex(int x, int y)
	{
		/*ThreeFloats retval;

		retval.x = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) * 
				   cosf(((float)y/(float)NumSections) * 6.283185306f);
		retval.y = sinf((((float)x/(float)NumSections)-0.5f) * 3.141592653f);
		retval.z = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) *
				   sinf(((float)y/(float)NumSections) * 6.283185306f);

		return retval;*/
		if(x >= NumSections)
			x -= NumSections;
		if(y >= NumSections)
			y -= NumSections;

		return vertices[(y*NumSections)+x].pos;
	}

	///	Helper method to update vertices after the performer has modified them.
	/*!
		MUST BE CALLED WITHIN A critSec.lock()/critSec.unlock() PAIR!!!!
	 */
	void updateVertices();

	///	Struct representing a vertex in the sphere.
	struct Vertex
	{
		///	The Vertex' current position.
		ThreeFloats pos;
		///	The Vertex' default position.
		ThreeFloats defaultPos;
		///	The Vertex' sculpted position.
		ThreeFloats sculptPos;
	};

	///	Enum containing the number of sides we use to draw the sphere.
	enum
	{
		NumSections = 32
	};

	///	Used to protect the vertices while they're being drawn.
	CriticalSection critSec;

	///	How much the sphere is being stretched by.
	float stretch;
	///	How much the sphere is being twisted by.
	float twist;
	///	Left sculptor control.
	float leftSculpt;
	///	Right sculptor control.
	float rightSculpt;
	///	Left grabber control.
	float leftGrab;
	///	Right grabber control.
	float rightGrab;
	///	How much the left side is stretched by.
	float leftStretch;
	///	How much the right side is stretched by.
	float rightStretch;

	///	All the Vertices in the sphere.
	Vertex vertices[NumSections*NumSections];
	///	Current left sculpt length position.
	int leftSculptL;
	///	Current left sculpt twist position.
	int leftSculptT;
	///	Current right sculpt length position.
	int rightSculptL;
	///	Current right sculpt twist position.
	int rightSculptT;

	///	Used to update the vertices in graphicDoStuff() when necessary.
	bool verticesNeedUpdated;
};

}

#endif
