//	DeformableSphere2.cpp - A sphere which can be deformed in various ways.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "DeformableSphere2.h"
#include "Camera.h"
#include "StringParser.h"
//#include "AvailableExtensions.h"
#include "GLHeaders.h"

#include <iostream>

using namespace std;

//using namespace HeilanX3D::AvailableExtensions;

namespace HeilanX3D
{

NodeConstructor<DeformableSphere2> DeformableSphere2::constructor;

//------------------------------------------------------------------------------
DeformableSphere2::DeformableSphere2():
stretch(1.0f),
twist(0.0f),
leftSculpt(0.0f),
rightSculpt(0.0f),
leftGrab(0.0f),
rightGrab(0.0f),
leftStretch(0.5f),
rightStretch(0.5f),
xAccel(0.0f),
yAccel(0.0f),
zAccel(0.0f),
spikyMode(false),
tex1X(0.0f),
tex1Y(0.0f),
tex1SX(1.0f),
tex1SY(1.0f),
tex1Rot(0.0f),
tex2X(0.0f),
tex2Y(0.0f),
tex2SX(1.0f),
tex2SY(1.0f),
tex2Rot(0.0f),
texBlend(0.8f),
vertices(0),
leftSculptL(24),
leftSculptT(8),
rightSculptL(8),
rightSculptT(8),
verticesNeedUpdated(false),
vertexPositions(0),
multiTextureSupported(false),
noiseTex(0),
perlinTex(0),
sineTex(0),
thresholdPerlinTex(0)
{
	setType("DeformableSphere2");
	setVisual(true);
}

//------------------------------------------------------------------------------
DeformableSphere2::~DeformableSphere2()
{
	if(vertices)
		delete [] vertices;
	if(vertexPositions)
		delete [] vertexPositions;
	if(noiseTex)
		delete noiseTex;
	if(perlinTex)
		delete perlinTex;
	if(sineTex)
		delete sineTex;
	if(thresholdPerlinTex)
		delete thresholdPerlinTex;
}

//------------------------------------------------------------------------------
void DeformableSphere2::initialise()
{
	int i;
	int x, y;

	if(vertices)
		delete [] vertices;
	vertices = new Vertex[NumSections*NumSections];
	if(vertexPositions)
		delete [] vertexPositions;
	vertexPositions = new float[SurfaceWidth*NumSubdivisions*SurfaceHeight*NumSubdivisions];
	if(noiseTex)
		delete noiseTex;
	noiseTex = new RubbishNoiseTexture();
	if(perlinTex)
		delete perlinTex;
	perlinTex = new PerlinTexture();
	if(sineTex)
		delete sineTex;
	sineTex = new StraightSineTexture();
	if(thresholdPerlinTex)
		delete thresholdPerlinTex;
	thresholdPerlinTex = new ThresholdPerlinTexture();

	//Generate default positions.
	for(y=0;y<NumSections;++y)
	{
		for(x=0;x<NumSections;++x)
		{
			vertices[(y*NumSections)+x].defaultPos.x = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) * 
				   									   cosf(((float)y/(float)NumSections) * 6.283185306f);
			vertices[(y*NumSections)+x].defaultPos.y = sinf((((float)x/(float)NumSections)-0.5f) * 3.141592653f);
			vertices[(y*NumSections)+x].defaultPos.z = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) *
				   									   sinf(((float)y/(float)NumSections) * 6.283185306f);
			vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].defaultPos;
			vertices[(y*NumSections)+x].sculptPos = vertices[(y*NumSections)+x].defaultPos;
		}
	}

	//Calculate Tao surface stuff.
	for(i=0;i<(SurfaceHeight*SurfaceWidth);++i)
	{
		positions[i] = 0.0f;
		forces[i] = 0.0f;
		velocities[i] = 0.0f;
	}

	forces[SurfaceWidth+1] = 1.0f;

	//Calculate cells' mass.
	float decay = (1.0f/16.0f);
	float intendedFreq = 440.0f;
	float actualFreq;
	float compensationFactor;
	defaultVelocityMultiplier = 1.0-(0.00036f/decay);
	actualFreq = (24000.0f*30.0f/30.0f)/(SurfaceWidth+1);
	compensationFactor=powf(4.0f, log10f(actualFreq/intendedFreq)/log10f(2.0f));
	mass = 3.5f * compensationFactor;
	invMass = 1.0f/mass;

	velocityMultiplier = defaultVelocityMultiplier;

	for(i=0;i<(SurfaceWidth*NumSubdivisions*SurfaceHeight*NumSubdivisions);++i)
		vertexPositions[i] = 0.0f;


	//Check if multi-texturing's supported.
#ifndef GL_ARB_multitexture
	if(glActiveTextureARB && glMultiTexCoord2fARB)
		multiTextureSupported = true;
#endif

	//Create textures.
	noiseTex->generateOpenGLTexture(noiseId);
	sineTex->generateOpenGLTexture(sineId);
	perlinTex->generateOpenGLTexture(perlinId);
	thresholdPerlinTex->generateOpenGLTexture(thresholdPerlinId);
}

//------------------------------------------------------------------------------
void DeformableSphere2::graphicDoStuff()
{
	int x, y;
	//int i, j;
	GLfloat constColor[4] = {1.0f, 1.0f, 1.0f, texBlend};

	updateVertices2();

#ifdef TWO_PASS
	glDisable(GL_LIGHTING);
#endif

	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);

	glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_REPLACE);

	glBindTexture(GL_TEXTURE_2D, perlinId);

	//Translate, scale and rotate perlin texture.
	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(tex1X, tex1Y, 0.0f);
	glScalef(tex1SX, tex1SY, 1.0f);
	glRotatef(tex1Rot, 0.0f, 0.0f, 1.0f);


	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);

	//Set constant colour (alpha channel of which is used to interpolate between
	//the two textures).
	glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, constColor);

	//Interpolate between the two textures.
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
	glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_INTERPOLATE_ARB);
	glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_PREVIOUS_ARB);
	glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
	glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_TEXTURE);
	glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
	glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE2_RGB_ARB, GL_CONSTANT_ARB);
	glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND2_RGB_ARB, GL_SRC_ALPHA);

	glBindTexture(GL_TEXTURE_2D, thresholdPerlinId);

	//Translate, scale and rotate sine texture.
	//glMatrixMode(GL_TEXTURE);
	//glPushMatrix();
	glLoadIdentity();
	glTranslatef(tex2X, tex2Y, 0.0f);
	glScalef(tex2SX, tex2SY, 1.0f);
	glRotatef(tex2Rot, 0.0f, 0.0f, 1.0f);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glScalef((1.0f+xAccel), (1.0f+yAccel), (1.0f+zAccel));

	glRotatef(90.0f, 0.0f, 0.0f, 1.0f);

	critSec.lock();
	{
		glBegin(GL_QUADS);
		{
			//Main sphere.
			for(y=0;y<NumSections;++y)
			{
				for(x=1;x<(NumSections-2);++x)
				{
					calcVertex(x, y);
					calcVertex(x+1, y);
					calcVertex(x+1, y+1);
					if(x < (NumSections-2))
						calcVertex(x, y+1);
					else
						calcVertex(x+1, y+1);
				}
			}
		}
		glEnd();

#ifdef TWO_PASS
		//2nd-pass.
		glEnable(GL_LIGHTING);
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glDisable(GL_TEXTURE_2D);
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glDisable(GL_TEXTURE_2D);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		glBlendFunc(GL_DST_COLOR,GL_SRC_COLOR);
		glEnable(GL_BLEND);
		glBegin(GL_QUADS);
		{
			//Main sphere.
			for(y=0;y<NumSections;++y)
			{
				for(x=1;x<(NumSections-2);++x)
				{
					calcVertex(x, y);
					calcVertex(x+1, y);
					calcVertex(x+1, y+1);
					if(x < (NumSections-2))
						calcVertex(x, y+1);
					else
						calcVertex(x+1, y+1);
				}
			}
		}
		glEnd();
		glDisable(GL_BLEND);
		glEnable(GL_TEXTURE_2D);
#endif

#ifndef TWO_PASS //'cos I'm lazy...
		glBegin(GL_TRIANGLES);
		{
			ThreeFloats verts1[3];

			//Top of the sphere.
			for(x=0;x<NumSections;++x)
			{
				verts1[0].x = 0.0f;
				//verts1[0].y = 1.0f;
				verts1[0].y = leftStretch;
				verts1[0].z = 0.0f;
				verts1[1] = getVertex((NumSections-1), x);
				if(x < (NumSections-1))
					verts1[2] = getVertex((NumSections-1), x+1);
				else
					verts1[2] = getVertex((NumSections-1), 0);

				glNormal3f(0.0f, 1.0f, 0.0f);
				if(multiTextureSupported)
				{
					glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
											 (((float)x+0.5f)/(float)NumSections),
											 1.0f);
					glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
											 (((float)x+0.5f)/(float)NumSections),
											 1.0f);
				}
				else
				{
					glTexCoord2f((((float)x+0.5f)/(float)NumSections),
								 1.0f);
				}
				glVertex3f(verts1[0].x, verts1[0].y, verts1[0].z);

				if(multiTextureSupported)
				{
					glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
											 ((float)x/(float)NumSections),
											 ((float)(NumSections-1)/(float)NumSections));
					glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
											 ((float)x/(float)NumSections),
											 ((float)(NumSections-1)/(float)NumSections));
				}
				else
				{
					glTexCoord2f(((float)x/(float)NumSections),
								 ((float)(NumSections-1)/(float)NumSections));
				}
				glVertex3f(verts1[1].x, verts1[1].y, verts1[1].z);

				if(multiTextureSupported)
				{
					glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
											 ((float)(x+1)/(float)NumSections),
											 ((float)(NumSections-1)/(float)NumSections));
					glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
											 ((float)(x+1)/(float)NumSections),
											 ((float)(NumSections-1)/(float)NumSections));
				}
				else
				{
					glTexCoord2f(((float)(x+1)/(float)NumSections),
								 ((float)(NumSections-1)/(float)NumSections));
				}
				glVertex3f(verts1[2].x, verts1[2].y, verts1[2].z);
			}

			//Bottom of the sphere.
			for(x=0;x<NumSections;++x)
			{
				verts1[0].x = 0.0f;
				//verts1[0].y = -1.0f;
				verts1[0].y = -rightStretch;
				verts1[0].z = 0.0f;
				verts1[1] = getVertex(2, x);
				if(x < (NumSections-1))
					verts1[2] = getVertex(2, x+1);
				else
					verts1[2] = getVertex(2, 0);

				glNormal3f(0.0f, -1.0f, 0.0f);
				if(multiTextureSupported)
				{
					glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
											 (((float)x+0.5f)/(float)NumSections),
											 1.0f);
					glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
											 (((float)x+0.5f)/(float)NumSections),
											 1.0f);
				}
				else
				{
					glTexCoord2f((((float)x+0.5f)/(float)NumSections),
								 1.0f);
				}
				glVertex3f(verts1[0].x, verts1[0].y, verts1[0].z);

				if(multiTextureSupported)
				{
					glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
											 ((float)x/(float)NumSections),
											 ((float)(NumSections-1)/(float)NumSections));
					glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
											 ((float)x/(float)NumSections),
											 ((float)(NumSections-1)/(float)NumSections));
				}
				else
				{
					glTexCoord2f(((float)x/(float)NumSections),
								 ((float)(NumSections-1)/(float)NumSections));
				}
				glVertex3f(verts1[1].x, verts1[1].y, verts1[1].z);

				if(multiTextureSupported)
				{
					glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
											 ((float)(x+1)/(float)NumSections),
											 ((float)(NumSections-1)/(float)NumSections));
					glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
											 ((float)(x+1)/(float)NumSections),
											 ((float)(NumSections-1)/(float)NumSections));
				}
				else
				{
					glTexCoord2f(((float)(x+1)/(float)NumSections),
								 ((float)(NumSections-1)/(float)NumSections));
				}
				glVertex3f(verts1[2].x, verts1[2].y, verts1[2].z);
			}
		}
		glEnd();
#endif
	}
	critSec.unlock();

	//'Disable' multi-texturing.
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glDisable(GL_TEXTURE_2D);

	glActiveTextureARB(GL_TEXTURE0_ARB);

	glMatrixMode(GL_TEXTURE);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

//------------------------------------------------------------------------------
void DeformableSphere2::setAttribute(const string& attribute, EventValue& val)
{
	//int i;
	StringParser parser(val);

	if(attribute == "stretch")
	{
		stretch = val;
		verticesNeedUpdated = true;
	}
	else if(attribute == "twist")
	{
		twist = val;
		verticesNeedUpdated = true;
	}
	else if(attribute == "leftSculpt")
	{
		leftSculpt = val;
		verticesNeedUpdated = true;
	}
	else if(attribute == "rightSculpt")
	{
		rightSculpt = val;
		verticesNeedUpdated = true;
	}
	else if(attribute == "leftGrab")
		leftGrab = val;
	else if(attribute == "rightGrab")
		rightGrab = val;
	else if(attribute == "xAccel")
		xAccel = val;
	else if(attribute == "yAccel")
		yAccel = val;
	else if(attribute == "zAccel")
		zAccel = val;
	else if(attribute == "spikyMode")
		spikyMode = (float)val > 0.5f;
	else if(attribute == "tex1X")
		tex1X = val;
	else if(attribute == "tex1Y")
		tex1Y = val;
	else if(attribute == "tex1SX")
		tex1SX = val;
	else if(attribute == "tex1SY")
		tex1SY = val;
	else if(attribute == "tex1Rot")
		tex1Rot = val;
	else if(attribute == "tex2X")
		tex2X = val;
	else if(attribute == "tex2Y")
		tex2Y = val;
	else if(attribute == "tex2SX")
		tex2SX = val;
	else if(attribute == "tex2SY")
		tex2SY = val;
	else if(attribute == "tex2Rot")
		tex2Rot = val;
	else if(attribute == "texBlend")
		texBlend = val;
}

//------------------------------------------------------------------------------
void DeformableSphere2::calcVertex(int x, int y)
{
	ThreeFloats verts1[4];
	ThreeFloats verts2[4];
	ThreeFloats verts3[4];
	ThreeFloats verts4[4];
	ThreeFloats n1;
	ThreeFloats n2;
	ThreeFloats n3;
	ThreeFloats n4;
	ThreeFloats n5;

	//Current quad.
	verts1[0] = getVertex(x+1, y+1);
	verts1[1] = getVertex(x, y+1);
	verts1[2] = getVertex(x, y);
	verts1[3] = getVertex(x+1, y);

	//Next quad.
	verts2[0] = getVertex(x+2, y+1);
	verts2[1] = getVertex(x+1, y+1);
	verts2[2] = getVertex(x+1, y);
	verts2[3] = getVertex(x+2, y);

	//Quad above.
	verts3[0] = getVertex(x+1, y+2);
	verts3[1] = getVertex(x, y+2);
	verts3[2] = getVertex(x, y+1);
	verts3[3] = getVertex(x+1, y+1);

	//Quad above and along.
	verts3[0] = getVertex(x+2, y+2);
	verts3[1] = getVertex(x+1, y+2);
	verts3[2] = getVertex(x+1, y+1);
	verts3[3] = getVertex(x+2, y+1);

	n1 = calcNormal2(verts1[0], verts1[1], verts1[2]);
	n2 = calcNormal2(verts2[0], verts2[1], verts2[2]);
	n3 = calcNormal2(verts3[0], verts3[1], verts3[2]);
	n4 = calcNormal2(verts4[0], verts4[1], verts4[2]);
	n5 = calcSmooth(n1, n2, n3, n4);

	glNormal3f(n5.x, n5.y, n5.z);
	if(multiTextureSupported)
	{
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
							 ((float)(y+1)/(float)NumSections),
							 ((float)x/(float)NumSections));
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
							 ((float)(y+1)/(float)NumSections),
							 ((float)x/(float)NumSections));
	}
	else
	{
		glTexCoord2f(((float)(y+1)/(float)NumSections),
					 ((float)x/(float)NumSections));
	}
	glVertex3f(verts1[0].x, verts1[0].y, verts1[0].z);
}

//------------------------------------------------------------------------------
void DeformableSphere2::updateVertices()
{
	int i, x, y;
	ThreeFloats rotMatrix[3];
	ThreeFloats tempThree;

	//Update sculpting positions.
	if(leftGrab < 0.5f)
	{
		leftSculptL = static_cast<int>((stretch * (1.0f-leftGrab)) * 13) + 17;
		leftSculptL %= 29;
		leftSculptL += 1;
		leftSculptT = static_cast<int>((twist * (1.0f-leftGrab)) * 30)+1;
		leftSculptT %= 29;
		leftSculptT += 1;
	}
	else
		leftStretch = stretch;
	if(rightGrab < 0.5f)
	{
		rightSculptL = static_cast<int>((stretch * (1.0f-rightGrab)) * 13) + 1;
		rightSculptL %= 30;
		rightSculptL += 1;
		rightSculptT = static_cast<int>((twist * (1.0f-rightGrab)) * 30)+1;
		rightSculptT %= 30;
		rightSculptT += 1;
	}
	else
		rightStretch = stretch;

	//Calculate leftSculpt.
	//i = (8*NumSections)+24;
	i = (leftSculptT * NumSections) + leftSculptL;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-leftSculpt);
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-leftSculpt);
	++i;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(leftSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(leftSculpt*0.5f));
	i -= 2;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(leftSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(leftSculpt*0.5f));
	++i;
	i -= NumSections;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(leftSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(leftSculpt*0.5f));
	//++i;
	i += NumSections * 2;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(leftSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(leftSculpt*0.5f));

	//Calculate rightSculpt.
	//i = (8*NumSections)+8;
	i = (rightSculptT * NumSections) + rightSculptL;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-rightSculpt);
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-rightSculpt);
	++i;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(rightSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(rightSculpt*0.5f));
	i -= 2;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(rightSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(rightSculpt*0.5f));
	++i;
	i -= NumSections;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(rightSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(rightSculpt*0.5f));
	//++i;
	i += NumSections * 2;
	vertices[i].sculptPos.x = vertices[i].defaultPos.x *
						(1.0f-(rightSculpt*0.5f));
	vertices[i].sculptPos.z = vertices[i].defaultPos.z *
						(1.0f-(rightSculpt*0.5f));

	//Calculate stretch.
	/*for(i=0;i<(NumSections*NumSections);++i)
	{
		vertices[i].pos = vertices[i].defaultPos;
		vertices[i].pos.y *= stretch;
	}*/
	/*for(y=0;y<(NumSections/2);++y)
	{
		for(x=0;x<NumSections;++x)
		{
			vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].defaultPos;
			vertices[(y*NumSections)+x].pos.y *= leftStretch;
		}
	}
	for(y=(NumSections/2);y<NumSections;++y)*/
	for(y=0;y<NumSections;++y)
	{
		for(x=0;x<(NumSections/2);++x)
		{
			//vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].defaultPos;
			vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].sculptPos;
			vertices[(y*NumSections)+x].pos.y *= rightStretch;
		}
		for(x=(NumSections/2);x<NumSections;++x)
		{
			//vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].defaultPos;
			vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].sculptPos;
			vertices[(y*NumSections)+x].pos.y *= leftStretch;
		}
	}

	//Calculate rotation.
	for(y=0;y<NumSections;++y)
	{
		for(x=0;x<(NumSections/2);++x)
		{
			float tempf = (1.0f-((float)x/((float)NumSections/2.0f)));

			tempf *= (leftGrab + rightGrab) * 0.5f;
			rotMatrix[0].x = cosf(twist * tempf * 2.0f * 3.141592653f);
			rotMatrix[0].z = sinf(twist * tempf * 2.0f * 3.141592653f);
			/*rotMatrix[2].x = -sinf(twist * tempf * 2.0f * 3.141592653f);
			rotMatrix[2].z = cosf(twist * tempf * 2.0f * 3.141592653f);*/

			tempThree.x = (vertices[(y*NumSections)+x].pos.x * rotMatrix[0].x) +
						  (vertices[(y*NumSections)+x].pos.z * rotMatrix[0].z);
			tempThree.y = vertices[(y*NumSections)+x].pos.y;
			tempThree.z = (vertices[(y*NumSections)+x].pos.x * -rotMatrix[0].z) +
						  (vertices[(y*NumSections)+x].pos.z * rotMatrix[0].x);

			vertices[(y*NumSections)+x].pos = tempThree;
		}
		for(x=(NumSections/2)+1;x<NumSections;++x)
		{
			float tempf = (((float)x-(NumSections/2))/((float)NumSections/2.0f));

			tempf *= (leftGrab + rightGrab) * 0.5f;
			rotMatrix[0].x = cosf(-twist * tempf * 2.0f * 3.141592653f);
			rotMatrix[0].z = sinf(-twist * tempf * 2.0f * 3.141592653f);
			//rotMatrix[2].x = -sinf(-twist * tempf * 2.0f * 3.141592653f);
			//rotMatrix[2].z = cosf(-twist * tempf * 2.0f * 3.141592653f);

			tempThree.x = (vertices[(y*NumSections)+x].pos.x * rotMatrix[0].x) +
						  (vertices[(y*NumSections)+x].pos.z * rotMatrix[0].z);
			tempThree.y = vertices[(y*NumSections)+x].pos.y;
			tempThree.z = (vertices[(y*NumSections)+x].pos.x * -rotMatrix[0].z) +
						  (vertices[(y*NumSections)+x].pos.z * rotMatrix[0].x);

			vertices[(y*NumSections)+x].pos = tempThree;
		}
	}
}

//------------------------------------------------------------------------------
void DeformableSphere2::updateVertices2()
{
	int i, j, x, y;
	//ThreeFloats rotMatrix[3];
	ThreeFloats tempThree;
	int tempint;
	int vertexCount = 0;
	//static int thingCount = 0;

	//Calculate Tao surface.

	//Calculate forces.
	{
		//Propagate forces.
		j = 0;
		for(i=1;i<(SurfaceWidth-1);++i)
		{
			tempint = (j*SurfaceHeight)+i;
			forces[tempint] = (positions[tempint-1] +
							   positions[tempint+1] +
							   positions[tempint+SurfaceHeight-1] +
							   positions[tempint+SurfaceHeight] +
							   positions[tempint+SurfaceHeight+1] +
							   positions[(SurfaceWidth*(SurfaceHeight-1))+tempint-1] +
							   positions[(SurfaceWidth*(SurfaceHeight-1))+tempint] +
							   positions[(SurfaceWidth*(SurfaceHeight-1))+tempint+1]) -
							  (8.0f * positions[tempint]);
		}
		tempint = (j*SurfaceHeight);
		forces[tempint] = (positions[tempint+1] +
						   positions[tempint+SurfaceHeight] +
						   positions[tempint+SurfaceHeight+1] +
						   positions[(SurfaceWidth*(SurfaceHeight-1))+tempint] +
						   positions[(SurfaceWidth*(SurfaceHeight-1))+tempint+1]) -
						  (5.0f * positions[tempint]);

		tempint = (j*SurfaceHeight)+(SurfaceWidth-1);
		forces[tempint] = (positions[tempint-1] +
						   positions[tempint+SurfaceHeight] +
						   positions[tempint+SurfaceHeight-1] +
						   positions[(SurfaceWidth*(SurfaceHeight-1))+tempint] +
						   positions[(SurfaceWidth*(SurfaceHeight-1))+tempint-1]) -
						  (5.0f * positions[tempint]);

		for(j=1;j<(SurfaceHeight-1);++j)
		{
			for(i=1;i<(SurfaceWidth-1);++i)
			{
				tempint = (j*SurfaceHeight)+i;
				forces[tempint] = (positions[tempint-1] +
								   positions[tempint+1] +
								   positions[tempint-SurfaceHeight-1] +
								   positions[tempint-SurfaceHeight] +
								   positions[tempint-SurfaceHeight+1] +
								   positions[tempint+SurfaceHeight-1] +
								   positions[tempint+SurfaceHeight] +
								   positions[tempint+SurfaceHeight+1]) -
								  (8.0f * positions[tempint]);
			}
			tempint = (j*SurfaceHeight);
			forces[tempint] = (positions[tempint+1] +
							   positions[tempint-SurfaceHeight] +
							   positions[tempint-SurfaceHeight+1] +
							   positions[tempint+SurfaceHeight] +
							   positions[tempint+SurfaceHeight+1]) -
							  (5.0f * positions[tempint]);

			tempint = (j*SurfaceHeight)+(SurfaceWidth-1);
			forces[tempint] = (positions[tempint-1] +
							   positions[tempint-SurfaceHeight] +
							   positions[tempint-SurfaceHeight-1] +
							   positions[tempint+SurfaceHeight] +
							   positions[tempint+SurfaceHeight-1]) -
							  (5.0f * positions[tempint]);
		}

		for(i=1;i<(SurfaceWidth-1);++i)
		{
			tempint = (j*SurfaceHeight)+i;
			forces[tempint] = (positions[tempint-1] +
							   positions[tempint+1] +
							   positions[tempint-SurfaceHeight-1] +
							   positions[tempint-SurfaceHeight] +
							   positions[tempint-SurfaceHeight+1] +
							   positions[tempint-(j*SurfaceHeight)-1] +
							   positions[tempint-(j*SurfaceHeight)] +
							   positions[tempint-(j*SurfaceHeight)+1]) -
							  (8.0f * positions[tempint]);
		}
		tempint = (j*SurfaceHeight);
		forces[tempint] = (positions[tempint+1] +
						   positions[tempint-SurfaceHeight] +
						   positions[tempint-SurfaceHeight+1] +
						   positions[tempint-(j*SurfaceHeight)] +
						   positions[tempint-(j*SurfaceHeight)+1]) -
						  (5.0f * positions[tempint]);

		tempint = (j*SurfaceHeight)+(SurfaceWidth-1);
		forces[tempint] = (positions[tempint-1] +
						   positions[tempint-(j*SurfaceHeight)] +
						   positions[tempint-(j*SurfaceHeight)-1]) -
						  (3.0f * positions[tempint]);

		//Update force sensor values.
		forces[SurfaceWidth+1] += leftGrab;
		forces[(3*SurfaceWidth)+1] += leftSculpt;
		forces[SurfaceWidth+3] += rightGrab;
		forces[(3*SurfaceWidth)+3] += rightSculpt;
	}

	//Calculate positions.
	for(i=0;i<SurfaceHeight;++i)
	{
		for(j=1;j<(SurfaceWidth-1);++j)
		{
			tempint = (i*SurfaceWidth)+j;

			velocities[tempint] += forces[tempint] * invMass;
			velocities[tempint] *= velocityMultiplier;

			if(velocityMultiplier != defaultVelocityMultiplier)
				positions[tempint] = 0.0f;
			else
				positions[tempint] += velocities[tempint];
		}
	}

	if(fabs(positions[SurfaceWidth+1]) > 1.0f)
		velocityMultiplier *= 0.0f;
	else
		velocityMultiplier = defaultVelocityMultiplier;

	//Update vertices.
	for(y=0;y<SurfaceHeight;++y)
	{
		for(x=0;x<SurfaceWidth;++x)
		{
			bezierSubdivide(x, y, vertexCount);
			vertexCount += NumSubdivisions*NumSubdivisions;
		}
	}

	/*int inty = NumSections-1;
	cout << vertices[inty].sculptPos.x << endl;

	glDisable(GL_LIGHTING);
	glBegin(GL_LINE_STRIP);
	{
		glColor3f(1.0f, 1.0f, 1.0f);
		glVertex3f(vertices[inty].sculptPos.x,
				   vertices[inty].sculptPos.y,
				   vertices[inty].sculptPos.z + 4.0f);
		glVertex3f(vertices[inty].sculptPos.x,
				   vertices[inty].sculptPos.y,
				   vertices[inty].sculptPos.z - 4.0f);
	}
	glEnd();
	glEnable(GL_LIGHTING);*/

	for(y=0;y<NumSections;++y)
	{
		for(x=0;x<NumSections;++x)
			vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].sculptPos;
	}

	//Calculate stretch.
	/*for(y=0;y<NumSections;++y)
	{
		for(x=0;x<(NumSections/2);++x)
		{
			vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].sculptPos;
			vertices[(y*NumSections)+x].pos.y *= rightStretch;
		}
		for(x=(NumSections/2);x<NumSections;++x)
		{
			vertices[(y*NumSections)+x].pos = vertices[(y*NumSections)+x].sculptPos;
			vertices[(y*NumSections)+x].pos.y *= leftStretch;
		}
	}

	//Calculate rotation.
	for(y=0;y<NumSections;++y)
	{
		for(x=0;x<(NumSections/2);++x)
		{
			float tempf = (1.0f-((float)x/((float)NumSections/2.0f)));

			tempf *= (leftGrab + rightGrab) * 0.5f;
			rotMatrix[0].x = cosf(twist * tempf * 2.0f * 3.141592653f);
			rotMatrix[0].z = sinf(twist * tempf * 2.0f * 3.141592653f);

			tempThree.x = (vertices[(y*NumSections)+x].pos.x * rotMatrix[0].x) +
						  (vertices[(y*NumSections)+x].pos.z * rotMatrix[0].z);
			tempThree.y = vertices[(y*NumSections)+x].pos.y;
			tempThree.z = (vertices[(y*NumSections)+x].pos.x * -rotMatrix[0].z) +
						  (vertices[(y*NumSections)+x].pos.z * rotMatrix[0].x);

			vertices[(y*NumSections)+x].pos = tempThree;
		}
		for(x=(NumSections/2)+1;x<NumSections;++x)
		{
			float tempf = (((float)x-(NumSections/2))/((float)NumSections/2.0f));

			tempf *= (leftGrab + rightGrab) * 0.5f;
			rotMatrix[0].x = cosf(-twist * tempf * 2.0f * 3.141592653f);
			rotMatrix[0].z = sinf(-twist * tempf * 2.0f * 3.141592653f);

			tempThree.x = (vertices[(y*NumSections)+x].pos.x * rotMatrix[0].x) +
						  (vertices[(y*NumSections)+x].pos.z * rotMatrix[0].z);
			tempThree.y = vertices[(y*NumSections)+x].pos.y;
			tempThree.z = (vertices[(y*NumSections)+x].pos.x * -rotMatrix[0].z) +
						  (vertices[(y*NumSections)+x].pos.z * rotMatrix[0].x);

			vertices[(y*NumSections)+x].pos = tempThree;
		}
	}*/
}

}
