//	DeformableSphere2.h - A sphere which can be deformed in various ways.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef DEFORMABLESPHERE2_H_
#define DEFORMABLESPHERE2_H_

#include "AbstractNode.h"
#include "CriticalSection.h"
#include "../helper/PerlinTexture.h"
#include "../helper/StraightSineTexture.h"
#include "../helper/ThresholdPerlinTexture.h"

#include "GLHeaders.h"
#include <cmath>

/*+
	Name: DeformableSphere2
	Type: ChildNode.
	Description: A sphere which may be deformed by manipulating it's attributes.
			 	 This is a test node for my final audiovisual instrument, hence
				 the slightly counterintuitive attributes (they make more sense
				 when considering the physical interface I'm designing for it).
	Attribute:	leftGrab
	AttType:	SFFloat (input)
	AttDesc:	'Grabs' the left side of the sphere, an operation which
				modulates the other attributes.

	Attribute:	rightGrab
	AttType:	SFFloat (input)
	AttDesc:	'Grabs' the right side of the sphere, an operation which
				modulates the other attributes.

	Attribute:	stretch
	AttType:	SFFloat (input)
	AttDesc:	When the grabs are 0, stretches the sphere along the x
				dimension.  Otherwise changes the sculpting position (x-axis) of
				the left and/or right sides.

	Attribute:	twist
	AttType:	SFFloat (input)
	AttDesc:	When the grabs are 0, twists the sphere around the x axis.
				Otherwise changes the sculpting position (rotating around the
				object) of the left and/or right sides.

	Attribute:	leftSculpt
	AttType:	SFFloat (input)
	AttDesc:	When leftGrab is non-zero, controls the radius (from the
				sphere's centre) of the vertex at the current left sculpting
				position.

	Attribute:	rightSculpt
	AttType:	SFFloat (input)
	AttDesc:	When rightGrab is non-zero, controls the radius (from the
				sphere's centre) of the vertex at the current right sculpting
				position.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY" headlight="FALSE"/&gt;

		&lt;Transform DEF="temp" translation="0.0 0.0 -8.0"&gt;
			&lt;DeformableSphere2 DEF="bob"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	A sphere which can be deformed in various ways.
/*!
	\ingroup libheilanextras
 */
class DeformableSphere2 : public ChildNode
{
  public:
	///	Constructor.
	DeformableSphere2();
	///	Destructor.
	~DeformableSphere2();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the sphere.
	void graphicDoStuff();
	///	Not used, overridden to do nothing.
	/*AmbisonicData *audioDoStuff(int numSamples,
								const ThreeFloats& pos) {return 0;};*/

	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set the sphere's parameters.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<DeformableSphere2> constructor;
  private:
	///	Helper method to calculate the normal of a polygon.
	inline void calcNormal(const ThreeFloats& v1,
						   const ThreeFloats& v2,
						   const ThreeFloats& v3)
	{
		ThreeFloats a;
		ThreeFloats b;
		ThreeFloats normal;
		float length;

		a.x = v1.x - v2.x;
		a.y = v1.y - v2.y;
		a.z = v1.z - v2.z;

		b.x = v2.x - v3.x;
		b.y = v2.y - v3.y;
		b.z = v2.z - v3.z;

		normal.x = (a.y * b.z) - (a.z * b.y);
		normal.y = (a.z * b.x) - (a.x * b.z);
		normal.z = (a.x * b.y) - (a.y * b.x);

		length = sqrtf((normal.x*normal.x)+(normal.y*normal.y)+(normal.z*normal.z));
		if(length == 0.0f)
			length = 1.0f;
		else
			length = 1.0f/length;

		glNormal3f((normal.x*length), (normal.y*length), (normal.z*length));
	};
	///	Second version, to return the normal instead of drawing it.
	inline ThreeFloats calcNormal2(const ThreeFloats& v1,
								   const ThreeFloats& v2,
								   const ThreeFloats& v3)
	{
		ThreeFloats a;
		ThreeFloats b;
		ThreeFloats normal;
		float length;

		a.x = v1.x - v2.x;
		a.y = v1.y - v2.y;
		a.z = v1.z - v2.z;

		b.x = v2.x - v3.x;
		b.y = v2.y - v3.y;
		b.z = v2.z - v3.z;

		normal.x = (a.y * b.z) - (a.z * b.y);
		normal.y = (a.z * b.x) - (a.x * b.z);
		normal.z = (a.x * b.y) - (a.y * b.x);

		length = sqrtf((normal.x*normal.x)+(normal.y*normal.y)+(normal.z*normal.z));
		if(length == 0.0f)
			length = 1.0f;
		else
			length = 1.0f/length;

		normal.x *= length;
		normal.y *= length;
		normal.z *= length;

		return normal;
	};
	///	To calculate smooth normals.
	inline ThreeFloats calcSmooth(ThreeFloats& n1,
								  ThreeFloats& n2,
								  ThreeFloats& n3,
								  ThreeFloats& n4)
	{
		ThreeFloats retval;

		retval.x = (n1.x + n2.x + n3.x + n4.x)*0.25f;
		retval.y = (n1.y + n2.y + n3.y + n4.y)*0.25f;
		retval.z = (n1.z + n2.z + n3.z + n4.z)*0.25f;

		return retval;
	};

	///	Helper method to calculate a vertex on the sphere.
	void calcVertex(int x, int y);
	///	Helper method for calcVertex().
	inline ThreeFloats& getVertex(int x, int y)
	{
		/*ThreeFloats retval;

		retval.x = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) * 
				   cosf(((float)y/(float)NumSections) * 6.283185306f);
		retval.y = sinf((((float)x/(float)NumSections)-0.5f) * 3.141592653f);
		retval.z = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) *
				   sinf(((float)y/(float)NumSections) * 6.283185306f);

		return retval;*/
		if(x >= NumSections)
			x -= NumSections;
		if(y >= NumSections)
			y -= NumSections;

		return vertices[(y*NumSections)+x].pos;
	}

	///	Helper method to update vertices after the performer has modified them.
	/*!
		MUST BE CALLED WITHIN A critSec.lock()/critSec.unlock() PAIR!!!!
	 */
	void updateVertices();
	///	Helper method to update vertices according to the Tao surface.
	void updateVertices2();

	///	Linear interpolation helper method.
	/*!
		Assumes val is of the range 0->1.
	 */
	inline float linearInterp(float left, float right, float val)
	{
		float retval;

		retval = left + (val*(right-left));

		return retval;
	};
	///	B-spline interpolation helper method.
	inline float bSplineInterp(float x0,
							   float x1,
							   float x2,
							   float x3,
							   float val)
	{
		// 4-point, 3rd-order B-spline (x-form)
		float ym1py1 = x0+x2;
		float c0 = (1.0f/6.0f) * ym1py1 + (2.0f/3.0f) * x1;
		float c1 = (1.0f/2.0f) * (x2-x0);
		float c2 = (1.0f/2.0f) * ym1py1 - x1;
		float c3 = (1.0f/2.0f) * (x1-x2) + (1.0f/6.0f) * (x3-x0);
		return ((c3*val+c2)*val+c1)*val+c0;
	};
	///	Helper method to subdivide and draw a quad.
	/*!
		\param x The x position within positions of the lower left vertex.
		\param y The y position within positions of the lower left vertex.
		\param vertexCount The vertex in vertices that we're focussing on.
	 */
	inline void bezierSubdivide(int x, int y, int vertexCount)
	{
		int i, j;
		float tempi, tempj;
		float temp1, temp2, temp3, temp4;
		int tempX, tempY;
		float tempVertices[NumSubdivisions+1][NumSubdivisions+1];
		float corners[4][4];
		int tempint;

		//Calculate corner positions.
		for(i=-1;i<3;++i)
		{
			for(j=-1;j<3;++j)
			{
				tempX = x + j;
				if(tempX < 0)
					tempX += SurfaceWidth;
				else if(tempX >= SurfaceWidth)
					tempX -= SurfaceWidth;

				tempY = y + i;
				if(tempY < 0)
					tempY += SurfaceHeight;
				else if(tempY >= SurfaceHeight)
					tempY -= SurfaceHeight;

				corners[i+1][j+1] = positions[tempX + (tempY*SurfaceWidth)] * 2.0f;
			}
		}

		//y-axis
		for(i=0;i<NumSubdivisions;++i)
		{
			tempi = (float)i/(float)(NumSubdivisions-1);

			//x-axis
			for(j=0;j<NumSubdivisions;++j)
			{
				tempj = (float)j/(float)(NumSubdivisions-1);

				if(spikyMode)
				{
					temp1 = linearInterp(corners[1][1], corners[1][2], tempj);
					temp2 = linearInterp(corners[2][1], corners[2][2], tempj);
					tempVertices[j][i] = linearInterp(temp1, temp2, tempi);
				}
				else
				{
					temp1 = bSplineInterp(corners[0][0],
										  corners[0][1],
										  corners[0][2],
										  corners[0][3],
										  tempj);
					temp2 = bSplineInterp(corners[1][0],
										  corners[1][1],
										  corners[1][2],
										  corners[1][3],
										  tempj);
					temp3 = bSplineInterp(corners[2][0],
										  corners[2][1],
										  corners[2][2],
										  corners[2][3],
										  tempj);
					temp4 = bSplineInterp(corners[3][0],
										  corners[3][1],
										  corners[3][2],
										  corners[3][3],
										  tempj);
					tempVertices[j][i] = bSplineInterp(temp1,
													   temp2,
													   temp3,
													   temp4,
													   tempi);
				}
			}
		}

		tempint = vertexCount;
		if(spikyMode)
		{
			float thing1 = 0.0f;
			float thing2 = 0.0f;
			for(i=0;i<NumSubdivisions;++i)
			{
				for(j=0;j<NumSubdivisions;++j)
				{
					vertices[tempint].sculptPos.x = vertices[tempint].defaultPos.x *
													(1.0f + tempVertices[j][i]) *
													(thing1*thing2);
					vertices[tempint].sculptPos.z = vertices[tempint].defaultPos.z *
													(1.0f + tempVertices[j][i]) *
													(thing1*thing2);
					vertices[tempint].sculptPos.y = vertices[tempint].defaultPos.y *
													//(thing1*thing2);
													thing1;
					++tempint;
					thing1 = 1.0f-thing1;
				}
				tempint -= NumSubdivisions;
				tempint += NumSections;
				thing2 = 1.0f-thing2;
			}

		}
		else
		{
			for(i=0;i<NumSubdivisions;++i)
			{
				for(j=0;j<NumSubdivisions;++j)
				{
					vertices[tempint].sculptPos.x = vertices[tempint].defaultPos.x *
													(1.0f + tempVertices[j][i]);
					vertices[tempint].sculptPos.z = vertices[tempint].defaultPos.z *
													(1.0f + tempVertices[j][i]);
					vertices[tempint].sculptPos.y = vertices[tempint].defaultPos.y;
					//++vertexCount;
					++tempint;
				}
				tempint -= NumSubdivisions;
				tempint += NumSections;
			}
		}
	};

	///	Struct representing a vertex in the sphere.
	struct Vertex
	{
		///	The Vertex' current position.
		ThreeFloats pos;
		///	The Vertex' default position.
		ThreeFloats defaultPos;
		///	The Vertex' sculpted position.
		ThreeFloats sculptPos;
	};

	///	Enum containing the number of sides we use to draw the sphere.
	enum
	{
		SurfaceWidth = 5,
		SurfaceHeight = 5,
		NumSubdivisions = 6,
		NumSections = (NumSubdivisions*SurfaceWidth),
		TextureSize = 256 //Width and height both equal TextureSize.
	};

	///	Used to protect the vertices while they're being drawn.
	CriticalSection critSec;

	///	How much the sphere is being stretched by.
	float stretch;
	///	How much the sphere is being twisted by.
	float twist;
	///	Left sculptor control.
	float leftSculpt;
	///	Right sculptor control.
	float rightSculpt;
	///	Left grabber control.
	float leftGrab;
	///	Right grabber control.
	float rightGrab;
	///	How much the left side is stretched by.
	float leftStretch;
	///	How much the right side is stretched by.
	float rightStretch;
	///	Current x-axis acceleration.
	float xAccel;
	///	Current y-axis acceleration.
	float yAccel;
	///	Current z-axis acceleration.
	float zAccel;
	///	Whether the sphere's in spiky mode (linear interp), or smooth mode (bezier).
	bool spikyMode;

	///	X-axis translation of the 1st texture.
	float tex1X;
	///	Y-axis translation of the 1st texture.
	float tex1Y;
	///	X-axis scale of the 1st texture.
	float tex1SX;
	///	Y-axis scale of the 1st texture.
	float tex1SY;
	///	Rotation angle of the 1st texture.
	float tex1Rot;
	///	X-axis translation of the 2nd texture.
	float tex2X;
	///	Y-axis translation of the 2nd texture.
	float tex2Y;
	///	X-axis scale of the 2nd texture.
	float tex2SX;
	///	Y-axis scale of the 2nd texture.
	float tex2SY;
	///	Rotation angle of the 2nd texture.
	float tex2Rot;
	///	How much the two textures are blended by (0 == tex1, 1 == tex2).
	float texBlend;

	///	All the Vertices in the sphere.
	Vertex *vertices;
	//Vertex vertices[NumSections*NumSections];
	///	Current left sculpt length position.
	int leftSculptL;
	///	Current left sculpt twist position.
	int leftSculptT;
	///	Current right sculpt length position.
	int rightSculptL;
	///	Current right sculpt twist position.
	int rightSculptT;

	///	Used to update the vertices in graphicDoStuff() when necessary.
	bool verticesNeedUpdated;

	///	Delay line for the surface (i.e. the positions of the cells).
	float positions[SurfaceWidth*SurfaceHeight];
	///	Forces for the string cells.
	float forces[SurfaceWidth*SurfaceHeight];
	///	Velocities for the string cells.
	float velocities[SurfaceWidth*SurfaceHeight];
	///	Mass of the individual cells.
	float mass;
	///	Inverse mass of the individual cells.
	float invMass;
	///	How much to multiply the cells' velocities by.
	float velocityMultiplier;
	///	The default velocity multiplier.
	float defaultVelocityMultiplier;
	///	The subdivided vertex positions from the surface.
	float *vertexPositions;
	//float vertexPositions[SurfaceWidth*NumSubdivisions*SurfaceHeight*NumSubdivisions];

	///	True if the current graphics card supports multitexturing.
	bool multiTextureSupported;

	///	Noise Texture.
	RubbishNoiseTexture *noiseTex;
	///	Perlin noise Texture.
	PerlinTexture *perlinTex;
	///	Sine Texture.
	StraightSineTexture *sineTex;
	///	Perlin noise Texture.
	ThresholdPerlinTexture *thresholdPerlinTex;
	///	Index to the noise texture.
	GLuint noiseId;
	///	Index to the perlin texture.
	GLuint perlinId;
	///	Index to grey texture.
	GLuint sineId;
	///	Index to the thresholded perlin texture.
	GLuint thresholdPerlinId;
};

}

#endif
