//	DynamicTexture.h - A procedural texture which changes over time.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef DYNAMICTEXTURE_H_
#define DYNAMICTEXTURE_H_

#include "AbstractNode.h"
#include "CriticalSection.h"

#include "GLHeaders.h"
#include <cmath>
#include <iostream>

/*+
	Name: DynamicTexture
	Type: ChildNode.
	Description: A basic procedurally-generated texture based on Perlin noise.
	Attribute:	repeatS
	AttType:	SFBool (input)
	AttDesc:	Whether to repeat the texture past its bounds or not.

	Attribute:	repeatT
	AttType:	SFBool (input)
	AttDesc:	Whether to repeat the texture past its bounds or not.

	Attribute:	persistence
	AttType:	SFFloat (input)
	AttDesc:	How smooth the noise is.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;Transform translation="0.0 0.0 0.0"&gt;
			&lt;Shape&gt;
				&lt;Appearance&gt;
					&lt;DynamicTexture DEF="bob"/&gt;
				&lt;/Appearance&gt;
				&lt;Box/&gt;
			&lt;/Shape&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Class representing a basic texture.
/*!
	\ingroup libheilanextras
 */
class DynamicTexture : public ChildNode
{
  public:
	///	Constructor.
	DynamicTexture();
	///	Destructor.
	~DynamicTexture();

	///	Initialises the textures.
	void initialise();

	///	Draws the DynamicTexture.
	void graphicDoStuff();
	///	Not used, overridden to do nothing.
	/*AmbisonicData *audioDoStuff(int numSamples,
								const ThreeFloats& pos) {return 0;};*/

	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set the material's attributes.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<DynamicTexture> constructor;
  private:
	///	Helper method to generate the noise texture.
	void generateNoise();
	///	Helper method to interpolate between pixel values.
	inline float bSplineInterp(float x0,
							   float x1,
							   float x2,
							   float x3,
							   float val)
	{
		// 4-point, 3rd-order B-spline (x-form)
		float ym1py1 = x0+x2;
		float c0 = (1.0f/6.0f) * ym1py1 + (2.0f/3.0f) * x1;
		float c1 = (1.0f/2.0f) * (x2-x0);
		float c2 = (1.0f/2.0f) * ym1py1 - x1;
		float c3 = (1.0f/2.0f) * (x1-x2) + (1.0f/6.0f) * (x3-x0);
		return ((c3*val+c2)*val+c1)*val+c0;
	};
	///	Helper method to generate the main texture.
	inline void generateTexture()
	{
		int i, j, k, l;
		int x, y;
		int tempint;
		float tempX, tempY;
		int tempX2, tempY2;
		float corners[4][4];
		float temp1, temp2, temp3, temp4;
		float tempP;

		critSec.lock();
		{

		//Clear main texture.
		for(y=0;y<TextureSize;++y)
		{
			for(x=0;x<TextureSize;++x)
				mainTexture[(y*TextureSize)+x] = 0.5f;
		}

		//Loop through noise octaves.
		for(i=4, j=0;i<TextureSize;i*=2, ++j)
		//for(i=4, j=0;i<64;i*=2, ++j)
		{
			tempP = powf(persistence, (float)i);

			//Go through the texture's pixels, add to mainTexture.
			for(y=0;y<TextureSize;++y)
			{
				tempY = (float)y/(float)TextureSize;
				tempY *= (float)i;

				for(x=0;x<TextureSize;++x)
				{
					tempX = (float)x/(float)TextureSize;
					tempX *= (float)i;

					tempint = (y*TextureSize)+x;

					//Get corner pixel values.
					for(k=-1;k<3;++k)
					{
						for(l=-1;l<3;++l)
						{
							tempX2 = (int)tempX + l;
							if(tempX2 < 0)
								tempX2 += i;
							else if(tempX2 >= i)
								tempX2 -= i;

							tempY2 = (int)tempY + k;
							if(tempY2 < 0)
								tempY2 += i;
							else if(tempY2 >= i)
								tempY2 -= i;

							corners[k+1][l+1] = noiseTexture[tempX2 +
															 (tempY2*TextureSize)];
						}
					}

					//Interpolate.
					temp1 = bSplineInterp(corners[0][0],
										  corners[0][1],
										  corners[0][2],
										  corners[0][3],
										  (tempX-(float)((int)tempX)));
					temp2 = bSplineInterp(corners[1][0],
										  corners[1][1],
										  corners[1][2],
										  corners[1][3],
										  (tempX-(float)((int)tempX)));
					temp3 = bSplineInterp(corners[2][0],
										  corners[2][1],
										  corners[2][2],
										  corners[2][3],
										  (tempX-(float)((int)tempX)));
					temp4 = bSplineInterp(corners[3][0],
										  corners[3][1],
										  corners[3][2],
										  corners[3][3],
										  (tempX-(float)((int)tempX)));
					mainTexture[tempint] += (bSplineInterp(temp1,
												  	       temp2,
												   	       temp3,
												   	       temp4,
												   	       (tempY-(float)((int)tempY)))-0.5f) *
										     tempP;
				}
			}
		}

		//Add final noise texture.
		tempP = powf(persistence, (float)i);
		for(y=0;y<TextureSize;++y)
		{
			for(x=0;x<TextureSize;++x)
			{
				tempint = (y*TextureSize)+x;
				mainTexture[tempint] += (noiseTexture[tempint]-0.5f) * tempP;
			}
		}

		//Write to texture.
		glBindTexture(GL_TEXTURE_2D, texture);

		glTexImage2D(GL_TEXTURE_2D,
					 0,
					 GL_RGBA,
					 TextureSize,
					 TextureSize,
					 0,
					 GL_LUMINANCE,
					 GL_FLOAT,
					 mainTexture);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		}
		critSec.unlock();
	};

	///	Some constants.
	enum
	{
		TextureSize = 256 //Width and height both equal TextureSize.
	};

	///	Whether the texture should repeat in the S direction.
	bool repeatS;
	///	Whether the texture should repeat in the T direction.
	bool repeatT;
	///	The persistence value used to generate the noise.
	float persistence;

	///	The texture we use.
	GLuint texture;

	///	The basic noise texture we base the whole thing on.
	float noiseTexture[TextureSize*TextureSize];

	///	The actual perlin noise texture.
	float mainTexture[TextureSize*TextureSize];

	///	Critical section to protext mainTexture.
	CriticalSection critSec;
};

}

#endif
