//	GestureDetector.h - A simple visualisation of gestures detected from a 3D
//						accelerometer.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef GESTUREDETECTOR_H_
#define GESTUREDETECTOR_H_

#include "AbstractNode.h"
#include "Ashitaka/Accel2Pos.h"

#include "GLHeaders.h"
#include <iostream>

/*+
	Name: GestureDetector
	Type: ChildNode.
	Description: A basic gesture detector, used to detect clockwise and
				 anti-clockwise circles, upside-down v shapes, and sharp jerks
				 to the left and the right.
	Attribute:	accelX
	AttType:	SFFloat (input/output)
	AttDesc:	The current x-axis value of the accelerometer input, used to
				detect the known gestures, and passed on as an output.

	Attribute:	accelY
	AttType:	SFFloat (input/output)
	AttDesc:	The current y-axis value of the accelerometer input, used to
				detect the known gestures, and passed on as an output.

	Attribute:	accelZ
	AttType:	SFFloat (input/output)
	AttDesc:	The current z-axis value of the accelerometer input, used to
				detect the known gestures, and passed on as an output.

	Attribute:	debugVisuals
	AttType:	SFBool (input)
	AttDesc:	Shows the current state of the detector.

	Attribute:	useParticles
	AttType:	SFBool (input)
	AttDesc:	When a gesture is detected, visualise it as a trail of
				particles.

	Attribute:	match_1
	AttType:	SFBool (output)
	AttDesc:	Sends true when an anti-clockwise circle (around the z-axis) is
				detected.

	Attribute:	match_2
	AttType:	SFBool (output)
	AttDesc:	Sends true when an clockwise circle (around the z-axis) is
				detected.

	Attribute:	match_3
	AttType:	SFBool (output)
	AttDesc:	Sends true when an upside-down v (along the x-y plane) is
				detected.

	Attribute:	leftJerk
	AttType:	SFBool (output)
	AttDesc:	Sends true when a jerk towards the left is detected.

	Attribute:	rightJerk
	AttType:	SFBool (output)
	AttDesc:	Sends true when a jerk towards the right is detected.

	Example:
&lt;?xml version='1.0' encoding='utf-8'?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile='Full'&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type='ANY' headlight='FALSE'/&gt;

		&lt;Ashitaka2 DEF='bob'/&gt;
		&lt;Billboard&gt;
			&lt;GestureDetector DEF='gertrude' useParticles='true'/&gt;
		&lt;/Billboard&gt;

		&lt;ROUTE fromNode='gertrude' fromField='match_1' toNode='bob' toField='split1'/&gt;
		&lt;ROUTE fromNode='gertrude' fromField='match_2' toNode='bob' toField='split2'/&gt;
		&lt;ROUTE fromNode='gertrude' fromField='match_3' toNode='bob' toField='split3'/&gt;
		&lt;ROUTE fromNode='gertrude' fromField='leftJerk' toNode='bob' toField='rotateRight'/&gt;
		&lt;ROUTE fromNode='gertrude' fromField='rightJerk' toNode='bob' toField='rotateLeft'/&gt;

		&lt;ROUTE fromNode='gertrude' fromField='accelX' toNode='bob' toField='accelX'/&gt;
		&lt;ROUTE fromNode='gertrude' fromField='accelY' toNode='bob' toField='accelY'/&gt;
		&lt;ROUTE fromNode='gertrude' fromField='accelZ' toNode='bob' toField='accelZ'/&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

class ParticleGenerator;

///	A simple realtime visualisation of gestures detected from a 3D accelerometer.
/*!
	Designed to be hooked up to Ashitaka2 to trigger certain actions when
	particular gestures are detected.
 */
class GestureDetector : public ChildNode
{
  public:
	///	Constructor.
	GestureDetector();
	///	Destructor.
	~GestureDetector();

	///	Used to load the textures.
	void initialise();

	///	Draws the AccelLine.
	void graphicDoStuff();

	///	Overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set this node's attributes.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<GestureDetector> constructor;
  private:
	///	Represents a gesture, with various methods for doing stuff.
	/*!
		A Gesture is just a sequence of 10 ThreeFloats.
	 */
	class Gesture
	{
	  public:
		///	Constrctor.
		Gesture() {};
		///	Destrctor.
		~Gesture() {};

		///	Returns the indexed point on the gesture (does no range checking).
		ThreeFloats& operator[](int index)
		{
			return points[index];
		};
		///	Copies the gesture data from another Gesture.
		Gesture& operator=(const Gesture& op2)
		{
			int i;

			for(i=0;i<10;++i)
				points[i] = op2.points[i];

			return *this;
		};

		///	Normalises the gesture to have a size of 1.
		/*!
			\return The amount we had to scale by.
		 */
		float normalise()
		{
			int i;
			float scale;
			ThreeFloats size;
			//The data we get from the interface is -20 -> +20.
			ThreeFloats minVal(20.0f);
			ThreeFloats maxVal(-20.0f);

			for(i=0;i<10;++i)
			{
				if(minVal.x > points[i].x)
					minVal.x = points[i].x;
				if(maxVal.x < points[i].x)
					maxVal.x = points[i].x;
				if(minVal.y > points[i].y)
					minVal.y = points[i].y;
				if(maxVal.y < points[i].y)
					maxVal.y = points[i].y;
				if(minVal.z > points[i].z)
					minVal.z = points[i].z;
				if(maxVal.z < points[i].z)
					maxVal.z = points[i].z;
			}
			if(maxVal != minVal)
			{
				size = maxVal - minVal;

				if((size.x > size.y) && (size.x > size.z))
					scale = 1.0f/size.x;
				else if((size.y > size.x) && (size.y > size.z))
					scale = 1.0f/size.y;
				else
					scale = 1.0f/size.z;

				for(i=0;i<10;++i)
					points[i] *= scale;

				return scale;
			}
			return 0.0f;
		};

		///	Distributes the 10 points evenly.
		/*!
			i.e. Make the distances between points the same.
		 */
		void spacePoints()
		{
			int i, j, k;
			float tempf;
			float avgLength = 0.0f;
			ThreeFloats tempP[10];

			for(i=1;i<10;++i)
				avgLength += points[i].getDistance(points[i-1]);
			avgLength *= 0.1f;

			if(avgLength == 0.0f)
				return;

			tempP[0] = points[0];
			tempP[9] = points[9];
			j = 1;
			k = 0;
			//This hurts my head...
			for(i=1;i<9;++i)
			{
				k = j;
				tempf = points[j].getDistance(tempP[i-1]);
				while(tempf < avgLength)
				{
					++j;
					tempf += points[j].getDistance(points[j-1]);
				}

				if(k < j)
					tempf -= points[j].getDistance(points[j-1]);
				//The position within the current segment.
				if(avgLength > tempf)
					tempf = (avgLength-tempf)/points[j].getDistance(points[j-1]);
				else
					tempf = (tempf-avgLength)/points[j].getDistance(points[j-1]);

				//Linear interpolation.
				//tempP[i] = (points[j] * tempf) + (points[j-1] * (1.0f-tempf));
				//B-Spline interpolation.
				tempP[i].x = bSplineInterp(tempf,
										   points[j>1?j-2:9].x,
										   points[j-1].x,
										   points[j].x,
										   points[(j+1)%10].x);
				tempP[i].y = bSplineInterp(tempf,
										   points[j>1?j-2:9].y,
										   points[j-1].y,
										   points[j].y,
										   points[(j+1)%10].y);
				tempP[i].z = bSplineInterp(tempf,
										   points[j>1?j-2:9].z,
										   points[j-1].z,
										   points[j].z,
										   points[(j+1)%10].z);

				tempf = points[j].getDistance(points[j-1]) - tempf;
				if(tempf < avgLength)
					++j;
			}
			for(i=1;i<9;++i)
				points[i] = tempP[i];
		};

		///	Centres the gesture on (0,0,0).
		void centre()
		{
			int i;
			ThreeFloats temp3f;

			for(i=0;i<10;++i)
				temp3f += points[i];
			temp3f *= 0.1f;

			for(i=0;i<10;++i)
				points[i] -= temp3f;
		};

		///	Calculates the dot product between this gesture and another.
		float dot(const Gesture& gest) const
		{
			int i;
			float retval = 0.0f;

			for(i=0;i<10;++i)
			{
				//I'm confused.  How is it possible for me to access
				//gest.points directly?  Methinks I need to read up on my c++...
				retval += (points[i].x * gest.points[i].x);
				retval += (points[i].y * gest.points[i].y);
				retval += (points[i].z * gest.points[i].z);
			}

			return retval;
		};
	  private:
		///	Helper b-spline interpolation method.
		inline float bSplineInterp(float val,
								   float x0,
								   float x1,
								   float x2,
								   float x3)
		{
			// 4-point, 3rd-order B-spline (x-form)
			float ym1py1 = x0+x2;
			float c0 = (1.0f/6.0f) * ym1py1 + (2.0f/3.0f) * x1;
			float c1 = (1.0f/2.0f) * (x2-x0);
			float c2 = (1.0f/2.0f) * ym1py1 - x1;
			float c3 = (1.0f/2.0f) * (x1-x2) + (1.0f/6.0f) * (x3-x0);
			return ((c3*val+c2)*val+c1)*val+c0;
		};

		///	The points on the gesture.
		ThreeFloats points[10];
	};

	///	Indicates whether we've found a match.
	enum MatchingGesture
	{
		NoMatch,
		SomeMatch,
		FullMatch
	};

	///	Whether to display debug visuals or not.
	bool showDebugVisuals;
	///	Whether to display particles or textures when a good match is detected.
	bool useParticles;

	///	The previous 10 positions (i.e. the gesture to test).
	Gesture gesture;
	///	The recorded gesture.
	Gesture recGest;
	///	Our stored gestures.
	Gesture testGesture[3];
	///	Whether we've got a match.
	MatchingGesture match;
	///	Which gesture we've matched against.
	int matchedGesture;
	///	Whether we've detected a left(1) or right(2) jerk, or not(0).
	int jerk;

	///	Counter used to determine when to update gesture.
	int gestureCount;
	///	Counter used to ensure we don't send events repeatedly for a single match.
	int matchCount;
	///	Counter used to ensure we don't send events repeatedly for a single jerk.
	int jerkCount;

	///	Our three match textures, to be displayed when there's a good match.
	GLuint matchTextures[3];
	///	Used to fade out the match textures.
	float matchFade;
	///	Which match texture to display.
	int whichTexture;

	///	Our two jerk textures, to be displayed when the performer jerks it interface.
	GLuint jerkTextures[2];
	///	Used to fade out the jerk textures.
	float jerkFade;
	///	Which jerk texture to display.
	int whichJerk;

	///	Used to draw gestures when they're detected.
	ParticleGenerator *particles;

	///	The current X-position.
	Accel2Pos accelX;
	///	The current Y-position.
	Accel2Pos accelY;
	///	The current Z-position.
	Accel2Pos accelZ;
};

}

///	PNG textures as char arrays, generated from JUCE' BinaryBuilder.
namespace GestureTextures
{
    extern const char*  clockwisegesture_png;
    const int           clockwisegesture_pngSize = 82800;

    extern const char*  anticlockwisegesture_png;
    const int           anticlockwisegesture_pngSize = 83176;

    extern const char*  ngesture_png;
    const int           ngesture_pngSize = 35630;

    extern const char*  rightarrow_png;
    const int           rightarrow_pngSize = 22068;

    extern const char*  leftarrow_png;
    const int           leftarrow_pngSize = 22238;
};

#endif
