//	GhostImage.cpp - Node which renders all its children to a texture, then
//				   	 overlays it on top of them, alpha-blended.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "GhostImage.h"
#include "Camera.h"
#include "StringParser.h"

using namespace std;

namespace HeilanX3D
{

NodeConstructor<GhostImage> GhostImage::constructor;

//------------------------------------------------------------------------------
GhostImage::GhostImage():
ParentNode(),
scale(0.0f),
trigger(0.0f),
pass(false),
currentScale(0.0f)
{
	setType("GhostImage");
	setVisual(true);
	setShouldBeStart(true);
}

//------------------------------------------------------------------------------
GhostImage::~GhostImage()
{
	
}

//------------------------------------------------------------------------------
void GhostImage::setAttribute(const string& attribute, EventValue& val)
{
	StringParser parser(val);

	if(attribute == "scale")
	{
		scale = val;
		currentScale = scale;
	}
	else if(attribute == "trigger")
	{
		if((trigger < 0.5f) && ((float)val >= 0.5f))
		{
			trigger = 1.0f;
			currentScale = 0.0f;
		}
		else if((float)val < 0.5f)
			trigger = 0.0f;
	}
}

//------------------------------------------------------------------------------
void GhostImage::parentInitialise()
{
	int i;

	//Setup background texture.
	for(i=0;i<(TexWidth*TexHeight*3);++i)
		backgroundPixels[i] = 0;

	glGenTextures(1, &backgroundTexture);
	glBindTexture(GL_TEXTURE_2D, backgroundTexture);
	glTexImage2D(GL_TEXTURE_2D,
				 0,
				 3,
				 TexWidth,
				 TexHeight,
				 0,
				 GL_RGB,
				 GL_UNSIGNED_BYTE,
				 backgroundPixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

//------------------------------------------------------------------------------
void GhostImage::preChildDraw()
{
	if(!pass)
	{
		const int viewWidth = Camera::getInstance()->getViewWidth();
		const int viewHeight = Camera::getInstance()->getViewHeight();

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();

		//Set viewport to the texture's dimensions.
		glViewport(0, 0, TexWidth, TexHeight);

		//Switch to an ortho mode.
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, viewWidth, viewHeight, 0, -1, 16);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		//Switch back to perspective mode.
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}
}

//------------------------------------------------------------------------------
void GhostImage::postChildDraw()
{
	if(!pass)
	{
		const int viewWidth = Camera::getInstance()->getViewWidth();
		const int viewHeight = Camera::getInstance()->getViewHeight();

		if(currentScale == 0.0f)
		{
			glBindTexture(GL_TEXTURE_2D, backgroundTexture);
			glCopyTexImage2D(GL_TEXTURE_2D,
							 0,
							 GL_RGB8,
							 0,
							 0,
							 TexWidth,
							 TexHeight,
							 0);
		}
	
		glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		//Reset the viewport.
		glViewport(0,
				   0,
				   viewWidth,
				   viewHeight);

		//Draw children again.
		glBindTexture(GL_TEXTURE_2D, 0);
		pass = true;
		graphicDoStuff();
		glBindTexture(GL_TEXTURE_2D, backgroundTexture);

		//Switch to an ortho mode.
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, viewWidth, viewHeight, 0, -1, 16);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		//glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		glDisable(GL_DEPTH_TEST);

		//Draw test thing.
		/*glBegin(GL_QUADS);
		{
			glColor3f(1.0f, 1.0f, 0.0f);
			glVertex3f(-0.5f, -0.5f, 0.0f);
			glVertex3f(-0.5f, 0.5f, 0.0f);
			glVertex3f(0.5f, 0.5f, 0.0f);
			glVertex3f(0.5f, -0.5f, 0.0f);
		}
		glEnd();*/

		/*count += 0.02f;
		if(count > 1.0f)
			count = 1.0f;*/

		if(currentScale < scale)
		{
			currentScale += 0.02f;
			if(currentScale > scale)
				currentScale = scale;
		}

		glMatrixMode(GL_TEXTURE);
		glPushMatrix();
		glTranslatef(0.5f-((1.0f/((currentScale*3.0f)+1.0f))*0.5f), 0.5f-((1.0f/((currentScale*3.0f)+1.0f))*0.5f), 0.0f);
		glScalef((1.0f/((currentScale*3.0f)+1.0f)), (1.0f/((currentScale*3.0f)+1.0f)), 1.0f);
		glMatrixMode(GL_MODELVIEW);

		//Draw background texture.
		glDisable(GL_LIGHTING);
		glBegin(GL_QUADS);
		{
			//glColor4ub(255, 255, 255, 127);
			glColor4f(1.0f, 1.0f, 1.0f, (1.0f-currentScale));
			glTexCoord2f(0.0f, 1.0f);
			glVertex3i(0, 0, -16);

			glTexCoord2f(0.0f, 0.0f);
			glVertex3i(0, viewHeight, -16);

			glTexCoord2f(1.0f, 0.0f);
			glVertex3i(viewWidth, viewHeight, -16);

			glTexCoord2f(1.0f, 1.0f);
			glVertex3i(viewWidth, 0, -16);
		}
		glEnd();
		glEnable(GL_LIGHTING);

		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);

		glMatrixMode(GL_TEXTURE);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);

		glBindTexture(GL_TEXTURE_2D, 0);

		//Switch back to perspective mode.
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		glPopMatrix();
	}
	else
		pass = false;
}

}
