//	GhostImage.h - Node which renders all its children to a texture, then
//				   overlays it on top of them, alpha-blended.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef GHOSTIMAGE_H_
#define GHOSTIMAGE_H_

#include "ParentNode.h"

#include "GLHeaders.h"

/*+
	Name: GhostImage
	Type: ParentNode
	Description: A node which renders all it's children twice: first to a
				 texture, and then to the screen, with the texture overlaid,
				 scaled and alpha blended according to the scale parameter.
				 Note that the drawing of any visual nodes before this one in
				 the X3D file heirarchy will be erased when GhostImage does its
				 drawing.
	Attribute:	scale
	AttType:	SFFloat (input)
	AttDesc:	The scale and amount of alpha blending applied to the ghost
				texture (0 == no alpha, no scaling, 1 = full alpha, 4*scaling).

	Attribute:	trigger
	AttType:	SFFloat (input)
	AttDesc:	Used to trigger a refresh of the ghost texture.  Triggering
				happens when the value goes past 0.5.  Once triggered, the scale
				parameter is interpolated from 0 to the current user-set scale
				value.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;GhostImage DEF="bob"&gt;
			&lt;Transform translation="-1 0 -8"&gt;
				&lt;Shape&gt;
					&lt;Appearance&gt;
						&lt;Material diffuseColor="1 0 0"/&gt;
					&lt;/Appearance&gt;
					&lt;Box/&gt;
				&lt;/Shape&gt;
			&lt;/Transform&gt;

			&lt;Transform translation="2 0 -10"&gt;
				&lt;Shape&gt;
					&lt;Appearance&gt;
						&lt;Material diffuseColor="0 0 1"/&gt;
					&lt;/Appearance&gt;
					&lt;Box/&gt;
				&lt;/Shape&gt;
			&lt;/Transform&gt;
		&lt;/GhostImage&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Node which renders all its children to a texture, then overlays it on top of them, alpha-blended.
/*!
	\ingroup libheilanextras
 */
class GhostImage : public ParentNode
{
  public:
	///	Constructor.
	GhostImage();
	///	Destructor.
	~GhostImage();

	///	Where the various attributes get set.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<GhostImage> constructor;
  private:
	///	Initialise the texture.
	void parentInitialise();
	///	We override this to set the texture-rendering options.
	void preChildDraw();
	///	Here we store the texture we just rendered, and render the children again.
	void postChildDraw();

	///	Implemented to do nothing.
	void audioBlockSizeChanged(int newSize) {};

	///	Some constants.
	enum
	{
		TexWidth = 512,
		TexHeight = 512
	};

	///	How much the ghost texture is scaled and faded by.
	float scale;
	///	Used to trigger a refresh of the ghost texture.
	float trigger;

	///	Our background texture.
	GLuint backgroundTexture;
	///	Our background texture pixel data.
	unsigned char backgroundPixels[TexWidth*TexHeight*3];

	///	Variable indicating which drawing pass we're on.
	/*!
		Draw 1st to texture, then to screen.
	 */
	bool pass;

	///	Used after a texture refresh has been triggered, to scale up to scale over time.
	float currentScale;
};

}

#endif
