//	KeyTrigger.h - Class which generates boolean events when it receives a key.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef KEYTRIGGER_H_
#define KEYTRIGGER_H_

#include "AbstractNode.h"

/*+
	Name: KeyTrigger
	Type: ChildNode.
	Description: This node is used to (for example) start/stop TimeSensors
				 according to keyboard inputs.  The idea is you connect this
				 node between a KeySensor and TimeSensor's enabled attribute,
				 and the TimeSensor will then only trigger events when you hold
				 down the specified key(s).  You could also use a BooleanToggle
				 node to get a toggling action.
	Attribute:	set_activeVal
	AttType:	SFBool (input)
	AttDesc:	The boolean value of the following set_key event.  This is
				generated by KeySensor's isActive event, and tells you if the
				following set_key event is a key up or a key down event.

	Attribute:	set_key
	AttType:	SFString (input)
	AttDesc:	To be connected to KeySensor's keyPress and keyRelease outputs.
				This will generate a <key>_changed event, determined by the
				key that was passed in.  Keys are represented as a single
				character string.  So, calling set_key with a value of "a"
				would generate an a_changed event, with a boolean value
				determined by the last value sent to set_activeVal.

	Attribute:	a_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "a".

	Attribute:	b_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "b".

	Attribute:	c_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "c".

	Attribute:	d_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "d".

	Attribute:	e_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "e".

	Attribute:	f_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "f".

	Attribute:	g_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "g".

	Attribute:	h_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "h".

	Attribute:	i_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "i".

	Attribute:	j_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "j".

	Attribute:	k_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "k".

	Attribute:	l_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "l".

	Attribute:	m_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "m".

	Attribute:	n_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "n".

	Attribute:	o_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "o".

	Attribute:	p_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "p".

	Attribute:	q_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "q".

	Attribute:	r_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "r".

	Attribute:	s_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "s".

	Attribute:	t_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "t".

	Attribute:	u_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "u".

	Attribute:	v_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "v".

	Attribute:	w_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "w".

	Attribute:	x_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "x".

	Attribute:	y_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "y".

	Attribute:	z_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "z".

	Attribute:	0_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "0".

	Attribute:	1_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "1".

	Attribute:	2_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "2".

	Attribute:	3_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "3".

	Attribute:	4_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "4".

	Attribute:	5_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "5".

	Attribute:	6_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "6".

	Attribute:	7_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "7".

	Attribute:	8_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "8".

	Attribute:	9_changed
	AttType:	SFBool (output)
	AttDesc:	Generated when KeyTrigger receives a set_key event with a value
				of "9".

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform DEF="boxPos" translation="0 0 0"&gt;
			&lt;Shape&gt;
				&lt;Appearance&gt;
					&lt;Material DEF="boxColour" diffuseColor="1.0 0.0 0.0 "/&gt;
				&lt;/Appearance&gt;
				&lt;Box/&gt;
			&lt;/Shape&gt;
		&lt;/Transform&gt;

		&lt;KeySensor DEF="keybo"/&gt;
		&lt;KeyTrigger DEF="keyma"/&gt;
		&lt;TimeSensor DEF="testSensor" enabled="FALSE" cycleInterval="5.0" loop="TRUE"/&gt;
		&lt;PositionInterpolator DEF="PositionInterp" key="0.0 0.5 1.0" keyValue="1 0 0 0 1 0 0 0 1"/&gt;

		&lt;ROUTE fromNode="keybo" fromField="isActive" toNode="keyma" toField="set_activeVal"/&gt;
		&lt;ROUTE fromNode="keybo" fromField="keyPress" toNode="keyma" toField="set_key"/&gt;
		&lt;ROUTE fromNode="keybo" fromField="keyRelease" toNode="keyma" toField="set_key"/&gt;
		&lt;ROUTE fromNode="keyma" fromField="s_changed" toNode="testSensor" toField="enabled"/&gt;

		&lt;ROUTE fromNode="testSensor" fromField="fraction_changed" toNode="PositionInterp" toField="set_fraction"/&gt;
		&lt;ROUTE fromNode="PositionInterp" fromField="value_changed" toNode="boxPos" toField="translation"/&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Class which generates boolean events when it receives a key.
/*!
	\ingroup libheilanextras

	The idea is to connect this up to KeySensor, so you can start/stop things
	like TimeSensors by pressing keys.
 */
class KeyTrigger : public ChildNode
{
  public:
	///	Constructor.
	KeyTrigger();
	///	Destructor.
	~KeyTrigger();

	///	Over-ridden to do nothing.
	void graphicDoStuff() {};
	///	Over-ridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Where the main work this class does happens.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<KeyTrigger> constructor;
  private:
	///	The last set_activeVal event we received.
	bool lastActive;
};

}

#endif
