//	Kodama.cpp - An audiovisual instrument using a physical model as a mapping
//			     intermediary.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Kodama.h"
#include "Camera.h"
#include "StringParser.h"
#include "AmbisonicData.h"
#include <iostream>
#include <fstream>
#include <cassert>

using namespace std;

#define Pi 3.1415926535897932384626433832795f

namespace HeilanX3D
{
	
NodeConstructor<Kodama> Kodama::constructor;

//------------------------------------------------------------------------------
Kodama::Kodama():
ChildNode(),
exagVal(0.0f),
texturePan(0.0f),
texturePan2(0.0f),
beenPlucked(false),
pluckVal(1.0f),
pluckPos(StringSize/2),
bowAmplitude(0.0f),
bowValue(-1.0f),
bowInc(0.01f),
x0(0.0f),
x1(0.0f),
audioBuffer(0),
bufferSize(0),
lastAtten(0.0f)
{
	setType("Kodama");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
Kodama::~Kodama()
{
	if(audioBuffer)
		delete [] audioBuffer;
}

//------------------------------------------------------------------------------
void Kodama::initialise()
{
	int i;

	for(i=0;i<StringSize;++i)
		string[i] = 0.0f;

	for(i=0;i<StringSize;++i)
	{
		upperDelayLine[i] = 0.0f;
		lowerDelayLine[i] = 0.0f;
	}

	for(i=0;i<StringSize;++i)
	{
		//tones[i].frequency = 110.0f * static_cast<float>(i);
		if(i%8 == 0)
			tones[i].frequency = 110.0f * static_cast<float>(i/8);
		else if(i%7 == 0)
			tones[i].frequency = 130.81f * static_cast<float>(i/8);
		else if(i%6 == 0)
			tones[i].frequency = 164.81f * static_cast<float>(i/8);
		else if(i%5 == 0)
			tones[i].frequency = 195.99f * static_cast<float>(i/8);
		else if(i%4 == 0)
			tones[i].frequency = 246.94f * static_cast<float>(i/8);
		else if(i%3 == 0)
			tones[i].frequency = 293.66f * static_cast<float>(i/8);
		else if(i%2 == 0)
			tones[i].frequency = 349.23f * static_cast<float>(i/8);
		else
			tones[i].frequency = 493.88f * static_cast<float>(i/8);
		tones[i].amplitude = 0.0f;
		tones[i].coeff = 2.0f * sinf(Pi*(tones[i].frequency/44100.0f));
		tones[i].x0 = 0.0f;
		tones[i].x1 = 1.0f;
	}

	generatePlasma1();
	generatePlasma2();

	glGenTextures(1, &texture1);

	glBindTexture(GL_TEXTURE_2D, texture1);

	glTexImage2D(GL_TEXTURE_2D,
				 0,
				 4,
				 pixelWidth,
				 pixelHeight,
				 0,
				 GL_RGBA,
				 GL_UNSIGNED_BYTE,
				 textureData1);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glGenTextures(1, &texture2);

	glBindTexture(GL_TEXTURE_2D, texture2);

	glTexImage2D(GL_TEXTURE_2D,
				 0,
				 4,
				 pixelWidth,
				 pixelHeight,
				 0,
				 GL_RGBA,
				 GL_UNSIGNED_BYTE,
				 textureData2);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

//------------------------------------------------------------------------------
void Kodama::graphicDoStuff()
{
	int i, j;
	float lastVal, lastVal2;
	float tempx, tempy;
	float tempx2, tempy2;

	glBindTexture(GL_TEXTURE_2D, texture1);

	glDisable(GL_LIGHTING);
	glBegin(GL_QUADS);
	{
		lastVal = texturePan2;
		critSec.lock();
		for(j=0;j<StringSize;++j)
		{
			lastVal = 0.0f;
			lastVal2 = 0.0f;
			tempx = sinf(0.0f) * ((float)(StringSize-j)/(float)StringSize);
			tempx += string[j] * 8.0f;
			tempy = cosf(0.0f) * ((float)(StringSize-j)/(float)StringSize);
			tempx2 = sinf(0.0f) * ((float)(StringSize-(j+1))/(float)StringSize);
			if(j < (StringSize-1))
				tempx2 += string[j+1] * 8.0f;
			tempy2 = cosf(0.0f) * ((float)(StringSize-(j+1))/(float)StringSize);
			for(i=0;i<64;++i)
			{
				glVertex3f(tempx, tempy, (float)-j);
				glTexCoord2f(texturePan+lastVal, texturePan2+lastVal2);

				glVertex3f(tempx2, tempy2, (float)(-(j+1)));
				glTexCoord2f(texturePan+0.125f+lastVal, texturePan2+lastVal2);

				tempx2 = sinf(((float)(i+1)/64.0f)*(2.0f*Pi)) * ((float)(StringSize-(j+1))/(float)StringSize);
				if(j < (StringSize-1))
					tempx2 += string[j+1] * 8.0f;
				tempy2 = cosf(((float)(i+1)/64.0f)*(2.0f*Pi)) * ((float)(StringSize-(j+1))/(float)StringSize);
				glVertex3f(tempx2, tempy2, (float)(-(j+1)));
				glTexCoord2f(texturePan+0.125f+lastVal, texturePan2+1.0f+lastVal2);

				tempx = sinf(((float)(i+1)/64.0f)*(2.0f*Pi)) * ((float)(StringSize-j)/(float)StringSize);
				tempx += string[j] * 8.0f;
				tempy = cosf(((float)(i+1)/64.0f)*(2.0f*Pi)) * ((float)(StringSize-j)/(float)StringSize);
				lastVal += 0.125f;
				glVertex3f(tempx, tempy, (float)-j);
				glTexCoord2f(texturePan+lastVal, texturePan2+1.0f+lastVal2);
			}
		}
		critSec.unlock();
	}
	glEnd();

	glBindTexture(GL_TEXTURE_2D, texture2);

	glDisable(GL_DEPTH_TEST);
	glColor4f(1.0f, 1.0f, 1.0f, exagVal);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glEnable(GL_BLEND);
	glBegin(GL_QUADS);
	{
		lastVal = texturePan2;
		critSec.lock();
		for(j=0;j<StringSize;++j)
		{
			lastVal = 0.0f;
			lastVal2 = 0.0f;
			tempx = sinf(0.0f) * ((float)(StringSize-j)/(float)StringSize) * 0.99f;
			tempx += string[j] * 8.0f;
			tempy = cosf(0.0f) * ((float)(StringSize-j)/(float)StringSize) * 0.99f;
			tempx2 = sinf(0.0f) * ((float)(StringSize-(j+1))/(float)StringSize) * 0.99f;
			if(j < (StringSize-1))
				tempx2 += string[j+1] * 8.0f;
			tempy2 = cosf(0.0f) * ((float)(StringSize-(j+1))/(float)StringSize) * 0.99f;
			for(i=0;i<64;++i)
			{
				glVertex3f(tempx, tempy, (float)-j);
				glTexCoord2f(texturePan+lastVal, texturePan2+lastVal2);

				glVertex3f(tempx2, tempy2, (float)(-(j+1)));
				glTexCoord2f(texturePan+0.125f+lastVal, texturePan2+lastVal2);

				tempx2 = sinf(((float)(i+1)/64.0f)*(2.0f*Pi)) * ((float)(StringSize-(j+1))/(float)StringSize) * 0.99f;
				if(j < (StringSize-1))
					tempx2 += string[j+1] * 8.0f;
				tempy2 = cosf(((float)(i+1)/64.0f)*(2.0f*Pi)) * ((float)(StringSize-(j+1))/(float)StringSize) * 0.99f;
				glVertex3f(tempx2, tempy2, (float)(-(j+1)));
				glTexCoord2f(texturePan+0.125f+lastVal, texturePan2+1.0f+lastVal2);

				tempx = sinf(((float)(i+1)/64.0f)*(2.0f*Pi)) * ((float)(StringSize-j)/(float)StringSize) * 0.99f;
				tempx += string[j] * 8.0f;
				tempy = cosf(((float)(i+1)/64.0f)*(2.0f*Pi)) * ((float)(StringSize-j)/(float)StringSize) * 0.99f;
				lastVal += 0.125f;
				glVertex3f(tempx, tempy, (float)-j);
				glTexCoord2f(texturePan+lastVal, texturePan2+1.0f+lastVal2);
			}
		}
		critSec.unlock();
	}
	glEnd();

	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);

	glEnable(GL_LIGHTING);

	texturePan -= 0.005f;
	if(exagVal > 0.0f)
		texturePan2 -= (0.05f * exagVal) + (0.005f * (1.0f-exagVal));
	else
		texturePan2 -= 0.005f;
}

//------------------------------------------------------------------------------
AmbisonicData *Kodama::getMonoAudio(int numSamples)
{
	int i, j;
	float tempf = 0.0f;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Update mapping string.
		if(i%2 == 0)
		{
			/*if(beenPlucked)
			{
				beenPlucked = false;

				//Pluck the string...
				//tempint = static_cast<int>(pluckPos*(float)(StringSize-2)) + 1;
				tempint = pluckPos;
				for(j=0;j<(StringSize-1);++j)
				{
					if(j<tempint)
						upperDelayLine[j] = ((float)j/(float)tempint) * pluckVal;
					else
						upperDelayLine[j] = ((float)(j-tempint)/((StringSize-1)-(float)tempint)) * pluckVal;
					lowerDelayLine[(StringSize-1)-j] = upperDelayLine[j];
				}
				exagVal = 1.0f;
			}*/
			if(bowAmplitude > 0.0f)
			{
				//upperDelayLine[pluckPos] = bowValue * bowAmplitude;
				//lowerDelayLine[(StringSize-1)-pluckPos] = upperDelayLine[pluckPos];
				if(i%4 == 0)
				{
				for(j=0;j<(StringSize-1);++j)
				{
					if(j<pluckPos)
						upperDelayLine[j] = ((float)j/(float)pluckPos) * bowValue * bowAmplitude;
					else
						upperDelayLine[j] = ((float)(StringSize-1-j)/(float)(StringSize-1-pluckPos)) * bowValue * bowAmplitude;
					lowerDelayLine[(StringSize-1)-j] = upperDelayLine[j];
				}
				}

				bowValue += bowInc;
				if(bowValue > 1.0f)
					bowValue = -1.0f;

				bowAmplitude *= 0.9999f;
				if(bowAmplitude < 1e-10)
					bowAmplitude = 0.0f;

				exagVal = bowAmplitude;
			}
			/*if(exagVal > 0.0f)
				exagVal -= 0.00005f;*/

			tempf = upperDelayLine[(StringSize-1)];
			upperDelayLine[0] = -1.0f * ((0.33f*lowerDelayLine[(StringSize-1)]) +
										 (0.33f*x0) +
										 (0.33f*x1));
			x1 = x0;
			x0 = lowerDelayLine[(StringSize-1)];
			for(j=(StringSize-1);j>0;--j)
				upperDelayLine[j] = upperDelayLine[j-1];

			lowerDelayLine[0] = -1.0f * tempf;
			for(j=(StringSize-1);j>0;--j)
				lowerDelayLine[j] = lowerDelayLine[j-1];
		}

		//Update the tones' amplitudes.
		for(j=0;j<StringSize;++j)
			tones[j].amplitude = (upperDelayLine[j] + lowerDelayLine[StringSize-1-j] + 1.0f) * 0.5f * (1.0f/24.0f);

		//Calculate monophonic audio.
		tempf = 0.0f;
		for(j=0;j<StringSize;++j)
		{
			//if(tones[j].amplitude > 0.0f)
			{
				tones[j].x0 = tones[j].x0 - (tones[j].x1 * tones[j].coeff);
				tones[j].x1 = tones[j].x1 + (tones[j].x0 * tones[j].coeff);
				tempf += tones[j].x0 * tones[j].amplitude;
			}
		}

		//Convert float sine tone to Ambisonic, put it in the buffer.
		audioBuffer[i].W = tempf * bowAmplitude;
	}

	critSec.lock();
	for(j=0;j<StringSize;++j)
	{
		string[j] = upperDelayLine[j] + lowerDelayLine[StringSize-1-j];
		//tones[j].amplitude = fabs(string[j]) * (1.0f/24.0f);
		//tones[j].amplitude = (string[j] + 1.0f) * 0.5f * (1.0f/24.0f);
	}
	critSec.unlock();

	return audioBuffer;
}

//------------------------------------------------------------------------------
void Kodama::setBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}

//------------------------------------------------------------------------------
void Kodama::setAttribute(const std::string& attribute, EventValue& val)
{
	if(attribute == "pluck")
	{
		beenPlucked = true;
		pluckVal = val;

		if(fabs(pluckVal) > bowAmplitude)
			bowAmplitude = fabs(pluckVal);
	}
	else if(attribute == "pluckPos")
		pluckPos = val;
	else if(attribute == "frequency")
	{
		float freq = val;
		for(int i=0;i<StringSize;++i)
		{
			if(i%8 == 0)
				tones[i].frequency = freq * static_cast<float>(i/8);
			else if(i%7 == 0)
				tones[i].frequency = (freq*1.189181818f) * static_cast<float>(i/8);
			else if(i%6 == 0)
				tones[i].frequency = (freq*1.498545455f) * static_cast<float>(i/8);
			else if(i%5 == 0)
				tones[i].frequency = (freq*1.781727273f) * static_cast<float>(i/8);
			else if(i%4 == 0)
				tones[i].frequency = (freq*2.244909091f) * static_cast<float>(i/8);
			else if(i%3 == 0)
				tones[i].frequency = (freq*2.669636364f) * static_cast<float>(i/8);
			else if(i%2 == 0)
				tones[i].frequency = (freq*3.174818182f) * static_cast<float>(i/8);
			else
				tones[i].frequency = (freq*4.489818182) * static_cast<float>(i/8);
			tones[i].coeff = 2.0f * sinf(Pi*(tones[i].frequency/44100.0f));
			tones[i].x0 = 0.0f;
			tones[i].x1 = 1.0f;
		}
	}
}

//------------------------------------------------------------------------------
void Kodama::generatePlasma1()
{
	int x, y;
	unsigned char tempch;
	float tempf = (float)paletteSize/2.0f;

	//Generate the colour palette.
	for(x=0;x<paletteSize;++x)
	{
		palette1[x] = (unsigned int)((float)0x000000FF * ((float)(paletteSize-x)/paletteSize));
		palette1[x] += (unsigned int)((float)0x0080FF80 * ((float)x/paletteSize));
	}

	//Generate the texture.
	for(y=0;y<pixelWidth;++y)
	{
		for(x=0;x<pixelHeight;++x)
		{
			tempch = (unsigned char)(tempf + tempf * sinf(hypot(pixelWidth-(x+tempf), pixelHeight-(y+tempf))/16.0f));
			//tempch = (unsigned char)(tempf + tempf * sinf((float)x/(37.0f+15.0f*cosf((float)y/74.0f)) ) * cosf((float)y/(31.0f+11.0f*sinf((float)x/57.0f))));
			textureData1[(y*pixelWidth)+x] = palette1[tempch];
		}
	}
}

//------------------------------------------------------------------------------
void Kodama::generatePlasma2()
{
	int x, y;
	unsigned char tempch;
	float tempf = (float)paletteSize/2.0f;

	//Generate the colour palette.
	for(x=0;x<paletteSize;++x)
	{
		palette2[x] = (unsigned int)((float)0xFFFF0000 * ((float)(paletteSize-x)/paletteSize));
		palette2[x] += (unsigned int)((float)0xFF00FF00 * ((float)x/paletteSize));
	}

	//Generate the texture.
	for(y=0;y<pixelWidth;++y)
	{
		for(x=0;x<pixelHeight;++x)
		{
			//tempch = (unsigned char)(tempf + tempf * sinf(hypotf(pixelWidth-(x+tempf), pixelHeight-(y+tempf))/16.0f));
			tempch = (unsigned char)(tempf + tempf * sinf((float)x/(37.0f+15.0f*cosf((float)y/74.0f)) ) * cosf((float)y/(31.0f+11.0f*sinf((float)x/57.0f))));
			textureData2[(y*pixelWidth)+x] = palette2[tempch];
		}
	}
}

}
