//	Kodama.h - An audiovisual instrument using a physical model as a mapping
//			   intermediary.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef KODAMA_H_
#define KODAMA_H_

#include "AbstractNode.h"
#include "CriticalSection.h"
#include "GLHeaders.h"

/*+
	Name: Kodama
	Type: ChildNode.
	Description: A simple audiovisual instrument.  Uses a physical model as a
			 	 mapping intermediary, where the performer controls the model,
				 which then controls the audio and visuals.  The visuals consist
				 of a tunnel-esque representation of the inside of the string,
				 while the audio consists of an additive synth, the partials of
				 which are controlled by the positions of the cells on the
				 string.
	Attribute:	pluck
	AttType:	SFFloat (input)
	AttDesc:	Contrary to the name, this actually controls the amplitude of a
				simplistic bowing model.

	Attribute:	pluckPos
	AttType:	SFInt32 (input)
	AttDesc:	Controls where the bow is applied to the string.  Range:
				0-&gt;127.

	Attribute:	frequency
	AttType:	SFFloat (input)
	AttDesc:	The root frequency of the additive synth.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0.0 0.0 -1.0"&gt;
			&lt;Appearance&gt;
				&lt;ImageTexture url="/home/niall/Projects/NiallsX3DBrowser/src/woodthing.jpg"/&gt;
			&lt;/Appearance&gt;
			&lt;Kodama DEF="bob"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	An audiovisual instrument using a physical model as a mapping intermediary.
/*!
	\ingroup libheilanextras
 */
class Kodama : public ChildNode
{
  public:
	///	Constructor.
	Kodama();
	///	Destructor.
	~Kodama();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the visuals.
	void graphicDoStuff();
	///	To handle the instrument's audio output.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Called to set the current audio block size.
	void setBlockSize(int newSize);

	///	Called to set the value of the named attribute.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<Kodama> constructor;
  private:
	///	Generates the 1st plasma texture.
	void generatePlasma1();
	///	Generates the 2nd plasma texture.
	void generatePlasma2();

	///	Various constants.
	enum
	{
		pixelWidth = 256,
		pixelHeight = 256,
		StringSize = 128,
		paletteSize = 256
	};

	///	Simple struct representing a single tone in the additive synthesizer.
	struct Tone
	{
		///	The frequency of the tone.
		float frequency;
		///	The amplitude of the tone.
		float amplitude;
		///	Coefficient for the sine generation.
		float coeff;
		///	Coefficient for the sine generation.
		float x0;
		///	Coefficient for the sine generation.
		float x1;
	};
	///	The tones for the additive synthesizer.
	Tone tones[StringSize];

	///	Image data.
	unsigned char image[pixelWidth*pixelHeight];
	///	Plasma texture 1.
	unsigned int textureData1[pixelWidth*pixelHeight];
	///	Plasma texture 2.
	unsigned int textureData2[pixelWidth*pixelHeight];
	///	Colour palette for textureData1.
	unsigned int palette1[paletteSize];
	///	Colour palette for textureData2.
	unsigned int palette2[paletteSize];
	///	Texture to map (1).
	GLuint texture1;
	///	Texture to map (2).
	GLuint texture2;
	///	Used to determine how much to blend the 2nd texture by.
	float exagVal;

	///	Texture panning var.
	float texturePan;
	float texturePan2;

	///	Positions of the mapping string.
	float string[StringSize];

	///	The upper delay line of the mapping string.
	float upperDelayLine[StringSize];
	///	The lower delay line of the mapping string.
	float lowerDelayLine[StringSize];
	///	Whether the string has just been plucked or not.
	bool beenPlucked;
	///	The value to pluck the string with.
	float pluckVal;
	///	The position on the string that's been plucked.
	int pluckPos;

	///	The amplitude of the bowing.
	float bowAmplitude;
	///	The current bowing value.
	float bowValue;
	///	How much bowValue is incremented each time.
	float bowInc;

	///	Filter coefficient.
	float x0;
	///	Filter coefficient.
	float x1;

	///	Critical section for the string data.
	CriticalSection critSec;

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;
	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;
};

};

#endif
