//	Koroku.cpp - An audiovisual instrument where the performer controls the
//			   	 audio and visuals which both also control each other.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Koroku.h"
#include "Camera.h"
#include "StringParser.h"
#include "AmbisonicData.h"

#include <iostream>
#include <cassert>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<Koroku> Koroku::constructor;

//------------------------------------------------------------------------------
void nurbsError(GLenum errorCode)
{
	cout << "NURBS error: " << gluErrorString(errorCode) << endl;
}

//------------------------------------------------------------------------------
Koroku::Koroku():
ChildNode(),
slider1Val(0.0f),
slider2Val(0.0f),
slider3Val(0.0f),
slider4Val(0.0f),
slider1Rate(0.0f),
controlRate(441),
slider34Diff(0.0f),
overlap(0.0f),
frontBackSymmetry(1.0f),
colourTemperature(0.0f),
energyContent(0.0f),
instantAmp(0.0f),
brightness(0.0f),
nurbsObject(0),
symmetry(1.0f),
rotSpeedSym(0.0f),
rotSpeedAsym(0.0f),
scaleSym(1.0f),
scaleAsym(1.0f),
coordInt(0.0f),
colourInt(0.0f),
rotValSym(0.0f),
rotValAsym(0.0f),
feedbackAmp(0.0f),
dampingValue(0.0f),
dampingRight(0),
bowForce(0.0f),
bowVelocity(0.1f),
bowMode(false),
bowPointPosition(0.0f),
bowForceExerted(0.0f),
dcSpeed(0.0f),
dcPos(0.0f),
#ifdef HAVE_FFTW
fftPlan(0),
fftCount(0),
#endif
audioBuffer(0),
bufferSize(0),
lastAtten(0.0f)
{
	setType("Koroku");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
Koroku::~Koroku()
{
	if(audioBuffer)
		delete [] audioBuffer;

	if(nurbsObject)
		gluDeleteNurbsRenderer(nurbsObject);

#ifdef HAVE_FFTW
	if(fftPlan)
		fftw_destroy_plan(fftPlan);
#endif
}

//------------------------------------------------------------------------------
void Koroku::initialise()
{
	int i;

	//Audio settings.
	setMaxFrontAudioRadius(50.0f);
	setMaxBackAudioRadius(50.0f);
	setMinFrontAudioRadius(25.0f);
	setMinBackAudioRadius(25.0f);

	for(i=0;i<StringLength;++i)
	{
		delayLine[i] = 0.0f;
		forces[i] = 0.0f;
		velocities[i] = 0.0f;
	}

	expCoeff = 1.0f - (1.0f/0.1f);
	logCoeff = 1.0f - (1.0f/0.9f);

	//Calculate cells' mass.
	float defaultVelocityMultiplier;
	float decay = 5.0f; //5s?
	float intendedFreq = 110.0f;
	float actualFreq;
	float compensationFactor;
	defaultVelocityMultiplier = (1.0-((.00012f*44100.0f)/44100.0f)/decay);
	actualFreq = (24000.0f*44100.0f/44100.0f)/(StringLength+1);
	compensationFactor=powf(4.0f, log10f(actualFreq/intendedFreq)/log10f(2.0f));
	mass = 3.5f * compensationFactor;
	invMass = 1.0f/mass;

	velocityMultiplier = defaultVelocityMultiplier;
	dampMultiplier = 1.0f-(powf(10000.0f, dampingValue)/10000.0f);

	for(i=0;i<RmsSize;++i)
		audioRms[i] = 0.0f;

#ifdef HAVE_FFTW
	//Set up fft.
	if(!fftPlan)
		fftPlan = fftw_plan_r2r_1d(FftSize, fftInput, fftOutput, FFTW_R2HC, FFTW_MEASURE);
#endif

	//Visuals settings.
	if(!nurbsObject)
		nurbsObject = gluNewNurbsRenderer();
	if(nurbsObject)
	{
		gluNurbsProperty(nurbsObject, GLU_SAMPLING_TOLERANCE, 50.0);
		gluNurbsProperty(nurbsObject, GLU_DISPLAY_MODE, GLU_FILL);
	}

	setupCoordsColours();

	//Set up knots.
	uKnots[0] = 0.0f;
	uKnots[1] = 0.0f;
	uKnots[2] = 0.0f;
	uKnots[3] = 1.0f;
	uKnots[4] = 2.0f;
	uKnots[5] = 2.0f;
	uKnots[6] = 2.0f;

	vKnots[0] = 0.0f;
	vKnots[1] = 0.0f;
	vKnots[2] = 0.0f;
	vKnots[3] = 1.0f;
	vKnots[4] = 2.0f;
	vKnots[5] = 2.0f;
	vKnots[6] = 2.0f;
}

//------------------------------------------------------------------------------
void Koroku::graphicDoStuff()
{
	int i, j, k;
	int tempCoordInt;
	int tempColourInt;
	float tempf[4];
	float tempColTemp;
	float big, small;

	//Update control parameters.
	if(symmetry < 0.5f)
	{
		symmetry = fakeExp(slider1Rate*8.0f) +
				   (slider2Val*0.125f) +
				   ((1.0f-dampingValue)*0.125f);
	}
	else
	{
		symmetry = fakeLog(slider1Rate*8.0f) +
				   (slider2Val*0.125f) +
				   ((1.0f-dampingValue)*0.125f);
	}

	colourInt = (fakeExp(slider3Val) * 0.25f) +
				(fakeLog(slider4Val) * 0.25f) +
				((1.0f-dampingValue) * 0.25f) +
				(brightness * 0.25f);

	coordInt = (fakeLog(energyContent) * 0.25f) +
			   (fakeExp(instantAmp) * 0.25f) +
			   (bowForce * 0.25f) +
			   (fakeLog(brightness) * 0.25f);

	rotSpeedSym = (fakeExp(instantAmp) * 0.25f) +
				  (bowForce * 0.5f) +
				  (brightness * 0.25f);
	rotSpeedAsym = rotSpeedSym * 2.0f;

	if(rotValSym > 270.0f)
		big = 360-rotValSym;
	else if(rotValSym > 180.0f)
		big = fmod(rotValSym, 90.0f);
	else if(rotValSym > 90.0f)
		big = 180-rotValSym;
	else
		big = rotValSym;

	if(rotValAsym > 270.0f)
		small = 360-rotValAsym;
	else if(rotValAsym > 180.0f)
		small = fmod(rotValAsym, 90.0f);
	else if(rotValAsym > 90.0f)
		small = 180-rotValAsym;
	else
		small = rotValAsym;
	overlap = fmod(fabs((small-90.0f)-big), 90.0f)/90.0f;

	scaleSym = fakeLog(energyContent) * 2.0f;
	scaleAsym = (1.0f-(scaleSym*0.5f)) * 2.0f;

	tempCoordInt = static_cast<int>(coordInt);
	tempColourInt = static_cast<int>(colourInt);

	//Fill points and colours.
	for(i=0,j=0,k=0;i<NumPoints;++i,j+=3,k+=4)
	{
		points[j] = interp(coordInt,
						   coordsSym[tempCoordInt][i].x,
						   coordsSym[tempCoordInt+1][i].x);
		points[j+1] = interp(coordInt,
						     coordsSym[tempCoordInt][i].y,
						     coordsSym[tempCoordInt+1][i].y);
		points[j+2] = interp(coordInt,
						     coordsSym[tempCoordInt][i].z,
						     coordsSym[tempCoordInt+1][i].z);

		colours[k] = interp(colourInt,
						    coloursSym[tempColourInt][i].x,
						    coloursSym[tempColourInt+1][i].x);
		colours[k+1] = interp(colourInt,
						      coloursSym[tempColourInt][i].y,
						      coloursSym[tempColourInt+1][i].y);
		colours[k+2] = interp(colourInt,
						      coloursSym[tempColourInt][i].z,
						      coloursSym[tempColourInt+1][i].z);
		colours[k+3] = 1.0f;
	}

	//Update frontBackSymmetry.
	for(i=0;i<4;++i)
	{
		//x
		tempf[i] = points[i*3] - points[(i+12)*3];
		//y
		tempf[i] += points[(i*3)+1] - points[((i+12)*3)+1];

		tempf[i] *= 0.5f;
	}
	//We take the average.
	frontBackSymmetry = (tempf[0]+tempf[1]+tempf[2]+tempf[3])*0.25f;

	//Update tempColTemp.
	if(tempColourInt < 4)
	{
		tempColTemp = interp(colourInt,
							 colourTempsSym[tempColourInt],
							 colourTempsSym[tempColourInt+1]);
	}
	else
	{
		tempColTemp = interp(colourInt,
							 colourTempsSym[tempColourInt],
							 colourTempsSym[0]);
	}

	//Update rotation.
	rotValSym += rotSpeedSym;
	fmod(rotValSym, 360.0f);
	rotValAsym += rotSpeedAsym;
	fmod(rotValAsym, 360.0f);

	glDisable(GL_LIGHTING);

	//Symmetrical parts.
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glScalef(scaleSym, scaleSym, scaleSym);
	glRotatef(-rotValSym, 0.0f, 0.0f, 1.0f);

	//Top right section.
	gluBeginSurface(nurbsObject);
	{
		gluNurbsSurface(nurbsObject,
						NumKnots,
						uKnots,
						NumKnots,
						vKnots,
						16,
						4,
						colours,
						3,
						3,
						GL_MAP2_COLOR_4);
		gluNurbsSurface(nurbsObject,
						NumKnots,
						uKnots,
						NumKnots,
						vKnots,
						12,
						3,
						points,
						3,
						3,
						GL_MAP2_VERTEX_3);
	}
	gluEndSurface(nurbsObject);

	//bottom left section.
	glRotatef(180.0f, 0.0f, 0.0f, 1.0f);
	gluBeginSurface(nurbsObject);
	{
		gluNurbsSurface(nurbsObject,
						NumKnots,
						uKnots,
						NumKnots,
						vKnots,
						16,
						4,
						colours,
						3,
						3,
						GL_MAP2_COLOR_4);
		gluNurbsSurface(nurbsObject,
						NumKnots,
						uKnots,
						NumKnots,
						vKnots,
						12,
						3,
						points,
						3,
						3,
						GL_MAP2_VERTEX_3);
	}
	gluEndSurface(nurbsObject);

	glPopMatrix();

	//Asymmetrical parts.
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	//Calculate the asymmetrical points differently if necessary.
	if(symmetry > 0.0f)
	{
		for(i=0,j=0,k=0;i<NumPoints;++i,j+=3,k+=4)
		{
			points[j] = interpSym(symmetry,
							   	  points[j],
							   	  interp(coordInt,
								 		 coordsAsym[tempCoordInt][i].x,
								 		 coordsAsym[tempCoordInt+1][i].x));
			points[j+1] = interpSym(symmetry,
							   	  	points[j+1],
							   	  	interp(coordInt,
								 		   coordsAsym[tempCoordInt][i].y,
								 		   coordsAsym[tempCoordInt+1][i].y));
			points[j+2] = interpSym(symmetry,
							   	  	points[j+2],
							   	  	interp(coordInt,
								 		   coordsAsym[tempCoordInt][i].z,
								 		   coordsAsym[tempCoordInt+1][i].z));
	
			colours[k] = interpSym(symmetry,
							   	   colours[k],
							   	   interp(colourInt,
								 		  coloursAsym[tempColourInt][i].x,
								 		  coloursAsym[tempColourInt+1][i].x));
			colours[k+1] = interpSym(symmetry,
							   	     colours[k+1],
							   	     interp(colourInt,
								 		    coloursAsym[tempColourInt][i].y,
								 		    coloursAsym[tempColourInt+1][i].y));
			colours[k+2] = interpSym(symmetry,
							   	     colours[k+2],
							   	     interp(colourInt,
								 		    coloursAsym[tempColourInt][i].z,
								 		    coloursAsym[tempColourInt+1][i].z));
			colours[k+3] = 1.0f;
		}
		float tempf = (scaleAsym * symmetry) + (scaleSym * (1.0f-symmetry));
		glScalef(tempf, tempf, tempf);
		tempf = (fmod(rotValAsym, 360.0f) * symmetry) + (fmod(rotValSym, 360.0f) * (1.0f-symmetry));
		glRotatef(-tempf, 0.0f, 0.0f, 1.0f);

		//Update tempColTemp again.
		if(tempColourInt < 4)
		{
			tempColTemp += interp(colourInt,
								  colourTempsSym[tempColourInt],
								  colourTempsSym[tempColourInt+1]);
		}
		else
		{
			tempColTemp += interp(colourInt,
								  colourTempsSym[tempColourInt],
								  colourTempsSym[0]);
		}
		tempColTemp *= 0.5f;
	}
	else
	{
		glScalef(scaleSym, scaleSym, scaleSym);
		glRotatef(rotValSym, 0.0f, 0.0f, 1.0f);
	}
	colourTemperature = tempColTemp;

	//bottom right section.
	glRotatef(-90.0f, 0.0f, 0.0f, 1.0f);
	gluBeginSurface(nurbsObject);
	{
		gluNurbsSurface(nurbsObject,
						NumKnots,
						uKnots,
						NumKnots,
						vKnots,
						16,
						4,
						colours,
						3,
						3,
						GL_MAP2_COLOR_4);
		gluNurbsSurface(nurbsObject,
						NumKnots,
						uKnots,
						NumKnots,
						vKnots,
						12,
						3,
						points,
						3,
						3,
						GL_MAP2_VERTEX_3);
	}
	gluEndSurface(nurbsObject);

	//top left section.
	glRotatef(180.0f, 0.0f, 0.0f, 1.0f);
	gluBeginSurface(nurbsObject);
	{
		gluNurbsSurface(nurbsObject,
						NumKnots,
						uKnots,
						NumKnots,
						vKnots,
						16,
						4,
						colours,
						3,
						3,
						GL_MAP2_COLOR_4);
		gluNurbsSurface(nurbsObject,
						NumKnots,
						uKnots,
						NumKnots,
						vKnots,
						12,
						3,
						points,
						3,
						3,
						GL_MAP2_VERTEX_3);
	}
	gluEndSurface(nurbsObject);

	glPopMatrix();

	glEnable(GL_LIGHTING);
}

//------------------------------------------------------------------------------
AmbisonicData *Koroku::getMonoAudio(int numSamples)
{
	int i, j;
	float tempf = 0.0f;
	float tempf2 = 0.0f;
	//float tempPitchf;
	//int tempPitchi;
	//float tempSaw = 0.0f;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Update control parameters.
		++controlCount;
		if(controlCount >= controlRate)
		{
			controlCount = 0;

			//Slider1Rate.
			for(j=(Slider1BufSize)-1;j>0;--j)
				slider1Prev[j] = slider1Prev[j-1];
			slider1Prev[0] = slider1Val;

			tempf2 = 0.0f;
			for(j=1;j<Slider1BufSize;++j)
			{
				tempf2 += (slider1Prev[j]-slider1Prev[j-1]) *
						  (slider1Prev[j]-slider1Prev[j-1]);
			}
			slider1Rate = sqrtf(tempf2/(float)Slider1BufSize);

			//BowForce.
			bowForce = slider1Rate +
					   ((fmod(rotValAsym, 360.0f)/360.0f)*0.25f) +
					   (colourTemperature*0.25f);

			//StringFreq.
			updateStringFreq((((fakeExp(slider2Val)*0.75) +
							   ((fmod(rotValSym, 360.0f)/360.0f)*0.125f) +
							   (colourTemperature*0.125f))
							  *165.0f)+55.0f);

			//DampingValue.
			dampingValue = (fakeExp(slider34Diff)*0.5f) + (overlap*0.5f);
			dampMultiplier = 1.0f-(powf(10000.0f, dampingValue)/10000.0f);

			//DampingRight.
			dampingRight = (int)((((1.0f-overlap)*0.75f)+
								  (frontBackSymmetry*0.25f))*(StringLength/2));

			//FeedbackAmp.
			/*feedbackAmp = (fakeExp(slider4Val)*0.75f)+
						  (fakeLog(scaleSym*0.5f)*0.25f);*/
			feedbackAmp = (fakeExp(slider4Val)*0.75f)+
						  (fakeLog(((scaleSym*0.25f)+(scaleAsym*0.25f))*0.5f)*0.25f);
		}

		//Calculate monophonic audio.
		tempf = 0.0f;

		//Calculate forces.
		{
			for(j=1;j<(StringLength-1);++j)
			{
				forces[j] = (delayLine[j-1] + delayLine[j+1]) -
							(2.0f * delayLine[j]);
			}
			forces[0] = forces[1] - forces[0];
			forces[StringLength-1] = forces[StringLength-2] -
									 forces[StringLength-1];
		}

		//Add feedback.
		if(feedbackAmp > 0.0f)
		{
			delayLine[7*(StringLength/8)] += tanh(delayLine[FeedbackPos] *
												  feedbackAmp * 3.1415f) *
												  0.005f;
		}
		//Calculate bow stuff.
		if(bowForce > 0.0f)
		{
			float relativeVelocity;

			bowPointPosition += bowVelocity;
			relativeVelocity = bowVelocity - velocities[BowPosition];

			if(!bowMode)
			{
				float stickingForce = bowPointPosition - delayLine[BowPosition];

				if(stickingForce > bowForce)
					bowMode = true;
				else
					bowForceExerted = stickingForce;
			}
			else
			{
				float slippingForce = bowForce/(1.0f+fabs(relativeVelocity));

				if(velocities[BowPosition] >= 0.0f)
				{
					bowMode = false;
					bowPointPosition = delayLine[BowPosition];
				}
				else
					bowForceExerted = slippingForce;
			}

			forces[BowPosition] += bowForceExerted;
			forces[BowPosition-1] += bowForceExerted * 0.5f;
			forces[BowPosition+1] += bowForceExerted * 0.5f;
		}

		//Calculate positions.
		for(j=1;j<(StringLength-1);++j)
		{
			velocities[j] += forces[j] * invMass;
			if(j < dampingRight)
				velocities[j] *= dampMultiplier;
			else
				velocities[j] *= velocityMultiplier;
			delayLine[j] += velocities[j];
		}

		//Calculate audio.
		tempf = delayLine[3*(StringLength/4)];

		//Apply DC filter.
		dcSpeed = dcSpeed + (tempf - dcPos) * 0.000004567f;
		dcPos = dcPos + dcSpeed;
		dcSpeed = dcSpeed * 0.96f;
		tempf = (tempf-dcPos);

		//Calculate instantaneous amplitude.
		if(fabs(tempf) > instantAmp)
			instantAmp += 0.00001f;
			//instantAmp = fabs(tempf);
		else if(instantAmp >= 0.00001f)
			instantAmp -= 0.00001f;

		//Update audioRms so we can calculate energyContent.
		for(j=(RmsSize-1);j>1;--j)
			audioRms[j] = audioRms[j-1];
		audioRms[0] = fabs(tempf) * 0.5f;

		for(j=0;j<RmsSize;++j)
			tempf2 += (audioRms[j]*audioRms[j]);
		tempf2 = sqrtf(tempf2/(float)RmsSize);
		if(tempf2 > energyContent)
			energyContent += 0.0001f;
		else if(energyContent > 0.0f)
			energyContent -= 0.0001f;

#ifdef HAVE_FFTW
		//Update fft.
		fftInput[fftCount] = (double)tempf;
		++fftCount;
		if(fftCount > (FftSize-1))
		{
			float tempSum = 0.0f;

			fftCount = 0;
			fftw_execute(fftPlan);

			//Update brightness.
			/*for(j=0;j<((FftSize/2)+1);++j)
				tempSum += (float)(fftOutput[j][0] * ((float)j/(float)FftSize));
			tempSum /= (float)(FftSize) * 2.0f;*/

			float magnitude;
			float averageFA = 0.0f;
			float averageA = 0.0f;
			for(j=0;j<FftSize;++j)
			//for(j=0;j<FftSize;++j)
			{
				magnitude = fabs((float)fftOutput[j]);
				averageFA += ((float)j * magnitude);
				averageA += magnitude;
			}
			if(averageA > 0.0f)
				tempSum = averageFA/averageA;
			else
				tempSum = 0.0f;
			tempSum /= FftSize;

			if(tempSum > brightness)
			{
				brightness = tempSum;
			}
			else if(brightness > 0.0f)
			{
				brightness -= 0.005f;
				if(brightness < 0.0f)
					brightness = 0.0f;
			}
		}
#endif

		audioBuffer[i].W = tempf * 0.5f;
	}

	return audioBuffer;
}

//------------------------------------------------------------------------------
void Koroku::setBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}

//------------------------------------------------------------------------------
void Koroku::setSamplerate(float val)
{
	controlRate = static_cast<int>(0.01 * val);
}

//------------------------------------------------------------------------------
void Koroku::setAttribute(const std::string& attribute, EventValue& val)
{
	if(attribute == "slider1")
		slider1Val = val;
	else if(attribute == "slider2")
		slider2Val = val;
	else if(attribute == "slider3")
	{
		slider3Val = val;
		slider34Diff = fabs(slider4Val-slider3Val);
	}
	else if(attribute == "slider4")
	{
		slider4Val = val;
		slider34Diff = fabs(slider4Val-slider3Val);
	}
	//Temporary variables.
	else if(attribute == "symmetry")
		symmetry = val;
	else if(attribute == "rotSpeed")
	{
		rotSpeedSym = val;
		rotSpeedAsym = rotSpeedSym * 2.0f;
	}
	else if(attribute == "scale")
	{
		scaleSym = (float)val * 2.0f;
		scaleAsym = (1.0f-(scaleSym*0.5f)) * 2.0f;
	}
	else if(attribute == "coordInt")
		coordInt = val;
	else if(attribute == "colourInt")
		colourInt = val;
	else if(attribute == "stringFreq")
	{
		float intendedFreq = val;
		float actualFreq;
		float compensationFactor;
		if(intendedFreq < 55.0f)
			intendedFreq = 55.0f;
		else if(intendedFreq > 220.0f)
			intendedFreq = 220.0f;
		actualFreq = (24000.0f*44100.0f/44100.0f)/(StringLength+1);
		compensationFactor=powf(4.0f, log10f(actualFreq/intendedFreq)/log10f(2.0f));
		mass = 3.5f * compensationFactor;
		invMass = 1.0f/mass;
	}
	else if(attribute == "feedbackAmp")
		feedbackAmp = val;
	else if(attribute == "bowForce")
		bowForce = val;
	else if(attribute == "dampingValue")
	{
		dampingValue = val;
		dampMultiplier = 1.0f-(powf(10000.0f, dampingValue)/10000.0f);
	}
	else if(attribute == "dampingRight")
	{
		int tempint = val;

		if(tempint < 0)
			tempint = 0;
		else if(tempint > (StringLength-1))
			tempint = StringLength-1;

		dampingRight = tempint;
	}
}

//------------------------------------------------------------------------------
void Koroku::setupCoordsColours()
{
	//Symmetrical Coordinates 1.
	coordsSym[0][0] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsSym[0][1] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsSym[0][2] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsSym[0][3] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsSym[0][4] = ThreeFloats(0.0f, 1.0f, -0.3f);
	coordsSym[0][5] = ThreeFloats(0.5f, 1.0f, -0.3f);
	coordsSym[0][6] = ThreeFloats(1.0f, 0.5f, -0.3f);
	coordsSym[0][7] = ThreeFloats(1.0f, 0.0f, -0.3f);
	coordsSym[0][8] = ThreeFloats(0.0f, 1.0f, 0.3f);
	coordsSym[0][9] = ThreeFloats(0.5f, 1.0f, 0.3f);
	coordsSym[0][10] = ThreeFloats(1.0f, 0.5f, 0.3f);
	coordsSym[0][11] = ThreeFloats(1.0f, 0.0f, 0.3f);
	coordsSym[0][12] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coordsSym[0][13] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coordsSym[0][14] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coordsSym[0][15] = ThreeFloats(0.0f, 0.0f, 1.0f);

	//Symmetrical Coordinates 2.
	coordsSym[1][0] = ThreeFloats(0.0f, 2.0f, -1.0f);
	coordsSym[1][1] = ThreeFloats(1.0f, 2.0f, -1.0f);
	coordsSym[1][2] = ThreeFloats(2.0f, 1.0f, -1.0f);
	coordsSym[1][3] = ThreeFloats(2.0f, 0.0f, -1.0f);
	coordsSym[1][4] = ThreeFloats(0.0f, 1.0f, -0.3f);
	coordsSym[1][5] = ThreeFloats(0.5f, 1.0f, -0.3f);
	coordsSym[1][6] = ThreeFloats(1.0f, 0.5f, -0.3f);
	coordsSym[1][7] = ThreeFloats(1.0f, 0.0f, -0.3f);
	coordsSym[1][8] = ThreeFloats(0.0f, 1.0f, 0.3f);
	coordsSym[1][9] = ThreeFloats(0.5f, 1.0f, 0.3f);
	coordsSym[1][10] = ThreeFloats(1.0f, 0.5f, 0.3f);
	coordsSym[1][11] = ThreeFloats(1.0f, 0.0f, 0.3f);
	coordsSym[1][12] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coordsSym[1][13] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coordsSym[1][14] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coordsSym[1][15] = ThreeFloats(0.0f, 0.0f, 1.0f);

	//Symmetrical Coordinates 3.
	coordsSym[2][0] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsSym[2][1] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsSym[2][2] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsSym[2][3] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsSym[2][4] = ThreeFloats(0.0f, 1.0f, -0.3f);
	coordsSym[2][5] = ThreeFloats(0.5f, 1.0f, -0.3f);
	coordsSym[2][6] = ThreeFloats(1.0f, 0.5f, -0.3f);
	coordsSym[2][7] = ThreeFloats(1.0f, 0.0f, -0.3f);
	coordsSym[2][8] = ThreeFloats(0.0f, 1.0f, 0.3f);
	coordsSym[2][9] = ThreeFloats(0.5f, 1.0f, 0.3f);
	coordsSym[2][10] = ThreeFloats(1.0f, 0.5f, 0.3f);
	coordsSym[2][11] = ThreeFloats(1.0f, 0.0f, 0.3f);
	coordsSym[2][12] = ThreeFloats(0.0f, 1.0f, 1.0f);
	coordsSym[2][13] = ThreeFloats(2.0f, 1.0f, 1.0f);
	coordsSym[2][14] = ThreeFloats(1.0f, 2.0f, 1.0f);
	coordsSym[2][15] = ThreeFloats(1.0f, 0.0f, 1.0f);

	//Symmetrical Coordinates 4.
	coordsSym[3][0] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsSym[3][1] = ThreeFloats(2.0f, 2.0f, -1.0f);
	coordsSym[3][2] = ThreeFloats(2.0f, 2.0f, -1.0f);
	coordsSym[3][3] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsSym[3][4] = ThreeFloats(0.0f, 1.0f, -0.3f);
	coordsSym[3][5] = ThreeFloats(0.5f, 1.0f, -0.3f);
	coordsSym[3][6] = ThreeFloats(1.0f, 0.5f, -0.3f);
	coordsSym[3][7] = ThreeFloats(1.0f, 0.0f, -0.3f);
	coordsSym[3][8] = ThreeFloats(0.0f, 1.0f, 0.3f);
	coordsSym[3][9] = ThreeFloats(0.5f, 1.0f, 0.3f);
	coordsSym[3][10] = ThreeFloats(1.0f, 0.5f, 0.3f);
	coordsSym[3][11] = ThreeFloats(1.0f, 0.0f, 0.3f);
	coordsSym[3][12] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coordsSym[3][13] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coordsSym[3][14] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coordsSym[3][15] = ThreeFloats(0.0f, 0.0f, 1.0f);

	//Symmetrical Coordinates 5.
	coordsSym[4][0] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsSym[4][1] = ThreeFloats(0.0f, 2.0f, -1.0f);
	coordsSym[4][2] = ThreeFloats(0.0f, 2.0f, -1.0f);
	coordsSym[4][3] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsSym[4][4] = ThreeFloats(0.0f, 1.0f, -0.3f);
	coordsSym[4][5] = ThreeFloats(0.5f, 1.0f, -0.3f);
	coordsSym[4][6] = ThreeFloats(1.0f, 0.5f, -0.3f);
	coordsSym[4][7] = ThreeFloats(1.0f, 0.0f, -0.3f);
	coordsSym[4][8] = ThreeFloats(0.0f, 1.0f, 0.3f);
	coordsSym[4][9] = ThreeFloats(0.5f, 1.0f, 0.3f);
	coordsSym[4][10] = ThreeFloats(1.0f, 0.5f, 0.3f);
	coordsSym[4][11] = ThreeFloats(1.0f, 0.0f, 0.3f);
	coordsSym[4][12] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coordsSym[4][13] = ThreeFloats(2.0f, 0.0f, 1.0f);
	coordsSym[4][14] = ThreeFloats(2.0f, 0.0f, 1.0f);
	coordsSym[4][15] = ThreeFloats(0.0f, 0.0f, 1.0f);

	//Asymmetrical Coordinates 1.
	coordsAsym[0][0] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coordsAsym[0][1] = ThreeFloats(0.0f, 0.0f, -0.5f);
	coordsAsym[0][2] = ThreeFloats(0.0f, 0.0f, -0.5f);
	coordsAsym[0][3] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coordsAsym[0][4] = ThreeFloats(0.0f, 0.5f, -0.3f);
	coordsAsym[0][5] = ThreeFloats(0.25f, 1.0f, -0.3f);
	coordsAsym[0][6] = ThreeFloats(1.0f, 0.25f, -0.3f);
	coordsAsym[0][7] = ThreeFloats(0.5f, 0.0f, -0.3f);
	coordsAsym[0][8] = ThreeFloats(0.0f, 0.5f, 0.3f);
	coordsAsym[0][9] = ThreeFloats(0.25f, 1.0f, 0.3f);
	coordsAsym[0][10] = ThreeFloats(1.0f, 0.25f, 0.3f);
	coordsAsym[0][11] = ThreeFloats(0.5f, 0.0f, 0.3f);
	coordsAsym[0][12] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coordsAsym[0][13] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coordsAsym[0][14] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coordsAsym[0][15] = ThreeFloats(0.0f, 0.0f, 0.0f);

	//Asymmetrical Coordinates 2.
	coordsAsym[1][0] = ThreeFloats(0.0625f, 0.5f, -1.0f);
	coordsAsym[1][1] = ThreeFloats(0.125f, 0.25f, -0.5f);
	coordsAsym[1][2] = ThreeFloats(0.25f, 0.125f, -0.5f);
	coordsAsym[1][3] = ThreeFloats(0.5f, 0.0625f, -1.0f);
	coordsAsym[1][4] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coordsAsym[1][5] = ThreeFloats(0.0f, 0.0f, -0.5f);
	coordsAsym[1][6] = ThreeFloats(0.0f, 0.0f, -0.5f);
	coordsAsym[1][7] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coordsAsym[1][8] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coordsAsym[1][9] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coordsAsym[1][10] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coordsAsym[1][11] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coordsAsym[1][12] = ThreeFloats(0.0625f, 0.5f, 1.0f);
	coordsAsym[1][13] = ThreeFloats(0.125f, 0.25f, 0.5f);
	coordsAsym[1][14] = ThreeFloats(0.25f, 0.125f, 0.5f);
	coordsAsym[1][15] = ThreeFloats(0.5f, 0.0625f, 1.0f);

	//Asymmetrical Coordinates 3.
	coordsAsym[2][0] = ThreeFloats(0.0f, 0.0f, -0.5f);
	coordsAsym[2][1] = ThreeFloats(0.0f, 0.0f, -0.5f);
	coordsAsym[2][2] = ThreeFloats(0.0f, 0.0f, -0.5f);
	coordsAsym[2][3] = ThreeFloats(0.0f, 0.0f, -0.5f);
	coordsAsym[2][4] = ThreeFloats(0.0f, 0.0f, -0.35f);
	coordsAsym[2][5] = ThreeFloats(0.5f, 1.0f, -0.35f);
	coordsAsym[2][6] = ThreeFloats(1.0f, 0.5f, -0.35f);
	coordsAsym[2][7] = ThreeFloats(0.0f, 0.0f, -0.35f);
	coordsAsym[2][8] = ThreeFloats(0.0f, 0.0f, 0.35f);
	coordsAsym[2][9] = ThreeFloats(0.5f, 1.0f, 0.35f);
	coordsAsym[2][10] = ThreeFloats(1.0f, 0.5f, 0.35f);
	coordsAsym[2][11] = ThreeFloats(0.0f, 0.0f, 0.35f);
	coordsAsym[2][12] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coordsAsym[2][13] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coordsAsym[2][14] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coordsAsym[2][15] = ThreeFloats(0.0f, 0.0f, 0.5f);

	//Asymmetrical Coordinates 4.
	coordsAsym[3][0] = ThreeFloats(0.0f, 1.0f, 0.0f);
	coordsAsym[3][1] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsAsym[3][2] = ThreeFloats(0.0f, 0.0f, -1.0f);
	coordsAsym[3][3] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coordsAsym[3][4] = ThreeFloats(0.0f, 1.0f, 0.0f);
	coordsAsym[3][5] = ThreeFloats(0.2f, 0.4f, -0.3f);
	coordsAsym[3][6] = ThreeFloats(0.4f, 0.2f, -0.3f);
	coordsAsym[3][7] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coordsAsym[3][8] = ThreeFloats(0.0f, 1.0f, 0.0f);
	coordsAsym[3][9] = ThreeFloats(0.2f, 0.4f, 0.3f);
	coordsAsym[3][10] = ThreeFloats(0.4f, 0.2f, -0.3f);
	coordsAsym[3][11] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coordsAsym[3][12] = ThreeFloats(0.0f, 1.0f, 0.0f);
	coordsAsym[3][13] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coordsAsym[3][14] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coordsAsym[3][15] = ThreeFloats(1.0f, 0.0f, 0.0f);

	//Asymmetrical Coordinates 5.
	coordsAsym[4][0] = ThreeFloats(0.0f, 1.5f, 0.5f);
	coordsAsym[4][1] = ThreeFloats(0.25f, 0.5f, 0.5f);
	coordsAsym[4][2] = ThreeFloats(0.5f, 0.25f, 0.5f);
	coordsAsym[4][3] = ThreeFloats(1.5f, 0.0f, 0.5f);
	coordsAsym[4][4] = ThreeFloats(0.0f, 1.0f, -0.3f);
	coordsAsym[4][5] = ThreeFloats(0.25f, 0.5f, -0.3f);
	coordsAsym[4][6] = ThreeFloats(0.5f, 0.25f, -0.3f);
	coordsAsym[4][7] = ThreeFloats(1.0f, 0.0f, -0.3f);
	coordsAsym[4][8] = ThreeFloats(0.0f, 1.0f, 0.3f);
	coordsAsym[4][9] = ThreeFloats(0.25f, 0.5f, 0.3f);
	coordsAsym[4][10] = ThreeFloats(0.5f, 0.25f, 0.3f);
	coordsAsym[4][11] = ThreeFloats(1.0f, 0.0f, 0.3f);
	coordsAsym[4][12] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coordsAsym[4][13] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coordsAsym[4][14] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coordsAsym[4][15] = ThreeFloats(0.0f, 0.0f, 0.5f);

	//Symmetrical Colours 1.
	coloursSym[0][0] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursSym[0][1] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursSym[0][2] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursSym[0][3] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursSym[0][4] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursSym[0][5] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursSym[0][6] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursSym[0][7] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursSym[0][8] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursSym[0][9] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursSym[0][10] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursSym[0][11] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursSym[0][12] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursSym[0][13] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursSym[0][14] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursSym[0][15] = ThreeFloats(0.75f, 0.75f, 1.0f);

	//Symmetrical Colours 2.
	coloursSym[1][0] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[1][1] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[1][2] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[1][3] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[1][4] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[1][5] = ThreeFloats(0.5f, 0.5f, 1.0f);
	coloursSym[1][6] = ThreeFloats(0.5f, 0.5f, 1.0f);
	coloursSym[1][7] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[1][8] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[1][9] = ThreeFloats(0.5f, 0.5f, 1.0f);
	coloursSym[1][10] = ThreeFloats(0.5f, 0.5f, 1.0f);
	coloursSym[1][11] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[1][12] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[1][13] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursSym[1][14] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursSym[1][15] = ThreeFloats(0.0f, 0.0f, 0.0f);

	//Symmetrical Colours 3.
	coloursSym[2][0] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coloursSym[2][1] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[2][2] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[2][3] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coloursSym[2][4] = ThreeFloats(1.0f, 0.5f, 0.0f);
	coloursSym[2][5] = ThreeFloats(0.75f, 0.0f, 0.0f);
	coloursSym[2][6] = ThreeFloats(0.75f, 0.0f, 0.0f);
	coloursSym[2][7] = ThreeFloats(1.0f, 0.5f, 0.0f);
	coloursSym[2][8] = ThreeFloats(1.0f, 0.5f, 0.0f);
	coloursSym[2][9] = ThreeFloats(0.75f, 0.0f, 0.0f);
	coloursSym[2][10] = ThreeFloats(0.75f, 0.0f, 0.0f);
	coloursSym[2][11] = ThreeFloats(1.0f, 0.5f, 0.0f);
	coloursSym[2][12] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coloursSym[2][13] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[2][14] = ThreeFloats(0.0f, 0.0f, 0.0f);
	coloursSym[2][15] = ThreeFloats(1.0f, 0.0f, 0.0f);

	//Symmetrical Colours 4.
	coloursSym[3][0] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coloursSym[3][1] = ThreeFloats(1.0f, 0.5f, 0.0f);
	coloursSym[3][2] = ThreeFloats(1.0f, 0.5f, 0.0f);
	coloursSym[3][3] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coloursSym[3][4] = ThreeFloats(1.0f, 0.5f, 0.0f);
	coloursSym[3][5] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursSym[3][6] = ThreeFloats(0.9f, 0.9f, 1.0f);
	coloursSym[3][7] = ThreeFloats(1.0f, 0.5f, 0.0f);
	coloursSym[3][8] = ThreeFloats(1.0f, 0.5f, 0.0f);
	coloursSym[3][9] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursSym[3][10] = ThreeFloats(0.9f, 0.9f, 1.0f);
	coloursSym[3][11] = ThreeFloats(1.0f, 0.5f, 0.0f);
	coloursSym[3][12] = ThreeFloats(0.5f, 0.5f, 1.0f);
	coloursSym[3][13] = ThreeFloats(0.5f, 0.5f, 1.0f);
	coloursSym[3][14] = ThreeFloats(0.5f, 0.5f, 1.0f);
	coloursSym[3][15] = ThreeFloats(0.5f, 0.5f, 1.0f);

	//Symmetrical Colours 5.
	coloursSym[4][0] = ThreeFloats(0.5f, 1.0f, 0.75f);
	coloursSym[4][1] = ThreeFloats(0.25f, 0.5f, 0.5f);
	coloursSym[4][2] = ThreeFloats(0.25f, 0.5f, 0.5f);
	coloursSym[4][3] = ThreeFloats(0.5f, 1.0f, 0.75f);
	coloursSym[4][4] = ThreeFloats(0.5f, 1.0f, 0.75f);
	coloursSym[4][5] = ThreeFloats(0.25f, 0.5f, 0.25f);
	coloursSym[4][6] = ThreeFloats(0.25f, 0.5f, 0.25f);
	coloursSym[4][7] = ThreeFloats(0.5f, 1.0f, 0.75f);
	coloursSym[4][8] = ThreeFloats(0.5f, 1.0f, 0.75f);
	coloursSym[4][9] = ThreeFloats(0.25f, 0.5f, 0.25f);
	coloursSym[4][10] = ThreeFloats(0.25f, 0.5f, 0.25f);
	coloursSym[4][11] = ThreeFloats(0.5f, 1.0f, 0.75f);
	coloursSym[4][12] = ThreeFloats(0.5f, 1.0f, 0.75f);
	coloursSym[4][13] = ThreeFloats(0.25f, 0.5f, 0.25f);
	coloursSym[4][14] = ThreeFloats(0.25f, 0.5f, 0.25f);
	coloursSym[4][15] = ThreeFloats(0.5f, 1.0f, 0.75f);

	//Asymmetrical Colours 1.
	coloursAsym[0][0] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursAsym[0][1] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[0][2] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[0][3] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursAsym[0][4] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursAsym[0][5] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[0][6] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[0][7] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursAsym[0][8] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursAsym[0][9] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[0][10] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[0][11] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursAsym[0][12] = ThreeFloats(0.75f, 0.75f, 1.0f);
	coloursAsym[0][13] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[0][14] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[0][15] = ThreeFloats(0.75f, 0.75f, 1.0f);

	//Asymmetrical Colours 2.
	coloursAsym[1][0] = ThreeFloats(0.5f, 0.75f, 1.0f);
	coloursAsym[1][1] = ThreeFloats(0.25f, 0.75f, 0.5f);
	coloursAsym[1][2] = ThreeFloats(0.25f, 0.75f, 0.5f);
	coloursAsym[1][3] = ThreeFloats(0.5f, 0.75f, 1.0f);
	coloursAsym[1][4] = ThreeFloats(0.5f, 0.75f, 0.75f);
	coloursAsym[1][5] = ThreeFloats(0.5f, 0.75f, 1.0f);
	coloursAsym[1][6] = ThreeFloats(0.5f, 0.75f, 1.0f);
	coloursAsym[1][7] = ThreeFloats(0.5f, 0.75f, 0.75f);
	coloursAsym[1][8] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[1][9] = ThreeFloats(0.125f, 0.125f, 0.5f);
	coloursAsym[1][10] = ThreeFloats(0.125f, 0.125f, 0.5f);
	coloursAsym[1][11] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[1][12] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[1][13] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[1][14] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[1][15] = ThreeFloats(1.0f, 1.0f, 1.0f);

	//Asymmetrical Colours 3.
	coloursAsym[2][0] = ThreeFloats(1.0f, 0.75f, 0.0f);
	coloursAsym[2][1] = ThreeFloats(1.0f, 0.25f, 0.25f);
	coloursAsym[2][2] = ThreeFloats(0.5f, 0.25f, 0.75f);
	coloursAsym[2][3] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coloursAsym[2][4] = ThreeFloats(1.0f, 0.75f, 0.0f);
	coloursAsym[2][5] = ThreeFloats(1.0f, 0.25f, 0.25f);
	coloursAsym[2][6] = ThreeFloats(0.5f, 0.25f, 0.75f);
	coloursAsym[2][7] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coloursAsym[2][8] = ThreeFloats(1.0f, 0.75f, 0.0f);
	coloursAsym[2][9] = ThreeFloats(1.0f, 0.25f, 0.25f);
	coloursAsym[2][10] = ThreeFloats(0.5f, 0.25f, 0.75f);
	coloursAsym[2][11] = ThreeFloats(0.0f, 0.0f, 0.5f);
	coloursAsym[2][12] = ThreeFloats(1.0f, 0.75f, 0.0f);
	coloursAsym[2][13] = ThreeFloats(1.0f, 0.25f, 0.25f);
	coloursAsym[2][14] = ThreeFloats(0.5f, 0.25f, 0.75f);
	coloursAsym[2][15] = ThreeFloats(0.0f, 0.0f, 0.5f);

	//Asymmetrical Colours 4.
	coloursAsym[3][0] = ThreeFloats(1.0f, 1.0f, 0.0f);
	coloursAsym[3][1] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coloursAsym[3][2] = ThreeFloats(0.75f, 0.25f, 1.0f);
	coloursAsym[3][3] = ThreeFloats(0.3f, 0.3f, 1.0f);
	coloursAsym[3][4] = ThreeFloats(0.5f, 0.5f, 0.4f);
	coloursAsym[3][5] = ThreeFloats(0.75f, 0.125f, 0.5f);
	coloursAsym[3][6] = ThreeFloats(0.875f, 0.25f, 0.5f);
	coloursAsym[3][7] = ThreeFloats(0.6f, 0.6f, 0.5f);
	coloursAsym[3][8] = ThreeFloats(0.25f, 0.25f, 0.8f);
	coloursAsym[3][9] = ThreeFloats(0.5f, 0.25f, 0.75f);
	coloursAsym[3][10] = ThreeFloats(1.0f, 0.25f, 0.25f);
	coloursAsym[3][11] = ThreeFloats(1.0f, 1.0f, 0.0f);
	coloursAsym[3][12] = ThreeFloats(0.25f, 0.25f, 0.8f);
	coloursAsym[3][13] = ThreeFloats(0.5f, 0.25f, 0.75f);
	coloursAsym[3][14] = ThreeFloats(1.0f, 0.25f, 0.25f);
	coloursAsym[3][15] = ThreeFloats(1.0f, 1.0f, 0.0f);

	//Asymmetrical Colours 5.
	coloursAsym[4][0] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coloursAsym[4][1] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coloursAsym[4][2] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[4][3] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[4][4] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[4][5] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[4][6] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coloursAsym[4][7] = ThreeFloats(1.0f, 0.0f, 0.0f);
	coloursAsym[4][8] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coloursAsym[4][9] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coloursAsym[4][10] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[4][11] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[4][12] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[4][13] = ThreeFloats(1.0f, 1.0f, 1.0f);
	coloursAsym[4][14] = ThreeFloats(0.0f, 0.0f, 1.0f);
	coloursAsym[4][15] = ThreeFloats(0.0f, 0.0f, 1.0f);

	//Symmetrical colour temperatures.
	colourTempsSym[0] = 0.0f;
	colourTempsSym[1] = 0.6f;
	colourTempsSym[2] = 0.8f;
	colourTempsSym[3] = 0.9f;
	colourTempsSym[4] = 0.4f;

	//Asymmetrical colour temperatures.
	colourTempsAsym[0] = 0.0f;
	colourTempsAsym[1] = 0.3f;
	colourTempsAsym[2] = 0.9f;
	colourTempsAsym[3] = 1.0f;
	colourTempsAsym[4] = 1.0f;
}

}
