//	MassSpringDamper.h - Simple physical model to route to other nodes.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef MASSSPRINGDAMPER_H_
#define MASSSPRINGDAMPER_H_

#include "AbstractNode.h"
#include "TwoFloats.h"

/*+
	Name: MassSpringDamper
	Type: ChildNode
	Description: This node implements a simple damped mass spring model.  When
				 it receives a set_position event, it will move towards that
				 position, oscillating around it according to its mass,
				 stiffness and damping attributes.  It constantly outputs its
				 current position as events, so you can route it to other nodes
				 to create more interesting (and slightly more familiar from
				 real life) motions.
	Attribute:	position
	AttType:	SFFloat (input)
	AttDesc:	Use this to trigger the node into motion - this is the centre
				point around which the mass will oscillate.

	Attribute:	massPosition
	AttType:	SFFloat (input)
	AttDesc:	An output-only attribute, this is the actual position of the
				mass, centred on position.

	Attribute:	stiffness
	AttType:	SFFloat (input)
	AttDesc:	The spring stiffness of the simulation, affects how the mass
				moves.

	Attribute:	mass
	AttType:	SFFloat (input)
	AttDesc:	The mass used in the simulation, affects how the mass moves.

	Attribute:	damping
	AttType:	SFFloat (input)
	AttDesc:	How much damping is applied to the mass' motion.

	Attribute:	massPosition
	AttType:	SFFloat (output)
	AttDesc:	The current position of the mass with respect to its centre.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY" headlight="FALSE"/&gt;

		&lt;Transform DEF="bob" translation="0 0 -32"&gt;
			&lt;Shape&gt;
				&lt;Box/&gt;
			&lt;/Shape&gt;
		&lt;/Transform&gt;

		&lt;TimeSensor DEF="timer" cycleInterval="10" loop="FALSE"/&gt;
		&lt;ScalarInterpolator DEF="massPositioner" key="0 0.25 0.75 1" keyValue="1 2 2 1"/&gt;
		&lt;MassSpringDamper DEF="mass" mass="1" damping="0.5"/&gt;
		&lt;PositionInterpolator DEF="bobPosition" key="0 3" keyValue="-4 0 -16		4 0 -16"/&gt;

		&lt;ROUTE fromNode="timer" fromField="fraction_changed" toNode="massPositioner" toField="set_fraction"/&gt;
		&lt;ROUTE fromNode="massPositioner" fromField="value_changed" toNode="mass" toField="set_position"/&gt;
		&lt;ROUTE fromNode="mass" fromField="massPosition" toNode="bobPosition" toField="set_fraction"/&gt;
		&lt;ROUTE fromNode="bobPosition" fromField="value_changed" toNode="bob" toField="translation"/&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Simple physical model to route to other nodes.
/*!
	\ingroup libheilanextras

	The idea is that this node can be used to create damped mass-spring-type
	motion in other nodes.

	We use the RK4 method here to calculate the motion.
 */
class MassSpringDamper : public ChildNode
{
  public:
	///	Constructor.
	MassSpringDamper();
	///	Destructor.
	~MassSpringDamper();

	///	Overridden to do nothing.
	void graphicDoStuff() {};
	///	All our calculations take place here.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Overridden to do nothing.
	void setBlockSize(int newSize) {};
	///	We need the samplerate to calculate the motion correctly.
	void setSamplerate(float val);

	///	Called to set the value of the named attribute.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<MassSpringDamper> constructor;
  private:
	///	Helper method to evaluate a single Euler step.
	inline TwoFloats evaluate(float dt, const TwoFloats& d)
	{
		TwoFloats state;
		state.x = position.x + d.x*dt;
		state.y = position.y + d.y*dt;

		TwoFloats output;
		output.x = state.y;
		output.y = -stiffnessAndMass*state.x - damping*state.y;
		return output;
	};
	///	Helper method to evaluate a single Euler step.
	inline TwoFloats evaluate()
	{
		TwoFloats output;
		output.x = position.y;
		output.y = -stiffnessAndMass*position.x - damping*position.y;
		return output;
	};
	///	Calculates the next position of the mass.
	inline void integrate(float dt)
	{
		TwoFloats a = evaluate();
		TwoFloats b = evaluate(dt*0.5f, a);
		TwoFloats c = evaluate(dt*0.5f, b);
		TwoFloats d = evaluate(dt, c);

		const float dxdt = 1.0f/6.0f * (a.x + 2.0f*(b.x + c.x) + d.x);
		const float dvdt = 1.0f/6.0f * (a.y + 2.0f*(b.y + c.y) + d.y);

		position.x = position.x + dxdt * dt;
		position.y = position.y + dvdt * dt;
	};

	///	The current samplerate.
	float samplerate;

	///	The current position and velocity of the mass.
	TwoFloats position;

	///	The mass' intended position.
	float intendedPos;
	///	The spring stiffness.
	float stiffness;
	///	The mass.
	float mass;
	///	The damping applied to the mass' motion.
	float damping;
	///	The amplitude of the mass' motion.
	float amplitude;

	///	stiffness/mass.
	float stiffnessAndMass;
	///	The current sample period.
	float samplePeriod;

	///	Counter so we know when to send the next event.
	int contCounter;
};

}

#endif
