//	Moro.cpp - An audiovisual instrument where the performer controls the
//			   audio and visuals which both also control each other.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Moro.h"
#include "Camera.h"
#include "StringParser.h"
#include "AmbisonicData.h"

#include "GLHeaders.h"
#include <cassert>
#include <iostream>

using namespace std;

namespace HeilanX3D
{
	
NodeConstructor<Moro> Moro::constructor;

//------------------------------------------------------------------------------
Moro::Moro():
ChildNode(),
slider1Val(0.0f),
slider2Val(0.0f),
zDistance(0.0f),
curvyness(0.275f),
spiralness(0.0f),
radius(0.5f),
pointDistance(0.0f),
controlDistance(0.0f),
averageDistance(0.0f),
energyContent(0.0f),
brightness(0.0f),
bowVal(-1.0f),
bowFreq(110.0f),
bowAmp(0.0f),
pitch(0.0f),
#ifdef HAVE_FFTW
fftPlan(0),
fftCount(0),
#endif
audioBuffer(0),
bufferSize(0),
lastAtten(0.0f)
{
	setType("Moro");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
Moro::~Moro()
{
	if(audioBuffer)
		delete [] audioBuffer;

#ifdef HAVE_FFTW
	if(fftPlan)
		fftw_destroy_plan(fftPlan);
#endif
}

//-----------------------------------------------------------------------------
void Moro::initialise()
{
	int i;

	setMaxFrontAudioRadius(50.0f);
	setMaxBackAudioRadius(50.0f);
	setMinFrontAudioRadius(25.0f);
	setMinBackAudioRadius(25.0f);

	//Calculate visual stuff.
	//Top right section.
	beziers[0].points[0].x = 0.0f;
	beziers[0].points[0].y = radius;
	beziers[0].points[1].x = radius;
	beziers[0].points[1].y = 0.0f;

	beziers[0].controls[0].x = curvyness;
	beziers[0].controls[0].y = 0.5f;
	beziers[0].controls[1].x = 0.5f;
	beziers[0].controls[1].y = curvyness;

	calcBezierCoefficients(beziers[0]);

	//Bottom right section.
	beziers[1].points[0].x = radius;
	beziers[1].points[0].y = 0.0f;
	beziers[1].points[1].x = 0.0f;
	beziers[1].points[1].y = -radius;

	beziers[1].controls[0].x = 0.5f;
	beziers[1].controls[0].y = -curvyness;
	beziers[1].controls[1].x = curvyness;
	beziers[1].controls[1].y = -0.5f;

	calcBezierCoefficients(beziers[1]);

	//Bottom left section.
	beziers[2].points[0].x = 0.0f;
	beziers[2].points[0].y = -radius;
	beziers[2].points[1].x = -radius;
	beziers[2].points[1].y = 0.0f;

	beziers[2].controls[0].x = -curvyness;
	beziers[2].controls[0].y = -0.5f;
	beziers[2].controls[1].x = -0.5f;
	beziers[2].controls[1].y = -curvyness;

	calcBezierCoefficients(beziers[2]);

	//Top left section.
	beziers[3].points[0].x = -radius;
	beziers[3].points[0].y = 0.0f;
	beziers[3].points[1].x = 0.0f;
	beziers[3].points[1].y = radius;

	beziers[3].controls[0].x = -0.5f;
	beziers[3].controls[0].y = curvyness;
	beziers[3].controls[1].x = -curvyness;
	beziers[3].controls[1].y = 0.5f;

	calcBezierCoefficients(beziers[3]);

	//Calculate audio stuff.
	for(i=0;i<StringLength;++i)
	{
		delayLine[i] = 0.0f;
		forces[i] = 0.0f;
		velocities[i] = 0.0f;
	}
	for(i=0;i<RmsSize;++i)
		audioRms[i] = 0.0f;

#ifdef HAVE_FFTW
	//Set up fft.
	if(!fftPlan)
		fftPlan = fftw_plan_r2r_1d(FftSize, fftInput, fftOutput, FFTW_R2HC, FFTW_MEASURE);
#endif

	sinFreq = bowFreq;
	x0 = 0.0f;
	x1 = 1.0f;
	coeff = 2.0f * sinf(3.141592653f*(sinFreq/44100.0f));
}

//-----------------------------------------------------------------------------
void Moro::graphicDoStuff()
{
	int i, j;
	float tempX, tempY, tempZ, tempT;
	float avgDist = 0.0f;
	const int numPoints = 32;
	const float invNumPoints = 1.0f/(float)numPoints;
	const float invNumPointsTwice = 1.0f/(float)(2*numPoints);

	if(0)
	{
		glDisable(GL_LIGHTING);
		glColor3f(0.0f, 0.0f, 0.0f);
		glLineWidth(2.0f);
	
		glBegin(GL_LINE_STRIP);
		{
			for(i=0;i<StringLength;++i)
			{
				glVertex3f(((float)(i-StringLength/2)*0.01f),
						   delayLine[i],
						   0.0f);
			}
		}
		glEnd();

		glEnable(GL_LIGHTING);
	}
	else
	{

	//Update parameters.
	curvyness = (slider1Val * 0.75f) + (brightness * 0.25f);
	zDistance = (slider2Val * 0.75f) + (energyContent * 0.25f);
	zDistance -= 0.01f;
	spiralness = (energyContent * 0.5f) + pitch;
	radius = (brightness * 0.5f) + (transients * 0.5f);

	//Update bezier points.
	beziers[0].points[0].x = 0.0f;
	beziers[0].points[0].y = radius;
	beziers[0].points[1].x = radius;
	beziers[0].points[1].y = 0.0f;

	beziers[1].points[0].x = radius;
	beziers[1].points[0].y = 0.0f;
	beziers[1].points[1].x = 0.0f;
	beziers[1].points[1].y = -radius;

	beziers[2].points[0].x = 0.0f;
	beziers[2].points[0].y = -radius;
	beziers[2].points[1].x = -radius;
	beziers[2].points[1].y = 0.0f;

	beziers[3].points[0].x = -radius;
	beziers[3].points[0].y = 0.0f;
	beziers[3].points[1].x = 0.0f;
	beziers[3].points[1].y = radius;

	//Update bezier control points.
	beziers[0].controls[0].x = (((curvyness * 1.25f)*2.0f)-1.25f);
	beziers[0].controls[0].y = 0.5f;
	beziers[0].controls[1].x = 0.5f + (spiralness * 1.0f);
	beziers[0].controls[1].y = (((curvyness * 1.25f)*2.0f)-1.25f);;

	beziers[1].controls[0].x = 0.5f;
	beziers[1].controls[0].y = -(((curvyness * 1.25f)*2.0f)-1.25f);;
	beziers[1].controls[1].x = (((curvyness * 1.25f)*2.0f)-1.25f);;
	beziers[1].controls[1].y = -0.5f - (spiralness * 1.0f);

	beziers[2].controls[0].x = -(((curvyness * 1.25f)*2.0f)-1.25f);;
	beziers[2].controls[0].y = -0.5f;
	beziers[2].controls[1].x = -0.5f - (spiralness * 1.0f);
	beziers[2].controls[1].y = -(((curvyness * 1.25f)*2.0f)-1.25f);;

	beziers[3].controls[0].x = -0.5f;
	beziers[3].controls[0].y = (((curvyness * 1.25f)*2.0f)-1.25f);;
	beziers[3].controls[1].x = -(((curvyness * 1.25f)*2.0f)-1.25f);;
	beziers[3].controls[1].y = 0.5f + (spiralness * 1.0f);

	calcBezierCoefficients(beziers[0]);
	calcBezierCoefficients(beziers[1]);
	calcBezierCoefficients(beziers[2]);
	calcBezierCoefficients(beziers[3]);

	//Calculate visual->audio parameters.
	pointDistance = sqrtf(((beziers[0].points[1].x-beziers[0].points[0].x)*
						   (beziers[0].points[1].x-beziers[0].points[0].x))+
						  ((beziers[0].points[1].y-beziers[0].points[0].y)*
						   (beziers[0].points[1].y-beziers[0].points[0].y)));
	controlDistance = sqrtf(((beziers[0].controls[0].x-beziers[0].points[0].x)*
						     (beziers[0].controls[0].x-beziers[0].points[0].x))+
						    ((beziers[0].controls[0].y-beziers[0].points[0].y)*
						     (beziers[0].controls[0].y-beziers[0].points[0].y)));

	//Draw the bezier curves.
	glDisable(GL_LIGHTING);
	glColor3f(0.0f, 0.0f, 0.0f);
	glLineWidth(2.0f);

	glBegin(GL_LINE_STRIP);
	{
		for(i=0;i<4;++i)
		{
			for(j=0;j<numPoints;++j)
			{
				tempT = (float)j * invNumPoints;

				tempX = beziers[i].aX * (tempT*tempT*tempT);
				tempX += beziers[i].bX * (tempT*tempT);
				tempX += beziers[i].cX * tempT;
				tempX += beziers[i].points[0].x;

				tempY = beziers[i].aY * (tempT*tempT*tempT);
				tempY += beziers[i].bY * (tempT*tempT);
				tempY += beziers[i].cY * tempT;
				tempY += beziers[i].points[0].y;

				tempZ = (float)(i*numPoints)+(float)j;
				tempZ *= invNumPointsTwice;
				tempZ -= 1.0f;
				tempZ *= zDistance * 4.0f;
				glVertex3f(tempX, tempY, tempZ);

				avgDist += sqrtf((tempX*tempX)+(tempY*tempY)+(tempZ*tempZ));
			}
		}
		averageDistance = avgDist * invNumPoints;

		//Connect back to the start.
		tempX = beziers[3].points[1].x;
		tempY = beziers[3].points[1].y;
		tempZ = 1.0f;
		tempZ *= zDistance * 4.0f;
		glVertex3f(tempX, tempY, tempZ);
	}
	glEnd();

	glEnable(GL_LIGHTING);
	}
}

//-----------------------------------------------------------------------------
AmbisonicData *Moro::getMonoAudio(int numSamples)
{
	int i, j;
	float tempf = 0.0f;
	float tempPitchf;
	int tempPitchi;
	float tempf2 = 0.0f;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		tempf = 0.0f;

		//Update parameters.
		pitch = ((slider1Val * 0.75f) + (pointDistance * 0.25f)) * 0.5f;
		bowAmp = ((1.0f-slider2Val) * 0.75f) + (controlDistance * 0.25f);
		bowFreq = (zDistance * 0.75f) + (averageDistance * 0.25f);
		bowFreq *= 412.5f;
		bowFreq += 27.5f;

		//Calculate forces.
		{
			for(j=1;j<(StringLength-1);++j)
			{
				forces[j] = (delayLine[j-1] + delayLine[j+1]) -
							(2.0f * delayLine[j]);
			}
			forces[0] = forces[1] - forces[0];
			forces[StringLength-1] = forces[StringLength-2] -
									 forces[StringLength-1];
		}

		//Update pitch stuff.
		{
			//0.1f because pitch is only 0->0.5 anyway.
			tempPitchf = pitch * (float)StringLength;
			tempPitchi = (int)tempPitchf;
		}

		//Update excitation stuff.
		{
			/*bowVal += (bowFreq * (1.0f/44100.0f));
			if(bowVal > 1.0f)
				bowVal = -1.0f;*/
			bowVal = sinf(x0 * 2.0f * 3.141592653f);
			x0 += bowFreq * (1.0f/44100.0f);
			//x0 += 44100.0f/bowFreq;
			if(x0 > 1.0f)
				x0 -= 1.0f;

			delayLine[7*(StringLength/8)] += bowVal * 0.1f * bowAmp;
		}

		//Calculate positions.
		for(j=1;j<(StringLength-1);++j)
		{
			//velocities[j] += forces[j] * 0.285714286; //0.285714286 = 1/3.5
			velocities[j] += forces[j] * 0.096812616; //0.096812616 = 1/10.329
			if(j > tempPitchi)
				velocities[j] *= 0.999976f;
			else
				velocities[j] *= 0.369042656f;
			delayLine[j] += velocities[j];
		}

		//Calculate audio.
		tempf = delayLine[3*(StringLength/4)];

		//Update audioRms.
		for(j=(RmsSize-1);j>1;--j)
			audioRms[j] = audioRms[j-1];
		audioRms[0] = fabs(tempf) * 0.5f;

		/*energyContent = sqrtf(((audioRms[3]*audioRms[3])+
							   (audioRms[2]*audioRms[2])+
							   (audioRms[1]*audioRms[1])+
							   (audioRms[0]*audioRms[0]))*0.25f);*/
		for(j=0;j<RmsSize;++j)
			tempf2 += (audioRms[j]*audioRms[j]);
		tempf2 = sqrtf(tempf2/(float)RmsSize);
		if(tempf2 > energyContent)
			energyContent = tempf2;
			//energyContent += 0.0001f;
		else
			energyContent -= 0.0001f;

		//Detect transients.
		if((fabs(fabs(audioRms[0])-fabs(audioRms[1])) > 0.5f)&&(transients < 1.0f))
			transients += 0.0001f;
		else if(transients > 0.0f)
			transients -= 0.00001f;
		else
			transients = 0.0f;

#ifdef HAVE_FFTW
		//Update fft.
		fftInput[fftCount] = (double)tempf;
		++fftCount;
		if(fftCount > (FftSize-1))
		{
			float tempSum = 0.0f;

			fftCount = 0;
			fftw_execute(fftPlan);

			//Update brightness.
			/*for(j=0;j<((FftSize/2)+1);++j)
				tempSum += (float)(fftOutput[j][0] * ((float)j/(float)FftSize));
			tempSum /= (float)(FftSize) * 2.0f;*/

			float magnitude;
			float averageFA = 0.0f;
			float averageA = 0.0f;
			for(j=0;j<FftSize;++j)
			//for(j=0;j<FftSize;++j)
			{
				magnitude = fabs((float)fftOutput[j]);
				averageFA += ((float)j * magnitude);
				averageA += magnitude;
			}
			if(averageA > 0.0f)
				tempSum = averageFA/averageA;
			else
				tempSum = 0.0f;
			tempSum /= FftSize;

			if(tempSum > brightness)
			{
				brightness = tempSum;
			}
			else if(brightness > 0.0f)
			{
				brightness -= 0.005f;
				if(brightness < 0.0f)
					brightness = 0.0f;
			}
		}
#endif

		//Convert float sine tone to Ambisonic, put it in the buffer.
		audioBuffer[i].W = tempf * 0.5f;
	}

	return audioBuffer;
}

//-----------------------------------------------------------------------------
void Moro::setBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}

//-----------------------------------------------------------------------------
void Moro::setAttribute(const string& attribute, EventValue& val)
{
	StringParser parser(val);

	if(attribute == "slider1")
		slider1Val = val;
	else if(attribute == "slider2")
		slider2Val = val;
}

}
