//	Moro.h - An audiovisual instrument where the performer controls the
//			 audio and visuals which both also control each other.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef MORO_H_
#define MORO_H_

#include "AbstractNode.h"

#ifdef HAVE_FFTW
#include <fftw3.h>
#endif

/*+
	Name: Moro
	Type: ChildNode.
	Description: A simple audiovisual instrument.  Uses simple monochrome bezier
			 	 curves for the visuals and a physically-modelled string for the
			 	 audio.  There are no rate-of-change inputs in this instrument.
	Attribute:	slider1
	AttType:	SFFloat (input)

	Attribute:	slider2
	AttType:	SFFloat (input)

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0.0 0.0 -2.0"&gt;
			&lt;Moro DEF="bob"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	An audiovisual instrument where the performer controls the audio and visuals which both also control each other.
/*!
	\ingroup libheilanextras
 */
class Moro : public ChildNode
{
  public:
	///	Constructor.
	Moro();
	///	Destructor.
	~Moro();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the visuals.
	void graphicDoStuff();
	///	To handle the instrument's audio output.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Called to set the current audio block size.
	void setBlockSize(int newSize);

	///	Called to set the value of the named attribute.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<Moro> constructor;
  private:
	///	Represents a 2d point.
	struct BezierPoint
	{
		float x;
		float y;
	};

	///	Represents the important parts of a bezier section.
	struct BezierSection
	{
		///	Start and end points.
		BezierPoint points[2];
		///	Control points.
		BezierPoint controls[2];

		///	X value coefficient.
		float cX;
		///	X value coefficient.
		float bX;
		///	X value coefficient.
		float aX;

		///	Y value coefficient.
		float cY;
		///	Y value coefficient.
		float bY;
		///	Y value coefficient.
		float aY;
	};

	///	To calculate the coefficents of a bezier section.
	inline void calcBezierCoefficients(BezierSection& section)
	{
		section.cX = 3.0f * (section.controls[0].x-section.points[0].x);
		section.bX = 3.0f * (section.controls[1].x - section.controls[0].x);
		section.bX -= section.cX;
		section.aX = section.points[1].x - section.points[0].x - section.cX;
		section.aX -= section.bX;

		section.cY = 3.0f * (section.controls[0].y-section.points[0].y);
		section.bY = 3.0f * (section.controls[1].y - section.controls[0].y);
		section.bY -= section.cY;
		section.aY = section.points[1].y - section.points[0].y - section.cY;
		section.aY -= section.bY;
	};

	///	The four sections of the bezier shape.
	BezierSection beziers[4];

	///	Slider 1 current value.
	float slider1Val;
	///	Slider 2 current value.
	float slider2Val;

	///	Current visual z-distance value.
	float zDistance;
	///	Current visual curvy-ness value.
	float curvyness;
	///	Current visual spiral-ness value.
	float spiralness;
	///	Current visual radius value.
	float radius;

	///	Current 2d distance between 1st & 2nd visual points.
	float pointDistance;
	///	Current visual distance between 1st control point and its parent.
	float controlDistance;
	///	Current average 3d distance of points to origin.
	float averageDistance;
	///	Current audio energy content.
	float energyContent;
	///	Current audio brightness.
	float brightness;
	///	Audio transient detector.
	float transients;

	///	Various constants.
	enum
	{
		StringLength = 127,
		RmsSize = 32
	};

	///	Delay line for the string (i.e. the positions of the cells).
	float delayLine[StringLength];
	///	Forces for the string cells.
	float forces[StringLength];
	///	Velocities for the string cells.
	float velocities[StringLength];

	///	Bowing value.
	float bowVal;
	///	Frequency of the bowing.
	float bowFreq;
	///	Amplitude of the bowing.
	float bowAmp;

	///	The current pitch (stop) value (as 0->0.5 of the length of the string).
	float pitch;

	///	Previous RmsSize (abs) amplitude values.
	float audioRms[RmsSize];
#ifdef HAVE_FFTW
	///	Constant containing the size of the fft.
	enum
	{
		FftSize = 1024
	};
	///	Input audio buffer.
	/*!
		Used to calculate the 'brightness' of the audio output.
	 */
	double fftInput[FftSize];
	///	FFT output.
	/*!
		Used to calculate the 'brightness' of the audio output.
	 */
	double fftOutput[FftSize];
	///	Plan for the FFT.
	fftw_plan fftPlan;

	///	Counter used to determine when to update fftOutput.
	int fftCount;
#endif

	///	test.
	float sinFreq;
	///	test.
	float x0;
	///	test.
	float x1;
	///	test.
	float coeff;

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;
	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;
};

}

#endif
