//	MotionBlur.cpp - Simple motion blur effect.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "MotionBlur.h"
#include "Camera.h"
#include "StringParser.h"

#include <iostream>
#include <cmath>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<MotionBlur> MotionBlur::constructor;

//------------------------------------------------------------------------------
MotionBlur::MotionBlur():
ParentNode(),
writePos(0),
pass(false),
texWidth(0),
texHeight(0)
{
	int i;

	setType("MotionBlur");
	setVisual(true);
	setShouldBeStart(true);

	for(i=0;i<NumTextures;++i)
		backgroundPixels[i] = 0;
}

//------------------------------------------------------------------------------
MotionBlur::~MotionBlur()
{
	int i;

	for(i=0;i<NumTextures;++i)
	{
		if(backgroundPixels[i])
			delete [] backgroundPixels[i];
	}
}

//------------------------------------------------------------------------------
void MotionBlur::setAttribute(const string& attribute, EventValue& val)
{
	
}

//------------------------------------------------------------------------------
void MotionBlur::parentInitialise()
{
	int i, j;

	setTexSize();

	//Setup background textures.
	for(i=0;i<NumTextures;++i)
	{
		if(!backgroundPixels[i])
			backgroundPixels[i] = new unsigned char[texWidth*texHeight*3];
		for(j=0;j<(texWidth*texHeight*3);++j)
			backgroundPixels[i][j] = 0;

		glGenTextures(1, &backgroundTexture[i]);
		glBindTexture(GL_TEXTURE_2D, backgroundTexture[i]);
		glTexImage2D(GL_TEXTURE_2D,
					 0,
					 3,
					 texWidth,
					 texHeight,
					 0,
					 GL_RGB,
					 GL_UNSIGNED_BYTE,
					 backgroundPixels[i]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
}

//------------------------------------------------------------------------------
void MotionBlur::preChildDraw()
{
	if(!pass)
	{
		const int viewWidth = Camera::getInstance()->getViewWidth();
		const int viewHeight = Camera::getInstance()->getViewHeight();

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();

		//Set viewport to the texture's dimensions.
		glViewport(0, 0, texWidth, texHeight);

		//Switch to an ortho mode.
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, viewWidth, viewHeight, 0, -1, 16);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		//Switch back to perspective mode.
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}
}

//------------------------------------------------------------------------------
void MotionBlur::postChildDraw()
{
	int i;
	float invNumTextures = 1.0f/(float)(NumTextures-1);

	if(!pass)
	{
		const int viewWidth = Camera::getInstance()->getViewWidth();
		const int viewHeight = Camera::getInstance()->getViewHeight();

		glBindTexture(GL_TEXTURE_2D, backgroundTexture[writePos]);
		glCopyTexImage2D(GL_TEXTURE_2D,
						 0,
						 GL_RGB8,
						 0,
						 0,
						 texWidth,
						 texHeight,
						 0);
	
		glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		//Reset the viewport.
		glViewport(0,
				   0,
				   viewWidth,
				   viewHeight);

		//Draw children again.
		glBindTexture(GL_TEXTURE_2D, 0);
		pass = true;
		graphicDoStuff();
		glBindTexture(GL_TEXTURE_2D, backgroundTexture[0]);

		//Switch to an ortho mode.
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, viewWidth, viewHeight, 0, -1, 16);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		glDisable(GL_DEPTH_TEST);

		glMatrixMode(GL_TEXTURE);
		glPushMatrix();
		glMatrixMode(GL_MODELVIEW);

		//Draw background texture.
		glDisable(GL_LIGHTING);

		for(i=0;i<NumTextures;++i)
		{
			int tempPos;

			tempPos = writePos-i;
			if(tempPos < 0)
				tempPos += NumTextures;
			glBindTexture(GL_TEXTURE_2D, backgroundTexture[tempPos]);
			glBegin(GL_QUADS);
			{
				glColor4f(1.0f, 1.0f, 1.0f, invNumTextures);
				glTexCoord2f(0.0f, 1.0f);
				glVertex3i(0, 0, -16);
	
				glTexCoord2f(0.0f, 0.0f);
				glVertex3i(0, viewHeight, -16);
	
				glTexCoord2f(1.0f, 0.0f);
				glVertex3i(viewWidth, viewHeight, -16);
	
				glTexCoord2f(1.0f, 1.0f);
				glVertex3i(viewWidth, 0, -16);
			}
			glEnd();
		}

		++writePos;
		if(writePos > (NumTextures-1))
			writePos = 0;

		glEnable(GL_LIGHTING);

		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);

		glMatrixMode(GL_TEXTURE);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);

		glBindTexture(GL_TEXTURE_2D, 0);

		//Switch back to perspective mode.
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		glPopMatrix();
	}
	else
		pass = false;
}

//------------------------------------------------------------------------------
void MotionBlur::setTexSize()
{
	float tempint;
	int actualWidth = Camera::getInstance()->getViewWidth();
	int actualHeight = Camera::getInstance()->getViewHeight();

	tempint = 1.0f;
	while(texWidth == 0)
	{
		if(powf(2.0f, tempint) > actualWidth)
		{
			texWidth = static_cast<int>(powf(2.0f, tempint-1.0f));
			break;
		}
		tempint += 1.0f;
	}

	tempint = 1.0f;
	while(texHeight == 0)
	{
		if(pow(2.0f, tempint) > actualHeight)
		{
			texHeight = static_cast<int>(powf(2.0f, tempint-1.0f));
			break;
		}
		tempint += 1.0f;
	}

	cout << "texWidth: " << texWidth << endl;
	cout << "texHeight: " << texHeight << endl;
}

}
