//	MotionBlur.h - Simple motion blur effect.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef MOTIONBLUR_H_
#define MOTIONBLUR_H_

#include "ParentNode.h"

#include "GLHeaders.h"

/*+
	Name: MotionBlur
	Type: ParentNode.
	Description: Simplistic motion blur effect, applies to all child nodes.  Not
				 very good at the moment...
	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;MotionBlur&gt;
			&lt;Background skyColor="1 1 1 0.66 0.66 1" skyAngle="0 1" groundColor="0.125 0.5 0.125 0.5 0.75 0.4 1 1 1" groundAngle="0 1.3 1.570796327"/&gt;

			&lt;Transform translation="-2 0 -10"&gt;
				&lt;Shape&gt;
					&lt;Appearance&gt;
						&lt;Material DEF="boxColour" diffuseColor="0 0 1"/&gt;
					&lt;/Appearance&gt;
					&lt;Box/&gt;
				&lt;/Shape&gt;
			&lt;/Transform&gt;

			&lt;Transform translation="2 1 -10"&gt;
				&lt;Shape&gt;
					&lt;Appearance&gt;
						&lt;Material DEF="boxColour" diffuseColor="0.1 0.8 0.1"/&gt;
					&lt;/Appearance&gt;
					&lt;Box/&gt;
				&lt;/Shape&gt;
			&lt;/Transform&gt;
		&lt;/MotionBlur&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Simple motion blur effect.
/*!
	\ingroup libheilanextras
 */
class MotionBlur : public ParentNode
{
  public:
	///	Constructor.
	MotionBlur();
	///	Destructor.
	~MotionBlur();

	///	Where the various attributes get set.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<MotionBlur> constructor;
  private:
	///	Initialises the textures.
	void parentInitialise();

	///	We override this to set the texture-rendering options.
	void preChildDraw();
	///	Here we store the texture we just rendered, and render the children again.
	void postChildDraw();

	///	Implemented to do nothing.
	void audioBlockSizeChanged(int newSize) {};

	///	Helper method to set texWidth and texHeight optimally.
	void setTexSize();

	///	Some constants.
	enum
	{
		NumTextures = 8
	};

	///	Our background texture.
	GLuint backgroundTexture[NumTextures];
	///	Our background texture pixel data.
	unsigned char *backgroundPixels[NumTextures];
	///	The current texture to write to.
	int writePos;

	///	Variable indicating which drawing pass we're on.
	/*!
		Draw 1st to texture, then to screen.
	 */
	bool pass;

	///	The width of the texture.
	int texWidth;
	///	The height of the texture.
	int texHeight;
};

}

#endif
