//	Nago.cpp - An audiovisual instrument where the performer controls the audio
//			   which controls the visuals.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Nago.h"
#include "Camera.h"
#include "StringParser.h"
#include "AmbisonicData.h"

#include "GLHeaders.h"
#include <cassert>
#include <iostream>

using namespace std;

#define Pi 3.1415926535897932384626433832795f

namespace HeilanX3D
{

NodeConstructor<Nago> Nago::constructor;

//------------------------------------------------------------------------------
Nago::Nago():
ChildNode(),
slider1Rms(0.0f),
slider1Attached(false),
slider1CountDown(0.0f),
bowVal(-1.0f),
justPlucked(false),
pitch(0.0f),
pX0(0.0f),
pX1(1.0f),
wobbleEnv(0.0f),
displayString(false),
displayBrightness(false),
envelope(0.0f),
brightness(0.0f),
#ifdef HAVE_FFTW
fftPlan(0),
fftCount(0),
#endif
audioBuffer(0),
bufferSize(0),
lastAtten(0.0f)
{
	setType("Nago");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
Nago::~Nago()
{
	if(audioBuffer)
		delete [] audioBuffer;

#ifdef HAVE_FFTW
	if(fftPlan)
		fftw_destroy_plan(fftPlan);
#endif
}

//------------------------------------------------------------------------------
void Nago::initialise()
{
	int i;

	for(i=0;i<StringLength;++i)
	{
		delayLine[i] = 0.0f;
		forces[i] = 0.0f;
		velocities[i] = 0.0f;
	}

	for(i=0;i<4;++i)
		slider1[i] = 0.0f;

	pCoeff = 2.0f * sinf(Pi*(30.0f/44100.0f));

	//Top right section.
	beziers[0].points[0][0] = 0.0f;
	beziers[0].points[0][1] = 1.0f;
	beziers[0].points[1][0] = 1.0f;
	beziers[0].points[1][1] = 0.0f;

	beziers[0].controls[0][0] = 0.5f;
	beziers[0].controls[0][1] = 1.0f;
	beziers[0].controls[1][0] = 1.0f;
	beziers[0].controls[1][1] = 0.5f;

	calcBezierCoefficients(beziers[0]);

	//Bottom right section.
	beziers[1].points[0][0] = 1.0f;
	beziers[1].points[0][1] = 0.0f;
	beziers[1].points[1][0] = 0.0f;
	beziers[1].points[1][1] = -1.0f;

	beziers[1].controls[0][0] = 1.0f;
	beziers[1].controls[0][1] = -0.5f;
	beziers[1].controls[1][0] = 0.5f;
	beziers[1].controls[1][1] = -1.0f;

	calcBezierCoefficients(beziers[1]);

	//Bottom left section.
	beziers[2].points[0][0] = 0.0f;
	beziers[2].points[0][1] = -1.0f;
	beziers[2].points[1][0] = -1.0f;
	beziers[2].points[1][1] = 0.0f;

	beziers[2].controls[0][0] = -0.5f;
	beziers[2].controls[0][1] = -1.0f;
	beziers[2].controls[1][0] = -1.0f;
	beziers[2].controls[1][1] = -0.5f;

	calcBezierCoefficients(beziers[2]);

	//Top left section.
	beziers[3].points[0][0] = -1.0f;
	beziers[3].points[0][1] = 0.0f;
	beziers[3].points[1][0] = 0.0f;
	beziers[3].points[1][1] = 1.0f;

	beziers[3].controls[0][0] = -1.0f;
	beziers[3].controls[0][1] = 0.5f;
	beziers[3].controls[1][0] = -0.5f;
	beziers[3].controls[1][1] = 1.0f;

	calcBezierCoefficients(beziers[3]);

#ifdef HAVE_FFTW
	//Set up fft.
	if(!fftPlan)
		fftPlan = fftw_plan_r2r_1d(FftSize, fftInput, fftOutput, FFTW_R2HC, FFTW_MEASURE);
#endif

	setMaxFrontAudioRadius(50.0f);
	setMaxBackAudioRadius(50.0f);
	setMinFrontAudioRadius(25.0f);
	setMinBackAudioRadius(25.0f);
}

//------------------------------------------------------------------------------
void Nago::graphicDoStuff()
{
	int i, j;
	float tempX, tempY, tempT;
	float tempSin;
	//float tempBright = brightness * 16.0f;
	float tempBright = brightness;
	float tempPitch = pitch * 2.0f;
	const int numPoints = 16;
	const float invNumPoints = 1.0f/(float)numPoints;

	if(displayString)
	{
		glDisable(GL_LIGHTING);
		glColor3f(0.0f, 0.0f, 0.0f);
		glLineWidth(2.0f);
	
		glBegin(GL_LINE_STRIP);
		{
			for(i=0;i<StringLength;++i)
			{
				glVertex3f(((float)(i-StringLength/2)*0.01f),
						   delayLine[i],
						   0.0f);
			}
		}
		glEnd();

		if(displayBrightness)
		{
			if(tempBright > 1.0f)
				tempBright = 1.0f;
			glBegin(GL_QUADS);
			{
				glVertex3f(-tempBright, tempBright, 0.0f);
				glVertex3f(tempBright, tempBright, 0.0f);
				glVertex3f(tempBright, -tempBright, 0.0f);
				glVertex3f(-tempBright, -tempBright, 0.0f);
			}
			glEnd();
		}

		glEnable(GL_LIGHTING);
	}
	else
	{
		glDisable(GL_LIGHTING);
		glColor3f(0.0f, 0.0f, 0.0f);
		glLineWidth(2.0f);

		//Recalculate coefficients.
		tempSin = pX0 * wobbleEnv;
		tempSin *= 0.05f;

		beziers[0].points[0][0] = 0.0f;
		beziers[0].points[0][1] = 1.0f + (tempBright * 2.0f);
		beziers[0].points[1][0] = 1.0f;
		beziers[0].points[1][1] = 0.0f;
		beziers[0].controls[0][0] = 0.5f - tempSin - tempPitch;
		beziers[0].controls[0][1] = 1.0f - (tempBright * 4.0f) - tempSin;
		beziers[0].controls[1][0] = 1.0f - tempSin - tempPitch;
		beziers[0].controls[1][1] = 0.5f - (tempBright * 4.0f) - tempSin;
		calcBezierCoefficients(beziers[0]);

		beziers[1].points[0][0] = 1.0f;
		beziers[1].points[0][1] = 0.0f;
		beziers[1].points[1][0] = 0.0f;
		beziers[1].points[1][1] = -1.0f - (tempBright * 2.0f);
		beziers[1].controls[0][0] = 1.0f - tempSin - tempPitch;
		beziers[1].controls[0][1] = -0.5f + (tempBright * 4.0f) + tempSin;
		beziers[1].controls[1][0] = 0.5f - tempSin - tempPitch;
		beziers[1].controls[1][1] = -1.0f + (tempBright * 4.0f) + tempSin;
		calcBezierCoefficients(beziers[1]);

		beziers[2].points[0][0] = 0.0f;
		beziers[2].points[0][1] = -1.0f - (tempBright * 2.0f);
		beziers[2].points[1][0] = -1.0f;
		beziers[2].points[1][1] = 0.0f;
		beziers[2].controls[0][0] = -0.5f + tempSin + tempPitch;
		beziers[2].controls[0][1] = -1.0f + (tempBright * 4.0f) + tempSin;
		beziers[2].controls[1][0] = -1.0f + tempSin + tempPitch;
		beziers[2].controls[1][1] = -0.5f + (tempBright * 4.0f) + tempSin;
		calcBezierCoefficients(beziers[2]);

		beziers[3].points[0][0] = -1.0f;
		beziers[3].points[0][1] = 0.0f;
		beziers[3].points[1][0] = 0.0f;
		beziers[3].points[1][1] = 1.0f + (tempBright * 2.0f);
		beziers[3].controls[0][0] = -1.0f + tempSin + tempPitch;
		beziers[3].controls[0][1] = 0.5f - (tempBright * 4.0f) - tempSin;
		beziers[3].controls[1][0] = -0.5f + tempSin + tempPitch;
		beziers[3].controls[1][1] = 1.0f - (tempBright * 4.0f) - tempSin;
		calcBezierCoefficients(beziers[3]);

		glBegin(GL_LINE_STRIP);
		{
			for(i=0;i<4;++i)
			{
				for(j=0;j<numPoints;++j)
				{
					tempT = (float)j * invNumPoints;

					tempX = beziers[i].aX * (tempT*tempT*tempT);
					tempX += beziers[i].bX * (tempT*tempT);
					tempX += beziers[i].cX * tempT;
					tempX += beziers[i].points[0][0];
					tempX *= (envelope+0.25f);

					tempY = beziers[i].aY * (tempT*tempT*tempT);
					tempY += beziers[i].bY * (tempT*tempT);
					tempY += beziers[i].cY * tempT;
					tempY += beziers[i].points[0][1];
					tempY *= (envelope+0.25f);

					glVertex3f(tempX, tempY, 0.0f);
				}
			}

			//Connect back to the start.
			tempX = beziers[0].points[0][0];
			tempX *= (envelope+0.25f);
			tempY = beziers[0].points[0][1];
			tempY *= (envelope+0.25f);
			glVertex3f(tempX, tempY, 0.0f);
		}
		glEnd();

		/*glBegin(GL_QUADS);
		{
			glVertex3f(-tempBright, tempBright, 0.0f);
			glVertex3f(tempBright, tempBright, 0.0f);
			glVertex3f(tempBright, -tempBright, 0.0f);
			glVertex3f(-tempBright, -tempBright, 0.0f);
		}
		glEnd();*/

		glEnable(GL_LIGHTING);
	}
}

//------------------------------------------------------------------------------
AmbisonicData *Nago::getMonoAudio(int numSamples)
{
	int i, j;
	float tempf = 0.0f;
	float tempPitchf;
	int tempPitchi;
	int tempint;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		tempf = 0.0f;

		//Calculate forces.
		for(j=1;j<(StringLength-1);++j)
		{
			forces[j] = (delayLine[j-1] + delayLine[j+1]) -
						(2.0f * delayLine[j]);
		}
		forces[0] = forces[1] - forces[0];
		forces[StringLength-1] = forces[StringLength-2] -
								 forces[StringLength-1];

		//Update pitch stuff.
		if(wobbleEnv > 0.0f)
		{
			pX0 = pX0 - (pX1 * pCoeff);
			pX1 = pX1 + (pX0 * pCoeff);

			wobbleEnv -= 0.0001f;

			//0.1f because pitch is only 0->0.5 anyway.
			tempPitchf = fabs(pitch + (pX0 * wobbleEnv * 0.1f)) *
							 (float)StringLength;
			tempPitchi = (int)tempPitchf;

			//Allow for fractional pitches?
			tempf = tempPitchf - (float)tempPitchi;
			//delayLine[tempPitchi] *= tempf;
			/*delayLine[tempPitchi] = 0.0f;
			delayLine[tempPitchi+1] *= 1.0f - tempf;*/
		}
		else
		{
			tempPitchf = fabs(pitch) *(float)StringLength;
			tempPitchi = (int)tempPitchf;

			//Allow for fractional pitches?
			tempf = tempPitchf - (float)tempPitchi;
			//delayLine[tempPitchi] *= tempf;
			/*delayLine[tempPitchi] = 0.0f;
			delayLine[tempPitchi+1] *= 1.0f - tempf;*/
		}

		//Update slider1 if necessary.
		if(slider1CountDown > 0.0f)
		{
			tempf = 0.0f;
			slider1CountDown -= 0.000022676f; //=1/44100
			if(slider1CountDown <= 1.0f)
			{
				slider1[3] = slider1[2];
				slider1[2] = slider1[1];
				slider1[1] = slider1[0];
				slider1[0] *= slider1CountDown;

				for(j=0;j<3;++j)
					tempf += (slider1[j+1]-slider1[j]) * (slider1[j+1]-slider1[j]);
				slider1Rms = sqrtf(tempf * 0.333333f);
			}
		}

		//Update excitation stuff.
		if(slider1Attached)
		{
			//If slider 1's attached, damp the string a bit.
			/*delayLine[7*(StringLength/8)] *= 0.75f;
			delayLine[7*(StringLength/8)] += (slider1Rms * 0.25f);*/
			//bowVal += 0.01f * (2.0f*slider1Rms);
			//bowVal += 0.01f + ((slider1Rms-0.5f) * 0.001f);
			bowVal += (0.005f + (0.01f * pitch)) + ((slider1Rms-0.5f) * 0.001f);
			if(bowVal > 1.0f)
				bowVal = -1.0f;

			delayLine[7*(StringLength/8)] = bowVal * slider1Rms * 2.0f;
		}
		if(justPlucked)
		{
			tempint = (7*(StringLength/8));
			for(j=tempPitchi;j<(StringLength-1);++j)
			{
				if(j < tempint)
					delayLine[j] = (float)(j-tempPitchi)/(float)tempint;
				else
					delayLine[j] = (float)(StringLength-j-tempPitchi)/((float)(StringLength-tempint));

				if(delayLine[j] > 1.0f)
					delayLine[j] = 1.0f;
			}
			justPlucked = false;
		}

		//Calculate positions.
		for(j=1;j<(StringLength-1);++j)
		{
			//velocities[j] += forces[j] * 0.285714286; //0.285714286 = 1/3.5
			velocities[j] += forces[j] * 0.096812616; //0.096812616 = 1/10.329
			if(j > tempPitchi)
				velocities[j] *= 0.999976f;
			else
				velocities[j] *= 0.369042656f;
			delayLine[j] += velocities[j];
		}

		//Calculate audio.
		tempf = delayLine[3*(StringLength/4)];

#ifdef HAVE_FFTW
		//Update fft.
		fftInput[fftCount] = (double)tempf;
		++fftCount;
		if(fftCount > (FftSize-1))
		{
			float tempSum = 0.0f;

			fftCount = 0;
			fftw_execute(fftPlan);

			//Update brightness.
			/*for(j=0;j<((FftSize/2)+1);++j)
				tempSum += (float)(fftOutput[j][0] * ((float)j/(float)FftSize));
			tempSum /= (float)(FftSize) * 2.0f;*/

			float magnitude;
			float averageFA = 0.0f;
			float averageA = 0.0f;
			for(j=0;j<FftSize;++j)
			//for(j=0;j<FftSize;++j)
			{
				magnitude = fabs((float)fftOutput[j]);
				averageFA += ((float)j * magnitude);
				averageA += magnitude;
			}
			if(averageA > 0.0f)
				tempSum = averageFA/averageA;
			else
				tempSum = 0.0f;
			tempSum /= FftSize;
			tempSum *= slider1Rms;

			if(tempSum > brightness)
			{
				brightness = tempSum;
			}
			else if(brightness > 0.0f)
			{
				brightness -= 0.005f;
				if(brightness < 0.0f)
					brightness = 0.0f;
			}
		}
#endif

		//Update envelope.
		if((fabs(tempf)*0.75f) > envelope)
			envelope = fabs(tempf) * 0.75f;
		else if(envelope > 0.0f)
		{
			envelope -= 0.00001f;
			if(envelope < 0.0f)
				envelope = 0.0f;
		}

		//Convert float sine tone to Ambisonic, put it in the buffer.
		audioBuffer[i].W = tempf;
	}

	return audioBuffer;
}

//------------------------------------------------------------------------------
void Nago::setBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}

//------------------------------------------------------------------------------
void Nago::setAttribute(const string& attribute, EventValue& val)
{
	int i;
	float tempf = 0.0f;

	if(attribute == "slider1")
	{
		slider1[3] = slider1[2];
		slider1[2] = slider1[1];
		slider1[1] = slider1[0];
		slider1[0] = val;
		slider1CountDown = 2.0f;

		/*for(i=0;i<4;++i)
			tempf += slider1[i]*slider1[i];
		slider1Rms = sqrtf(tempf*0.25f);*/
		for(i=0;i<3;++i)
			tempf += (slider1[i+1]-slider1[i]) * (slider1[i+1]-slider1[i]);
		slider1Rms = sqrtf(tempf * 0.333333f);

		if(fabs(slider1[0]-slider1[1]) > 0.3f)
		{
			//Need to let go of the string.
			slider1Attached = false;

			//Make sure we hear it get plucked.
			justPlucked = true;

			//Add in pitch wobble.
			wobbleEnv = 1.0f;
		}
		else if(fabs(slider1[0]-slider1[1]) < 0.1f)
		{
			//Need to be attached to the string.
			//Will this work okay?
			slider1Attached = true;
		}
	}
	else if(attribute == "slider2")
	{
		tempf = val;
		if(tempf > 0.5f)
			tempf = 0.5f;
		else if(tempf < 0.0f)
			tempf = 0.0f;

		pitch = tempf;
	}
	else if(attribute == "displayString")
	{
		if((float)val > 0.5f)
			displayString = true;
		else
			displayString = false;
	}
	else if(attribute == "displayBrightness")
	{
		if((float)val > 0.5f)
			displayBrightness = true;
		else
			displayBrightness = false;
	}
}

}
