//	Nago.h - An audiovisual instrument where the performer controls the audio
//			 which controls the visuals.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef NAGO_H_
#define NAGO_H_

#include "AbstractNode.h"
#ifdef HAVE_FFTW
#include <fftw3.h>
#endif

/*+
	Name: Nago
	Type: ChildNode.
	Description: A simple audiovisual instrument.  Uses simple monochrome bezier
			 	 curves for the visuals and a physically-modelled string for the
			 	 audio.
	Attribute:	slider1
	AttType:	SFFloat (input)

	Attribute:	slider2
	AttType:	SFFloat (input)

	Attribute:	displayString
	AttType:	SFFloat (input)
	AttDesc:	When greater than 0.5, displays the physically-modelled string.

	Attribute:	displayBrightness
	AttType:	SFFloat (input)
	AttDesc:	When greate than 0.5, displays the brightness of the audio as a
				square.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0.0 0.0 -6.0"&gt;
			&lt;Nago DEF="bob"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	An audiovisual instrument where the performer controls the audio which controls the visuals.
/*!
	\ingroup libheilanextras
 */
class Nago : public ChildNode
{
  public:
	///	Constructor.
	Nago();
	///	Destructor.
	~Nago();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the visuals.
	void graphicDoStuff();
	///	To handle the instrument's audio output.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Called to set the current audio block size.
	void setBlockSize(int newSize);

	///	Called to set the value of the named attribute.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<Nago> constructor;
  private:
	///	Represents the important parts of a bezier section.
	struct BezierSection
	{
		///	Start and end points.
		float points[2][2];
		///	Control points.
		float controls[2][2];

		///	X value coefficient.
		float cX;
		///	X value coefficient.
		float bX;
		///	X value coefficient.
		float aX;

		///	Y value coefficient.
		float cY;
		///	Y value coefficient.
		float bY;
		///	Y value coefficient.
		float aY;
	};

	///	To calculate the coefficents of a bezier section.
	inline void calcBezierCoefficients(BezierSection& section)
	{
		section.cX = 3.0f * (section.controls[0][0]-section.points[0][0]);
		section.bX = 3.0f * (section.controls[1][0] - section.controls[0][0]);
		section.bX -= section.cX;
		section.aX = section.points[1][0] - section.points[0][0] - section.cX;
		section.aX -= section.bX;

		section.cY = 3.0f * (section.controls[0][1]-section.points[0][1]);
		section.bY = 3.0f * (section.controls[1][1] - section.controls[0][1]);
		section.bY -= section.cY;
		section.aY = section.points[1][1] - section.points[0][1] - section.cY;
		section.aY -= section.bY;
	};

	///	Various constants.
	enum
	{
		StringLength = 127
	};

	///	Delay line for the string (i.e. the positions of the cells).
	float delayLine[StringLength];
	///	Forces for the string cells.
	float forces[StringLength];
	///	Velocities for the string cells.
	float velocities[StringLength];

	///	Stores the last 4 values of slider 1, to calculate the rms of them.
	float slider1[4];
	///	Stores the rms of slider1.
	float slider1Rms;
	///	Whether slider 1 is currently attached to (damping) the string.
	bool slider1Attached;
	///	Used to make sure slider1 decays when the user's not moving it.
	float slider1CountDown;
	///	Value set from setAttribute().
	float newSlider1Value;

	///	Bowing value.
	float bowVal;
	///	Used to make sure we hear the string plucked.
	bool justPlucked;

	///	The current pitch value (as 0->0.5 of the length of the string).
	float pitch;
	///	Used to calculate pitch wobble.
	float pX0;
	///	Used to calculate pitch wobble.
	float pX1;
	///	Used to calculate pitch wobble.
	float pCoeff;
	///	The envelope of the pitch wobble.
	float wobbleEnv;

	///	Whether to display the string, or the bezier shape.
	bool displayString;
	///	Whether to display the current brightness of the audio as a square.
	bool displayBrightness;
	///	The four sections of the bezier shape.
	BezierSection beziers[4];
	///	Audio envelope, used to set the size of the bezier shape.
	float envelope;

	///	The 'brightness' of the audio output.
	float brightness;
#ifdef HAVE_FFTW
	///	Constant containing the size of the fft.
	enum
	{
		FftSize = 1024
	};
	///	Input audio buffer.
	/*!
		Used to calculate the 'brightness' of the audio output.
	 */
	double fftInput[FftSize];
	///	FFT output.
	/*!
		Used to calculate the 'brightness' of the audio output.
	 */
	double fftOutput[FftSize];
	///	Plan for the FFT.
	fftw_plan fftPlan;

	///	Counter used to determine when to update fftOutput.
	int fftCount;
#endif

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;
	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;
};
	
};

#endif
