//	OSCOutput.h - Any messages sent to this node get transmitted as OSC
//				  messages.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef OSCOUTPUT_H_
#define OSCOUTPUT_H_

#include "AbstractNode.h"

/*+
	Name: OSCOutput
	Type: ChildNode.
	Description: Any messages sent to this node get transmitted as OSC messages.
	Attribute:	oscAddress
	AttType:	SFString (input)
	AttDesc:	The OSC address to send any OSC messages to.

	Attribute:	ipAddress
	AttType:	SFString (input)
	AttDesc:	The IP address to send OSC messages to.  If blank (the default)
				messages will be sent to the address specified on the command
				line.

	Attribute:	port
	AttType:	SFInt32 (input)
	AttDesc:	The port to send OSC messages to.  If 0 (the default) messages
				will be sent to the port specified on the command line.

	Attribute:	sffloatOutput
	AttType:	SFFloat (input)
	AttDesc:	Transmits SFFloat values as OSC messages.

	Attribute:	sfintOutput
	AttType:	SFInt32 (input)
	AttDesc:	Transmits SFInt32 values as OSC messages.

	Example:

				(OSC output to another instance of Heilan)
&lt;?xml version='1.0' encoding='utf-8'?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile='Interactive'&gt;
   &lt;Scene&gt;

        &lt;Transform translation='0 0 0'&gt;
            &lt;Shape&gt;
                &lt;Appearance&gt;
                    &lt;Material DEF='matt'/&gt;
                    &lt;ImageTexture url='TestTexture.png'/&gt;
                &lt;/Appearance&gt;
                &lt;Box/&gt;
            &lt;/Shape&gt;
        &lt;/Transform&gt;

        &lt;TimeSensor DEF='timmy' cycleInterval='5' loop='true'/&gt;
        &lt;ScalarInterpolator DEF='scaly' key='0 0.5 1' keyValue='1 0 1'/&gt;

        &lt;OSCOutput DEF='oo' oscAddress='/heilan/matt/transparency'/&gt;

        &lt;ROUTE fromNode='timmy' fromField='fraction_changed' toNode='scaly' toField='set_fraction'/&gt;
        &lt;ROUTE fromNode='scaly' fromField='value_changed' toNode='matt' toField='transparency'/&gt;
        &lt;ROUTE fromNode='scaly' fromField='value_changed' toNode='oo' toField='sffloatOutput'/&gt;

    &lt;/Scene&gt;
&lt;/X3D&gt;

				(OSC input from the previous file)
&lt;?xml version='1.0' encoding='utf-8'?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile='Interactive'&gt;
   &lt;Scene&gt;

        &lt;Transform translation='0 0 0'&gt;
            &lt;Shape&gt;
                &lt;Appearance&gt;
                    &lt;Material DEF='matt'/&gt;
                    &lt;ImageTexture url='TestTexture.png'/&gt;
                &lt;/Appearance&gt;
                &lt;Box/&gt;
            &lt;/Shape&gt;
        &lt;/Transform&gt;

    &lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Any messages sent to this node get transmitted as OSC messages.
/*!
	\ingroup libheilanextras
 */
class OSCOutput : public ChildNode
{
  public:
	///	Constructor.
	OSCOutput();
	///	Destructor.
	~OSCOutput();

	///	Not used, overridden to do nothing.
	void graphicDoStuff() {};
	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to send OSC messages.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<OSCOutput> constructor;
  private:
	///	The OSC address to send messages to.
	std::string oscAddress;
	///	The IP address to send messages to if we're not using the default.
	std::string ipAddress;
	///	The port  to send messages to if we're not using the default.
	int port;
};

}

#endif
