//	Okkoto.cpp - An audiovisual instrument where the performer controls the
//			   visuals which controls the audio.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Okkoto.h"
#include "Camera.h"
#include "StringParser.h"
#include "AmbisonicData.h"

#include "GLHeaders.h"
#include <cassert>
#include <iostream>

using namespace std;

namespace HeilanX3D
{
	
NodeConstructor<Okkoto> Okkoto::constructor;

//------------------------------------------------------------------------------
Okkoto::Okkoto():
ChildNode(),
slider1Val(0.0f),
slider2Val(0.0f),
slider1Rms(0.0f),
slider2Rms(0.0f),
rmsDiff(0.0f),
bowVal(-1.0f),
bowFreq(30.0f),
bowAmp(0.0f),
justPlucked(0.0f),
lastSlider2(0.0f),
pitch(0.0f),
audioBuffer(0),
bufferSize(0),
lastAtten(0.0f)
{
	setType("Okkoto");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
Okkoto::~Okkoto()
{
	if(audioBuffer)
		delete [] audioBuffer;
}

//------------------------------------------------------------------------------
void Okkoto::initialise()
{
	int i;

	//Calculate visual stuff.
	//Top right section.
	beziers[0].points[0].x = 1.0f;
	beziers[0].points[0].y = -0.5f;
	beziers[0].points[1].x = 0.5f;
	beziers[0].points[1].y = -1.0f;

	beziers[0].controls[0].x = 0.75f;
	beziers[0].controls[0].y = -0.5f;
	beziers[0].controls[1].x = 0.75f;
	beziers[0].controls[1].y = -0.5f;

	calcBezierCoefficients(beziers[0]);

	//Bottom right section.
	beziers[1].points[0].x = 0.5f;
	beziers[1].points[0].y = -0.5f;
	beziers[1].points[1].x = 0.0f;
	beziers[1].points[1].y = -0.5f;

	beziers[1].controls[0].x = 0.25f;
	beziers[1].controls[0].y = -0.5f;
	beziers[1].controls[1].x = 0.25f;
	beziers[1].controls[1].y = -0.5f;

	calcBezierCoefficients(beziers[1]);

	//Bottom left section.
	beziers[2].points[0].x = 0.0f;
	beziers[2].points[0].y = -0.5f;
	beziers[2].points[1].x = -0.5f;
	beziers[2].points[1].y = -0.5f;

	beziers[2].controls[0].x = -0.25f;
	beziers[2].controls[0].y = -0.5f;
	beziers[2].controls[1].x = -0.25f;
	beziers[2].controls[1].y = -0.5f;

	calcBezierCoefficients(beziers[2]);

	//Top left section.
	beziers[3].points[0].x = -0.5f;
	beziers[3].points[0].y = -0.5f;
	beziers[3].points[1].x = -1.0f;
	beziers[3].points[1].y = -0.5f;

	beziers[3].controls[0].x = -0.75f;
	beziers[3].controls[0].y = -0.5f;
	beziers[3].controls[1].x = -0.75f;
	beziers[3].controls[1].y = -0.5f;

	calcBezierCoefficients(beziers[3]);

	//Set flat and circle positions.
	flatPos[0].x = -1.0f;
	flatPos[0].y = -0.5f;
	flatPos[1].x = -0.5f;
	flatPos[1].y = -0.5f;
	flatPos[2].x = 0.5f;
	flatPos[2].y = -0.5f;
	flatPos[3].x = 1.0f;
	flatPos[3].y = -0.5f;

	circlePos[0].x = -0.5f;
	circlePos[0].y = 0.0f;
	circlePos[1].x = 0.0f;
	circlePos[1].y = 0.5f;
	circlePos[2].x = 0.5f;
	circlePos[2].y = 0.0f;

	for(i=0;i<4;++i)
	{
		slider1Old[i] = 0.0f;
		slider2Old[i] = 0.0f;
	}

	//Calculate audio stuff.
	for(i=0;i<StringLength;++i)
	{
		delayLine[i] = 0.0f;
		forces[i] = 0.0f;
		velocities[i] = 0.0f;
	}

	setMaxFrontAudioRadius(50.0f);
	setMaxBackAudioRadius(50.0f);
	setMinFrontAudioRadius(25.0f);
	setMinBackAudioRadius(25.0f);
}

//------------------------------------------------------------------------------
void Okkoto::graphicDoStuff()
{
	int i, j;
	float tempf;
	float tempX, tempY, tempT;
	const int numPoints = 16;
	const float invNumPoints = 1.0f/(float)numPoints;

	//Update RMS values.
	slider1Old[3] = slider1Old[2];
	slider1Old[2] = slider1Old[1];
	slider1Old[1] = slider1Old[0];
	slider1Old[0] = slider1Val;
	tempf = 0.0f;
	for(i=0;i<3;++i)
		tempf += (slider1Old[i+1]-slider1Old[i]) * (slider1Old[i+1]-slider1Old[i]);
	tempf = sqrtf(tempf * 0.333333f);
	if(tempf > slider1Rms)
		slider1Rms = tempf;
	else if(slider1Rms > 0.0f)
	{
		slider1Rms -= 0.005f;
		if(slider1Rms < 0.0f)
			slider1Rms = 0.0f;
	}

	slider2Old[3] = slider2Old[2];
	slider2Old[2] = slider2Old[1];
	slider2Old[1] = slider2Old[0];
	slider2Old[0] = slider2Val;
	tempf = 0.0f;
	for(i=0;i<3;++i)
		tempf += (slider2Old[i+1]-slider2Old[i]) * (slider2Old[i+1]-slider2Old[i]);
	tempf = sqrtf(tempf * 0.333333f);
	if(tempf > slider2Rms)
		slider2Rms = tempf;
	else if(slider2Rms > 0.0f)
	{
		slider2Rms -= 0.005f;
		if(slider2Rms < 0.0f)
			slider2Rms = 0.0f;
	}

	rmsDiff = slider2Rms - slider1Rms;

	//Update bezier points.
	tempf = slider1Rms + 0.2f * slider2Rms;
	tempf *= 4.5f;
	if(tempf > 1.0f)
		tempf = 1.0f;

	beziers[3].points[1].x = (1.0f - tempf) * flatPos[0].x;
	beziers[3].points[1].x += tempf * circlePos[1].x;
	beziers[3].points[1].y = (1.0f - tempf) * flatPos[0].y;
	beziers[3].points[1].y += tempf * circlePos[1].y;

	beziers[3].points[0].x = (1.0f - tempf) * flatPos[1].x;
	beziers[3].points[0].x += tempf * circlePos[0].x;
	beziers[3].points[0].y = (1.0f - tempf) * flatPos[1].y;
	beziers[3].points[0].y += tempf * circlePos[0].y;

	beziers[2].points[1].x = beziers[3].points[0].x;
	beziers[2].points[1].y = beziers[3].points[0].y;

	beziers[1].points[0].x = (1.0f - tempf) * flatPos[2].x;
	beziers[1].points[0].x += tempf * circlePos[2].x;
	beziers[1].points[0].y = (1.0f - tempf) * flatPos[2].y;
	beziers[1].points[0].y += tempf * circlePos[2].y;

	beziers[0].points[1].x = beziers[1].points[0].x;
	beziers[0].points[1].y = beziers[1].points[0].y;

	beziers[0].points[0].x = (1.0f - tempf) * flatPos[3].x;
	beziers[0].points[0].x += tempf * circlePos[1].x;
	beziers[0].points[0].y = (1.0f - tempf) * flatPos[3].y;
	beziers[0].points[0].y += tempf * circlePos[1].y;

	for(i=0;i<4;++i)
	{
		calcBezierControlPoints(beziers[i], (rmsDiff*4.0f), slider2Val);
		calcBezierCoefficients(beziers[i]);
	}

	//Calculate pitch from distance between point1,1 and point4,2.
	tempX = beziers[3].points[1].x - beziers[0].points[0].x;
	tempY = beziers[3].points[1].y - beziers[0].points[0].y;
	pitch = sqrtf((tempX * tempX) + (tempY * tempY)) * 0.25f;

	//Calculate bowFreq and inverse bowAmp from rmsDiff.
	bowFreq = rmsDiff * 0.5f;
	//bowAmp = (1.0f - rmsDiff) * 0.5f;
	bowAmp = rmsDiff * 0.25f;

	//Draw the bezier curves.
	glDisable(GL_LIGHTING);
	glColor3f(0.0f, 0.0f, 0.0f);
	glLineWidth(2.0f);

	glBegin(GL_LINE_STRIP);
	{
		for(i=0;i<4;++i)
		{
			for(j=0;j<numPoints;++j)
			{
				tempT = (float)j * invNumPoints;

				tempX = beziers[i].aX * (tempT*tempT*tempT);
				tempX += beziers[i].bX * (tempT*tempT);
				tempX += beziers[i].cX * tempT;
				tempX += beziers[i].points[0].x;

				tempY = beziers[i].aY * (tempT*tempT*tempT);
				tempY += beziers[i].bY * (tempT*tempT);
				tempY += beziers[i].cY * tempT;
				tempY += beziers[i].points[0].y;

				glVertex3f(tempX, tempY, 0.0f);
			}
		}

		//Connect back to the start.
		tempX = beziers[3].points[1].x;
		tempY = beziers[3].points[1].y;
		glVertex3f(tempX, tempY, 0.0f);
	}
	glEnd();

	glEnable(GL_LIGHTING);
}

//------------------------------------------------------------------------------
AmbisonicData *Okkoto::getMonoAudio(int numSamples)
{
	int i, j;
	float tempf = 0.0f;
	float tempPitchf;
	int tempPitchi;
	int tempint;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		tempf = 0.0f;

		//Calculate forces.
		{
			for(j=1;j<(StringLength-1);++j)
			{
				forces[j] = (delayLine[j-1] + delayLine[j+1]) -
							(2.0f * delayLine[j]);
			}
			forces[0] = forces[1] - forces[0];
			forces[StringLength-1] = forces[StringLength-2] -
									 forces[StringLength-1];
		}

		//Update pitch stuff.
		{
			//0.1f because pitch is only 0->0.5 anyway.
			tempPitchf = (pitch + (justPlucked*0.1f)) * (float)StringLength;
			tempPitchi = (int)tempPitchf;
		}

		//Update excitation stuff.
		{
			//bowVal += (0.005f + (0.01f * (bowFreq + (slider2Val * 0.5f))));
			bowVal += (0.005f + (0.01f * (bowFreq + (slider1Val * 2.0f))));
			if(bowVal > 1.0f)
				bowVal = -1.0f;

			delayLine[7*(StringLength/8)] += bowVal * bowAmp * 2.0f;

			if(fabs(slider2Val - lastSlider2) > 0.2f)
				justPlucked = 1.0f;
			lastSlider2 = slider2Val;

			if(justPlucked > 0.0f)
			{
				tempint = (7*(StringLength/8));
				for(j=tempPitchi;j<(StringLength-1);++j)
				{
					tempf = delayLine[j];
					if(j < tempint)
						tempf += (float)(j-tempPitchi)/(float)tempint;
					else
						tempf += (float)(StringLength-j-tempPitchi)/((float)(StringLength-tempint));

					tempf *= (1.0f-justPlucked);
					delayLine[j] = tempf;

					/*if(delayLine[j] > 1.0f)
						delayLine[j] = 1.0f;*/
				}

				justPlucked -= 0.01f;
			}
		}

		//Calculate positions.
		for(j=1;j<(StringLength-1);++j)
		{
			//velocities[j] += forces[j] * 0.285714286; //0.285714286 = 1/3.5
			velocities[j] += forces[j] * 0.096812616; //0.096812616 = 1/10.329
			if(j > tempPitchi)
				velocities[j] *= 0.999976f;
			else
				velocities[j] *= 0.369042656f;
			delayLine[j] += velocities[j];
		}

		//Calculate audio.
		tempf = delayLine[3*(StringLength/4)];

		//Convert float sine tone to Ambisonic, put it in the buffer.
		audioBuffer[i].W = tempf * 0.5f;
	}

	return audioBuffer;
}

//------------------------------------------------------------------------------
void Okkoto::setBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}

//------------------------------------------------------------------------------
void Okkoto::setAttribute(const string& attribute, EventValue& val)
{
	StringParser parser(val);

	if(attribute == "slider1")
		slider1Val = val;
	else if(attribute == "slider2")
		slider2Val = val;
}

}
