//	Okkoto.h - An audiovisual instrument where the performer controls the
//			   visuals which controls the audio.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef OKKOTO_H_
#define OKKOTO_H_

#include "AbstractNode.h"
#include <cmath>

/*+
	Name: Okkoto
	Type: ChildNode.
	Description: A simple audiovisual instrument.  Uses simple monochrome bezier
			 	 curves for the visuals and a physically-modelled string for the
			 	 audio.
	Attribute:	slider1
	AttType:	SFFloat (input)

	Attribute:	slider2
	AttType:	SFFloat (input)

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0.0 0.0 -2.0"&gt;
			&lt;Okkoto DEF="bob"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	An audiovisual instrument where the performer controls the visuals which controls the audio.
/*!
	\ingroup libheilanextras
 */
class Okkoto : public ChildNode
{
  public:
	///	Constructor.
	Okkoto();
	///	Destructor.
	~Okkoto();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the visuals.
	void graphicDoStuff();
	///	To handle the instrument's audio output.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Called to set the current audio block size.
	void setBlockSize(int newSize);

	///	Called to set the value of the named attribute.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<Okkoto> constructor;
  private:
	///	Represents a 2d point.
	struct BezierPoint
	{
		float x;
		float y;
	};

	///	Represents the important parts of a bezier section.
	struct BezierSection
	{
		///	Start and end points.
		BezierPoint points[2];
		///	Control points.
		BezierPoint controls[2];

		///	X value coefficient.
		float cX;
		///	X value coefficient.
		float bX;
		///	X value coefficient.
		float aX;

		///	Y value coefficient.
		float cY;
		///	Y value coefficient.
		float bY;
		///	Y value coefficient.
		float aY;
	};

	///	To calculate the coefficents of a bezier section.
	inline void calcBezierCoefficients(BezierSection& section)
	{
		section.cX = 3.0f * (section.controls[0].x-section.points[0].x);
		section.bX = 3.0f * (section.controls[1].x - section.controls[0].x);
		section.bX -= section.cX;
		section.aX = section.points[1].x - section.points[0].x - section.cX;
		section.aX -= section.bX;

		section.cY = 3.0f * (section.controls[0].y-section.points[0].y);
		section.bY = 3.0f * (section.controls[1].y - section.controls[0].y);
		section.bY -= section.cY;
		section.aY = section.points[1].y - section.points[0].y - section.cY;
		section.aY -= section.bY;
	};

	///	To calculate the position of a bezier section's control points.
	inline void calcBezierControlPoints(BezierSection& section,
										   float amount,
										   float pointy)
	{
		float x, y;

		if(section.points[1].x > section.points[0].x)
		{
			x = section.points[1].x-section.points[0].x;
			section.controls[0].x = section.points[0].x + (0.5f * x);
		}
		else
		{
			x = section.points[0].x-section.points[1].x;
			section.controls[0].x = section.points[1].x + (0.5f * x);
		}
		section.controls[1].x = section.controls[0].x;
		section.controls[0].x += amount * section.points[0].x * pointy;
		section.controls[1].x += amount * section.points[1].x * pointy;

		if(section.points[1].y > section.points[0].y)
		{
			y = section.points[1].y-section.points[0].y;
			section.controls[0].y = section.points[0].y + (0.5f * y);
		}
		else
		{
			y = section.points[0].y-section.points[1].y;
			section.controls[0].y = section.points[1].y + (0.5f * y);
		}
		section.controls[1].y = section.controls[0].y;
		section.controls[0].y += amount * section.points[0].y;
		section.controls[1].y += amount * section.points[1].y;
	};

	///	The four sections of the bezier shape.
	BezierSection beziers[4];
	///	'Flat' positions.
	BezierPoint flatPos[4];
	///	'Circle' positions.
	BezierPoint circlePos[3];

	///	Slider 1 current value.
	float slider1Val;
	///	Slider 2 current value.
	float slider2Val;
	///	Slider 1 RMS value.
	float slider1Rms;
	///	Slider 2 RMS value.
	float slider2Rms;
	///	Difference between slider1Rms & slider2Rms.
	float rmsDiff;

	///	Values used to calculate slider1Rms.
	float slider1Old[4];
	///	Values used to calculate slider2Rms.
	float slider2Old[4];

	///	Various constants.
	enum
	{
		StringLength = 127
	};

	///	Delay line for the string (i.e. the positions of the cells).
	float delayLine[StringLength];
	///	Forces for the string cells.
	float forces[StringLength];
	///	Velocities for the string cells.
	float velocities[StringLength];

	///	Bowing value.
	float bowVal;
	///	Frequency of the bowing.
	float bowFreq;
	///	Amplitude of the bowing.
	float bowAmp;
	///	Used to make sure we hear the string plucked.
	float justPlucked;
	///	The last slider2Val, used to detect transients.
	float lastSlider2;

	///	The current pitch (stop) value (as 0->0.5 of the length of the string).
	float pitch;

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;
	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;
};

}

#endif
