//	OscPlayback.h - Node which plays back a file of saved OSC data.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef OSCPLAYBACK_H_
#define OSCPLAYBACK_H_

#include <map>

#include "AbstractNode.h"

/*+
	Name: OscPlayback
	Type: ChildNode.
	Description: Used to play back a file of saved OSC data.  When one of these
				 is in a scene, it will play back the OSC events contained
				 within the saved file, and send the events to the browser's
				 OSC handler to pass on to the appropriate nodes.
	Attribute:	url
	AttType:	SFString (input)
	AttDesc:	Url of the file with the saved OSC data.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;OscPlayback url="test59_oscplayback.hosc"/&gt;

		&lt;GhostImage DEF="bob"&gt;
			&lt;Transform translation="-1 0 -8"&gt;
				&lt;Shape&gt;
					&lt;Appearance&gt;
						&lt;Material diffuseColor="1 0 0"/&gt;
					&lt;/Appearance&gt;
					&lt;Box/&gt;
				&lt;/Shape&gt;
			&lt;/Transform&gt;

			&lt;Transform translation="2 0 -10"&gt;
				&lt;Shape&gt;
					&lt;Appearance&gt;
						&lt;Material diffuseColor="0 0 1"/&gt;
					&lt;/Appearance&gt;
					&lt;Box/&gt;
				&lt;/Shape&gt;
			&lt;/Transform&gt;
		&lt;/GhostImage&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Node which plays back a file of saved OSC data.
/*!
	\ingroup libheilanextras

	When it reaches the sample timestamp in the file, this node then broadcasts
	the message at that timestamp using OpenSoundController::handleMessage().

	\todo Add loop parameter.
 */
class OscPlayback : public ChildNode
{
  public:
	///	Constructor.
	OscPlayback();
	///	Destructor.
	~OscPlayback();

	///	Overridden to do nothing.
	void graphicDoStuff() {};
	///	We use this to get the current audio sample to test against our timestamps.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Where we load the file.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<OscPlayback> constructor;
  private:
	///	Used to store a message.
	struct Message
	{
		///	This is the full address (/heilan/\<address\>/\<parameter\>).
		std::string address;
		///	The value to be sent.
		std::string value;
	};

	///	All the messages we have to output.
	std::multimap<unsigned int, Message> messages;
	///	Our current audio sample value.
	unsigned int samples;
};

}

#endif
