//	PaintingBackground.cpp - Background which is updated in realtime with
//						     splodges sent via ROUTEs.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "PaintingBackground.h"
#include "Camera.h"
#include "StringParser.h"
#include "AttributeHelper.h"

#include "SDL.h"
#include <cmath>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<PaintingBackground> PaintingBackground::constructor;

//------------------------------------------------------------------------------
PaintingBackground::PaintingBackground():
ChildNode(),
fDStretch(0.0f),
bDStretch(0.0f),
splodgeScale(0.025f),
//splodgeScale(1.0f),
mainTexture(0),
colour(0.0f, 0.0f, 1.0f)
{
	setType("PaintingBackground");
	setVisual(true);
	setShouldBeStart(true);
}

//------------------------------------------------------------------------------
PaintingBackground::~PaintingBackground()
{
	if(mainTexture)
		delete [] mainTexture;
}

//------------------------------------------------------------------------------
void PaintingBackground::initialise()
{
	int i, x, y;
	float tempf;
	int tempint;
	unsigned char *pixelPointer = splodge;

	if(!mainTexture)
		mainTexture = new unsigned char[MainTextureWidth*MainTextureWidth*3];

	for(i=0;i<(MainTextureWidth*MainTextureHeight*3);++i)
		mainTexture[i] = 0;

	//Generate splodge.
	for(y=0;y<SplodgeSize;++y)
	{
		for(x=0;x<SplodgeSize;++x)
		{
			tempf = sinf(((float)x/(float)SplodgeSize)*3.1415926535897932384626433832795f);
			tempf *= sinf(((float)y/(float)SplodgeSize)*3.1415926535897932384626433832795f);
			tempf = (exp(tempf)-1.0f)/(exp(1.0f)-1.0f)*1.5f-0.1f;
			if(tempf < 0.0f)
				tempf = 0.0f;
			else if(tempf > 1.0f)
				tempf = 1.0f;
			*pixelPointer = 255;
			*(pixelPointer+1) = (unsigned char)(tempf * 255.0f);
			pixelPointer += 2;

			tempint = (x*3)+(y*SplodgeSize*3);
			tempTex[tempint] = 0;
			tempTex[tempint+1] = 0;
			tempTex[tempint+2] = 0;
		}
	}

	//Setup main texture.
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexImage2D(GL_TEXTURE_2D,
				 0,
				 3,
				 MainTextureWidth,
				 MainTextureWidth,
				 0,
				 GL_RGB,
				 GL_UNSIGNED_BYTE,
				 mainTexture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);
}

//------------------------------------------------------------------------------
void PaintingBackground::graphicDoStuff()
{
	int width = Camera::getInstance()->getViewWidth();
	int height = Camera::getInstance()->getViewHeight();
	float texWidth, texHeight;
	TwoFloats temp2f;

	glBindTexture(GL_TEXTURE_2D, texture);

	critSec.lock();
	{
		//Update mainTexture with any new splodges in splodgePos.
		while(!splodgePos.empty())
		{
			if(!splodgeCols.empty())
			{
				colour = splodgeCols.front();
				splodgeCols.pop();
			}
			addSplodge(getPixelCoords(splodgePos.front(), width, height),
					   width,
					   height);
			splodgePos.pop();
		}
	}
	critSec.unlock();

	//Work out the texture coordinates.
	if(width < MainTextureWidth)
		texWidth = (float)width/(float)MainTextureWidth;
	else
		texWidth = 1.0f;
	if(height < MainTextureWidth)
		texHeight = (float)height/(float)MainTextureWidth;
	else
		texHeight = 1.0f;

	glDisable(GL_LIGHTING);
	glDepthMask(GL_FALSE);
	glDisable(GL_DEPTH_TEST);

	//Setup ortho projection.
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glOrtho(0, width, height, 0, -1, 150);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glDisable(GL_BLEND);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	//Draw background.
	glBegin(GL_QUADS);
	{
		glTexCoord2f(0.0f, 0.0f);
		glVertex2f(-(fDStretch*(float)width),
				   -(fDStretch*(float)height));
		glTexCoord2f(texWidth, 0.0f);
		glVertex2f(((float)width+(bDStretch*(float)width)),
				   -(bDStretch*(float)height));
		glTexCoord2f(texWidth, texHeight);
		glVertex2f(((float)width+(fDStretch*(float)width)),
				   ((float)height+(fDStretch*(float)height)));
		glTexCoord2f(0.0f, texHeight);
		glVertex2f(-(bDStretch*(float)width),
				   ((float)height+(bDStretch*(float)height)));
	}
	glEnd();
	glBindTexture(GL_TEXTURE_2D, 0);

	//Draw gradient on top of background.
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glBegin(GL_QUADS);
	{
		glColor4f(0.0f, 0.0f, 0.0f, 0.0f);
		glVertex2d(0, 0);
		glVertex2d(width, 0);
		glColor4f(0.0f, 0.0f, 0.0f, 1.0f);
		glVertex2d(width, height);
		glVertex2d(0, height);
	}
	glEnd();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glEnable(GL_LIGHTING);
	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
}

//------------------------------------------------------------------------------
void PaintingBackground::setAttribute(const string& attribute, EventValue& val)
{
	int i;
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "colour")
	{
		StringParser parse(val);
		AttributeHelper<ThreeFloats> helper(&colour);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parse.getNumThreeFloats() > 0)
			colour = parse.getThreeFloats();
	}
	else if(attribute == "splodgePos")
	{
		StringParser parse(val);

		critSec.lock();
		{
			for(i=0;i<parse.getNumThreeFloats();++i)
				splodgePos.push(parse.getThreeFloats(i));
		}
		critSec.unlock();
	}
	else if(attribute == "splodgePosCol")
	{
		StringParser parse(val);
		bool thing = false;

		if((parse.getNumThreeFloats()%2) == 0)
		{
			critSec.lock();
			{
				for(i=0;i<parse.getNumThreeFloats();++i)
				{
					if(!thing)
						splodgePos.push(parse.getThreeFloats(i));
					else
						splodgeCols.push(parse.getThreeFloats(i));
					thing = !thing;
				}
			}
			critSec.unlock();
		}
		else
			cout << "PaintingBackground splodgePosCol: Need an even number of values." << endl;
	}
	else if(attribute == "fDStretch")
		fDStretch = val;
	else if(attribute == "bDStretch")
		bDStretch = val;
}

}
