//	PaintingBackground.h - Background which is updated in realtime with
//						   splodges sent via ROUTEs.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PAINTINGBACKGROUND_H_
#define PAINTINGBACKGROUND_H_

#include "AbstractNode.h"
#include "TwoFloats.h"
#include "CriticalSection.h"
#include "MiscHelper.h"

#include "SDL.h"
#include "GLHeaders.h"
#include <queue>
#include <iostream>

/*+
	Name: PaintingBackground
	Type: ChildNode.
	Description: Originally to be used with the Ashitaka instrument.  Whenever
				 it receives a splodgePos, or splodgePosCol event, it draws a
				 blob on its texture, according to the current colour, or the
				 colour passed in.  This texture is then drawn as a background
				 to the scene.  It can also be stretched.
	Attribute:	colour
	AttType:	SFVec3f (input)
	AttDesc:	Sets the current painting colour.

	Attribute:	splodgePos
	AttType:	SFVec3f (input)
	AttDesc:	Draws a splodge at the passed-in position (assuming the position
				is within the viewing window.

	Attribute:	splodgePosCol
	AttType:	MFVec3f (input)
	AttDesc:	Same as splodgePos, only a second SFVec3f is passed in to set
				the colour to draw the splodge as.  It will accept any number of
				splodgePos's and colours.

	Attribute:	fDStretch
	AttType:	SFFloat (input)
	AttDesc:	Stretches the texture along the forward diagonal (/).

	Attribute:	bDStretch
	AttType:	SFFloat (input)
	AttDesc:	Stretches the texture along the backward diagonal (\).

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY" headlight="FALSE"/&gt;

		&lt;PaintingBackground DEF="backg"/&gt;

		&lt;Transform DEF="temp" translation="0.0 0.0 0.0"&gt;
			&lt;DirectionalLight DEF="lightDir" color="1.0 1.0 1.0" direction="-0.25 0.5 1.0"/&gt;
			&lt;DirectionalLight DEF="lightDir2" color="0.25 0.4 0.25" direction="0.0 -1.0 1.0"/&gt;

			&lt;Shape&gt;
				&lt;Ashitaka DEF="bob"/&gt;
			&lt;/Shape&gt;
		&lt;/Transform&gt;

		&lt;ROUTE fromNode="bob" fromField="quadsPosCol" toNode="backg" toField="splodgePosCol"/&gt;
		&lt;ROUTE fromNode="bob" fromField="wobble1" toNode="backg" toField="bDStretch"/&gt;
		&lt;ROUTE fromNode="bob" fromField="wobble2" toNode="backg" toField="fDStretch"/&gt;

	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Background which is updated in realtime with splodges sent via ROUTEs.
/*!
	\ingroup libheilanextras

	When the node receives a spotPos message via a ROUTE, it paints a faint
	spot at that point, in the current colour.  As more (differently-coloured)
	spots are painted, the background will tend to white.
 */
class PaintingBackground : public ChildNode
{
  public:
	///	Constructor.
	PaintingBackground();
	///	Destructor.
	~PaintingBackground();

	///	Sets up some stuff.
	void initialise();

	///	Draws the background.
	void graphicDoStuff();

	///	Used to set this node's attributes.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<PaintingBackground> constructor;
  private:
	///	Overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Helper method to add a new splodge to mainTexture.
	inline void addSplodge(const TwoFloats& pos, int width, int height)
	{
		int i;
		int x, y;
		int startPosX, startPosY;
		int tempint, tempint2;
		float tempf, tempf2;
		int tempWidth, tempHeight;
		float tempX, tempY;

		if(((pos.x-(SplodgeSize/2)) > 0.0f) &&
		   ((pos.x+(SplodgeSize/2)) < (float)MainTextureWidth) &&
		   ((pos.y-(SplodgeSize/2)) > 0.0f) &&
		   ((pos.y+(SplodgeSize/2)) < (float)MainTextureHeight))
		{
			tempX = pos.x * (float)MainTextureWidth/(float)width;
			startPosX = (int)tempX - (SplodgeSize/2);
			tempY = pos.y * (float)MainTextureHeight/(float)height;
			startPosY = (int)tempY - (SplodgeSize/2);

			tempint = (startPosX*3)+(startPosY*MainTextureWidth*3);
			if((tempint >= 0) &&
			   (tempint < (MainTextureWidth*MainTextureHeight*3)))
			{
				tempint2 = tempint;
				i = 0;
				for(y=0;y<SplodgeSize;++y)
				{
					for(x=0;x<SplodgeSize;++x)
					{
						if(((tempint+2) >= 0) &&
						   ((tempint+2) < (MainTextureWidth*MainTextureWidth*3)))
						{
							tempf = (float)splodge[i]/255.0f;
							tempf *= (float)splodge[i+1]/255.0f;
							tempf *= 255.0f * splodgeScale;
							i += 2;

							tempf2 = (float)mainTexture[tempint] +
									 (tempf * colour.x);
							if(tempf2 > 255.0f)
								tempf2 = 255.0f;
							mainTexture[tempint] = (int)tempf2;
							++tempint2;
							++tempint;

							tempf2 = (float)mainTexture[tempint] +
									 (tempf * colour.y);
							if(tempf2 > 255.0f)
								tempf2 = 255.0f;
							mainTexture[tempint] = (int)tempf2;
							++tempint2;
							++tempint;

							tempf2 = (float)mainTexture[tempint] +
									 (tempf * colour.z);
							if(tempf2 > 255.0f)
								tempf2 = 255.0f;
							mainTexture[tempint] = (int)tempf2;
							++tempint2;
							++tempint;
						}
						else
						{
							tempint += 3;
							i += 2;
						}
					}
					tempint += ((MainTextureWidth*3)-(SplodgeSize*3));
				}
			}
			else
			{
				startPosX = 0;
				startPosY = 0;
			}
			//exit(0);

			tempWidth = (int)((float)SplodgeSize *
							  (float)width/(float)MainTextureWidth);
			//glTexSubImage2D seems to expect 4byte boundaries for some reason.
			while((tempWidth%4) != 0)
				++tempWidth;
			tempHeight = (int)((float)SplodgeSize *
							   (float)height/(float)MainTextureHeight);
			while((tempHeight%4) != 0)
				++tempHeight;
			blitToTempTex(startPosX, startPosY, tempWidth, tempHeight);

			startPosX = (int)pos.x - (tempWidth/2);
			startPosY = (int)pos.y - (tempHeight/2);

			if(startPosX < 0)
			{
				tempWidth = startPosX + tempWidth;
				startPosX = 0;
			}
			else if((startPosX+tempWidth) >= MainTextureWidth)
				tempWidth = MainTextureWidth - startPosX;
			if(startPosY < 0)
			{
				tempHeight = startPosY + tempHeight;
				startPosY = 0;
			}
			else if((startPosY+SplodgeSize) >= MainTextureHeight)
				tempHeight = MainTextureHeight - startPosY;

			glTexSubImage2D(GL_TEXTURE_2D,
							0,
							startPosX,
							startPosY,
							tempWidth,
							tempHeight,
							GL_RGB,
							GL_UNSIGNED_BYTE,
							tempTex);
		}
	};
	///	Helper method to transform a 3d coordinate to 2d screen pixels.
	/*!
		Remember to scale it into MainTextureWidth x MainTextureHeight when
		you get the return value...
	 */
	inline TwoFloats getPixelCoords(const ThreeFloats& coords,
									int width,
									int height)
	{
		TwoFloats retval;
		double modelMatrix[16];
		double projMatrix[16];
		GLint view[4];
		double winX, winY, winZ;

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		gluPerspective(45.0f,
				   static_cast<float>(width)/static_cast<float>(height),
				   0.1f,
				   500.0f);

		glGetDoublev(GL_MODELVIEW_MATRIX, modelMatrix);
		glGetDoublev(GL_PROJECTION_MATRIX, projMatrix);
		glGetIntegerv(GL_VIEWPORT, view);
		gluProject(coords.x,
				   coords.y,
				   coords.z,
				   modelMatrix,
				   projMatrix,
				   view,
				   &winX,
				   &winY,
				   &winZ);

		retval.x = winX;
		retval.y = height-winY;

		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);

		return retval;
	};
	///	Helper method to blit a 32x32 section of mainTexture to tempTex.
	/*!
		\param x X-axis position in mainTexture to start at.
		\param y Y-axis position in mainTexture to start at.
		\param width Width to scale the 32x32 section to.
		\param height Height to scale the 32x32 section to.
	 */
	inline void blitToTempTex(int x, int y, int width, int height)
	{
		float mTX, mTY;
		int tTX, tTY;
		float xInc, yInc;
		int tempWidth, tempHeight;
		int i;
		int tempint;

		if(width > (SplodgeSize*2))
			tempWidth = 64;
		else
			tempWidth = width;
		if(height > (SplodgeSize*2))
			tempHeight = 64;
		else
			tempHeight = height;

		mTX = (float)x;
		mTY = (float)y;
		xInc = (float)SplodgeSize/(float)tempWidth;
		yInc = (float)SplodgeSize/(float)tempHeight;

		i = 0;
		for(tTY=0;tTY<tempHeight;++tTY)
		{
			mTX = (float)x;
			for(tTX=0;tTX<tempWidth;++tTX)
			{
				tempint = (heilanRound(mTX)*3) +
						  (heilanRound(mTY)*MainTextureWidth*3);
				tempTex[i] = mainTexture[tempint];
				++i;
				tempTex[i] = mainTexture[tempint+1];
				++i;
				tempTex[i] = mainTexture[tempint+2];
				++i;
				mTX += xInc;
			}
			mTY += yInc;
		}
	};

	///	Some constants.
	enum
	{
		MainTextureWidth = 1024, ///<This <em>MUST</em> be &gt;= MainTextureHeight.
		MainTextureHeight = 1024,
		SplodgeSize = 48
	};

	///	Used to stretch the texture in the forward diagonal.
	float fDStretch;
	///	Used to stretch the texture in the backward diagonal.
	float bDStretch;

	///	Scale value used to determine how much of the splodge is added to mainTexture.
	float splodgeScale;

	///	The main texture we use.
	GLuint texture;
	///	Our main texture pixel data.
	//unsigned char mainTexture[MainTextureWidth*MainTextureWidth*3];
	unsigned char *mainTexture;
	///	Our splodge pixel data.
	unsigned char splodge[SplodgeSize*SplodgeSize*2];
	///	Used to write new splodges to the texture.
	unsigned char tempTex[SplodgeSize*2*SplodgeSize*2*3];

	///	Current colour to draw splodges.
	ThreeFloats colour;
	///	Temporary storage of new splodge positions.
	std::queue<ThreeFloats> splodgePos;
	///	Temporary storage of new splodge colours.
	std::queue<ThreeFloats> splodgeCols;

	///	CriticalSection used to protect splodgePos.
	CriticalSection critSec;
};

}

#endif
