//	ParticleGenerator.cpp - Node which constantly fires off particles.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "ParticleGenerator.h"
#include "StringParser.h"
#include "AttributeHelper.h"
#include "FileFinder.h"
#include "MiscHelper.h"

#include "SDL_image.h"
#include <cmath>
#include <algorithm>
#include <iostream>

using namespace std;

namespace HeilanX3D
{

bool particleZOrder(const Particle& left, const Particle& right)
{
	return left.getPosition().z < right.getPosition().z;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
Particle::Particle():
lifespan(0.0f),
gravity(0.05f),
speed(0.0f)
{
	direction.y = 1.0f;
	gravityDirection.y = -1.0f;
	colour = 1.0f;
	endColour = 1.0f;
}

//------------------------------------------------------------------------------
Particle::~Particle()
{
	
}

//------------------------------------------------------------------------------
void Particle::fire()
{
	position = startPosition;
}

//------------------------------------------------------------------------------
void Particle::move()
{
	ThreeFloats oldPos = position;

	position += 0.1f * direction;
	//position += gravity * gravityDirection;
	direction += (gravity * (gravityDirection - direction));
	lifespan -= 0.01f;
	if(lifespan < 0.0f)
		lifespan = 0.0f;

	actualDirection = oldPos - position;
	speed = sqrtf((actualDirection.x * actualDirection.x) +
				  (actualDirection.y * actualDirection.y) +
				  (actualDirection.z * actualDirection.z));
}

//------------------------------------------------------------------------------
void Particle::setStartPosition(const ThreeFloats& val)
{
	startPosition = val;
}

//------------------------------------------------------------------------------
void Particle::setDirection(const ThreeFloats& val)
{
	direction = val;
}

//------------------------------------------------------------------------------
void Particle::setGravity(float val)
{
	gravity = val;
}

//------------------------------------------------------------------------------
void Particle::setGravityDirection(const ThreeFloats& val)
{
	gravityDirection = val;
}

//------------------------------------------------------------------------------
void Particle::setLifespan(float val)
{
	lifespan = val;
}

//------------------------------------------------------------------------------
void Particle::setColour(const ThreeFloats& val)
{
	colour = val;
}

//------------------------------------------------------------------------------
void Particle::setEndColour(const ThreeFloats& val)
{
	endColour = val;
}

NodeConstructor<ParticleGenerator> ParticleGenerator::constructor;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
ParticleGenerator::ParticleGenerator():
ChildNode(),
numParticles(8),
gravity(0.05f),
avgLifespan(1.0f),
particleSize(0.125f),
stop(false),
shrink(false),
textureWidth(256),
textureHeight(256),
pixels(0)
{
	setType("ParticleGenerator");
	setVisual(true);

	gravityDirection.y = -1.0f;
	colour = 1.0f;
	endColour = 1.0f;
}

//------------------------------------------------------------------------------
ParticleGenerator::~ParticleGenerator()
{
	if(pixels)
		delete [] pixels;
}

//------------------------------------------------------------------------------
void ParticleGenerator::initialise()
{
	int i;
	//int x, y;
	//float tempf;
	//unsigned char *pixelPointer;

	if(particles.size() < (unsigned)numParticles)
	{
		for(i=0;i<numParticles;++i)
		{
			particles.push_back(Particle());
	
			particles[i].setGravity(gravity);
			particles[i].setGravityDirection(gravityDirection);
		}
	}

	/*if(!pixels)
	{
		pixels = new unsigned char[textureWidth*textureHeight*2];
		pixelPointer = pixels;
		//Generate texture.
		for(y=0;y<256;++y)
		{
			for(x=0;x<256;++x)
			{
				tempf = sinf(((float)x/256.0f)*3.1415926535897932384626433832795f);
				tempf *= sinf(((float)y/256.0f)*3.1415926535897932384626433832795f);
				tempf = (exp(tempf)-1.0f)/(exp(1.0f)-1.0f)*1.5f-0.1f;
				if(tempf < 0.0f)
					tempf = 0.0f;
				else if(tempf > 1.0f)
					tempf = 1.0f;
				*pixelPointer = 255;
				*(pixelPointer+1) = (unsigned char)(tempf * 255.0f);
				pixelPointer += 2;
			}
		}
	}

	//Setup texture.
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexImage2D(GL_TEXTURE_2D,
				 0,
				 2,
				 256,
				 256,
				 0,
				 GL_LUMINANCE_ALPHA,
				 GL_UNSIGNED_BYTE,
				 pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);*/
	loadDefaultTexture();

	//Generate the display list.
	glNewList(displayList, GL_COMPILE);
	displayList = glGenLists(1);
	{
		glBegin(GL_QUADS);
		{
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(-0.125f, 0.125f, 0.0f);
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(0.125f, 0.125f, 0.0f);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(0.125f, -0.125f, 0.0f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-0.125f, -0.125f, 0.0f);
		}
		glEnd();
	}
	glEndList();

	if(url != "")
		loadTexture();
}

//------------------------------------------------------------------------------
void ParticleGenerator::graphicDoStuff()
{
	int i;
	float tempf2;
	int count = 0;
	float tempRand;
	ThreeFloats tempf;

	for(i=0;i<numParticles;++i)
	{
		if((!particles[i].getIsActive()) && (!stop))
		{
			tempRand = (float)rand()/(float)RAND_MAX;
			tempRand -= 0.5f;
			tempRand *= avgLifespan;
			particles[i].setLifespan(avgLifespan+tempRand);

			tempRand = (float)rand()/(float)RAND_MAX;
			tempRand -= 0.5f;
			tempf.x = tempRand;
			tempRand = (float)rand()/(float)RAND_MAX;
			tempRand -= 0.5f;
			tempf.y = tempRand;
			tempRand = (float)rand()/(float)RAND_MAX;
			tempRand -= 0.5f;
			tempf.z = tempRand;
			particles[i].setDirection(tempf);

			particles[i].setColour(colour);
			particles[i].setEndColour(endColour);

			particles[i].fire();

			//Used to maintain a fairly constant number of particles output,
			//rather than having them all output in one big clump at the
			//start.
			++count;
			if(count > 6)
				break;
		}
		else
			particles[i].move();
	}

	//Make sure the z order's right.
	//sort(particles.begin(), particles.end(), particleZOrder);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);

	critSec.lock();
	{
		glBindTexture(GL_TEXTURE_2D, texture);

		glDisable(GL_LIGHTING);

		for(i=0;i<numParticles;++i)
		{
			if(particles[i].getIsActive())
			{
				//tempf2 = heilanLog(particles[i].getLifespan());
				tempf2 = particles[i].getLifespan();
				ThreeFloats tempCol = particles[i].getColour();
				tempCol *= tempf2;
				tempCol += particles[i].getEndColour() * (1.0f-tempf2);
				glColor4f(tempCol.x,
						  tempCol.y,
						  tempCol.z,
						  particles[i].getLifespan());

				glTranslatef(particles[i].getPosition().x,
							 particles[i].getPosition().y,
							 particles[i].getPosition().z);

				if(!shrink)
					tempf2 = 1.0f;
				//glCallList(displayList);
				glBegin(GL_QUADS);
				{
					glTexCoord2f(0.0f, 1.0f);
					glVertex3f(-particleSize*tempf2, particleSize*tempf2, 0.0f);
					glTexCoord2f(1.0f, 1.0f);
					glVertex3f(particleSize*tempf2, particleSize*tempf2, 0.0f);
					glTexCoord2f(1.0f, 0.0f);
					glVertex3f(particleSize*tempf2, -particleSize*tempf2, 0.0f);
					glTexCoord2f(0.0f, 0.0f);
					glVertex3f(-particleSize*tempf2, -particleSize*tempf2, 0.0f);
				}
				glEnd();

				glTranslatef(-particles[i].getPosition().x,
							 -particles[i].getPosition().y,
							 -particles[i].getPosition().z);
			}
		}

		glBindTexture(GL_TEXTURE_2D, 0);
		glEnable(GL_LIGHTING);
		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);
	}
	critSec.unlock();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glColor3f(1.0f, 1.0f, 1.0f);
}

//------------------------------------------------------------------------------
void ParticleGenerator::setAttribute(const string& attribute, EventValue& val)
{
	int i;
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "numParticles")
	{
		particles.clear();
		for(i=0;i<(int)val;++i)
		{
			/*ThreeFloats tempCol((float)rand()/(float)RAND_MAX,
								(float)rand()/(float)RAND_MAX,
								(float)rand()/(float)RAND_MAX);*/
			ThreeFloats tempCol(1.0f, 1.0f, 0.25f);

			particles.push_back(Particle());

			particles[i].setGravity(gravity);
			particles[i].setGravityDirection(gravityDirection);
			particles[i].setColour(tempCol);
		}
		numParticles = val;
	}
	else if(realAtt == "gravity")
	{
		gravity = val;

		for(i=0;i<numParticles;++i)
			particles[i].setGravity(gravity);
	}
	else if(realAtt == "gravityDirection")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&gravityDirection);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			gravityDirection = parser.getThreeFloats();

		for(i=0;i<numParticles;++i)
			particles[i].setGravityDirection(gravityDirection);
	}
	else if(realAtt == "avgLifespan")
		avgLifespan = val;
	else if(realAtt == "colour")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&colour);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			colour = parser.getThreeFloats();

		/*for(i=0;i<numParticles;++i)
			particles[i].setColour(colour);*/
	}
	else if(realAtt == "endColour")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&endColour);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			endColour = parser.getThreeFloats();

		/*for(i=0;i<numParticles;++i)
			particles[i].setEndColour(endColour);*/
	}
	else if(realAtt == "particleSize")
		particleSize = val;
	else if(realAtt == "url")
	{
		url = (string)val;

		critSec.lock();
		{
			if(pixels)
				delete [] pixels;
			pixels = 0;

			loadTexture();
		}
		critSec.unlock();
	}
	else if(realAtt == "position")
	{
		ThreeFloats tempP;
		StringParser parser(val);

		if(parser.getNumThreeFloats() > 0)
			tempP = parser.getThreeFloats();

		for(i=0;i<numParticles;++i)
			particles[i].setStartPosition(tempP);
	}
	else if(realAtt == "stop")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			stop = parser.getBool();
	}
	else if(realAtt == "shrink")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			shrink = parser.getBool();
	}
}

//------------------------------------------------------------------------------
void ParticleGenerator::loadTexture()
{
	string tempPath;
	SDL_Surface *tempTex = 0;
	SDL_Surface *tempTex2 = 0;
	SDL_Surface *tempTex3 = 0;

	tempPath = FileFinder::getInstance().findFile(url);
	if(tempPath != "")
	{
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
	
		//This is the easiest way I could find to create a sensible SDL_PixelFormat
		//for SDL_ConvertSurface().
#ifdef __BIG_ENDIAN__
		tempTex3 = SDL_CreateRGBSurface(SDL_SWSURFACE,
										1,
										1,
										32,
										0xff000000,
										0x00ff0000,
										0x0000ff00,
										0x000000ff);
#else
		tempTex3 = SDL_CreateRGBSurface(SDL_SWSURFACE,
										1,
										1,
										32,
										0x000000ff,
										0x0000ff00,
										0x00ff0000,
										0xff000000);
#endif
	
		if((tempTex = IMG_Load(tempPath.c_str())))
		{
			//We first convert the image into a specific format, before we call
			//glTexImage2D.
			if((tempTex2 = SDL_ConvertSurface(tempTex,
				tempTex3->format,
				SDL_SWSURFACE)))
			{
				glGenTextures(1, &texture);
	
				glBindTexture(GL_TEXTURE_2D, texture);
	
				glTexImage2D(GL_TEXTURE_2D,
							 0,
							 GL_RGBA,
							 tempTex2->w,
							 tempTex2->h,
							 0,
							 GL_RGBA,
							 GL_UNSIGNED_BYTE,
							 tempTex2->pixels);
	
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	
				SDL_FreeSurface(tempTex2);
				SDL_FreeSurface(tempTex);
				glBindTexture(GL_TEXTURE_2D, 0);
			}
			else
			{
				SDL_FreeSurface(tempTex);
				cout << "ParticleGenerator: Could not load image: " << tempPath << endl;
			}
		}
		else
			cout << "ParticleGenerator: Could not load image: " << tempPath << endl;
	
		SDL_FreeSurface(tempTex3);
	}
	else
		cout << "ParticleGenerator: " << url << " does not seem to exist?" << endl;
}

//------------------------------------------------------------------------------
void ParticleGenerator::loadDefaultTexture()
{
	SDL_Surface *tempIm;

	if((tempIm = IMG_Load_RW(SDL_RWFromConstMem(GestureTextures::particle_png,
											    GestureTextures::particle_pngSize), 1)) != 0)
	{
		glGenTextures(1, &texture);

		glBindTexture(GL_TEXTURE_2D, texture);

		glTexImage2D(GL_TEXTURE_2D,
					 0,
					 GL_RGBA,
					 tempIm->w,
					 tempIm->h,
					 0,
					 GL_RGBA,
					 GL_UNSIGNED_BYTE,
					 tempIm->pixels);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		glBindTexture(GL_TEXTURE_2D, 0);

		SDL_FreeSurface(tempIm);
	}
}

}
