//	ParticleGenerator.h - Node which constantly fires off particles.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PARTICLEGENERATOR_H_
#define PARTICLEGENERATOR_H_

#include "AbstractNode.h"
#include "CriticalSection.h"

#include "GLHeaders.h"
#include <vector>

/*+
	Name: ParticleGenerator
	Type: ChildNode
	Description: A fairly simple particle generator.  Has attributes to control
				 the amount and direction of gravity applied to the particles,
				 as well as their average lifespan (how long they take to fade
				 away), and their colour.  No collision detection is applied.
	Attribute:	numParticles
	AttType:	SFInt32 (input)
	AttDesc:	How many particles may be visible at any one time.

	Attribute:	gravity
	AttType:	SFFloat (input)
	AttDesc:	The amount of gravity applied to the particles.

	Attribute:	gravityDirection
	AttType:	SFVec3f (input)
	AttDesc:	The direction of the gravity applied to the particles (should be
				normalised).

	Attribute:	avgLifespan
	AttType:	SFFloat (input)
	AttDesc:	How long (on average) the particles will live before they fade
				out completely.

	Attribute:	colour
	AttType:	SFVec3f (input)
	AttDesc:	The colour of the particles.

	Attribute:	particleSize
	AttType:	SFFloat (input)
	AttDesc:	The size of the particles.

	Attribute:	url
	AttType:	SFString (input)
	AttDesc:	Url to a texture to use for the particles, instead of the
				default procedural one.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY" headlight="FALSE"/&gt;

		&lt;Background skkyColor="0 0 0"/&gt;

		&lt;Transform DEF="temp" translation="0.0 0.0 -16.0"&gt;
			&lt;ParticleGenerator DEF="bob" numParticles="128" colour="0.25 0.25 1"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

//------------------------------------------------------------------------------
///	Simple particle class.
/*!
	\ingroup libheilanextras
 */
class Particle
{
  public:
	///	Constructor.
	Particle();
	///	Destructor.
	~Particle();

	///	Makes the particle visible, and starts it moving.
	void fire();
	///	Updates the particle's position if it's visible.
	void move();

	///	Returns the particle's current position.
	const ThreeFloats& getPosition() const {return position;};
	///	Returns the particle's current lifespan.
	float getLifespan() const {return lifespan;};
	///	Returns whether or not the particle's currently visible.
	bool getIsActive() const {return (lifespan != 0.0f);};
	///	Returns the colour of the particle.
	const ThreeFloats& getColour() const {return colour;};
	///	Returns the end colour of the particle.
	const ThreeFloats& getEndColour() const {return endColour;};
	///	Returns the direction of the particle.
	const ThreeFloats& getDirection() const {return direction;};
	///	Returns the gravity direction of the particle.
	const ThreeFloats& getGravityDirection() const {return gravityDirection;};
	///	Returns the actual direction of the particle.
	const ThreeFloats& getActualDirection() const {return actualDirection;};
	///	Returns the speed of the particle.
	float getSpeed() const {return speed;};

	///	Sets the particle's starting position.
	void setStartPosition(const ThreeFloats& val);
	///	Sets the direction the particle should move in.
	void setDirection(const ThreeFloats& val);
	///	Sets the amount of gravity the particle experiences.
	void setGravity(float val);
	///	Sets the direction of the gravity applied to the particle.
	void setGravityDirection(const ThreeFloats& val);
	///	Sets the lifespan of the particle.
	/*!
		This *must* be called immediately prior to fire(), otherwise there is
		a risk that the particle will not fire at all.
	 */
	void setLifespan(float val);
	///	Sets the colour of the particle.
	void setColour(const ThreeFloats& val);
	///	Sets the end colour of the particle.
	void setEndColour(const ThreeFloats& val);
  private:
	///	The particle's current position.
	ThreeFloats position;
	///	The particle's current direction.
	ThreeFloats direction;
	///	The particle's current lifespan (how long till it dies).
	float lifespan;
	///	The particle's starting position.
	ThreeFloats startPosition;
	///	The amount of gravity applied to the particle.
	float gravity;
	///	The direction of the gravity applied to the particle.
	ThreeFloats gravityDirection;
	///	The colour of the particle.
	ThreeFloats colour;
	///	The end colour of the particle.
	ThreeFloats endColour;

	///	The actual direction of the particle.
	ThreeFloats actualDirection;
	///	The speed of the particle.
 	float speed;
};

//------------------------------------------------------------------------------
///	Node which constantly fires off particles.
/*!
	\ingroup libheilanextras
 */
class ParticleGenerator : public ChildNode
{
  public:
	///	Constructor.
	ParticleGenerator();
	///	Destructor.
	~ParticleGenerator();

	///	Initialises the particles.
	void initialise();

	///	Draws the ParticleGenerator.
	void graphicDoStuff();
	///	Not used, overridden to do nothing.
	/*AmbisonicData *audioDoStuff(int numSamples,
								const ThreeFloats& pos) {return 0;};*/

	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set the size and solidity of the box.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<ParticleGenerator> constructor;
  private:
	///	Tries to load the texture in url.
	void loadTexture();
	///	Loads the default texture.
	void loadDefaultTexture();

	///	The number of particles used.
	int numParticles;
	///	The amount of gravity used.
	float gravity;
	///	The direction of the gravity.
	ThreeFloats gravityDirection;
	///	The average lifespan of a particle.
	float avgLifespan;
	///	The colour of the particles.
	ThreeFloats colour;
	///	The end colour of the particles.
	ThreeFloats endColour;
	///	The size of the particles.
	float particleSize;
	///	The texture to use for the particles.
	std::string url;
	///	Used to stop the production of new particles.
	bool stop;
	///	Used to shrink the particles as their life goes on.
	bool shrink;

	///	Vector holding our particles.
	std::vector<Particle> particles;
	///	Display list for a single particle.
	GLuint displayList;
	///	Our particle texture.
	GLuint texture;
	///	The width of the texture.
	int textureWidth;
	///	The height of the texture.
	int textureHeight;
	///	Our pixel data.
	unsigned char *pixels;

	///	Critical section to protect the loading of textures.
	CriticalSection critSec;
};

}

///	PNG textures as char arrays, generated from JUCE' BinaryBuilder.
namespace GestureTextures
{
    extern const char*  particle_png;
    const int           particle_pngSize = 59933;
};

#endif
