//	PhysicalModel.cpp - Node to test & develop the physical model to be used in
//						the final instrument.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "PhysicalModel.h"
#include "Camera.h"
#include "StringParser.h"
#include "AmbisonicData.h"

#include "GLHeaders.h"
#include <cassert>
#include <iostream>

using namespace std;

namespace HeilanX3D
{
	
NodeConstructor<PhysicalModel> PhysicalModel::constructor;

//------------------------------------------------------------------------------
PhysicalModel::PhysicalModel():
ChildNode(),
stopPos(0.0f),
pluckAmp(0.0f),
lastPluck(1.0f),
sawtoothAmp(0.0f),
playSawtooth(false),
feedbackPos(StringLength/4),
feedbackAmp(0.0f),
dampingValue(0.0f),
dampingLeft(0),
dampingRight(0),
plucking(false),
bowForce(0.0f),
bowVelocity(0.0f),
bowPosition(7*(StringLength/8)),
bowMode(false),
bowPointPosition(0.0f),
bowForceExerted(0.0f),
force1(0.0f),
force2(0.0f),
force3(0.0f),
force4(0.0f),
audioBuffer(0),
bufferSize(0),
lastAtten(0.0f)
{
	setType("PhysicalModel");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
PhysicalModel::~PhysicalModel()
{
	if(audioBuffer)
		delete [] audioBuffer;
}

//------------------------------------------------------------------------------
void PhysicalModel::initialise()
{
	int i;
	float tempFreq;

	setMaxFrontAudioRadius(50.0f);
	setMaxBackAudioRadius(50.0f);
	setMinFrontAudioRadius(25.0f);
	setMinBackAudioRadius(25.0f);

	//Calculate audio stuff.
	for(i=0;i<StringLength;++i)
	{
		delayLine[i] = 0.0f;
		forces[i] = 0.0f;
		velocities[i] = 0.0f;
	}
	for(i=0;i<NumSines;++i)
	{
		tempFreq = 110.0f * (float)(i+1);
		sawtooth[i].coeff = 2.0f * sinf(3.141592653f*(tempFreq/44100.0f));
	}

	pluckWave.coeff = 2.0f * sinf(3.141592653f*(2756.25f/44100.0f));
	pluckWave.x0 = 0.0f;
	pluckWave.x1 = -1.0f;

	//Calculate cells' mass.
	float defaultVelocityMultiplier;
	float decay = 5.0f; //5s?
	float intendedFreq = 110.0f;
	float actualFreq;
	float compensationFactor;
	defaultVelocityMultiplier = (1.0-((.00012f*44100.0f)/44100.0f)/decay);
	actualFreq = (24000.0f*44100.0f/44100.0f)/(StringLength+1);
	compensationFactor=powf(4.0f, log10f(actualFreq/intendedFreq)/log10f(2.0f));
	mass = 3.5f * compensationFactor;
	invMass = 1.0f/mass;

	velocityMultiplier = defaultVelocityMultiplier;
	stopMultiplier = 1.0f-(powf(10000.0f, 0.95f)/10000.0f);
	dampMultiplier = 1.0f-(powf(10000.0f, dampingValue)/10000.0f);
}

//------------------------------------------------------------------------------
void PhysicalModel::graphicDoStuff()
{
	int i;

	glDisable(GL_LIGHTING);
	glColor3f(0.0f, 0.0f, 0.0f);
	glLineWidth(2.0f);

	glBegin(GL_LINE_STRIP);
	{
		for(i=0;i<StringLength;++i)
		{
			glVertex3f(((float)(i-StringLength/2)*0.01f),
					   delayLine[i],
					   0.0f);
		}
	}
	glEnd();

	glEnable(GL_LIGHTING);
}

//------------------------------------------------------------------------------
AmbisonicData *PhysicalModel::getMonoAudio(int numSamples)
{
	int i, j;
	float tempf = 0.0f;
	float tempPitchf;
	int tempPitchi;
	int tempint;
	float tempSaw = 0.0f;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		tempf = 0.0f;

		//Calculate forces.
		{
			for(j=1;j<(StringLength-1);++j)
			{
				forces[j] = (delayLine[j-1] + delayLine[j+1]) -
							(2.0f * delayLine[j]);
			}
			forces[0] = forces[1] - forces[0];
			forces[StringLength-1] = forces[StringLength-2] -
									 forces[StringLength-1];
		}

		//Update pitch stuff.
		{
			tempPitchf = stopPos * (float)StringLength;
			tempPitchi = (int)tempPitchf;
		}

		//Update excitation stuff.
		if(sawtoothAmp > 0.0f)
		{
			tempSaw = 0.0f;
			for(j=0;j<NumSines;++j)
			{
				sawtooth[j].x0 = sawtooth[j].x0 - (sawtooth[j].x1 *
												   sawtooth[j].coeff);
				sawtooth[j].x1 = sawtooth[j].x1 + (sawtooth[j].x0 *
												   sawtooth[j].coeff);
				tempSaw += sawtooth[j].x0 * (1.0f/(float)(j+1));
			}
			delayLine[7*(StringLength/8)] += tempSaw * 0.05f * sawtoothAmp;
		}
		if(plucking)
		{
			pluckWave.x0 = pluckWave.x0 - (pluckWave.x1 * pluckWave.coeff);
			pluckWave.x1 = pluckWave.x1 + (pluckWave.x0 * pluckWave.coeff);

			//delayLine[7*(StringLength/8)] += (1.0f-((pluckWave.x0*0.5f)+0.5f));

			tempint = (7*(StringLength/8));
			for(j=tempPitchi;j<(StringLength-1);++j)
			{
				if(j < tempint)
				{
					delayLine[j] += ((float)(j-tempPitchi)/(float)tempint) *
									((pluckWave.x1*0.5f)+0.5f);
									//(1.0f-((pluckWave.x0*0.5f)+0.5f));
				}
				else
				{
					delayLine[j] += (float)(StringLength-j-tempPitchi)/((float)(StringLength-tempint)) *
									((pluckWave.x1*0.5f)+0.5f);
									//(1.0f-((pluckWave.x0*0.5f)+0.5f));
				}

				if(delayLine[j] > 1.0f)
					delayLine[j] = 1.0f;
			}

			if(pluckWave.x1 < -0.9f)
			{
				plucking = false;
				pluckWave.x0 = 0.0f;
				pluckWave.x1 = -1.0f;
			}
		}
		//Add force sensors.
		{
			int tempPos = StringLength/5;
			delayLine[tempPos] += force1;
			tempPos += StringLength/5;
			delayLine[tempPos] += force2;
			tempPos += StringLength/5;
			delayLine[tempPos] += force3;
			tempPos += StringLength/5;
			delayLine[tempPos] += force4;
		}
		//Add feedback.
		if(feedbackAmp > 0.0f)
		{
			delayLine[7*(StringLength/8)] += tanh(delayLine[feedbackPos] *
												  feedbackAmp * 3.1415f) *
											 0.005f;
		}
		//Calculate bow stuff.
		if((bowForce > 0.0f) && (bowVelocity > 0.0f))
		{
			float relativeVelocity;

			bowPointPosition += bowVelocity;
			relativeVelocity = bowVelocity - velocities[bowPosition];

			if(!bowMode)
			{
				float stickingForce = bowPointPosition - delayLine[bowPosition];

				if(stickingForce > bowForce)
					bowMode = true;
				else
					bowForceExerted = stickingForce;
			}
			else
			{
				float slippingForce = bowForce/(1.0f+fabs(relativeVelocity));

				if(velocities[bowPosition] >= 0.0f)
				{
					bowMode = false;
					bowPointPosition = delayLine[bowPosition];
				}
				else
					bowForceExerted = slippingForce;
			}

			forces[bowPosition] += bowForceExerted;
			if(bowPosition > 1)
				forces[bowPosition-1] += bowForceExerted * 0.5f;
			if(bowPosition < (StringLength-2))
				forces[bowPosition+1] += bowForceExerted * 0.5f;
		}

		//Calculate positions.
		for(j=1;j<(StringLength-1);++j)
		{
			//velocities[j] += forces[j] * 0.285714286; //0.285714286 = 1/3.5
			//velocities[j] += forces[j] * 0.096812616; //0.096812616 = 1/10.329
			velocities[j] += forces[j] * invMass;
			if((j >= dampingLeft) && (j < dampingRight))
				velocities[j] *= dampMultiplier;
			else if(j > tempPitchi)
				velocities[j] *= velocityMultiplier;
			else
				velocities[j] *= stopMultiplier;
			delayLine[j] += velocities[j];
		}

		//Calculate audio.
		tempf = delayLine[3*(StringLength/4)];

		//Convert float sine tone to Ambisonic, put it in the buffer.
		if(!playSawtooth)
			audioBuffer[i].W = tempf * 0.5f;
		else
			audioBuffer[i].W = tempSaw * sawtoothAmp;
	}

	return audioBuffer;
}

//------------------------------------------------------------------------------
void PhysicalModel::setBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}

//------------------------------------------------------------------------------
void PhysicalModel::setAttribute(const string& attribute, EventValue& val)
{
	StringParser parser(val);

	if(attribute == "stopPos")
		stopPos = val;
	else if(attribute == "pluck")
	{
		if(((float)val >= 0.5f) && (lastPluck < 0.5f))
		{
			pluckAmp = 0.0f;
			plucking = true;
		}
		lastPluck = val;
	}
	else if(attribute == "sawtoothAmp")
		sawtoothAmp = val;
	else if(attribute == "playSawtooth")
	{
		if((float)val < 0.5f)
			playSawtooth = false;
		else
			playSawtooth = true;
	}
	else if(attribute == "stringFreq")
	{
		float intendedFreq = val;
		float actualFreq;
		float compensationFactor;
		if(intendedFreq < 55.0f)
			intendedFreq = 55.0f;
		else if(intendedFreq > 220.0f)
			intendedFreq = 220.0f;
		actualFreq = (24000.0f*44100.0f/44100.0f)/(StringLength+1);
		compensationFactor=powf(4.0f, log10f(actualFreq/intendedFreq)/log10f(2.0f));
		mass = 3.5f * compensationFactor;
		invMass = 1.0f/mass;
	}
	else if(attribute == "feedbackPos")
		feedbackPos = (int)((float)val * (float)StringLength);
	else if(attribute == "feedbackAmp")
		feedbackAmp = val;
	else if(attribute == "bowPosition")
		bowPosition = val;
	else if(attribute == "bowForce")
		bowForce = val;
	else if(attribute == "bowVelocity")
		bowVelocity = val;
	else if(attribute == "dampingValue")
	{
		dampingValue = val;
		dampMultiplier = 1.0f-(powf(10000.0f, dampingValue)/10000.0f);
	}
	else if(attribute == "dampingLeft")
	{
		int tempint = val;

		if(tempint < 0)
			tempint = 0;
		else if(tempint > (StringLength-1))
			tempint = StringLength-1;

		dampingLeft = tempint;
	}
	else if(attribute == "dampingRight")
	{
		int tempint = val;

		if(tempint < 0)
			tempint = 0;
		else if(tempint > (StringLength-1))
			tempint = StringLength-1;

		dampingRight = tempint;
	}
	else if(attribute == "force1")
		force1 = val;
	else if(attribute == "force2")
		force2 = val;
	else if(attribute == "force3")
		force3 = val;
	else if(attribute == "force4")
		force4 = val;
}

}
