//	PhysicalModel.h - Node to test & develop the physical model to be used in
//					  the final instrument.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PHYSICALMODEL_H_
#define PHYSICALMODEL_H_

#include "AbstractNode.h"

/*+
	Name: PhysicalModel
	Type: ChildNode.
	Description: A node representing a physically-modelled string (which is used
			 	 in most of the audiovisual instruments here, and will be used
				 in the final audiovisual instrument).
	Attribute:	stopPos
	AttType:	SFFloat (input)
	AttDesc:	The position of the stop on the string (range: 0-&gt;1).

	Attribute:	pluck
	AttType:	SFFloat (input)
	AttDesc:	When it goes past the 0.5 mark, plucks the string.

	Attribute:	sawtoothAmp
	AttType:	SFFloat (input)
	AttDesc:	Amplitude of a sawtooth generator applied to the string.

	Attribute:	playSawtooth
	AttType:	SFFloat (input)
	AttDesc:	Whether to apply the sawtooth to the string or not (&lt;0.5 ==
				no, &gt;0.5 == yes).

	Attribute:	stringFreq
	AttType:	SFFloat (input)
	AttDesc:	The frequency the string will resonate at equivalent to
				stretching the string on a guitar) (range: 55-&gt;220).

	Attribute:	feedbackPos
	AttType:	SFFloat (input)
	AttDesc:	The position on the string that the feedback element takes its
				input from (range: 0-&gt;1).

	Attribute:	feedbackAmp
	AttType:	SFFloat (input)
	AttDesc:	How much feedback is applied.

	Attribute:	bowPosition
	AttType:	SFInt32 (input)
	AttDesc:	The position on the string that the bow is applied at (range:
				0-&gt;127).

	Attribute:	bowForce
	AttType:	SFFloat (input)
	AttDesc:	The force applied to the bow (range: 0-&gt;0.1).

	Attribute:	bowVelocity
	AttType:	SFFloat (input)
	AttDesc:	The velocity of the bow (range: 0-&gt;0.1).

	Attribute:	dampingValue
	AttType:	SFFloat (input)
	AttDesc:	How much to damp the string by.

	Attribute:	dampingLeft
	AttType:	SFInt32 (input)
	AttDesc:	The left position of the damped section (range: 0-&gt;127).

	Attribute:	dampingRight
	AttType:	SFInt32 (input)
	AttDesc:	The right position of the damped section (range: 0-&gt;127).

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0.0 0.0 -2.0"&gt;
			&lt;PhysicalModel DEF="bob"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Node to test & develop the physical model to be used in the final instrument.
/*!
	\ingroup libheilanextras
 */
class PhysicalModel : public ChildNode
{
  public:
	///	Constructor.
	PhysicalModel();
	///	Destructor.
	~PhysicalModel();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the visuals.
	void graphicDoStuff();
	///	To handle the instrument's audio output.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Called to set the current audio block size.
	void setBlockSize(int newSize);

	///	Called to set the value of the named attribute.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<PhysicalModel> constructor;
  private:
	///	Various constants.
	enum
	{
		StringLength = 128,
		NumSines = 32
	};

	///	Delay line for the string (i.e. the positions of the cells).
	float delayLine[StringLength];
	///	Forces for the string cells.
	float forces[StringLength];
	///	Velocities for the string cells.
	float velocities[StringLength];
	///	Mass of the individual cells.
	float mass;
	///	Inverse mass of the individual cells.
	float invMass;
	///	How much to multiply the cells' velocities by (default).
	float velocityMultiplier;
	///	How much to multiply the cells' velocities by (stopped).
	float stopMultiplier;
	///	How much to multiply the cells' velocities by (damped).
	float dampMultiplier;

	///	Position of the stop.
	float stopPos;
	///	Current pluck amplitude.
	float pluckAmp;
	///	Last pluck value (only pluck when value initially goes past 50%).
	float lastPluck;
	///	Current amplitude of the sawtooth wave.
	float sawtoothAmp;
	///	True to let us hear the sawtooth wave.
	bool playSawtooth;
	///	Position on the string to get feedback from.
	int feedbackPos;
	///	Amplitude of the feedback.
	float feedbackAmp;
	///	Damping value.
	float dampingValue;
	///	Damping left position.
	int dampingLeft;
	///	Damping right position.
	int dampingRight;

	///	Simple struct representing a sine wave oscillator.
	struct Sinewave
	{
		float x0;
		float x1;
		float coeff;

		Sinewave()
		{
			x0 = 0.0f;
			x1 = 1.0f;
		}
	};
	///	Sine waves used to construct a sawtooth wave.
	Sinewave sawtooth[NumSines];
	///	Sine wave used to create a raised-cosine pluck.
	Sinewave pluckWave;
	///	If we should be plucking at the moment.
	bool plucking;

	///	Bow force.
	float bowForce;
	///	Bow velocity.
	float bowVelocity;
	///	Bow position.
	int bowPosition;
	///	Bow mode.
	bool bowMode;
	///	Position (y-axis) of the bow.
	float bowPointPosition;
	///	The actual force exerted by the bow.
	float bowForceExerted;

	///	Force sensor 1 force.
	float force1;
	///	Force sensor 2 force.
	float force2;
	///	Force sensor 3 force.
	float force3;
	///	Force sensor 4 force.
	float force4;

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;
	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;
};

}

#endif
