//	PhysicalSurface.h - A simple visual plane based on Tao's physical modelling
//						techniques.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PHYSICALSURFACE_H_
#define PHYSICALSURFACE_H_

#include "AbstractNode.h"

#include "GLHeaders.h"
#include <cmath>
#include <iostream>

/*+
	Name: PhysicalSurface
	Type: ChildNode.
	Description: A simple visual plane based on Tao's physical modelling
				 techniques.  I made this to test out the use of Tao's
				 techniques in a deformable sphere (the DeformableSphere2
				 node), and also to test ways of subdividing a plane (we use
				 a tiny Tao surface (v.few cells) and interpolate between
				 them to get a smooth surface).
	Attribute:	force1
	AttType:	SFFloat (input)
	AttDesc:	Force applied to the lower left cell.

	Attribute:	force2
	AttType:	SFFloat (input)
	AttDesc:	Force applied to the upper left cell.

	Attribute:	force3
	AttType:	SFFloat (input)
	AttDesc:	Force applied to the lower right cell.

	Attribute:	force4
	AttType:	SFFloat (input)
	AttDesc:	Force applied to the upper right cell.

	Attribute:	drawLines
	AttType:	SFFloat (input)
	AttDesc:	Whether to draw lines representing the Tao surface or not.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY" headlight="FALSE"/&gt;

		&lt;Transform DEF="temp" translation="0 -2 -8"&gt;
			&lt;PhysicalSurface DEF="bob"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	A simple visual plane based on Tao's physical modelling techniques.
/*!
	\ingroup libheilanextras
 */
class PhysicalSurface : public ChildNode
{
  public:
	///	Constructor.
	PhysicalSurface();
	///	Destructor.
	~PhysicalSurface();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the surface.
	void graphicDoStuff();
	///	Not used, overridden to do nothing.
	/*AmbisonicData *audioDoStuff(int numSamples,
								const ThreeFloats& pos) {return 0;};*/

	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set the sphere's parameters.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<PhysicalSurface> constructor;
  private:
	///	Linear interpolation helper method.
	/*!
		Assumes val is of the range 0->1.
	 */
	inline float linearInterp(float left, float right, float val)
	{
		float retval;

		retval = left + (val*(right-left));

		return retval;
	};
	///	B-spline interpolation helper method.
	inline float bSplineInterp(float x0,
							   float x1,
							   float x2,
							   float x3,
							   float val)
	{
		// 4-point, 3rd-order B-spline (x-form)
		float ym1py1 = x0+x2;
		float c0 = (1.0f/6.0f) * ym1py1 + (2.0f/3.0f) * x1;
		float c1 = (1.0f/2.0f) * (x2-x0);
		float c2 = (1.0f/2.0f) * ym1py1 - x1;
		float c3 = (1.0f/2.0f) * (x1-x2) + (1.0f/6.0f) * (x3-x0);
		return ((c3*val+c2)*val+c1)*val+c0;
	};
	///	Helper method to subdivide and draw a quad.
	/*!
		\param x The x position within positions of the lower left vertex.
		\param y The y position within positions of the lower left vertex.
	 */
	inline void simpleSubdivide(int x, int y)
	{
		int i, j;
		float tempi, tempj;
		float temp1, temp2;
		int tempint;
		float vertices[NumSubdivisions+1][NumSubdivisions+1];
		float corners[4];

		//Calculate corner positions.
		tempint = x + (y*SurfaceWidth);
		corners[0] = positions[tempint];
		tempint = x + 1 + (y*SurfaceWidth);
		corners[1] = positions[tempint];
		if((y+1) >= SurfaceHeight)
			tempint = x + 1;
		else
			tempint = x + 1 + ((y+1)*SurfaceWidth);
		corners[2] = positions[tempint];
		if((y+1) >= SurfaceHeight)
			tempint = x;
		else
			tempint = x + ((y+1)*SurfaceWidth);
		corners[3] = positions[tempint];

		//y-axis
		for(i=0;i<NumSubdivisions;++i)
		{
			//x-axis
			for(j=0;j<NumSubdivisions;++j)
			{

				tempi = (float)i/(float)(NumSubdivisions-1);
				tempj = (float)j/(float)(NumSubdivisions-1);

				temp1 = linearInterp(corners[0], corners[3], tempi);
				temp2 = linearInterp(corners[1], corners[2], tempi);
				vertices[j][i] = linearInterp(temp1, temp2, tempj);
			}
		}

		for(i=0;i<(NumSubdivisions-1);++i)
		{
			for(j=0;j<(NumSubdivisions-1);++j)
			{
				float tempi = (float)i/(float)(NumSubdivisions-1);
				float tempj = (float)j/(float)(NumSubdivisions-1);

				glColor3f(1.0f, tempi, tempj);

				glVertex3f(((float)x/(float)SurfaceWidth)-0.5f+(tempj/SurfaceWidth),
						   vertices[j][i],
						   ((float)y/(float)SurfaceHeight)-0.5f+(tempi/SurfaceHeight));
				glVertex3f(((float)x/(float)SurfaceWidth)-0.5f+(((float)(j+1)/(float)(NumSubdivisions-1))/SurfaceWidth),
						   vertices[j+1][i],
						   ((float)y/(float)SurfaceHeight)-0.5f+(tempi/SurfaceHeight));
				glVertex3f(((float)x/(float)SurfaceWidth)-0.5f+(((float)(j+1)/(float)(NumSubdivisions-1))/SurfaceWidth),
						   vertices[j+1][i+1],
						   ((float)y/(float)SurfaceHeight)-0.5f+(((float)(i+1)/(float)(NumSubdivisions-1))/SurfaceHeight));
				glVertex3f(((float)x/(float)SurfaceWidth)-0.5f+(tempj/SurfaceWidth),
						   vertices[j][i+1],
						   ((float)y/(float)SurfaceHeight)-0.5f+(((float)(i+1)/(float)(NumSubdivisions-1))/SurfaceHeight));
			}
		}
	};
	///	Helper method to subdivide and draw a quad.
	/*!
		\param x The x position within positions of the lower left vertex.
		\param y The y position within positions of the lower left vertex.
	 */
	inline void bezierSubdivide(int x, int y)
	{
		int i, j;
		float tempi, tempj;
		float temp1, temp2, temp3, temp4;
		int tempX, tempY;
		float vertices[NumSubdivisions+1][NumSubdivisions+1];
		float corners[4][4];

		//Calculate corner positions.
		for(i=-1;i<3;++i)
		{
			for(j=-1;j<3;++j)
			{
				tempX = x + j;
				if(tempX < 0)
					tempX += SurfaceWidth;
				else if(tempX >= SurfaceWidth)
					tempX -= SurfaceWidth;

				tempY = y + i;
				if(tempY < 0)
					tempY += SurfaceHeight;
				else if(tempY >= SurfaceHeight)
					tempY -= SurfaceHeight;

				corners[i+1][j+1] = positions[tempX + (tempY*SurfaceWidth)];
			}
		}

		//y-axis
		for(i=0;i<NumSubdivisions;++i)
		{
			//x-axis
			for(j=0;j<NumSubdivisions;++j)
			{
				tempi = (float)i/(float)(NumSubdivisions-1);
				tempj = (float)j/(float)(NumSubdivisions-1);

				/*temp1 = linearInterp(corners[1][1], corners[2][1], tempi);
				temp2 = linearInterp(corners[1][2], corners[2][2], tempi);*/
				temp1 = bSplineInterp(corners[0][0],
									  corners[0][1],
									  corners[0][2],
									  corners[0][3],
									  tempj);
				temp2 = bSplineInterp(corners[1][0],
									  corners[1][1],
									  corners[1][2],
									  corners[1][3],
									  tempj);
				temp3 = bSplineInterp(corners[2][0],
									  corners[2][1],
									  corners[2][2],
									  corners[2][3],
									  tempj);
				temp4 = bSplineInterp(corners[3][0],
									  corners[3][1],
									  corners[3][2],
									  corners[3][3],
									  tempj);
				vertices[j][i] = bSplineInterp(temp1,
											   temp2,
											   temp3,
											   temp4,
											   tempi);
			}
		}

		for(i=0;i<(NumSubdivisions-1);++i)
		{
			for(j=0;j<(NumSubdivisions-1);++j)
			{
				float tempi = (float)i/(float)(NumSubdivisions-1);
				float tempj = (float)j/(float)(NumSubdivisions-1);

				glColor3f(1.0f, tempi, tempj);

				glVertex3f(((float)x/(float)SurfaceWidth)-0.5f+(tempj/SurfaceWidth),
						   vertices[j][i],
						   ((float)y/(float)SurfaceHeight)-0.5f+(tempi/SurfaceHeight));
				glVertex3f(((float)x/(float)SurfaceWidth)-0.5f+(((float)(j+1)/(float)(NumSubdivisions-1))/SurfaceWidth),
						   vertices[j+1][i],
						   ((float)y/(float)SurfaceHeight)-0.5f+(tempi/SurfaceHeight));
				glVertex3f(((float)x/(float)SurfaceWidth)-0.5f+(((float)(j+1)/(float)(NumSubdivisions-1))/SurfaceWidth),
						   vertices[j+1][i+1],
						   ((float)y/(float)SurfaceHeight)-0.5f+(((float)(i+1)/(float)(NumSubdivisions-1))/SurfaceHeight));
				glVertex3f(((float)x/(float)SurfaceWidth)-0.5f+(tempj/SurfaceWidth),
						   vertices[j][i+1],
						   ((float)y/(float)SurfaceHeight)-0.5f+(((float)(i+1)/(float)(NumSubdivisions-1))/SurfaceHeight));
			}
		}
	};

	///	Some constants.
	enum
	{
		SurfaceWidth = 5,
		SurfaceHeight = 5,
		NumSubdivisions = 6
	};

	///	Force sensor 1 value.
	float force1;
	///	Force sensor 2 value.
	float force2;
	///	Force sensor 3 value.
	float force3;
	///	Force sensor 4 value.
	float force4;
	///	Whether to draw the lines representing the actual surface or not.
	bool drawLines;

	///	Delay line for the surface (i.e. the positions of the cells).
	float positions[SurfaceWidth*SurfaceHeight];
	///	Forces for the string cells.
	float forces[SurfaceWidth*SurfaceHeight];
	///	Velocities for the string cells.
	float velocities[SurfaceWidth*SurfaceHeight];
	///	Mass of the individual cells.
	float mass;
	///	Inverse mass of the individual cells.
	float invMass;
	///	How much to multiply the cells' velocities by (default).
	float velocityMultiplier;
};

}

#endif
