//	PixelArray.cpp - Large array of cubes (pixels) which can display images the
//				     same way monitors do.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "PixelArray.h"
#include "AudioClip.h"
#include "AmbisonicData.h"
#include "Camera.h"
#include "StringParser.h"
#include "FileFinder.h"

#include <iostream>
#include <cassert>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<PixelArray> PixelArray::constructor;

//------------------------------------------------------------------------------
PixelArray::PixelArray():
ParentNode(),
fileSize(0),
index(0),
totalIndex(0),
audioBuffer(0),
bufferSize(0),
lastAtten(0.0f),
samplerate(44100.0f),
currentAmp(0.0f),
image(0),
camDistance(-1.0f)
{
	setType("PixelArray");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
PixelArray::~PixelArray()
{
	if(audioBuffer)
		delete [] audioBuffer;
	if(image)
		SDL_FreeSurface(image);
}

//------------------------------------------------------------------------------
AmbisonicData *PixelArray::getAmbisonicAudio(int numSamples,
											 const ThreeFloats& pos)
{
	int i;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;
	AmbisonicData *retval = getMonoAudio(numSamples);
	const ThreeFloats camPos = Camera::getInstance()->getAudioPosition();
	const Quaternion& camRot = Camera::getInstance()->getAudioRotation();
	//const float rotMult = 6.283185306f/360.0f;
	ThreeFloats adjustedPos = pos;
	ThreeFloats tempCam = Camera::getInstance()->getAudioPosition();

	//So the sound is above the zero point.
	adjustedPos.y += 1.0f;

	camDistance = pos.z - (tempCam.z);

	if(retval)
	{
		//Get current Ambisonic position (see AmbisonicData::operator=()), relative
		//to the camera.

		//So what we do here is:
		//1.) Ensure relativePos is equal to this node's spatial position in the
		//    scene.
		relativePos = adjustedPos;
		//2.) Rotate relativePos by the rotation of the camera, by multiplying
		//    it by the camera's orientation Quaternion.
		relativePos = camRot * relativePos;
		//3.) Translate relativePos by the position of the camera.
		relativePos -= camPos;

		lastAtten = calcRelativeAttenuation(lastPos);
		if(lastPos != relativePos)
		{
			ambiStart = lastPos;
			ambiEnd = relativePos;
			//We increment the position by a constant value, so we don't get
			//clicks or zipper noise.  Of course, there are disadvantages to
			//this approach...
			ambiInc = 0.005f;
			currentAmbi = ambiStart;
			newAtten = calcRelativeAttenuation(relativePos);
		}
		else
		{
			currentAmbi = relativePos;
			newAtten = lastAtten;
		}
		currentAtten = lastAtten;

		//Encode audio.
		for(i=0;i<numSamples;++i)
		{
			//Update currentAmbi and currentAtten if necessary.
			if(ambiInc > 0.0f)
			{
				//Calc currentAmbi.
				currentAmbi = ((1.0f-ambiIndex) * ambiStart) +
							  (ambiIndex * ambiEnd);
				//Calc currentAtten.
				currentAtten = ((1.0f-ambiIndex) * lastAtten) +
							   (ambiIndex * newAtten);

				ambiIndex += ambiInc;

				//Stop updating index if we've reached our destination.
				//I get weird rounding errors here, hence the .0000001f etc...
				if(ambiIndex > 1.0000001f)
				{
					ambiInc = 0.0f;
					currentAtten = newAtten;
				}
			}

			//Convert float sound file to Ambisonic, put it in the buffer.
			retval[i] = currentAmbi * (retval[i].W * currentAtten);
		}
		//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
		if(ambiIndex >= 1.0f)
		{
			lastPos = relativePos;
			lastAtten = currentAtten;
		}
		else
		{
			lastPos = ((1.0f-ambiIndex) * lastPos) + (ambiIndex * relativePos);
			lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
		}
	}

	return retval;
}

//------------------------------------------------------------------------------
AmbisonicData *PixelArray::getMonoAudio(int numSamples)
{
	int i;
	float tempf;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		if(totalIndex < fileSize)
			tempf = currentBlock[static_cast<int>(index)];
		else
			tempf = 0.0f;

		//Zero the buffer.
		audioBuffer[i] = 0.0f;

		//Convert float sound file to Ambisonic, put it in the buffer.
		audioBuffer[i].W += tempf;

		if(totalIndex < fileSize)
		{
			//Update index and totalIndex.
			++totalIndex;
			++index;
			if(totalIndex >= fileSize)
			{
				//Loop back to the start of the file.
				totalIndex = fmod(totalIndex, fileSize);
				index = static_cast<int>(fmod(totalIndex, AudioClip::streamBlockSize));
				currentBlock = audioClip->getNextBlock(this);
			}
			else if(index >= AudioClip::streamBlockSize)
			{
				//Load the next block.
				index = static_cast<int>(fmod(totalIndex, AudioClip::streamBlockSize));
				currentBlock = audioClip->getNextBlock(this);
			}
		}
		//Update envelope.
		if(fabs(tempf) > currentAmp)
			currentAmp += 0.005f;
		else if(currentAmp > 0.0f)
			currentAmp -= 0.00001f;
		else
			currentAmp = 0.0f;
	}

	return audioBuffer;
}

//------------------------------------------------------------------------------
void PixelArray::setAttribute(const string& attribute, EventValue& val)
{
	if(attribute == "maxBack")
		setMaxBackAudioRadius(val);
	else if(attribute == "maxFront")
		setMaxFrontAudioRadius(val);
	else if(attribute == "minBack")
		setMinBackAudioRadius(val);
	else if(attribute == "minFront")
		setMinFrontAudioRadius(val);
	else if(attribute == "image")
	{
		string tempPath;

		tempPath = FileFinder::getInstance().findFile(val);
		if(tempPath != "")
		{
			image = SDL_LoadBMP(tempPath.c_str());
			if(!image)
				cout << "Could not load image: " << tempPath << endl;
			else
				cout << "Loaded image: " << tempPath << endl;
		}
		else
			cout << "PixelArray: " << (string)val << "does not seem to exist?" << endl;
	}
}

//------------------------------------------------------------------------------
void PixelArray::childrenLoaded()
{
	unsigned int i;

	for(i=0;i<(unsigned)getNumChildren();++i)
	{
		if(getChild(i)->getType() == "AudioClip")
		{
			AudioClip *tempAudioClip = dynamic_cast<AudioClip *>(getChild(i));

			if(tempAudioClip->getFileSize())
			{
				//Set audioClip.
				audioClip = tempAudioClip;

				//Set currentBlock and fileSize.
				currentBlock = audioClip->getCurrentBlock();
				fileSize = audioClip->getFileSize();
			}

			break;
		}
	}
}

//------------------------------------------------------------------------------
void PixelArray::parentInitialise()
{
	//Generate the display list.
	pixel = glGenLists(1);
	glNewList(pixel, GL_COMPILE);
		glBegin(GL_QUADS);
			//Front face.
			glVertex3f(-0.5f, -0.5f, 0.5f);
			glVertex3f(0.5f, -0.5f, 0.5f);
			glVertex3f(0.5f, 0.5f, 0.5f);
			glVertex3f(-0.5f, 0.5f, 0.5f);

			//Rear face.
			glVertex3f(-0.5f, -0.5f, -0.5f);
			glVertex3f(-0.5f,  0.5f, -0.5f);
			glVertex3f(0.5f,  0.5f, -0.5f);
			glVertex3f(0.5f, -0.5f, -0.5f);

			// Top Face
			glVertex3f(-0.5f, 0.5f, -0.5f);
			glVertex3f(-0.5f, 0.5f, 0.5f);
			glVertex3f(0.5f, 0.5f, 0.5f);
			glVertex3f(0.5f, 0.5f, -0.5f);
		
			// Bottom Face
			glVertex3f(-0.5f, -0.5f, -0.5f);
			glVertex3f(0.5f, -0.5f, -0.5f);
			glVertex3f(0.5f, -0.5f, 0.5f);
			glVertex3f(-0.5f, -0.5f, 0.5f);

			// Right face
			glVertex3f(0.5f, -0.5f, -0.5f);
			glVertex3f(0.5f, 0.5f, -0.5f);
			glVertex3f(0.5f, 0.5f, 0.5f);
			glVertex3f(0.5f, -0.5f, 0.5f);

			// Left Face
			glVertex3f(-0.5f, -0.5f, -0.5f);
			glVertex3f(-0.5f, -0.5f, 0.5f);
			glVertex3f(-0.5f, 0.5f, 0.5f);
			glVertex3f(-0.5f, 0.5f, -0.5f);
		glEnd();
	glEndList();

}

//------------------------------------------------------------------------------
void PixelArray::samplerateChanged(float newRate)
{
	samplerate = newRate;
}

//------------------------------------------------------------------------------
void PixelArray::preChildDraw()
{
	int x, y;
	Uint8 index = 0;
	Uint8 *imageThing = 0;
	SDL_Color colTemp;
	SDL_PixelFormat *fmt = 0;
	float pixDist = 1.0f;

	//If the pixel array's behind us, we can ignore it.
	if(camDistance < 2.0f)
	{
		if(camDistance > -90.0f)
			pixDist = (1.0f + (2.0f*(1.0f-(camDistance/-90.0f))));

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();

		glDisable(GL_LIGHTING);
		glColor3f(0.0f, 0.0, 0.0f);
		if(image)
		{
			fmt = image->format;
			SDL_LockSurface(image);
			imageThing = (Uint8 *)(image->pixels);
		}
		for(y=0;y<ArrayHeight;++y)
		{
			for(x=0;x<ArrayWidth;++x)
			{
				glPushMatrix();
	
				if(image)
				{
					index = static_cast<Uint8>(static_cast<float>(imageThing[(y*ArrayWidth)+x])*currentAmp);
					colTemp = fmt->palette->colors[index];
					glColor3ub(colTemp.r, colTemp.g, colTemp.b);
				}
				glTranslatef(((-(ArrayWidth/2)+x)*pixDist),
							 (((ArrayHeight/2)-y)*pixDist),
							 0.0f);
				glCallList(pixel);
	
				glPopMatrix();
			}
		}
		if(image)
			SDL_UnlockSurface(image);
		glEnable(GL_LIGHTING);
	}
}

//------------------------------------------------------------------------------
void PixelArray::postChildDraw()
{
	const GLfloat diffuse[]={1.0f, 1.0f, 1.0f};
	const GLfloat ambient[]={0.0f, 0.0f, 0.0f};
	const GLfloat emissive[]={0.0f, 0.0f, 0.0f};
	const GLfloat specular[]={0.5f, 0.5f, 0.5f};
	const GLfloat shininess = 0.0f;

	//Not sure if there's a better way to do this?
	glBindTexture(GL_TEXTURE_2D, 0);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
	glMaterialfv(GL_FRONT, GL_AMBIENT, ambient);
	glMaterialfv(GL_FRONT, GL_EMISSION, emissive);
	glMaterialf(GL_FRONT, GL_SHININESS, shininess); 
	glMaterialfv(GL_FRONT, GL_SPECULAR, specular);

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

//------------------------------------------------------------------------------
/*AmbisonicData *PixelArray::preChildAudio(int numSamples,
										 const ThreeFloats& newPos)
{
	int i;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;
	float tempf;
	ThreeFloats adjustedPos = newPos;
	ThreeFloats tempCam = Camera::getInstance()->getAudioPosition();

	//So the sound is above the zero point.
	adjustedPos.y += 1.0f;

	camDistance = newPos.z - (tempCam.z);

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Get current Ambisonic position (see AmbisonicData::operator=()), relative
	//to the camera.
	relativePos = adjustedPos;
	relativePos -= tempCam;
	lastAtten = calcRelativeAttenuation(lastPos);
	if(lastPos != relativePos)
	{
		ambiStart = lastPos;
		ambiEnd = relativePos;
		//ambiInc = 1.0f/numSamples;
		//We increment the position by a constant value, so we don't get clicks
		//or zipper noise.  Of course, there are disadvantages to this
		//approach...
		ambiInc = 0.005f;
		currentAmbi = ambiStart;
		newAtten = calcRelativeAttenuation(relativePos);
	}
	else
		currentAmbi = relativePos;
	currentAtten = lastAtten;

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		if(totalIndex < fileSize)
			tempf = currentBlock[static_cast<int>(index)];
		else
			tempf = 0.0f;

		//Update currentAmbi and currentAtten if necessary.
		if(ambiInc > 0.0f)
		{
			//Calc currentAmbi.
			currentAmbi = ((1.0f-ambiIndex) * ambiStart) +
						  (ambiIndex * ambiEnd);
			//Calc currentAtten.
			currentAtten = ((1.0f-ambiIndex) * lastAtten) +
						   (ambiIndex * newAtten);

			ambiIndex += ambiInc;

			//Stop updating index if we've reached our destination.
			//I get weird rounding errors here, hence the .0000001f etc...
			if(ambiIndex > 1.0000001f)
			{
				ambiInc = 0.0f;
				currentAtten = newAtten;
			}
		}
		//Zero the buffer.
		audioBuffer[i] = 0.0f;

		//Convert float sound file to Ambisonic, put it in the buffer.
		audioBuffer[i] += currentAmbi * (tempf * currentAtten);

		if(totalIndex < fileSize)
		{
			//Update index and totalIndex.
			++totalIndex;
			++index;
			if(totalIndex >= fileSize)
			{
				//Loop back to the start of the file.
				totalIndex = fmod(totalIndex, fileSize);
				index = static_cast<int>(fmod(totalIndex, AudioClip::streamBlockSize));
				currentBlock = audioClip->getNextBlock(this);
			}
			else if(index >= AudioClip::streamBlockSize)
			{
				//Load the next block.
				index = static_cast<int>(fmod(totalIndex, AudioClip::streamBlockSize));
				currentBlock = audioClip->getNextBlock(this);
			}
		}
		//Update envelope.
		if(fabs(tempf) > currentAmp)
			currentAmp += 0.005f;
		else if(currentAmp > 0.0f)
			currentAmp -= 0.00001f;
		else
			currentAmp = 0.0f;
	}
	//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
	if(ambiIndex >= 1.0f)
	{
		lastPos = relativePos;
		lastAtten = currentAtten;
	}
	else
	{
		lastPos = ((1.0f-ambiIndex) * lastPos) + (ambiIndex * relativePos);
		lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
	}

	return audioBuffer;
}*/

//------------------------------------------------------------------------------
void PixelArray::audioBlockSizeChanged(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}
	
}
