//	PixelArray.h - Large array of cubes (pixels) which can display images the
//				   same way monitors do.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PIXELARRAY_H_
#define PIXELARRAY_H_

#include "ParentNode.h"
#include "AmbisonicData.h"

#include "SDL.h"
#include "GLHeaders.h"
#include <vector>

/*+
	Name: PixelArray
	Type: ParentNode, to be used with a child AudioClip node.
	Description: Displays a bitmap image (must be size 160x120 and 8-bit with a
			 	 colour palette!!) as cubes representing the pixels.  The
				 amplitude envelope of the audio is used to basically cycle
				 through the colour palette, creating a jittery kind of motion.
				 When the camera gets close, the distance between the cubes
				 increases, opening up gaps between the pixels.
	Attribute:	maxBack
	AttType:	SFFloat (input)
	AttDesc:	The same as Sound's maxBack attribute.

	Attribute:	maxFront
	AttType:	SFFloat (input)
	AttDesc:	The same as Sound's maxFront attribute.

	Attribute:	minBack
	AttType:	SFFloat (input)
	AttDesc:	The same as Sound's minBack attribute.

	Attribute:	minFront
	AttType:	SFFloat (input)
	AttDesc:	The same as Sound's minFront attribute.

	Attribute:	image
	AttType:	SFString (input)
	AttDesc:	Url to the .bmp image to be used.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="-0.5 0.5 -140.0 "&gt;
			&lt;PixelArray image="DarkBlueRadial.bmp" minFront="150" maxFront="300" minBack="150" maxBack="300"&gt;
				&lt;AudioClip url="/home/niall/music/temp/savedWavs/fadeInDroneL.wav"/&gt;
			&lt;/PixelArray&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

class AudioClip;

///	Large array of cubes (pixels) which can display images the same way monitors do.
/*!
	\ingroup libheilanextras
 */
class PixelArray : public ParentNode
{
  public:
	///	Constructor.
	PixelArray();
	///	Destructor.
	~PixelArray();

	///	Overridden so we can get the position relative to the camera.
	AmbisonicData *getAmbisonicAudio(int numSamples, const ThreeFloats& pos);
	///	Generates the audio for this node.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Used to set this node's attributes.
	/*!
		Also used by child AudioClip nodes to communicate that they've loaded
		their sound file, and it's ready for streaming.
	 */
	void setAttribute(const std::string& attribute, EventValue& val);
	///	Used to load the AudioClip.
	void childrenLoaded();

	static NodeConstructor<PixelArray> constructor;
  private:
	///	Initialises various bits and pieces.
	void parentInitialise();

	///	Called so the node knows what samplerate it's running at.
	void samplerateChanged(float newRate);

	///	Called before the node starts drawing its children.
	void preChildDraw();
	///	Called after the node has finished drawing its children.
	void postChildDraw();
	///	Called before the node starts retrieving audio from it's children.
	//AmbisonicData *preChildAudio(int numSamples, const ThreeFloats& newPos);
	///	Called when the soundcard block size changes.
	void audioBlockSizeChanged(int newSize);

	///	The size of the sound file we're streaming from.
	/*!
		We assume all 4 AudioClips are the same size.
	 */
	float fileSize;

	///	Index we're currently reading from the current block.
	int index;
	///	Index we're reading from the overall sound file.
	float totalIndex;

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;

	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;

	///	The current samplerate.
	float samplerate;

	///	The current amplitude of the attached AudioClip.
	float currentAmp;

	///	The AudioClip used.
	AudioClip *audioClip;
	///	The current audio blocks we're reading from for each AudioClip.
	float *currentBlock;

	///	Storage for one 'pixel'.
	GLuint pixel;
	///	The image we're drawing.
	SDL_Surface *image;

	///	The z-distance from the pixel array to the camera.
	/*!
		-ve means the pixel array's in front of the camera, +ve means it's
		behind it, and we don't have to draw it.
	 */
	float camDistance;

	///	Enum containing the dimensions of the pixel array.
	enum
	{
		ArrayWidth = 160,
		ArrayHeight = 120
	};
};

}

#endif
