//	PulseCubes.cpp - A cube which fires off smaller cubes when it detects
//				   	 transients.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "PulseCubes.h"
#include "AudioClip.h"
#include "AmbisonicData.h"
#include "Camera.h"
#include "StringParser.h"

#include "GLHeaders.h"
#include <cassert>
#include <cmath>
#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<PulseCubes> PulseCubes::constructor;

//------------------------------------------------------------------------------
PulseCubes::PulseCubes():
ParentNode(),
audioClip(0),
currentBlock(0),
fileSize(0),
index(0),
totalIndex(0),
audioBuffer(0),
bufferSize(0),
lastAtten(0.0f),
samplerate(44100.0f),
currentAmp(0.0f),
pulseCount(0),
pulseCount2(0),
waitNum(0),
waitPoint(0.0f),
waitFraction(0.0f),
generatePulses(true),
currentPulse(1)
{
	setType("PulseCubes");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
PulseCubes::~PulseCubes()
{
	if(audioBuffer)
		delete [] audioBuffer;
}

//------------------------------------------------------------------------------
AmbisonicData *PulseCubes::getAmbisonicAudio(int numSamples,
											 const ThreeFloats& pos)
{
	int i, j;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;
	const ThreeFloats camPos = Camera::getInstance()->getAudioPosition();
	const Quaternion& camRot = Camera::getInstance()->getAudioRotation();
	//const float rotMult = 6.283185306f/360.0f;
	float tempf;
	AmbisonicData *tempPulse;

	//Skip everything if we don't have an AudioClip.
	if(!audioClip)
		return 0;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Get current Ambisonic position (see AmbisonicData::operator=()), relative
	//to the camera.

	//So what we do here is:
	//1.) Ensure relativePos is equal to this node's spatial position in the
	//    scene.
	relativePos = pos;
	//2.) Rotate relativePos by the rotation of the camera, by multiplying
	//    it by the camera's orientation Quaternion.
	relativePos = camRot * relativePos;
	//3.) Translate relativePos by the position of the camera.
	relativePos -= camPos;

	lastAtten = calcRelativeAttenuation(lastPos);
	if(lastPos != relativePos)
	{
		ambiStart = lastPos;
		ambiEnd = relativePos;
		//ambiInc = 1.0f/numSamples;
		//We increment the position by a constant value, so we don't get clicks
		//or zipper noise.  Of course, there are disadvantages to this
		//approach...
		ambiInc = 0.005f;
		currentAmbi = ambiStart;
		newAtten = calcRelativeAttenuation(relativePos);
	}
	else
		currentAmbi = relativePos;
	currentAtten = lastAtten;

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		if(totalIndex < fileSize)
		{
			tempf = currentBlock[static_cast<int>(index)];

			//Update index and totalIndex.
			++totalIndex;
			++index;
			if(totalIndex >= fileSize)
			{
				//Loop back to the start of the file.
				totalIndex = fmod(totalIndex, fileSize);
				index = static_cast<int>(fmod(totalIndex, AudioClip::streamBlockSize));
				currentBlock = audioClip->getNextBlock(this);
			}
			else if(index >= AudioClip::streamBlockSize)
			{
				//Load the next block.
				index = static_cast<int>(fmod(totalIndex, AudioClip::streamBlockSize));
				currentBlock = audioClip->getNextBlock(this);
			}
		}
		else
			tempf = 0.0f;

		//Update currentAmbi and currentAtten if necessary.
		if(ambiInc > 0.0f)
		{
			//Calc currentAmbi.
			currentAmbi = ((1.0f-ambiIndex) * ambiStart) +
						  (ambiIndex * ambiEnd);
			//Calc currentAtten.
			currentAtten = ((1.0f-ambiIndex) * lastAtten) +
						   (ambiIndex * newAtten);

			ambiIndex += ambiInc;

			//Stop updating index if we've reached our destination.
			//I get weird rounding errors here, hence the .0000001f etc...
			if(ambiIndex > 1.0000001f)
			{
				ambiInc = 0.0f;
				currentAtten = newAtten;
			}
		}

		//Convert float sound file to Ambisonic, put it in the buffer.
		audioBuffer[i] = currentAmbi * (tempf * currentAtten);

		//Update envelope.
		if((fabs(tempf)-currentAmp) > 0.2f)
		{
			currentAmp = fabs(tempf);
			if(waitFraction >= waitPoint)
			{
				if(pulseCount < 1)
				{
					if(pulseCount2 == 1)
					{
						if(currentPulse > waitNum)
						{
							for(j=0;j<12;++j)
								pulses[j].trigger();
						}
						else
							++currentPulse;
						pulseCount2 = 0;
					}
					else
						pulseCount2 = 1;
					pulseCount = 20500;
				}
			}
		}
		else if(fabs(tempf) > currentAmp)
			//currentAmp = tempf;
			currentAmp += 0.0001f;
		else if(currentAmp > 0.0f)
			currentAmp -= 0.00001f;
		else
			currentAmp = 0.0f;

		if(pulseCount > 0)
			--pulseCount;
	}
	//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
	if(ambiIndex >= 1.0f)
	{
		lastPos = relativePos;
		lastAtten = currentAtten;
	}
	else
	{
		lastPos = ((1.0f-ambiIndex) * lastPos) + (ambiIndex * relativePos);
		lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
	}

	//Render pulses' audio.
	for(i=0;i<12;++i)
	{
		tempPulse = pulses[i].getAudioBlock(numSamples, pos);
		if(tempPulse)
		{
			for(j=0;j<numSamples;++j)
				audioBuffer[j] += tempPulse[j];
		}
	}

	return audioBuffer;
}

//------------------------------------------------------------------------------
AmbisonicData *PulseCubes::getMonoAudio(int numSamples)
{
	//We do everything in getAmbisonicAudio(), since it was easier to do that
	//than convert the old code properly to the new scheme.
	return 0;
}

//------------------------------------------------------------------------------
void PulseCubes::setAttribute(const string& attribute, EventValue& val)
{
	if(attribute == "maxBack")
		setMaxBackAudioRadius(val);
	else if(attribute == "maxFront")
		setMaxFrontAudioRadius(val);
	else if(attribute == "minBack")
		setMinBackAudioRadius(val);
	else if(attribute == "minFront")
		setMinFrontAudioRadius(val);
	else if(attribute == "pulseWait")
	{
		waitNum = val;
		if(waitNum > 0)
			currentPulse = 0;
	}
	else if(attribute == "waitPoint")
	{
		waitPoint = val;
		if((waitPoint > 0.0f) && (waitFraction <= waitPoint))
			generatePulses = false;
		else
			generatePulses = true;
	}
	else if(attribute == "waitFraction")
	{
		waitFraction = val;
		if(waitFraction > waitPoint)
			generatePulses = true;
	}
}

//------------------------------------------------------------------------------
void PulseCubes::childrenLoaded()
{
	unsigned int i;

	for(i=0;i<(unsigned)getNumChildren();++i)
	{
		if(getChild(i)->getType() == "AudioClip")
		{
			AudioClip *tempAudioClip = dynamic_cast<AudioClip *>(getChild(i));

			if(tempAudioClip->getFileSize())
			{
				//Set audioClip.
				audioClip = tempAudioClip;

				//Set currentBlock and fileSize.
				currentBlock = audioClip->getCurrentBlock();
				fileSize = audioClip->getFileSize();
			}

			break;
		}
	}
}

//------------------------------------------------------------------------------
void PulseCubes::parentInitialise()
{
	ThreeFloats tempInc;
	float tempf = 0.15f;

	tempInc.x = -tempf;
	tempInc.y = tempf;
	tempInc.z = -tempf;
	pulses[0].setPositionIncrement(tempInc);
	tempInc.x = tempf;
	tempInc.y = tempf;
	tempInc.z = -tempf;
	pulses[1].setPositionIncrement(tempInc);
	tempInc.x = -tempf;
	tempInc.y = -tempf;
	tempInc.z = -tempf;
	pulses[2].setPositionIncrement(tempInc);
	tempInc.x = tempf;
	tempInc.y = -tempf;
	tempInc.z = -tempf;
	pulses[3].setPositionIncrement(tempInc);

	tempInc.x = -tempf;
	tempInc.y = tempf;
	tempInc.z = tempf;
	pulses[4].setPositionIncrement(tempInc);
	tempInc.x = tempf;
	tempInc.y = tempf;
	tempInc.z = tempf;
	pulses[5].setPositionIncrement(tempInc);
	tempInc.x = -tempf;
	tempInc.y = -tempf;
	tempInc.z = tempf;
	pulses[6].setPositionIncrement(tempInc);
	tempInc.x = tempf;
	tempInc.y = -tempf;
	tempInc.z = tempf;
	pulses[7].setPositionIncrement(tempInc);

	tempInc.x = tempf;
	tempInc.y = 0.0f;
	tempInc.z = 0.0f;
	pulses[8].setPositionIncrement(tempInc);
	tempInc.x = -tempf;
	tempInc.y = 0.0f;
	tempInc.z = 0.0f;
	pulses[9].setPositionIncrement(tempInc);
	tempInc.x = 0.0f;
	tempInc.y = tempf;
	tempInc.z = 0.0f;
	pulses[10].setPositionIncrement(tempInc);
	tempInc.x = 0.0f;
	tempInc.y = -tempf;
	tempInc.z = 0.0f;
	pulses[11].setPositionIncrement(tempInc);
}

//------------------------------------------------------------------------------
void PulseCubes::samplerateChanged(float newRate)
{
	int i;

	samplerate = newRate;
	for(i=0;i<12;++i)
		pulses[i].setSamplerate(samplerate);
}

//------------------------------------------------------------------------------
void PulseCubes::preChildDraw()
{
	
}

//------------------------------------------------------------------------------
void PulseCubes::postChildDraw()
{
	int i;
	const GLfloat diffuse[]={1.0f, 1.0f, 1.0f};
	const GLfloat ambient[]={0.0f, 0.0f, 0.0f};
	const GLfloat emissive[]={0.0f, 0.0f, 0.0f};
	const GLfloat specular[]={0.5f, 0.5f, 0.5f};
	const GLfloat shininess = 0.0f;

	for(i=0;i<12;++i)
		pulses[i].renderCube();

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glScalef(currentAmp, currentAmp, currentAmp);
	glRotatef(45.0f, 1.0f, 1.0f, 0.0f);

	if(currentAmp > 0.0f)
	{
		glBegin(GL_QUADS);
			//Front face.
			glNormal3f(0.0f, 0.0f, 1.0f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-0.5f, -0.5f, 0.5f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(0.5f, -0.5f, 0.5f);
			glTexCoord2f(0.5f, 0.5f);
			glVertex3f(0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.0f, 0.5f);
			glVertex3f(-0.5f, 0.5f, 0.5f);
	
			//Rear face.
			glNormal3f(0.0f, 0.0f, -1.0f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(-0.5f, -0.5f, -0.5f);
			glTexCoord2f(0.5f, 0.5f);
			glVertex3f(-0.5f,  0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.5f);
			glVertex3f(0.5f,  0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(0.5f, -0.5f, -0.5f);
	
			// Top Face
			glNormal3f(0.0f, 1.0f, 0.0f);
			glTexCoord2f(0.0f, 0.5f);
			glVertex3f(-0.5f, 0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.5f, 0.5f);
			glVertex3f(0.5f, 0.5f, -0.5f);
		
			// Bottom Face
			glNormal3f(0.0f, -1.0f, 0.0f);
			glTexCoord2f(0.5f, 0.5f);
			glVertex3f(-0.5f, -0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.5f);
			glVertex3f(0.5f, -0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(0.5f, -0.5f, 0.5f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(-0.5f, -0.5f, 0.5f);
	
			// Right face
			glNormal3f(1.0f, 0.0f, 0.0f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(0.5f, -0.5f, -0.5f);
			glTexCoord2f(0.5f, 0.5f);
			glVertex3f(0.5f, 0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.5f);
			glVertex3f(0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(0.5f, -0.5f, 0.5f);
	
			// Left Face
			glNormal3f(-1.0f, 0.0f, 0.0f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-0.5f, -0.5f, -0.5f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(-0.5f, -0.5f, 0.5f);
			glTexCoord2f(0.5f, 0.5f);
			glVertex3f(-0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.0f, 0.5f);
			glVertex3f(-0.5f, 0.5f, -0.5f);
		glEnd();
	}

	glPopMatrix();

	//Not sure if there's a better way to do this?
	glBindTexture(GL_TEXTURE_2D, 0);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
	glMaterialfv(GL_FRONT, GL_AMBIENT, ambient);
	glMaterialfv(GL_FRONT, GL_EMISSION, emissive);
	glMaterialf(GL_FRONT, GL_SHININESS, shininess); 
	glMaterialfv(GL_FRONT, GL_SPECULAR, specular);

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

//------------------------------------------------------------------------------
/*AmbisonicData *PulseCubes::preChildAudio(int numSamples,
										 const ThreeFloats& newPos)
{
	int i, j;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;
	float tempf;
	AmbisonicData *tempPulse;

	//Skip everything if we don't have an AudioClip.
	if(!audioClip)
		return 0;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Get current Ambisonic position (see AmbisonicData::operator=()), relative
	//to the camera.
	relativePos = newPos;
	relativePos -= Camera::getInstance()->getAudioPosition();
	lastAtten = calcRelativeAttenuation(lastPos);
	if(lastPos != relativePos)
	{
		ambiStart = lastPos;
		ambiEnd = relativePos;
		//ambiInc = 1.0f/numSamples;
		//We increment the position by a constant value, so we don't get clicks
		//or zipper noise.  Of course, there are disadvantages to this
		//approach...
		ambiInc = 0.005f;
		currentAmbi = ambiStart;
		newAtten = calcRelativeAttenuation(relativePos);
	}
	else
		currentAmbi = relativePos;
	currentAtten = lastAtten;

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		if(totalIndex < fileSize)
		{
			tempf = currentBlock[static_cast<int>(index)];

			//Update index and totalIndex.
			++totalIndex;
			++index;
			if(totalIndex >= fileSize)
			{
				//Loop back to the start of the file.
				totalIndex = fmod(totalIndex, fileSize);
				index = static_cast<int>(fmod(totalIndex, AudioClip::streamBlockSize));
				currentBlock = audioClip->getNextBlock(this);
			}
			else if(index >= AudioClip::streamBlockSize)
			{
				//Load the next block.
				index = static_cast<int>(fmod(totalIndex, AudioClip::streamBlockSize));
				currentBlock = audioClip->getNextBlock(this);
			}
		}
		else
			tempf = 0.0f;

		//Update currentAmbi and currentAtten if necessary.
		if(ambiInc > 0.0f)
		{
			//Calc currentAmbi.
			currentAmbi = ((1.0f-ambiIndex) * ambiStart) +
						  (ambiIndex * ambiEnd);
			//Calc currentAtten.
			currentAtten = ((1.0f-ambiIndex) * lastAtten) +
						   (ambiIndex * newAtten);

			ambiIndex += ambiInc;

			//Stop updating index if we've reached our destination.
			//I get weird rounding errors here, hence the .0000001f etc...
			if(ambiIndex > 1.0000001f)
			{
				ambiInc = 0.0f;
				currentAtten = newAtten;
			}
		}

		//Convert float sound file to Ambisonic, put it in the buffer.
		audioBuffer[i] = currentAmbi * (tempf * currentAtten);

		//Update envelope.
		if((fabs(tempf)-currentAmp) > 0.2f)
		{
			currentAmp = fabs(tempf);
			if(waitFraction >= waitPoint)
			{
				if(pulseCount < 1)
				{
					if(pulseCount2 == 1)
					{
						if(currentPulse > waitNum)
						{
							for(j=0;j<12;++j)
								pulses[j].trigger();
						}
						else
							++currentPulse;
						pulseCount2 = 0;
					}
					else
						pulseCount2 = 1;
					pulseCount = 20500;
				}
			}
		}
		else if(fabs(tempf) > currentAmp)
			//currentAmp = tempf;
			currentAmp += 0.0001f;
		else if(currentAmp > 0.0f)
			currentAmp -= 0.00001f;
		else
			currentAmp = 0.0f;

		if(pulseCount > 0)
			--pulseCount;
	}
	//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
	if(ambiIndex >= 1.0f)
	{
		lastPos = relativePos;
		lastAtten = currentAtten;
	}
	else
	{
		lastPos = ((1.0f-ambiIndex) * lastPos) + (ambiIndex * relativePos);
		lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
	}

	//Render pulses' audio.
	for(i=0;i<12;++i)
	{
		tempPulse = pulses[i].getAudioBlock(numSamples, newPos);
		if(tempPulse)
		{
			for(j=0;j<numSamples;++j)
				audioBuffer[j] += tempPulse[j];
		}
	}

	return audioBuffer;
}*/

//------------------------------------------------------------------------------
void PulseCubes::audioBlockSizeChanged(int newSize)
{
	int i;

	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];

	for(i=0;i<12;++i)
		pulses[i].setAudioBlockSize(bufferSize);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
PulseCubes::Pulse::Pulse():
audioBuffer(0),
bufferSize(0),
currentSize(0.0f),
samplerate(44100.0f),
rampPos(-1.0f),
rampAmp(1.0f),
rampAmpInc(0)
{
	rampFreq = static_cast<float>(rand())/static_cast<float>(RAND_MAX);
	rampFreq *= 5.0f;
	//rampFreq += 217.5f;
	rampFreq = 220.0f;

	rampInc = (rampFreq*0.5f)/samplerate;

	cutoff = 1760.0f;
	f = 2.0f * sinf(3.1415926535897932384626433832795f*cutoff/samplerate);
	q = 0.1f;
	low = high = band = 0.0f;
}

//------------------------------------------------------------------------------
PulseCubes::Pulse::~Pulse()
{
	if(audioBuffer)
		delete [] audioBuffer;
}

//------------------------------------------------------------------------------
void PulseCubes::Pulse::trigger()
{
	//if(currentSize <= 0.0f)
	//{
		currentSize = 1.0f;
		currentPosition = 0.0f;
	//}
}

//------------------------------------------------------------------------------
AmbisonicData *PulseCubes::Pulse::getAudioBlock(int numSamples,
												const ThreeFloats& newPos)
{
	int i;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;

	//Return if the pulse is not currently active.
	if(currentSize <= 0.0f)
		return 0;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Get current Ambisonic position (see AmbisonicData::operator=()), relative
	//to the camera.
	relativePos = newPos;
	relativePos += currentPosition;
	relativePos -= Camera::getInstance()->getAudioPosition();
	lastAtten = calcRelativeAttenuation(lastPos);
	if(lastPos != relativePos)
	{
		ambiStart = lastPos;
		ambiEnd = relativePos;
		//ambiInc = 1.0f/numSamples;
		//We increment the position by a constant value, so we don't get clicks
		//or zipper noise.
		ambiInc = 0.005f;
		currentAmbi = ambiStart;
		newAtten = calcRelativeAttenuation(relativePos);
	}
	else
		currentAmbi = relativePos;
	currentAtten = lastAtten;

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		rampPos += rampInc;
		if(rampPos > 0.25f)
		{
			rampPos = -0.25f;
			++rampAmpInc;
			if(rampAmpInc > 15)
			{
				rampAmp = static_cast<float>(rand())/static_cast<float>(RAND_MAX);
				if(rampAmp > 0.5f)
					rampAmp = 1.0f;
				else
					rampAmp = 0.0f;
				/*rampAmp = 1.0f - rampAmp;*/
				rampAmpInc = 0;
			}
		}

		//Calculate filter.
		low = low + (f * band);
		high = (q * rampPos) - low - (q * band);
		band = (f * high) + band;
		notch = high + low;

		//Update currentAmbi and currentAtten if necessary.
		if(ambiInc > 0.0f)
		{
			//Calc currentAmbi.
			currentAmbi = ((1.0f-ambiIndex) * ambiStart) +
						  (ambiIndex * ambiEnd);
			//Calc currentAtten.
			currentAtten = ((1.0f-ambiIndex) * lastAtten) +
						   (ambiIndex * newAtten);

			ambiIndex += ambiInc;

			//Stop updating index if we've reached our destination.
			//I get weird rounding errors here, hence the .0000001f...
			if(ambiIndex > 1.0000001f)
			{
				ambiInc = 0.0f;
				currentAtten = newAtten;
			}
		}

		//Convert float audio to Ambisonic, put it in the buffer.
		audioBuffer[i] = currentAmbi * (high * 3.0f * rampAmp * currentSize * currentAtten);
	}
	//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
	if(ambiIndex >= 1.0f)
	{
		lastPos = relativePos;
		lastAtten = currentAtten;
	}
	else
	{
		lastPos = ((1.0f-ambiIndex) * lastPos) + (ambiIndex * relativePos);
		lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
	}

	return audioBuffer;
}

//------------------------------------------------------------------------------
void PulseCubes::Pulse::renderCube()
{
	if(currentSize <= 0.0f)
		return;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glTranslatef(currentPosition.x, currentPosition.y, currentPosition.z);
	glRotatef(45.0f, increment.x, increment.y, increment.z);
	glScalef(currentSize*0.5f, currentSize*0.5f, currentSize*0.5f);
	currentPosition += increment;
	currentSize -= 0.01f;

	glBegin(GL_QUADS);
		//Front face.
		glNormal3f(0.0f, 0.0f, 1.0f);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(-0.5f, -0.5f, 0.5f);
		glTexCoord2f(0.5f, 0.0f);
		glVertex3f(0.5f, -0.5f, 0.5f);
		glTexCoord2f(0.5f, 0.5f);
		glVertex3f(0.5f, 0.5f, 0.5f);
		glTexCoord2f(0.0f, 0.5f);
		glVertex3f(-0.5f, 0.5f, 0.5f);
	
		//Rear face.
		glNormal3f(0.0f, 0.0f, -1.0f);
		glTexCoord2f(0.5f, 0.0f);
		glVertex3f(-0.5f, -0.5f, -0.5f);
		glTexCoord2f(0.5f, 0.5f);
		glVertex3f(-0.5f,  0.5f, -0.5f);
		glTexCoord2f(0.0f, 0.5f);
		glVertex3f(0.5f,  0.5f, -0.5f);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(0.5f, -0.5f, -0.5f);
	
		// Top Face
		glNormal3f(0.0f, 1.0f, 0.0f);
		glTexCoord2f(0.0f, 0.5f);
		glVertex3f(-0.5f, 0.5f, -0.5f);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(-0.5f, 0.5f, 0.5f);
		glTexCoord2f(0.5f, 0.0f);
		glVertex3f(0.5f, 0.5f, 0.5f);
		glTexCoord2f(0.5f, 0.5f);
		glVertex3f(0.5f, 0.5f, -0.5f);
		
		// Bottom Face
		glNormal3f(0.0f, -1.0f, 0.0f);
		glTexCoord2f(0.5f, 0.5f);
		glVertex3f(-0.5f, -0.5f, -0.5f);
		glTexCoord2f(0.0f, 0.5f);
		glVertex3f(0.5f, -0.5f, -0.5f);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(0.5f, -0.5f, 0.5f);
		glTexCoord2f(0.5f, 0.0f);
		glVertex3f(-0.5f, -0.5f, 0.5f);
	
		// Right face
		glNormal3f(1.0f, 0.0f, 0.0f);
		glTexCoord2f(0.5f, 0.0f);
		glVertex3f(0.5f, -0.5f, -0.5f);
		glTexCoord2f(0.5f, 0.5f);
		glVertex3f(0.5f, 0.5f, -0.5f);
		glTexCoord2f(0.0f, 0.5f);
		glVertex3f(0.5f, 0.5f, 0.5f);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(0.5f, -0.5f, 0.5f);
	
		// Left Face
		glNormal3f(-1.0f, 0.0f, 0.0f);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(-0.5f, -0.5f, -0.5f);
		glTexCoord2f(0.5f, 0.0f);
		glVertex3f(-0.5f, -0.5f, 0.5f);
		glTexCoord2f(0.5f, 0.5f);
		glVertex3f(-0.5f, 0.5f, 0.5f);
		glTexCoord2f(0.0f, 0.5f);
		glVertex3f(-0.5f, 0.5f, -0.5f);
	glEnd();

	glPopMatrix();
}

//------------------------------------------------------------------------------
void PulseCubes::Pulse::setPositionIncrement(const ThreeFloats& inc)
{
	increment = inc;
}

//------------------------------------------------------------------------------
void PulseCubes::Pulse::setSamplerate(float val)
{
	samplerate = val;
	rampInc = (rampFreq*0.5f)/samplerate;
}

//------------------------------------------------------------------------------
void PulseCubes::Pulse::setAudioBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}

//------------------------------------------------------------------------------
inline float PulseCubes::Pulse::calcRelativeAttenuation(const ThreeFloats& relPos)
{
	float mindist, camdist, attdist, dropoffdist;
	float retval;

	mindist = 3.0f;
	camdist = sqrtf((relPos.x * relPos.x)+
				    (relPos.y * relPos.y)+
				    (relPos.z * relPos.z));
	attdist = camdist - 3.0f;
	//dropoffdist = 14.0f - 3.0f; //What it should be.
	dropoffdist = 1.0f/11.0f; //Should be better optimised.

	if(camdist < mindist)
		retval = 1.0f;
	else
	{
		retval = -20.0f * (attdist*dropoffdist);
		retval = powf(10.0f, (retval*0.1f));
	}

	return retval;
}
	
}
