//	PulseCubes.h - A cube which fires off smaller cubes when it detects
//				   transients.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PULSECUBES_H_
#define PULSECUBES_H_

#include "ParentNode.h"

/*+
	Name: PulseCubes
	Type: ParentNode, to be used with a child AudioClip node.
	Description: A cube (the size of which is controlled by the amplitude
				 envelope of the accompanying AudioClip) which fires off smaller
				 cubes when it detects transients in the AudioClip's audio.
	Attribute:	maxBack
	AttType:	SFFloat (input)
	AttDesc:	The same as Sound's maxBack attribute.

	Attribute:	maxFront
	AttType:	SFFloat (input)
	AttDesc:	The same as Sound's maxFront attribute.

	Attribute:	minBack
	AttType:	SFFloat (input)
	AttDesc:	The same as Sound's minBack attribute.

	Attribute:	minFront
	AttType:	SFFloat (input)
	AttDesc:	The same as Sound's minFront attribute.

	Attribute:	pulseWait
	AttType:	SFInt32 (input)
	AttDesc:	The number of transients to wait before first firing off the
				smaller cubes.

	Attribute:	waitPoint
	AttType:	SFFloat (input)
	AttDesc:	Time (in seconds) at which the cube should start firing off the
				smaller cubes.

	Attribute:	waitFraction
	AttType:	SFFloat (input)
	AttDesc:	To be connected to a TimeSensor, and used with waitPoint to let
				the cube know what the current time is.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0.0 0.0 -8.0 "&gt;
			&lt;PulseCubes DEF="pulses" waitPoint="0.5" minFront="20.0 " maxFront="40.0 " minBack="20.0 " maxBack="40.0 "&gt;
				&lt;Appearance&gt;
					&lt;Material diffuseColor="1.0 0.0 0.0"/&gt;
				&lt;/Appearance&gt;
				&lt;AudioClip url="/home/niall/music/concertPiece/oldSounds/HeartbeatSoundTemp.wav"/&gt;
			&lt;/PulseCubes&gt;
		&lt;/Transform&gt;

		&lt;TimeSensor DEF="timer" cycleInterval="20.0" loop="TRUE"/&gt;

		&lt;ROUTE fromNode="timer" fromField="fraction_changed" toNode="pulses" toField="waitFraction"/&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

class AudioClip;

//------------------------------------------------------------------------------
///	A cube which fires off smaller cubes when it detects transients.
/*!
	\ingroup libheilanextras
 */
class PulseCubes : public ParentNode
{
  public:
	///	Constructor.
	PulseCubes();
	///	Destructor.
	~PulseCubes();

	///	So we can position the pulses correctly.
	AmbisonicData *getAmbisonicAudio(int numSamples, const ThreeFloats& pos);
	///	Generates the audio for this node.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Used to set this node's attributes.
	/*!
		Also used by child AudioClip nodes to communicate that they've loaded
		their sound file, and it's ready for streaming.
	 */
	void setAttribute(const std::string& attribute, EventValue& val);
	///	Used to load the AudioClip.
	void childrenLoaded();

	static NodeConstructor<PulseCubes> constructor;
  private:
	///	Initialises various bits and pieces.
	void parentInitialise();

	///	Called so the node knows what samplerate it's running at.
	void samplerateChanged(float newRate);

	///	Called before the node starts drawing its children.
	void preChildDraw();
	///	Called after the node has finished drawing its children.
	void postChildDraw();
	///	Called before the node starts retrieving audio from it's children.
	//AmbisonicData *preChildAudio(int numSamples, const ThreeFloats& newPos);
	///	Called when the soundcard block size changes.
	void audioBlockSizeChanged(int newSize);

	///	Pointer to the current AudioClip node we're streaming sound from.
	AudioClip *audioClip;
	///	Pointer to the current block of audio we're reading from.
	float *currentBlock;
	///	The size of the sound file we're streaming from.
	float fileSize;

	///	Index we're currently reading from the current block.
	int index;
	///	Index we're reading from the overall sound file.
	float totalIndex;

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;

	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;

	///	The current samplerate.
	float samplerate;

	///	The current amplitude of the attached AudioClip.
	float currentAmp;
	///	Used to make sure we don't trigger the Pulses in too-quick succession.
	int pulseCount;
	///	Used to make sure we only trigger the Pulses on the second heartbeat.
	int pulseCount2;

	///	The number of pulses to wait before first triggering the Pulses.
	int waitNum;
	///	Fraction beyond which the pulses should be generated.
	/*!
		The node waits until it gets a waitFraction event with a value greater
		than or equal to this value, before it starts generating pulses.  A
		value of 0 (the default) means it starts immediately.
	 */
	float waitPoint;
	///	Used to determine whether we've passed waitPoint.
	float waitFraction;
	///	Whether we're generating pulses or not.
	bool generatePulses;
	///	The index of the last pulse.
	int currentPulse;

	//--------------------------------------------------------------------------
	///	Class representing the smaller cubes fired off by PulseCubes.
	class Pulse
	{
	  public:
		///	Constructor.
		Pulse();
		///	Destructor.
		~Pulse();

		///	Triggers the cube to start moving (always starts from 0,0,0).
		void trigger();
		///	Returns the current block of Ambisonic audio.
		/*!
			\return 0 if the cube is not active.
		 */
		AmbisonicData *getAudioBlock(int numSamples, const ThreeFloats& newPos);
		///	Renders the cube.
		void renderCube();

		///	Sets the value to increment the cube's position by.
		/*!
			This is used to calculate how to move the cube - i.e. if inc was
			(1,0,0) it would move to the right, by one unit per frame.
		 */
		void setPositionIncrement(const ThreeFloats& inc);
		///	Sets the samplerate we're running at.
		void setSamplerate(float val);
		///	Sets the size of the AmbisonicData buffer to use.
		void setAudioBlockSize(int newSize);
	  private:
		///	Used to determine the pulse' amplitude.
		inline float calcRelativeAttenuation(const ThreeFloats& relPos);

		///	The AmbisonicData buffer to return from getAudioBlock().
		AmbisonicData *audioBuffer;
		///	The size of audioBuffer.
		int bufferSize;

		///	Used to interpolate the node's position if the camera (or it) is moving.
		ThreeFloats lastPos;
		///	Used to interpolate the node's attenuation if the camera (or it) is moving.
		float lastAtten;

		///	The cube's current position.
		ThreeFloats currentPosition;
		///	The cube's current size.
		float currentSize;
		///	The cube's motion increment.
		ThreeFloats increment;

		///	Current samplerate.
		float samplerate;
		///	Current position in the ramp wave.
		float rampPos;
		///	Frequency of the ramp wave.
		float rampFreq;
		///	Increment for the ramp wave.
		float rampInc;
		///	Used to amplitude modulate the ramp wave.
		float rampAmp;
		///	Used to work out when to change rampAmp.
		int rampAmpInc;

		///	Filter cutoff.
		float cutoff;
		///	Filter variable.
		float f;
		///	Filter resonance.
		float q;
		///	Filter variable.
		float low;
		///	Filter variable.
		float band;
		///	Filter variable.
		float high;
		///	Filter variable.
		float notch;
	};
	//--------------------------------------------------------------------------

	///	Array of Pulses to shoot off on detection of a transient.
	Pulse pulses[12];
};

}

#endif
