//	SineTest.h - Simple sine tone test node.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SINETEST_H_
#define SINETEST_H_

#include "AbstractNode.h"

/*+
	Name: SineTest
	Type: ChildNode.
	Description: Simple test node, constantly outputs a sine wave.
	Attribute:	freq
	AttType:	SFFloat (input)
	AttDesc:	Frequency of the sine wave.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0.0 0.0 -1.0 "&gt;
			&lt;SineTest freq="880"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

struct AmbisonicData;

namespace HeilanX3D
{

///	Simple sine tone test node.
/*!
	\ingroup libheilanextras
 */
class SineTest : public ChildNode
{
  public:
	///	Constructor.
	SineTest();
	///	Destructor.
	~SineTest();

	///	Draw the node's position as a simple line shape.
	void graphicDoStuff();
	///	Output a constant sine wave.
	AmbisonicData *getMonoAudio(int numSamples);
	///	Sets the size of the node's temporary audio buffer.
	void setBlockSize(int newSize);

	///	Used to set the sine tone's frequency.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<SineTest> constructor;
  private:
	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;

	///	Frequency of the sine tone.
	float sinFreq;
	///	Coefficient for the sine tone.
	float sinCoeff;
	///	Thing for the sine tone.
	float sinX0;
	///	Thing for the sine tone.
	float sinX1;

	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;
};

}

#endif
