//	SparkParticles.cpp - Simple particle generator intended to look like sparks.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "SparkParticles.h"
#include "StringParser.h"
#include "Camera.h"

#include "GLHeaders.h"
#include <cmath>
#include <iostream>
#include <cstdlib>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<SparkParticles> SparkParticles::constructor;

//------------------------------------------------------------------------------
SparkParticles::SparkParticles():
AbstractNode(),
numParticles(32),
colour(1.0f, 1.0f, 0.1f)
{
	setType("SparkParticles");
	setVisual(true);
};

//------------------------------------------------------------------------------
SparkParticles::~SparkParticles()
{
		
}

//------------------------------------------------------------------------------
void SparkParticles::initialise()
{
	int i, j;
	int x, y;
	float tempf;
	unsigned char *pixelPointer = pixels;

	//Initialise particles.
	for(i=0;i<numParticles;++i)
	{
		particles.push_back(Particle());
		for(j=0;j<8;++j)
			oldPos[j].push_back(ThreeFloats(0.0f));

		particles[i].setGravity(0.025f);
		particles[i].setGravityDirection(ThreeFloats(0.0f, 0.0f, 0.0f));
	}

	//Generate texture.
	for(y=0;y<256;++y)
	{
		for(x=0;x<256;++x)
		{
			tempf = sinf(((float)x/256.0f)*3.1415926535897932384626433832795f);
			tempf *= sinf(((float)y/256.0f)*3.1415926535897932384626433832795f);
			tempf = (exp(tempf)-1.0f)/(exp(1.0f)-1.0f)*1.5f-0.1f;
			if(tempf < 0.0f)
				tempf = 0.0f;
			else if(tempf > 1.0f)
				tempf = 1.0f;
			*pixelPointer = 255;
			*(pixelPointer+1) = (unsigned char)(tempf * 255.0f);
			pixelPointer += 2;
		}
	}

	//Setup texture.
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexImage2D(GL_TEXTURE_2D,
				 0,
				 2,
				 256,
				 256,
				 0,
				 GL_LUMINANCE_ALPHA,
				 GL_UNSIGNED_BYTE,
				 pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);
}

//------------------------------------------------------------------------------
void SparkParticles::graphicDoStuff()
{
	int i;
	float tempRand;
	ThreeFloats tempf;
	const float avgLifespan = 0.5f;
	const float particleSize = (1.0f/16.0f);

	for(i=0;i<numParticles;++i)
	{
		if(!particles[i].getIsActive())
		{
			tempRand = (float)rand()/(float)RAND_MAX;
			tempRand -= 0.5f;
			tempRand *= avgLifespan;
			particles[i].setLifespan(avgLifespan+tempRand);

			tempRand = (float)rand()/(float)RAND_MAX;
			tempRand -= 0.5f;
			tempf.x = tempRand;
			tempRand = (float)rand()/(float)RAND_MAX;
			tempRand -= 0.5f;
			tempf.y = tempRand;
			tempRand = (float)rand()/(float)RAND_MAX;
			tempRand -= 0.5f;
			tempf.z = tempRand;
			particles[i].setDirection(tempf);

			particles[i].setColour(colour);

			particles[i].fire();
		}
		else
			particles[i].move();
	}

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glEnable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);

	glBindTexture(GL_TEXTURE_2D, texture);

	glDisable(GL_LIGHTING);

	for(i=0;i<numParticles;++i)
	{
		if(particles[i].getIsActive())
		{
			float angle;
			//const float speed = 1.0f - (particles[i].getSpeed() * 32.0f);
			const float speed = 0.25f;
			const ThreeFloats& tempDir = particles[i].getActualDirection();
			//const ThreeFloats& tempCol = particles[i].getColour();
			glColor4f(colour.x,
					  colour.y,
					  colour.z,
					  particles[i].getLifespan());

			angle = atanf(tempDir.y/tempDir.x);
			glTranslatef(particles[i].getPosition().x,
						 particles[i].getPosition().y,
						 particles[i].getPosition().z);
			glRotatef((angle * (180.0f/3.141592653f)), 0.0f, 0.0f, 1.0f); 

			//Primary particle.
			glBegin(GL_QUADS);
			{
				/*glTexCoord2f(0.0f, 1.0f);
				glVertex3f(-particleSize, (particleSize*0.25f), 0.0f);
				glTexCoord2f(1.0f, 1.0f);
				glVertex3f(particleSize, (particleSize*0.25f), 0.0f);
				glTexCoord2f(1.0f, 0.0f);
				glVertex3f(particleSize, -(particleSize*0.25f), 0.0f);
				glTexCoord2f(0.0f, 0.0f);
				glVertex3f(-particleSize, -(particleSize*0.25f), 0.0f);*/

				glTexCoord2f(0.0f, 1.0f);
				glVertex3f(-particleSize, (particleSize*speed), 0.0f);
				glTexCoord2f(1.0f, 1.0f);
				glVertex3f(particleSize, (particleSize*speed), 0.0f);
				glTexCoord2f(1.0f, 0.0f);
				glVertex3f(particleSize, -(particleSize*speed), 0.0f);
				glTexCoord2f(0.0f, 0.0f);
				glVertex3f(-particleSize, -(particleSize*speed), 0.0f);
			}
			glEnd();

			glRotatef(-(angle*(180.0f/3.141592653f)), 0.0f, 0.0f, 1.0f);

			glTranslatef(-particles[i].getPosition().x,
						 -particles[i].getPosition().y,
						 -particles[i].getPosition().z);

			//Trails.
			/*for(j=0;j<8;++j)
			{
				float size = 0.5f;
				glTranslatef(oldPos[j][i].x,
							 oldPos[j][i].y,
							 oldPos[j][i].z);
				glBegin(GL_QUADS);
				{
					glTexCoord2f(0.0f, 1.0f);
					glVertex3f(-(particleSize*size), (particleSize*size), 0.0f);
					glTexCoord2f(1.0f, 1.0f);
					glVertex3f((particleSize*size), (particleSize*size), 0.0f);
					glTexCoord2f(1.0f, 0.0f);
					glVertex3f((particleSize*size), -(particleSize*size), 0.0f);
					glTexCoord2f(0.0f, 0.0f);
					glVertex3f(-(particleSize*size), -(particleSize*size), 0.0f);
				}
				glEnd();
				glTranslatef(-oldPos[j][i].x,
							 -oldPos[j][i].y,
							 -oldPos[j][i].z);
			}
			for(j=7;j>0;--j)
				oldPos[j][i] = oldPos[j-1][i];
			oldPos[0][i] = particles[i].getPosition();*/
		}
	}

	glBindTexture(GL_TEXTURE_2D, 0);
	glEnable(GL_LIGHTING);
	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glColor3f(1.0f, 1.0f, 1.0f);
}

//------------------------------------------------------------------------------
void SparkParticles::setAttribute(const string& attribute, EventValue& val)
{
	if(attribute == "numParticles")
		numParticles = val;
	else if(attribute == "colour")
	{
		StringParser parse(val);

		if(parse.getNumThreeFloats() > 0)
			colour = parse.getThreeFloats();
	}
}

}
