//	SparkParticles.h - Simple particle generator intended to look like sparks.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SPARKPARTICLES_H_
#define SPARKPARTICLES_H_

#include "AbstractNode.h"
#include "ParticleGenerator.h"

#include "GLHeaders.h"

/*+
	Name: SparkParticles
	Type: ChildNode
	Description: A particle generator tweaked to try and make it look like
				 sparks.
	Attribute:	numParticles
	AttType:	SFInt32 (input)
	AttDesc:	How many particles may be visible at any one time.

	Attribute:	colour
	AttType:	SFVec3f (input)
	AttDesc:	The colour of the particles.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY" headlight="FALSE"/&gt;

		&lt;Background skyColor="0 0 0" skyAngle="0" groundColor="0 0 0" groundAngle="0"/&gt;

		&lt;Billboard&gt;
			&lt;Transform translation="0 0 -4"&gt;
				&lt;SparkParticles DEF="bob" colour="0.35 0.35 1"/&gt;
			&lt;/Transform&gt;
		&lt;/Billboard&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

//------------------------------------------------------------------------------
///	Simple particle generator intended to look like sparks.
/*!
	\ingroup libheilanextras
 */
class SparkParticles : public ChildNode
{
  public:
	///	Constructor.
	SparkParticles();
	///	Destructor.
	~SparkParticles();

	///	Initialises the texture.
	void initialise();

	///	Draws the SparkParticles.
	void graphicDoStuff();

	///	Overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set this node's attributes.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<SparkParticles> constructor;
  private:
	///	The texture we use.
	GLuint texture;
	///	Our pixel data.
	unsigned char pixels[256*256*2];

	///	Number of particles to display.
	int numParticles;
	///	The colour of the particles.
	ThreeFloats colour;
	///	Vector holding our particles.
	std::vector<Particle> particles;

	///	Test.
	std::vector<ThreeFloats> oldPos[8];
};

}

#endif
