//	StarCubes.cpp - Five cubes, whose size (& radius from the centre cube) are
//				    controlled by the associated AudioClip.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "StarCubes.h"
#include "AudioClip.h"
#include "AmbisonicData.h"
#include "Camera.h"
#include "StringParser.h"
#include <cassert>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<StarCubes> StarCubes::constructor;

//------------------------------------------------------------------------------
StarCubes::StarCubes():
ParentNode(),
audioClip(0),
currentBlock(0),
fileSize(0),
index(0),
totalIndex(0),
audioBuffer(0),
bufferSize(0),
lastAtten(0.0f),
samplerate(44100.0f),
currentAmp(0.0f)
{
	setType("StarCubes");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
StarCubes::~StarCubes()
{
	if(audioBuffer)
		delete [] audioBuffer;
}

//------------------------------------------------------------------------------
AmbisonicData *StarCubes::getMonoAudio(int numSamples)
{
	int i;
	float tempf;

	//Skip everything if we don't have an AudioClip.
	if(!audioClip)
		return 0;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		if(totalIndex < fileSize)
		{
			tempf = currentBlock[static_cast<int>(index)];

			//Update index and totalIndex.
			++totalIndex;
			++index;
			if(totalIndex >= fileSize)
			{
				//Loop back to the start of the file.
				totalIndex = fmod(totalIndex, fileSize);
				index = static_cast<int>(fmod(totalIndex, AudioClip::streamBlockSize));
				currentBlock = audioClip->getNextBlock(this);
			}
			else if(index >= AudioClip::streamBlockSize)
			{
				//Load the next block.
				index = static_cast<int>(fmod(totalIndex, AudioClip::streamBlockSize));
				currentBlock = audioClip->getNextBlock(this);
			}
		}
		else
			tempf = 0.0f;

		//Convert float sound file to Ambisonic, put it in the buffer.
		audioBuffer[i].W = tempf;

		//Update envelope.
		if((fabs(tempf)-currentAmp) > 0.1f)
			currentAmp = tempf;
		else if(fabs(tempf) > currentAmp)
			//currentAmp = tempf;
			currentAmp += 0.0001f;
		else if(currentAmp > 0.0f)
			currentAmp -= 0.00001f;
		else
			currentAmp = 0.0f;
	}

	return audioBuffer;
}

//------------------------------------------------------------------------------
void StarCubes::setAttribute(const string& attribute, EventValue& val)
{
	if(attribute == "maxBack")
		setMaxBackAudioRadius(val);
	else if(attribute == "maxFront")
		setMaxFrontAudioRadius(val);
	else if(attribute == "minBack")
		setMinBackAudioRadius(val);
	else if(attribute == "minFront")
		setMinFrontAudioRadius(val);
}

//------------------------------------------------------------------------------
void StarCubes::childrenLoaded()
{
	unsigned int i;

	for(i=0;i<(unsigned)getNumChildren();++i)
	{
		if(getChild(i)->getType() == "AudioClip")
		{
			AudioClip *tempAudioClip = dynamic_cast<AudioClip *>(getChild(i));

			if(tempAudioClip->getFileSize())
			{
				//Set audioClip.
				audioClip = tempAudioClip;

				//Set currentBlock and fileSize.
				currentBlock = audioClip->getCurrentBlock();
				fileSize = audioClip->getFileSize();
			}

			break;
		}
	}
}

//------------------------------------------------------------------------------
void StarCubes::parentInitialise()
{
	//Generate the display list.
	pixel = glGenLists(1);
	glNewList(pixel, GL_COMPILE);
		glBegin(GL_QUADS);
			//Front face.
			glNormal3f(0.0f, 0.0f, 1.0f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-0.5f, -0.5f, 0.5f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(0.5f, -0.5f, 0.5f);
			glTexCoord2f(0.5f, 0.5f);
			glVertex3f(0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.0f, 0.5f);
			glVertex3f(-0.5f, 0.5f, 0.5f);
	
			//Rear face.
			glNormal3f(0.0f, 0.0f, -1.0f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(-0.5f, -0.5f, -0.5f);
			glTexCoord2f(0.5f, 0.5f);
			glVertex3f(-0.5f,  0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.5f);
			glVertex3f(0.5f,  0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(0.5f, -0.5f, -0.5f);
	
			// Top Face
			glNormal3f(0.0f, 1.0f, 0.0f);
			glTexCoord2f(0.0f, 0.5f);
			glVertex3f(-0.5f, 0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.5f, 0.5f);
			glVertex3f(0.5f, 0.5f, -0.5f);
		
			// Bottom Face
			glNormal3f(0.0f, -1.0f, 0.0f);
			glTexCoord2f(0.5f, 0.5f);
			glVertex3f(-0.5f, -0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.5f);
			glVertex3f(0.5f, -0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(0.5f, -0.5f, 0.5f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(-0.5f, -0.5f, 0.5f);
	
			// Right face
			glNormal3f(1.0f, 0.0f, 0.0f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(0.5f, -0.5f, -0.5f);
			glTexCoord2f(0.5f, 0.5f);
			glVertex3f(0.5f, 0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.5f);
			glVertex3f(0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(0.5f, -0.5f, 0.5f);
	
			// Left Face
			glNormal3f(-1.0f, 0.0f, 0.0f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-0.5f, -0.5f, -0.5f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(-0.5f, -0.5f, 0.5f);
			glTexCoord2f(0.5f, 0.5f);
			glVertex3f(-0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.0f, 0.5f);
			glVertex3f(-0.5f, 0.5f, -0.5f);
		glEnd();
	glEndList();
}

//------------------------------------------------------------------------------
void StarCubes::samplerateChanged(float newRate)
{
	samplerate = newRate;
}

//------------------------------------------------------------------------------
void StarCubes::preChildDraw()
{
	
}

//------------------------------------------------------------------------------
void StarCubes::postChildDraw()
{
	const GLfloat diffuse[]={1.0f, 1.0f, 1.0f};
	const GLfloat ambient[]={0.0f, 0.0f, 0.0f};
	const GLfloat emissive[]={0.0f, 0.0f, 0.0f};
	const GLfloat specular[]={0.5f, 0.5f, 0.5f};
	const GLfloat shininess = 0.0f;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	if(currentAmp > 0.0f)
	{
		//Centre cube.
		glPushMatrix();
		glScalef(currentAmp, currentAmp, currentAmp);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();

		//Top cube 1.
		glPushMatrix();
		glRotatef((currentAmp*90.0f), 0.0f, 0.0f, 1.0f);
		glScalef(currentAmp, currentAmp, currentAmp);
		glTranslatef(0.0f, 4.0f, 0.0f);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();

		//Bottom cube 1.
		glPushMatrix();
		glRotatef((currentAmp*90.0f), 0.0f, 0.0f, 1.0f);
		glScalef(currentAmp, currentAmp, currentAmp);
		glTranslatef(0.0f, -4.0f, 0.0f);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();

		//Right cube 1.
		glPushMatrix();
		glRotatef((currentAmp*90.0f), 0.0f, 0.0f, 1.0f);
		glScalef(currentAmp, currentAmp, currentAmp);
		glTranslatef(4.0f, 0.0f, 0.0f);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();

		//Left cube 1.
		glPushMatrix();
		glRotatef((currentAmp*90.0f), 0.0f, 0.0f, 1.0f);
		glScalef(currentAmp, currentAmp, currentAmp);
		glTranslatef(-4.0f, 0.0f, 0.0f);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();

		//Top cube 2.
		glPushMatrix();
		glRotatef(-(currentAmp*90.0f), 0.0f, 0.0f, 1.0f);
		glScalef(currentAmp, currentAmp, currentAmp);
		glTranslatef(0.0f, 8.0f, 0.0f);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();

		//Bottom cube 2.
		glPushMatrix();
		glRotatef(-(currentAmp*90.0f), 0.0f, 0.0f, 1.0f);
		glScalef(currentAmp, currentAmp, currentAmp);
		glTranslatef(0.0f, -8.0f, 0.0f);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();

		//Right cube 2.
		glPushMatrix();
		glRotatef(-(currentAmp*90.0f), 0.0f, 0.0f, 1.0f);
		glScalef(currentAmp, currentAmp, currentAmp);
		glTranslatef(8.0f, 0.0f, 0.0f);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();

		//Left cube 2.
		glPushMatrix();
		glRotatef(-(currentAmp*90.0f), 0.0f, 0.0f, 1.0f);
		glScalef(currentAmp, currentAmp, currentAmp);
		glTranslatef(-8.0f, 0.0f, 0.0f);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();

		//Top Right cube 2.
		glPushMatrix();
		glRotatef(-(currentAmp*90.0f), 0.0f, 0.0f, 1.0f);
		glScalef(currentAmp, currentAmp, currentAmp);
		glTranslatef(5.65f, 5.65f, 0.0f);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();

		//Bottom Right cube 2.
		glPushMatrix();
		glRotatef(-(currentAmp*90.0f), 0.0f, 0.0f, 1.0f);
		glScalef(currentAmp, currentAmp, currentAmp);
		glTranslatef(5.65f, -5.65f, 0.0f);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();

		//Bottom Left cube 2.
		glPushMatrix();
		glRotatef(-(currentAmp*90.0f), 0.0f, 0.0f, 1.0f);
		glScalef(currentAmp, currentAmp, currentAmp);
		glTranslatef(-5.65f, -5.65f, 0.0f);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();

		//Top Left cube 2.
		glPushMatrix();
		glRotatef(-(currentAmp*90.0f), 0.0f, 0.0f, 1.0f);
		glScalef(currentAmp, currentAmp, currentAmp);
		glTranslatef(-5.65f, 5.65f, 0.0f);
		glRotatef(45.0f, 1.0f, 1.0f, 0.0f);
		glCallList(pixel);
		glPopMatrix();
	}

	glPopMatrix();

	//Not sure if there's a better way to do this?
	glBindTexture(GL_TEXTURE_2D, 0);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
	glMaterialfv(GL_FRONT, GL_AMBIENT, ambient);
	glMaterialfv(GL_FRONT, GL_EMISSION, emissive);
	glMaterialf(GL_FRONT, GL_SHININESS, shininess); 
	glMaterialfv(GL_FRONT, GL_SPECULAR, specular);

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

//------------------------------------------------------------------------------
/*AmbisonicData *StarCubes::preChildAudio(int numSamples, const ThreeFloats& newPos)
{
	int i;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;
	float tempf;

	//Skip everything if we don't have an AudioClip.
	if(!audioClip)
		return 0;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Get current Ambisonic position (see AmbisonicData::operator=()), relative
	//to the camera.
	relativePos = newPos;
	relativePos -= Camera::getInstance()->getAudioPosition();
	lastAtten = calcRelativeAttenuation(lastPos);
	if(lastPos != relativePos)
	{
		ambiStart = lastPos;
		ambiEnd = relativePos;
		//ambiInc = 1.0f/numSamples;
		//We increment the position by a constant value, so we don't get clicks
		//or zipper noise.  Of course, there are disadvantages to this
		//approach...
		ambiInc = 0.005f;
		currentAmbi = ambiStart;
		newAtten = calcRelativeAttenuation(relativePos);
	}
	else
		currentAmbi = relativePos;
	currentAtten = lastAtten;

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		if(totalIndex < fileSize)
		{
			tempf = currentBlock[static_cast<int>(index)];

			//Update index and totalIndex.
			++totalIndex;
			++index;
			if(totalIndex >= fileSize)
			{
				//Loop back to the start of the file.
				totalIndex = fmod(totalIndex, fileSize);
				index = static_cast<int>(fmod(totalIndex, AudioClip::streamBlockSize));
				currentBlock = audioClip->getNextBlock(this);
			}
			else if(index >= AudioClip::streamBlockSize)
			{
				//Load the next block.
				index = static_cast<int>(fmod(totalIndex, AudioClip::streamBlockSize));
				currentBlock = audioClip->getNextBlock(this);
			}
		}
		else
			tempf = 0.0f;

		//Update currentAmbi and currentAtten if necessary.
		if(ambiInc > 0.0f)
		{
			//Calc currentAmbi.
			currentAmbi = ((1.0f-ambiIndex) * ambiStart) +
						  (ambiIndex * ambiEnd);
			//Calc currentAtten.
			currentAtten = ((1.0f-ambiIndex) * lastAtten) +
						   (ambiIndex * newAtten);

			ambiIndex += ambiInc;

			//Stop updating index if we've reached our destination.
			//I get weird rounding errors here, hence the .0000001f etc...
			if(ambiIndex > 1.0000001f)
			{
				ambiInc = 0.0f;
				currentAtten = newAtten;
			}
		}

		//Convert float sound file to Ambisonic, put it in the buffer.
		audioBuffer[i] = currentAmbi * (tempf * currentAtten);

		//Update envelope.
		if((fabs(tempf)-currentAmp) > 0.1f)
			currentAmp = tempf;
		else if(fabs(tempf) > currentAmp)
			//currentAmp = tempf;
			currentAmp += 0.0001f;
		else if(currentAmp > 0.0f)
			currentAmp -= 0.00001f;
		else
			currentAmp = 0.0f;
	}
	//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
	if(ambiIndex >= 1.0f)
	{
		lastPos = relativePos;
		lastAtten = currentAtten;
	}
	else
	{
		lastPos = ((1.0f-ambiIndex) * lastPos) + (ambiIndex * relativePos);
		lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
	}

	return audioBuffer;
}*/

//------------------------------------------------------------------------------
void StarCubes::audioBlockSizeChanged(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}
	
}
