//	StarCubes.h - Five cubes, whose size (& radius from the centre cube) are
//				  controlled by the associated AudioClip.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef STARCUBES_H_
#define STARCUBES_H_

#include "ParentNode.h"
#include "GLHeaders.h"

/*+
	Name: StarCubes
	Type: ParentNode, to be used with a child AudioClip node.
	Description: Consists of five nodes, the size of which is controlled by the
			 	 amplitude envelope of the accompanying AudioClip.  For four of
				 the cubes, their radius from the centre cube is also controlled
				 by the amplitude envelope.
	Attribute:	maxBack
	AttType:	SFFloat (input)
	AttDesc:	The same as Sound's maxBack attribute.

	Attribute:	maxFront
	AttType:	SFFloat (input)
	AttDesc:	The same as Sound's maxFront attribute.

	Attribute:	minBack
	AttType:	SFFloat (input)
	AttDesc:	The same as Sound's minBack attribute.

	Attribute:	minFront
	AttType:	SFFloat (input)
	AttDesc:	The same as Sound's minFront attribute.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0.0 0.0 -8.0 "&gt;
			&lt;StarCubes&gt;
				&lt;Appearance&gt;
					&lt;Material diffuseColor="0.2 0.2 0.2 "/&gt;
				&lt;/Appearance&gt;
				&lt;AudioClip url="SineThing2.wav"/&gt;
			&lt;/StarCubes&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

class AudioClip;

///	Five cubes, whose size (& radius from the centre cube) are controlled by the associated AudioClip.
/*!
	\ingroup libheilanextras
 */
class StarCubes : public ParentNode
{
  public:
	///	Constructor.
	StarCubes();
	///	Destructor.
	~StarCubes();

	///	Generates the audio for this node.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Used to set this node's attributes.
	/*!
		Also used by child AudioClip nodes to communicate that they've loaded
		their sound file, and it's ready for streaming.
	 */
	void setAttribute(const std::string& attribute, EventValue& val);
	///	Used to load the AudioClip.
	void childrenLoaded();

	static NodeConstructor<StarCubes> constructor;
  private:
	///	Initialises various bits and pieces.
	void parentInitialise();

	///	Called so the node knows what samplerate it's running at.
	void samplerateChanged(float newRate);

	///	Called before the node starts drawing its children.
	void preChildDraw();
	///	Called after the node has finished drawing its children.
	void postChildDraw();
	///	Called before the node starts retrieving audio from it's children.
	//AmbisonicData *preChildAudio(int numSamples, const ThreeFloats& newPos);
	///	Called when the soundcard block size changes.
	void audioBlockSizeChanged(int newSize);

	///	Pointer to the current AudioClip node we're streaming sound from.
	AudioClip *audioClip;
	///	Pointer to the current block of audio we're reading from.
	float *currentBlock;
	///	The size of the sound file we're streaming from.
	float fileSize;

	///	Index we're currently reading from the current block.
	int index;
	///	Index we're reading from the overall sound file.
	float totalIndex;

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;

	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;

	///	The current samplerate.
	float samplerate;

	///	The current amplitude of the attached AudioClip.
	float currentAmp;

	///	Storage for one 'pixel'.
	GLuint pixel;
};

}

#endif
