//	StretchyBackground.cpp - Test node to implement a stretchy background, derived
//						   	 from the foreground objects.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "StretchyBackground.h"
#include "Camera.h"
#include "StringParser.h"
#include "AttributeHelper.h"

#include <cmath>
#include <cstdlib>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<StretchyBackground> StretchyBackground::constructor;

//------------------------------------------------------------------------------
StretchyBackground::StretchyBackground():
ChildNode(),
stretchiness(64.0f),
rotation(0.0f),
cubeMotion(0.0f)
{
	setType("StretchyBackground");
	setVisual(true);
	setShouldBeStart(true);
}

//------------------------------------------------------------------------------
StretchyBackground::~StretchyBackground()
{
	
}

//------------------------------------------------------------------------------
void StretchyBackground::initialise()
{
	int i;

	//Setup displayList.
	displayList = glGenLists(1);
	glNewList(displayList, GL_COMPILE);
	{
		glBegin(GL_QUADS);
		{
			//Front face.
			glNormal3f(0.0f, 0.0f, 1.0f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-0.5f, -0.5f, 0.5f);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(0.5f, -0.5f, 0.5f);
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(-0.5f, 0.5f, 0.5f);

			//Rear face.
			glNormal3f(0.0f, 0.0f, -1.0f);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(-0.5f, -0.5f, -0.5f);
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(-0.5f,  0.5f, -0.5f);
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(0.5f,  0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(0.5f, -0.5f, -0.5f);

			// Top Face
			glNormal3f(0.0f, 1.0f, 0.0f);
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(-0.5f, 0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-0.5f, 0.5f, 0.5f);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(0.5f, 0.5f, 0.5f);
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(0.5f, 0.5f, -0.5f);

			// Bottom Face
			glNormal3f(0.0f, -1.0f, 0.0f);
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(-0.5f, -0.5f, -0.5f);
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(0.5f, -0.5f, -0.5f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(0.5f, -0.5f, 0.5f);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(-0.5f, -0.5f, 0.5f);

			// Right face
			glNormal3f(1.0f, 0.0f, 0.0f);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(0.5f, -0.5f, -0.5f);
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(0.5f, 0.5f, -0.5f);
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(0.5f, -0.5f, 0.5f);

			// Left Face
			glNormal3f(-1.0f, 0.0f, 0.0f);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-0.5f, -0.5f, -0.5f);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(-0.5f, -0.5f, 0.5f);
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(-0.5f, 0.5f, 0.5f);
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(-0.5f, 0.5f, -0.5f);
		}
		glEnd();
	}
	glEndList();

	//Setup foreground cube positions.
	for(i=0;i<NumCubes;++i)
	{
		cubePos[i].x = (((float)rand()/(float)RAND_MAX) * 4.0f) - 2.0f;
		cubePos[i].y = (((float)rand()/(float)RAND_MAX) * 8.0f) - 4.0f;
		cubePos[i].z = (((float)rand()/(float)RAND_MAX) * 4.0f) - 2.0f;
		cubeRots[i] = ((float)rand()/(float)RAND_MAX) * 90.0f;
	}

	//Setup background texture.
	for(i=0;i<(TexWidth*TexHeight*3);++i)
		backgroundPixels[i] = 0;

	glGenTextures(1, &backgroundTexture);
	glBindTexture(GL_TEXTURE_2D, backgroundTexture);
	glTexImage2D(GL_TEXTURE_2D,
				 0,
				 3,
				 TexWidth,
				 TexHeight,
				 0,
				 GL_RGB,
				 GL_UNSIGNED_BYTE,
				 backgroundPixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

//------------------------------------------------------------------------------
void StretchyBackground::graphicDoStuff()
{
	int i;
/*	float x[4];
	float y[4];*/
	const int viewWidth = Camera::getInstance()->getViewWidth();
	const int viewHeight = Camera::getInstance()->getViewHeight();
	float tempf = sinf(cubeMotion);

	cubeMotion += 0.01f;
	if(cubeMotion > 6.283185306f)
		cubeMotion = 0.0f;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	//Set viewport to the texture's dimensions.
	glViewport(0, 0, TexWidth, TexHeight);

	//Switch to an ortho mode.
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, viewWidth, viewHeight, 0, -1, 16);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	//Draw background gradient.
	glDisable(GL_LIGHTING);
	glBegin(GL_QUADS);
	{
		glColor3ub(192, 192, 224);
		glVertex3i(0, 0, -16);

		glVertex3i(viewWidth, 0, -16);

		glColor3ub(240, 240, 255);
		glVertex3i(viewWidth, viewHeight, -16);

		glVertex3i(0, viewHeight, -16);
	}
	glEnd();
	glEnable(GL_LIGHTING);

	//Switch back to perspective mode.
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	//Draw cubes.
	for(i=0;i<NumCubes;++i)
	{
		glRotatef(cubeRots[i], 1.0f, 0.0f, 1.0f);
		glTranslatef(cubePos[i].x+tempf,
					 cubePos[i].y+tempf,
					 cubePos[i].z+tempf);
		glCallList(displayList);
		glTranslatef(-(cubePos[i].x+tempf),
					 -(cubePos[i].y+tempf),
					 -(cubePos[i].z+tempf));
		glRotatef(cubeRots[i], -1.0f, 0.0f, -1.0f);
	}

	//Copy drawn cubes to the texture.
	glBindTexture(GL_TEXTURE_2D, backgroundTexture);
	glCopyTexImage2D(GL_TEXTURE_2D,
					 0,
					 GL_RGB8,
					 0,
					 0,
					 TexWidth,
					 TexHeight,
					 0);

	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//Reset the viewport.
	glViewport(0,
			   0,
			   viewWidth,
			   viewHeight);

	//Switch to an ortho mode.
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, viewWidth, viewHeight, 0, -1, 16);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	//Calculate texture transforms.
	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	glTranslatef(0.5f-((1.0f/stretchiness)*0.5f), 0.0f, 0.0f);
	glScalef((1.0f/stretchiness), 1.0f, 1.0f);
	glRotatef(rotation*360.0f, 0.0f, 0.0f, 1.0f);
	glMatrixMode(GL_MODELVIEW);

	//Now perform various translations etc.

	//Draw background texture.
	glDisable(GL_LIGHTING);
	glBegin(GL_QUADS);
	{
		glColor3ub(255, 255, 255);
		glTexCoord2f(0.0f, 1.0f);
		glVertex3i(0, 0, -16);

		glTexCoord2f(0.0f, 0.0f);
		glVertex3i(0, viewHeight, -16);

		glTexCoord2f(1.0f, 0.0f);
		glVertex3i(viewWidth, viewHeight, -16);

		glTexCoord2f(1.0f, 1.0f);
		glVertex3i(viewWidth, 0, -16);
	}
	glEnd();
	glEnable(GL_LIGHTING);

	glMatrixMode(GL_TEXTURE);
	glPopMatrix();

	//Switch back to perspective mode.
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glBindTexture(GL_TEXTURE_2D, 0);

	//Draw cubes again so we see them as well as the blurred version.
	for(i=0;i<NumCubes;++i)
	{
		glRotatef(cubeRots[i], 1.0f, 0.0f, 1.0f);
		glTranslatef(cubePos[i].x+tempf,
					 cubePos[i].y+tempf,
					 cubePos[i].z+tempf);
		glCallList(displayList);
		glTranslatef(-(cubePos[i].x+tempf),
					 -(cubePos[i].y+tempf),
					 -(cubePos[i].z+tempf));
		glRotatef(cubeRots[i], -1.0f, 0.0f, -1.0f);
	}

	glPopMatrix();
}

//------------------------------------------------------------------------------
void StretchyBackground::setAttribute(const string& attribute, EventValue& val)
{
	if(attribute == "stretchiness")
		stretchiness = val;
	else if(attribute == "rotation")
		rotation = val;
}

}
