//	StretchyBackground.h - Test node to implement a stretchy background, derived
//						   from the foreground objects.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef STRETCHYBACKGROUND_H_
#define STRETCHYBACKGROUND_H_

#include "AbstractNode.h"

#include "GLHeaders.h"

/*+
	Name: StretchyBackground
	Type: ChildNode.
	Description: 32 cubes are positioned randomly, and set to move along random
			 	 vectors.  This is then rendered to a texture which is applied
				 to the background of the scene, and which can be stretched and
				 rotated by the user via the node's attributes.
	Attribute:	stretchiness
	AttType:	SFFloat (input)
	AttDesc:	How much the texture is stretched by in the x direction (bigger
				numbers == better).

	Attribute:	rotation
	AttType:	SFFloat (input)
	AttDesc:	How much the texture is rotated (range: 0-&gt;1).

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0 0 -12"&gt;
			&lt;Shape&gt;
				&lt;Appearance&gt;
					&lt;Material diffuseColor="0.75 1 0.85"/&gt;
				&lt;/Appearance&gt;
				&lt;StretchyBackground DEF="bob"/&gt;
			&lt;/Shape&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Test node to implement a stretchy background, derived from the foreground objects.
/*!
	\ingroup libheilanextras
 */
class StretchyBackground : public ChildNode
{
  public:
	///	Constructor.
	StretchyBackground();
	///	Destructor.
	~StretchyBackground();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the StretchyBackground.
	void graphicDoStuff();
	///	Not used, overridden to do nothing.
	/*AmbisonicData *audioDoStuff(int numSamples,
								const ThreeFloats& pos) {return 0;};*/

	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set the amount of stretchiness etc.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<StretchyBackground> constructor;
  private:
	///	Some constants.
	enum
	{
		NumCubes = 32,
		TexWidth = 512,
		TexHeight = 512
	};

	///	The amount of stretchiness.
	float stretchiness;
	///	How much the background's rotated by.
	float rotation;

	///	The positions of the foreground cubes.
	ThreeFloats cubePos[NumCubes];
	///	The rotations of the foreground cubes.
	float cubeRots[NumCubes];
	///	Used to move the cubes in a sine motion.
	float cubeMotion;
	///	Display list used to draw the foreground cubes.
	GLuint displayList;

	///	Our background texture.
	GLuint backgroundTexture;
	///	Our background texture pixel data.
	unsigned char backgroundPixels[TexWidth*TexHeight*3];
};

}

#endif
