//	Toki.h - An audiovisual instrument where the performer controls the
//			 audio and visuals which both also control each other.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef TOKI_H_
#define TOKI_H_

#include "AbstractNode.h"

#include "GLHeaders.h"
#ifdef HAVE_FFTW
#include <fftw3.h>
#endif
#include <cmath>

/*+
	Name: Toki
	Type: ChildNode.
	Description: A simple audiovisual instrument.  Uses NURBS for the visuals,
				 and a physically-modelled string for the audio.
	Attribute:	slider1
	AttType:	SFFloat (input)

	Attribute:	slider2
	AttType:	SFFloat (input)

	Attribute:	slider3
	AttType:	SFFloat (input)

	Attribute:	slider4
	AttType:	SFFloat (input)

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0.0 0.0 -6.0"&gt;
			&lt;Toki DEF="bob"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	An audiovisual instrument where the performer controls the audio and visuals which both also control each other.
/*!
	\ingroup libheilanextras
 */
class Toki : public ChildNode
{
  public:
	///	Constructor.
	Toki();
	///	Destructor.
	~Toki();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the visuals.
	void graphicDoStuff();
	///	To handle the instrument's audio output.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Called to set the current audio block size.
	void setBlockSize(int newSize);
	///	Called so we can update controlRate.
	void setSamplerate(float val);

	///	Called to set the value of the named attribute.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<Toki> constructor;
  private:
	///	Method to fill coordsSym/Asym and coloursSym/Asym.
	void setupCoordsColours();

	///	Simple linear interpolation method.
	inline float interp(float index, float left, float right)
	{
		float retval;
		const float tempf = index - static_cast<float>(static_cast<int>(index));

		retval = (tempf * right) + ((1.0f-tempf) * left);

		return retval;
	}

	///	Further simplified linear interpolation method for the symmetrical calcs.
	inline float interpSym(float index, float left, float right)
	{
		float retval;

		retval = (index * right) + ((1.0f-index) * left);

		return retval;
	}

	///	A cheap, faked exp curve.
	/*!
		Only works on a 0->1 range.
	 */
	inline float fakeExp(float val)
	{
		return val/(val + (expCoeff*(val-1.0f)));
	}
	///	A cheap, faked log curve.
	/*!
		Only works on a 0->1 range.
	 */
	inline float fakeLog(float val)
	{
		return val/(val + (logCoeff*(val-1.0f)));
	}

	///	Updates the frequency of the string.
	inline void updateStringFreq(float val)
	{
		static const float actualFreq = (24000.0f*44100.0f/44100.0f)/(StringLength+1);
		static const float invLog = 1.0f/log10f(2.0f);
		float compensationFactor;
		if(val < 55.0f)
			val = 55.0f;
		else if(val > 200.0f)
			val = 200.0f;
		compensationFactor=powf(4.0f, log10f(actualFreq/val)*invLog);
		mass = 3.5f * compensationFactor;
		invMass = 1.0f/mass;
	}

	///	A parameter smoothing filter class, used to remove jerkiness.
	class SmoothingFilter
	{
	  public:
		///	Constructor.
		SmoothingFilter();
	  	///	Destructor.
		~SmoothingFilter();

		///	Returns the smoothed version of the input.
		inline float operator()(double input)
		{
			aReg = (aCoeff*aReg) + input;
			bReg = (bCoeff*bReg) + input;
			cReg = (cCoeff*cReg) + input;

			return (float)((aGain*aReg)+(bGain*bReg)+(cGain*cReg));
		};
	  private:
		///	A coefficient.
		double aCoeff;
		///	B coefficient.
		double bCoeff;
		///	C coefficient.
		double cCoeff;
		///	A register.
		double aReg;
		///	B register.
		double bReg;
		///	C register.
		double cReg;
		///	A gain.
		double aGain;
		///	B gain.
		double bGain;
		///	C gain.
		double cGain;
	};

	///	Enum containing various constants.
	enum
	{
		Order = 3,
		NumPoints = 16,
		NumKnots = (NumPoints/4)+Order,
		StringLength = 128,
		FeedbackPos = (StringLength/4),
		BowPosition = 7*(StringLength/8),
		Slider1BufSize = 8,
		RmsSize = 32
	};

	//----Input parameters----

	///	Current value of the slider 1 input.
	float slider1Val;
	///	Current value of the slider 2 input.
	float slider2Val;
	///	Current value of the slider 3 input.
	float slider3Val;
	///	Current value of the slider 4 input.
	float slider4Val;

	//----Control parameters----

	///	The current speed at which slider 1 is moving.
	float slider1Rate;
	///	Whether slider1Rate should be interpreted as exp or log for damping amt/symmetry.
	bool slider1Type;
	///	Buffer used to calculate slider1Rate.
	float slider1Prev[Slider1BufSize];
	///	Rate(num samples) at which control parameters (slider1Rate...) are updated.
	int controlRate;
	///	Used to determine when to update control parameters.
	int controlCount;

	///	SmoothingFilter used to smooth slider1Rate.
	SmoothingFilter slider1Filter;

	///	Coefficient for exponential curve.
	float expCoeff;
	///	Coefficient for log curve.
	float logCoeff;

	///	The current energy content of the audio.
	float energyContent;
	///	Previous audio amplitude values.
	float audioRms[RmsSize];
	///	The current (instantaneous) amplitude of the audio.
	float instantAmp;
	///	Current audio brightness.
	float brightness;

	//----Visuals parameters----

	///	The GLU nurbs object we use to draw the shape.
	GLUnurbsObj *nurbsObject;
	///	The NURBS knots in the u direction.
	float uKnots[NumKnots];
	///	The NURBS knots in the v direction.
	float vKnots[NumKnots];
	///	The NURBS control points.
	float points[NumPoints*3];
	///	The NURBS colours.
	float colours[NumPoints*4];

	///	Visuals parameter: Whether all 4 sections move as 1, or they move in groups of 2.
	/*!
		We use this value to interpolate between the symmetrical and
		asymmetrical parameters for 2 of the sections (the other 2 are always
		symmetrical).
	 */
	float symmetry;
	///	Visuals parameter: The speed of rotation (symmetrical).
	float rotSpeedSym;
	///	Visuals parameter: The speed of rotation (asymmetrical).
	float rotSpeedAsym;
	///	Visuals parameter: The scale of the sections (symmetrical).
	float scaleSym;
	///	Visuals parameter: The scale of the sections (asymmetrical).
	float scaleAsym;
	///	Visuals parameter: The coordinate interpolation value.
	/*!
		Various coordinate settings are stored at points along a 1d line, with
		this value determining what the current coordinates should be.
	 */
	float coordInt;
	///	Visuals parameter: The colour interpolation value.
	/*!
		Various colour settings are stored at points along a 1d line, with
		this value determining what the current coordinates should be.
	 */
	float colourInt;

	///	The current rotation value (symmetrical).
	float rotValSym;
	///	The current rotation value (asymmetrical).
	float rotValAsym;

	///	The symmetrical array of Coordinates.
	ThreeFloats coordsSym[5][16];
	///	The asymmetrical array of Coordinates.
	ThreeFloats coordsAsym[5][16];
	///	The symmetrical array of Colours.
	ThreeFloats coloursSym[5][16];
	///	The asymmetrical array of Colours.
	ThreeFloats coloursAsym[5][16];
	///	The symmetrical colour temperatures.
	float colourTempsSym[5];
	///	The asymmetrical colour temperatures.
	float colourTempsAsym[5];

	//----Audio parameters----

	///	Delay line for the string (i.e. the positions of the cells).
	float delayLine[StringLength];
	///	Forces for the string cells.
	float forces[StringLength];
	///	Velocities for the string cells.
	float velocities[StringLength];
	///	Mass of the individual cells.
	float mass;
	///	Inverse mass of the individual cells.
	float invMass;
	///	How much to multiply the cells' velocities by (default).
	float velocityMultiplier;
	///	How much to multiply the cells' velocities by (damped).
	float dampMultiplier;

	///	Amplitude of the feedback.
	float feedbackAmp;
	///	Damping value.
	float dampingValue;
	///	Damping right position.
	int dampingRight;

	///	Bow force.
	float bowForce;
	///	Bow velocity.
	float bowVelocity;
	///	Bow mode.
	bool bowMode;
	///	Position (y-axis) of the bow.
	float bowPointPosition;
	///	The actual force exerted by the bow.
	float bowForceExerted;

	///	DC filter coefficient.
	float dcSpeed;
	///	DC filter coefficient.
	float dcPos;

#ifdef HAVE_FFTW
	///	Constant containing the size of the fft.
	enum
	{
		FftSize = 1024
	};
	///	Input audio buffer.
	/*!
		Used to calculate the 'brightness' of the audio output.
	 */
	double fftInput[FftSize];
	///	FFT output.
	/*!
		Used to calculate the 'brightness' of the audio output.
	 */
	double fftOutput[FftSize];
	///	Plan for the FFT.
	fftw_plan fftPlan;

	///	Counter used to determine when to update fftOutput.
	int fftCount;
#endif

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;
	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;
};

}

#endif
