//	TransientDetector.h - TransientDetector node, detects transients in
//						  AudioClips.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef TRANSIENTDETECTOR_H_
#define TRANSIENTDETECTOR_H_

#include "ParentNode.h"

/*+
	Name: TransientDetector
	Type: ParentNode, to be used with a child AudioClip.
	Description: This node is mainly identical to Sound, but includes an
				 envelope detector and transient detector, both of which output
				 events using X3D's ROUTE method.  In this way you can easily
				 set up some basic audiovisual mappings based on the amplitude
				 envelope of a sound file.  The events output are named
				 <i>amplitudeEnvelope</i> and
				 <i>transientDetected</i>, where amplitudeEnvelope
				 has a range 0->1, and transientDetected outputs a 1 followed
				 by a 0 when it detects a transient.  Note that
				 amplitudeEnvelope events are sent every 0.01s, while
				 transientDetected events are sent when the node detects them.
				 The node has all the same attributes that the standard X3D
				 Sound node has, in addition to those below.
	Attribute:	isMute
	AttType:	SFBool (input)
	AttDesc:	If true, the node will not output any sound, just send the
				envelope and transient information detected.

	Attribute:	amplitudeEnvelope
	AttType:	SFFloat (output)
	AttDesc:	The current value of the AudioClip's amplitude (0-1).

	Attribute:	transientDetected
	AttType:	SFFloat (output)
	AttDesc:	Set to 1 when a transient's detected, then 0.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;GhostImage DEF="bob" scale="1"&gt;
			&lt;Transform translation="0.0 0.0 -4.0 "&gt;
				&lt;TransientDetector DEF="trans"&gt;
					&lt;AudioClip url="/home/niall/music/samples/thing.wav" loop="TRUE"/&gt;
				&lt;/TransientDetector&gt;
				&lt;Shape&gt;
					&lt;Appearance&gt;
						&lt;Material diffuseColor="0.0 0.0 1.0 "/&gt;
					&lt;/Appearance&gt;
					&lt;Box DEF="box" size="0.125 0.5 0.125"/&gt;
				&lt;/Shape&gt;
			&lt;/Transform&gt;
		&lt;/GhostImage&gt;

		&lt;ROUTE fromNode="trans" fromField="transientDetected" toNode="bob" toField="trigger"/&gt;
		&lt;ROUTE fromNode="trans" fromField="amplitudeEnvelope" toNode="box" toField="size.y"/&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

class AudioClip;

///	TransientDetector node, detects transients in AudioClips.
/*!
	\ingroup libheilanextras

	This node is mainly identical to Sound, but includes an envelope detector
	and transient detector, both of which output events using X3D's ROUTE
	method.  In this way you can set up some basic audiovisual mappings based on
	the amplitude envelope of a sound file.
 */
class TransientDetector : public ParentNode
{
  public:
	///	Constructor.
	TransientDetector();
	///	Destructor.
	~TransientDetector();

	///	Overridden to take the location attribute into account.
	/*!
		We don't sum any child nodes' audio because (I don't think) Sound's
		supposed to be able to have child sound-emitting nodes.
	 */
	AmbisonicData *getAmbisonicAudio(int numSamples, const ThreeFloats& pos);
	///	Generates the audio for this node.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Used to set this node's attributes.
	/*!
		Also used by child AudioClip nodes to communicate that they've loaded
		their sound file, and it's ready for streaming.
	 */
	void setAttribute(const std::string& attribute, EventValue& val);
	///	Used to load the AudioClip.
	void childrenLoaded();

	static NodeConstructor<TransientDetector> constructor;
  private:
	///	Called so the node knows what samplerate it's running at.
	void samplerateChanged(float newRate);

	///	Used to stream audioClip to the soundcard.
	//AmbisonicData *preChildAudio(int numSamples, const ThreeFloats& newPos);
	///	Called when the soundcard block size changes.
	void audioBlockSizeChanged(int newSize);

	///	Whether or not the sound file should be played looped or not.
	bool loop;
	///	The node's location attribute.
	ThreeFloats location;
	///	The node's intensity.
	float intensity;
	///	Whether or not the node should output sound.
	bool isMute;

	///	Pointer to the current AudioClip node we're streaming sound from.
	AudioClip *audioClip;
	///	Pointer to the current block of audio we're reading from.
	float *currentBlock;
	///	The size of the sound file we're streaming from.
	float fileSize;

	///	Index we're currently reading from the current block.
	float index;
	///	Index we're reading from the overall sound file.
	float totalIndex;
	///	Current increment (used to read from the sound file at different speeds).
	float increment;

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;

	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;

	///	The current samplerate.
	float samplerate;

	///	The time (in seconds) this node should start playing.
	float startTime;
	///	The time (in seconds) this node should stop playing.
	float stopTime;
	///	The time (in seconds) this node should pause playing.
	float pauseTime;
	///	The time (in seconds) this node should resume playing.
	float resumeTime;
	///	The time (in samples) this node should start playing.
	int startSamples;
	///	The time (in samples) this node should stop playing.
	int stopSamples;
	///	The time (in samples) this node should pause playing.
	int pauseSamples;
	///	The time (in samples) this node should resume playing.
	int resumeSamples;
	///	The elapsed time (in samples).
	int elapsedSamples;

	///	Our current amplitude envelope value.
	float envelope;
	///	Counter used to determine when to send the continuous signals.
	int contCounter;
};

}

#endif
