//	WebcamTexture.h - A texture that displays the output of a webcam.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef WEBCAMTEXTURE_H_
#define WEBCAMTEXTURE_H_

#include "AbstractNode.h"
#include "CriticalSection.h"
#include "Webcam.h"

#include "GLHeaders.h"

/*+
	Name: WebcamTexture
	Type: ChildNode.
	Description: Streams the output from a webcam into a texture. Uses the first
				 camera on your system, and defaults to a resolution of
				 320x240.  This means that the resultant texture will have quite
				 a lot of blank pixels, as OpenGL requires textures to have
				 power-of-two dimensions (512x512 in this case).
	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;Transform translation="0 0 -8" scale="4.45 4.45 1"&gt;
			&lt;Shape&gt;
				&lt;Appearance&gt;
					&lt;WebcamTexture/&gt;
				&lt;/Appearance&gt;
				&lt;IndexedFaceSet coordIndex="0 1 2 3 -1" texCoordIndex="0 1 2 3 -1"&gt;
					&lt;Coordinate point="2 1.5 0	-2 1.5 0	-2 -1.5 0		2 -1.5 0"/&gt;
					&lt;TextureCoordinate point="0.625 0	0 0		0 0.46875		0.625 0.46875"/&gt;
				&lt;/IndexedFaceSet&gt;
			&lt;/Shape&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	A texture that displays the output of a webcam.
/*!
	\ingroup libheilanextras
 */
class WebcamTexture : public ChildNode,
					  public WebcamCallback
{
  public:
	///	Constructor.
	WebcamTexture();
	///	Destructor.
	~WebcamTexture();

	///	Called when the webcam has a new frame ready.
	void callback(unsigned char *data);

	///	Starts the Webcam object capturing.
	void initialise();

	///	Draws the texture.
	void graphicDoStuff();

	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Not used, overridden to do nothing.
	void setAttribute(const std::string& attribute, EventValue& val) {};

	static NodeConstructor<WebcamTexture> constructor;
  private:
	///	The size of our webcam images.
	enum
	{
		Width = 320,
		Height = 240,
		TexSize = 512
	};

	///	Our Webcam object.
	Webcam *webcam;
	///	Critical section to protect pixels.
	CriticalSection critSec;

	///	Our texture.
	GLuint texture;
	///	The webcam pixel data for our texture.
	//unsigned char pixels[Width*Height*3];
	unsigned char pixels[TexSize*TexSize*3];
};

}

#endif
