//	Yakul.cpp - A simple audiovisual instrument based on a smooth-coarse
//			  continuum.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Yakul.h"
#include "Camera.h"
#include "StringParser.h"
#include "AmbisonicData.h"

#include "GLHeaders.h"
#include <cmath>
#include <cassert>
#include <iostream>
#include <cstdlib>

using namespace std;

#define Pi 3.1415926535897932384626433832795f

namespace HeilanX3D
{
	
NodeConstructor<Yakul> Yakul::constructor;

//------------------------------------------------------------------------------
Yakul::Yakul():
ChildNode(),
controlParam(0.0f),
controlParamChanged(false),
averageROC(0.0f),
lowEnvelope(0.0f),
lowEnvelopeDecay(0.00001f),
highPartialEnvelope(0.0f),
lastControlParamDecay(0.0001f),
highPartialEnvelopeDecay(0.00005f),
audioBuffer(0),
bufferSize(0),
lastAtten(0.0f)
{
	setType("Yakul");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
Yakul::~Yakul()
{
	if(audioBuffer)
		delete [] audioBuffer;
}

//------------------------------------------------------------------------------
void Yakul::initialise()
{
	int i;

	for(i=0;i<NumTones;++i)
	{
		if(rand() > RAND_MAX/2)
			tones[i].frequency = (110.0f+(float)rand()/(float)RAND_MAX) * static_cast<float>(i);
		else
			tones[i].frequency = 0.0f;
		tones[i].amplitude = 0.0f;
		tones[i].coeff = 2.0f * sinf(Pi*(tones[i].frequency/44100.0f));
		tones[i].x0 = 0.0f;
		tones[i].x1 = 1.0f;
	}

	for(i=0;i<5;++i)
		lastControlParams[i] = 0.0f;

	lowPartials[0] = 0.0f;
	for(i=1;i<(NumTones>>4);++i)
		lowPartials[i] = 1.0f;
}

//------------------------------------------------------------------------------
void Yakul::graphicDoStuff()
{
	int x;

	glLineWidth(2.0f);
	glPointSize(4.0f);
	glBegin(GL_TRIANGLES);
		//glVertex2f(0.0f, 0.0f);
		//for(y=0;y<64;++y)
		{
			glVertex2f(sinf(((float)0/64.0f)*(2.0f*Pi))*lowEnvelope,
							   cosf(((float)0/64.0f)*(2.0f*Pi))*lowEnvelope);
			for(x=1;x<64;++x)
			{
				if(((x%2)==0) /*&& ((y%2)==0)*/)
				{
					glVertex2f(sinf(((float)x/64.0f)*(2.0f*Pi))*lowEnvelope,
							   cosf(((float)x/64.0f)*(2.0f*Pi))*lowEnvelope);
				}
				else
				{
					glVertex2f(sinf(((float)x/64.0f)*(2.0f*Pi))*(1.0f-highPartialEnvelope)*lowEnvelope,
							   cosf(((float)x/64.0f)*(2.0f*Pi))*(1.0f-highPartialEnvelope)*lowEnvelope);
				}
				glVertex2f(0.0f, 0.0f);
				if(((x%2)==0) /*&& ((y%2)==0)*/)
				{
					glVertex2f(sinf(((float)x/64.0f)*(2.0f*Pi))*lowEnvelope,
							   cosf(((float)x/64.0f)*(2.0f*Pi))*lowEnvelope);
				}
				else
				{
					glVertex2f(sinf(((float)x/64.0f)*(2.0f*Pi))*(1.0f-highPartialEnvelope)*lowEnvelope,
							   cosf(((float)x/64.0f)*(2.0f*Pi))*(1.0f-highPartialEnvelope)*lowEnvelope);
				}
			}
		}
		glVertex2f(sinf(0.0f)*lowEnvelope, cosf(0.0f)*lowEnvelope);
		glVertex2f(0.0f, 0.0f);
	glEnd();
}

//------------------------------------------------------------------------------
AmbisonicData *Yakul::getMonoAudio(int numSamples)
{
	int i, j;
	float tempf = 0.0f;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Update lastControlParams.
		for(j=4;j>0;--j)
			lastControlParams[j] = lastControlParams[j-1];
		if(controlParamChanged)
		{
			if(fabs(controlParam - lastControlParams[0]) >
			   fabs(lastControlParams[0] - lastControlParams[1]))
			{
				//cout << "fabs(controlParam - lastControlParams[0]): " << fabs(controlParam - lastControlParams[0]) << endl;
				lastControlParams[0] = controlParam;
			}
			controlParamChanged = false;
		}
		/*else if(lastControlParams[0] > 1e-10)
			lastControlParams[0] *= 0.9995f;
		else
			lastControlParams[0] = 0.0f;*/

		//Update averageROC.
		averageROC = 0.0f;
		for(j=0;j<4;++j)
			averageROC += fabs(lastControlParams[j]-lastControlParams[j+1]) * 0.25f;

		//Update lowEnvelope.
		if(averageROC > 0.0f)
		{
			lowEnvelope += (lowEnvelopeDecay * 10000.0f);
			if(lowEnvelope > 1.0f)
				lowEnvelope = 1.0f;
		}
		else if(lowEnvelope > 0.0f)
		{
			lowEnvelope -= lowEnvelopeDecay;
			if(lowEnvelope < 0.0f)
				lowEnvelope = 0.0f;
		}

		//Calculate highPartialEnvelope.
		tempf = fabs((1.0f+lastControlParams[0])-(1.0f+lastControlParams[1]));
		if(tempf > highPartialEnvelope)
			highPartialEnvelope = tempf;
		else if(highPartialEnvelope > 0.0f)
			highPartialEnvelope -= highPartialEnvelopeDecay*0.25f;
		else
			highPartialEnvelope = 0.0f;

		//Update lowPartials.
		for(j=0;j<(NumTones>>4);++j)
		{
			if(lowPartials[j] < 1.0f)
			{
				lowPartials[j] += 0.00005f;
				if(lowPartials[j] >= 1.0f)
				{
					lowPartials[j] = 1.0f;
					lowPartials[(j+1)%(NumTones>>4)] = 0.0f;
				}
				break;
			}
		}

		//Update the tones' amplitudes.
		//j < 8, if NumTones == 128.
		for(j=0;j<(NumTones>>4);++j)
			tones[j].amplitude = lowEnvelope * lowPartials[j];
		for(j=(NumTones>>4);j<NumTones;++j)
			tones[j].amplitude = (highPartialEnvelope /* ((float)(NumTones-j)/(float)NumTones)*/) * 0.5f;

		//Calculate monophonic audio.
		tempf = 0.0f;
		for(j=0;j<NumTones;++j)
		{
			if(tones[j].amplitude > 0.0f)
			{
				tones[j].x0 = tones[j].x0 - (tones[j].x1 * tones[j].coeff);
				tones[j].x1 = tones[j].x1 + (tones[j].x0 * tones[j].coeff);
				tempf += tones[j].x0 * tones[j].amplitude;
			}
		}

		//Convert float sine tone to Ambisonic, put it in the buffer.
		audioBuffer[i].W = tempf;
	}

	return audioBuffer;
}

//------------------------------------------------------------------------------
void Yakul::setBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}

//------------------------------------------------------------------------------
void Yakul::setAttribute(const string& attribute, EventValue& val)
{
	if(attribute == "controlParam")
	{
		controlParam = val;
		controlParamChanged = true;
	}
	else if(attribute == "frequency")
	{
		for(int i=0;i<NumTones;++i)
		{
			if(tones[i].frequency > 0.0f)
				tones[i].frequency = (float)val * static_cast<float>(i);
			else
				tones[i].frequency = 0.0f;
			tones[i].coeff = 2.0f * sinf(Pi*(tones[i].frequency/44100.0f));
			tones[i].x0 = 0.0f;
			tones[i].x1 = 1.0f;
		}
	}
}

}
