//	Yakul.h - A simple audiovisual instrument based on a smooth-coarse
//			  continuum.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef YAKUL_H_
#define YAKUL_H_

#include "AbstractNode.h"

/*+
	Name: Yakul
	Type: ChildNode.
	Description: A simple audiovisual instrument.  This is the first instrument
			 	 built for my PhD.  The visuals consist of a sun-type polygon,
				 the size of which is controlled by the speed at which the
				 performer moves the controlParam attribute.  Sudden changes in
				 speed make the polygon more pointy.  The audio engine consists
				 of an additive synth, the high partials of which get amplified
				 when there are sudden changes of speed in controlParam.
	Attribute:	controlParam
	AttType:	SFFloat (input)

	Attribute:	frequency
	AttType:	SFFloat (input)
	AttDesc:	The root frequency of the additive synth.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="ANY"/&gt;

		&lt;Transform translation="0.0 0.0 -5.0"&gt;
			&lt;Appearance&gt;
				&lt;Material diffuseColor="0.8 0.5 0.0"/&gt;
			&lt;/Appearance&gt;
			&lt;Yakul DEF="bob"&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	A simple audiovisual instrument based on a smooth-coarse continuum.
/*!
	\ingroup libheilanextras

	This instrument does not implement any direct audio-visual mappings.  The
	connection between the two comes solely from the fact that they're
	controlled by the same gestural parameter.

	The instrument is controlled by a single parameter, of which it is the rate
	of change which matters.  When the parameter is not moving, the instrument
	outputs no sound, and the visual object is not visible.  When the parameter
	is moving slowly, the sound output is a (relatively) low frequency drone,
	with partials changing slowly.  Visually, the object is represented as a
	sphere in this case.  When the parameter is undergoing rapid, discontinuous
	changes, the audio responds by adding a lot of high frequency content which
	rapidly decays.  Visually, the object becomes coarse and full of sharp
	points.  As the motion smooths again, the object moves back to a simple
	sphere.
 */
class Yakul : public ChildNode
{
  public:
	///	Constructor.
	Yakul();
	///	Destructor.
	~Yakul();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the instrument.
	void graphicDoStuff();
	///	To handle the instrument's audio output.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Called to set the current audio block size.
	void setBlockSize(int newSize);

	///	Called to set the value of the named attribute.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<Yakul> constructor;
  private:
	///	Simple struct representing a single tone in the additive synthesizer.
	struct Tone
	{
		///	The frequency of the tone.
		float frequency;
		///	The amplitude of the tone.
		float amplitude;
		///	Coefficient for the sine generation.
		float coeff;
		///	Coefficient for the sine generation.
		float x0;
		///	Coefficient for the sine generation.
		float x1;
	};

	///	The number of tones in the additive synthesizer.
	enum
	{
		NumTones = 128
	};

	///	The tones for the additive synthesizer.
	Tone tones[NumTones];
	///	The current value (static) parameter used to control the instrument.
	float controlParam;
	///	true when controlParam has changed, and lastControlParams need updated.
	bool controlParamChanged;
	///	The average rate of change of controlParam.
	float averageROC;
	///	Values used to calculate averageROC.
	/*!
		lastControlParams[0] is the most recent one - it gets recalculated every
		sample, taking the value of controlParam if it's changed, or decaying
		down to 0 if not.  These values are propagated along the rest of the
		lastControlParams.

		If fabs(controlParam - lastControlParams[0]) > 0.2(?), we generate high
		partials, but setting highPartialEnvelope > 0.
	 */
	float lastControlParams[5];
	///	Envelope used to control the overall amplitude of the low partials.
	/*!
		This is controlled by averageROC.
	 */
	float lowEnvelope;
	///	The value by which lowEnvelope decays.
	float lowEnvelopeDecay;
	///	Envelope used to control the high partials of the additive synth.
	float highPartialEnvelope;
	///	The value by which lastControlParams[0] decays.
	float lastControlParamDecay;
	///	The value by which highPartialEnvelope decays.
	float highPartialEnvelopeDecay;
	///	Used to envelope the low partials.
	float lowPartials[NumTones>>4];

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;
	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;
};

}

#endif
