//	FPSNavigation.h - FPS-style navigation type.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef FPSNAVIGATION_H_
#define FPSNAVIGATION_H_

#include "Navigation.h"
#include "ThreeFloats.h"
#include "FourFloats.h"

namespace HeilanX3D
{

/*+
	Name: FPSNavigation
	Type: Navigation type
	Description: A navigation type based on the navigation typically found in
				 First Person Shooters.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="FPS"/&gt;

		&lt;Transform translation="3.0 0.0 -5.0" rotation="0.0 1.0 0.25 55.0"&gt;
			&lt;Shape&gt;
				&lt;Appearance&gt;
					&lt;Material diffuseColor="0.25 0.125 0.0 "/&gt;
				&lt;/Appearance&gt;
				&lt;Box size="0.25 0.25 0.25"/&gt;
			&lt;/Shape&gt;
		&lt;/Transform&gt;
		&lt;Transform translation="-2.0 0.0 -5.0" rotation="0.0 -1.0 0.0 35.0"&gt;
			&lt;Shape&gt;
				&lt;Appearance&gt;
					&lt;ImageTexture url="/home/niall/Projects/NiallsX3DBrowser/src/woodthing.jpg"/&gt;
				&lt;/Appearance&gt;
				&lt;Box size="1.0 1.0 1.0"/&gt;
			&lt;/Shape&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

///	FPS-style navigation type.
/*!
	\ingroup libheilanextras

	Note: This navigation type grabs mouse and keyboard input, so you won't be
	able to interact with other applications while you're using it.


	\todo Add up-down mouselook (the Camera class will need modified to handle
	rotation properly.
 */
class FPSNavigation : public Navigation
{
  public:
	///	Constructor.
	FPSNavigation();
	///	Destructor.
	~FPSNavigation();

	///	To handle the mouse look.
	void mouseMove(int x, int y);

	///	To handle the movement/strafing.
	void keyDown(SDLKey key);
	///	To handle the movement/strafing.
	void keyUp(SDLKey key);

	///	To update our position.
	void screenRefresh();

	static NavigationConstructor<FPSNavigation> constructor;
  private:
	///	Whether we're moving forward or not.
	bool movingForward;
	///	Whether we're moving backward or not.
	bool movingBackward;
	///	Whether we're strafing left or not.
	bool movingLeft;
	///	Whether we're strafing right or not.
	bool movingRight;

	///	How much we're rotating around the y axis.
	float xRotation;

	///	Keeps track of our current x rotation.
	float xRot;
	///	Keeps track of our current y rotation.
	float yRot;
	///	Our last x rotation value.
	float lastXRot;
	///	Our last y rotation value.
	float lastYRot;
	///	Our actual rotation.
	FourFloats rotation;

	///	Our current position.
	ThreeFloats position;
};

}

#endif
