//	SimpleViewpointNavigation.h - Moving the mouse interpolates between
//								  Viewpoints.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SIMPLEVIEWPOINTNAVIGATION_H_
#define SIMPLEVIEWPOINTNAVIGATION_H_

#include "Navigation.h"

/*+
	Name: SimpleViewpointNavigation
	Type: Navigation type.
	Description: A simple navigation type, where moving the mouse interpolates
			 	 between Viewpoints.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="SIMPLE_VIEWPOINT"/&gt;

		&lt;Viewpoint position="0.0 0.0 0.0 "/&gt;
		&lt;Viewpoint position="0.0 0.0 -8.0 "/&gt;
		&lt;Viewpoint position="1.0 0.0 -12.0 "/&gt;
		&lt;Viewpoint position="-4.0 0.0 -24.0 "/&gt;

		&lt;Transform translation="1.0 -0.05 -1.0"&gt;
			&lt;SineTest freq="880"/&gt;
		&lt;/Transform&gt;
		&lt;Transform translation="-1.0 0.05 -1.0"&gt;
			&lt;SineTest freq="440"/&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Moving the mouse interpolates between Viewpoints.
/*!
	\ingroup libheilanextras
 */
class SimpleViewpointNavigation : public Navigation
{
  public:
	///	Constructor.
	SimpleViewpointNavigation();
	///	Destructor.
	~SimpleViewpointNavigation();

	///	Called on a mouse button down event.
	void mouseDown(int button, int x, int y);
	///	Called on a mouse button move event.
	void mouseMove(int x, int y);
	///	Called on a mouse button up event.
	void mouseUp(int button, int x, int y);

	static NavigationConstructor<SimpleViewpointNavigation> constructor;
  private:
	///	Whether or not the left mouse button is currently down.
	bool leftIsDown;
	///	Whether or not the middle mouse button is currently down.
	bool middleIsDown;
	///	Whether or not the right mouse button is currently down.
	bool rightIsDown;

	///	The last y-position of the mouse.
	int lastY;

	///	The current viewpoint we're at.
	float viewpointIndex;
};

}

#endif
