//	TimedViewpointNavigation.cpp - Camera moves on it's own, according to the
//								   time variables of the current viewpoint.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "TimedViewpointNavigation.h"
#include "BindableNode.h"
#include "Viewpoint.h"
#include "Camera.h"

#include <cmath>
#include <iostream>

using namespace std;

namespace HeilanX3D
{

NavigationConstructor<TimedViewpointNavigation> TimedViewpointNavigation::constructor;

//------------------------------------------------------------------------------
TimedViewpointNavigation::TimedViewpointNavigation():
Navigation(),
viewpointIndex(0),
lastTicks(0),
currentIndex(0.0f),
inPause(true)
{
	BindableNodeSingleton *binder = BindableNodeSingleton::getInstance();
	Viewpoint *nextViewpoint = (Viewpoint *)binder->getIndexedNode("Viewpoint",
																   viewpointIndex);

	/*if(!Camera::getInstance()->getOfflineRendering())
		lastTicks = SDL_GetTicks();*/

	//Update camera's start position.
	Camera::getInstance()->setGraphicPosition(nextViewpoint->getPosition());
}

//------------------------------------------------------------------------------
TimedViewpointNavigation::~TimedViewpointNavigation()
{
	
}

//------------------------------------------------------------------------------
void TimedViewpointNavigation::screenRefresh()
{
	Uint32 newTicks;
	Uint32 tickDiff;
	float fracIndex;
	ThreeFloats newPos;
	ThreeFloats startPos;
	ThreeFloats endPos;
	Viewpoint *lastViewpoint;
	BindableNodeSingleton *binder = BindableNodeSingleton::getInstance();

	int numNodes = binder->getNumNodes("Viewpoint");
	if(numNodes < 2)
		return;

	Viewpoint *nextViewpoint = (Viewpoint *)binder->getIndexedNode("Viewpoint",
																   viewpointIndex);

	if((viewpointIndex-1) < 0)
	{
		lastViewpoint = (Viewpoint *)binder->getIndexedNode("Viewpoint",
															(numNodes-1));
	}
	else
	{
		lastViewpoint = (Viewpoint *)binder->getIndexedNode("Viewpoint",
															(viewpointIndex-1));
	}

	//Calculate difference in ticks.
	/*if(!getIsOffline())
		newTicks = SDL_GetTicks();
	else*/
		newTicks = Camera::getInstance()->getOfflineTicks();
	tickDiff = newTicks - lastTicks;
	lastTicks = newTicks;

	//If we're not currently in the pause section of the current Viewpoint.
	if(!inPause)
	{
		//Update currentIndex.
		currentIndex += (static_cast<float>(tickDiff)/1000.0f);
		//If we're now in the pause section.
		if(currentIndex > nextViewpoint->getTimeToGetHere())
		{
			inPause = true;
			if(nextViewpoint->getTimeToGetHere() > 0.0f)
			{
				currentIndex = fmod(currentIndex,
									nextViewpoint->getTimeToGetHere());
			}
			else
			{
				currentIndex = 0.0f;
				//Because we wouldn't get to update the camera's position
				//otherwise.
				Camera::getInstance()->setGraphicPosition(nextViewpoint->getPosition());
			}
		}
		else
		{
			//Calculate new camera position.
			startPos = lastViewpoint->getPosition();
			endPos = nextViewpoint->getPosition();
			if(nextViewpoint->getTimeToGetHere() > 0.0f)
				fracIndex = currentIndex/nextViewpoint->getTimeToGetHere();
			else
				fracIndex = 1.0f;
			newPos = (startPos * (1.0f-fracIndex)) + (endPos * fracIndex);

			//Update camera position.
			Camera::getInstance()->setGraphicPosition(newPos);
		}
	}
	else
	{
		//Update currentIndex.
		currentIndex += (static_cast<float>(tickDiff)/1000.0f);
		//If it's time to increment viewpointIndex.
		if(currentIndex > nextViewpoint->getPauseTime())
		{
			++viewpointIndex;
			viewpointIndex = viewpointIndex % numNodes;
			if(nextViewpoint->getPauseTime() > 0)
				currentIndex = fmod(currentIndex, nextViewpoint->getPauseTime());
			else
				currentIndex = 0.0f;
			inPause = false;
		}
	}
}

}
