//	TimedViewpointNavigation.h - Camera moves on it's own, according to the
//								 time variables of the current viewpoint.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef TIMEDVIEWPOINTNAVIGATION_H_
#define TIMEDVIEWPOINTNAVIGATION_H_

#include "Navigation.h"
#include "SDL.h"

/*+
	Name: TimedViewpointNavigation
	Type: Navigation type.
	Description: The camera moves between Viewpoints on its own, regardless of
				 user input.  See the
				 <a href="../viewpointdocs.htm">Timed Viewpoint</a> page for
				 more information.

	Example:
&lt;?xml version="1.0" encoding="utf-8"?&gt;
&lt;!DOCTYPE X3D&gt;
&lt;X3D profile="Full"&gt;
	&lt;Scene&gt;
		&lt;NavigationInfo type="TIMED_VIEWPOINT"/&gt;

		&lt;Viewpoint position="0.0 0.0 0.0 "&gt;
			&lt;MetadataString name="heilan_timeToGetHere" value="0.0 "/&gt;
			&lt;MetadataString name="heilan_pauseTime" value="45.0 "/&gt;
		&lt;/Viewpoint&gt;
		&lt;Viewpoint position="0.0 0.0 -140.0 "&gt;
			&lt;MetadataString name="heilan_timeToGetHere" value="15.25 "/&gt;
			&lt;MetadataString name="heilan_pauseTime" value="0.0 "/&gt;
		&lt;/Viewpoint&gt;
		&lt;Viewpoint position="0.0 0.0 -300.0 "&gt;
			&lt;MetadataString name="heilan_timeToGetHere" value="10.0 "/&gt;
			&lt;MetadataString name="heilan_pauseTime" value="2.0 "/&gt;
		&lt;/Viewpoint&gt;
		&lt;Viewpoint position="0.0 0.0 -350.0 "&gt;
			&lt;MetadataString name="heilan_timeToGetHere" value="10.0 "/&gt;
			&lt;MetadataString name="heilan_pauseTime" value="0.0 "/&gt;
		&lt;/Viewpoint&gt;

		&lt;Transform translation="0.0 1.0 -140.0 "&gt;
			&lt;PixelArray image="/home/niall/Projects/HeilanBrowser/src/DarkBlueRadial.bmp" minFront="150" maxFront="280" minBack="150" maxBack="280"&gt;
				&lt;AudioClip url="/mnt/disk2/music/concertPiece/fadeInDroneT2a.wav"/&gt;
			&lt;/PixelArray&gt;
		&lt;/Transform&gt;
	&lt;/Scene&gt;
&lt;/X3D&gt;
 */

namespace HeilanX3D
{

///	Camera moves on it's own, according to the time variables of the current viewpoint.
/*!
	\ingroup libheilanextras
 */
class TimedViewpointNavigation : public Navigation
{
  public:
	///	Constructor.
	TimedViewpointNavigation();
	///	Destructor.
	~TimedViewpointNavigation();

	///	Used to update the camera's position according to the current time.
	void screenRefresh();

	static NavigationConstructor<TimedViewpointNavigation> constructor;
  private:
	///	The current viewpoint we're at.
	int viewpointIndex;

	///	The ticks the previous screenRefresh() occurred at.
	Uint32 lastTicks;

	///	Index within the current section.
	/*!
		Could be timeToGetHere or pauseTime for any particular Viewpoint.
	 */
	float currentIndex;
	///	Whether we're currently in the timeToGetHere or pauseTime of the current Viewpoint.
	bool inPause;
};

}

#endif
