//	Application.cpp - Class handling the various application bits and pieces.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "GLHeaders.h"
#include "Application.h"
#include "FatalError.h"
#include "Browser.h"
#include "SpeakerConfig.h"
#include "CustomSpeakerConfig.h"
#include "NodeFactory.h"
#include "Camera.h"
#include "OscRecorder.h"
//#include "AvailableExtensions.h"
#include "GLHeaders.h"
#include "KeyboardInput.h"
#include "SoundOutput.h"

#ifdef HAVE_LIBJPEG
extern "C"
{
#include "jpeglib.h"
}
#endif

#ifdef HAVE_FFMPEG
extern "C"
{
#include "ffmpeg/avformat.h"
#include "ffmpeg/avcodec.h"
}
#endif

#ifndef WIN32
#include <dlfcn.h>
#endif

#include "SDL_loadso.h"
#include "sndfile.h"
#include <sstream>
#include <iostream>
#include <cstdio>
#include <cassert>


using namespace std;

//------------------------------------------------------------------------------
Application::Application(const CmdLineParams &commandLine):
browser(0),
timeToQuit(false),
lastRepaint(0),
width(0),
height(0),
flags(SDL_OPENGL|SDL_RESIZABLE),
fullscreen(false),
version("0.15"),
runningTime(0),
samplerate(44100),
numChannels(2)
{
	int aaLevel = 0;
	bool antialias = false;
	CmdLineParams::const_iterator it;
	stringstream tempConverter;
	string fileToLoad;
	float audioSamplerate = 44100.0f;
	int audioSpeakers = 1;
	SpeakerConfig *speakers = 0;
	int audioBlockSize = 0;
	int audioApi = -1;
	int audioDevice = -1;
	int audioInputDevice = -1;
	string oscMulticastGroup;
	int oscPort = 5678;
	string oscOutAddress;
	int oscOutPort = 5678;
	bool noAudio = false;
	bool noOSC = false;
	bool bloomFilter = false;
	bool cursor = false;
	int tempFlags;

	//Determine whether to print the help from the command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if((*it) == "--help")
		{
			cout << endl;
			cout << endl;
			cout << "HeilanBrowser: An X3D browser specialised for audio work.";
			cout << endl;
			cout << "Version: " << version;
			cout << endl << endl;
			cout << "Usage:" << endl;
			cout << "HeilanBrowser [flags] <x3d_file>";
			cout << endl << endl;
			cout << "Flags:" << endl;
			cout << "--width [val] | -w [val]:\tSet the width of the window.";
			cout << endl;
			cout << "--height [val] | -h [val]:\tSet the height of the window.";
			cout << endl;
			cout << "--antialias [val] | -a [val]:\tSets the antialiasing level.";
			cout << endl;
			cout << "--library [file] | -l [file]:\tAdds a (node/nav/sound) library to be loaded.";
			cout << endl;
			cout << "--audioAPI [val] | -aa [val]:\tSets the audio API to use.";
			cout << endl;
			cout << "--audioDevice [val] | -ad [val]: Sets the output audio device to use.";
			cout << endl;
			cout << "--audioInputDevice [val] | -ai [val]: Sets the input audio device to use.";
			cout << endl;
			cout << "--audioSamplerate [val] | -as [val]: Sets the audio samplerate to use.";
			cout << endl;
			cout << "--audioSpeakers [val] <file> | -ap [val] <file>: Sets the speaker configuration to use.";
			cout << endl;
			cout << "--audioBlockSize [val] | -ab [val]: Sets the audio block size to use.";
			cout << endl;
			cout << "--OSCMulticastAddress [val] | -om [val]: Sets the multicast group address to listen for OSC packets on.";
			cout << endl;
			cout << "--OSCPort [val] | -op [val]:\tSets the port to listen for OSC packets on.";
			cout << endl;
			cout << "--OSCOutputAddress [val] | -ooa [val]: Sets the address to transmit OSC packets to.";
			cout << endl;
			cout << "--OSCOutputPort [val] | -oop [val]: Sets the port to transmit OSC packets to.";
			cout << endl;
			cout << "--noAudio | -na:\t\tDisables sound.";
			cout << endl;
			cout << "--noOSC | -nO:\t\t\tDisables Open Sound Control support.";
			cout << endl;
			cout << "--fullscreen | -f:\t\tSets whether to go fullscreen or not.";
			cout << endl;
			cout << "--bloom | -b:\t\tTurns on a bloom filter post-processing.  This requires a graphics card with support for frame buffer objects and shaders.";
			cout << endl;
			cout << "--offline [seconds] [file] | -of [seconds] [file]: Renders visuals and audio offline for [seconds]s of the scene.  Renders video to a sequence of jpegs (named [file]xxxx.jpg), and audio to a number of wav files (named [file]xx.wav).";
			cout << endl;
			cout << "--recordOSC [file] | -rO [file]: Saves all OSC messages input to a text file for later playback.";
			cout << endl;
			cout << "--cursor | -c:\t\tSets whether to display a mouse cursor or not.";
			cout << endl;
			cout << "--help:\t\t\t\tDisplay this help.";
			cout << endl;
			cout << "--version | -v:\t\t\tDisplay version.";
			cout << endl;
			cout << "--nodes | -n:\t\t\tDisplay available X3D nodes.";
			cout << endl;
			cout << "--enumerateSoundcards:\t\t\tDisplay values for the available audio APIs and devices.";
			cout << endl;
			cout << endl;
			cout << "audioSpeakers Usage:";
			cout << endl;
			cout << "0: Mono";
			cout << endl;
			cout << "1: Stereo";
			cout << endl;
			cout << "2: Square";
			cout << endl;
			cout << "3: Pentagon";
			cout << endl;
			cout << "4: Hexagon";
			cout << endl;
			cout << "5: Octagon1";
			cout << endl;
			cout << "6: Octagon2";
			cout << endl;
			cout << "7: Surround1";
			cout << endl;
			cout << "8: Cube";
			cout << endl;
			cout << "9: Dodecahedron1";
			cout << endl;
			cout << "10: Dodecahedron2";
			cout << endl;
			cout << "11: Components - Bypasses the decoder, outputs the 4 Ambisonic components; W,X,Y,Z.";
			cout << endl;
			cout << "12 <file>: Custom - Loads the speaker configuration from <file>.";
			cout << endl;
			cout << "\t   (see SpeakerConfigurationExample.xml)";
			cout << endl;
			throw FatalError("\n");
		}
	}

	//Determine whether to print the version from the command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--version") || ((*it) == "-v"))
		{
			cout << endl;
			cout << endl;
			cout << "HeilanBrowser: An X3D browser specialised for audio work.";
			cout << endl;
			cout << "Version: " << version;
			cout << endl;
			throw FatalError("\n");
		}
	}

	//Determine whether to print the available nodes from the command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--nodes") || ((*it) == "-n"))
		{
			initFactory();

			//Load the standard X3D nodes.
			cout << endl;
			cout << endl;
			cout << "HeilanBrowser available nodes:" << endl;
			cout << "------------------------------" << endl;
			cout << endl;
			cout << "--------------------------------------------------------------------------------" << endl;
			cout << "Standard X3D nodes:" << endl;
			cout << endl;
			HeilanX3D::NodeFactory::getInstance().printNodes();
			cout << endl << HeilanX3D::NodeFactory::getInstance().getNumNodes();
			cout << " nodes." << endl;
			cout << endl;
			cout << endl;
			HeilanX3D::NodeFactory::getInstance().unregisterAll();

			//Load any libraries specified on the command line.
			CmdLineParams::const_iterator it2;
			for(it2=commandLine.begin();it2!= commandLine.end();++it2)
			{
				if(((*it2) == "--library") || ((*it2) == "-l"))
				{
					++it2;
					loadLibrary(*it2);

					cout << endl;
					cout << "--------------------------------------------------------------------------------" << endl;
					cout << *it2 << " nodes:" << endl;
					cout << endl;
					HeilanX3D::NodeFactory::getInstance().printNodes();
					cout << endl << HeilanX3D::NodeFactory::getInstance().getNumNodes();
					cout << " nodes." << endl;
					cout << endl;
					cout << endl;
					HeilanX3D::NodeFactory::getInstance().unregisterAll();
				}
			}

			throw FatalError("\n");
		}
	}

	//Determine whether to print the available nodes from the command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if((*it) == "--enumerateSoundcards")
		{
			SoundOutput::enumerateSoundcards();

			throw FatalError("\n");
		}
	}

	//Initialise SDL.
	if(SDL_Init(SDL_INIT_TIMER|SDL_INIT_VIDEO) < 0)
	{
		ostringstream err;
		err << "Could not initialise SDL: " << SDL_GetError();
		throw FatalError(err.str());
	}

	//Set window name.
	SDL_WM_SetCaption("Heilan Browser", 0);

	//Initialise the NodeFactory and NavigationFactory.
	//initFactory();

	//Determine width from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--width") || ((*it) == "-w"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> width;
				tempConverter.clear();
			}
			else
				break;
		}
	}
	if(!width)
	{
		width = 640;
		cout << "Using default width = 640" << endl;
	}

	//Determine height from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--height") || ((*it) == "-h"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> height;
				tempConverter.clear();
			}
			else
				break;
		}
	}
	if(!height)
	{
		height = 480;
		cout << "Using default height = 480" << endl;
	}

	//Determine whether to go fullscreen from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--fullscreen") || ((*it) == "-f"))
			fullscreen = true;
	}

	//Determine whether to go fullscreen from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--bloom") || ((*it) == "-b"))
			bloomFilter = true;
	}

	//Determine whether to use antialiasing from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--antialias") || ((*it) == "-a"))
		{
			antialias = true;
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> aaLevel;
				tempConverter.clear();
			}
			else
				break;
		}
	}
	if(antialias && (aaLevel < 2))
	{
		aaLevel = 2;
		cout << "Using 2x antialiasing." << endl;
	}

	//Determine audio API from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--audioAPI") || ((*it) == "-aa"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> audioApi;
				tempConverter.clear();
			}
			else
				break;
		}
	}

	//Determine audio device from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--audioDevice") || ((*it) == "-ad"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> audioDevice;
				tempConverter.clear();
			}
			else
				break;
		}
	}

	//Determine audio device from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--audioInputDevice") || ((*it) == "-ai"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> audioInputDevice;
				tempConverter.clear();
			}
			else
				break;
		}
	}

	//Determine audio samplerate from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--audioSamplerate") || ((*it) == "-as"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> audioSamplerate;
				tempConverter.clear();
				samplerate = static_cast<unsigned int>(audioSamplerate);
			}
			else
				break;
		}
	}

	//Determine audio speakers from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--audioSpeakers") || ((*it) == "-ap"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> audioSpeakers;
				tempConverter.clear();

				if(audioSpeakers == 12)
				{
					++it;
					if(it != commandLine.end())
						speakers = new CustomSpeakerConfig(*it);
					else
						speakers = new StereoConfig();
					numChannels = speakers->getNumChannels();
				}
			}
			else
				break;
		}
	}
	switch(audioSpeakers)
	{
		case 0:
			speakers = new MonoConfig();
			break;
		case 1:
			speakers = new StereoConfig();
			break;
		case 2:
			speakers = new SquareConfig();
			break;
		case 3:
			speakers = new PentagonConfig();
			break;
		case 4:
			speakers = new HexagonConfig();
			break;
		case 5:
			speakers = new Octagon1Config();
			break;
		case 6:
			speakers = new Octagon2Config();
			break;
		case 7:
			speakers = new Surround1Config();
			break;
		case 8:
			speakers = new CubeConfig();
			break;
		case 9:
			speakers = new Dodecahedron1Config();
			break;
		case 10:
			speakers = new Dodecahedron2Config();
			break;
		case 11:
			speakers = new ComponentsConfig();
			break;
	}
	numChannels = speakers->getNumChannels();

	//Determine audio block size from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--audioBlockSize") || ((*it) == "-ab"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> audioBlockSize;
				tempConverter.clear();
			}
			else
				break;
		}
	}

	//Determine OSC multicast group address from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--OSCMulticastAddress") || ((*it) == "-om"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> oscMulticastGroup;
				tempConverter.clear();
			}
			else
				break;
		}
	}

	//Determine OSC port from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--OSCPort") || ((*it) == "-op"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> oscPort;
				tempConverter.clear();
			}
			else
				break;
		}
	}

	//Determine OSC output address from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--OSCOutputAddress") || ((*it) == "-ooa"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> oscOutAddress;
				tempConverter.clear();
			}
			else
				break;
		}
	}

	//Determine OSC output port from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--OSCOutputPort") || ((*it) == "-oop"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> oscOutPort;
				tempConverter.clear();
			}
			else
				break;
		}
	}

	//Determine whether audio should be disabled from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--noAudio") || ((*it) == "-na"))
			noAudio = true;
	}

	//Determine whether OSC should be disabled from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--noOSC") || ((*it) == "-nO"))
			noOSC = true;
	}

	//Determine whether to render offline.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--offline") || ((*it) == "-of"))
		{
			++it;
			if(it != commandLine.end())
			{
				tempConverter << *it;
				tempConverter >> runningTime;
				tempConverter.str("");

				//Convert runningTime to samples.
				runningTime *= samplerate;

				++it;
				if(it != commandLine.end())
					offscreenName = *it;
			}
		}
	}

	//Determine whether to record OSC messages.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--recordOSC") || ((*it) == "-rO"))
		{
			++it;
			if(it != commandLine.end())
			{
				string tempFile;

				tempConverter << *it;
				tempConverter >> tempFile;
				tempConverter.str("");

				OscRecorder::getInstance()->setFile(tempFile);
			}
		}
	}

	//Determine whether to display a mouse cursor from command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--cursor") || ((*it) == "-c"))
			cursor = true;
	}

	//Load any libraries specified on the command line.
	for(it=commandLine.begin();it!= commandLine.end();++it)
	{
		if(((*it) == "--library") || ((*it) == "-l"))
		{
			++it;
			loadLibrary(*it);
		}
	}

	//Get the X3D file to load.
	fileToLoad = commandLine[commandLine.size()-1];
	//Error checking.
	if(fileToLoad.length() > 4)
	{
		//Is not an x3d file.
		if(fileToLoad.substr(fileToLoad.length()-4) != ".x3d")
		{
			cout << "No x3d file selected." << endl;
			fileToLoad = "";
		}
		else //Add filename to window title.
		{
			stringstream tempConv;

			tempConv << fileToLoad << " - HeilanBrowser";
			SDL_WM_SetCaption(tempConv.str().c_str(), 0);
		}
	}
	else //Can't possibly be an x3d file (doesn't have enough characters).
	{
		cout << "No x3d file selected." << endl;
		fileToLoad = "";
	}

	//Set OpenGL attributes (we use 32 bit colour).
	if(SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8) < 0)
		cout << "Could not set red pixel size to 8..." << endl;
	if(SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8) < 0)
		cout << "Could not set green pixel size to 8..." << endl;
	if(SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8) < 0)
		cout << "Could not set blue pixel size to 8..." << endl;
	if(SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8) < 0)
		cout << "Could not set alpha pixel size to 8..." << endl;
	if(SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16) < 0)
		cout << "Could not set depth size to 16..." << endl;
	if(SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1) < 0)
		cout << "Could not set double buffering..." << endl;
	if(antialias)
	{
		if(SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1) < 0)
			cout << "Could not set multi sample buffers..." << endl;
		if(SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, aaLevel) < 0)
			cout << "Could not set multi sample samples..." << endl;
	}

	//Set the video mode.
	tempFlags = flags;
	if(fullscreen)
		tempFlags |= SDL_FULLSCREEN;

	if(SDL_SetVideoMode(width, height, 0, tempFlags) == 0)
	{
		ostringstream err;
		err << "Could not set video mode: " << SDL_GetError();
		throw FatalError(err.str());
	}

	//Initialise GLEW.
	GLenum glewErr = glewInit();
	if(glewErr != GLEW_OK)
	{
		cout << "Could not initialise GLEW, there may be monsters ahead..." << endl;
		cout << "GLEW error: " << glewGetErrorString(glewErr) << endl;
	}

#ifdef HAVE_FFMPEG
	//Initialise ffmpeg.
	av_register_all();
#endif

	//Try to setup vertical syncing.
/*#ifdef HAVE_GLXEXT
	int tempint32 = glXSwapIntervalSGI(1);
	if(tempint32 == GLX_BAD_VALUE)
		cout << "glXSwapIntervalSGI() = GLX_BAD_VALUE" << endl;
	else if(tempint32 == GLX_BAD_CONTEXT)
		cout << "glXSwapIntervalSGI() = GLX_BAD_CONTEXT" << endl;
	else
		cout << "Swap control is now on?" << endl;
#endif*/

	//Hide the mouse cursor.
	SDL_ShowCursor(SDL_DISABLE);

	//Initialise OpenGL stuff.
	initGraphics();
	//HeilanX3D::AvailableExtensions::initExtensions();

	//Initialise the NodeFactory and NavigationFactory.
	initFactory();

	if(runningTime)
		audioBlockSize = 1;
	browser = new Browser(audioSamplerate,
						  *speakers,
						  audioBlockSize,
						  audioApi,
						  audioDevice,
						  audioInputDevice);
	if(noAudio || runningTime)
		browser->setNoAudio();
	if(noOSC)
		browser->setNoOSC();
	else
	{
		browser->setOSCMulticastGroup(oscMulticastGroup);
		browser->setOSCPort(oscPort);
		browser->setOSCOutDetails(oscOutAddress, oscOutPort);
	}

	if(bloomFilter)
		browser->setBloomFilter();
	browser->setCursor(cursor);

	if(runningTime)
	{
		HeilanX3D::Camera::getInstance()->setOfflineRendering(true);
		browser->setOffline();
	}

	browser->loadFile(fileToLoad);
	//Make sure any OpenGL operations that need to happen in node constructors
	//get through.
	SDL_GL_SwapBuffers();

	if(speakers)
		delete speakers;
}

//------------------------------------------------------------------------------
Application::~Application()
{
	unsigned int i;

	if(browser)
		delete browser;

	HeilanX3D::NodeFactory::getInstance().unregisterAll();
	for(i=0;i<libraries.size();++i)
		SDL_UnloadObject(libraries[i]);
}

//------------------------------------------------------------------------------
void Application::eventLoop()
{
	Uint32 tempint;
	SDL_Event event;

	if((runningTime) && (offscreenName != ""))
	{
		cout << "Starting offscreen render." << endl;
		renderToFile();
		return;
	}

	while(1)
	{
		//Handle events.
		while(SDL_PollEvent(&event))
		{
			switch(event.type)
			{
				case SDL_KEYDOWN:
					keyDown(event.key.keysym);
					break;
				case SDL_KEYUP:
					keyUp(event.key.keysym);
					break;
				case SDL_MOUSEBUTTONDOWN:
					if(event.button.button == SDL_BUTTON_WHEELUP)
						mouseWheelUp();
					else if(event.button.button == SDL_BUTTON_WHEELDOWN)
						mouseWheelDown();
					else
					{
						mouseDown(event.button.button,
							  event.button.x,
							  event.button.y);
					}
					break;
				case SDL_MOUSEBUTTONUP:
					if((event.button.button != SDL_BUTTON_WHEELUP) &&
					   (event.button.button != SDL_BUTTON_WHEELDOWN))
					{
						mouseUp(event.button.button,
							event.button.x,
							event.button.y);
					}
					break;
				case SDL_MOUSEMOTION:
					mouseMove(event.motion.x, event.motion.y);
					break;
				case SDL_VIDEORESIZE:
					width = event.resize.w;
					height = event.resize.h;
					if(!SDL_SetVideoMode(width, height, 0, flags))
						cout << "Error changing video mode." << endl;
					initGraphics();
					break;
				case SDL_QUIT:
					timeToQuit = true;
					break;
			}
		}

		if(timeToQuit)
			break;

		//Repaint the screen, or wait a bit first.
		//We aim for a 30fps frame rate here.
		tempint = lastRepaint;
		lastRepaint = SDL_GetTicks();
		if((lastRepaint - tempint) < 30)
			SDL_Delay(30 - (lastRepaint - tempint));
		browser->paint();
		glFinish();
		SDL_GL_SwapBuffers();
	}
}

//------------------------------------------------------------------------------
void Application::renderToFile()
{
	int i;
	bool timeToQuit = false;
	SDL_Event event;

	//Used to keep track of where we are in the timeline (the user will set
	//how long we should run for).
	unsigned int totalCounter = 0;
	//Keeps track of the sample we're at in the current second.
	unsigned int audioCounter = 0;
	//Keeps track of the current visual frame we're at in the current second.
	//(we run at 30fps)
	unsigned int visualCounter = 0;
	//How many audio samples to update the visuals at.
	unsigned int visualSamples = (unsigned int)((float)samplerate/30.0f);
	//Keeps track of the current visual frame we're at in the whole timeline.
	unsigned int visualFrame = 0;

	//Used to get the name of the various files written.
	stringstream tempstr;
	stringstream tempstr2;

	//Pixel data to store images in.
	unsigned char *pixels = new unsigned char[width*height*3];
	cout << width << "x" << height << endl;

	//Array of SNDFILEs for each of the audio files to be written.
	SNDFILE **audioOutput = new SNDFILE *[numChannels];
	//Array of SF_INFOs for each of the audio files to be written.
	SF_INFO *audioInfo = new SF_INFO[numChannels];
	//Buffer of numChannels audio samples.
	float *audioBuffer = new float[numChannels];
	//CallbackStruct from SoundOutput.
	CallbackStruct *callbackStruct = browser->getSoundOutput()->getCallbackStruct();

	assert(runningTime > 0);

#ifndef HAVE_LIBJPEG
	cout << "libjpeg support was not compiled - cannot write images." << endl;
#endif

	for(i=0;i<numChannels;++i)
	{
		tempstr << offscreenName << i << ".wav";
		audioInfo[i].samplerate = samplerate;
		audioInfo[i].channels = 1;
		audioInfo[i].format = SF_FORMAT_WAV|SF_FORMAT_PCM_16;
		audioOutput[i] = sf_open(tempstr.str().c_str(),
								 SFM_WRITE,
								 &(audioInfo[i]));
		if(!audioOutput[i])
			cout << "Could not create file: " << tempstr.str() << endl;
		tempstr.str("");
	}

	while(totalCounter < runningTime)
	{
		//Handle events, so we can still quit the app before it's due.
		while(SDL_PollEvent(&event))
		{
			switch(event.type)
			{
				case SDL_KEYDOWN:
					if(event.key.keysym.sym == SDLK_ESCAPE)
						timeToQuit = true;
					break;
				case SDL_KEYUP:
					keyUp(event.key.keysym);
					break;
				case SDL_QUIT:
					timeToQuit = true;
					break;
			}
		}

		if(timeToQuit)
			break;

		//Update audio, write to file.
		SoundOutput::portAudioCallback(0,
									   audioBuffer,
									   1,
									   0,
									   0,
									   (void *)callbackStruct);
		for(i=0;i<numChannels;++i)
		{
			//This is a major hack, and I really shouldn't be doing it,
			//but it's easier than attenuating the offending audio files...
			*(audioBuffer+i) *= 0.75f;

			if(*(audioBuffer+i) > 1.0f)
				*(audioBuffer+i) = 1.0f;
			else if(*(audioBuffer+i) < -1.0f)
				*(audioBuffer+i) = -1.0f;
			sf_writef_float(audioOutput[i], (audioBuffer+i), 1);
		}

		if((visualCounter%visualSamples) == 0)
		{
			//Update visuals, write to file.

			browser->paint();
			glFinish();

			SDL_GL_SwapBuffers();

			memset(pixels, (width*height*3), 0);
			//Get pixels from opengl frame buffer.
			glReadPixels(0,					//X start pos.
						 0,					//Y start pos.
						 width,				//Width.
						 height,			//Height.
						 GL_RGB,			//What data we're reading.
						 GL_UNSIGNED_BYTE,	//Format to write to.
						 (void *)pixels);	//Buffer to write to.
			//Not really necessary, I don't think, but I might as well...
			//SDL_GL_SwapBuffers();

			//Write to a jpeg file.
#ifdef HAVE_LIBJPEG
			tempstr.str("");
			tempstr2.str("");
			tempstr2.width(5);
			tempstr2.fill('0');
			tempstr2 << visualFrame;
			tempstr << offscreenName << tempstr2.str() << ".jpg";
			writeJpeg(pixels, tempstr.str().c_str());
			cout << "Wrote visual frame: " << tempstr.str() << endl;
#endif
			++visualFrame;
		}

		++totalCounter;
		++visualCounter;
		++audioCounter;
		if(audioCounter >= samplerate)
		{
			audioCounter = 0;
			visualCounter = 0;
			cout << "Wrote " << totalCounter/samplerate << " seconds." << endl;
		}
	}

	for(i=0;i<numChannels;++i)
		sf_close(audioOutput[i]);
	cout << "done." << endl;

	delete [] audioBuffer;
	delete [] audioInfo;
	delete [] audioOutput;
	delete [] pixels;
}

//------------------------------------------------------------------------------
void Application::keyDown(SDL_keysym keyPressed)
{
	if(keyPressed.sym == SDLK_ESCAPE)
		timeToQuit = true;
	else if(keyPressed.sym == SDLK_f)
	{
		fullscreen = !fullscreen;

		if(fullscreen)
			SDL_SetVideoMode(width, height, 0, flags|SDL_FULLSCREEN);
		else
			SDL_SetVideoMode(width, height, 0, flags);
		initGraphics();
		browser->init();
	}

	browser->keyDown(keyPressed.sym, keyPressed.mod);
	HeilanX3D::KeyboardInput::getInstance().keyDown(keyPressed.sym);
}

//------------------------------------------------------------------------------
void Application::keyUp(SDL_keysym keyPressed)
{
	browser->keyUp(keyPressed.sym, keyPressed.mod);
	HeilanX3D::KeyboardInput::getInstance().keyUp(keyPressed.sym);
}

//------------------------------------------------------------------------------
void Application::mouseDown(int button, int posX, int posY)
{
	browser->mouseDown(button, posX, posY);
}

//------------------------------------------------------------------------------
void Application::mouseUp(int button, int posX, int posY)
{
	browser->mouseUp(button, posX, posY);
}

//------------------------------------------------------------------------------
void Application::mouseMove(int posX, int posY)
{
	browser->mouseMove(posX, posY);
}

//------------------------------------------------------------------------------
void Application::mouseWheelUp()
{
	browser->mouseWheelUp();
}

//------------------------------------------------------------------------------
void Application::mouseWheelDown()
{
	browser->mouseWheelDown();
}

//------------------------------------------------------------------------------
void Application::initGraphics()
{
	glViewport(0, 0, width, height);

	HeilanX3D::Camera::getInstance()->setViewWidth(width);
	HeilanX3D::Camera::getInstance()->setViewHeight(height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f,
				   static_cast<float>(width)/static_cast<float>(height),
				   0.1f,
				   500.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	glEnable(GL_TEXTURE_2D);
	glShadeModel(GL_SMOOTH);
	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	/*glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);*/

	//Lighting stuff.
	GLfloat lightAmbient[] = {0.25f, 0.25f, 0.25f, 1.0f};
	GLfloat lightDiffuse[] = {1.0f, 1.0f, 1.0f, 1.0f};
	GLfloat lightPosition[] = {0.0f, 0.0f, 0.0f, 1.0f};
	glLightfv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
	glLightfv(GL_LIGHT0, GL_POSITION,lightPosition);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);
	glEnable(GL_AUTO_NORMAL);
	glEnable(GL_RESCALE_NORMAL);

	glClear(GL_ACCUM_BUFFER_BIT);
}

//------------------------------------------------------------------------------
void Application::loadLibrary(const std::string& libName)
{
	void *libraryObject;

	libraryObject = SDL_LoadObject(libName.c_str());
	if(libraryObject)
	{
		LibraryFunction libFunc = (LibraryFunction)SDL_LoadFunction(libraryObject,
												   					"libraryInitialise");
		if(libFunc)
		{
			libFunc();
			libraries.push_back(libraryObject);
		}
		else
		{
			cout << "Could not find libraryInitialise() in: " << libName << endl;
			SDL_UnloadObject(libraryObject);
		}
	}
	else
	{
		cout << "Could not load library: " << libName << endl;
		cout << "Error: " << SDL_GetError() << endl;
	}
}

//------------------------------------------------------------------------------
void Application::writeJpeg(unsigned char *pixels, const string& filename)
{
#ifdef HAVE_LIBJPEG
	//This code pretty much copied from libjpeg's example.c.
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	FILE *outfile;
	JSAMPROW row_pointer[1];
	int row_stride;

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);

	if((outfile = fopen(filename.c_str(), "wb")) == NULL)
		cout << "Could not create file: " << filename << endl;
	jpeg_stdio_dest(&cinfo, outfile);

	cinfo.image_width = width;
	cinfo.image_height = height;
	cinfo.input_components = 3;
	cinfo.in_color_space = JCS_RGB;
	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, 95, true);

	jpeg_start_compress(&cinfo, true);

	row_stride = width * 3;	/* JSAMPLEs per row in image_buffer */

	while(cinfo.next_scanline < cinfo.image_height)
	{
		row_pointer[0] = (JSAMPLE *)(&pixels[(cinfo.image_height-1-cinfo.next_scanline) * row_stride]);
		jpeg_write_scanlines(&cinfo, row_pointer, 1);
	}

	jpeg_finish_compress(&cinfo);
	fclose(outfile);

	jpeg_destroy_compress(&cinfo);
#endif
}
