//	Application.h - Class handling the various application bits and pieces.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef APPLICATION_H_
#define APPLICATION_H_

//Stupid MSVC v6 warning about the STL...
#ifdef WIN32
#pragma warning(disable:4786)
#endif

#include "SDL.h"
#include <vector>
#include <string>

typedef std::vector<std::string> CmdLineParams;
class Browser;

/*!
	\mainpage

	\section Introduction

	Heilan is a cross-platform OpenGL X3D browser written in C++ and designed
	for audio work. Specifically, it forms the environment within which I am
	developing an audiovisual instrument for my PhD. As such, it has a number of
	features not commonly found in X3D browsers:

    - Low latency audio courtesy of PortAudio (ASIO, DirectX, MME on Windows;
	Jack, ALSA, OSS on Linux; CoreAudio on OSX).
    - A 1st order B-format Ambisonic audio engine, allowing for full 3d sound,
	capable of accommodating virtually any speaker configuration (at the moment
	this is limited to preset configurations, but that will change in the
	future).
    - Open Sound Control support for all nodes, allowing a node's attributes to
	be manipulated in realtime, potentially by multiple users.
    - A multi-threaded audio engine, able to split off different nodes' audio
	processing into separate threads (and potentially, separate cpus/cores).
	This can be configured by the scene author.

	Heilan aims to conform to the X3D Interchange profile, with additional
	support for certain nodes such as Sound and NurbsPatchSurface (see here for
	a list of supported nodes). It may be extended with libraries which can
	provide extra nodes, navigation types, and sound file loaders. It comes with
	a library containing my own experimental audiovisual nodes, and a couple of
	extra navigation types.

	The browser itself is a command line program designed to be used offline
	(there's no browser plugin or http/ftp support), though there is a GUI
	frontend available as well (included in the binary package). Both are open
	source, licensed under the GPL.


	Developed by Niall Moody at the Centre for Music Technology, University of
	Glasgow.

	\section devInfo Information for Developers

	- \ref NodeTypeCreation : Things to know when creating a new node type.
	- \ref Libraries : The steps involved in creating a library for Heilan.
	- \ref AshitakaParameters : Parameter mapping techniques for the Ashitka
	instrument.
 */

/*!
	\page Libraries Creating a library

	Heilan supports the use of libraries (of which libheilanextras is an
	example) to add extra node types, navigation types, or sound file loaders.
	You should probably be able to figure out the steps to create a library
	yourself from libheilanextras, but here are the most important things to
	note.

\section libraryInitialise libraryInitialise()
	For Heilan to be able to load your library, you must have a
	libraryInitialise() function, in which you register your nodes etc. with the
	browser's NodeFactory instance. A simple example is:
\code
#include "NodeFactory.h"

#include "Thing.h"

using namespace HeilanX3D;

extern "C"
{

//------------------------------------------------------------------------------
void libraryInitialise()
{
	NodeFactory::getInstance().registerNode(&Thing::constructor);
}

}
\endcode

	\section linuxLinking Linking on Linux

	You should link your library to the heilanbrowser executable. To do this on
	Linux, you just need to make sure your CXXFLAGS for the compilation have the
	following flags:
\code
CXXFLAGS="-Bdynamic -shared"
\endcode

	\section windowsLinking Linking on Windows

	On Windows, Heilan is built using MSVC v6. To build your library, you will
	first have to build Heilan from source, which will create an import library
	file, 'HeilanBrowser.lib' along with the executable. Just add this file to
	your list of libraries in the project settings, and your library should
	compile and work with Heilan. 
 */
	
///	The type of the function loaded from external libraries.
/*!
	See the contents of Application::initFactory() for how this function should
	be used in a library.
 */
typedef void (*LibraryFunction)();

///	Class handling the various application bits and pieces.
class Application
{
  public:
	///	Constructor.
	/*!
		\param commandLine A vector containing all the command line parameters
		passed in.

		This will call SDL_Init() with the SDL_INIT_VIDEO and SDL_INIT_TIMER
		flags.  It will also open a window according to the command line
		arguments (e.g. fullscreen/windowed, geometry, bit depth etc.).
	 */
	Application(const CmdLineParams &commandLine);
	///	Destructor.
	/*!
		Will call SDL_Quit().
	 */
	~Application();

	///	The main event loop occurs inside this method.
	void eventLoop();
  private:
	///	Renders the whole scene to a series of files.
	/*!
		This essentially replaces the contents of eventLoop(), to render the
		scene as a series of .jpg image files, and a number of .wav audio files.
		This is so 'compositions' can be easily rendered into an mpeg, or other
		video file.

		Audio in this case is processed as single samples, rather than the usual
		block-based processing.

		The rendering is done in a single thread, but if the user has specified
		parallel audio processing in the x3d file, each audio object will render
		its audio in a separate thread (whether this will be an advantage is
		hard to tell, considering the audio will be processed on a per-sample
		basis).
	 */
	void renderToFile();

	///	Called when a key down event is received.
	void keyDown(SDL_keysym keyPressed);
	///	Called when a key up event is received.
	void keyUp(SDL_keysym keyPressed);

	///	Called when a mouse down event is received.
	void mouseDown(int button, int posX, int posY);
	///	Called when a mouse up event is received.
	void mouseUp(int button, int posX, int posY);
	///	Called when a mouse move event is received.
	void mouseMove(int posX, int posY);
	///	Called when the mouse wheel is scrolled up.
	void mouseWheelUp();
	///	Called when the mouse wheel is scrolled down.
	void mouseWheelDown();

	///	Called to initialise the factory with all our currently-implemented nodes.
	/*!
		This only initialises the nodes from the X3D spec - my own ones are
		in separate libraries, to be loaded with loadLibrary().
	 */
	void initFactory();
	///	Called to initialise opengl stuff.
	void initGraphics();

	///	Called to load the named library.
	void loadLibrary(const std::string& libName);

	///	Helper method to write the contents of pixels to a JPEG file.
	void writeJpeg(unsigned char *pixels, const std::string& filename);

	///	Pointer to our browser.
	Browser *browser;

	///	We set this to true when it's time to quit the program.
	bool timeToQuit;
	///	We use this to determine if we need to repaint the screen immediately.
	Uint32 lastRepaint;

	///	The width of the window.
	int width;
	///	The height of the window.
	int height;
	///	The flags used to set the video mode.
	int flags;
	///	Whether we're in fullscreen mode or not.
	bool fullscreen;

	///	Current version of the software.
	std::string version;

	///	Array containing pointers to any libraries loaded.
	std::vector<void *> libraries;

	///	If we're rendering to a file, this will be the length of the file to write, in samples.
	unsigned int runningTime;
	///	The samplerate we're running at.
	unsigned int samplerate;
	///	Name of the files to write to in offscreen mode.
	std::string offscreenName;
	///	Number of audio channels we're using.
	int numChannels;
};

#endif
