//	Browser.cpp - The browser class.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Browser.h"
#include "AbstractNode.h"
#include "NavigationInfo.h"
#include "X3DFileLoader.h"
#include "Camera.h"
#include "Navigation.h"
#include "FatalError.h"
#include "OpenSoundController.h"
#include "Lighting.h"
#include "OscRecorder.h"
#include "Images.h"
#include "ThreadManager.h"

#include "GLHeaders.h"
#include "SDL.h"
#include "SDL_image.h"
#include <sstream>
#include <iostream>

using namespace std;
using namespace HeilanX3D;

#define Pi 3.1415926535897932384626433832795f

//------------------------------------------------------------------------------
Browser::Browser(float samplerate,
				 const SpeakerConfig& speakers,
				 int blockSize,
				 int api,
				 int device,
				 int inputDevice):
sceneRoot(0),
routedNodes(0),
noAudio(false),
noOSC(false),
offlineRendering(false),
headlight(true),
offlineTicks(0),
tempBuf(0),
bloomFilter(0),
cursor(false),
cursorSize(0.03f, 0.03f),
audioFadeIn(0.0f),
ignoreFirstBlock(true),
clearEventsCount(60)
{
	SDL_Surface *tempIm;

	cout << "Api: " << api << endl;
	cout << "Samplerate: " << samplerate << endl;
	cout << "Block size: " << blockSize << endl;
	cout << "Device: " << device << endl;
	soundOutput.setApi(api);
	soundOutput.setSamplerate(samplerate);
	soundOutput.setCallback(this);
	soundOutput.setSpeakerConfig(speakers);
	soundOutput.setBlockSize(blockSize);
	soundOutput.setDevice(device);
	soundOutput.setInputDevice(inputDevice);

	tempBuf = new float[Camera::getInstance()->getViewWidth() * Camera::getInstance()->getViewHeight()];

	//Generate cursor texture.
	if((tempIm = IMG_Load_RW(SDL_RWFromConstMem(Images::cursor_png,
											    Images::cursor_pngSize), 1)) != 0)
	{
		glGenTextures(1, &cursorTex);

		glBindTexture(GL_TEXTURE_2D, cursorTex);

		glTexImage2D(GL_TEXTURE_2D,
					 0,
					 GL_RGBA,
					 tempIm->w,
					 tempIm->h,
					 0,
					 GL_RGBA,
					 GL_UNSIGNED_BYTE,
					 tempIm->pixels);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		glBindTexture(GL_TEXTURE_2D, 0);

		SDL_FreeSurface(tempIm);
	}

	//Generate no file was loaded texture.
	if((tempIm = IMG_Load_RW(SDL_RWFromConstMem(Images::nofileselected_png,
											    Images::nofileselected_pngSize), 1)) != 0)
	{
		glGenTextures(1, &noFileTex);

		glBindTexture(GL_TEXTURE_2D, noFileTex);

		glTexImage2D(GL_TEXTURE_2D,
					 0,
					 GL_RGBA,
					 tempIm->w,
					 tempIm->h,
					 0,
					 GL_RGBA,
					 GL_UNSIGNED_BYTE,
					 tempIm->pixels);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		glBindTexture(GL_TEXTURE_2D, 0);

		SDL_FreeSurface(tempIm);
	}
}

//------------------------------------------------------------------------------
Browser::~Browser()
{
	if(!noOSC)
		OpenSoundController::getInstance()->stopListening();

	ThreadManager::getInstance().stopThreads();

	if(soundOutput.isRunning())
		soundOutput.stopAudio();

	if(routedNodes)
		delete routedNodes;

	//This is probably completely unnecessary, but we might as well stick to
	//the convention we've already used.
	if(sceneRoot)
	{
		sceneRoot->clearEvents();
		sceneRoot->decRefCount();
	}

	//This is so we don't get segfaults if we're using a Navigation type from a
	//library.
	Camera::getInstance()->setNavigationTypes("EXAMINE");

	if(tempBuf)
		delete [] tempBuf;

	if(bloomFilter)
		delete bloomFilter;
}

//------------------------------------------------------------------------------
void Browser::loadFile(string file)
{
	NavigationInfo *sceneNav = 0;
	X3DFileLoader loader(file);

	if(soundOutput.isRunning())
		soundOutput.stopAudio();

	if(loader.isValidX3DFile())
	{
		//Load the scene.
		sceneRoot = loader.loadScene(true);
		loader.handleIMPORTS();
		routedNodes = loader.getRoutedNodes();
		//Setup the navigation mode.
		sceneNav = static_cast<NavigationInfo *>(BindableNodeSingleton::getInstance()->getActiveNode("NavigationInfo"));
		if(sceneNav)
			Camera::getInstance()->setNavigationTypes(sceneNav->getTypes());
		else
		{
			cout << "No NavigationInfo node in this scene, defaulting to: ";
			cout << "EXAMINE" << endl;
		}

		ThreadManager::getInstance().startThreads();

		//Setup and start audio.
		sceneRoot->setBlockSize(soundOutput.getBlockSize());
		sceneRoot->setSamplerate(soundOutput.getSamplerate());
		if(!noAudio)
			soundOutput.startAudio();

		//Start listening for OSC messages.
		if(!noOSC)
			OpenSoundController::getInstance()->startListening();

		//Initialise the camera.
		Camera::getInstance()->initialise();
	}
	else
	{
		ostringstream err;
		err << "Could not load file: " << file << " (invalid file)";
		//throw FatalError(err.str());
	}
}

//------------------------------------------------------------------------------
void Browser::paint()
{
	unsigned int i;

	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

	Lighting::getInstance().drawHeadlight();

	if(bloomFilter)
		bloomFilter->setup();

	if(sceneRoot)
	{
		//Update navigation if it needs updating.
		//Camera::getInstance()->getNavigation()->screenRefresh();
		//Setup camera.
		Camera::getInstance()->draw();
		//Draw scene.
		sceneRoot->graphicDoStuff();

		//Handle any events.
		--clearEventsCount;
		if(!clearEventsCount)
		{
			clearEventsCount = 60;
			sceneRoot->clearEvents();
		}
		else
		{
			if(routedNodes)
			{
				for(i=0;i<routedNodes->size();++i)
					(*routedNodes)[i]->handleEvents();
			}
		}
	}
	else
	{
		glDisable(GL_LIGHTING);
		glBindTexture(GL_TEXTURE_2D, noFileTex);
		glBegin(GL_QUADS);
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(-1.0f, -1.0f, -1.25f);

			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(1.0f, -1.0f, -1.25f);

			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(1.0f, 1.0f, -1.25f);

			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-1.0f, 1.0f, -1.25f);
		glEnd();
		glBindTexture(GL_TEXTURE_2D, 0);
		glEnable(GL_LIGHTING);
	}

	if(bloomFilter)
		bloomFilter->draw();

	if(cursor)
	{
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();

		glLoadIdentity();
		gluOrtho2D(0.0f, 1.0f, 1.0f, 0.0f);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		glBindTexture(GL_TEXTURE_2D, cursorTex);
		glDisable(GL_LIGHTING);
		glColor3f(1.0f, 1.0f, 1.0f); //Just in case.
		glBegin(GL_QUADS);
		{
			glTexCoord2f(0.0f, 1.0f);
			glVertex2f(mousePos.x-(cursorSize.x*0.33f),
					   mousePos.y+(cursorSize.y*0.66f));
			glTexCoord2f(1.0f, 1.0f);
			glVertex2f(mousePos.x+(cursorSize.x*0.66f),
					   mousePos.y+(cursorSize.y*0.66f));
			glTexCoord2f(1.0f, 0.0f);
			glVertex2f(mousePos.x+(cursorSize.x*0.66f),
					   mousePos.y-(cursorSize.y*0.33f));
			glTexCoord2f(0.0f, 0.0f);
			glVertex2f(mousePos.x-(cursorSize.x*0.33f),
					   mousePos.y-(cursorSize.y*0.33f));
		}
		glEnd();
		glEnable(GL_LIGHTING);
		glBindTexture(GL_TEXTURE_2D, 0);

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
	}

	//Draw depth buffer to screen.
	/*if(tempBuf)
	{
		//glPixelTransferf(GL_DEPTH_SCALE, 0.25f);
		//glPixelTransferf(GL_DEPTH_BIAS, 0.5f);
		glReadPixels(0,
					 0,
					 Camera::getInstance()->getViewWidth(),
					 Camera::getInstance()->getViewHeight(),
					 GL_DEPTH_COMPONENT,
					 GL_FLOAT,
					 tempBuf);
		glDrawPixels(Camera::getInstance()->getViewWidth(),
					 Camera::getInstance()->getViewHeight(),
					 GL_LUMINANCE,
					 GL_FLOAT,
					 tempBuf);
	}*/

	/*glAccum(GL_MULT, 0.8f);
	glAccum(GL_ACCUM, 1.0f-0.8f);
	glAccum(GL_RETURN, 1.0f);*/
}

//------------------------------------------------------------------------------
void Browser::init()
{
	SDL_Surface *tempIm = 0;

	if(sceneRoot)
		sceneRoot->initialise();

	if((tempIm = IMG_Load_RW(SDL_RWFromConstMem(Images::cursor_png,
											   	Images::cursor_pngSize), 1)) != 0)
	{
		glGenTextures(1, &cursorTex);

		glBindTexture(GL_TEXTURE_2D, cursorTex);

		glTexImage2D(GL_TEXTURE_2D,
					 0,
					 GL_RGBA,
					 tempIm->w,
					 tempIm->h,
					 0,
					 GL_RGBA,
					 GL_UNSIGNED_BYTE,
					 tempIm->pixels);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		SDL_FreeSurface(tempIm);
	}
}

//------------------------------------------------------------------------------
void Browser::keyDown(int key, int modifier)
{
	Camera::getInstance()->getNavigation()->keyDown(static_cast<SDLKey>(key));

	if(key == SDLK_l)
	{
		headlight = !headlight;
		Lighting::getInstance().switchHeadlight(headlight);
	}
}

//------------------------------------------------------------------------------
void Browser::keyUp(int key, int modifier)
{
	Camera::getInstance()->getNavigation()->keyUp(static_cast<SDLKey>(key));
}

//------------------------------------------------------------------------------
void Browser::mouseDown(int button, int posX, int posY)
{
	Camera::getInstance()->getNavigation()->mouseDown(button, posX, posY);
}

//------------------------------------------------------------------------------
void Browser::mouseUp(int button, int posX, int posY)
{
	Camera::getInstance()->getNavigation()->mouseUp(button, posX, posY);
}

//------------------------------------------------------------------------------
void Browser::mouseMove(int posX, int posY)
{
	const float width = (float)Camera::getInstance()->getViewWidth();
	const float height = (float)Camera::getInstance()->getViewHeight();

	Camera::getInstance()->getNavigation()->mouseMove(posX, posY);
	mousePos.x = (float)posX/width;
	mousePos.y = (float)posY/height;

	if(width > height)
	{
		cursorSize.x = 0.04f;
		cursorSize.y = (width/height) * 0.04f;
	}
	else
	{
		cursorSize.x = (height/width) * 0.04f;
		cursorSize.y = 0.04f;
	}
}

//------------------------------------------------------------------------------
void Browser::mouseWheelUp()
{
	Camera::getInstance()->getNavigation()->mouseWheelUp();
}

//------------------------------------------------------------------------------
void Browser::mouseWheelDown()
{
	Camera::getInstance()->getNavigation()->mouseWheelDown();
}

//------------------------------------------------------------------------------
void Browser::setCursor(bool val)
{
	cursor = val;
}

//------------------------------------------------------------------------------
AmbisonicData *Browser::processBlock(unsigned int numSamples)
{
	unsigned int i;
	ThreeFloats tempPos;
	AmbisonicData *audio = 0;
	AmbisonicData tempAmb;
	FourFloats tempRot;
	const float samplerate = soundOutput.getSamplerate();

	//Update Camera variables.
	Camera::getInstance()->updateAudioRotation();
	Camera::getInstance()->updateAudioPosition();

	//Set input audio for the scene.
	AbstractNode::getNumInputAudioChannels(getNumInputChannels());
	AbstractNode::getInputAudio(getInputAudio(), true);

	//Get audio for the scene.
	if(sceneRoot)
	{
		sceneRoot->audioBlockStart();
		//audio = sceneRoot->audioDoStuff(numSamples, tempPos);
		audio = sceneRoot->getAmbisonicAudio(numSamples, tempPos);

		//We discard the first block because for some reason we get a nasty
		//click on OSX.  This is a fairly ugly hack, but I can't see where
		//the click is actually coming from (all the buffers should be
		//zeroed), so it'll have to do for now.
		if(ignoreFirstBlock)
		{
			ignoreFirstBlock = false;
			for(i=0;i<numSamples;++i)
				audio[i] = 0.0f;
			return audio;
		}

		if(audioFadeIn < 1.0f)
		{
			for(i=0;i<numSamples;++i)
			{
				audio[i] *= audioFadeIn;
				audioFadeIn += 0.0001f;
				if(audioFadeIn >= 1.0f)
				{
					audioFadeIn = 1.0f;
					break;
				}
			}
		}
	}

	//Rotate sound scene by the camera's rotation.
	/*tempRot = Camera::getInstance()->getAudioRotation();
	tempRot.x *= (Pi/180.0f);
	tempRot.y *= (Pi/180.0f);
	for(i=0;i<numSamples;++i)
	{
		tempAmb = audio[i];

		//We only rotate around the (Ambi) Z and Y axes.
		audio[i].X = (tempAmb.X * cosf(-tempRot.x) * cosf(-tempRot.y)) -
					 (tempAmb.Y * sinf(-tempRot.x) * cosf(-tempRot.y)) -
					 (tempAmb.Z * sinf(-tempRot.y));
		audio[i].Y = (tempAmb.X * sinf(-tempRot.x)) +
					 (tempAmb.Y * cosf(-tempRot.x));
		audio[i].Z = (tempAmb.X * cosf(-tempRot.x) * sinf(-tempRot.y)) -
					 (tempAmb.Y * sinf(-tempRot.x) * sinf(-tempRot.y)) +
					 (tempAmb.Z * cosf(-tempRot.y));
	}*/

	//Clear AbstractNode::inputAudio.
	AbstractNode::getInputAudio(getInputAudio(), true);

	offlineTicks += numSamples;
	//if(noAudio)
		Camera::getInstance()->setOfflineTicks((unsigned int)((float)offlineTicks/(samplerate/1000.0f)));

	//Update navigation position in the audio thread to ensure we get the same
	//timing when rendering offline as in realtime.
	Camera::getInstance()->getNavigation()->screenRefresh();

	//Update OscRecorder with the current sample value.
	OscRecorder::getInstance()->updateSampleVal(numSamples);

	return audio;
}

//------------------------------------------------------------------------------
void Browser::setOSCMulticastGroup(const string& address)
{
	OpenSoundController::getInstance()->setMulticastGroup(address);
}

//------------------------------------------------------------------------------
void Browser::setOSCPort(int port)
{
	OpenSoundController::getInstance()->setPort(port);
}

//------------------------------------------------------------------------------
void Browser::setOSCOutDetails(const string& address, int port)
{
	OpenSoundController::getInstance()->setOutDetails(address, port);
}

//------------------------------------------------------------------------------
void Browser::setNoAudio()
{
	noAudio = true;
	if(soundOutput.isRunning())
		soundOutput.stopAudio();
}

//------------------------------------------------------------------------------
void Browser::setNoOSC()
{
	noOSC = true;
}

//------------------------------------------------------------------------------
void Browser::setOffline()
{
	offlineRendering = true;
}

//------------------------------------------------------------------------------
void Browser::setBloomFilter()
{
	bloomFilter = new BloomFilter();
}
