//	Browser.h - The browser class.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef BROWSER_H_
#define BROWSER_H_

#include "SoundOutput.h"
#include "BloomFilter.h"
#include "TwoFloats.h"

#include <string>
#include <vector>

namespace HeilanX3D
{
class AbstractNode;
typedef std::vector<AbstractNode *> RoutedNodes;
}

//------------------------------------------------------------------------------
///	The browser class.
/*!
	Takes care of the various vectors of nodes, drawing the scene, audio, and
	OSC.

	\todo Get audio rotation working correctly.
 */
class Browser : public SoundOutputCallback
{
  public:
	///	Constructor.
	/*!
		\param samplerate The samplerate to use for the audio.
		\param speakers The speaker configuration to use for the audio.
		\param blockSize The block size to use for the audio (0 = default/use
		portaudio preffered size).
		\param api Index to the audio API to use (-1 = use default API,
		according to PortAudio).
		\param device Index to the audio device to use (-1 = use default
		device).
		\param inputDevice Index to the input audio device to use (-1 = use
		default device).
	 */
	Browser(float samplerate = 44100.0f,
			const SpeakerConfig& speakers = StereoConfig(),
			int blockSize = 0,
			int api = -1,
			int device = -1,
			int inputDevice = -1);
	///	Destructor.
	~Browser();

	///	Called to load the X3D scene to display.
	void loadFile(std::string file);

	///	Draws the scene.
	void paint();

	///	Calls initialise() on all nodes.
	void init();

	///	Called when a key is pressed.
	void keyDown(int key, int modifier);
	///	Called when a key is released.
	void keyUp(int key, int modifier);

	///	Called when a mouse button is clicked.
	void mouseDown(int button, int posX, int posY);
	///	Called when a mouse button is released.
	void mouseUp(int button, int posX, int posY);
	///	Called when the mouse moves.
	void mouseMove(int posX, int posY);
	///	Called when the mouse wheel scrolls up.
	void mouseWheelUp();
	///	Called when the mouse wheel scrolls down.
	void mouseWheelDown();
	///	Turns on/off the mouse cursor.
	void setCursor(bool val);

	///	Method called for every audio block, to process/generate audio from the scene.
	AmbisonicData *processBlock(unsigned int numSamples);

	///	Call this to set the multicast group to listen for OSC packets on.
	void setOSCMulticastGroup(const std::string& address);
	///	Call this to set the port to listen for OSC packets on.
	void setOSCPort(int port);
	///	Call this to set the address and port to send OSC messages to.
	void setOSCOutDetails(const std::string& address, int port);

	///	Call this to switch audio off.
	void setNoAudio();
	///	Call this to switch OSC support off.
	void setNoOSC();
	///	Call this to render everything offline.
	void setOffline();
	///	Call this to activate the bloom filter.
	void setBloomFilter();

	///	Returns a pointer to Browser's SoundOutput object.
	SoundOutput *getSoundOutput() {return &soundOutput;};
  private:
	///	The root node of the current X3D scene, used to draw it.
	HeilanX3D::AbstractNode *sceneRoot;
	///	Sound output device.
	SoundOutput soundOutput;
	///	Vector of any nodes connected via ROUTEs.
	HeilanX3D::RoutedNodes *routedNodes;

	///	True if the user does not want audio.
	bool noAudio;
	///	True if the user does not want OSC.
	bool noOSC;
	///	True if we're rendering offline.
	bool offlineRendering;

	///	Used to turn the headlight on or off.
	bool headlight;

	///	Our current sample in the audio (to calculate our offline ticks value).
	unsigned int offlineTicks;

	///	temp.
	float *tempBuf;

	///	Bloom filter for post-processing.
	BloomFilter *bloomFilter;

	///	Whether or not to display a mouse cursor.
	bool cursor;
	///	The texture to draw the cursor as.
	GLuint cursorTex;
	///	The current screen position of the mouse.
	TwoFloats mousePos;
	///	The cursor's current size (dependent on window dimensions).
	TwoFloats cursorSize;

	///	Used to fade in the audio on startup.
	float audioFadeIn;
	///	Used to ignore the first audio block to avoid a big click on OSX.
	bool ignoreFirstBlock;

	///	The texture to draw when no file was selected.
	GLuint noFileTex;

	///	Used to determine when to call AbstractNode::clearEvents().
	int clearEventsCount;
};

#endif
