//	BloomFilter.cpp - Class to simplify applying a post-processing bloom filter.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "BloomFilter.h"
#include "GLHeaders.h"
#include "Camera.h"

#include <iostream>
#include <cmath>

using namespace std;
using namespace HeilanX3D;

const char *thresholdShaderVertChar = "uniform sampler2D displacementMap;"\
	" "\
	"void main(void)"\
	"{"\
	"	gl_TexCoord[0] = gl_MultiTexCoord0;"\
	"	gl_Position = ftransform();"\
	"	gl_FrontColor = gl_Color;"\
	"}";

const char *thresholdShaderFragChar = "uniform sampler2D colorMap;"\
	" "\
	"void main(void)"\
	"{"\
	"	vec4 sum;"\
	"	float tempf;"\
	" "\
	"	sum = texture2D(colorMap, gl_TexCoord[0].st);"\
	"	tempf = clamp(step(1.5, (sum[0]+sum[1]+sum[2])), 0.0, 1.0);"\
	"	sum = vec4(tempf, tempf, tempf, 0.0);"\
	" "\
	"	gl_FragColor = sum + gl_Color;"\
	"}";

const char *blurShaderVertChar = "uniform sampler2D displacementMap;"\
	" "\
	"void main(void)"\
	"{"\
	"	gl_TexCoord[0] = gl_MultiTexCoord0;"\
	"	gl_Position = ftransform();"\
	"	gl_FrontColor = gl_Color;"\
	"}";

const char *blurShaderFragChar = "float kernel[9];"\
	" "\
	"uniform sampler2D colorMap;"\
	" "\
	"uniform vec2 offsets[9];"\
	" "\
	"void main(void)"\
	"{"\
	"int i = 0;"\
	"vec4 sum = vec4(0.0);"\
	" "\
	"if(gl_TexCoord[0].st[0] > 0.0)"\
	"{"\
	"kernel[0] = 1.0/16.0; 	kernel[1] = 2.0/16.0;	kernel[2] = 1.0/16.0;"\
	"kernel[3] = 2.0/16.0;	kernel[4] = 4.0/16.0;	kernel[5] = 2.0/16.0;"\
	"kernel[6] = 1.0/16.0;	kernel[7] = 2.0/16.0;	kernel[8] = 1.0/16.0;"\
	" "\
	"for(i=0;i<9;i++)"\
	"{"\
	"vec4 tmp = texture2D(colorMap, gl_TexCoord[0].st + offsets[i]);"\
	"sum += tmp * kernel[i];"\
	"}"\
	" "\
	"sum[3] = gl_Color[3];"\
	"}"\
	"gl_FragColor = sum;"\
	"}";

//------------------------------------------------------------------------------
BloomFilter::BloomFilter():
extensionsAvailable(false)
{
	int width = getPowerOf2(Camera::getInstance()->getViewWidth());
	int height = getPowerOf2(Camera::getInstance()->getViewHeight());

	if(glCreateShaderObjectARB && glGenFramebuffersEXT)
	{
		extensionsAvailable = true;

		//----Setup Frame Buffer Objects----
		glGenFramebuffersEXT(1, &frameBuffer);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, frameBuffer);

		// Create the render buffer for depth	
		glGenRenderbuffersEXT(1, &depthBuffer);
		glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depthBuffer);
		glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,
								 GL_DEPTH_COMPONENT,
								 width,
								 height);

		// Now setup a texture to render to.
		glGenTextures(1, &offscreenTexture);
		glBindTexture(GL_TEXTURE_2D, offscreenTexture);
		glTexImage2D(GL_TEXTURE_2D,
					 0,
					 GL_RGBA8,
					 width,
					 height,
					 0,
					 GL_RGBA,
					 GL_UNSIGNED_BYTE,
					 NULL);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

		// And attach it to the FBO so we can render to it
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
								  GL_COLOR_ATTACHMENT0_EXT,
								  GL_TEXTURE_2D,
								  offscreenTexture,
								  0);

		// Attach the depth render buffer to the FBO as it's depth attachment
		glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
									 GL_DEPTH_ATTACHMENT_EXT,
									 GL_RENDERBUFFER_EXT,
									 depthBuffer);

		//Setup thresholdTexture.
		glGenTextures(1, &thresholdTexture);
		glBindTexture(GL_TEXTURE_2D, thresholdTexture);
		glTexImage2D(GL_TEXTURE_2D,
					 0,
					 GL_RGBA8,
					 width,
					 height,
					 0,
					 GL_RGBA,
					 GL_UNSIGNED_BYTE,
					 NULL);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

		//And attach it to the FBO so we can render to it
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
								  GL_COLOR_ATTACHMENT1_EXT,
								  GL_TEXTURE_2D,
								  thresholdTexture,
								  0);
		if(glGetError() != GL_NO_ERROR)
			cout << "thresholdTexture error." << endl;

		//Setup downsampledTexture.
		glGenTextures(1, &downsampledTexture);
		glBindTexture(GL_TEXTURE_2D, downsampledTexture);
		glTexImage2D(GL_TEXTURE_2D,
					 0,
					 GL_RGBA8,
					 width,
					 height,
					 0,
					 GL_RGBA,
					 GL_UNSIGNED_BYTE,
					 NULL);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

		//And attach it to the FBO so we can render to it
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
								  GL_COLOR_ATTACHMENT2_EXT,
								  GL_TEXTURE_2D,
								  downsampledTexture,
								  0);
		if(glGetError() != GL_NO_ERROR)
			cout << "downsampledTexture error." << endl;

		GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
		if(status != GL_FRAMEBUFFER_COMPLETE_EXT)
		{
			cout << "Frame buffer error: " << status << endl;
			switch(status)
			{
				case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
					cout << "FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT" << endl;
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
					cout << "FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT" << endl;
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
					cout << "FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT" << endl;
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
					cout << "FRAMEBUFFER_INCOMPLETE_FORMATS_EXT" << endl;
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
					cout << "FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT" << endl;
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
					cout << "FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT" << endl;
					break;
				case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
					cout << "FRAMEBUFFER_UNSUPPORTED_EXT" << endl;
					break;
			}
		}

		//Unbind the FBO for now
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);


		//----Setup shaders----
		thresholdShaderVert = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
		thresholdShaderFrag = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
	
		glShaderSourceARB(thresholdShaderVert,
						  1,
						  (const GLcharARB **)(&thresholdShaderVertChar),
						  NULL);
		glShaderSourceARB(thresholdShaderFrag,
						  1,
						  (const GLcharARB **)(&thresholdShaderFragChar),
						  NULL);
	
		glCompileShaderARB(thresholdShaderVert);
		glCompileShaderARB(thresholdShaderFrag);

		cout << "BloomFilter Threshold Shader InfoLog:" << endl;
		HeilanGL::displayInfoLog(thresholdShaderVert);
		HeilanGL::displayInfoLog(thresholdShaderFrag);
	
		thresholdProgram = glCreateProgramObjectARB();
		glAttachObjectARB(thresholdProgram, thresholdShaderVert);
		glAttachObjectARB(thresholdProgram, thresholdShaderFrag);
	
		glLinkProgramARB(thresholdProgram);
	
		HeilanGL::displayInfoLog(thresholdProgram);


		blurShaderVert = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
		blurShaderFrag = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
	
		glShaderSourceARB(blurShaderVert,
						  1,
						  (const GLcharARB **)(&blurShaderVertChar),
						  NULL);
		glShaderSourceARB(blurShaderFrag,
						  1,
						  (const GLcharARB **)(&blurShaderFragChar),
						  NULL);
	
		glCompileShaderARB(blurShaderVert);
		glCompileShaderARB(blurShaderFrag);

		cout << "BloomFilter Gaussian Blur Shader InfoLog:" << endl;
		HeilanGL::displayInfoLog(blurShaderVert);
		HeilanGL::displayInfoLog(blurShaderFrag);
	
		blurProgram = glCreateProgramObjectARB();
		glAttachObjectARB(blurProgram, blurShaderVert);
		glAttachObjectARB(blurProgram, blurShaderFrag);
	
		glLinkProgramARB(blurProgram);
	
		HeilanGL::displayInfoLog(blurProgram);
	}
	else
	{
		cout << "No shader and/or frame buffer object support: ";
		cout << "bloom filter disabled." << endl;
	}
}

//------------------------------------------------------------------------------
BloomFilter::~BloomFilter()
{
	if(extensionsAvailable)
	{
		glDeleteFramebuffersEXT(1, &frameBuffer);
		glDeleteRenderbuffersEXT(1, &depthBuffer);
		glDeleteTextures(1, &offscreenTexture);
		glDeleteTextures(1, &thresholdTexture);
		glDeleteTextures(1, &downsampledTexture);
	}
}

//------------------------------------------------------------------------------
void BloomFilter::setup()
{
	int width = Camera::getInstance()->getViewWidth();
	int height = Camera::getInstance()->getViewHeight();

	if(extensionsAvailable)
	{
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, frameBuffer);
		glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);

		glPushAttrib(GL_VIEWPORT_BIT);
		glViewport(0, 0, width, height);
		glClearColor(1.0f, 1.0f, 1.0f, 0.5f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();
	}
}

//------------------------------------------------------------------------------
void BloomFilter::draw()
{
	int screenWidth = Camera::getInstance()->getViewWidth();
	int screenHeight = Camera::getInstance()->getViewHeight();
	int width = getPowerOf2(screenWidth);
	int height = getPowerOf2(screenHeight);
	float tempW = (float)screenWidth/(float)width;
	float tempH = (float)screenHeight/(float)height;
	float tempW2 = (float)DownsampleSize/(float)width;
	float tempH2 = (float)DownsampleSize/(float)height;

	if(extensionsAvailable)
	{
		glPopAttrib();

		glPushAttrib(GL_TEXTURE_BIT);

		glDisable(GL_LIGHTING);

		//----Threshold offscreenTexture----
		glUseProgramObjectARB(thresholdProgram);

		//Select the thresholdTexture as the output.
		glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT);

		glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, screenWidth, screenHeight, 0, -1, 16);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glBindTexture(GL_TEXTURE_2D, offscreenTexture);
		glEnable(GL_TEXTURE_2D);

		glTranslatef(0.0f, 0.0f, -2.0f);

		glDisable(GL_LIGHTING);
		glColor4f(0.0f, 0.0f, 0.0f, 1.0f);

		glBegin(GL_QUADS);
		{
			glTexCoord2f(0.0f, tempH);
			glVertex2d(0, 0);
			glTexCoord2f(tempW, tempH);
			glVertex2d(screenWidth, 0);
			glTexCoord2f(tempW, 0.0f);
			glVertex2d(screenWidth, screenHeight);
			glTexCoord2f(0.0f, 0.0f);
			glVertex2d(0, screenHeight);
		}
		glEnd();

		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		glUseProgramObjectARB(0);

		//----Downsample thresholdTexture----
		glUseProgramObjectARB(blurProgram);

		//Setup offsets.
		{
			GLint loc;
			float w = 1.0f/(float)DownsampleSize;
			float h = 1.0f/(float)DownsampleSize;
			float offsets[18] = {-w, -h, 0.0f, -h, w, -h,
								 -w, 0.0f, 0.0f, 0.0f, w, 0.0f,
								 -w, h, 0.0f, h, w, h};

			loc = glGetUniformLocationARB(blurProgram, "offsets");
			glUniform2fvARB(loc, 9, offsets);
		}

		glDrawBuffer(GL_COLOR_ATTACHMENT2_EXT);

		glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glPushAttrib(GL_VIEWPORT_BIT);
		glViewport(0, 0, DownsampleSize, DownsampleSize);
		glOrtho(0, DownsampleSize, DownsampleSize, 0, -1, 16);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glBindTexture(GL_TEXTURE_2D, thresholdTexture);
		glEnable(GL_TEXTURE_2D);

		glTranslatef(0.0f, 0.0f, -2.0f);

		glDisable(GL_LIGHTING);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

		glBegin(GL_QUADS);
		{
			glTexCoord2f(0.0f, tempH);
			glVertex2d(0, 0);
			glTexCoord2f(tempW, tempH);
			glVertex2d(DownsampleSize, 0);
			glTexCoord2f(tempW, 0.0f);
			glVertex2d(DownsampleSize, DownsampleSize);
			glTexCoord2f(0.0f, 0.0f);
			glVertex2d(0, DownsampleSize);
		}
		glEnd();

		glPopAttrib();

		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		glUseProgramObjectARB(0);

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

		//----Render to screen----

		glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// Clear Screen And Depth Buffer
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, screenWidth, screenHeight, 0, -1, 16);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glBindTexture(GL_TEXTURE_2D, offscreenTexture);
		glEnable(GL_TEXTURE_2D);

		glTranslatef(0.0f,0.0f,-2.0f);

		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

		glBegin(GL_QUADS);
		{
			// Front Face
			glTexCoord2f(0.0f, tempH);
			glVertex2d(0, 0);
			glTexCoord2f(tempW, tempH);
			glVertex2d(screenWidth, 0);
			glTexCoord2f(tempW, 0.0f);
			glVertex2d(screenWidth, screenHeight);
			glTexCoord2f(0.0f, 0.0f);
			glVertex2d(0, screenHeight);
		}
		glEnd();

		//Blur downsampled texture again on resizing.
		glUseProgramObjectARB(blurProgram);

		//Setup offsets.
		{
			GLint loc;
			float w = 1.0f/(float)width;
			float h = 1.0f/(float)height;
			float offsets[18] = {-w, -h, 0.0f, -h, w, -h,
								 -w, 0.0f, 0.0f, 0.0f, w, 0.0f,
								 -w, h, 0.0f, h, w, h};

			loc = glGetUniformLocationARB(blurProgram, "offsets");
			glUniform2fvARB(loc, 9, offsets);
		}

		glDisable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glColor4f(0.0f, 0.0f, 0.0f, 0.4f);
		glBindTexture(GL_TEXTURE_2D, downsampledTexture);

		glBegin(GL_QUADS);
		{
			// Front Face
			glTexCoord2f(0.0f, tempH2);
			glVertex2d(0, 0);
			glTexCoord2f(tempW2, tempH2);
			glVertex2d(screenWidth, 0);
			glTexCoord2f(tempW2, 0.0f);
			glVertex2d(screenWidth, screenHeight);
			glTexCoord2f(0.0f, 0.0f);
			glVertex2d(0, screenHeight);
		}
		glEnd();

		glUseProgramObjectARB(0);

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		//glDisable(GL_TEXTURE_2D);
		glEnable(GL_DEPTH_TEST);
		glDisable(GL_BLEND);
		glEnable(GL_LIGHTING);

		glPopAttrib();
	}
}

//------------------------------------------------------------------------------
/*void BloomFilter::displayInfoLog(GLhandleARB handle)
{
	int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

	glGetObjectParameterivARB(handle,
							  GL_OBJECT_INFO_LOG_LENGTH_ARB,
                              &infologLength);

    if (infologLength > 0)
    {
        infoLog = new char[infologLength];
        glGetInfoLogARB(handle, infologLength, &charsWritten, infoLog);
		cout << endl << "InfoLog (" << handle << "):" << endl << endl;
		cout << infoLog;
		delete [] infoLog;
    }
}*/

//------------------------------------------------------------------------------
int BloomFilter::getPowerOf2(int val)
{
	float tempf;
	int retval = 0;

	tempf = 1.0f;
	while(retval == 0)
	{
		if(powf(2.0f, tempf) > val)
		{
			retval = static_cast<int>(powf(2.0f, tempf-1.0f));
			break;
		}
		tempf += 1.0f;
	}

	return retval*2;
}
