//	BloomFilter.h - Class to simplify applying a post-processing bloom filter.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef BLOOMFILTER_H_
#define BLOOMFILTER_H_

#include "GLHeaders.h"

///	Class to simplify applying a post-processing bloom filter.
/*!
	To use it, you'd do the following in your drawing routine (assumes you've
	already created filter):
\code
//Setup filter.
filter.setup();

//Draw your OpenGL scene here.

//Apply filter, draw to screen.
filter.draw();
\endcode
 */
class BloomFilter
{
  public:
	///	Constructor.
	BloomFilter();
	///	Destructor.
	~BloomFilter();

	///	Sets up the filter before drawing.
	void setup();
	///	Applies the filter and draws to screen.
	void draw();
  private:
	///	Helper method to display the infoLog for a shader or program object.
	//void displayInfoLog(GLhandleARB handle);
	///	Helper method to get the closest power-of-two size to the input.
	int getPowerOf2(int val);

	///	The size of the bloom texture.
	enum
	{
		DownsampleSize = 64
	};

	///	Vertex shader.
	GLhandleARB thresholdShaderVert;
	///	Fragment shader.
	GLhandleARB thresholdShaderFrag;
	///	Program object.
	GLhandleARB thresholdProgram;
	///	Vertex shader.
	GLhandleARB blurShaderVert;
	///	Fragment shader.
	GLhandleARB blurShaderFrag;
	///	Program object.
	GLhandleARB blurProgram;

	///	Handle to our frame buffer.
	GLuint frameBuffer;
	///	Handle to our depth buffer.
	GLuint depthBuffer;
	///	Handle to the main texture (contents of the scene unaltered).
	GLuint offscreenTexture;
	///	Handle to the thresholded texture (offscreenTexture thresholded to a binary image).
	GLuint thresholdTexture;
	///	Handle to the downsampled texture (based on thresholdTexture).
	GLuint downsampledTexture;
	///	Handle to the downsampled texture (based on thresholdTexture).
	GLuint downsampledTexture2;

	///	Whether or not the extensions we need are available.
	bool extensionsAvailable;
};

extern const char *thresholdShaderVertChar;

extern const char *thresholdShaderFragChar;

extern const char *blurShaderVertChar;

extern const char *blurShaderFragChar;

#endif
