//	Event.h - Event structure used for ROUTEd events.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef EVENT_H_
#define EVENT_H_

#include "DllStuff.h"
#include "StringParser.h"

#include <string>
#include <sstream>

namespace HeilanX3D
{

class AbstractNode;

//------------------------------------------------------------------------------
///	Helper class used to automatically convert between event value types.
/*!
	This is so you can convert between the possible event data types without
	having to know what type the event you're acting on actually has.

	Here's some code to demonstrate:
\code
EventValue floatExample(1.0f);
EventValue intExample(2);
EventValue stringExample("123");

float tempf;
int tempi;
string temps;

//Automatically converts floatExample's 1.0f to a string and assigns it to
//temps.
temps = floatExample;

//Automatically converts stringExample's "123" to an int and assigns it to
//tempi.  tempi would be set to 0 if the string did not contain a number.
tempi = stringExample;

//Automatically converts floatExample's 1.0f to a string and assigns it to
//temps.
tempf = intExample;


//You can also do...

//floatExample is now a string value.
floatExample = "blurple";

//intExample is now a float value.
intExample = 23.4f;

//stringExample is now an int value.
stringExample = 345;
\endcode
 */
struct EventValue
{
  public:
	///	Default constructor (no value).
	EventValue():
	floatVal(0.0f),
	intVal(0),
	boolVal(false),
	type(None)
	{
		
	};
	///	Float constructor.
	EventValue(float val):
	floatVal(val),
	intVal(0),
	boolVal(false),
	type(Float)
	{
		
	};
	///	Int constructor.
	EventValue(int val):
	floatVal(0.0f),
	intVal(val),
	boolVal(false),
	type(Int)
	{
		
	};
	///	String constructor.
	EventValue(const std::string& val):
	floatVal(0.0f),
	intVal(0),
	stringVal(val),
	boolVal(false),
	type(String)
	{
		
	};
	///	2nd string constructor.
	EventValue(const char *val):
	floatVal(0.0f),
	intVal(0),
	stringVal(val),
	boolVal(false),
	type(String)
	{
		
	};
	///	Bool constructor.
	EventValue(bool val):
	floatVal(0.0f),
	intVal(0),
	boolVal(val),
	type(Int)
	{
		
	};

	///	Sets the value to a float.
	EventValue& operator=(float val)
	{
		floatVal = val;
		type = Float;

		return *this;
	};
	///	Sets the value to an int.
	EventValue& operator=(int val)
	{
		intVal = val;
		type = Int;

		return *this;
	};
	///	Sets the value to a string.
	EventValue& operator=(std::string val)
	{
		stringVal = val;
		type = String;

		return *this;
	};
	///	Sets the value to a bool.
	EventValue& operator=(bool val)
	{
		boolVal = val;
		type = Int;

		return *this;
	};
	///	Copies one EventValue.
	EventValue& operator=(const EventValue& val)
	{
		floatVal = val.floatVal;
		intVal = val.intVal;
		stringVal = val.stringVal;
		boolVal = val.boolVal;
		type = val.type;

		return *this;
	};

	///	Returns the value as a float.
	operator float()
	{
		//Is there a more elegant way to do this?
		switch(type)
		{
			case Float:
				return floatVal;
			case Int:
				return (float)intVal;
			case String:
			{
				StringParser parser(stringVal);

				if(parser.getNumNumbers() > 0)
					floatVal = parser.getFloat();

				return floatVal;
			}
			case Bool:
				return (boolVal == true) ? 1.0f : 0.0f;
			default:
				return 0.0f;
		}
	};
	///	Returns the value as an int.
	operator int()
	{
		//Is there a more elegant way to do this?
		switch(type)
		{
			case Int:
				return intVal;
			case Float:
				return (int)floatVal;
			case String:
			{
				StringParser parser(stringVal);

				if(parser.getNumNumbers() > 0)
					intVal = parser.getInt();

				return intVal;
			}
			case Bool:
				return (boolVal == true) ? 1 : 0;
			default:
				return 0;
		}
	};
	///	Returns the value as a string.
	operator const std::string&()
	{
		//Is there a more elegant way to do this?
		switch(type)
		{
			case String:
				return stringVal;
			case Float:
			{
				std::stringstream tempConv;

				tempConv << floatVal;
				stringVal = tempConv.str();

				return stringVal;
			}
			case Int:
			{
				std::stringstream tempConv;

				tempConv << intVal;
				stringVal = tempConv.str();

				return stringVal;
			}
			case Bool:
			{
				std::stringstream tempConv;

				tempConv << boolVal;
				stringVal = tempConv.str();

				return stringVal;
			}
			default:
				return stringVal;
		}
	};
	///	Returns the value as a bool.
	operator bool()
	{
		//Is there a more elegant way to do this?
		switch(type)
		{
			case Int:
				return (intVal > 0) ? true : false;
			case Float:
				return (floatVal > 0.5f) ? true : false;
			case String:
				return ((stringVal == "true")||(stringVal == "TRUE")) ? true : false;
			case Bool:
				return boolVal;
			default:
				return 0;
		}
	};

  private:
	///	The various possible types.
	enum ValueType
	{
		None,
		Float,
		Int,
		String,
		Bool
	};

	///	The stored float value.
	float floatVal;
	///	The stored int value.
	int intVal;
	///	The stored string value.
	std::string stringVal;
	///	The stored bool value.
	bool boolVal;
	///	The type of the stored value.
	ValueType type;
};

//------------------------------------------------------------------------------
///	Event structure used for ROUTEd events.
struct dllStuff Event
{
	///	The sending node.
	AbstractNode *sender;
	///	The attribute it's aimed at.
	std::string attribute;
	///	The value.
	EventValue value;
};

}

#endif
