//	AbstractNode.cpp - Pure virtual base class for all nodes.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "AbstractNode.h"
#include "OpenSoundController.h"
#include "AmbisonicData.h"
#include "Camera.h"

#include <sstream>
#include <iostream>
#include <cmath>
#include <algorithm>

using namespace std;

namespace HeilanX3D
{

//------------------------------------------------------------------------------
AbstractNode::AbstractNode():
refCount(1),
visual(false),
aural(false),
minFrontRadius(1.0f),
minBackRadius(1.0f),
maxFrontRadius(10.0f),
maxBackRadius(10.0f),
shouldAppearAtStart(false)
{
	
}

//------------------------------------------------------------------------------
AbstractNode::~AbstractNode()
{
	Event *tempEvent;

	//Ensure the input and output queues get cleared.
	while(!inputQueue.empty())
	{
		tempEvent = inputQueue.front();
		inputQueue.pop();
		delete tempEvent;
	}

	while(!outputQueue.empty())
	{
		tempEvent = outputQueue.front();
		outputQueue.pop();
		delete tempEvent;
	}
}

//------------------------------------------------------------------------------
void AbstractNode::addParent(ParentNode *parentNode)
{
	parents.push_back(parentNode);
}

//------------------------------------------------------------------------------
void AbstractNode::removeParent(ParentNode *parentNode)
{
	vector<ParentNode *>::iterator it;

	it = find(parents.begin(), parents.end(), parentNode);
	if(it != parents.end())
		parents.erase(it);
}

//------------------------------------------------------------------------------
AmbisonicData *AbstractNode::getAmbisonicAudio(int numSamples,
											   const ThreeFloats& pos)
{
	int i;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats camPos;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;
	AmbisonicData *retval = getMonoAudio(numSamples);
	const Quaternion& camRot = Camera::getInstance()->getAudioRotation();
	//const float rotMult = 6.283185306f/360.0f;

	if(retval)
	{
		//Get current Ambisonic position (see AmbisonicData::operator=()),
		//relative to the camera.

		//So what we do here is:
		//1.) Ensure relativePos is equal to this node's spatial position in the
		//    scene.
		relativePos = pos;
		//2.) Rotate relativePos by the rotation of the camera, by multiplying
		//    it by the camera's orientation Quaternion.
		relativePos = camRot * relativePos;
		//3.) Translate relativePos by the position of the camera.
		relativePos -= camPos;

		lastAtten = calcRelativeAttenuation(lastPos);
		if(lastPos != relativePos)
		{
			ambiStart = lastPos;
			ambiEnd = relativePos;
			//We increment the position by a constant value, so we don't get
			//clicks or zipper noise.  Of course, there are disadvantages to
			//this approach...
			ambiInc = 0.005f;
			currentAmbi = ambiStart;
			newAtten = calcRelativeAttenuation(relativePos);
		}
		else
		{
			currentAmbi = relativePos;
			newAtten = lastAtten;
		}
		currentAtten = lastAtten;

		//Encode audio.
		for(i=0;i<numSamples;++i)
		{
			//Update currentAmbi and currentAtten if necessary.
			if(ambiInc > 0.0f)
			{
				//Calc currentAmbi.
				currentAmbi = ((1.0f-ambiIndex) * ambiStart) +
							  (ambiIndex * ambiEnd);
				//Calc currentAtten.
				currentAtten = ((1.0f-ambiIndex) * lastAtten) +
							   (ambiIndex * newAtten);

				ambiIndex += ambiInc;

				//Stop updating index if we've reached our destination.
				//I get weird rounding errors here, hence the .0000001f etc...
				if(ambiIndex > 1.0000001f)
				{
					ambiInc = 0.0f;
					currentAtten = newAtten;
				}
			}

			//Convert float sound file to Ambisonic, put it in the buffer.
			retval[i] = currentAmbi * (retval[i].W * currentAtten);
		}
		//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
		if(ambiIndex >= 1.0f)
		{
			lastPos = relativePos;
			lastAtten = currentAtten;
		}
		else
		{
			lastPos = ((1.0f-ambiIndex) * lastPos) + (ambiIndex * relativePos);
			lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
		}
	}

	return retval;
}

//------------------------------------------------------------------------------
const float * const AbstractNode::getInputAudio(float *block, bool set)
{
	static float *inputAudio = 0;

	if(set)
		inputAudio = block;

	return inputAudio;
}

//------------------------------------------------------------------------------
int AbstractNode::getNumInputAudioChannels(int chan)
{
	static int inputChannels = 0;

	if(chan != -1)
		inputChannels = chan;

	return inputChannels;
}

//------------------------------------------------------------------------------
void AbstractNode::setAndSendAttribute(const string& attribute, EventValue& val)
{
	setAttribute(attribute, val);
	sendEvent(attribute, val);
}

//------------------------------------------------------------------------------
void AbstractNode::setOscName(const string& newName)
{
	oscName = newName;
	OpenSoundController::getInstance()->registerNode(this);
}

//------------------------------------------------------------------------------
void AbstractNode::setDefName(const std::string& newName)
{
	defName = newName;
	OpenSoundController::getInstance()->registerNode(this);
}

//------------------------------------------------------------------------------
AbstractNode *AbstractNode::findNode(const std::string& name)
{
	if(defName == name)
		return this;
	else
		return 0;
}

//------------------------------------------------------------------------------
void AbstractNode::addRoute(const std::string& fromAtt,
						    AbstractNode *toNode,
						    const std::string& toAtt)
{
	RouteData tempRoute;

	tempRoute.toNode = toNode;
	tempRoute.toAttribute = toAtt;

	routeNodes.insert(make_pair(fromAtt, tempRoute));
}

//------------------------------------------------------------------------------
void AbstractNode::addEvent(Event *event)
{
	inputQueue.push(event);
}

//------------------------------------------------------------------------------
void AbstractNode::handleEvents()
{
	Event *tempEvent;
	Event *tempEvent2;
	multimap<string, RouteData>::iterator it;

	//Handle any input Events.
	while(!inputQueue.empty())
	{
		tempEvent = inputQueue.front();
		setAndSendAttribute(tempEvent->attribute, tempEvent->value);
		inputQueue.pop();
		delete tempEvent;
	}

	//Handle any waiting output Events.
	eventCritSec.lock();
	while(!outputQueue.empty())
	{
		tempEvent2 = outputQueue.front();
		it = routeNodes.find(tempEvent2->attribute);
		if(it != routeNodes.end())
		{
			for(it=routeNodes.lower_bound(tempEvent2->attribute);
				it!=routeNodes.upper_bound(tempEvent2->attribute);
				++it)
			{
				tempEvent = new Event;
				tempEvent->sender = this;
				tempEvent->attribute = tempEvent2->attribute;
				tempEvent->value = tempEvent2->value;

				//Update tempEvent to apply to the receiving node.
				tempEvent->attribute = it->second.toAttribute;

				//Send the event.
				it->second.toNode->addEvent(tempEvent);
			}
		}

		outputQueue.pop();
		delete tempEvent2;
	}
	eventCritSec.unlock();
}

//------------------------------------------------------------------------------
void AbstractNode::clearEvents()
{
	handleEvents();
}

//------------------------------------------------------------------------------
void AbstractNode::setType(const std::string& newType)
{
	type = newType;
}

//------------------------------------------------------------------------------
void AbstractNode::setVisual(bool val)
{
	visual = val;
}

//------------------------------------------------------------------------------
void AbstractNode::setAural(bool val)
{
	aural = val;
}

//------------------------------------------------------------------------------
void AbstractNode::setShouldBeStart(bool val)
{
	shouldAppearAtStart = val;
}

//------------------------------------------------------------------------------
void AbstractNode::setMinFrontAudioRadius(float val)
{
	minFrontRadius = val;
}

//------------------------------------------------------------------------------
void AbstractNode::setMinBackAudioRadius(float val)
{
	minBackRadius = val;
}

//------------------------------------------------------------------------------
void AbstractNode::setMaxFrontAudioRadius(float val)
{
	maxFrontRadius = val;
}

//------------------------------------------------------------------------------
void AbstractNode::setMaxBackAudioRadius(float val)
{
	maxBackRadius = val;
}

//------------------------------------------------------------------------------
void AbstractNode::sendEvent(const string& attribute, EventValue value)
{
	if(routeNodes.size() > 0)
	{
		Event *tempEvent = new Event;

		tempEvent->sender = this;
		tempEvent->attribute = attribute;
		tempEvent->value = value;

		eventCritSec.lock();
		outputQueue.push(tempEvent);
		eventCritSec.unlock();
	}
}

//------------------------------------------------------------------------------
void AbstractNode::incRefCount()
{
	++refCount;
}

//------------------------------------------------------------------------------
void AbstractNode::decRefCount()
{
	--refCount;

	if(refCount <= 0)
		delete this;
}

}
