//	AbstractNode.h - Pure virtual base class for all nodes.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef ABSTRACTNODE_H_
#define ABSTRACTNODE_H_

//Stupid MSVC v6 warning about the STL...
#ifdef WIN32
#pragma warning(disable:4786)
#endif

#include "ThreeFloats.h"
#include "FourFloats.h"
#include "CriticalSection.h"
#include "Event.h"

#include <string>
#include <queue>
#include <map>

struct AmbisonicData;

namespace HeilanX3D
{

class SceneConstructor;
class ParentNode;

/*!
	\page NodeTypeCreation Creating a Heilan X3D node type

	\section Basics

	Child nodes should inherit from the ChildNode class (this is in fact a
	typedef of the AbstractNode class - the AbstractNode class is the
	equivalent of X3DNode in the X3D specification).  Parent nodes should
	inherit from ParentNode.  Make sure to read the documentation for
	ParentNode's methods - a number of them replace the methods used in
	ChildNode (i.e. you generally shouldn't override
	ParentNode::graphicDoStuff() in a ParentNode subclass - use
	ParentNode::preChildDraw() and ParentNode::postChildDraw() instead).

	\section Ambisonic Spatialisation

	In previous versions of Heilan (< v0.12), you would have to handle the
	Ambisonic spatialisation yourself, but that's changed now.  If you're
	generating audio, you just fill a buffer with a <em>mono</em> stream in
	AbstractNode::getMonoAudio(), and it will be encoded for you.  The one thing
	to note is that AbstractNode::getMonoAudio() returns an array of
	AmbisonicData, rather than single floats.  As such, you should construct and
	maintain this array in your subclass, and in AbstractNode::getMonoAudio(),
	write your audio data to the W component of the AmbisonicData structures.

	Here's some basic example code:
\code
//------------------------------------------------------------------------------
AmbisonicData *Example::getMonoAudio(int numSamples)
{
	int i;
	static float index = 0.0f;

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		audioBuffer[i].W = sinf(index);

		index += 0.001f;
		if(index > TWO_PI)
			index = fmod(index);
	}

	return audioBuffer;
}



//------------------------------------------------------------------------------
void Example::setBlockSize(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}
\endcode

	\section Construction

	In order for nodes to be constructed on demand when an X3D file is parsed,
	a factory design pattern is used. For a node to make its constructor
	available to the factory, 3 things must be done (well, 2 really):

	- Add a static 'NodeConstructor' instance to your class, like so:
\code
class Thing : public ChildNode
{
  public:
	...

	static NodeConstructor<Thing> constructor;
};
\endcode

	- Make sure you define it in the source file:
\code
NodeConstructor Thing::constructor;
\endcode

	- Make sure you register the node with the NodeFactory instance somewhere:
\code
NodeFactory::getInstance().registerNode(&Thing::constructor);
\endcode

If you do this, you should be able to use your node in the browser. 
 */

//------------------------------------------------------------------------------
///	Pure virtual base class for all nodes.
/*!
	This node implements reference counting, because USE'd nodes will have
	multiple parents.  Whenever X3DFileLoader finds a USE'd node, it will
	increment the relevant DEF'd node's reference count.  To delete an object,
	make sure you call decRefCount() instead of delete.

	\sa NodeConstructor
 */
class dllStuff AbstractNode
{
  public:
	///	Constructor.
	/*!
		Things that *MUST* be done in the constructor of subclasses:
		- Call setType() with the textual name (i.e. what appears in the x3d
		file) of your subclass.
		- Call setVisual() and setAural() to make sure they're set correctly.
	 */
	AbstractNode();
	///	Destructor.
	virtual ~AbstractNode();

	///	Called after all the nodes in the scene have been constructed.
	/*!
		Normally, I wouldn't bother with a function like this - any
		initialisation stuff I would put in the constructor - but since
		NodeFactory actually constructs and destroys an instance of each
		class registered with it every time the browser's started, I decided it
		made sense to allow classes to have somewhere to initialise things
		that don't need to be initialised when the factory first registers them.

		Note that this function can be called more than once (when the user
		switches to/from fullscreen), so if you're allocating memory, make sure
		you don't leak memory if it's called for a second (third, fourth...) time.
	 */
	virtual void initialise() {};

	///	Called to add to this node's parent nodes.
	void addParent(ParentNode *parentNode);
	///	Called to remove parentNode from this node's list of parents.
	void removeParent(ParentNode *parentNode);
	///	Returns this node's parents.
	/*!
		A child may have more than one parent if it's USEd.
	 */
	std::vector<ParentNode *>& getParent() {return parents;};

	///	This will be called by Browser or the ParentNode to draw the node in the scene.
	/*!
		Browser or the ParentNode will first determine whether this node wants
		to be drawn by calling isVisual() (to try and avoid unnecessary
		operations).
	 */
	virtual void graphicDoStuff() = 0;

	
	// THIS IS DEPRECATED NOW - SEE THE 2 METHODS BELOW IT

	///	This will be called by Browser(?) to get the node's audio output.
	/*!
		\param numSamples The number of samples of audio to generate.
		\param pos The position of the node in the 3d scene.  This is necessary
		because most nodes will be moved in the scene by Transform ParentNodes,
		and therefore won't have direct access to their current position.  This
		variable lets the position be propagated through the scene (and it will
		work when the audio gets multi-threaded).

		Browser(?) will sum the ambisonic audio output from all the nodes in
		the scene, then convert the resultant data into however many speaker
		feeds we're using.

		Browser(?) will first determine whether this node wants to output audio
		by calling isAural() (to try and avoid unnecessary operations).
	 */
	/*virtual AmbisonicData *audioDoStuff(int numSamples,
										const ThreeFloats& pos) = 0;*/

	///	Called by any ParentNodes (or the Browser) to get this node's encoded audio.
	/*!
		\param numSamples The number of samples to calculate for this audio
		block.
		\param pos The current 3d position of the node (used for the Ambisonic
		encoding).
		\return 0 if the subclass does not produce audio.

		This calls getMonoAudio(), then encodes the resulting audio to Ambisonic
		B-format.  So for a node to generate sound, you just override
		getMonoAudio(), and this method will make sure the sound is encoded
		correctly.

		This method takes into account the minBack, minFront, maxBack and
		maxFront attributes to determine how the sound will fall off as the
		viewer moves away.

		Only ParentNodes should have to override this method.  Unless your
		subclass has an attribute that changes its position, there's no reason
		for you to override it.

		\sa getMonoAudio, AmbisonicData
	 */
	virtual AmbisonicData *getAmbisonicAudio(int numSamples,
											 const ThreeFloats& pos);
	///	Sound-producing nodes' audio output is calculated here.
	/*!
		\param numSamples The number of samples needing calculated for this
		audio block.

		This method gets called by getAmbisonicAudio(), which will then encode
		this method's monophonic output into Ambisonic B-format.

		Note: It is the subclass' responsibility to maintain an AmbisonicData
		buffer for the output - getAmbisonicAudio() will encode the audio to
		this buffer as well.  The process is:
	
		- ParentNode calls getAmbisonicAudio() for this node.
		- getAmbisonicAudio() calls getMonoAudio() for this node.
		- getMonoAudio() returns an AmbisonicData buffer to getAmbisonicAudio(),
		  with only the W components filled out (i.e. a mono audio stream).
		- getAmbisonicAudio() then encodes the mono audio from the W component
		  in this buffer to Ambisonic B-format audio, and writes it back to the
		  AmbisonicData buffer that getMonoAudio() returned.
		- getAmbisonicAudio() returns the newly-encoded audio to its parent
		  node.

		\todo Should this be protected?

		\sa getAmbisonicAudio, AmbisonicData
	 */
	virtual AmbisonicData *getMonoAudio(int numSamples) {return 0;};

	///	Called prior to calling audioDoStuff(), for nodes which need to reset things.
	virtual void audioBlockStart() {};
	///	Called to set the current audio block size.
	/*!
		This should be used to allocate memory for returning from
		getMonoAudio().
	 */
	virtual void setBlockSize(int newSize) = 0;
	///	Called to set the current audio samplerate.
	virtual void setSamplerate(float val) {};
	///	Returns the input audio for this block.
	/*!
		\param block If set is true, this sets the static block variable.
		Otherwise it does nothing.
		\param set If this is true, the static block variable will be set to the
		value of block.
		\return The current input audio (only valid during an audio processing
		block.

		This method should not be called from outside the audio thread(s).

		This is maybe a hacky way of doing things, but it was the way that
		offered the least disruption of the existing Heilan code, so...
	 */
	static const float * const getInputAudio(float *block = 0, bool set = false);
	///	Returns the number of channels in the input audio block.
	/*!
		\param chan If this is -1, nothing happens, otherwise it sets the 
		static numChannels variable.
	 */
	static int getNumInputAudioChannels(int chan = -1);

	///	Wraps setAttribute.  Also calls sendEvent, so events can be chained.
	void setAndSendAttribute(const std::string& attribute, EventValue& val);
	///	Called to set the value of the named attribute.
	/*!
		\param attribute The name of the attribute.
		\param val The attribute's new value, as it is in the .x3d file.  It is
		the subclass' job to convert the string into the appropriate type (you
		can use StringParser for this purpose).
	 */
	virtual void setAttribute(const std::string& attribute,
							  EventValue& val) = 0;

	///	Called to get the type of the node (e.g. "Box", "ImageTexture" etc.).
	std::string getType() const {return type;};

	///	Called by Browser(?) to determine whether the node wants to be drawn.
	bool isVisual() const {return visual;};
	///	Called by Browser(?) to determine whether the node outputs audio.
	bool isAural() const {return aural;};

	///	Called by X3DFileLoader when constructing an X3D scene.
	/*!
		This is used by ParentNodes to inform X3DFileLoader which ParentNode to
		add children nodes to.
	 */
	virtual void sceneConstructionEnter(SceneConstructor *constructor) {};
	///	Called by X3DFileLoader when constructing an X3D scene.
	/*!
		This is used by ParentNodes to inform X3DFileLoader which ParentNode to
		add children nodes to (or in this case, stop adding children to).
	 */
	virtual void sceneConstructionLeave(SceneConstructor *constructor) {};
	///	Informs the node about certain Heilan-specific metadata.
	/*!
		Heilan uses metadata to expand upon the X3D spec, without rendering
		Heilan X3D files unusable to other browsers.  When the X3DFileLoader
		sees the name and value fields in a Metadata node, it calls
		this method in the parent node, so the parent node can act on those
		values, if necessary.  This is used to activate per-node parallelism
		in the audio engine, and to give Viewpoint nodes time parameters
		specifying how long it should take to travel between them (among other
		things?).

		For simplicity, Heilan treats all metadata values as strings, and
		converts from double, float and int for their particular nodes.

		Do not override this if you're inheriting from ParentNode, use
		setParentMetadata() instead (since ParentNode overrides setMetadata() to
		implement per-node parallelism).
	 */
	virtual void setMetadata(const std::string& name,
							 const std::string& value) {};
	///	Called from X3DFileLoader to set the node's OSC name.
	void setOscName(const std::string& newName);
	///	Returns this node's OSC name.
	const std::string& getOscName() const {return oscName;};

	///	Called from X3DFileLoader to set the node's DEF'd name.
	void setDefName(const std::string& newName);
	///	Returns this node's DEF'd name.
	const std::string& getDefName() const {return defName;};
	///	Method used to search for a particular DEF-named node in a scene.
	/*!
		\param name DEF name of the node we're searching for.
		\return 0 if the named node could not be found.

		This method just checks if this node's name matches that passed in, and
		returns true or false accordingly.
	 */
	virtual AbstractNode *findNode(const std::string& name);
	///	Adds a ROUTEd node to this node's routeNodes vector.
	/*!
		\param fromAtt The attribute whose value should be output to the
		receiving node.
		\param toNode Pointer to the receiving node.
		\param toAtt The attribute to be updated (on toNode) when
		fromAtt changes.
	 */
	void addRoute(const std::string& fromAtt,
				  AbstractNode *toNode,
				  const std::string& toAtt);
	///	Adds an Event to this node's inputQueue.
	void addEvent(Event *event);
	///	Called from Browser::paint() to handle any waiting events.
	/*!
		Rather than having Browser traverse the whole scene, calling
		handleEvents() for every node (regardless of whether or not they're
		connected via ROUTEs), it just stores a vector of the nodes which are
		connected with ROUTEs, to speed things up a bit.
	 */
	void handleEvents();
	///	Used to make sure unROUTEd nodes get their output queues emptied periodically.
	/*!
		Even though a node may not be connected to any others via a ROUTE, it
		may still receive OSC messages, which would trigger it to send an Event
		to its output queue (since all Events in Heilan are automatically passed
		on to allow Event chaining).  This method ensures that we don't end up
		with a constantly-increasing output queue.

		It essentially wraps handleEvents(), but where handleEvents() is called
		for every visual frame (and only for those nodes which are ROUTEd),
		clearEvents() is only called once every 2 seconds (and for every node in
		the scene).  It is reimplemented in ParentNode to allow for ChildNodes
		to have their queues emptied by their parents.
	 */
	virtual void clearEvents();

	///	Increments the reference count for this object.
	void incRefCount();
	///	Decrements the reference count for this object and deletes it if necessary.
	void decRefCount();

	///	Returns true if this node type should appear at the start of it's parent's list of children.
	/*!
		This method is necessary due to a design flaw in Heilan: the scene is
		drawn in the order the nodes appear in the X3D file.  For most things
		this isn't a problem, but some nodes' graphical commands need to appear
		before those of other nodes, or the scene won't be drawn normally
		(e.g. in a Shape group, Appearance needs to come before Box...).  Any
		node types which need to appear at the start of their parent's list of
		children then, need to call setShouldBeStart() in their constructors,
		and X3DFileLoader will try and put them in the correct place in the
		scene graph.  There may still be occasions where things are screwed up
		(e.g if you have two sibling nodes which both return true here,
		X3DFileLoader will still put them in the order they appear in the file,
		which may be wrong) but short of creating a far more complex system of
		relative importances, or rewriting the entire app, this should be a
		reasonable compromise.
	 */
	bool shouldBeStart() const {return shouldAppearAtStart;};
  protected:
	///	Called by subclasses to set their type.
	void setType(const std::string& newType);
	///	Called by subclasses to indicate that they want to be drawn.
	void setVisual(bool val = true);
	///	Called by subclasses to indicate that they want to output audio.
	void setAural(bool val = true);
	///	Called by subclasses to indicate they should appear at the start of their parent's list of children.
	/*!
		\sa shouldBeStart()
	 */
	void setShouldBeStart(bool val = true);

	///	Called by subclasses to set the minimum (front) sound radius.
	/*!
		The front and back minimum and maximum radii specify how the sound of
		a node's level drops off as the camera moves away.  If the camera is
		within the minimum radius, the sound will be heard at full volume.  The
		maximum radius corresponds to the sound being attenuated to -20dB of
		it's normal level.  The current level for the camera is calculated
		using linear interpolation derived from the minimum and maximum radius
		points (see http://www.web3d.org/x3d/specifications/ISO-IEC-19775-X3DAbstractSpecification/ ).
	 */
	void setMinFrontAudioRadius(float val);
	///	Called by subclasses to set the minimum (back) sound radius.
	/*!
		\sa setMinFrontAudioRadius()
	 */
	void setMinBackAudioRadius(float val);
	///	Called by subclasses to set the maximum (front) sound radius.
	/*!
		\sa setMinFrontAudioRadius()
	 */
	void setMaxFrontAudioRadius(float val);
	///	Called by subclasses to set the maximum (back) sound radius.
	/*!
		\sa setMinFrontAudioRadius()
	 */
	void setMaxBackAudioRadius(float val);

	///	Returns the minimum (front) sound radius.
	float getMinFrontAudioRadius() const {return minFrontRadius;};
	///	Returns the minimum (back) sound radius.
	float getMinBackAudioRadius() const {return minBackRadius;};
	///	Returns the maximum (front) sound radius.
	float getMaxFrontAudioRadius() const {return maxFrontRadius;};
	///	Returns the maximum (back) sound radius.
	float getMaxBackAudioRadius() const {return maxBackRadius;};

	///	Helper method to return the audio attenuation relative to a point.
	/*!
		\param relPos This will usually be the position of the node relative to
		the camera.  This method will calculate the correct audio attenuation
		according to the node's distance from the camera, and the
		minFrontRadius, minBackRadius, maxFrontRadius and maxBackRadius
		variables.

		\todo Make it take the back radius values as well, rather than assuming
		a circular sound field.
	 */
	inline float calcRelativeAttenuation(const ThreeFloats& relPos)
	{
		float mindist, camdist, attdist, dropoffdist;
		float retval;

		mindist = minFrontRadius;
		camdist = sqrtf((relPos.x * relPos.x)+
						(relPos.y * relPos.y)+
						(relPos.z * relPos.z));
		attdist = camdist - minFrontRadius;
		dropoffdist = maxFrontRadius - minFrontRadius;

		if(camdist < mindist)
			retval = 1.0f;
		else
		{
			retval = -20.0f * (attdist/dropoffdist);
			retval = powf(10.0f, (retval*0.1f));
		}

		return retval;
	};

	///	Called from this node to send an Event to another node.
	/*!
		\param attribute Name of the attribute (in this node) that has changed.
		\param value The value of the attribute that has changed.
	 */
	void sendEvent(const std::string& attribute, EventValue value);

	///	Helper method used in getAmbisonicAudio().
	inline void rotateAroundY(ThreeFloats& pos, float angle)
	{
		if(angle != 0.0f)
		{
			float tempX = pos.x;
			float sinAngle = sinf(angle);
			float cosAngle = cosf(angle);

			pos.x = (pos.x * cosAngle) + (pos.z * sinAngle);
			pos.z = -(tempX * sinAngle) + (pos.z * cosAngle);
		}
	};
	///	Helper method used in getAmbisonicAudio().
	inline void rotateAroundX(ThreeFloats& pos, float angle)
	{
		if(angle != 0.0f)
		{
			float tempY = pos.y;
			float sinAngle = sinf(angle);
			float cosAngle = cosf(angle);

			pos.y = (pos.y * cosAngle) - (pos.z * sinAngle);
			pos.z = (tempY * sinAngle) + (pos.z * cosAngle);
		}
	};
	///	Helper method used in getAmbisonicAudio() (test).
	inline void rotateAroundYX(ThreeFloats& pos, float yAngle, float xAngle)
	{
		if((yAngle != 0.0f) && (xAngle != 0.0f))
		{
			float tempX = pos.x;
			float tempY = pos.y;
			float sinXAngle = sinf(xAngle);
			float cosXAngle = cosf(xAngle);
			float sinYAngle = sinf(yAngle);
			float cosYAngle = cosf(yAngle);

			pos.x = (pos.x * cosYAngle) + (pos.z * sinYAngle);
			pos.y = (pos.y * cosXAngle) - (pos.z * sinXAngle);
			pos.z = (tempY * sinXAngle) + (pos.z * cosXAngle) -
					(tempX * sinYAngle) + (pos.z * cosYAngle);
		}
	};
	///	(test).
	inline void glRotateTest(ThreeFloats& pos,
						 float angle,
						 float x,
						 float y,
						 float z)
	{
		FourFloats m0, m1, m2, m3;
		float cosAngle = cosf(angle);
		//float cosAngle2 = 1.0f - cosAngle;
		float sinAngle = sinf(angle);
		ThreeFloats retval;

		//Setup rotation matrix.
		//1st row.
		m0.x = (x*x)*(1.0f-cosAngle) + cosAngle;
		m0.y = (x*y)*(1.0f-cosAngle) - (z*sinAngle);
		m0.z = (x*z)*(1.0f-cosAngle) + (y*sinAngle);
		m0.r = 0.0f;
		//2nd row.
		m1.x = (y*x)*(1.0f-cosAngle) + (z*sinAngle);
		m1.y = (y*y)*(1.0f-cosAngle) + cosAngle;
		m1.z = (y*z)*(1.0f-cosAngle) - (x*sinAngle);
		m1.r = 0.0f;
		//3rd row.
		m2.x = (x*y)*(1.0f-cosAngle) - (y*sinAngle);
		m2.y = (y*z)*(1.0f-cosAngle) + (x*sinAngle);
		m2.z = (z*z)*(1.0f-cosAngle) + cosAngle;
		m2.r = 0.0f;
		//4th row.
		m3.x = 0.0f;
		m3.y = 0.0f;
		m3.z = 0.0f;
		m3.r = 1.0f;

		//Rotate pos by rotation matrix.
		retval.x = (m0.x * pos.x) + (m0.y * pos.y) + (m0.z * pos.z);
		retval.y = (m1.x * pos.x) + (m1.y * pos.y) + (m1.z * pos.z);
		retval.z = (m2.x * pos.x) + (m2.y * pos.y) + (m2.z * pos.z);
		pos = retval;
	};
  private:
	///	The reference count for this object.
	int refCount;

	///	Holds the type of the class.
	std::string type;

	///	Value indicating whether the node wants to be drawn.
	/*!
		False by default.
	 */
	bool visual;
	///	Value indicating whether the node wants to output audio.
	/*!
		False by default.
	 */
	bool aural;

	///	This node's OSC name.
	/*!
		This is used to identify separate nodes in the OSC address space.
	 */
	std::string oscName;

	///	Simple struct to represent the data required for a ROUTE.
	struct RouteData
	{
		///	The attribute in this node to send data from.
		/*!
			This is now used as the key to the routeNodes map, to (hopefully)
			speed up searching for the correct RouteData entry.
		 */
		//std::string fromAttribute;
		///	Pointer to the node to send any Events to.
		AbstractNode *toNode;
		///	The attribute in toNode to send data to.
		std::string toAttribute;
	};

	///	This node's DEF'd name.
	/*!
		This is used by ROUTE and USE(?) statements.
	 */
	std::string defName;
	///	This node's input event queue.
	std::queue<Event *> inputQueue;
	///	This node's output event queue.
	std::queue<Event *> outputQueue;
	///	Holds data on any ROUTEs this node has.
	std::multimap<std::string, RouteData> routeNodes;
	///	CriticalSection to guard the outputQueue.
	CriticalSection eventCritSec;

	///	This node's parent node.
	std::vector<ParentNode *> parents;

	///	Minimum audio radius, front.
	float minFrontRadius;
	///	Minimum audio radius, back.
	float minBackRadius;
	///	Maximum audio radius, front.
	float maxFrontRadius;
	///	Maximum audio radius, back.
	float maxBackRadius;

	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;

	///	Whether this node should appear at the start of its parent's list of children.
	bool shouldAppearAtStart;
};

//------------------------------------------------------------------------------
///	Base class for NodeConstructor.
/*!
	\sa NodeConstructor
 */
class dllStuff NodeConstructorBase
{
  public:
	///	Constructor (does nothing).
	NodeConstructorBase() {};
	///	Destructor (does nothing).
	virtual ~NodeConstructorBase() {};

	///	Returns a new instance of an AbstractNode subclass.
	virtual AbstractNode *createInstance() const = 0;

	///	Returns the name of the node's type.
	virtual std::string getType() const = 0;
};

//------------------------------------------------------------------------------
///	A simple template class used to construct nodes from a factory.
/*!
	All subclasses of AbstractNode that can occur in an X3D file (i.e. not
	ParentNode, but things like Box, ImageTexture etc.) should have one of
	these.  You would do something like the following, and then register
	thingConstructor with NodeFactory...
	\code
class Thing : public ChildNode
{
  public:
	...
	static NodeConstructor<Thing> thingConstructor;
};
	\endcode
 */
template <class T>
class dllStuff NodeConstructor : public NodeConstructorBase
{
  public:
	///	Constructor (does nothing).
	NodeConstructor() {};
	///	Destructor (does nothing).
	~NodeConstructor() {};

	///	Returns a new instance of an AbstractNode subclass.
	AbstractNode *createInstance() const {return new T;};

	///	Returns the name of the node's type.
	std::string getType() const {T retval; return retval.getType();};
};

//------------------------------------------------------------------------------
///	ChildNode is essentially an AbstractNode anyway.
/*!
	I decided to do this because there didn't seem much point in creating a
	subclass of AbstractNode that didn't actually add anything to AbstractNode,
	but I still wanted to be able to differentiate between ParentNodes and
	ChildNodes.  Hopefully it isn't too confusing...
 */
typedef AbstractNode ChildNode;

}

#endif
