//	AudioClip.cpp - AudioClip node, used to load sound files from disk.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "AudioClip.h"
#include "SoundFileLoader.h"
#include "ParentNode.h"
#include "StringParser.h"

#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<AudioClip> AudioClip::constructor;

//------------------------------------------------------------------------------
AudioClip::AudioClip():
ChildNode(),
ThreadObject("AudioClip"),
loop(false),
pitch(1.0f),
startTime(0.0f),
stopTime(0.0f),
resumeTime(0.0f),
pauseTime(0.0f),
soundFile(0),
currentBlock(0),
backToStart(false),
blockLoaded(false)
{
	int i, j;

	setType("AudioClip");

	for(i=0;i<numBlocks;++i)
	{
		blockArray.push_back(new float[streamBlockSize]);
		blockNeedsLoaded.push_back(false);
		for(j=0;j<streamBlockSize;++j)
			blockArray[i][j] = 0.0f;
	}
}

//------------------------------------------------------------------------------
AudioClip::~AudioClip()
{
	int i;

	if(blockArray.size() > 0)
	{
		for(i=(numBlocks-1);i>=0;--i)
			delete [] (blockArray[i]);
	}

	if(soundFile)
		delete soundFile;
}

//------------------------------------------------------------------------------
void AudioClip::initialise()
{
	ThreadManager::getInstance().initialiseThread(this, 62);
}

//------------------------------------------------------------------------------
float *AudioClip::getNextBlock(ParentNode *parent)
{
	if(soundFile)
	{
		critSec.lock();
		{
			//If the file size is less than the size of a block, there's not
			//much point loading multiple blocks...
			if((streamBlockSize < soundFile->getFileSize()) && (!blockLoaded))
			{
				blockNeedsLoaded[currentBlock] = true;
				++currentBlock;
				currentBlock %= numBlocks;
				blockLoaded = true;
			}
		}
		critSec.unlock();
	}

	return blockArray[currentBlock];
}

//------------------------------------------------------------------------------
float *AudioClip::getFirstBlock(ParentNode *parent)
{
	if((streamBlockSize > soundFile->getFileSize()) && (!blockLoaded))
	{
		critSec.lock();
		{
			backToStart = true;
			currentBlock = 0;
			blockNeedsLoaded[currentBlock] = true;
			//To load the block.
			threadIteration();

			blockLoaded = true;
		}
		critSec.unlock();
	}

	return blockArray[currentBlock];
}

//------------------------------------------------------------------------------
int AudioClip::getFileSize() const
{
	if(soundFile)
		return soundFile->getFileSize();
	else
		return 0;
}

//------------------------------------------------------------------------------
void AudioClip::audioBlockStart()
{
	blockLoaded = false;
}

//------------------------------------------------------------------------------
void AudioClip::threadIteration()
{
	int i;

	if(soundFile)
	{
		//Check if we're supposed to go back to the start of the file.
		if(backToStart)
		{
			cout << "backToStart" << endl;
			soundFile->seekg(0);
			backToStart = false;
		}

		for(i=0;i<numBlocks;++i)
		{
			//If this block needs loaded, load it.
			if(blockNeedsLoaded[i])
			{
				(*soundFile) >> blockArray[i];
				blockNeedsLoaded[i] = false;
			}
		}
	}
}

//------------------------------------------------------------------------------
void AudioClip::setAttribute(const string& attribute, EventValue& val)
{
	int i, j;

	if(attribute == "url")
	{
		if(soundFile)
			delete soundFile;

		//Create appropriate SoundFileLoader.
		soundFile = SFLoaderFactory::getInstance().getFileLoader(val);

		//Set block size.
		soundFile->width(streamBlockSize);

		//Load the first block.
		(*soundFile) >> blockArray[0];
		blockNeedsLoaded[0] = false;

		//Load or clear the other blocks.
		if(streamBlockSize < soundFile->getFileSize())
		{
			for(i=1;i<numBlocks;++i)
			{
				(*soundFile) >> blockArray[i];
				blockNeedsLoaded[i] = false;
			}
		}
		else //Just in case...
		{
			for(i=1;i<numBlocks;++i)
			{
				for(j=0;j<streamBlockSize;++j)
					blockArray[i][j] = 0.0f;
			}
		}

		//Reset currentBlock so we start from the beginning.
		currentBlock = 0;

		//Inform parents our url's changed.
		/*vector<ParentNode *>& parentNodes = getParent();
		for(k=0;k<parentNodes.size();++k)
			parentNodes[k]->setAttribute("audioClipReady", "true");*/
		stupid = (string)val;
	}
	else if(attribute == "loop")
	{
		StringParser parse((const string&)val);

		if(parse.getNumBools() > 0)
			loop = parse.getBool();
	}
	else if(attribute == "pitch")
		pitch = val;
	else if(attribute == "startTime")
		startTime = val;
	else if(attribute == "stopTime")
		stopTime = val;
	else if(attribute == "pauseTime")
		pauseTime = val;
	else if(attribute == "resumeTime")
		resumeTime = val;
}

}
