//	AudioClip.h - AudioClip node, used to load sound files from disk.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef AUDIOCLIP_H_
#define AUDIOCLIP_H_

#include "DllStuff.h"
#include "AbstractNode.h"
#include "CriticalSection.h"
#include "ThreadManager.h"

#include <vector>

namespace HeilanX3D
{

class SoundFileLoader;

///	AudioClip node, used to load sound files from disk.
/*!
	The way it works should be very simple - Heilan streams any audio from disk,
	so AudioClip splits its audio into blocks, and it's the playback node's
	responsibility to call getNextBlock() when it needs the next block.  Note
	that getNextBlock() automatically loops round to the start of the audio file
	when it reaches the end, to ensure that it always returns audio.

	The audio is actually streamed off the disk in the threadIteration() method,
	because we don't want to be doing any disk i/o in the audio thread.
 */
class dllStuff AudioClip : public ChildNode,
						   public ThreadObject
{
  public:
	///	Constructor.
	AudioClip();
	///	Destructor.
	~AudioClip();

	///	Initialises the blocks.
	void initialise();

	///	Returns a pointer to the current block of audio.
	/*!
		The audio pointed to by the returned pointer will be of size
		streamBlockSize.
	 */
	float *getCurrentBlock() const {return blockArray[currentBlock];};
	///	Increments the currentBlock parameter, returns a pointer to the new block.
	/*!
		The audio pointed to by the returned pointer will be of size
		streamBlockSize.
	 */
	float *getNextBlock(ParentNode *parent);
	///	Returns a pointer to the first block of the file.
	/*!
		Note that this actually loads the block itself, so should not be called
		from the audio thread.
	 */
	float *getFirstBlock(ParentNode *parent);
	///	Returns the size of the audio file.
	int getFileSize() const;

	///	Overridden to do nothing.
	void graphicDoStuff() {};
	///Overridden to do nothing.
	//AmbisonicData *audioDoStuff(int numSamples, const ThreeFloats& pos) {return 0;};
	///	So we can update the blockLoaded variable.
	void audioBlockStart();
	///	Overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	We use this to load the next block of the sound file.
	void threadIteration();

	///	To set the url etc.
	/*!
		Setting the url attribute loads the file.
	 */
	void setAttribute(const std::string& attribute, EventValue& val);

	///	Returns the value of the node's loop attribute.
	bool getLoop() const {return loop;};
	///	Returns the value of the node's pitch attribute.
	float getPitch() const {return pitch;};
	///	Returns the value of the node's startTime attribute.
	float getStartTime() const {return startTime;};
	///	Returns the value of the node's stopTime attribute.
	float getStopTime() const {return stopTime;};
	///	Returns the value of the node's resumeTime attribute.
	float getResumeTime() const {return resumeTime;};
	///	Returns the value of the node's pauseTime attribute.
	float getPauseTime() const {return pauseTime;};

	///	The size of the stream blocks is fixed, as is the number of blocks.
	enum
	{
		streamBlockSize = 11025,
		numBlocks = 3
	};

	static NodeConstructor<AudioClip> constructor;
  private:
	///	The node's loop attribute.
	bool loop;
	///	The node's pitch attribute.
	float pitch;
	///	The node's startTime attribute.
	float startTime;
	///	The node's stopTime attribute.
	float stopTime;
	///	The node's resumeTime attribute.
	float resumeTime;
	///	The node's pauseTime attribute.
	float pauseTime;

	///	Pointer to the sound file.
	SoundFileLoader *soundFile;

	///	Our array of blocks.
	std::vector<float *> blockArray;
	///	Array indicating whether the blocks need loading.
	std::vector<bool> blockNeedsLoaded;

	///	Index to our current block.
	int currentBlock;

	///	Whether we need to go back to the start of the file in idle().
	bool backToStart;

	///	Used to make sure we don't load a new block twice if the node's being USE'd.
	bool blockLoaded;
	///	Used to protect get...Block() when we're multithreaded and USE'd.
	CriticalSection critSec;

 	std::string stupid;
};
	
}

#endif
