//	Background.cpp - Node determining the background of the scene.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Background.h"
#include "StringParser.h"
#include "AttributeHelper.h"
#include "Camera.h"
#include "FileFinder.h"

#include "SDL_image.h"
#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<Background> Background::constructor;

//------------------------------------------------------------------------------
Background::Background():
BindableNode(),
backTexture(0),
bottomTexture(0),
frontTexture(0),
leftTexture(0),
rightTexture(0),
topTexture(0),
skyNeedsUpdating(false),
groundNeedsUpdating(false)
{
	setType("Background");
	setVisual(true);
}

//------------------------------------------------------------------------------
Background::~Background()
{
	BindableNodeSingleton::getInstance()->unregisterNode(this);
}

//------------------------------------------------------------------------------
void Background::initialise()
{
	int x, y;

	//Calculate sky hemisphere vertices.
	//Main sphere.
	for(y=NumVertSegments;y<(NumSegments-1);++y)
	{
		for(x=0;x<NumSegments;++x)
			skyVerts[y-NumVertSegments][x] = getVertex(x, y);
	}
	//Sphere top.
	for(x=0;x<NumSegments;++x)
	{
		skyVerts[NumVertSegments-1][x].x = 0.0f;
		skyVerts[NumVertSegments-1][x].y = 1.0f;
		skyVerts[NumVertSegments-1][x].z = 0.0f;
	}

	//Calculate ground hemisphere vertices.
	//Main sphere.
	for(y=1;y<NumVertSegments;++y)
	{
		for(x=0;x<NumSegments;++x)
			groundVerts[y][x] = getVertex(x, y);
	}
	for(x=0;x<NumSegments;++x)
		groundVerts[NumVertSegments][x] = getVertex(x, NumVertSegments);
	//Sphere bottom.
	for(x=0;x<NumSegments;++x)
	{
		groundVerts[0][x].x = 0.0f;
		groundVerts[0][x].y = -1.0f;
		groundVerts[0][x].z = 0.0f;
	}

	//Register with the BindableNodeSingleton.
	BindableNodeSingleton::getInstance()->registerNode(this);
}

//------------------------------------------------------------------------------
void Background::graphicDoStuff()
{
	int x, y;
	const float boxWidth = 0.125f;
	float tempMatrix[16];

	if(skyNeedsUpdating)
	{
		updateSkyColours();
		skyNeedsUpdating = false;
	}
	if(groundNeedsUpdating)
	{
		updateGroundColours();
		groundNeedsUpdating = false;
	}

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	if((skyColor.size() > 0) && (groundColor.size() == 0))
		glClearColor(skyColor[0].x, skyColor[0].y, skyColor[0].z, 0.0f);

	//Move so we're centred around the camera.
	//I got this idea from http://www.levp.de/3d/index.html
	glGetFloatv(GL_MODELVIEW_MATRIX, tempMatrix);
	tempMatrix[12] = 0.0f;
	tempMatrix[13] = 0.0f;
	tempMatrix[14] = 0.0f;
	glLoadMatrixf(tempMatrix);

	//Disable writing to the z-buffer, so everything appears behind the actual
	//scene.
    glDepthMask(GL_FALSE);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);

	//Draw sky hemisphere.
	glBegin(GL_QUADS);
	{
		for(y=0;y<(NumVertSegments-1);++y)
		{
			for(x=0;x<(NumSegments-1);++x)
			{
				glColor3f(skyColours[y][x].x,
						  skyColours[y][x].y,
						  skyColours[y][x].z);
				glVertex3f(skyVerts[y][x].x,
						   skyVerts[y][x].y,
						   skyVerts[y][x].z);
				glColor3f(skyColours[y][x+1].x,
						  skyColours[y][x+1].y,
						  skyColours[y][x+1].z);
				glVertex3f(skyVerts[y][x+1].x,
						   skyVerts[y][x+1].y,
						   skyVerts[y][x+1].z);
				glColor3f(skyColours[y+1][x+1].x,
						  skyColours[y+1][x+1].y,
						  skyColours[y+1][x+1].z);
				glVertex3f(skyVerts[y+1][x+1].x,
						   skyVerts[y+1][x+1].y,
						   skyVerts[y+1][x+1].z);
				glColor3f(skyColours[y+1][x].x,
						  skyColours[y+1][x].y,
						  skyColours[y+1][x].z);
				glVertex3f(skyVerts[y+1][x].x,
						   skyVerts[y+1][x].y,
						   skyVerts[y+1][x].z);
			}

			glColor3f(skyColours[y][x].x,
					  skyColours[y][x].y,
					  skyColours[y][x].z);
			glVertex3f(skyVerts[y][x].x,
					   skyVerts[y][x].y,
					   skyVerts[y][x].z);
			glColor3f(skyColours[y][0].x,
					  skyColours[y][0].y,
					  skyColours[y][0].z);
			glVertex3f(skyVerts[y][0].x,
					   skyVerts[y][0].y,
					   skyVerts[y][0].z);
			glColor3f(skyColours[y+1][0].x,
					  skyColours[y+1][0].y,
					  skyColours[y+1][0].z);
			glVertex3f(skyVerts[y+1][0].x,
					   skyVerts[y+1][0].y,
					   skyVerts[y+1][0].z);
			glColor3f(skyColours[y+1][x].x,
					  skyColours[y+1][x].y,
					  skyColours[y+1][x].z);
			glVertex3f(skyVerts[y+1][x].x,
					   skyVerts[y+1][x].y,
					   skyVerts[y+1][x].z);
		}
	}
	glEnd();

	//Draw ground hemisphere.
	if(groundColor.size() > 0)
	{
		glBegin(GL_QUADS);
		{
			//for(y=0;y<(NumVertSegments-1);++y)
			for(y=0;y<NumVertSegments;++y)
			{
				for(x=0;x<(NumSegments-1);++x)
				{
					glColor3f(groundColours[y][x].x,
						  	  groundColours[y][x].y,
						  	  groundColours[y][x].z);
					glVertex3f(groundVerts[y][x].x,
						   	   groundVerts[y][x].y,
						   	   groundVerts[y][x].z);
					glColor3f(groundColours[y][x+1].x,
						  	  groundColours[y][x+1].y,
						  	  groundColours[y][x+1].z);
					glVertex3f(groundVerts[y][x+1].x,
						   	   groundVerts[y][x+1].y,
							   groundVerts[y][x+1].z);
					glColor3f(groundColours[y+1][x+1].x,
							  groundColours[y+1][x+1].y,
							  groundColours[y+1][x+1].z);
					glVertex3f(groundVerts[y+1][x+1].x,
							   groundVerts[y+1][x+1].y,
							   groundVerts[y+1][x+1].z);
					glColor3f(groundColours[y+1][x].x,
							  groundColours[y+1][x].y,
							  groundColours[y+1][x].z);
					glVertex3f(groundVerts[y+1][x].x,
							   groundVerts[y+1][x].y,
							   groundVerts[y+1][x].z);
				}
	
				glColor3f(groundColours[y][x].x,
						  groundColours[y][x].y,
						  groundColours[y][x].z);
				glVertex3f(groundVerts[y][x].x,
						   groundVerts[y][x].y,
						   groundVerts[y][x].z);
				glColor3f(groundColours[y][0].x,
						  groundColours[y][0].y,
						  groundColours[y][0].z);
				glVertex3f(groundVerts[y][0].x,
						   groundVerts[y][0].y,
						   groundVerts[y][0].z);
				glColor3f(groundColours[y+1][0].x,
						  groundColours[y+1][0].y,
						  groundColours[y+1][0].z);
				glVertex3f(groundVerts[y+1][0].x,
						   groundVerts[y+1][0].y,
						   groundVerts[y+1][0].z);
				glColor3f(groundColours[y+1][x].x,
						  groundColours[y+1][x].y,
						  groundColours[y+1][x].z);
				glVertex3f(groundVerts[y+1][x].x,
						   groundVerts[y+1][x].y,
						   groundVerts[y+1][x].z);
			}
		}
		glEnd();
	}
	glColor3f(1.0f, 1.0f, 1.0f);

	//Draw textures if they exist.
	if(bottomTexture)
	{
		glBindTexture(GL_TEXTURE_2D, bottomTexture);

		glBegin(GL_QUADS);
		{
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(-boxWidth, -boxWidth, -boxWidth);
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(boxWidth, -boxWidth, -boxWidth);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(boxWidth, -boxWidth, boxWidth);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(-boxWidth, -boxWidth, boxWidth);
		}
		glEnd();

		glBindTexture(GL_TEXTURE_2D, 0);
	}
	if(backTexture)
	{
		glBindTexture(GL_TEXTURE_2D, frontTexture);

		glBegin(GL_QUADS);
		{
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(-boxWidth, -boxWidth, -boxWidth);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-boxWidth,  boxWidth, -boxWidth);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(boxWidth,  boxWidth, -boxWidth);
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(boxWidth, -boxWidth, -boxWidth);
		}
		glEnd();

		glBindTexture(GL_TEXTURE_2D, 0);
	}
	if(rightTexture)
	{
		glBindTexture(GL_TEXTURE_2D, rightTexture);

		glBegin(GL_QUADS);
		{
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(boxWidth, -boxWidth, -boxWidth);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(boxWidth, boxWidth, -boxWidth);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(boxWidth, boxWidth, boxWidth);
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(boxWidth, -boxWidth, boxWidth);
		}
		glEnd();

		glBindTexture(GL_TEXTURE_2D, 0);
	}
	if(frontTexture)
	{
		glBindTexture(GL_TEXTURE_2D, backTexture);

		glBegin(GL_QUADS);
		{
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(-boxWidth, -boxWidth, boxWidth);
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(boxWidth, -boxWidth, boxWidth);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(boxWidth, boxWidth, boxWidth);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(-boxWidth, boxWidth, boxWidth);
		}
		glEnd();

		glBindTexture(GL_TEXTURE_2D, 0);
	}
	if(leftTexture)
	{
		glBindTexture(GL_TEXTURE_2D, leftTexture);

		glBegin(GL_QUADS);
		{
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(-boxWidth, -boxWidth, -boxWidth);
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(-boxWidth, -boxWidth, boxWidth);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-boxWidth, boxWidth, boxWidth);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(-boxWidth, boxWidth, -boxWidth);
		}
		glEnd();

		glBindTexture(GL_TEXTURE_2D, 0);
	}
	if(topTexture)
	{
		glBindTexture(GL_TEXTURE_2D, topTexture);

		glBegin(GL_QUADS);
		{
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(-boxWidth, boxWidth, -boxWidth);
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(-boxWidth, boxWidth, boxWidth);
			glTexCoord2f(0.0f, 1.0f);
			glVertex3f(boxWidth, boxWidth, boxWidth);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(boxWidth, boxWidth, -boxWidth);
		}
		glEnd();

		glBindTexture(GL_TEXTURE_2D, 0);
	}

	glEnable(GL_LIGHTING);

	//Re-enable writing to the z-buffer.
    glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);

	glPopMatrix();
}

//------------------------------------------------------------------------------
void Background::setAttribute(const string& attribute, EventValue& val)
{
	unsigned int i;
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "groundAngle")
	{
		StringParser parser(val);
		AttributeHelper< vector<float> > helper(&groundAngle);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumNumbers() > 0)
		{
			groundAngle.clear();
			for(i=0;i<(unsigned)parser.getNumNumbers();++i)
				groundAngle.push_back(parser.getFloat(i));
		}

		groundNeedsUpdating = true;
	}
	else if(realAtt == "groundColor")
	{
		StringParser parser(val);
		AttributeHelper< vector<ThreeFloats> > helper(&groundColor);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
		{
			groundColor.clear();
			for(i=0;i<(unsigned)parser.getNumThreeFloats();++i)
				groundColor.push_back(parser.getThreeFloats(i));
			if(groundAngle.size() != groundColor.size())
			{
				groundAngle.clear();
				for(i=0;i<groundColor.size();++i)
					groundAngle.push_back(((float)i/(float)groundColor.size())*1.570796327f);
			}
		}

		groundNeedsUpdating = true;
	}
	else if(realAtt == "backUrl")
	{
		backUrl = (string)val;

		loadTexture(backTexture, backUrl);
	}
	else if(realAtt == "bottomUrl")
	{
		bottomUrl = (string)val;

		loadTexture(bottomTexture, bottomUrl);
	}
	else if(realAtt == "frontUrl")
	{
		frontUrl = (string)val;

		loadTexture(frontTexture, frontUrl);
	}
	else if(realAtt == "leftUrl")
	{
		leftUrl = (string)val;

		loadTexture(leftTexture, leftUrl);
	}
	else if(realAtt == "rightUrl")
	{
		rightUrl = (string)val;

		loadTexture(rightTexture, rightUrl);
	}
	else if(realAtt == "topUrl")
	{
		topUrl = (string)val;

		loadTexture(topTexture, topUrl);
	}
	else if(realAtt == "skyAngle")
	{
		StringParser parser(val);
		AttributeHelper< vector<float> > helper(&skyAngle);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumNumbers() > 0)
		{
			skyAngle.clear();
			for(i=0;i<(unsigned)parser.getNumNumbers();++i)
				skyAngle.push_back(parser.getFloat(i));
		}

		skyNeedsUpdating = true;
	}
	else if(realAtt == "skyColor")
	{
		StringParser parser(val);
		AttributeHelper< vector<ThreeFloats> > helper(&skyColor);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
		{
			skyColor.clear();
			for(i=0;i<(unsigned)parser.getNumThreeFloats();++i)
				skyColor.push_back(parser.getThreeFloats(i));
			if(skyAngle.size() != skyColor.size())
			{
				skyAngle.clear();
				for(i=0;i<skyColor.size();++i)
					skyAngle.push_back(((float)i/(float)skyColor.size())*1.570796327f);
			}
		}

		skyNeedsUpdating = true;
	}
}

//------------------------------------------------------------------------------
void Background::loadTexture(GLuint& texture, const std::string& url)
{
	string tempPath;
	SDL_Surface *tempTex = 0;
	SDL_Surface *tempTex2 = 0;
	SDL_Surface *tempTex3 = 0;

	tempPath = FileFinder::getInstance().findFile(url);
	if(tempPath != "")
	{
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
	
		//This is the easiest way I could find to create a sensible SDL_PixelFormat
		//for SDL_ConvertSurface().
#ifdef __BIG_ENDIAN__
		tempTex3 = SDL_CreateRGBSurface(SDL_SWSURFACE,
										1,
										1,
										32,
										0xff000000,
										0x00ff0000,
										0x0000ff00,
										0x000000ff);
#else
		tempTex3 = SDL_CreateRGBSurface(SDL_SWSURFACE,
										1,
										1,
										32,
										0x000000ff,
										0x0000ff00,
										0x00ff0000,
										0xff000000);
#endif
	
		if((tempTex = IMG_Load(tempPath.c_str())))
		{
			//We first convert the image into a specific format, before we call
			//glTexImage2D.
			if((tempTex2 = SDL_ConvertSurface(tempTex,
				tempTex3->format,
				SDL_SWSURFACE)))
			{
				glGenTextures(1, &texture);
	
				glBindTexture(GL_TEXTURE_2D, texture);
	
				glTexImage2D(GL_TEXTURE_2D,
							 0,
							 GL_RGBA,
							 tempTex2->w,
							 tempTex2->h,
							 0,
							 GL_RGBA,
							 GL_UNSIGNED_BYTE,
							 tempTex2->pixels);
	
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	
				SDL_FreeSurface(tempTex2);
				SDL_FreeSurface(tempTex);
				glBindTexture(GL_TEXTURE_2D, 0);
			}
			else
			{
				SDL_FreeSurface(tempTex);
				cout << "Background: Could not load image: " << tempPath << endl;
			}
		}
		else
			cout << "Background: Could not load image: " << tempPath << endl;
	
		SDL_FreeSurface(tempTex3);
	}
	else
		cout << "Background: " << url << " does not seem to exist?" << endl;
}

//------------------------------------------------------------------------------
void Background::updateSkyColours()
{
	int i, j;
	int numColours;
	unsigned int currentAngle = 0;
	float tempCount = 0.0f;
	float tempf;
	ThreeFloats tempCol;

	if(skyColor.size() > NumVertSegments)
		numColours = NumVertSegments;
	else
		numColours = skyColor.size();

	if(skyAngle.size() != skyColor.size())
	{
		cout << "Background: skyAngle's size does not match skyColor's" << endl;
		cout << "skyAngle.size = " << skyAngle.size() << endl;
		cout << "skyColor.size = " << skyColor.size() << endl;
		return;
	}

	if(numColours == 1)
	{
		for(i=0;i<NumVertSegments;++i)
		{
			for(j=0;j<NumSegments;++j)
				skyColours[i][j] = skyColor[0];
		}
	}
	else
	{
		for(i=0;i<NumVertSegments;++i)
		{
			if(currentAngle < (skyColor.size()-1))
			{
				tempf = (tempCount - skyAngle[currentAngle]);
				tempf /= (skyAngle[currentAngle+1]-skyAngle[currentAngle]);
				tempCol = ((1.0f-tempf) * skyColor[currentAngle]) +
						  (tempf * skyColor[currentAngle+1]);
			}
			else
				tempCol = skyColor[currentAngle];

			for(j=0;j<NumSegments;++j)
				skyColours[i][j] = tempCol;
				//skyColours[(i*NumSegments)+j] = tempCol;

			tempCount += (1.0f/static_cast<float>(NumVertSegments)) *
						 1.570796327f;
			if(currentAngle < (skyColor.size()-1))
			{
				if(tempCount > skyAngle[currentAngle+1])
					++currentAngle;
			}
			else
				currentAngle = (skyColor.size()-1);
		}
	}
}

//------------------------------------------------------------------------------
void Background::updateGroundColours()
{
	int i, j;
	int numColours;
	unsigned int currentAngle = 0;
	float tempCount = 0.0f;
	float tempf;
	ThreeFloats tempCol;

	if(groundColor.size() > NumVertSegments)
		numColours = NumVertSegments;
	else
		numColours = groundColor.size();

	if(groundAngle.size() != groundColor.size())
	{
		cout << "Background: groundAngle's size does not match groundColor's" << endl;
		cout << "groundAngle.size = " << groundAngle.size() << endl;
		cout << "groundColor.size = " << groundColor.size() << endl;
		return;
	}

	if(numColours == 1)
	{
		for(i=0;i<NumVertSegments;++i)
		{
			for(j=0;j<NumSegments;++j)
				groundColours[i][j] = groundColor[0];
		}
		if(groundColor.size() > 0)
		{
			for(i=0;i<NumSegments;++i)
				groundColours[NumVertSegments][i] = groundColor[groundColor.size()-1];
		}
	}
	else
	{
		for(i=0;i<NumVertSegments;++i)
		{
			if(currentAngle < (groundColor.size()-1))
			{
				tempf = (tempCount - groundAngle[currentAngle]);
				tempf /= (groundAngle[currentAngle+1]-groundAngle[currentAngle]);
				tempCol = ((1.0f-tempf) * groundColor[currentAngle]) +
						  (tempf * groundColor[currentAngle+1]);
			}
			else
				tempCol = groundColor[currentAngle];

			for(j=0;j<NumSegments;++j)
				groundColours[i][j] = tempCol;
				//groundColours[(i*NumSegments)+j] = tempCol;

			tempCount += (1.0f/static_cast<float>(NumVertSegments)) *
						 1.570796327f;
			if(currentAngle < (groundColor.size()-1))
			{
				if(tempCount > groundAngle[currentAngle+1])
					++currentAngle;
			}
			else
				currentAngle = (groundColor.size()-1);
		}
		if(groundColor.size() > 0)
		{
			for(i=0;i<NumSegments;++i)
				groundColours[NumVertSegments][i] = groundColor[groundColor.size()-1];
		}
	}
}

}
