//	Background.h - Node determining the background of the scene.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef BACKGROUND_H_
#define BACKGROUND_H_

#include "BindableNode.h"

#include "GLHeaders.h"
#include <cmath>

namespace HeilanX3D
{

///	Node determining the background of the scene.
class Background : public BindableNode
{
  public:
	///	Constructor.
	Background();
	///	Destructor.
	~Background();

	///	Initialises various bits and pieces.
	void initialise();

	///	Overridden to draw the background.
	void graphicDoStuff();
	///	Overridden to do nothing.
	/*AmbisonicData *audioDoStuff(int numSamples,
								const ThreeFloats& pos) {return 0;};*/

	///	Overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Called to set the value of the named attribute.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<Background> constructor;
  private:
	///	Helper method to load a texture.
	void loadTexture(GLuint& texture, const std::string& url);
	///	Helper method to update the sky colours correctly.
	void updateSkyColours();
	///	Helper method to update the ground colours correctly.
	void updateGroundColours();
	///	Helper method for calcVertex().
	inline ThreeFloats getVertex(int y, int x)
	{
		ThreeFloats retval;

		/*if(x >= NumSections)
			x -= NumSections;
		if(y >= NumSections)
			y -= NumSections;*/

		//x => phi, y => theta.

		retval.x = cosf((((float)x/(float)NumSegments)-0.5f) * 3.141592653f) * 
				   cosf(((float)y/(float)NumSegments) * 6.283185306f);
		retval.y = sinf((((float)x/(float)NumSegments)-0.5f) * 3.141592653f);
		retval.z = cosf((((float)x/(float)NumSegments)-0.5f) * 3.141592653f) *
				   sinf(((float)y/(float)NumSegments) * 6.283185306f);

		return retval;
	}

	///	Enum containing constants for the number of segments used.
	enum
	{
		NumSegments = 32,
		NumVertSegments = 16
	};

	///	The array of ground angles (to generate concentric bands of colour).
	std::vector<float> groundAngle;
	///	The array of ground colours.
	std::vector<ThreeFloats> groundColor;
	///	Url of the back texture.
	std::string backUrl;
	///	Url of the bottom texture.
	std::string bottomUrl;
	///	Url of the front texture.
	std::string frontUrl;
	///	Url of the left texture.
	std::string leftUrl;
	///	Url of the right texture.
	std::string rightUrl;
	///	Url of the top texture.
	std::string topUrl;
	///	The array of sky angles (to generate concentric bands of colour).
	std::vector<float> skyAngle;
	///	The current sky colour.
	std::vector<ThreeFloats> skyColor;

	///	Back texture.
	GLuint backTexture;
	///	Bottom texture.
	GLuint bottomTexture;
	///	Front texture.
	GLuint frontTexture;
	///	Left texture.
	GLuint leftTexture;
	///	Right texture.
	GLuint rightTexture;
	///	Top texture.
	GLuint topTexture;

	///	Vertices for the sky.
	ThreeFloats skyVerts[NumVertSegments][NumSegments];
	///	Colours for the sky.
	ThreeFloats skyColours[NumVertSegments][NumSegments];
	///	Vertices for the ground.
	ThreeFloats groundVerts[NumVertSegments+1][NumSegments];
	///	Colours for the ground.
	ThreeFloats groundColours[NumVertSegments+1][NumSegments];

	///	Used to tell us we need to update the colours of the sky hemisphere.
	bool skyNeedsUpdating;
	///	Used to tell us we need to update the colours of the ground hemisphere.
	bool groundNeedsUpdating;
};

}

#endif
