//	BindableNode.cpp - Base class for all bindable nodes.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "BindableNode.h"

#include <cassert>
#include <iostream>
#include <algorithm>

using namespace std;

namespace HeilanX3D
{

//------------------------------------------------------------------------------
BindableNode::BindableNode():
ChildNode()
{
	
}

//------------------------------------------------------------------------------
BindableNode::~BindableNode()
{
	
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
BindableNodeSingleton *BindableNodeSingleton::getInstance()
{
	static BindableNodeSingleton retval;

	return &retval;
}

//------------------------------------------------------------------------------
void BindableNodeSingleton::registerNode(BindableNode *node)
{
	AllNodes::iterator it;

	//Search for other nodes of the same type.
	it = allNodes.find(node->getType());

	//If we've found them, push node onto the end of the vector.
	if(it != allNodes.end())
	{
		//If we switch to fullscreen this method will get called again, so we
		//have to make sure we don't accidentally add the same node twice,
		//thrice, etc.
		if(find(it->second->begin(), it->second->end(), node) == it->second->end())
			it->second->push_back(node);
	}
	else //Otherwise, create new Nodes entry in allNodes.
	{
		Nodes *newNodes = new Nodes;
		newNodes->push_back(node);
		allNodes.insert(std::make_pair(node->getType(), newNodes));

		//And make this node the currently active one for this type.
		activeNodes.insert(std::make_pair(node->getType(), node));
	}
}

//------------------------------------------------------------------------------
void BindableNodeSingleton::unregisterNode(BindableNode *node)
{
	AllNodes::iterator it;
	Nodes::iterator it2;
	ActiveNodes::iterator it3;

	//Search for the correct node type.
	it = allNodes.find(node->getType());

	//Make sure they exist.
	//assert(it!=allNodes.end());
	if(it == allNodes.end())
		return;

	//Search for this particular node.
	for(it2=it->second->begin();it2!=it->second->end();++it2)
	{
		if((*it2)==node)
		{
			it->second->erase(it2);
			break;
		}
	}

	//Remove from activeNodes and allNodes if necessary.
	it3 = activeNodes.find(node->getType());
	//assert(it3!=activeNodes.end());
	if(it3 == activeNodes.end())
		return;
	if(it->second->size() < 1)
	{
		Nodes *tempNodes = it->second;
		allNodes.erase(node->getType());
		delete tempNodes;
		activeNodes.erase(node->getType());
	}
	else if(it3->second == node)
	{
		it2 = it->second->begin();
		it3->second = *it2;
	}
}

//------------------------------------------------------------------------------
BindableNode *BindableNodeSingleton::getActiveNode(const std::string& type) const
{
	ActiveNodes::const_iterator it;

	//Search for active node of the correct type.
	it = activeNodes.find(type);

	if(it != activeNodes.end())
		return it->second;
	else
		return 0;
}

//------------------------------------------------------------------------------
void BindableNodeSingleton::nextNode(const std::string& type)
{
	ActiveNodes::iterator actIt;
	AllNodes::iterator allIt;
	Nodes *currentNodes;
	Nodes::iterator curIt;

	//Search for active node of the correct type.
	actIt = activeNodes.find(type);
	//Search for all nodes of the correct type.
	allIt = allNodes.find(type);

	//If allIt is empty, then there are no nodes of that type in the scene.
	if(allIt == allNodes.end())
		return;

	//If actIt is empty, but allIt isn't, it means there's an error somewhere.
	assert(actIt != activeNodes.end());

	//Find the currently active node in allNodes.
	currentNodes = allIt->second;
	for(curIt=currentNodes->begin();curIt!=currentNodes->end();++curIt)
	{
		if((*curIt) == actIt->second)
		{
			//Move onto the next node in currentNodes, set it as active.
			++curIt;
			if(curIt == currentNodes->end())
				curIt = currentNodes->begin();
			actIt->second = *curIt;
		}
	}
}

//------------------------------------------------------------------------------
void BindableNodeSingleton::previousNode(const std::string& type)
{
	ActiveNodes::iterator actIt;
	AllNodes::iterator allIt;
	Nodes *currentNodes;
	Nodes::iterator curIt;

	//Search for active node of the correct type.
	actIt = activeNodes.find(type);
	//Search for all nodes of the correct type.
	allIt = allNodes.find(type);

	//If allIt is empty, then there are no nodes of that type in the scene.
	if(allIt == allNodes.end())
		return;

	//If actIt is empty, but allIt isn't, it means there's an error somewhere.
	assert(actIt != activeNodes.end());

	//Find the currently active node in allNodes.
	currentNodes = allIt->second;
	for(curIt=currentNodes->begin();curIt!=currentNodes->end();++curIt)
	{
		if((*curIt) == actIt->second)
		{
			//Move onto the previous node in currentNodes, set it as active.
			if(curIt == currentNodes->begin())
				curIt = currentNodes->end();
			--curIt;
			actIt->second = *curIt;
		}
	}
}

//------------------------------------------------------------------------------
int BindableNodeSingleton::getNumNodes(const std::string& type) const
{
	AllNodes::const_iterator it;

	//Search for node array of the correct type.
	it = allNodes.find(type);

	if(it != allNodes.end())
		return it->second->size();
	else
		return 0;
}

//------------------------------------------------------------------------------
BindableNode *BindableNodeSingleton::getIndexedNode(const std::string& type,
													int index) const
{
	AllNodes::const_iterator it;

	//Search for node array of the correct type.
	it = allNodes.find(type);

	if(it != allNodes.end())
		return (*(it->second))[index];
	else
		return 0;
}

//------------------------------------------------------------------------------
void BindableNodeSingleton::setIndexedNode(const std::string& type, int index)
{
	int i;
	ActiveNodes::iterator actIt;
	AllNodes::iterator allIt;
	Nodes *currentNodes;
	Nodes::iterator curIt;

	//Search for active node of the correct type.
	actIt = activeNodes.find(type);
	//Search for all nodes of the correct type.
	allIt = allNodes.find(type);

	//If allIt is empty, then there are no nodes of that type in the scene.
	if(allIt == allNodes.end())
		return;

	//If actIt is empty, but allIt isn't, it means there's an error somewhere.
	assert(actIt != activeNodes.end());

	//Set the currently active node in allNodes.
	currentNodes = allIt->second;
	for(i=0, curIt=currentNodes->begin();
		curIt!=currentNodes->end();
		++i, ++curIt)
	{
		if(i == index)
			actIt->second = *curIt;
	}
}

//------------------------------------------------------------------------------
BindableNodeSingleton::BindableNodeSingleton()
{
	
}

//------------------------------------------------------------------------------
BindableNodeSingleton::~BindableNodeSingleton()
{
	AllNodes::iterator it;

	for(it=allNodes.begin();it!=allNodes.end();++it)
		delete it->second;
}

}
