//	Cylinder.cpp - Class representing an X3D cylinder node.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Cylinder.h"
#include "StringParser.h"

using namespace std;

namespace HeilanX3D
{

NodeConstructor<Cylinder> Cylinder::constructor;

//------------------------------------------------------------------------------
Cylinder::Cylinder():
ChildNode(),
bottom(true),
radius(1.0f),
height(2.0f),
side(true),
solid(true),
top(true)
{
	setType("Cylinder");
	setVisual(true);
}

//------------------------------------------------------------------------------
Cylinder::~Cylinder()
{
	
}

//------------------------------------------------------------------------------
void Cylinder::initialise()
{
	int i;
	float xVerts[NumSections];
	float yVerts[NumSections];
	ThreeFloats tf1;
	ThreeFloats tf2;
	ThreeFloats tf3;
	ThreeFloats tf4;
	ThreeFloats tf5;
	ThreeFloats tf6;
	ThreeFloats verts1[3];
	ThreeFloats verts2[3];
	ThreeFloats verts3[3];
	ThreeFloats n1;
	ThreeFloats n2;
	ThreeFloats n3;
	ThreeFloats n4;

	for(i=0;i<NumSections;++i)
	{
		xVerts[i] = sinf(((float)i/(float)NumSections) * 6.283185306f);
		yVerts[i] = cosf(((float)i/(float)NumSections) * 6.283185306f);
	}

	//Generate the sides display list.
	sidesDisplay = glGenLists(1);
	glNewList(sidesDisplay, GL_COMPILE);
	{
		glBegin(GL_TRIANGLES);
		{
			for(i=0;i<(NumSections-1);++i)
			{
				tf1.x = (xVerts[i+1]*radius);
				tf1.y = (height*0.5f);
				tf1.z = (yVerts[i+1]*radius);
				tf2.x = (xVerts[i]*radius);
				tf2.y = (height*0.5f);
				tf2.z = (yVerts[i]*radius);
				tf3.x = (xVerts[i]*radius);
				tf3.y = -(height*0.5f);
				tf3.z = (yVerts[i]*radius);

				tf4.x = (xVerts[i+1]*radius);
				tf4.y = (height*0.5f);
				tf4.z = (yVerts[i+1]*radius);
				tf5.x = (xVerts[i]*radius);
				tf5.y = -(height*0.5f);
				tf5.z = (yVerts[i]*radius);
				tf6.x = (xVerts[i+1]*radius);
				tf6.y = -(height*0.5f);
				tf6.z = (yVerts[i+1]*radius);

				verts1[0].x = (xVerts[i+1]*radius);
				verts1[0].y = (height*0.5f);
				verts1[0].z = (yVerts[i+1]*radius);
				verts1[1].x = (xVerts[i]*radius);
				verts1[1].y = (height*0.5f);
				verts1[1].z = (yVerts[i]*radius);
				verts1[2].x = (xVerts[i]*radius);
				verts1[2].y = -(height*0.5f);
				verts1[2].z = (yVerts[i]*radius);

				if(i > 0)
				{
					verts2[0].x = (xVerts[i]*radius);
					verts2[0].y = (height*0.5f);
					verts2[0].z = (yVerts[i]*radius);
					verts2[1].x = (xVerts[i-1]*radius);
					verts2[1].y = -(height*0.5f);
					verts2[1].z = (yVerts[i-1]*radius);
					verts2[2].x = (xVerts[i]*radius);
					verts2[2].y = -(height*0.5f);
					verts2[2].z = (yVerts[i]*radius);
				}
				else
				{
					verts2[0].x = (xVerts[i]*radius);
					verts2[0].y = (height*0.5f);
					verts2[0].z = (yVerts[i]*radius);
					verts2[1].x = (xVerts[(NumSections-1)]*radius);
					verts2[1].y = -(height*0.5f);
					verts2[1].z = (yVerts[(NumSections-1)]*radius);
					verts2[2].x = (xVerts[i]*radius);
					verts2[2].y = -(height*0.5f);
					verts2[2].z = (yVerts[i]*radius);
				}

				if(i < (NumSections-2))
				{
					verts3[0].x = (xVerts[i+2]*radius);
					verts3[0].y = (height*0.5f);
					verts3[0].z = (yVerts[i+2]*radius);
					verts3[1].x = (xVerts[i+1]*radius);
					verts3[1].y = (height*0.5f);
					verts3[1].z = (yVerts[i+1]*radius);
					verts3[2].x = (xVerts[i+1]*radius);
					verts3[2].y = -(height*0.5f);
					verts3[2].z = (yVerts[i+1]*radius);
				}
				else
				{
					verts3[0].x = (xVerts[0]*radius);
					verts3[0].y = (height*0.5f);
					verts3[0].z = (yVerts[0]*radius);
					verts3[1].x = (xVerts[i+1]*radius);
					verts3[1].y = (height*0.5f);
					verts3[1].z = (yVerts[i+1]*radius);
					verts3[2].x = (xVerts[i+1]*radius);
					verts3[2].y = -(height*0.5f);
					verts3[2].z = (yVerts[i+1]*radius);
				}

				n1 = calcNormal2(verts1[0], verts1[1], verts1[2]);
				n2 = calcNormal2(verts2[0], verts2[1], verts2[2]);
				n3 = calcSmooth(n1, n2);
				n2 = calcNormal2(verts3[0], verts3[1], verts3[2]);
				n4 = calcSmooth(n1, n2);

				glNormal3f(n4.x, n4.y, n4.z);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
									 (((float)i+1.0f)/(float)NumSections),
									 0.0f);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
									 (((float)i+1.0f)/(float)NumSections),
									 0.0f);
				glVertex3f(tf1.x, tf1.y, tf1.z);

				glNormal3f(n3.x, n3.y, n3.z);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
									 (float)i/(float)NumSections,
									 0.0f);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
									 (float)i/(float)NumSections,
									 0.0f);
				glVertex3f(tf2.x, tf2.y, tf2.z);

				glNormal3f(n3.x, n3.y, n3.z);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
									 ((float)i/(float)NumSections),
									 1.0f);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
									 ((float)i/(float)NumSections),
									 1.0f);
				glVertex3f(tf3.x, tf3.y, tf3.z);


				glNormal3f(n4.x, n4.y, n4.z);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
									 (((float)i+1.0f)/(float)NumSections),
									 0.0f);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
									 (((float)i+1.0f)/(float)NumSections),
									 0.0f);
				glVertex3f(tf4.x, tf4.y, tf4.z);

				glNormal3f(n3.x, n3.y, n3.z);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
									 (float)i/(float)NumSections,
									 1.0f);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
									 (float)i/(float)NumSections,
									 1.0f);
				glVertex3f(tf5.x, tf5.y, tf5.z);

				glNormal3f(n4.x, n4.y, n4.z);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
									 (((float)i+1.0f)/(float)NumSections),
									 1.0f);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
									 (((float)i+1.0f)/(float)NumSections),
									 1.0f);
				glVertex3f(tf6.x, tf6.y, tf6.z);
			}
			tf1.x = (xVerts[0]*radius);
			tf1.y = (height*0.5f);
			tf1.z = (yVerts[0]*radius);
			tf2.x = (xVerts[i]*radius);
			tf2.y = (height*0.5f);
			tf2.z = (yVerts[i]*radius);
			tf3.x = (xVerts[i]*radius);
			tf3.y = -(height*0.5f);
			tf3.z = (yVerts[i]*radius);

			tf4.x = (xVerts[0]*radius);
			tf4.y = (height*0.5f);
			tf4.z = (yVerts[0]*radius);
			tf5.x = (xVerts[i]*radius);
			tf5.y = -(height*0.5f);
			tf5.z = (yVerts[i]*radius);
			tf6.x = (xVerts[0]*radius);
			tf6.y = -(height*0.5f);
			tf6.z = (yVerts[0]*radius);

			verts1[0].x = (xVerts[0]*radius);
			verts1[0].y = (height*0.5f);
			verts1[0].z = (yVerts[0]*radius);
			verts1[1].x = (xVerts[i]*radius);
			verts1[1].y = (height*0.5f);
			verts1[1].z = (yVerts[i]*radius);
			verts1[2].x = (xVerts[i]*radius);
			verts1[2].y = -(height*0.5f);
			verts1[2].z = (yVerts[i]*radius);

			verts2[0].x = (xVerts[i]*radius);
			verts2[0].y = (height*0.5f);
			verts2[0].z = (yVerts[i]*radius);
			verts2[1].x = (xVerts[i-1]*radius);
			verts2[1].y = -(height*0.5f);
			verts2[1].z = (yVerts[i-1]*radius);
			verts2[2].x = (xVerts[i]*radius);
			verts2[2].y = -(height*0.5f);
			verts2[2].z = (yVerts[i]*radius);

			verts3[0].x = (xVerts[1]*radius);
			verts3[0].y = (height*0.5f);
			verts3[0].z = (yVerts[1]*radius);
			verts3[1].x = (xVerts[0]*radius);
			verts3[1].y = (height*0.5f);
			verts3[1].z = (yVerts[0]*radius);
			verts3[2].x = (xVerts[0]*radius);
			verts3[2].y = -(height*0.5f);
			verts3[2].z = (yVerts[0]*radius);

			n1 = calcNormal2(verts1[0], verts1[1], verts1[2]);
			n2 = calcNormal2(verts2[0], verts2[1], verts2[2]);
			n3 = calcSmooth(n1, n2);
			n2 = calcNormal2(verts3[0], verts3[1], verts3[2]);
			n4 = calcSmooth(n1, n2);

			glNormal3f(n4.x, n4.y, n4.z);
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
								 (((float)i+1.0f)/(float)NumSections),
								 0.0f);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
								 (((float)i+1.0f)/(float)NumSections),
								 0.0f);
			glVertex3f(tf1.x, tf1.y, tf1.z);

			glNormal3f(n3.x, n3.y, n3.z);
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
								 (float)i/(float)NumSections,
								 0.0f);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
								 (float)i/(float)NumSections,
								 0.0f);
			glVertex3f(tf2.x, tf2.y, tf2.z);

			glNormal3f(n3.x, n3.y, n3.z);
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
								 ((float)i/(float)NumSections),
								 1.0f);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
								 ((float)i/(float)NumSections),
								 1.0f);
			glVertex3f(tf3.x, tf3.y, tf3.z);


			glNormal3f(n4.x, n4.y, n4.z);
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
								 (((float)i+1.0f)/(float)NumSections),
								 0.0f);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
								 (((float)i+1.0f)/(float)NumSections),
								 0.0f);
			glVertex3f(tf4.x, tf4.y, tf4.z);

			glNormal3f(n3.x, n3.y, n3.z);
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
								 (float)i/(float)NumSections,
								 1.0f);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
								 (float)i/(float)NumSections,
								 1.0f);
			glVertex3f(tf5.x, tf5.y, tf5.z);

			glNormal3f(n4.x, n4.y, n4.z);
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
								 (((float)i+1.0f)/(float)NumSections),
								 1.0f);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
								 (((float)i+1.0f)/(float)NumSections),
								 1.0f);
			glVertex3f(tf6.x, tf6.y, tf6.z);
		}
		glEnd();
	}
	glEndList();

	//Generate the top display list.
	topDisplay = glGenLists(1);
	glNewList(topDisplay, GL_COMPILE);
	{
		glBegin(GL_TRIANGLES);
		{
			for(i=0;i<(NumSections-1);++i)
			{
				tf1.x = (xVerts[i+1]*radius);
				tf1.y = (height*0.5f);
				tf1.z = (yVerts[i+1]*radius);
				tf2.x = 0.0f;
				tf2.y = (height*0.5f);
				tf2.z = 0.0f;
				tf3.x = (xVerts[i]*radius);
				tf3.y = (height*0.5f);
				tf3.z = (yVerts[i]*radius);

				calcNormal(tf1, tf2, tf3);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
									 (xVerts[i+1]*0.5f)+0.5f,
									 (yVerts[i+1]*0.5f)+0.5f);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
									 (xVerts[i+1]*0.5f)+0.5f,
									 (yVerts[i+1]*0.5f)+0.5f);
				glVertex3f(tf1.x, tf1.y, tf1.z);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.5f, 0.5f);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0.5f, 0.5f);
				glVertex3f(tf2.x, tf2.y, tf2.z);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
									 (xVerts[i]*0.5f)+0.5f,
									 (yVerts[i]*0.5f)+0.5f);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
									 (xVerts[i]*0.5f)+0.5f,
									 (yVerts[i]*0.5f)+0.5f);
				glVertex3f(tf3.x, tf3.y, tf3.z);
			}
			tf1.x = (xVerts[0]*radius);
			tf1.y = (height*0.5f);
			tf1.z = (yVerts[0]*radius);
			tf2.x = 0.0f;
			tf2.y = (height*0.5f);
			tf2.z = 0.0f;
			tf3.x = (xVerts[i]*radius);
			tf3.y = (height*0.5f);
			tf3.z = (yVerts[i]*radius);

			calcNormal(tf1, tf2, tf3);
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
								 (xVerts[0]*0.5f)+0.5f,
								 (yVerts[0]*0.5f)+0.5f);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
								 (xVerts[0]*0.5f)+0.5f,
								 (yVerts[0]*0.5f)+0.5f);
			glVertex3f(tf1.x, tf1.y, tf1.z);
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.5f, 0.5f);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0.5f, 0.5f);
			glVertex3f(tf2.x, tf2.y, tf2.z);
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
								 (xVerts[i]*0.5f)+0.5f,
								 (yVerts[i]*0.5f)+0.5f);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
								 (xVerts[i]*0.5f)+0.5f,
								 (yVerts[i]*0.5f)+0.5f);
			glVertex3f(tf3.x, tf3.y, tf3.z);
		}
		glEnd();
	}
	glEndList();

	//Generate the bottom display list.
	bottomDisplay = glGenLists(1);
	glNewList(bottomDisplay, GL_COMPILE);
	{
		glBegin(GL_TRIANGLES);
		{
			for(i=0;i<(NumSections-1);++i)
			{
				tf1.x = (xVerts[i+1]*radius);
				tf1.y = -(height*0.5f);
				tf1.z = (yVerts[i+1]*radius);
				tf2.x = 0.0f;
				tf2.y = -(height*0.5f);
				tf2.z = 0.0f;
				tf3.x = (xVerts[i]*radius);
				tf3.y = -(height*0.5f);
				tf3.z = (yVerts[i]*radius);

				calcNormal(tf3, tf2, tf1);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
									 (xVerts[i+1]*0.5f)+0.5f,
									 (yVerts[i+1]*0.5f)+0.5f);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
									 (xVerts[i+1]*0.5f)+0.5f,
									 (yVerts[i+1]*0.5f)+0.5f);
				glVertex3f(tf1.x, tf1.y, tf1.z);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.5f, 0.5f);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0.5f, 0.5f);
				glVertex3f(tf2.x, tf2.y, tf2.z);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
									 (xVerts[i]*0.5f)+0.5f,
									 (yVerts[i]*0.5f)+0.5f);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
									 (xVerts[i]*0.5f)+0.5f,
									 (yVerts[i]*0.5f)+0.5f);
				glVertex3f(tf3.x, tf3.y, tf3.z);
			}
			tf1.x = (xVerts[0]*radius);
			tf1.y = -(height*0.5f);
			tf1.z = (yVerts[0]*radius);
			tf2.x = 0.0f;
			tf2.y = -(height*0.5f);
			tf2.z = 0.0f;
			tf3.x = (xVerts[i]*radius);
			tf3.y = -(height*0.5f);
			tf3.z = (yVerts[i]*radius);

			calcNormal(tf3, tf2, tf1);
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
								 (xVerts[0]*0.5f)+0.5f,
								 (yVerts[0]*0.5f)+0.5f);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
								 (xVerts[0]*0.5f)+0.5f,
								 (yVerts[0]*0.5f)+0.5f);
			glVertex3f(tf1.x, tf1.y, tf1.z);
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.5f, 0.5f);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0.5f, 0.5f);
			glVertex3f(tf2.x, tf2.y, tf2.z);
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
								 (xVerts[i]*0.5f)+0.5f,
								 (yVerts[i]*0.5f)+0.5f);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
								 (xVerts[i]*0.5f)+0.5f,
								 (yVerts[i]*0.5f)+0.5f);
			glVertex3f(tf3.x, tf3.y, tf3.z);
		}
		glEnd();
	}
	glEndList();
}

//------------------------------------------------------------------------------
void Cylinder::graphicDoStuff()
{
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glScalef(radius, (height*0.5f), radius);

	if(side)
		glCallList(sidesDisplay);
	if(bottom)
		glCallList(bottomDisplay);
	if(top)
		glCallList(topDisplay);

	glPopMatrix();
}

//------------------------------------------------------------------------------
void Cylinder::setAttribute(const string& attribute, EventValue& val)
{
	StringParser parser(val);

	if(attribute == "bottom")
	{
		if(parser.getNumBools() > 0)
			bottom = parser.getBool();
	}
	else if(attribute == "radius")
		radius = val;
	else if(attribute == "height")
		height = val;
	else if(attribute == "side")
	{
		if(parser.getNumBools() > 0)
			side = parser.getBool();
	}
	else if(attribute == "solid")
	{
		if(parser.getNumBools() > 0)
			solid = parser.getBool();
	}
	else if(attribute == "top")
	{
		if(parser.getNumBools() > 0)
			top = parser.getBool();
	}
}

}
