//	Cylinder.h - Class representing an X3D cylinder node.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef CYLINDER_H_
#define CYLINDER_H_

#include "AbstractNode.h"

#include "GLHeaders.h"
#include <cmath>

namespace HeilanX3D
{

///	Class representing an X3D cylinder node.
class Cylinder : public ChildNode
{
  public:
	///	Constructor.
	Cylinder();
	///	Destructor.
	~Cylinder();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the Cone.
	void graphicDoStuff();
	///	Not used, overridden to do nothing.
	/*AmbisonicData *audioDoStuff(int numSamples,
								const ThreeFloats& pos) {return 0;};*/

	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set the cone's parameters.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<Cylinder> constructor;
  private:
	///	Helper method to calculate the normal of a polygon.
	inline void calcNormal(const ThreeFloats& v1,
						   const ThreeFloats& v2,
						   const ThreeFloats& v3)
	{
		ThreeFloats a;
		ThreeFloats b;
		ThreeFloats normal;
		float length;

		a.x = v1.x - v2.x;
		a.y = v1.y - v2.y;
		a.z = v1.z - v2.z;

		b.x = v2.x - v3.x;
		b.y = v2.y - v3.y;
		b.z = v2.z - v3.z;

		normal.x = (a.y * b.z) - (a.z * b.y);
		normal.y = (a.z * b.x) - (a.x * b.z);
		normal.z = (a.x * b.y) - (a.y * b.x);

		length = sqrtf((normal.x*normal.x)+(normal.y*normal.y)+(normal.z*normal.z));
		if(length == 0.0f)
			length = 1.0f;
		else
			length = 1.0f/length;

		glNormal3f((normal.x*length), (normal.y*length), (normal.z*length));
	};
	///	Second version, to return the normal instead of drawing it.
	inline ThreeFloats calcNormal2(const ThreeFloats& v1,
								   const ThreeFloats& v2,
								   const ThreeFloats& v3)
	{
		ThreeFloats a;
		ThreeFloats b;
		ThreeFloats normal;
		float length;

		a.x = v1.x - v2.x;
		a.y = v1.y - v2.y;
		a.z = v1.z - v2.z;

		b.x = v2.x - v3.x;
		b.y = v2.y - v3.y;
		b.z = v2.z - v3.z;

		normal.x = (a.y * b.z) - (a.z * b.y);
		normal.y = (a.z * b.x) - (a.x * b.z);
		normal.z = (a.x * b.y) - (a.y * b.x);

		length = sqrtf((normal.x*normal.x)+(normal.y*normal.y)+(normal.z*normal.z));
		if(length == 0.0f)
			length = 1.0f;
		else
			length = 1.0f/length;

		normal.x *= length;
		normal.y *= length;
		normal.z *= length;

		return normal;
	};
	///	To calculate smooth normals.
	inline ThreeFloats calcSmooth(ThreeFloats& n1, ThreeFloats& n2)
	{
		ThreeFloats retval;

		retval.x = (n1.x + n2.x)*0.5f;
		retval.y = (n1.y + n2.y)*0.5f;
		retval.z = (n1.z + n2.z)*0.5f;

		return retval;
	};

	///	Whether the bottom of the cylinder should be drawn.
	bool bottom;
	///	The radius of the cylinder.
	float radius;
	///	The height of the cylinder.
	float height;
	///	Whether the sides of the cylinder should be drawn.
	bool side;
	///	Whether the cylinder is visible from inside it.
	bool solid;
	///	Whether the top of the cylinder should be drawn.
	bool top;

	///	Enum containing the number of sides we use to draw the cylinder.
	enum
	{
		NumSections = 64
	};

	///	Storage for the sides display list.
	/*!
		So we don't have to do so much in graphicDoStuff().
	 */
	GLuint sidesDisplay;
	///	Storage for the top display list.
	/*!
		So we don't have to do so much in graphicDoStuff().
	 */
	GLuint topDisplay;
	///	Storage for the bottom display list.
	/*!
		So we don't have to do so much in graphicDoStuff().
	 */
	GLuint bottomDisplay;
};

}

#endif
