//	DirectionalLight.h - Class representing an X3D box node.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef DIRECTIONALLIGHT_H_
#define DIRECTIONALLIGHT_H_

#include "AbstractNode.h"

namespace HeilanX3D
{

///	Class representing an X3D directional light.
/*!
	How we handle DirectionalLights in Heilan Browser is: OpenGL is basically
	limited to 8 lights (it does allow for more, but 8 is the minimum, and most
	implementations only implement 8 lights).  We use one of these as the
	headlight, so we only have 7 to play with.  Because of this we cheat a bit
	to allow unlimited (under certain conditions) light sources.  Basically,
	because a DirectionalLight is only supposed to apply to nodes below it in
	the XML heirarchy, we disable it when we leave its level in the heirarchy.
	i.e. the following sequence is how a simple scene would be drawn:

	enter Transform, apply any translations etc. -> enable light -> draw Box ->
	disable light -> leave Transform, disable any translations etc.

	This means that one branch of the heirarchy can have up to 7 lights, but the
	number of branches which simultaneously can have 7 lights is basically
	unlimited.

	\todo Switching global to true will only work 100% correctly when it's the
	last attribute in the list.
 */
class DirectionalLight : public ChildNode
{
  public:
	///	Constructor.
	DirectionalLight();
	///	Destructor.
	~DirectionalLight();

	///	Draws the DirectionalLight.
	void graphicDoStuff();
	///	Not used, overridden to do nothing.
	/*AmbisonicData *audioDoStuff(int numSamples,
								const ThreeFloats& pos) {return 0;};*/

	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set the attributes of the light.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<DirectionalLight> constructor;
  private:
	///	The ambient intensity of the light.
	float ambientIntensity;
	///	The colour of the light.
	ThreeFloats color;
	///	The direction of the light.
	ThreeFloats direction;
	///	Whether the light applies globally or not.
	bool global;
	///	The intensity of the light.
	float intensity;
	///	Whether the light is on or not.
	bool on;
};

}

#endif
