//	ElevationGrid.cpp - Class representing an X3D elevation grid.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "ElevationGrid.h"
#include "AttributeHelper.h"
#include "StringParser.h"
#include "MultiTextureCoordinate.h"
#include "GLHeaders.h"
#include "MiscHelper.h"

#include "GLHeaders.h"
#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<ElevationGrid> ElevationGrid::constructor;

//------------------------------------------------------------------------------
ElevationGrid::ElevationGrid():
ParentNode(),
ccw(true),
colorPerVertex(true),
creaseAngle(0.0f),
normalPerVertex(true),
solid(true),
xDimension(0),
xSpacing(1.0f),
zDimension(0),
zSpacing(1.0f),
colors(0),
colorsRGBA(0),
normals(0)
{
	setType("ElevationGrid");
	setVisual(true);

	textureCoords[0] = 0;
	textureCoords[1] = 0;
}

//------------------------------------------------------------------------------
ElevationGrid::~ElevationGrid()
{
	
}

//------------------------------------------------------------------------------
void ElevationGrid::setAttribute(const string& attribute, EventValue& val)
{
	int i;
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "ccw")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			ccw = parser.getBool();
	}
	else if(realAtt == "colorPerVertex")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			colorPerVertex = parser.getBool();
	}
	else if(realAtt == "creaseAngle")
		creaseAngle = val;
	else if((realAtt == "set_height") || (realAtt == "height"))
	{
		StringParser parser(val);
		AttributeHelper< vector<float> > helper(&height);

		critSec.lock();
		{
			if(helper.isValid(attribute))
				helper.convert(attribute, val);
			else
			{
				if(height.size() > 0)
					height.clear();

				for(i=0;i<parser.getNumNumbers();++i)
					height.push_back(parser.getFloat(i));
			}
		}
		critSec.unlock();
	}
	else if(realAtt == "normalPerVertex")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			normalPerVertex = parser.getBool();
	}
	else if(realAtt == "solid")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			solid = parser.getBool();
	}
	else if(realAtt == "xDimension")
		xDimension = val;
	else if(realAtt == "xSpacing")
		xSpacing = val;
	else if(realAtt == "zDimension")
		zDimension = val;
	else if(realAtt == "zSpacing")
		zSpacing = val;
}

//------------------------------------------------------------------------------
void ElevationGrid::postChildDraw()
{
	unsigned int col, nor;
	int tempint;
	int x, z;

	float tempX, tempY, tempZ;

	bool colOkay, norOkay, texOkay;

	if(height.size() < (unsigned)(xDimension * zDimension))
		return;

	//If we're supposed to draw colours onto the object, disable lighting,
	//otherwise you won't see anything.
	if(!textureCoords[0] && (colors || colorsRGBA))
		glDisable(GL_LIGHTING);

	if(colorPerVertex && (drawingColours.size() >= (unsigned)(xDimension * zDimension)))
		colOkay = true;
	else if(!colorPerVertex &&
			(drawingColours.size() >= (unsigned)((xDimension-1) * (zDimension-1))))
		colOkay = true;
	else
		colOkay = false;

	if(normalPerVertex && (drawingNormals.size() >= (unsigned)(xDimension * zDimension)))
		norOkay = true;
	else if(!normalPerVertex &&
			(drawingNormals.size() >= (unsigned)((xDimension-1) * (zDimension-1))))
		norOkay = true;
	else
		norOkay = false;

	if(drawingTex[0].size() >= (unsigned)(xDimension * zDimension))
		texOkay = true;
	else
		texOkay = false;

	glBegin(GL_QUADS);
	{
		col = 0;
		nor = 0;
		for(z=0;z<(zDimension-1);++z)
		{
			for(x=0;x<(xDimension-1);++x)
			{
				//Set up per face variables.
				if((!colorPerVertex) && colOkay)
				{
					glColor4f(drawingColours[col].x,
							  drawingColours[col].y,
							  drawingColours[col].z,
							  drawingColours[col].r);
					++col;
				}
				if((!normalPerVertex) && norOkay)
				{
					tempX = drawingNormals[nor].x;
					tempY = drawingNormals[nor].y;
					tempZ = drawingNormals[nor].z;
					glNormal3f(drawingNormals[nor].x,
							   drawingNormals[nor].y,
							   drawingNormals[nor].z);
					++nor;
				}

				//Vertex (x, z).
				tempint = x + (z * xDimension);
				if(colorPerVertex && colOkay)
				{
					glColor4f(drawingColours[tempint].x,
							  drawingColours[tempint].y,
							  drawingColours[tempint].z,
							  drawingColours[tempint].r);
				}
				if(normalPerVertex && norOkay)
				{
					glNormal3f(drawingNormals[tempint].x,
							   drawingNormals[tempint].y,
							   drawingNormals[tempint].z);
				}
				if(texOkay)
				{
					glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
										 drawingTex[0][tempint].x,
										 drawingTex[0][tempint].y);
					if(drawingTex[1].size() > 0)
					{
						glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
											 drawingTex[1][tempint].x,
											 drawingTex[1][tempint].y);
					}
				}
				glVertex3f(xSpacing * (float)x,
						   height[tempint],
						   zSpacing * (float)z);

				//Vertex (x+1, z).
				++tempint;
				if(colorPerVertex && colOkay)
				{
					glColor4f(drawingColours[tempint].x,
							  drawingColours[tempint].y,
							  drawingColours[tempint].z,
							  drawingColours[tempint].r);
				}
				if(normalPerVertex && norOkay)
				{
					glNormal3f(drawingNormals[tempint].x,
							   drawingNormals[tempint].y,
							   drawingNormals[tempint].z);
				}
				if(texOkay)
				{
					glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
										 drawingTex[0][tempint].x,
										 drawingTex[0][tempint].y);
					if(drawingTex[1].size() > 0)
					{
						glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
											 drawingTex[1][tempint].x,
											 drawingTex[1][tempint].y);
					}
				}
				glVertex3f(xSpacing * (float)(x+1),
						   height[tempint],
						   zSpacing * (float)z);

				//Vertex (x+1, z+1).
				tempint += xDimension;
				if(colorPerVertex && colOkay)
				{
					glColor4f(drawingColours[tempint].x,
							  drawingColours[tempint].y,
							  drawingColours[tempint].z,
							  drawingColours[tempint].r);
				}
				if(normalPerVertex && norOkay)
				{
					glNormal3f(drawingNormals[tempint].x,
							   drawingNormals[tempint].y,
							   drawingNormals[tempint].z);
				}
				if(texOkay)
				{
					glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
										 drawingTex[0][tempint].x,
										 drawingTex[0][tempint].y);
					if(drawingTex[1].size() > 0)
					{
						glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
											 drawingTex[1][tempint].x,
											 drawingTex[1][tempint].y);
					}
				}
				glVertex3f(xSpacing * (float)(x+1),
						   height[tempint],
						   zSpacing * (float)(z+1));

				//Vertex (x, z+1).
				--tempint;
				if(colorPerVertex && colOkay)
				{
					glColor4f(drawingColours[tempint].x,
							  drawingColours[tempint].y,
							  drawingColours[tempint].z,
							  drawingColours[tempint].r);
				}
				if(normalPerVertex && norOkay)
				{
					glNormal3f(drawingNormals[tempint].x,
							   drawingNormals[tempint].y,
							   drawingNormals[tempint].z);
				}
				if(texOkay)
				{
					glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
										 drawingTex[0][tempint].x,
										 drawingTex[0][tempint].y);
					if(drawingTex[1].size() > 0)
					{
						glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
											 drawingTex[1][tempint].x,
											 drawingTex[1][tempint].y);
					}
				}
				glVertex3f(xSpacing * (float)x,
						   height[tempint],
						   zSpacing * (float)(z+1));
			}
		}
	}
	glEnd();

	if(!textureCoords[0] && (colors || colorsRGBA))
		glEnable(GL_LIGHTING);
}

//------------------------------------------------------------------------------
void ElevationGrid::childrenLoaded()
{
	unsigned int i, j, k;

	for(i=0;i<(unsigned)getNumChildren();++i)
	{
		if(getChild(i)->getType() == "Normal")
		{
			normals = dynamic_cast<Normal *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = normals->getVectors();

			drawingNormals.clear();

			for(j=0;j<tempPoints.size();++j)
				drawingNormals.push_back(tempPoints[j]);
		}
		else if(getChild(i)->getType() == "TextureCoordinate")
		{
			textureCoords[0] = dynamic_cast<TextureCoordinate *>(getChild(i));

			const vector<TwoFloats>& tempPoints = textureCoords[0]->getCoords();

			drawingTex[0].clear();

			for(j=0;j<tempPoints.size();++j)
				drawingTex[0].push_back(tempPoints[j]);
		}
		else if(getChild(i)->getType() == "MultiTextureCoordinate")
		{
			MultiTextureCoordinate *tempCoord = 0;
			tempCoord = dynamic_cast<MultiTextureCoordinate *>(getChild(i));

			if(tempCoord)
			{
				for(k=0;k<(unsigned)tempCoord->getNumChildTexCoord();++k)
				{
					textureCoords[k] = tempCoord->getChildTexCoord(k);

					const vector<TwoFloats>& tempPoints = textureCoords[k]->getCoords();

					drawingTex[k].clear();

					for(j=0;j<tempPoints.size();++j)
						drawingTex[k].push_back(tempPoints[j]);
				}
			}
		}
		else if(getChild(i)->getType() == "Color")
		{
			colors = dynamic_cast<Color *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = colors->getColours();

			drawingColours.clear();

			for(j=0;j<tempPoints.size();++j)
			{
				FourFloats tempf(tempPoints[j].x,
								 tempPoints[j].y,
								 tempPoints[j].z,
								 1.0f);

				drawingColours.push_back(tempf);
			}
		}
		else if(getChild(i)->getType() == "ColorRGBA")
		{
			colorsRGBA = dynamic_cast<ColorRGBA *>(getChild(i));

			const vector<FourFloats>& tempPoints = colorsRGBA->getColours();

			drawingColours.clear();

			for(j=0;j<tempPoints.size();++j)
				drawingColours.push_back(tempPoints[j]);
		}
	}

	if(drawingNormals.size() == 0)
		autoCalcNormals();
	if(drawingTex[0].size() == 0)
		autoCalcTextureCoords();
}

//------------------------------------------------------------------------------
void ElevationGrid::autoCalcNormals()
{
	int x, z;
	int tempint;
	float tempf;
	ThreeFloats temp3f;
	ThreeFloats v1, v2, v3;
	vector<ThreeFloats> tempNormals;

	tempint = 0;
	for(z=0;z<(zDimension-1);++z)
	{
		for(x=0;x<(xDimension-1);++x)
		{
			//Because we're working with 4-vertex polys, we need to be careful
			//here when choosing which vertices to calculate the normal from,
			//or we could end up with the wrong value, hence this if statement.
			if((height[tempint] == height[tempint+1]) &&
			   (height[tempint]== height[tempint+xDimension]))
			{
				v1.x = (float)(x+1) * xSpacing;
				v1.y = height[tempint+xDimension+1];
				v1.z = (float)(z+1) * zSpacing;

				v2.x = (float)(x+1) * xSpacing;
				v2.y = height[tempint+1];
				v2.z = (float)z * zSpacing;

				v3.x = (float)x * xSpacing;
				v3.y = height[tempint+xDimension];
				v3.z = (float)(z+1) * zSpacing;
			}
			else
			{
				v1.x = (float)(x+1) * xSpacing;
				v1.y = height[tempint+1];
				v1.z = (float)z * zSpacing;

				v2.x = (float)x * xSpacing;
				v2.y = height[tempint];
				v2.z = (float)z * zSpacing;

				v3.x = (float)x * xSpacing;
				v3.y = height[tempint+xDimension];
				v3.z = (float)(z+1) * zSpacing;
			}

			drawingNormals.push_back(heilanGetNormal(v1, v2, v3));

			++tempint;
		}
		++tempint;
	}

	//For per-vertex normals, we take the average of the 4 neighbouring face
	//normals.
	if(normalPerVertex)
	{
		tempNormals = drawingNormals;
		drawingNormals.clear();

		tempint = 0;
		for(z=0;z<zDimension;++z)
		{
			if(z == 0)
			{
				for(x=0;x<xDimension;++x)
				{
					if(x == 0)
					{
						drawingNormals.push_back(tempNormals[tempint]);
						++tempint;
					}
					else if(x < (xDimension-1))
					{
						temp3f = tempNormals[tempint-1];
						temp3f += tempNormals[tempint];
						tempf = temp3f.getDistance(ThreeFloats(0.0f));
						if(tempf != 0.0f)
							temp3f /= tempf;
						drawingNormals.push_back(temp3f);
						++tempint;
					}
					else
						drawingNormals.push_back(tempNormals[tempint-1]);
				}
			}
			else if(z < (zDimension-1))
			{
				for(x=0;x<xDimension;++x)
				{
					if(x == 0)
					{
						temp3f = tempNormals[tempint-(xDimension-1)];
						temp3f += tempNormals[tempint];
						tempf = temp3f.getDistance(ThreeFloats(0.0f));
						if(tempf != 0.0f)
							temp3f /= tempf;
						drawingNormals.push_back(temp3f);
						++tempint;
					}
					else if(x < (xDimension-1))
					{
						temp3f = tempNormals[tempint-1];
						temp3f += tempNormals[tempint];
						temp3f += tempNormals[tempint-(xDimension-1)-1];
						temp3f += tempNormals[tempint-(xDimension-1)];
						tempf = temp3f.getDistance(ThreeFloats(0.0f));
						if(tempf != 0.0f)
							temp3f /= tempf;
						drawingNormals.push_back(temp3f);
						++tempint;
					}
					else
					{
						temp3f = tempNormals[tempint-(xDimension-1)];
						temp3f += tempNormals[tempint];
						tempf = temp3f.getDistance(ThreeFloats(0.0f));
						if(tempf != 0.0f)
							temp3f /= tempf;
						drawingNormals.push_back(temp3f);
					}
				}
			}
			else
			{
				for(x=0;x<xDimension;++x)
				{
					if(x == 0)
					{
						drawingNormals.push_back(tempNormals[tempint-(xDimension-1)]);
						++tempint;
					}
					else if(x < (xDimension-1))
					{
						temp3f = tempNormals[tempint-(xDimension-1)-1];
						temp3f += tempNormals[tempint-(xDimension-1)];
						tempf = temp3f.getDistance(ThreeFloats(0.0f));
						if(tempf != 0.0f)
							temp3f /= tempf;
						drawingNormals.push_back(temp3f);
						++tempint;
					}
					else
						drawingNormals.push_back(tempNormals[tempint-(xDimension-1)-1]);
				}
			}
		}
	}
}

//------------------------------------------------------------------------------
void ElevationGrid::autoCalcTextureCoords()
{
	int x, z;
	TwoFloats temp2f;

	for(z=0;z<zDimension;++z)
	{
		temp2f.y = (float)z/(float)zDimension;

		for(x=0;x<xDimension;++x)
		{
			temp2f.x = (float)x/(float)xDimension;
			drawingTex[0].push_back(temp2f);
			drawingTex[1].push_back(temp2f);
		}
	}
}
	
}
