//	ElevationGrid.h - Class representing an X3D elevation grid.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef ELEVATIONGRID_H_
#define ELEVATIONGRID_H_

#include "ParentNode.h"
#include "Normal.h"
#include "Color.h"
#include "ColorRGBA.h"
#include "TextureCoordinate.h"
#include "CriticalSection.h"

namespace HeilanX3D
{

///	Class representing an X3D elevation grid.
/*!
	\todo Implement the ccw attribute.
	\todo Implement support for automatically calculating normals ourselves.
 */
class ElevationGrid : public ParentNode
{
  public:
	///	Constructor.
	ElevationGrid();
	///	Destructor.
	~ElevationGrid();

	///	Sets the curve's attributes.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<ElevationGrid> constructor;
  private:
	///	Draws the landscape.
	void postChildDraw();

	///	Overridden to do nothing.
	void audioBlockSizeChanged(int newSize) {};

	///	Used to get the normals etc. of the grid from a child Normal etc. node.
	void childrenLoaded();

	///	Helper method to calculate normals automatically.
	void autoCalcNormals();
	///	Helper method to calculate texture coordinates automatically.
	void autoCalcTextureCoords();

	///	Whether the vertices are ordered counter-clockwise or not.
	bool ccw;
	///	Whether colour should be applied per vertex, or per face.
	bool colorPerVertex;
	///	At what angle normals should be smooth shaded/flat shaded.
	float creaseAngle;
	///	The heights of the vertices.
	std::vector<float> height;
	///	Whether normals should apply per vertex or per face.
	bool normalPerVertex;
	///	Whether both sides of the polygons should be visible.
	bool solid;
	///	Number of vertices along the x-axis.
	int xDimension;
	///	Distance between vertices along the x-axis.
	float xSpacing;
	///	Number of vertices along the z-axis.
	int zDimension;
	///	Distance between vertices along the z-axis.
	float zSpacing;

	///	Pointer to the Color node we may use to colour the object.
	Color *colors;
	///	Pointer to the ColorRGBA node may we use to colour the object.
	ColorRGBA *colorsRGBA;
	///	Pointer to a Normal node we may use to calculate the object's lighting.
	Normal *normals;
	///	Pointer to a TextureCoordinate node we may use to texture the object.
	/*!
		There are 2 of these so we can accommodate MultiTextureCoordinates.
	 */
	TextureCoordinate *textureCoords[2];

	///	A stored copy of colors' or colorsRGBA's colours used to draw the object.
	std::vector<FourFloats> drawingColours;
	///	A stored copy of normals' normals used to draw the object.
	std::vector<ThreeFloats> drawingNormals;
	///	A stored copy of textureCoords' coordinates used to draw the object.
	/*!
		There are 2 of these so we can accommodate MultiTextureCoordinates.
	 */
	std::vector<TwoFloats> drawingTex[2];

	///	Critical section used to guard coordinates while we're drawing.
	CriticalSection critSec;
};

}

#endif
