//	IndexedFaceSet.cpp - Class representing an X3D indexed face set.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "IndexedFaceSet.h"
#include "MultiTextureCoordinate.h"
#include "StringParser.h"
#include "AttributeHelper.h"

#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<IndexedFaceSet> IndexedFaceSet::constructor;

//------------------------------------------------------------------------------
IndexedFaceSet::IndexedFaceSet():
ParentNode(),
ccw(true),
colorPerVertex(true),
convex(true),
creaseAngle(0),
normalPerVertex(true),
solid(true),
coordinates(0),
normals(0),
colors(0),
colorsRGBA(0)
{
	setType("IndexedFaceSet");
	setVisual(true);

	textureCoords[0] = 0;
	textureCoords[1] = 0;
}

//------------------------------------------------------------------------------
IndexedFaceSet::~IndexedFaceSet()
{
	
}

//------------------------------------------------------------------------------
void IndexedFaceSet::setAttribute(const string& attribute, EventValue& val)
{
	int i;
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if((realAtt == "set_colorIndex") || (realAtt == "colorIndex"))
	{
		StringParser parser(val);
		AttributeHelper< vector<int> > helper(&colorIndex);

		critSec.lock();
		{
			if(helper.isValid(attribute))
				helper.convert(attribute, val);
			else
			{
				if(colorIndex.size() > 0)
					colorIndex.clear();
	
				for(i=0;i<parser.getNumNumbers();++i)
					colorIndex.push_back(parser.getInt(i));
			}
		}
		critSec.unlock();
	}
	else if((realAtt == "set_coordIndex") || (realAtt == "coordIndex"))
	{
		StringParser parser(val);
		AttributeHelper< vector<int> > helper(&coordIndex);

		critSec.lock();
		{
			if(helper.isValid(attribute))
				helper.convert(attribute, val);
			else
			{
				if(coordIndex.size() > 0)
					coordIndex.clear();
	
				for(i=0;i<parser.getNumNumbers();++i)
					coordIndex.push_back(parser.getInt(i));
			}
		}
		critSec.unlock();
	}
	else if((realAtt == "set_normalIndex") || (realAtt == "normalIndex"))
	{
		StringParser parser(val);
		AttributeHelper< vector<int> > helper(&normalIndex);

		critSec.lock();
		{
			if(helper.isValid(attribute))
				helper.convert(attribute, val);
			else
			{
				if(normalIndex.size() > 0)
					normalIndex.clear();
	
				for(i=0;i<parser.getNumNumbers();++i)
					normalIndex.push_back(parser.getInt(i));
			}
		}
		critSec.unlock();
	}
	else if((realAtt == "set_texCoordIndex") || (realAtt == "texCoordIndex"))
	{
		StringParser parser(val);
		AttributeHelper< vector<int> > helper(&texCoordIndex);

		critSec.lock();
		{
			if(helper.isValid(attribute))
				helper.convert(attribute, val);
			else
			{
				if(texCoordIndex.size() > 0)
					texCoordIndex.clear();
	
				for(i=0;i<parser.getNumNumbers();++i)
					texCoordIndex.push_back(parser.getInt(i));
			}
		}
		critSec.unlock();
	}
	else if(attribute == "ccw")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			ccw = parser.getBool();
	}
	else if(attribute == "colorPerVertex")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			colorPerVertex = parser.getBool();
	}
	else if(attribute == "convex")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			convex = parser.getBool();
	}
	else if(attribute == "creaseAngle")
		creaseAngle = val;
	else if(attribute == "normalPerVertex")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			normalPerVertex = parser.getBool();
	}
	else if(attribute == "solid")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			solid = parser.getBool();
	}
}

//------------------------------------------------------------------------------
void IndexedFaceSet::postChildDraw()
{
	unsigned int i = 0; //Coordinates iterator.
	unsigned int j = 0; //Normals iterator.
	unsigned int k = 0; //Textures iterator.
	unsigned int l = 0; //Colours iterator.

	if(coordinates)
	{
		//Re-load if necessary.
		if(coordinates->getHasChanged())
			childrenLoaded();

		//If we're supposed to draw colours onto the object, disable lighting,
		//otherwise you won't see anything.
		if(!textureCoords[0] && (colors || colorsRGBA))
		{
			glDisable(GL_LIGHTING);

			/*if(colorsRGBA)
			{
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				glEnable(GL_BLEND);
			}*/
		}

		while(i < coordIndex.size())
		{
			glBegin(GL_POLYGON);
			{
				//Handle normals if they're supposed to be per face.
				if(!normalPerVertex)
				{
					if(normals && (j < normalIndex.size()))
					{
						glNormal3f(drawingNormals[normalIndex[j]].x,
								   drawingNormals[normalIndex[j]].y,
								   drawingNormals[normalIndex[j]].z);
						++j;
						//This is a bit of a hack - things will look wrong if
						//there's a different number of faces and normals.
						if(normalIndex[j] == -1)
							++j;
					}
					else //Calculate normals ourselves.
					{
						
					}
				}

				//Handle colours if they're supposed to be per face.
				if((!colorPerVertex) &&
				   (!textureCoords[0]) &&
				   (colors || colorsRGBA))
				{
					glColor4f(drawingColours[colorIndex[l]].x,
							  drawingColours[colorIndex[l]].y,
							  drawingColours[colorIndex[l]].z,
							  drawingColours[colorIndex[l]].r);
					++l;
					//This is a bit of a hack - things will look wrong if
					//there's a different number of faces and colours.
					if(colorIndex[l] == -1)
						++l;
				}

				while((coordIndex[i] > -1) && (i < coordIndex.size()))
				{
					//Handle per vertex normals.
					if(normals && normalPerVertex && (j < normalIndex.size()))
					{
						glNormal3f(drawingNormals[normalIndex[j]].x,
								   drawingNormals[normalIndex[j]].y,
								   drawingNormals[normalIndex[j]].z);
						++j;
						//This is a bit of a hack - things will look wrong if
						//there's a different number of coordinates and normals.
						if(normalIndex[j] == -1)
							++j;
					}
					else //Calculate normals ourselves.
					{
						
					}

					//Handle (single) texturing.
					if(textureCoords[0] && (k < texCoordIndex.size()))
					{
						glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
											 drawingTex[0][texCoordIndex[k]].x,
									 		 drawingTex[0][texCoordIndex[k]].y);

						//Handle (multi) texturing.
						if(textureCoords[1] && (k < texCoordIndex.size()))
						{
							glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
												 drawingTex[1][texCoordIndex[k]].x,
												 drawingTex[1][texCoordIndex[k]].y);
						}
						++k;
						//This is a bit of a hack - things will look wrong if
						//there's a different number of coordinates and
						//texture coordinates.
						if(texCoordIndex[k] == -1)
							++k;
					}
					//Handle per vertex colouring.
					else if((colors || colorsRGBA) &&
							colorPerVertex &&
							(l < colorIndex.size()))
					{
						glColor4f(drawingColours[colorIndex[l]].x,
								  drawingColours[colorIndex[l]].y,
								  drawingColours[colorIndex[l]].z,
								  drawingColours[colorIndex[l]].r);
						++l;
						//This is a bit of a hack - things will look wrong if
						//there's a different number of coordinates and colours.
						if(colorIndex[l] == -1)
							++l;
					}

					//Draw the vertex.
					glVertex3f(drawingCoords[coordIndex[i]].x,
						   	   drawingCoords[coordIndex[i]].y,
						   	   drawingCoords[coordIndex[i]].z);
					++i;
				}
				++i;
			}
			glEnd();
		}

		if(!textureCoords[0] && (colors || colorsRGBA))
		{
			glEnable(GL_LIGHTING);

			/*if(colorsRGBA)
				glDisable(GL_BLEND);*/
		}
	}
}

//------------------------------------------------------------------------------
void IndexedFaceSet::childrenLoaded()
{
	unsigned int i, j, k;

	for(i=0;i<(unsigned)getNumChildren();++i)
	{
		if(getChild(i)->getType() == "Coordinate")
		{
			coordinates = dynamic_cast<Coordinate *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = coordinates->getPoints();

			drawingCoords.clear();

			for(j=0;j<tempPoints.size();++j)
				drawingCoords.push_back(tempPoints[j]);
		}
		else if(getChild(i)->getType() == "Normal")
		{
			normals = dynamic_cast<Normal *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = normals->getVectors();

			drawingNormals.clear();

			for(j=0;j<tempPoints.size();++j)
				drawingNormals.push_back(tempPoints[j]);
		}
		else if(getChild(i)->getType() == "TextureCoordinate")
		{
			textureCoords[0] = dynamic_cast<TextureCoordinate *>(getChild(i));

			const vector<TwoFloats>& tempPoints = textureCoords[0]->getCoords();

			drawingTex[0].clear();

			for(j=0;j<tempPoints.size();++j)
				drawingTex[0].push_back(tempPoints[j]);
		}
		else if(getChild(i)->getType() == "MultiTextureCoordinate")
		{
			MultiTextureCoordinate *tempCoord = 0;
			tempCoord = dynamic_cast<MultiTextureCoordinate *>(getChild(i));

			if(tempCoord)
			{
				for(k=0;k<(unsigned)tempCoord->getNumChildTexCoord();++k)
				{
					textureCoords[k] = tempCoord->getChildTexCoord(k);

					const vector<TwoFloats>& tempPoints = textureCoords[k]->getCoords();

					drawingTex[k].clear();

					for(j=0;j<tempPoints.size();++j)
						drawingTex[k].push_back(tempPoints[j]);
				}
			}
		}
		else if(getChild(i)->getType() == "Color")
		{
			colors = dynamic_cast<Color *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = colors->getColours();

			drawingColours.clear();

			for(j=0;j<tempPoints.size();++j)
			{
				FourFloats tempf(tempPoints[j].x,
								 tempPoints[j].y,
								 tempPoints[j].z,
								 1.0f);

				drawingColours.push_back(tempf);
			}
		}
		else if(getChild(i)->getType() == "ColorRGBA")
		{
			colorsRGBA = dynamic_cast<ColorRGBA *>(getChild(i));

			const vector<FourFloats>& tempPoints = colorsRGBA->getColours();

			drawingColours.clear();

			for(j=0;j<tempPoints.size();++j)
				drawingColours.push_back(tempPoints[j]);
		}
	}
}

}
