//	IndexedLineSet.cpp - Class representing an X3D indexed line set.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "IndexedLineSet.h"
#include "StringParser.h"
#include "AttributeHelper.h"

using namespace std;

namespace HeilanX3D
{

NodeConstructor<IndexedLineSet> IndexedLineSet::constructor;

//------------------------------------------------------------------------------
IndexedLineSet::IndexedLineSet():
ParentNode(),
colorPerVertex(true),
coordinates(0),
colors(0),
colorsRGBA(0)
{
	setType("IndexedLineSet");
	setVisual(true);
}

//------------------------------------------------------------------------------
IndexedLineSet::~IndexedLineSet()
{
	
}

//------------------------------------------------------------------------------
void IndexedLineSet::setAttribute(const string& attribute, EventValue& val)
{
	int i;
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if((realAtt == "set_coordIndex") || (realAtt == "coordIndex"))
	{
		StringParser parser(val);
		AttributeHelper< vector<int> > helper(&coordIndex);

		critSec.lock();
		{
			if(helper.isValid(attribute))
				helper.convert(attribute, val);
			else
			{
				if(coordIndex.size() > 0)
					coordIndex.clear();
	
				for(i=0;i<parser.getNumNumbers();++i)
					coordIndex.push_back(parser.getInt(i));
			}
		}
		critSec.unlock();
	}
	else if((realAtt == "set_colorIndex") || (realAtt == "colorIndex"))
	{
		StringParser parser(val);
		AttributeHelper< vector<int> > helper(&colorIndex);

		critSec.lock();
		{
			if(helper.isValid(attribute))
				helper.convert(attribute, val);
			else
			{
				if(colorIndex.size() > 0)
					colorIndex.clear();
	
				for(i=0;i<parser.getNumNumbers();++i)
					colorIndex.push_back(parser.getInt(i));
			}
		}
		critSec.unlock();
	}
	else if(attribute == "colorPerVertex")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			colorPerVertex = parser.getBool();
	}
}

//------------------------------------------------------------------------------
void IndexedLineSet::postChildDraw()
{
	unsigned int i = 0; //Coordinates iterator.
	unsigned int j = 0; //Colours iterator.

	if(coordinates)
	{
		//Re-load if necessary.
		if(coordinates->getHasChanged())
			childrenLoaded();

		//IndexedLineSets are unlit.
		glDisable(GL_LIGHTING);

		while(i < coordIndex.size())
		{
			glBegin(GL_LINE_STRIP);
			{
				//Handle colours if they're supposed to be per line.
				if((!colorPerVertex) && (colors || colorsRGBA))
				{
					glColor4f(drawingColours[colorIndex[j]].x,
							  drawingColours[colorIndex[j]].y,
							  drawingColours[colorIndex[j]].z,
							  drawingColours[colorIndex[j]].r);
					++j;
					//This is a bit of a hack - things will look wrong if
					//there's a different number of faces and colours.
					if(colorIndex[j] == -1)
						++j;
				}

				while((coordIndex[i] > -1) && (i < coordIndex.size()))
				{
					//Handle per vertex colouring.
					if((colors || colorsRGBA) &&
					   colorPerVertex &&
					   (j < colorIndex.size()))
					{
						glColor4f(drawingColours[colorIndex[j]].x,
								  drawingColours[colorIndex[j]].y,
								  drawingColours[colorIndex[j]].z,
								  drawingColours[colorIndex[j]].r);
						++j;
						//This is a bit of a hack - things will look wrong if
						//there's a different number of coordinates and colours.
						if(colorIndex[j] == -1)
							++j;
					}

					//Draw the vertex.
					glVertex3f(drawingCoords[coordIndex[i]].x,
						   	   drawingCoords[coordIndex[i]].y,
						   	   drawingCoords[coordIndex[i]].z);
					++i;
				}
				++i;
			}
			glEnd();
		}

		glEnable(GL_LIGHTING);
	}
}

//------------------------------------------------------------------------------
void IndexedLineSet::childrenLoaded()
{
	unsigned int i, j;

	for(i=0;i<(unsigned)getNumChildren();++i)
	{
		if(getChild(i)->getType() == "Coordinate")
		{
			coordinates = dynamic_cast<Coordinate *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = coordinates->getPoints();

			drawingCoords.clear();

			for(j=0;j<tempPoints.size();++j)
				drawingCoords.push_back(tempPoints[j]);
		}
		else if(getChild(i)->getType() == "Color")
		{
			colors = dynamic_cast<Color *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = colors->getColours();

			drawingColours.clear();

			for(j=0;j<tempPoints.size();++j)
			{
				FourFloats tempf(tempPoints[j].x,
								 tempPoints[j].y,
								 tempPoints[j].z,
								 1.0f);

				drawingColours.push_back(tempf);
			}
		}
		else if(getChild(i)->getType() == "ColorRGBA")
		{
			colorsRGBA = dynamic_cast<ColorRGBA *>(getChild(i));

			const vector<FourFloats>& tempPoints = colorsRGBA->getColours();

			drawingColours.clear();

			for(j=0;j<tempPoints.size();++j)
				drawingColours.push_back(tempPoints[j]);
		}
	}
}

}
