//	IndexedTriangleFanSet.h - Class representing an X3D indexed triangle fan set.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef INDEXEDTRIANGLEFANSET_H_
#define INDEXEDTRIANGLEFANSET_H_

#include "ParentNode.h"
#include "Coordinate.h"
#include "Normal.h"
#include "Color.h"
#include "ColorRGBA.h"
#include "TextureCoordinate.h"
#include "CriticalSection.h"

#include "GLHeaders.h"

namespace HeilanX3D
{

///	Class representing an X3D indexed triangle fan set.
/*!
	\todo Implement the ccw attribute.
	\todo Implement support for MultiTextures.
	\todo Implement support for automatically calculating normals ourselves.
 */
class IndexedTriangleFanSet : public ParentNode
{
  public:
	///	Constructor.
	IndexedTriangleFanSet();
	///	Destructor.
	~IndexedTriangleFanSet();

	///	Sets the curve's attributes.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<IndexedTriangleFanSet> constructor;
  private:
	///	Draws the curve.
	void postChildDraw();

	///	Overridden to do nothing.
	void audioBlockSizeChanged(int newSize) {};

	///	Used to get the coordinates of the set from a child Coordinate node.
	void childrenLoaded();

	///	Whether the vertices are ordered counter-clockwise or not.
	bool ccw;
	///	Whether colour should be applied per vertex, or per face.
	bool colorPerVertex;
	///	Whether normals should apply per vertex or per face.
	bool normalPerVertex;
	///	Whether both sides of the polygons should be visible.
	bool solid;
	///	The indices of the coordinates etc. used to draw the object.
	std::vector<int> index;

	///	Pointer to the Coordinate node we use to draw the object.
	Coordinate *coordinates;
	///	Pointer to a Normal node we may use to calculate the object's lighting.
	Normal *normals;
	///	Pointer to a TextureCoordinate node we may use to texture the object.
	TextureCoordinate *textureCoords[2];
	///	Pointer to the Color node we may use to colour the object.
	Color *colors;
	///	Pointer to the ColorRGBA node may we use to colour the object.
	ColorRGBA *colorsRGBA;

	///	A stored copy of coordinates' coords used to draw the object.
	std::vector<ThreeFloats> drawingCoords;
	///	A stored copy of normals' normals used to draw the object.
	std::vector<ThreeFloats> drawingNormals;
	///	A stored copy of textureCoords' coordinates used to draw the object.
	std::vector<TwoFloats> drawingTex[2];
	///	A stored copy of colors' or colorsRGBA's colours used to draw the object.
	std::vector<FourFloats> drawingColours;

	///	Critical section used to guard coordinates while we're drawing.
	CriticalSection critSec;
};

}

#endif
