//	LineSet.cpp - Class representing an X3D line set.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "LineSet.h"
#include "StringParser.h"
#include "AttributeHelper.h"

using namespace std;

namespace HeilanX3D
{

NodeConstructor<LineSet> LineSet::constructor;

//------------------------------------------------------------------------------
LineSet::LineSet():
ParentNode(),
coordinates(0),
colors(0),
colorsRGBA(0)
{
	setType("LineSet");
	setVisual(true);
}

//------------------------------------------------------------------------------
LineSet::~LineSet()
{
	
}

//------------------------------------------------------------------------------
void LineSet::setAttribute(const string& attribute, EventValue& val)
{
	int i;
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "vertexCount")
	{
		StringParser parser(val);
		AttributeHelper< vector<int> > helper(&vertexCount);

		critSec.lock();
		{
			if(helper.isValid(attribute))
				helper.convert(attribute, val);
			else
			{
				if(vertexCount.size() > 0)
					vertexCount.clear();

				for(i=0;i<parser.getNumNumbers();++i)
					vertexCount.push_back(parser.getInt(i));
			}
		}
		critSec.unlock();
	}
}

//------------------------------------------------------------------------------
void LineSet::postChildDraw()
{
	unsigned int i = 0; //Coordinates iterator.
	unsigned int j = 0; //vertexCount iterator.
	int k;

	if(coordinates)
	{
		//Re-load if necessary.
		if(coordinates->getHasChanged())
			childrenLoaded();

		//LineSets are unlit.
		glDisable(GL_LIGHTING);

		while((i < drawingCoords.size()) && (j <vertexCount.size()))
		{
			glBegin(GL_LINE_STRIP);
			{
				for(k=0;((k<vertexCount[j])&&(i<drawingCoords.size()));++k,++i)
				{
					if(drawingColours.size()>0)
					{
						glColor4f(drawingColours[i].x,
								  drawingColours[i].y,
								  drawingColours[i].z,
								  drawingColours[i].r);
					}

					//Draw the vertex.
					glVertex3f(drawingCoords[i].x,
						   	   drawingCoords[i].y,
						   	   drawingCoords[i].z);
				}
				++j;
			}
			glEnd();
		}

		glEnable(GL_LIGHTING);
	}
}

//------------------------------------------------------------------------------
void LineSet::childrenLoaded()
{
	unsigned int i, j;

	for(i=0;i<(unsigned)getNumChildren();++i)
	{
		if(getChild(i)->getType() == "Coordinate")
		{
			coordinates = dynamic_cast<Coordinate *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = coordinates->getPoints();

			drawingCoords.clear();

			for(j=0;j<tempPoints.size();++j)
				drawingCoords.push_back(tempPoints[j]);
		}
		else if(getChild(i)->getType() == "Color")
		{
			colors = dynamic_cast<Color *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = colors->getColours();

			drawingColours.clear();

			for(j=0;j<tempPoints.size();++j)
			{
				FourFloats tempf(tempPoints[j].x,
								 tempPoints[j].y,
								 tempPoints[j].z,
								 1.0f);

				drawingColours.push_back(tempf);
			}
		}
		else if(getChild(i)->getType() == "ColorRGBA")
		{
			colorsRGBA = dynamic_cast<ColorRGBA *>(getChild(i));

			const vector<FourFloats>& tempPoints = colorsRGBA->getColours();

			drawingColours.clear();

			for(j=0;j<tempPoints.size();++j)
				drawingColours.push_back(tempPoints[j]);
		}
	}
}

}
