//	MovieTexture.cpp - Class representing a movie texture.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "MovieTexture.h"
#include "StringParser.h"
#include "FileFinder.h"

#include <iostream>
#include <sstream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<MovieTexture> MovieTexture::constructor;

//------------------------------------------------------------------------------
MovieTexture::MovieTexture():
ChildNode(),
ThreadObject("MovieTexture"),
repeatS(true),
repeatT(true)
{
	setType("MovieTexture");
	setVisual(true);

#ifdef HAVE_FFMPEG
	formatContext = 0;
	videoStream = -1;
	buffer = 0;

	videoQueueRead = 0;
	videoQueueWrite = 0;
#endif
}

//------------------------------------------------------------------------------
MovieTexture::~MovieTexture()
{
#ifdef HAVE_FFMPEG
	if(formatContext)
		unloadTexture();
#endif
}

//------------------------------------------------------------------------------
void MovieTexture::initialise()
{
	glGenTextures(1, &texture);
	ThreadManager::getInstance().initialiseThread(this, 30);//****CHANGE 30!****
}

//------------------------------------------------------------------------------
void MovieTexture::graphicDoStuff()
{
#ifdef HAVE_FFMPEG
	GLenum err;

	if(formatContext)
	{
		//Get the next frame.
		//getNextFrame();

		//If there's a frame waiting, convert it to RGB.
		if(videoQueueRead != videoQueueWrite)
		{
			sws_scale(convertContext,
					  videoFrameQueue[videoQueueRead]->data,
					  videoFrameQueue[videoQueueRead]->linesize,
					  0,
					  codecContext->height,
					  rgbFrame->data,
					  rgbFrame->linesize);

			queueCritSec.lock();
			{
				++videoQueueRead;
				if(videoQueueRead >= VideoFrameQueueSize)
					videoQueueRead = 0;
			}
			queueCritSec.unlock();
		}

		//Copy the contents of the decoded buffer to our texture with
		//glTexSubImage2D().
		glBindTexture(GL_TEXTURE_2D, texture);
		glTexSubImage2D(GL_TEXTURE_2D,
						0,
						0,
						0,
						codecContext->width,
						codecContext->height,
						GL_RGB,
						GL_UNSIGNED_BYTE,
						buffer);
		err = glGetError();
		if(err != GL_NO_ERROR)
			cout << "MovieTexture: OpenGL error: " << gluErrorString(err) << endl;
	}
#endif
}

//------------------------------------------------------------------------------
void MovieTexture::threadIteration()
{
#ifdef HAVE_FFMPEG
	int err;
	int frameFinished;

	while(av_read_frame(formatContext, &packet) >= 0)
	{
		//There may be frames from streams other than the one we're interested
		//in.
		if(packet.stream_index == videoStream)
		{
			err = avcodec_decode_video(codecContext,
									   videoFrameQueue[videoQueueWrite],
									   &frameFinished,
									   packet.data,
									   packet.size);
			if(err < 0)
				ffmpegError(err);

			//Check we could decode a frame.
			if(frameFinished)
			{
				//Stick the frame on our queue.
				queueCritSec.lock();
				{
					++videoQueueWrite;
					if(videoQueueWrite >= VideoFrameQueueSize)
						videoQueueWrite = 0;
				}
				queueCritSec.unlock();

				//Free the packet.
				av_free_packet(&packet);
				break;
			}
		}

		//Free the packet.
		av_free_packet(&packet);
	}
#endif
}

//------------------------------------------------------------------------------
void MovieTexture::setAttribute(const string& attribute, EventValue& val)
{
	if(attribute == "url")
	{
		url = (string)val;
		loadTexture();
	}
	else if(attribute == "repeatS")
	{
		StringParser parse(val);

		if(parse.getNumBools() > 0)
		{
			repeatS = parse.getBool();

			if(repeatS)
			{
				glBindTexture(GL_TEXTURE_2D, texture);
				if(repeatS)
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
				else
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
				glBindTexture(GL_TEXTURE_2D, 0);
			}
		}
	}
	else if(attribute == "repeatT")
	{
		StringParser parse(val);

		if(parse.getNumBools() > 0)
		{
			repeatT = parse.getBool();

			if(repeatT)
			{
				glBindTexture(GL_TEXTURE_2D, texture);
				if(repeatT)
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
				else
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
				glBindTexture(GL_TEXTURE_2D, 0);
			}
		}
	}
}

//------------------------------------------------------------------------------
void MovieTexture::loadTexture()
{
#ifdef HAVE_FFMPEG
	int i;
	int err;
	int w, h;
	string filename = FileFinder::getInstance().findFile(url);

	if(formatContext)
		unloadTexture();

	//Try and open the file.
	err = av_open_input_file(&formatContext,
							 filename.c_str(),
							 NULL,
							 0,
							 NULL);
	if(err == 0)
	{
		//Get stream information.
		err = av_find_stream_info(formatContext);
		if(err >= 0)
		{
			//Dumps the stream information to stderr.
			//This seems like a helpful function, but if it dumps a huge pile of
			//stuff, remove it...
			dump_format(formatContext, 0, filename.c_str(), 0);

			//Find the first video stream (we ignore all the others for now).
			for(i=0;i<formatContext->nb_streams;++i)
			{
				if(formatContext->streams[i]->codec->codec_type == CODEC_TYPE_VIDEO)
				{
					videoStream = i;
					break;
				}
			}
			if(videoStream != -1)
			{
				codecContext = formatContext->streams[videoStream]->codec;

				//Find a decoder for this video.
				codec = avcodec_find_decoder(codecContext->codec_id);
				if(codec)
				{
					//Inform the codec that we can handle truncated bitstreams.
					if(codec->capabilities & CODEC_CAP_TRUNCATED)
						codecContext->flags |= CODEC_FLAG_TRUNCATED;

					//Open the codec.
					err = avcodec_open(codecContext, codec);
					if(err >= 0)
					{
						//Allocate RGB frame.
						rgbFrame = avcodec_alloc_frame();

						//Allocate video frame queue.
						for(i=0;i<VideoFrameQueueSize;++i)
							videoFrameQueue[i] = avcodec_alloc_frame();

						//Get resolution of the video.
						w = codecContext->width;
						h = codecContext->height;

						//Create a buffer to store the RGB pixel data.
						buffer = new uint8_t[w*h*3];
						avpicture_fill((AVPicture *)rgbFrame,
									   buffer,
									   PIX_FMT_RGB24,
									   w,
									   h);

						//Setup an SWS context for the RGB conversion.
						convertContext = sws_getContext(w,
														h,
														codecContext->pix_fmt,
														w,
														h,
														PIX_FMT_RGB24,
														SWS_BICUBIC,
														NULL,
														NULL,
														NULL);
						if(convertContext)
						{
							//Setup the texture.
							glBindTexture(GL_TEXTURE_2D, texture);
							glTexImage2D(GL_TEXTURE_2D,
										 0,
										 GL_RGB,
										 getPowerOfTwo(w),
										 getPowerOfTwo(h),
										 0,
										 GL_RGB,
										 GL_UNSIGNED_BYTE,
										 0);//0 = initialise to 0...
							glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
							glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
							if(repeatS)
								glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
							else
								glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
							if(repeatT)
								glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
							else
								glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
							glBindTexture(GL_TEXTURE_2D, 0);

							//Tell the user a bit about it...
							cout << "MovieTexture: Loaded " << url << endl;
							cout << "MovieTexture: video width = " << w << endl;
							cout << "              texture width = " << getPowerOfTwo(w) << endl;
							cout << "MovieTexture: video height = " << h << endl;
							cout << "              texture height = " << getPowerOfTwo(h) << endl;
						}
						else
						{
							delete [] buffer;
							av_free(rgbFrame);
							avcodec_close(codecContext);
							av_close_input_file(formatContext);
							formatContext = 0;
							videoStream = -1;
							buffer = 0;
							cout << "MovieTexture Error: Could not create an RGB conversion context." << endl;
						}
					}
					else
					{
						av_close_input_file(formatContext);
						formatContext = 0;
						ffmpegError(err);
					}
				}
				else
				{
					av_close_input_file(formatContext);
					formatContext = 0;
					cout << "MovieTexture Error: Could not find a decoder for this video file." << endl;
				}
			}
			else
			{
				av_close_input_file(formatContext);
				formatContext = 0;
				cout << "MovieTexture Error: Could not find any video streams." << endl;
			}
		}
		else
		{
			av_close_input_file(formatContext);
			formatContext = 0;
			ffmpegError(err);
		}
	}
	else
		ffmpegError(err);
#endif
}

//------------------------------------------------------------------------------
void MovieTexture::unloadTexture()
{
#ifdef HAVE_FFMPEG
	int i;
	
	sws_freeContext(convertContext);
	if(buffer)
		delete [] buffer;

	queueCritSec.lock();
	{
		for(i=0;i<VideoFrameQueueSize;++i)
			av_free(videoFrameQueue[i]);
		videoFrameQueue[0] = 0;
	}
	queueCritSec.unlock();

	av_free(rgbFrame);
	rgbFrame = 0;

	avcodec_close(codecContext);
	av_close_input_file(formatContext);
	formatContext = 0;
	videoStream = -1;
	buffer = 0;
#endif
}

#ifdef HAVE_FFMPEG
//------------------------------------------------------------------------------
void MovieTexture::getNextFrame()
{
	int err;
	int frameFinished;

	//Read the next frame.
	while(av_read_frame(formatContext, &packet) >= 0)
	{
		//There may be frames from streams other than the one we're interested
		//in.
		if(packet.stream_index == videoStream)
		{
			//Decode the frame.
			err = avcodec_decode_video(codecContext,
									   videoFrameQueue[0],
									   &frameFinished,
									   packet.data,
									   packet.size);
			if(err < 0)
				ffmpegError(err);

			//Check we could decode a frame.
			if(frameFinished)
			{
				//Convert image to RGB.
				sws_scale(convertContext,
						  videoFrameQueue[0]->data,
						  videoFrameQueue[0]->linesize,
						  0,
						  codecContext->height,
						  rgbFrame->data,
						  rgbFrame->linesize);

				//Free the packet.
				av_free_packet(&packet);
				break;
			}
		}

		//Free the packet.
		av_free_packet(&packet);
	}
}

//------------------------------------------------------------------------------
void MovieTexture::ffmpegError(int val)
{
	cout << "MovieTexture ffmpeg Error: ";
	switch(val)
	{
		case AVERROR_UNKNOWN:
			cout << "Unknown error (or invalid data)." << endl;
			break;
		case AVERROR_IO:
			cout << "I/O error." << endl;
			break;
		case AVERROR_NUMEXPECTED:
			cout << "Number syntax expected in filename." << endl;
			break;
		case AVERROR_NOMEM:
			cout << "Not enough memory." << endl;
			break;
		case AVERROR_NOFMT:
			cout << "Unknown format." << endl;
			break;
		case AVERROR_NOTSUPP:
			cout << "Operation not supported." << endl;
			break;
		case AVERROR_NOENT:
			cout << "No such file or directory." << endl;
			break;
		case AVERROR_PATCHWELCOME:
			cout << "Not yet implemented in ffmpeg." << endl;
			break;
		default:
			cout << "Unknown error." << endl;
	}
};
#endif

}
