//	MovieTexture.h - Class representing a movie texture.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef MOVIETEXTURE_H_
#define MOVIETEXTURE_H_

#include "AbstractNode.h"
#include "ThreadManager.h"

#ifdef HAVE_FFMPEG
extern "C"
{
#include <ffmpeg/avformat.h>
#include <ffmpeg/avcodec.h>
#include <ffmpeg/swscale.h>
}
#endif

#include "SDL.h"
#include "GLHeaders.h"

namespace HeilanX3D
{

///	Class representing a movie texture.
/*!
	Uses ffmpeg for the video decoding.  Original code taken from:
	http://www.inb.uni-luebeck.de/~boehme/using_libavcodec.html

	See also: http://www.dranger.com/ffmpeg/tutorial01.html
 */
class MovieTexture : public ChildNode,
					 public ThreadObject
{
  public:
	///	Constructor.
	MovieTexture();
	///	Destructor.
	~MovieTexture();

	///	Initialises the Xine stuff.
	void initialise();

	///	Draws the MovieTexture.
	void graphicDoStuff();

	///	The video is decoded in this method (which runs in its own thread).
	void threadIteration();

	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set the material's attributes.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<MovieTexture> constructor;
  private:
	///	Tries to load the texture in url.
	void loadTexture();
	///	Takes down all the video loading variables.
	void unloadTexture();
	///	Helper method to make sure we have a power of two-sized texture.
	int getPowerOfTwo(int val)
	{
		int retval = 2;

		while(retval < val)
			retval *= 2;

		return retval;
	};

	///	The url to the image file.
	std::string url;
	///	Whether the texture should repeat in the S direction.
	bool repeatS;
	///	Whether the texture should repeat in the T direction.
	bool repeatT;

	///	The texture we use.
	GLuint texture;

	///	The previous frame's tick value.
	/*!
		We use this to determine which frame to display.
	 */
	unsigned int lastTick;

	///	Used to protect the Video queue.
	CriticalSection queueCritSec;
	///	Necessary?
	bool needsRGBConvert;

#ifdef HAVE_FFMPEG
	///	Helper method to retrieve the next frame.
	void getNextFrame();
	///	Helper method to give us some idea what happened on an error.
	void ffmpegError(int val);

	///	Constants.
	enum
	{
		VideoFrameQueueSize = 32
	};

	///	Our ffmpeg format context.
	AVFormatContext *formatContext;
	///	Which video stream we're reading from.
	int videoStream;
	///	The codec context we're using.
	AVCodecContext *codecContext;
	///	The actual codec we're using.
	AVCodec *codec;
	///	Frame to render frames into.
	//AVFrame *frame;
	///	Frame to convert the video to RGB.
	AVFrame *rgbFrame;
	///	Buffer where the pixel data gets stored.
	uint8_t *buffer;
	///	The current packet we're decoding.
	AVPacket packet;
	///	Used to convert the YUV image to RGB.
	SwsContext *convertContext;

	///	Queue of AVFrames used to pass video frames to the display thread.
	AVFrame *videoFrameQueue[VideoFrameQueueSize];
	///	Our current read position in the video queue.
	int videoQueueRead;
	///	Our current write position in the video queue.
	int videoQueueWrite;
#endif
};

}

#endif
