//	NurbsCurve.cpp - Class representing a NURBS curve (i.e. a curving line).
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "NurbsCurve.h"
#include "StringParser.h"
#include "AttributeHelper.h"

#include "GLHeaders.h"
#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<NurbsCurve> NurbsCurve::constructor;

//------------------------------------------------------------------------------
NurbsCurve::NurbsCurve():
ParentNode(),
coordinates(0),
points(0),
numPoints(0),
tessellation(0),
closed(false),
order(3),
knotArray(0),
nurbsObject(0)
{
	setType("NurbsCurve");
	setVisual(true);
}

//------------------------------------------------------------------------------
NurbsCurve::~NurbsCurve()
{
	if(points)
		delete [] points;
	if(knotArray)
		delete [] knotArray;

	if(nurbsObject)
		gluDeleteNurbsRenderer(nurbsObject);
}

//------------------------------------------------------------------------------
void NurbsCurve::setAttribute(const string& attribute, EventValue& val)
{
	int i;
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(attribute == "tessellation")
		tessellation = val;
	else if(realAtt == "weight")
	{
		StringParser parser(val);
		AttributeHelper< vector<double> > helper(&weight);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else
		{
			if(weight.size() > 0)
				weight.clear();

			for(i=0;i<parser.getNumNumbers();++i)
				weight.push_back(parser.getDouble(i));
		}
	}
	else if(attribute == "closed")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			closed = parser.getBool();
	}
	else if(realAtt == "knot")
	{
		StringParser parser(val);
		AttributeHelper< vector<double> > helper(&knot);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else
		{
			if(knot.size() > 0)
				knot.clear();

			for(i=0;i<parser.getNumNumbers();++i)
				knot.push_back(parser.getDouble(i));

			critSec.lock();
			{
				if(knotArray)
					delete [] knotArray;
				knotArray = new float[knot.size()];

				for(i=0;(unsigned int)i<knot.size();++i)
					knotArray[i] = knot[i];
			}
			critSec.unlock();
		}
	}
	else if(attribute == "order")
		order = val;
}

//------------------------------------------------------------------------------
void NurbsCurve::parentInitialise()
{
	if(!nurbsObject)
		nurbsObject = gluNewNurbsRenderer();
	if(nurbsObject)
	{
		gluNurbsProperty(nurbsObject, GLU_SAMPLING_TOLERANCE, 25.0);
		gluNurbsProperty(nurbsObject, GLU_DISPLAY_MODE, GLU_FILL);
	}
}

//------------------------------------------------------------------------------
void NurbsCurve::postChildDraw()
{
	if(coordinates)
	{
		if(coordinates->getHasChanged())
			childrenLoaded();
	}

	if((points) && (nurbsObject))
	{
		glDisable(GL_LIGHTING);
		/*glColor3ub(0, 0, 0);
		glLineWidth(1.0f);

		int i;

		glBegin(GL_LINE_STRIP);
		{
			for(i=0;i<numPoints;i+=3)
				glVertex3d(points[i], points[i+1], points[i+2]);
		}
		glEnd();

		glLineWidth(2.0f);*/
		critSec.lock();
		{
			gluBeginCurve(nurbsObject);
			{
				gluNurbsCurve(nurbsObject,
							  knot.size(),
							  knotArray,
							  3,
							  points,
							  order,
							  GL_MAP1_VERTEX_3);
			}
			gluEndCurve(nurbsObject);
		}
		critSec.unlock();

		glEnable(GL_LIGHTING);
	}
}

//------------------------------------------------------------------------------
void NurbsCurve::childrenLoaded()
{
	int i, j, k;

	for(i=0;i<getNumChildren();++i)
	{
		if(getChild(i)->getType() == "Coordinate")
		{
			coordinates = dynamic_cast<Coordinate *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = coordinates->getPoints();

			if(!points)
			{
				numPoints = tempPoints.size() * 3;
				points = new float[numPoints];
			}

			k = 0;
			for(j=0;j<(numPoints/3);++j)
			{
				points[k] = tempPoints[j].x;
				++k;
				points[k] = tempPoints[j].y;
				++k;
				points[k] = tempPoints[j].z;
				++k;
			}

			break;
		}
	}
}

}
