//	NurbsCurve.h - Class representing a NURBS curve (i.e. a curving line).
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef NURBSCURVE_H_
#define NURBSCURVE_H_

#include "ParentNode.h"
#include "Coordinate.h"
#include "GLHeaders.h"

namespace HeilanX3D
{

///	Class representing a NURBS curve (i.e. a curving line).
/*!
	\todo Support solid attribute.
	\todo Support weight attribute (how?).
 */
class NurbsCurve : public ParentNode
{
  public:
	///	Constructor.
	NurbsCurve();
	///	Destructor.
	~NurbsCurve();

	///	Sets the curve's attributes.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<NurbsCurve> constructor;
  private:
	///	Initialises the nurbs object.
	void parentInitialise();

	///	Draws the curve.
	void postChildDraw();

	///	Overridden to do nothing.
	void audioBlockSizeChanged(int newSize) {};

	///	Used to get the control points of the curve from a child Coordinate node.
	void childrenLoaded();

	///	Pointer to the Coordinate node we use to draw the curve.
	Coordinate *coordinates;
	///	Our copy of coordinates' points.
	/*!
		We store them like this so they'll fit into the glu equations correctly.
	 */
	float *points;
	///	The number of values in points.
	/*!
		After points is first constructed, the number of values in it will not
		change.
	 */
	int numPoints;

	///	How the curve is tesselated.
	int tessellation;
	///	The weights of the control points.
	std::vector<double> weight;
	///	Whether the curve should be closed or not.
	bool closed;
	///	The knots for the curve.
	std::vector<double> knot;
	///	The order of the curve.
	int order;

	///	Knot array used to feed into gluNurbsCurve().
	/*!
		This is just a copy of knot.
	 */
	float *knotArray;
	///	Critical section used to protect knotArray.
	CriticalSection critSec;

	///	The GLU nurbs object we use to draw the curve.
	GLUnurbsObj *nurbsObject;
};

}

#endif
