//	NurbsPatchSurface.cpp - Class representing a NURBS surface.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "NurbsPatchSurface.h"
#include "StringParser.h"
#include "AttributeHelper.h"

#include <iostream>

using namespace std;

namespace HeilanX3D
{
	
NodeConstructor<NurbsPatchSurface> NurbsPatchSurface::constructor;

//------------------------------------------------------------------------------
NurbsPatchSurface::NurbsPatchSurface():
ParentNode(),
coordinates(0),
colors(0),
points(0),
colours(0),
numPoints(0),
numColours(0),
uTessellation(0),
vTessellation(0),
solid(true),
uClosed(false),
uDimension(0),
uOrder(3),
vClosed(false),
vDimension(0),
vOrder(3),
uKnotArray(0),
vKnotArray(0),
nurbsObject(0)
{
	setType("NurbsPatchSurface");
	setVisual(true);
}

//------------------------------------------------------------------------------
NurbsPatchSurface::~NurbsPatchSurface()
{
	if(points)
		delete [] points;
	if(uKnotArray)
		delete [] uKnotArray;
	if(vKnotArray)
		delete [] vKnotArray;

	if(nurbsObject)
		gluDeleteNurbsRenderer(nurbsObject);
}

//------------------------------------------------------------------------------
void NurbsPatchSurface::setAttribute(const string& attribute, EventValue& val)
{
	int i;
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(attribute == "uTessellation")
		uTessellation = val;
	else if(attribute == "vTessellation")
		vTessellation = val;
	else if(realAtt == "weight")
	{
		StringParser parser(val);
		AttributeHelper< vector<double> > helper(&weight);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else
		{
			if(weight.size() > 0)
				weight.clear();

			for(i=0;i<parser.getNumNumbers();++i)
				weight.push_back(parser.getDouble(i));
		}
	}
	else if(attribute == "solid")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			solid = parser.getBool();
	}
	else if(attribute == "uClosed")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			uClosed = parser.getBool();
	}
	else if(attribute == "uDimension")
		uDimension = val;
	else if(realAtt == "uKnot")
	{
		StringParser parser(val);
		AttributeHelper< vector<double> > helper(&uKnot);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else
		{
			if(uKnot.size() > 0)
				uKnot.clear();

			for(i=0;i<parser.getNumNumbers();++i)
				uKnot.push_back(parser.getDouble(i));

			critSec.lock();
			{
				if(uKnotArray)
					delete [] uKnotArray;
				uKnotArray = new float[uKnot.size()];

				for(i=0;(unsigned int)i<uKnot.size();++i)
					uKnotArray[i] = uKnot[i];
			}
			critSec.unlock();
		}
	}
	else if(attribute == "uOrder")
		uOrder = val;
	else if(attribute == "vClosed")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			vClosed = parser.getBool();
	}
	else if(attribute == "vDimension")
		vDimension = val;
	else if(realAtt == "vKnot")
	{
		StringParser parser(val);
		AttributeHelper< vector<double> > helper(&vKnot);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else
		{
			if(vKnot.size() > 0)
				vKnot.clear();

			for(i=0;i<parser.getNumNumbers();++i)
				vKnot.push_back(parser.getDouble(i));

			critSec.lock();
			{
				if(vKnotArray)
					delete [] vKnotArray;
				vKnotArray = new float[vKnot.size()];

				for(i=0;(unsigned int)i<vKnot.size();++i)
					vKnotArray[i] = vKnot[i];
			}
			critSec.unlock();
		}
	}
	else if(attribute == "vOrder")
		vOrder = val;
}

//------------------------------------------------------------------------------
void NurbsPatchSurface::parentInitialise()
{
	if(!nurbsObject)
		nurbsObject = gluNewNurbsRenderer();
	if(nurbsObject)
	{
		gluNurbsProperty(nurbsObject, GLU_SAMPLING_TOLERANCE, 50.0);
		gluNurbsProperty(nurbsObject, GLU_DISPLAY_MODE, GLU_FILL);
	}
}

//------------------------------------------------------------------------------
void NurbsPatchSurface::postChildDraw()
{
	if(coordinates)
	{
		if(coordinates->getHasChanged())
			childrenLoaded();
		else if(colors)
		{
			if(colors->getHasChanged())
				childrenLoaded();
		}
	}

	if((points) && (nurbsObject))
	{
		critSec.lock();
		{
			if(colours)
				glDisable(GL_LIGHTING);
			gluBeginSurface(nurbsObject);
			{
				if(colours)
				{
					gluNurbsSurface(nurbsObject,
							    	uKnot.size(),
									uKnotArray,
									vKnot.size(),
									vKnotArray,
									vDimension*4,
									4,
									colours,
									uOrder,
									vOrder,
									GL_MAP2_COLOR_4);
				}
				gluNurbsSurface(nurbsObject,
							    uKnot.size(),
							    uKnotArray,
							    vKnot.size(),
							    vKnotArray,
							    vDimension*3,
							    3,
							    points,
							    uOrder,
							    vOrder,
							    GL_MAP2_VERTEX_3);
			}
			gluEndSurface(nurbsObject);
			if(colours)
				glEnable(GL_LIGHTING);
		}
		critSec.unlock();
	}

	/*glDisable(GL_LIGHTING);
	glColor3ub(0, 0, 0);
	glBegin(GL_LINE_STRIP);
	{
		int i;

		for(i=0;i<numPoints;i+=3)
			glVertex3f(points[i], points[i+1], points[i+2]);
	}
	glEnd();
	glEnable(GL_LIGHTING);*/
}

//------------------------------------------------------------------------------
void NurbsPatchSurface::childrenLoaded()
{
	int i, j, k;

	for(i=0;i<getNumChildren();++i)
	{
		if((!coordinates) && (getChild(i)->getType() == "Coordinate"))
		{
			coordinates = dynamic_cast<Coordinate *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = coordinates->getPoints();

			if(!points)
			{
				numPoints = tempPoints.size() * 3;
				points = new float[numPoints];
			}

			k = 0;
			for(j=0;j<(numPoints/3);++j)
			{
				points[k] = tempPoints[j].x;
				++k;
				points[k] = tempPoints[j].y;
				++k;
				points[k] = tempPoints[j].z;
				++k;
			}
		}
		else if((!colors) && (getChild(i)->getType() == "Color"))
		{
			colors = dynamic_cast<Color *>(getChild(i));

			const vector<ThreeFloats>& tempColours = colors->getColours();

			if(!colours)
			{
				numColours = tempColours.size() * 4;
				colours = new float[numColours];
			}

			k = 0;
			for(j=0;(unsigned)j<tempColours.size();++j)
			{
				colours[k] = tempColours[j].x;
				++k;
				colours[k] = tempColours[j].y;
				++k;
				colours[k] = tempColours[j].z;
				++k;
				colours[k] = 1.0f; //gluNurbsSurface() needs an alpha value too.
				++k;
			}
		}
		else if((!colors) && (getChild(i)->getType() == "ColorRGBA"))
		{
			colorsRGBA = dynamic_cast<ColorRGBA *>(getChild(i));

			const vector<FourFloats>& tempColours = colorsRGBA->getColours();

			if(!colours)
			{
				numColours = tempColours.size() * 4;
				colours = new float[numColours];
			}

			k = 0;
			for(j=0;(unsigned)j<tempColours.size();++j)
			{
				colours[k] = tempColours[j].x;
				++k;
				colours[k] = tempColours[j].y;
				++k;
				colours[k] = tempColours[j].z;
				++k;
				colours[k] = tempColours[j].r;
				++k;
			}
		}
		else if(coordinates)
		{
			const vector<ThreeFloats>& tempPoints = coordinates->getPoints();

			k = 0;
			for(j=0;j<(numPoints/3);++j)
			{
				points[k] = tempPoints[j].x;
				++k;
				points[k] = tempPoints[j].y;
				++k;
				points[k] = tempPoints[j].z;
				++k;
			}

			if(colors)
			{
				const vector<ThreeFloats>& tempColours = colors->getColours();

				k = 0;
				for(j=0;(unsigned)j<tempColours.size();++j)
				{
					colours[k] = tempColours[j].x;
					++k;
					colours[k] = tempColours[j].y;
					++k;
					colours[k] = tempColours[j].z;
					++k;
					colours[k] = 1.0f; //gluNurbsSurface() needs an alpha value too.
					++k;
				}
			}
			else if(colorsRGBA)
			{
				const vector<FourFloats>& tempColours = colorsRGBA->getColours();

				k = 0;
				for(j=0;(unsigned)j<tempColours.size();++j)
				{
					colours[k] = tempColours[j].x;
					++k;
					colours[k] = tempColours[j].y;
					++k;
					colours[k] = tempColours[j].z;
					++k;
					colours[k] = tempColours[j].r;
					++k;
				}
			}
		}
		else if(colors)
		{
			const vector<ThreeFloats>& tempColours = colors->getColours();

			k = 0;
			for(j=0;(unsigned)j<tempColours.size();++j)
			{
				colours[k] = tempColours[j].x;
				++k;
				colours[k] = tempColours[j].y;
				++k;
				colours[k] = tempColours[j].z;
				++k;
				colours[k] = 1.0f; //gluNurbsSurface() needs an alpha value too.
				++k;
			}
		}
		else if(colorsRGBA)
		{
			const vector<FourFloats>& tempColours = colorsRGBA->getColours();

			k = 0;
			for(j=0;(unsigned)j<tempColours.size();++j)
			{
				colours[k] = tempColours[j].x;
				++k;
				colours[k] = tempColours[j].y;
				++k;
				colours[k] = tempColours[j].z;
				++k;
				colours[k] = tempColours[j].r;
				++k;
			}	
		}
	}
}

//------------------------------------------------------------------------------
void NurbsPatchSurface::nurbsError(GLenum errorCode)
{
	cout << "NURBS error: " << gluErrorString(errorCode) << endl;
}

}
