//	NurbsPatchSurface.h - Class representing a NURBS surface.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef NURBSPATCHSURFACE_H_
#define NURBSPATCHSURFACE_H_

#include "ParentNode.h"
#include "Coordinate.h"
#include "Color.h"
#include "ColorRGBA.h"

#include "GLHeaders.h"

namespace HeilanX3D
{

///	Class representing a NURBS surface.
class NurbsPatchSurface : public ParentNode
{
  public:
	///	Constructor.
	NurbsPatchSurface();
	///	Destructor.
	~NurbsPatchSurface();

	///	Sets the curve's attributes.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<NurbsPatchSurface> constructor;
  private:
	///	Initialises the nurbs object.
	void parentInitialise();

	///	Draws the surface.
	void postChildDraw();

	///	Overridden to do nothing.
	void audioBlockSizeChanged(int newSize) {};

	///	Used to get the control points of the curve from a child Coordinate node.
	void childrenLoaded();

	///	Used to get any NURBS errors.
	static void nurbsError(GLenum errorCode);

	///	Pointer to the Coordinate node we use to draw the surface.
	Coordinate *coordinates;
	///	Pointer to the Color node we use to colour the surface.
	Color *colors;
	///	Pointer to the ColorRGBA node we use to colour the surface.
	ColorRGBA *colorsRGBA;
	///	Our copy of coordinates' points.
	/*!
		We store them like this so they'll fit into the glu equations correctly.
	 */
	float *points;
	///	Our copy of colors' colours.
	/*!
		We store them like this so they'll fit into the glu equations correctly.
	 */
	float *colours;
	///	The number of values in points.
	/*!
		After points is first constructed, the number of values in it will not
		change.
	 */
	int numPoints;
	///	The number of values in colours.
	/*!
		After colours is first constructed, the number of values in it will not
		change.
	 */
	int numColours;

	///	How the surface is tessellated in the u direction.
	int uTessellation;
	///	How the surface is tessellated in the v direction.
	int vTessellation;
	///	The weights of the control points.
	std::vector<double> weight;
	///	Whether the surface should be solid or not.
	bool solid;
	///	Whether the surface should be closed in the u direction or not.
	bool uClosed;
	///	The number of control points in the u direction.
	int uDimension;
	///	The u dimension knots.
	std::vector<double> uKnot;
	///	The order of the u dimension.
	int uOrder;
	///	Whether the surface should be closed in the v direction or not.
	bool vClosed;
	///	The number of control points in the v direction.
	int vDimension;
	///	The v dimension knots.
	std::vector<double> vKnot;
	///	The order of the v dimension.
	int vOrder;

	///	Knot array used to feed into gluNurbsCurve().
	/*!
		This is just a copy of uKnot.
	 */
	float *uKnotArray;
	///	Knot array used to feed into gluNurbsCurve().
	/*!
		This is just a copy of vKnot.
	 */
	float *vKnotArray;
	///	Critical section used to protect uKnotArray and vKnotArray.
	CriticalSection critSec;

	///	The GLU nurbs object we use to draw the curve.
	GLUnurbsObj *nurbsObject;
};

}

#endif
