//	ParentNode.cpp - Base class for all nodes which hold child nodes.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "ParentNode.h"
#include "AmbisonicData.h"
#include "SceneConstructor.h"
#include "ScopedMutex.h"
#include "AudioThread.h"
#include "Lighting.h"
#include "Camera.h"

#include "SDL.h"
#include <iostream>

using namespace std;

namespace HeilanX3D
{

//------------------------------------------------------------------------------
ParentNode::ParentNode():
AbstractNode(),
summingBuffer(0),
blockSize(0),
paralleliseChildren(false),
pNumChildren(0),
pNumSamples(0)
{
	//ParentNodes *must* have aural set true, if there's a chance their children
	//might be generating audio.
	setAural();
	//Ditto for visual.
	setVisual();
}

//------------------------------------------------------------------------------
ParentNode::~ParentNode()
{
	//Delete children.
	clear();

	//Delete summingBuffer.
	if(summingBuffer)
		delete [] summingBuffer;
}

//------------------------------------------------------------------------------
void ParentNode::initialise()
{
	NodeArray::iterator it;

	parentInitialise();

	for(it=children.begin();it!= children.end();++it)
		(*it)->initialise();
}

//------------------------------------------------------------------------------
void ParentNode::graphicDoStuff()
{
	NodeArray::iterator it;

	Lighting::getInstance().pushLighting();
	preChildDraw();

	for(it=children.begin();it!= children.end();++it)
	{
		if((*it)->isVisual())
			(*it)->graphicDoStuff();
	}

	postChildDraw();
	Lighting::getInstance().popLighting();
}

//------------------------------------------------------------------------------
/*AmbisonicData *ParentNode::audioDoStuff(int numSamples, const ThreeFloats& pos)
{
	int i;
	unsigned int j;
	AmbisonicData *tempBuf = 0;
	NodeArray::iterator it;

	//First, make sure we've got the correct block size.
	if(numSamples > blockSize)
		setBlockSize(numSamples);

	//Now get audio from preChildAudio().
	tempBuf = preChildAudio(numSamples, pos);
	if(tempBuf)
	{
		for(i=0;i<numSamples;++i)
			summingBuffer[i] = tempBuf[i];
	}
	else
	{
		//Otherwise, clear summingBuffer.
		for(i=0;i<numSamples;++i)
			summingBuffer[i] = 0.0f;
	}

	//Now get audio from all the children outputting it.
	if(!paralleliseChildren)
	{
		for(it=children.begin();it!=children.end();++it)
		{
			if((*it)->isAural())
			{
				tempBuf = (*it)->audioDoStuff(numSamples, position);
				if(tempBuf)
				{
					for(i=0;i<numSamples;++i)
						summingBuffer[i] += tempBuf[i];
				}
			}
		}
	}
	else
	{
		pNumSamples = numSamples;
		//Signal the nodes to generate audio.
		for(j=0;j<audioThreads.size();++j)
			(audioThreads[j])->getAudioBlock(numSamples, pos);

		//When this returns, we'll have collected the audio from all our
		//children.
		if(pNumChildren < audioThreads.size())
			waiter.wait();
		pNumChildren = 0;
	}

	//Now get audio from postChildAudio().
	tempBuf = postChildAudio(numSamples, position);
	if(tempBuf)
	{
		for(i=0;i<numSamples;++i)
			summingBuffer[i] += tempBuf[i];
	}

	//Finally return summingBuffer.
	return summingBuffer;
}*/

//------------------------------------------------------------------------------
AmbisonicData *ParentNode::getAmbisonicAudio(int numSamples,
											 const ThreeFloats& pos)
{
	int i;
	unsigned int j;
	AmbisonicData *tempBuf = 0;
	NodeArray::iterator it;

	setPosition(pos);

	//First, make sure we've got the correct block size.
	if(numSamples > blockSize)
		setBlockSize(numSamples);

	//Now get audio from preChildAudio().
	/*tempBuf = preChildAudio(numSamples, pos);
	if(tempBuf)
	{
		for(i=0;i<numSamples;++i)
			summingBuffer[i] = tempBuf[i];
	}
	else*/
	{
		//Otherwise, clear summingBuffer.
		for(i=0;i<numSamples;++i)
			summingBuffer[i] = 0.0f;
	}

	//Now get audio from all the children outputting it.
	if(!paralleliseChildren)
	{
		for(it=children.begin();it!=children.end();++it)
		{
			if((*it)->isAural())
			{
				//tempBuf = (*it)->audioDoStuff(numSamples, position);
				tempBuf = (*it)->getAmbisonicAudio(numSamples, position);
				if(tempBuf)
				{
					for(i=0;i<numSamples;++i)
						summingBuffer[i] += tempBuf[i];
				}
			}
		}
	}
	else
	{
		pNumSamples = numSamples;
		//Signal the nodes to generate audio.
		for(j=0;j<audioThreads.size();++j)
			(audioThreads[j])->getAudioBlock(numSamples, position);

		//When this returns, we'll have collected the audio from all our
		//children.
		if(pNumChildren < audioThreads.size())
			waiter.wait();
		pNumChildren = 0;
	}

	//Now get any audio from this node.
	encodeBlock(numSamples, position, getMonoAudio(numSamples));
	if(tempBuf)
	{
		for(i=0;i<numSamples;++i)
			summingBuffer[i] += tempBuf[i];
	}

	//Finally return summingBuffer.
	return summingBuffer;
}

//------------------------------------------------------------------------------
void ParentNode::audioBlockStart()
{
	NodeArray::iterator it;

	//Inform subclasses.
	parentAudioBlockStart();

	//Inform all child nodes.
	for(it=children.begin();it!= children.end();++it)
		(*it)->audioBlockStart();
}

//------------------------------------------------------------------------------
void ParentNode::setBlockSize(int newSize)
{
	NodeArray::iterator it;

	//Update summingBuffer.
	if(summingBuffer)
		delete [] summingBuffer;
	summingBuffer = 0;

	blockSize = newSize;

	summingBuffer = new AmbisonicData[blockSize];

	//Now inform subclasses the block size has changed.
	audioBlockSizeChanged(blockSize);

	//And inform all child nodes.
	for(it=children.begin();it!= children.end();++it)
		(*it)->setBlockSize(blockSize);
}

//------------------------------------------------------------------------------
void ParentNode::setSamplerate(float val)
{
	NodeArray::iterator it;

	//Inform subclasses.
	samplerateChanged(val);

	//Inform children.
	for(it=children.begin();it!= children.end();++it)
		(*it)->setSamplerate(val);
}

//------------------------------------------------------------------------------
void ParentNode::addNode(AbstractNode *newChild, bool addToStart)
{
	int i;

	if((!addToStart) || (children.size() == 0))
		children.push_back(newChild);
	else
	{
		children.push_back(children[children.size()-1]);
		for(i=(children.size()-2);i>0;--i)
			children[i] = children[i-1];
		children[0] = newChild;
	}

	if((paralleliseChildren) && newChild->isAural())
	{
		AudioThread *tempAT = new AudioThread(newChild, this);
		audioThreads.push_back(tempAT);
	}
}

//------------------------------------------------------------------------------
void ParentNode::clear(bool dontDelete)
{
	unsigned int i;
	NodeArray::iterator it;

	for(i=0;i<audioThreads.size();++i)
		delete (audioThreads[i]);
	if(!dontDelete)
	{
		for(it=children.begin();it!= children.end();++it)
			(*it)->decRefCount();
	}

	children.clear();
	audioThreads.clear();
}

//------------------------------------------------------------------------------
void ParentNode::parallelAudioReady(AmbisonicData *buffer)
{
	int i;

	//So no more than one node tries to write to the buffer at any one time.
	ScopedMutex sm;

	if(buffer)
	{
		for(i=0;i<pNumSamples;++i)
			summingBuffer[i] += buffer[i];
	}
	else
		cout << "ParentNode parallel audio: No buffer!" << endl;

	//If we've received audio from all our child nodes, signal audioDoStuff()
	//that it can continue.
	++pNumChildren;
	if(pNumChildren >= audioThreads.size())
		waiter.signal();
}

//------------------------------------------------------------------------------
void ParentNode::setMetadata(const std::string& name, const std::string& value)
{
	unsigned int i;

	if((name == "heilan_parallelChildren") &&
	   ((value == "true") || (value == "true")))
	{
		paralleliseChildren = true;

		for(i=0;i<children.size();++i)
		{
			AudioThread *tempAT = new AudioThread(children[i], this);
			audioThreads.push_back(tempAT);
		}
	}
	else
		setParentMetadata(name, value);
}

//------------------------------------------------------------------------------
void ParentNode::sceneConstructionEnter(SceneConstructor *constructor)
{
	constructor->pushParent(this);
}

//------------------------------------------------------------------------------
void ParentNode::sceneConstructionLeave(SceneConstructor *constructor)
{
	constructor->popParent();
	childrenLoaded();
}

//------------------------------------------------------------------------------
AbstractNode *ParentNode::findNode(const std::string& name)
{
	unsigned int i;
	AbstractNode *retval = 0;

	if(getDefName() != name)
	{
		for(i=0;i<children.size();++i)
		{
			retval = children[i]->findNode(name);
			if(retval)
				break;
		}
	}
	else
		retval = this;

	return retval;
}

//------------------------------------------------------------------------------
void ParentNode::clearEvents()
{
	unsigned int i;

	for(i=0;i<children.size();++i)
		children[i]->handleEvents();

	handleEvents();
}

//------------------------------------------------------------------------------
void ParentNode::preChildDraw()
{
	
}

//------------------------------------------------------------------------------
void ParentNode::postChildDraw()
{
	
}

//------------------------------------------------------------------------------
/*AmbisonicData *ParentNode::preChildAudio(int numSamples,
										 const ThreeFloats& newPos)
{
	return 0;
}

//------------------------------------------------------------------------------
AmbisonicData *ParentNode::postChildAudio(int numSamples,
										  const ThreeFloats& newPos)
{
	return 0;
}*/

//------------------------------------------------------------------------------
AmbisonicData *ParentNode::encodeBlock(int numSamples,
									   const ThreeFloats& pos,
									   AmbisonicData *block)
{
	int i;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats camPos;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;
	const Quaternion& camRot = Camera::getInstance()->getAudioRotation();
	//const float rotMult = 6.283185306f/360.0f;

	if(block)
	{
		//Get current Ambisonic position (see AmbisonicData::operator=()),
		//relative to the camera.

		//So what we do here is:
		//1.) Ensure relativePos is equal to this node's spatial position in the
		//    scene.
		relativePos = pos;
		//2.) Rotate relativePos by the rotation of the camera, by multiplying
		//    it by the camera's orientation Quaternion.
		relativePos = camRot * relativePos;
		//3.) Translate relativePos by the position of the camera.
		relativePos -= camPos;

		lastAtten = calcRelativeAttenuation(lastPos);
		if(lastPos != relativePos)
		{
			ambiStart = lastPos;
			ambiEnd = relativePos;
			//We increment the position by a constant value, so we don't get
			//clicks or zipper noise.  Of course, there are disadvantages to
			//this approach...
			ambiInc = 0.005f;
			currentAmbi = ambiStart;
			newAtten = calcRelativeAttenuation(relativePos);
		}
		else
		{
			currentAmbi = relativePos;
			newAtten = lastAtten;
		}
		currentAtten = lastAtten;

		//Encode audio.
		for(i=0;i<numSamples;++i)
		{
			//Update currentAmbi and currentAtten if necessary.
			if(ambiInc > 0.0f)
			{
				//Calc currentAmbi.
				currentAmbi = ((1.0f-ambiIndex) * ambiStart) +
							  (ambiIndex * ambiEnd);
				//Calc currentAtten.
				currentAtten = ((1.0f-ambiIndex) * lastAtten) +
							   (ambiIndex * newAtten);

				ambiIndex += ambiInc;

				//Stop updating index if we've reached our destination.
				//I get weird rounding errors here, hence the .0000001f etc...
				if(ambiIndex > 1.0000001f)
				{
					ambiInc = 0.0f;
					currentAtten = newAtten;
				}
			}

			//Convert float sound file to Ambisonic, put it in the buffer.
			block[i] = currentAmbi * (block[i].W * currentAtten);
		}
		//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
		if(ambiIndex >= 1.0f)
		{
			lastPos = relativePos;
			lastAtten = currentAtten;
		}
		else
		{
			lastPos = ((1.0f-ambiIndex) * lastPos) + (ambiIndex * relativePos);
			lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
		}
	}

	return block;
}

//------------------------------------------------------------------------------
void ParentNode::setPosition(const ThreeFloats& newPos)
{
	position = newPos;
}

//------------------------------------------------------------------------------
AbstractNode *ParentNode::getChild(int index) const
{
	if(static_cast<unsigned int>(index) < children.size())
		return children[index];
	else
		return 0;
}

}
