//	ProximitySensor.cpp - Class representing an X3D ProximitySensor.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "ProximitySensor.h"
#include "StringParser.h"
#include "AttributeHelper.h"
#include "Camera.h"

#include <sstream>
#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<ProximitySensor> ProximitySensor::constructor;

//------------------------------------------------------------------------------
ProximitySensor::ProximitySensor():
ChildNode(),
enabled(true),
active(false),
currentTime(0.0),
lastTime(0.0),
samplerate(44100.0),
invSamplerate(1.0/44100.0)
{
	setType("ProximitySensor");
	setVisual(true);
	setAural(true);
}

//------------------------------------------------------------------------------
ProximitySensor::~ProximitySensor()
{
	
}

//------------------------------------------------------------------------------
void ProximitySensor::graphicDoStuff()
{
	const ThreeFloats& newCamPos = Camera::getInstance()->getGraphicPosition();
	const Quaternion& newCamRot = Camera::getInstance()->getGraphicRotation();

	if(enabled &&
	   ((size.x != 0.0f) && (size.y != 0.0f) && (size.z != 0.0f)))
	{
		if(!active)
		{
			//Check if the user's entered our bounding box.
			if(((newCamPos.x > (center.x-(size.x*0.5f))) &&
				(newCamPos.x < (center.x+(size.x*0.5f)))) &&
			   ((newCamPos.y > (center.y-(size.y*0.5f))) &&
				(newCamPos.y < (center.y+(size.y*0.5f)))) &&
			   ((newCamPos.z > (center.z-(size.z*0.5f))) &&
				(newCamPos.z < (center.z+(size.z*0.5f)))))
			{
				active = true;

				//We're supposed to interpolate to get the exact time the user
				//entered the box...
				//...but I can't work out how to do that, so I'll just cheat...
				sendEvent("isActive", "true");
				sendEvent("enterTime", (float)(lastTime+((currentTime-lastTime)*0.5)));
			}
		}
		else
		{
			//Check if the Camera's left our bounding box.
			if(((newCamPos.x < (center.x-(size.x*0.5f))) ||
				(newCamPos.x > (center.x+(size.x*0.5f)))) ||
			   ((newCamPos.y < (center.y-(size.y*0.5f))) ||
				(newCamPos.y > (center.y+(size.y*0.5f)))) ||
			   ((newCamPos.z < (center.z-(size.z*0.5f))) ||
				(newCamPos.z > (center.z+(size.z*0.5f)))))
			{
				active = false;

				//We're supposed to interpolate to get the exact time the user
				//left the box...
				//...but I can't work out how to do that, so I'll just cheat...
				sendEvent("isActive", "false");
				sendEvent("exitTime", (float)(lastTime+((currentTime-lastTime)*0.5)));
			}
			else
			{
				if(newCamPos != lastCamPos)
				{
					stringstream tempConv;

					tempConv << newCamPos.x << " ";
					tempConv << newCamPos.y << " ";
					tempConv << newCamPos.z;
					sendEvent("position_changed", tempConv.str());
				}
				if(newCamRot != lastCamRot)
				{
					stringstream tempConv;

					tempConv << newCamRot.w << " ";
					tempConv << newCamRot.x << " ";
					tempConv << newCamRot.y << " ";
					tempConv << newCamRot.z;
					sendEvent("orientation_changed", tempConv.str());
				}
			}
		}
	}

	lastCamPos = newCamPos;
	lastCamRot = newCamRot;
	lastTime = currentTime;
}

//------------------------------------------------------------------------------
AmbisonicData *ProximitySensor::getMonoAudio(int numSamples)
{
	currentTime += (double)numSamples * invSamplerate;

	return 0;
}

//------------------------------------------------------------------------------
void ProximitySensor::setSamplerate(float val)
{
	samplerate = (double)val;
	invSamplerate = 1.0/samplerate;
}

//------------------------------------------------------------------------------
void ProximitySensor::setAttribute(const string& attribute, EventValue& val)
{
	if(attribute == "center")
	{
		StringParser parse(val);
		AttributeHelper<ThreeFloats> helper(&center);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else
		{
			if(parse.getNumThreeFloats() > 0)
				center = parse.getThreeFloats();
		}
	}
	else if(attribute == "enabled")
	{
		StringParser parse(val);

		if(parse.getNumBools() > 0)
			enabled = parse.getBool();
	}
	else if(attribute == "size")
	{
		StringParser parse(val);
		AttributeHelper<ThreeFloats> helper(&size);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else
		{
			if(parse.getNumThreeFloats() > 0)
				size = parse.getThreeFloats();
		}
	}
}

}
