//	Sound.cpp - Sound node, used to play back AudioClips.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Sound.h"
#include "AudioClip.h"
#include "AmbisonicData.h"
#include "Camera.h"
#include "StringParser.h"
#include "AttributeHelper.h"

#include <cassert>
#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<Sound> Sound::constructor;

//------------------------------------------------------------------------------
Sound::Sound():
ParentNode(),
loop(true),
intensity(1.0f),
audioClip(0),
currentBlock(0),
fileSize(0),
index(0.0f),
totalIndex(0.0f),
increment(1.0f),
audioBuffer(0),
bufferSize(0),
samplerate(44100.0f),
startTime(0.0f),
stopTime(-1.0f),
pauseTime(-1.0f),
resumeTime(-1.0f),
startSamples(0),
stopSamples(-1),
pauseSamples(-1),
resumeSamples(-1),
elapsedSamples(0)
{
	setType("Sound");
	setVisual(false);
	setAural(true);
}

//------------------------------------------------------------------------------
Sound::~Sound()
{
	if(audioBuffer)
		delete [] audioBuffer;
}

//------------------------------------------------------------------------------
AmbisonicData *Sound::getAmbisonicAudio(int numSamples, const ThreeFloats& pos)
{
	int i;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;
	AmbisonicData *retval = getMonoAudio(numSamples);
	const ThreeFloats camPos = Camera::getInstance()->getAudioPosition();
	const Quaternion& camRot = Camera::getInstance()->getAudioRotation();
	//const float rotMult = 6.283185306f/360.0f;

	if(retval)
	{
		setPosition(location + pos);

		//Get current Ambisonic position (see AmbisonicData::operator=()),
		//relative to the camera.

		//So what we do here is:
		//1.) Ensure relativePos is equal to this node's spatial position in the
		//    scene.
		relativePos = location + pos;
		//2.) Rotate relativePos by the rotation of the camera, by multiplying
		//    it by the camera's orientation Quaternion.
		relativePos = camRot * relativePos;
		//3.) Translate relativePos by the position of the camera.
		relativePos -= camPos;

		lastAtten = calcRelativeAttenuation(lastPos);
		if(lastPos != relativePos)
		{
			ambiStart = lastPos;
			ambiEnd = relativePos;
			//We increment the position by a constant value, so we don't get
			//clicks or zipper noise.  Of course, there are disadvantages to
			//this approach...
			ambiInc = 0.005f;
			currentAmbi = ambiStart;
			newAtten = calcRelativeAttenuation(relativePos);
		}
		else
		{
			if(relativePos != 0.0f)
				currentAmbi = relativePos;
			else
			{
				currentAmbi.W = 0.707f;
				currentAmbi.X = 0;
				currentAmbi.Y = 0;
				currentAmbi.Z = 0;
			}
			newAtten = lastAtten;
		}
		currentAtten = lastAtten;

		//Encode audio.
		for(i=0;i<numSamples;++i)
		{
			//Update currentAmbi and currentAtten if necessary.
			if(ambiInc > 0.0f)
			{
				//Calc currentAmbi.
				currentAmbi = ((1.0f-ambiIndex) * ambiStart) +
							  (ambiIndex * ambiEnd);
				//Calc currentAtten.
				currentAtten = ((1.0f-ambiIndex) * lastAtten) +
							   (ambiIndex * newAtten);

				ambiIndex += ambiInc;

				//Stop updating index if we've reached our destination.
				//I get weird rounding errors here, hence the .0000001f etc...
				if(ambiIndex > 1.0000001f)
				{
					ambiInc = 0.0f;
					currentAtten = newAtten;
				}
			}

			//Convert float sound file to Ambisonic, put it in the buffer.
			retval[i] = currentAmbi * (retval[i].W * currentAtten);
		}
		//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
		if(ambiIndex >= 1.0f)
		{
			lastPos = relativePos;
			lastAtten = currentAtten;
		}
		else
		{
			lastPos = ((1.0f-ambiIndex) * lastPos) + (ambiIndex * relativePos);
			lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
		}
	}

	return retval;
}

//------------------------------------------------------------------------------
AmbisonicData *Sound::getMonoAudio(int numSamples)
{
	int i;
	float tempf;

	//Skip everything if we don't have an AudioClip.
	if(!audioClip)
		return 0;

	//Make sure we're up to date.
	loop = audioClip->getLoop();
	increment = audioClip->getPitch();
	startTime = audioClip->getStartTime();
	stopTime = audioClip->getStopTime();
	resumeTime = audioClip->getResumeTime();
	pauseTime = audioClip->getPauseTime();
	startSamples = static_cast<int>(samplerate * startTime);
	stopSamples = static_cast<int>(samplerate * stopTime);
	pauseSamples = static_cast<int>(samplerate * pauseTime);
	resumeSamples = static_cast<int>(samplerate * resumeTime);

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		if((loop || ((totalIndex-startSamples) < fileSize)) &&
		   (elapsedSamples >= startSamples) &&
		   (((stopSamples > 0) && (elapsedSamples < stopSamples)) ||
			(stopSamples <= 0)) &&
		   ((((pauseSamples > 0) &&
			  ((elapsedSamples > resumeSamples) ||
			   (elapsedSamples < pauseSamples)))) ||
			(pauseSamples <= 0))) //(Ouch!  This is horrible!)
		{
			tempf = currentBlock[static_cast<int>(index)];

			//Update index and totalIndex.
			totalIndex += increment;
			index += increment; //Faster than doing fmod().

			if(index >= AudioClip::streamBlockSize)
			{
				//Load the next block.
				//totalIndex = fmod(totalIndex, fileSize);
				index = fmod(index, AudioClip::streamBlockSize);
				currentBlock = audioClip->getNextBlock(this);
			}
		}
		else
			tempf = 0.0f;

		//Convert float sound file to Ambisonic, put it in the buffer.
		audioBuffer[i].W = tempf * intensity;

		//Update elapsed time.
		++elapsedSamples;
	}

	return audioBuffer;
}

//------------------------------------------------------------------------------
void Sound::setAttribute(const string& attribute, EventValue& val)
{
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "location")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&location);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			location = parser.getThreeFloats();
	}
	else if(attribute == "intensity")
		intensity = val;
	else if(attribute == "pitch")
		increment = val;
	else if(attribute == "maxBack")
		setMaxBackAudioRadius(val);
	else if(attribute == "maxFront")
		setMaxFrontAudioRadius(val);
	else if(attribute == "minBack")
		setMinBackAudioRadius(val);
	else if(attribute == "minFront")
		setMinFrontAudioRadius(val);
}

//------------------------------------------------------------------------------
void Sound::childrenLoaded()
{
	unsigned int i;

	for(i=0;i<(unsigned)getNumChildren();++i)
	{
		if(getChild(i)->getType() == "AudioClip")
		{
			AudioClip *tempAudioClip = dynamic_cast<AudioClip *>(getChild(i));

			if(tempAudioClip->getFileSize())
			{
				//Set audioClip.
				audioClip = tempAudioClip;

				//Set currentBlock and fileSize.
				currentBlock = audioClip->getCurrentBlock();
				fileSize = audioClip->getFileSize();
			}

			loop = audioClip->getLoop();
			increment = audioClip->getPitch();
			startTime = audioClip->getStartTime();
			stopTime = audioClip->getStopTime();
			resumeTime = audioClip->getResumeTime();
			pauseTime = audioClip->getPauseTime();

			break;
		}
	}
}

//------------------------------------------------------------------------------
void Sound::samplerateChanged(float newRate)
{
	samplerate = newRate;

	startSamples = static_cast<int>(samplerate * startTime);
	stopSamples = static_cast<int>(samplerate * stopTime);
	pauseSamples = static_cast<int>(samplerate * pauseTime);
	resumeSamples = static_cast<int>(samplerate * resumeTime);
}

//------------------------------------------------------------------------------
AmbisonicData *Sound::preChildAudio(int numSamples, const ThreeFloats& newPos)
{
	int i;
	float ambiIndex = 0.0f;
	float ambiInc = 0.0f;
	AmbisonicData currentAmbi;
	AmbisonicData ambiStart;
	AmbisonicData ambiEnd;
	ThreeFloats relativePos;
	float newAtten = 0.0f;
	float currentAtten;
	float tempf;

	//Skip everything if we don't have an AudioClip.
	if(!audioClip)
		return 0;

	//This should NEVER happen, unless there's an error somewhere else...
	assert(numSamples <= bufferSize);

	setPosition(location + newPos);

	//Get current Ambisonic position (see AmbisonicData::operator=()), relative
	//to the camera.
	relativePos = (location + newPos);
	relativePos -= Camera::getInstance()->getAudioPosition();
	lastAtten = calcRelativeAttenuation(lastPos);
	if(lastPos != relativePos)
	{
		ambiStart = lastPos;
		ambiEnd = relativePos;
		//ambiInc = 1.0f/numSamples;
		//We increment the position by a constant value, so we don't get clicks
		//or zipper noise.  Of course, there are disadvantages to this
		//approach...
		ambiInc = 0.005f;
		currentAmbi = ambiStart;
		newAtten = calcRelativeAttenuation(relativePos);
	}
	else
	{
		currentAmbi = relativePos;
		newAtten = lastAtten;
	}
	currentAtten = lastAtten;

	//Calculate audio, put in audioBuffer.
	for(i=0;i<numSamples;++i)
	{
		//Calculate monophonic audio.
		if((loop || (totalIndex < fileSize)) &&
		   (elapsedSamples >= startSamples) &&
		   (((stopSamples > 0) && (elapsedSamples < stopSamples)) ||
			(stopSamples <= 0)) &&
		   ((((pauseSamples > 0) &&
			  ((elapsedSamples > resumeSamples) ||
			   (elapsedSamples < pauseSamples)))) ||
			(pauseSamples <= 0))) //(Ouch!  This is horrible!)
		{
			tempf = currentBlock[static_cast<int>(index)];

			//Update index and totalIndex.
			totalIndex += increment;
			index += increment; //Faster than doing fmod().

			if(index >= AudioClip::streamBlockSize)
			{
				//Load the next block.
				//totalIndex = fmod(totalIndex, fileSize);
				index = fmod(index, AudioClip::streamBlockSize);
				currentBlock = audioClip->getNextBlock(this);
			}
		}
		else
			tempf = 0.0f;

		//Update currentAmbi and currentAtten if necessary.
		if(ambiInc > 0.0f)
		{
			//Calc currentAmbi.
			currentAmbi = ((1.0f-ambiIndex) * ambiStart) +
						  (ambiIndex * ambiEnd);
			//Calc currentAtten.
			currentAtten = ((1.0f-ambiIndex) * lastAtten) +
						   (ambiIndex * newAtten);

			ambiIndex += ambiInc;

			//Stop updating index if we've reached our destination.
			//I get weird rounding errors here, hence the .0000001f etc...
			if(ambiIndex > 1.0000001f)
			{
				ambiInc = 0.0f;
				currentAtten = newAtten;
			}
		}

		//Convert float sound file to Ambisonic, put it in the buffer.
		audioBuffer[i] = currentAmbi * (tempf * intensity * currentAtten);

		//Update elapsed time.
		++elapsedSamples;
	}
	//Set lastPos correctly (I'm assuming this is okay, Ambisonically?).
	if(ambiIndex >= 1.0f)
	{
		lastPos = relativePos;
		lastAtten = currentAtten;
	}
	else
	{
		lastPos = ((1.0f-ambiIndex) * lastPos) + (ambiIndex * relativePos);
		lastAtten = ((1.0f-ambiIndex) * lastAtten) + (ambiIndex * newAtten);
	}

	return audioBuffer;
}

//------------------------------------------------------------------------------
void Sound::audioBlockSizeChanged(int newSize)
{
	if(audioBuffer)
		delete [] audioBuffer;

	bufferSize = newSize;
	audioBuffer = new AmbisonicData[bufferSize];
}

}
