//	Sound.h - Sound node, used to play back AudioClips.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SOUND_H_
#define SOUND_H_

#include "ParentNode.h"

namespace HeilanX3D
{

class AudioClip;

///	Sound node, used to play back AudioClips.
/*!
	How this works is Sound is given an extra Heilan-only attribute
	(audioClipReady), which is set to true by an AudioClip child node once it's
	loaded it's sound file.  When Sound receives this audioClipReady 'message',
	it goes through it's children nodes to look for an AudioClip - when it finds
	one, it then stores the pointer, to be used to stream audio from disk.

	\todo Is this method of streaming sound files too complex?
	\todo Add direction attribute and convert maxBack etc. values elliptical
	as opposed to fixed circular.
 */
class Sound : public ParentNode
{
  public:
	///	Constructor.
	Sound();
	///	Destructor.
	~Sound();

	///	Overridden to take the location attribute into account.
	/*!
		We don't sum any child nodes' audio because (I don't think) Sound's
		supposed to be able to have child sound-emitting nodes.
	 */
	AmbisonicData *getAmbisonicAudio(int numSamples, const ThreeFloats& pos);
	///	Where the audio is calculated.
	AmbisonicData *getMonoAudio(int numSamples);

	///	Used to set this node's attributes.
	/*!
		Also used by child AudioClip nodes to communicate that they've loaded
		their sound file, and it's ready for streaming.

		\todo Sort out the loop attribute.
	 */
	void setAttribute(const std::string& attribute, EventValue& val);

	///	Used to determine whether a child AudioClip has been loaded.
	void childrenLoaded();

	static NodeConstructor<Sound> constructor;
  private:
	///	Called so the node knows what samplerate it's running at.
	void samplerateChanged(float newRate);

	///	Used to stream audioClip to the soundcard.
	AmbisonicData *preChildAudio(int numSamples, const ThreeFloats& newPos);
	///	Called when the soundcard block size changes.
	void audioBlockSizeChanged(int newSize);

	///	Whether or not the sound file should be played looped or not.
	bool loop;
	///	The node's location attribute.
	ThreeFloats location;
	///	The node's intensity.
	float intensity;

	///	Pointer to the current AudioClip node we're streaming sound from.
	AudioClip *audioClip;
	///	Pointer to the current block of audio we're reading from.
	float *currentBlock;
	///	The size of the sound file we're streaming from.
	float fileSize;

	///	Index we're currently reading from the current block.
	/*!
		\todo Could we make this an int to speed things up?
	 */
	float index;
	///	Index we're reading from the overall sound file.
	float totalIndex;
	///	Current increment (used to read from the sound file at different speeds).
	float increment;

	///	Temporary audio block returned from audioDoStuff().
	AmbisonicData *audioBuffer;
	///	Size of audioBuffer.
	int bufferSize;

	///	Used to interpolate the node's position if the camera (or it) is moving.
	ThreeFloats lastPos;
	///	Used to interpolate the node's attenuation if the camera (or it) is moving.
	float lastAtten;

	///	The current samplerate.
	float samplerate;

	///	The time (in seconds) this node should start playing.
	float startTime;
	///	The time (in seconds) this node should stop playing.
	float stopTime;
	///	The time (in seconds) this node should pause playing.
	float pauseTime;
	///	The time (in seconds) this node should resume playing.
	float resumeTime;
	///	The time (in samples) this node should start playing.
	int startSamples;
	///	The time (in samples) this node should stop playing.
	int stopSamples;
	///	The time (in samples) this node should pause playing.
	int pauseSamples;
	///	The time (in samples) this node should resume playing.
	int resumeSamples;
	///	The elapsed time (in samples).
	int elapsedSamples;
};

}

#endif
