//	Sphere.h - Class representing an X3D sphere node.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SPHERE_H_
#define SPHERE_H_

#include "AbstractNode.h"

#include "GLHeaders.h"
#include <cmath>

namespace HeilanX3D
{

///	Class representing an X3D cylinder node.
/*!
	\todo Fix normals for top & bottom of sphere.
 */
class Sphere : public ChildNode
{
  public:
	///	Constructor.
	Sphere();
	///	Destructor.
	~Sphere();

	///	Initialises various bits and pieces.
	void initialise();

	///	Draws the sphere.
	void graphicDoStuff();
	///	Not used, overridden to do nothing.
	/*AmbisonicData *audioDoStuff(int numSamples,
								const ThreeFloats& pos) {return 0;};*/

	///	Not used, overridden to do nothing.
	void setBlockSize(int newSize) {};

	///	Used to set the sphere's parameters.
	void setAttribute(const std::string& attribute, EventValue& val);

	static NodeConstructor<Sphere> constructor;
  private:
	///	Helper method to calculate the normal of a polygon.
	inline void calcNormal(const ThreeFloats& v1,
						   const ThreeFloats& v2,
						   const ThreeFloats& v3)
	{
		ThreeFloats a;
		ThreeFloats b;
		ThreeFloats normal;
		float length;

		a.x = v1.x - v2.x;
		a.y = v1.y - v2.y;
		a.z = v1.z - v2.z;

		b.x = v2.x - v3.x;
		b.y = v2.y - v3.y;
		b.z = v2.z - v3.z;

		normal.x = (a.y * b.z) - (a.z * b.y);
		normal.y = (a.z * b.x) - (a.x * b.z);
		normal.z = (a.x * b.y) - (a.y * b.x);

		length = sqrtf((normal.x*normal.x)+(normal.y*normal.y)+(normal.z*normal.z));
		if(length == 0.0f)
			length = 1.0f;
		else
			length = 1.0f/length;

		glNormal3f((normal.x*length), (normal.y*length), (normal.z*length));
	};
	///	Second version, to return the normal instead of drawing it.
	inline ThreeFloats calcNormal2(const ThreeFloats& v1,
								   const ThreeFloats& v2,
								   const ThreeFloats& v3)
	{
		ThreeFloats a;
		ThreeFloats b;
		ThreeFloats normal;
		float length;

		a.x = v1.x - v2.x;
		a.y = v1.y - v2.y;
		a.z = v1.z - v2.z;

		b.x = v2.x - v3.x;
		b.y = v2.y - v3.y;
		b.z = v2.z - v3.z;

		normal.x = (a.y * b.z) - (a.z * b.y);
		normal.y = (a.z * b.x) - (a.x * b.z);
		normal.z = (a.x * b.y) - (a.y * b.x);

		length = sqrtf((normal.x*normal.x)+(normal.y*normal.y)+(normal.z*normal.z));
		if(length == 0.0f)
			length = 1.0f;
		else
			length = 1.0f/length;

		normal.x *= length;
		normal.y *= length;
		normal.z *= length;

		return normal;
	};
	///	To calculate smooth normals.
	inline ThreeFloats calcSmooth(ThreeFloats& n1,
								  ThreeFloats& n2,
								  ThreeFloats& n3,
								  ThreeFloats& n4)
	{
		ThreeFloats retval;

		retval.x = (n1.x + n2.x + n3.x + n4.x)*0.25f;
		retval.y = (n1.y + n2.y + n3.y + n4.y)*0.25f;
		retval.z = (n1.z + n2.z + n3.z + n4.z)*0.25f;

		return retval;
	};

	///	Helper method to calculate a vertex on the sphere.
	void calcVertex(int x, int y);
	///	Helper method for calcVertex().
	inline ThreeFloats getVertex(int x, int y)
	{
		ThreeFloats retval;

		/*if(x >= NumSections)
			x -= NumSections;
		if(y >= NumSections)
			y -= NumSections;*/

		//x => phi, y => theta.

		retval.x = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) * 
				   cosf(((float)y/(float)NumSections) * 6.283185306f);
		retval.y = sinf((((float)x/(float)NumSections)-0.5f) * 3.141592653f);
		retval.z = cosf((((float)x/(float)NumSections)-0.5f) * 3.141592653f) *
				   sinf(((float)y/(float)NumSections) * 6.283185306f);

		return retval;
	}

	///	Radius of the sphere.
	float radius;
	///	Whether or not the sphere should be solid.
	bool solid;

	///	Enum containing the number of sides we use to draw the sphere.
	enum
	{
		NumSections = 32
	};

	///	The display list containing the sphere shape.
	GLuint displayList;
};

}

#endif
