//	TimeSensor.cpp - Node used to output time-based events.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "TimeSensor.h"
#include "StringParser.h"

#include <sstream>
#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<TimeSensor> TimeSensor::constructor;
//------------------------------------------------------------------------------
TimeSensor::TimeSensor():
ChildNode(),
samplerate(44100.0f),
cycleInterval(1.0),
enabled(true),
loop(false),
pauseTime(0.0),
resumeTime(0.0),
startTime(0.0),
stopTime(0.0),
cycleTime(0.0),
elapsedTime(0.0),
fractionChanged(0.0f),
isActive(false),
isPaused(false),
time(0.0),
invSamplerate(1.0/44100.0),
currentCycle(0),
lastCycle(0.0),
cyclesFinished(false),
contCounter(0)
{
	setType("TimeSensor");
	setVisual(false);
	setAural(true);
}

//------------------------------------------------------------------------------
TimeSensor::~TimeSensor()
{
	
}

//------------------------------------------------------------------------------
AmbisonicData *TimeSensor::getMonoAudio(int numSamples)
{
	int i;

	if(enabled)
	{
		//The sensor should send an isActive event when it's first enabled,
		//assuming the startTime etc. values agree.
		if((time == 0.0) && 
		   (startTime <= 0.0) &&
		   ((pauseTime != 0.0) || (pauseTime == resumeTime)))
		{
			isActive = true;
			sendEvent("isActive", "true");
		}

		for(i=0;i<numSamples;++i)
		{
			time += invSamplerate;

			if(isActive)
			{
				//Check if we're not paused, or stopped.
				if(((time < pauseTime) || (time >= resumeTime)) &&
				   ((time <= stopTime) || (stopTime <= startTime)))
				{
					elapsedTime += invSamplerate;

					//Send isPaused = false if it's currently true.
					if(isPaused)
					{
						isPaused = false;
						sendEvent("isPaused", "false");
					}

					//Check if we're on a new cycle, output cycleTime if so.
					if((int)(elapsedTime/cycleInterval) > currentCycle)
					{
						currentCycle = static_cast<int>(elapsedTime/cycleInterval);
						lastCycle = elapsedTime;
						if(loop)
							sendEvent("cycleTime", (float)time);
						else
						{
							isActive = false;
							cyclesFinished = true;
							sendEvent("isActive", "false");
						}
					}
				}
				//If we're paused.
				else if((time >= pauseTime) && (time < resumeTime))
				{
					if(!isPaused)
					{
						isPaused = true;
						sendEvent("isPaused", "true");
					}
				}
				//If we should stop.
				else if((time >= stopTime) && (stopTime > startTime))
				{
					isActive = false;
					sendEvent("isActive", "false");
				}
			}
			else
			{
				//Check if we've passed startTime, and should activated the
				//sensor.
				if((time >= startTime) &&
				   ((time <= stopTime) || (stopTime <= startTime)) &&
				   (!cyclesFinished))
				{
					isActive = true;
					sendEvent("isActive", "true");
				}
			}
		}
	}
	else
	{
		for(i=0;i<numSamples;++i)
			time += invSamplerate;
	}

	for(i=0;i<numSamples;++i)
	{
		++contCounter;

		if(contCounter > (int)(samplerate*0.01f))
		{
			contCounter = 0;

			//Send time and elapsedTime events, if necessary.
			if(isActive && !isPaused)
			{
				sendEvent("time", (float)time);
				sendEvent("elapsedTime", (float)elapsedTime);
			}

			//Send fraction_changed event, if necessary.
			if(isActive && !isPaused && !cyclesFinished)
			{
				fractionChanged = static_cast<float>((elapsedTime-lastCycle)/
										  			 cycleInterval);
				sendEvent("fraction_changed", fractionChanged);
			}
		}
	}

	return 0;
}

//------------------------------------------------------------------------------
void TimeSensor::setSamplerate(float val)
{
	samplerate = val;
	invSamplerate = 1.0/static_cast<double>(samplerate);
}

//------------------------------------------------------------------------------
void TimeSensor::setAttribute(const string& attribute, EventValue& val)
{
	if(attribute == "cycleInterval")
	{
		StringParser parse(val);

		if(parse.getNumNumbers() > 0)
			cycleInterval = parse.getDouble();
	}
	else if(attribute == "enabled")
	{
		StringParser parse(val);

		if(parse.getNumBools() > 0)
			enabled = parse.getBool();
	}
	else if(attribute == "loop")
	{
		StringParser parse(val);

		if(parse.getNumBools() > 0)
		{
			loop = parse.getBool();
			if(loop)
				cyclesFinished = false;
		}
	}
	else if(attribute == "pauseTime")
	{
		StringParser parse(val);

		if(parse.getNumNumbers() > 0)
			pauseTime = parse.getDouble();
	}
	else if(attribute == "resumeTime")
	{
		StringParser parse(val);

		if(parse.getNumNumbers() > 0)
			resumeTime = parse.getDouble();
	}
	else if(attribute == "startTime")
	{
		StringParser parse(val);

		if(parse.getNumNumbers() > 0)
			startTime = parse.getDouble();

		if(startTime > time)
		{
			//Reset cyclesFinished.
			cyclesFinished = false;
		}
	}
	else if(attribute == "stopTime")
	{
		StringParser parse(val);

		if(parse.getNumNumbers() > 0)
			stopTime = parse.getDouble();
	}
}

}
