//	Transform.cpp - Class representing the Transform parent node.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Transform.h"
#include "StringParser.h"
#include "AttributeHelper.h"

#include "GLHeaders.h"
#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<Transform> Transform::constructor;

//------------------------------------------------------------------------------
Transform::Transform():
ParentNode(),
center(0.0f),
rotation(0.0f, 0.0f, 0.0f, 1.0f),
scale(1.0f),
scaleOrientation(0.0f, 0.0f, 0.0f, 1.0f),
translation(0.0f),
bboxCenter(0.0f),
bboxSize(-1.0f)
{
	ThreeFloats temp3f;

	setType("Transform");
	setVisual(true);
	setAural(true);

	temp3f = ThreeFloats(rotation.x, rotation.y, rotation.z);
	rotation.setFromAxisAngle(rotation.w, temp3f);
	temp3f = ThreeFloats(scaleOrientation.x,
						 scaleOrientation.y,
						 scaleOrientation.z);
	scaleOrientation.setFromAxisAngle(scaleOrientation.w, temp3f);
}

//------------------------------------------------------------------------------
Transform::~Transform()
{
	
}

//------------------------------------------------------------------------------
AmbisonicData *Transform::getAmbisonicAudio(int numSamples,
											const ThreeFloats& newPos)
{
	ThreeFloats tempPos;

	//Rotate translation by the rotation.
	tempPos = rotation * translation;

	setPosition(tempPos + newPos);

	return ParentNode::getAmbisonicAudio(numSamples, (translation+newPos));
}

//------------------------------------------------------------------------------
void Transform::setAttribute(const string& attribute, EventValue& val)
{
	ThreeFloats temp3f;
	FourFloats temp4f;
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "center")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&center);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			center = parser.getThreeFloats();
	}
	else if(realAtt == "rotation")
	{
		StringParser parser(val);
		AttributeHelper<FourFloats> helper(&temp4f);

		if(helper.isValid(attribute))
		{
			helper.convert(attribute, val);
			temp3f = ThreeFloats(temp4f.r, temp4f.x, temp4f.y);
			rotation.setFromAxisAngle(temp4f.z*0.017453292519943295769236907684886,
									  temp3f);
		}
		else if(parser.getNumFourFloats() > 0)
		{
			temp4f = parser.getFourFloats();
			temp3f = ThreeFloats(temp4f.r, temp4f.x, temp4f.y);
			rotation.setFromAxisAngle(temp4f.z*0.017453292519943295769236907684886,
									  temp3f);
		}
	}
	else if(realAtt == "scale")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&scale);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			scale = parser.getThreeFloats();
	}
	else if(realAtt == "scaleOrientation")
	{
		StringParser parser(val);
		AttributeHelper<FourFloats> helper(&temp4f);

		if(helper.isValid(attribute))
		{
			helper.convert(attribute, val);
			temp3f = ThreeFloats(temp4f.r, temp4f.x, temp4f.y);
			scaleOrientation.setFromAxisAngle(temp4f.z*0.017453292519943295769236907684886,
											  temp3f);
		}
		else if(parser.getNumFourFloats() > 0)
		{
			temp4f = parser.getFourFloats();
			temp3f = ThreeFloats(temp4f.r, temp4f.x, temp4f.y);
			scaleOrientation.setFromAxisAngle(temp4f.z*0.017453292519943295769236907684886,
											  temp3f);
		}
	}
	else if(realAtt == "translation")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&translation);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			translation = parser.getThreeFloats();
	}
	else if(realAtt == "bboxCenter")
	{
		StringParser parser(val);
		AttributeHelper<ThreeFloats> helper(&bboxCenter);

		if(helper.isValid(attribute))
			helper.convert(attribute, val);
		else if(parser.getNumThreeFloats() > 0)
			bboxCenter = parser.getThreeFloats();
	}
	else if(realAtt == "bboxSize")
	{
		StringParser parser(val);

		if(parser.getNumThreeFloats() > 0)
			bboxSize = parser.getThreeFloats();
	}
}

//------------------------------------------------------------------------------
void Transform::preChildDraw()
{
	ThreeFloats axis;
	double angle = 0.0;

	//Setup matrix.
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	//Now perform various translations etc.

	//Apply the rotation.
	rotation.getAxisAngle(angle, axis);
	glRotatef((angle * 57.295779513082320876798154814105),
			  axis.x,
			  axis.y,
			  axis.z);

	//Apply the translation, offset from center.
	glTranslatef(center.x+translation.x,
				 center.y+translation.y,
				 center.z+translation.z);

	//First sort out the scale orientation.
	scaleOrientation.getAxisAngle(angle, axis);
	glRotatef((angle * 57.295779513082320876798154814105),
			  axis.x,
			  axis.y,
			  axis.z);
	//Now Apply our scale operation.
	glScalef(scale.x, scale.y, scale.z);
	//Revert to previous orientation.
	glRotatef(-(angle * 57.295779513082320876798154814105),
			  -axis.x,
			  -axis.y,
			  -axis.z);
}

//------------------------------------------------------------------------------
void Transform::postChildDraw()
{
	GLint tempint;

	//Reset matrix.
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	//Check if we should pop the texture matrix.
	//This is far from ideal, but I can't think of anywhere else to put it,
	//since TextureTransform is a child node...
	glGetIntegerv(GL_TEXTURE_STACK_DEPTH, &tempint);
	if(tempint > 1)
	{
		glMatrixMode(GL_TEXTURE);
		glPopMatrix();

		//Just in case...
		glMatrixMode(GL_MODELVIEW);
	}
}

//------------------------------------------------------------------------------
/*AmbisonicData *Transform::preChildAudio(int numSamples,
										const ThreeFloats& newPos)
{
	setPosition(translation + newPos);

	return 0;
}*/

}
