//	TriangleFanSet.cpp - Class representing an X3D triangle fan set.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "TriangleFanSet.h"
#include "MultiTextureCoordinate.h"
#include "StringParser.h"
#include "AttributeHelper.h"

#include <iostream>

using namespace std;

namespace HeilanX3D
{

NodeConstructor<TriangleFanSet> TriangleFanSet::constructor;

//------------------------------------------------------------------------------
TriangleFanSet::TriangleFanSet():
ParentNode(),
ccw(true),
colorPerVertex(true),
normalPerVertex(true),
solid(true),
coordinates(0),
normals(0),
colors(0),
colorsRGBA(0)
{
	setType("TriangleFanSet");
	setVisual(true);

	textureCoords[0] = 0;
	textureCoords[1] = 0;
}

//------------------------------------------------------------------------------
TriangleFanSet::~TriangleFanSet()
{
	
}

//------------------------------------------------------------------------------
void TriangleFanSet::setAttribute(const string& attribute, EventValue& val)
{
	int i;
	string realAtt = AttributeHelperBase::getAddress(attribute);

	if(realAtt == "fanCount")
	{
		StringParser parser(val);
		AttributeHelper< vector<int> > helper(&fanCount);

		critSec.lock();
		{
			if(helper.isValid(attribute))
				helper.convert(attribute, val);
			else
			{
				if(fanCount.size() > 0)
					fanCount.clear();
	
				for(i=0;i<parser.getNumNumbers();++i)
					fanCount.push_back(parser.getInt(i));
			}
		}
		critSec.unlock();
	}
	else if(attribute == "ccw")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			ccw = parser.getBool();
	}
	else if(attribute == "colorPerVertex")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			colorPerVertex = parser.getBool();
	}
	else if(attribute == "normalPerVertex")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			normalPerVertex = parser.getBool();
	}
	else if(attribute == "solid")
	{
		StringParser parser(val);

		if(parser.getNumBools() > 0)
			solid = parser.getBool();
	}
}

//------------------------------------------------------------------------------
void TriangleFanSet::postChildDraw()
{
	unsigned int i = 1;
	unsigned int j;
	unsigned int fanNum = 0;
	unsigned int fanStart = 0;

	if(coordinates)
	{
		//Re-load if necessary.
		if(coordinates->getHasChanged())
			childrenLoaded();

		//If we're supposed to draw colours onto the object, disable lighting,
		//otherwise you won't see anything.
		if(!textureCoords[0] && (colors || colorsRGBA))
			glDisable(GL_LIGHTING);

		while(i < drawingCoords.size())
		{
			while((i-1-fanStart) <= (unsigned)fanCount[fanNum+1])
			{
				glBegin(GL_TRIANGLES);
				{
					//Handle normals if they're supposed to be per face.
					if(!normalPerVertex)
					{
						if(normals && (i < drawingCoords.size()))
						{
							glNormal3f(drawingNormals[i].x,
									   drawingNormals[i].y,
									   drawingNormals[i].z);
						}
						else //Calculate normals ourselves.
						{
							
						}
					}

					//Handle colours if they're supposed to be per face.
					if((!colorPerVertex) &&
					   (!textureCoords[0]) &&
					   (colors || colorsRGBA))
					{
						glColor4f(drawingColours[i].x,
								  drawingColours[i].y,
								  drawingColours[i].z,
								  drawingColours[i].r);
					}

					for(j=0;j<3;++j)
					{
						int tempIndex;

						if((j == 0) || (i > drawingCoords.size()))
							tempIndex = fanStart;
						else
							tempIndex = i;

						if(tempIndex == -1)
							break;

						//Handle per vertex normals.
						if(normals && normalPerVertex &&
						   (i < drawingCoords.size()))
						{
							glNormal3f(drawingNormals[tempIndex].x,
									   drawingNormals[tempIndex].y,
									   drawingNormals[tempIndex].z);
						}
						else //Calculate normals ourselves.
						{
							
						}

						//Handle texturing.
						if(textureCoords[0] && (i < drawingCoords.size()))
						{
							glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
												 drawingTex[0][tempIndex].x,
									 			 drawingTex[0][tempIndex].y);

							//Handle multitexturing.
							if(textureCoords[1] && (i < drawingCoords.size()))
							{
								glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
													 drawingTex[1][tempIndex].x,
													 drawingTex[1][tempIndex].y);
							}
						}
						//Handle per vertex colouring.
						else if((colors || colorsRGBA) &&
								colorPerVertex &&
								(i < drawingCoords.size()))
						{
							glColor4f(drawingColours[tempIndex].x,
									  drawingColours[tempIndex].y,
									  drawingColours[tempIndex].z,
									  drawingColours[tempIndex].r);
						}

						//Draw the vertex.
						glVertex3f(drawingCoords[tempIndex].x,
								   drawingCoords[tempIndex].y,
								   drawingCoords[tempIndex].z);

						if(j == 1)
							++i;
					}
				}
				glEnd();

				if((i+1) >= drawingCoords.size())
					break;
			}
			++i;

			if(i < drawingCoords.size())
			{
				fanStart = i;
				++i;
			}
		}

		if(!textureCoords[0] && (colors || colorsRGBA))
			glEnable(GL_LIGHTING);
	}
}

//------------------------------------------------------------------------------
void TriangleFanSet::childrenLoaded()
{
	unsigned int i, j, k;

	for(i=0;i<(unsigned)getNumChildren();++i)
	{
		if(getChild(i)->getType() == "Coordinate")
		{
			coordinates = dynamic_cast<Coordinate *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = coordinates->getPoints();

			drawingCoords.clear();

			for(j=0;j<tempPoints.size();++j)
				drawingCoords.push_back(tempPoints[j]);
		}
		else if(getChild(i)->getType() == "Normal")
		{
			normals = dynamic_cast<Normal *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = normals->getVectors();

			drawingNormals.clear();

			for(j=0;j<tempPoints.size();++j)
				drawingNormals.push_back(tempPoints[j]);
		}
		else if(getChild(i)->getType() == "TextureCoordinate")
		{
			textureCoords[0] = dynamic_cast<TextureCoordinate *>(getChild(i));

			const vector<TwoFloats>& tempPoints = textureCoords[0]->getCoords();

			drawingTex[0].clear();

			for(j=0;j<tempPoints.size();++j)
				drawingTex[0].push_back(tempPoints[j]);
		}
		else if(getChild(i)->getType() == "MultiTextureCoordinate")
		{
			MultiTextureCoordinate *tempCoord = 0;
			tempCoord = dynamic_cast<MultiTextureCoordinate *>(getChild(i));

			if(tempCoord)
			{
				for(k=0;k<(unsigned)tempCoord->getNumChildTexCoord();++k)
				{
					textureCoords[k] = tempCoord->getChildTexCoord(k);

					const vector<TwoFloats>& tempPoints = textureCoords[k]->getCoords();

					drawingTex[k].clear();

					for(j=0;j<tempPoints.size();++j)
						drawingTex[k].push_back(tempPoints[j]);
				}
			}
		}
		else if(getChild(i)->getType() == "Color")
		{
			colors = dynamic_cast<Color *>(getChild(i));

			const vector<ThreeFloats>& tempPoints = colors->getColours();

			drawingColours.clear();

			for(j=0;j<tempPoints.size();++j)
			{
				FourFloats tempf(tempPoints[j].x,
								 tempPoints[j].y,
								 tempPoints[j].z,
								 1.0f);

				drawingColours.push_back(tempf);
			}
		}
		else if(getChild(i)->getType() == "ColorRGBA")
		{
			colorsRGBA = dynamic_cast<ColorRGBA *>(getChild(i));

			const vector<FourFloats>& tempPoints = colorsRGBA->getColours();

			drawingColours.clear();

			for(j=0;j<tempPoints.size();++j)
				drawingColours.push_back(tempPoints[j]);
		}
	}
}

}
