//	Camera.cpp - Singleton class representing the camera/listening point in the
//			   scene.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Camera.h"
#include "ExamineNavigation.h"
#include "FatalError.h"
#include "OpenSoundController.h"
#include "OSCMessage.h"

#include "GLHeaders.h"
#include <sstream>
#include <iostream>

using namespace std;

namespace HeilanX3D
{

//------------------------------------------------------------------------------
Camera *Camera::getInstance()
{
	static Camera retval;

	return &retval;
}

//------------------------------------------------------------------------------
void Camera::initialise()
{
	if(currentNav)
	{
		setGraphicPosition(currentNav->getPosition(), true);
		setGraphicRotation(currentNav->getRotation());
	}
}

//------------------------------------------------------------------------------
void Camera::setNavigationTypes(const std::string& text)
{
	string tempstr;

	navTypes = text;

	//Update currentNav accordingly.
	if(currentNav)
	{
		delete currentNav;
		currentNav = 0;
	}

	//Examine is the default.
	if((navTypes.find("ANY", 0) != string::npos)||
	   (navTypes.find("EXAMINE", 0) != string::npos))
		currentNav = NavigationFactory::getInstance()->createNav("EXAMINE");
	else //Try to construct the first type in navTypes.
	{
		//If there's more than one entry, select the first one.
		//(we skip the first character on the off chance there's a leading
		// space character)
		if(navTypes.find(' ', 1) != string::npos)
		{
			tempstr = navTypes.substr(0, (navTypes.find(' ', 1)-1));
			currentNav = NavigationFactory::getInstance()->createNav(tempstr);
			setGraphicPosition(currentNav->getPosition());
			setGraphicRotation(currentNav->getRotation());
		}
		else //Otherwise, try to create it from navTypes.
		{
			tempstr = navTypes;
			//cout << "currentNav before: " << (int)currentNav << endl;
			currentNav = NavigationFactory::getInstance()->createNav(tempstr);
			/*setGraphicPosition(currentNav->getPosition());
			setGraphicRotation(currentNav->getRotation());*/
			//cout << "currentNav after: " << (int)currentNav << endl;
		}

		/*setGraphicPosition(currentNav->getPosition());
		setGraphicRotation(currentNav->getRotation());		//Fallback if the above did not work.*/
		if(!currentNav)
		{
			currentNav = NavigationFactory::getInstance()->createNav("EXAMINE");
			setGraphicPosition(currentNav->getPosition());
			setGraphicRotation(currentNav->getRotation());
			cout << "Could not set navigation to: " << tempstr;
			cout << "  falling back to: EXAMINE" << endl;
		}
	}

	currentNav->setIsOffline(offline);
}

//------------------------------------------------------------------------------
void Camera::setNavigation(const std::string& text)
{
	if((navTypes.find(text, 0) == string::npos)&&
	   (navTypes.find("ANY", 0) == string::npos))
	{
		cout << "Tried to set navigation to: " << text;
		cout << ", which is not allowed in this scene." << endl;
		return;
	}

	if(currentNav)
	{
		delete currentNav;
		currentNav = 0;
	}

	currentNav = NavigationFactory::getInstance()->createNav(text);
	setGraphicPosition(currentNav->getPosition());
	setGraphicRotation(currentNav->getRotation());

	currentNav->setIsOffline(offline);
}

//------------------------------------------------------------------------------
void Camera::setGraphicPosition(const ThreeFloats& val, bool noOSC)
{
	//Lock critical section.
	if(posCritSec.lock())
	{
		//Set position.
		gPosition = val;
		//cout << "c: " << val.x << " " << val.y << " " << val.z << endl;

		//Unlock critical section.
		posCritSec.unlock();
	}

	//Send message to OpenSoundController.
	if(!noOSC)
	{
		OSC::Message mess;

		mess.setAddress("/heilan/camera/position.x");
		mess.addFloat32(gPosition.x);
		OpenSoundController::getInstance()->handleOSCMessage(&mess, true);

		mess.clearMessage();
		mess.setAddress("/heilan/camera/position.y");
		mess.addFloat32(gPosition.y);
		OpenSoundController::getInstance()->handleOSCMessage(&mess, true);

		mess.clearMessage();
		mess.setAddress("/heilan/camera/position.z");
		mess.addFloat32(gPosition.z);
		OpenSoundController::getInstance()->handleOSCMessage(&mess, true);
	}
}

//------------------------------------------------------------------------------
void Camera::updateAudioPosition()
{
	//Lock critical section.
	if(posCritSec.lock())
	{
		//Update position.
		aPosition = gPosition;

		//Unlock critical section.
		posCritSec.unlock();
	}
}

//------------------------------------------------------------------------------
void Camera::setGraphicRotation(const Quaternion& val)
{
	//Lock critical section.
	if(rotCritSec.lock())
	{
		//Set rotation.
		gRotation = val;
		//cout << "gRotation.w = " << gRotation.w << endl;

		//Unlock critical section.
		rotCritSec.unlock();
	}

	//Send message to OpenSoundController.
	{
		OSC::Message mess;

		mess.setAddress("/heilan/camera/rotation.x");
		mess.addFloat32(gRotation.x);
		OpenSoundController::getInstance()->handleOSCMessage(&mess, true);

		mess.clearMessage();
		mess.setAddress("/heilan/camera/rotation.y");
		mess.addFloat32(gRotation.y);
		OpenSoundController::getInstance()->handleOSCMessage(&mess, true);

		mess.clearMessage();
		mess.setAddress("/heilan/camera/rotation.z");
		mess.addFloat32(gRotation.z);
		OpenSoundController::getInstance()->handleOSCMessage(&mess, true);

		mess.clearMessage();
		mess.setAddress("/heilan/camera/rotation.w");
		mess.addFloat32(gRotation.w);
		OpenSoundController::getInstance()->handleOSCMessage(&mess, true);
	}
}

//------------------------------------------------------------------------------
void Camera::updateAudioRotation()
{
	//Lock critical section.
	if(rotCritSec.lock())
	{
		//Update position.
		aRotation = gRotation;

		//Unlock critical section.
		rotCritSec.unlock();
	}
}

//------------------------------------------------------------------------------
void Camera::draw()
{
	/*//Rotate camera to gRotation.
	//glRotatef(gRotation.r, gRotation.x, gRotation.y, gRotation.z);
	glRotatef(gRotation.x, 0.0f, 1.0f, 0.0f);
	glRotatef(gRotation.y, 1.0f, 0.0f, 0.0f);

	//Move camera to gPosition.
	//Why are x and z negative?
	glTranslatef(-gPosition.x, gPosition.y, -gPosition.z);*/
	glTranslatef(-gPosition.x, gPosition.y, -gPosition.z);
	//glMultMatrixd(gRotation.getMatrix()); //I'm not sure the getMatrix() method is entirely correct...
	ThreeFloats axis;
	double angle = 0.0;
	gRotation.getAxisAngle(angle, axis);
	glRotatef((angle * 57.295779513082320876798154814105),
			  axis.x,
			  axis.y,
			  axis.z);
}

//------------------------------------------------------------------------------
Camera::Camera():
currentNav(0),
offline(false),
offlineTicks(0),
viewWidth(640),
viewHeight(480)
{
	//By default, currentNav is set to an ExamineNavigation instance.
	currentNav = new ExamineNavigation();
	//setGraphicPosition(currentNav->getPosition(), true);
	//setGraphicRotation(currentNav->getRotation());
	//gPosition = currentNav->getPosition();
}

//------------------------------------------------------------------------------
Camera::~Camera()
{
	if(currentNav)
		delete currentNav;
}

}
