//	FileFinder.h - Singleton class to use when looking for a file.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef FILEFINDER_H_
#define FILEFINDER_H_

#include "CriticalSection.h"

#include <fstream>
#include <vector>
#include <string>

namespace HeilanX3D
{

///	Singleton class to use when looking for a file.
class dllStuff FileFinder
{
  public:
	///	Returns the sole instance of FileFinder.
	static FileFinder& getInstance();

	///	Looks for a file, returns a corrected path if it can find it.
	/*!
		\param path The path to the file.
		\param storePath If the path to the file is valid, store it (minus the
		filename) in our vector of paths, so future calls to findFile() will
		search that path.
		\return If it can find the file, returns the file's path, otherwise
		returns "".

		This is so we can include textures, sound files etc. in the same folder
		as the X3D file, without having to specify their absolute paths.  When
		X3DFileLoader first loads a scene, it calls findFile() with storePath
		as true, and the X3D file's path is stored, so any further file loading
		can be done relative to that path.
	 */
	std::string findFile(const std::string& path, bool storePath = false);

	///	Clears our vector of stored paths.
	void clearStoredPaths();

	///	Destructor.
	~FileFinder();
  private:
	///	Constructor.
	FileFinder();

	///	Helper method to test if a file exists.
	bool exists(const std::string& path);

	///	Our stored paths.
	std::vector<std::string> paths;
	///	The platform-specific separator character.
	char separator;

	///	Critical section to protect paths.
	CriticalSection critSec;
};

}

#endif
