//	KeyboardInput.cpp - Singleton class to handle keyboard input.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "KeyboardInput.h"

#include <string>
#include <cassert>
#include <algorithm>

using namespace std;

namespace HeilanX3D
{

//------------------------------------------------------------------------------
KeyboardInput& KeyboardInput::getInstance()
{
	static KeyboardInput instance;

	return instance;
}

//------------------------------------------------------------------------------
void KeyboardInput::registerCallback(KeyboardCallback *callback)
{
	assert(callback != 0);

	callbacks.push_back(callback);
}

//------------------------------------------------------------------------------
void KeyboardInput::unregisterCallback(KeyboardCallback *callback)
{
	vector<KeyboardCallback *>::iterator it;

	it = find(callbacks.begin(), callbacks.end(), callback);
	if(it != callbacks.end())
		callbacks.erase(it);
}

//------------------------------------------------------------------------------
void KeyboardInput::keyDown(SDLKey key)
{
	unsigned int i;
	string tempstr;

	if(isActionKey(key))
	{
		KeyboardCallback::ActionKey actionKey = getActionKey(key);

		for(i=0;i<callbacks.size();++i)
			callbacks[i]->actionKeyPressed(actionKey);
	}
	else
	{
		switch(key)
		{
			case SDLK_LALT:
			case SDLK_RALT:
				for(i=0;i<callbacks.size();++i)
					callbacks[i]->altKeyChanged(true);
				break;
			case SDLK_LCTRL:
			case SDLK_RCTRL:
				for(i=0;i<callbacks.size();++i)
					callbacks[i]->ctrlKeyChanged(true);
				break;
			case SDLK_LSHIFT:
			case SDLK_RSHIFT:
				for(i=0;i<callbacks.size();++i)
					callbacks[i]->shiftKeyChanged(true);
				break;
			//Is this okay?
			default:
				tempstr = (char)key;
				for(i=0;i<callbacks.size();++i)
					callbacks[i]->keyPressed(tempstr);
		}
	}
}

//------------------------------------------------------------------------------
void KeyboardInput::keyUp(SDLKey key)
{
	unsigned int i;
	string tempstr;

	if(isActionKey(key))
	{
		KeyboardCallback::ActionKey actionKey = getActionKey(key);

		for(i=0;i<callbacks.size();++i)
			callbacks[i]->actionKeyReleased(actionKey);
	}
	else
	{
		switch(key)
		{
			case SDLK_LALT:
			case SDLK_RALT:
				for(i=0;i<callbacks.size();++i)
					callbacks[i]->altKeyChanged(false);
				break;
			case SDLK_LCTRL:
			case SDLK_RCTRL:
				for(i=0;i<callbacks.size();++i)
					callbacks[i]->ctrlKeyChanged(false);
				break;
			case SDLK_LSHIFT:
			case SDLK_RSHIFT:
				for(i=0;i<callbacks.size();++i)
					callbacks[i]->shiftKeyChanged(false);
				break;
			//Is this okay?
			default:
				tempstr = (char)key;
				for(i=0;i<callbacks.size();++i)
					callbacks[i]->keyReleased(tempstr);
		}
	}
}

//------------------------------------------------------------------------------
KeyboardInput::~KeyboardInput()
{
	
}

//------------------------------------------------------------------------------
KeyboardInput::KeyboardInput()
{
	
}

//------------------------------------------------------------------------------
bool KeyboardInput::isActionKey(SDLKey key)
{
	bool retval;

	switch(key)
	{
		case SDLK_F1:
		case SDLK_F2:
		case SDLK_F3:
		case SDLK_F4:
		case SDLK_F5:
		case SDLK_F6:
		case SDLK_F7:
		case SDLK_F8:
		case SDLK_F9:
		case SDLK_F10:
		case SDLK_F11:
		case SDLK_F12:
		case SDLK_PAGEUP:
		case SDLK_PAGEDOWN:
		case SDLK_UP:
		case SDLK_DOWN:
		case SDLK_LEFT:
		case SDLK_RIGHT:
		case SDLK_BACKSPACE:
		case SDLK_RETURN:
			retval = true;
			break;
		default:
			retval = false;
	}

	return retval;
}

//------------------------------------------------------------------------------
KeyboardCallback::ActionKey KeyboardInput::getActionKey(SDLKey key)
{
	KeyboardCallback::ActionKey retval;

	switch(key)
	{
		case SDLK_F1:
			retval = KeyboardCallback::F1;
			break;
		case SDLK_F2:
			retval = KeyboardCallback::F2;
			break;
		case SDLK_F3:
			retval = KeyboardCallback::F3;
			break;
		case SDLK_F4:
			retval = KeyboardCallback::F4;
			break;
		case SDLK_F5:
			retval = KeyboardCallback::F5;
			break;
		case SDLK_F6:
			retval = KeyboardCallback::F6;
			break;
		case SDLK_F7:
			retval = KeyboardCallback::F7;
			break;
		case SDLK_F8:
			retval = KeyboardCallback::F8;
			break;
		case SDLK_F9:
			retval = KeyboardCallback::F9;
			break;
		case SDLK_F10:
			retval = KeyboardCallback::F10;
			break;
		case SDLK_F11:
			retval = KeyboardCallback::F11;
			break;
		case SDLK_F12:
			retval = KeyboardCallback::F12;
			break;
		case SDLK_PAGEUP:
			retval = KeyboardCallback::PgUp;
			break;
		case SDLK_PAGEDOWN:
			retval = KeyboardCallback::PgDown;
			break;
		case SDLK_UP:
			retval = KeyboardCallback::Up;
			break;
		case SDLK_DOWN:
			retval = KeyboardCallback::Down;
			break;
		case SDLK_LEFT:
			retval = KeyboardCallback::Left;
			break;
		case SDLK_RIGHT:
			retval = KeyboardCallback::Right;
			break;
		case SDLK_BACKSPACE:
			retval = KeyboardCallback::Backspace;
			break;
		case SDLK_RETURN:
			retval = KeyboardCallback::Return;
			break;
		default:
			retval = KeyboardCallback::None;
	}

	return retval;
}

}
