//	KeyboardInput.h - Singleton class to handle keyboard input.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef KEYBOARDINPUT_H_
#define KEYBOARDINPUT_H_

#include "SDL_keysym.h"

#include <vector>
#include <string>

namespace HeilanX3D
{

//------------------------------------------------------------------------------
///	Abstract callback used to receive data when the user presses a key.
class KeyboardCallback
{
  public:
	///	Destructor.
	virtual ~KeyboardCallback() {};

	///	Called by KeyboardInput when the user presses a key.
	virtual void keyPressed(const std::string& key) = 0;
	///	Called by KeyboardInput when the user releases a key.
	virtual void keyReleased(const std::string& key) = 0;
	///	Called by KeyboardInput when the alt key changes state.
	/*!
		\param val True == down, false == up.
	 */
	virtual void altKeyChanged(bool val) {};
	///	Called by KeyboardInput when the ctrl key changes state.
	/*!
		\param val True == down, false == up.
	 */
	virtual void ctrlKeyChanged(bool val) {};
	///	Called by KeyboardInput when the shift key changes state.
	/*!
		\param val True == down, false == up.
	 */
	virtual void shiftKeyChanged(bool val) {};

	///	Enum of the action keys, as specified by the X3D spec.
	enum ActionKey
	{
		F1 = 1,
		F2,
		F3,
		F4,
		F5,
		F6,
		F7,
		F8,
		F9,
		F10,
		F11,
		F12,
		Home,
		End,
		PgUp,
		PgDown,
		Up,
		Down,
		Left,
		Right,
		Backspace,
		Return,

		None
	};

	///	Called by KeyboardInput when the user presses an action key.
	virtual void actionKeyPressed(ActionKey key) {};
	///	Called by KeyboardInput when the user releases an action key.
	virtual void actionKeyReleased(ActionKey key) {};
};

//------------------------------------------------------------------------------
///	Singleton class to handle keyboard input.
class KeyboardInput
{
  public:
	///	Returns the sole instance of KeyboardInput.
	static KeyboardInput& getInstance();

	///	Registers a KeyboardCallback with us.
	void registerCallback(KeyboardCallback *callback);
	///	Unregisters a KeyboardCallback with us.
	void unregisterCallback(KeyboardCallback *callback);

	///	Called when we receive a key down event.
	void keyDown(SDLKey key);
	///	Called when we receive a key up event.
	void keyUp(SDLKey key);

	///	Destructor.
	/*!
		Should be private, but since MSVC v6 is used on Windows, we can't do
		that...
	 */
	~KeyboardInput();
  private:
	///	Constructor.
	KeyboardInput();

	///	Helper method to determine whether key is an ActionKey.
	bool isActionKey(SDLKey key);
	///	Helper method to convert an SDLKey to an ActionKey.
	KeyboardCallback::ActionKey getActionKey(SDLKey key);

	///	Our vector of KeyboardCallbacks.
	std::vector<KeyboardCallback *> callbacks;
};

}

#endif
