//	Lighting.cpp - Singleton class to handle lighting.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Lighting.h"

#include <iostream>

using namespace std;

namespace HeilanX3D
{

//------------------------------------------------------------------------------
Lighting& Lighting::getInstance()
{
	static Lighting retval;

	return retval;
}

//------------------------------------------------------------------------------
void Lighting::addDirectionalLight(const ThreeFloats& direction,
							 	   const ThreeFloats& colour,
								   float ambient)
{
	GLfloat lightDirection[4];
	GLfloat lightDiffuse[4];
	GLfloat lightAmbient[4];

	if(nextActive == 7)
	{
		cout << "Lighting: Too many lights active, unable to add DirectionalLight." << endl;
		return;
	}

	//Enable light.
	lightDirection[0] = direction.x;
	lightDirection[1] = direction.y;
	lightDirection[2] = direction.z;
	lightDirection[3] = 0.0f; //0 makes it a directional light rather than a point.
	lightDiffuse[0] = colour.x;
	lightDiffuse[1] = colour.y;
	lightDiffuse[2] = colour.z;
	lightDiffuse[3] = 1.0f;
	lightAmbient[0] = colour.x * ambient;
	lightAmbient[1] = colour.y * ambient;
	lightAmbient[2] = colour.z * ambient;
	glLightfv(lights[nextActive].id, GL_POSITION, lightDirection);
	glLightfv(lights[nextActive].id, GL_DIFFUSE, lightDiffuse);
	glLightfv(lights[nextActive].id, GL_AMBIENT, lightDiffuse);
	glEnable(lights[nextActive].id);

	//Add to activeLights.
	lights[nextActive].position = direction;
	lights[nextActive].colour = colour;
	lights[nextActive].global = false;
	activeLights.insert(make_pair(currentState, &(lights[nextActive])));
	++nextActive;
}

//------------------------------------------------------------------------------
void Lighting::addPointLight(const ThreeFloats& position,
						     const ThreeFloats& attenuation,
						     const ThreeFloats& colour,
						     float ambient)
{
	GLfloat lightDirection[4];
	GLfloat lightDiffuse[4];
	GLfloat lightAmbient[4];

	if(nextActive == 7)
	{
		cout << "Lighting: Too many lights active, unable to add PointLight." << endl;
		return;
	}

	//Enable light.
	lightDirection[0] = position.x;
	lightDirection[1] = position.y;
	lightDirection[2] = position.z;
	lightDirection[3] = 1.0f; //1 makes it a point light.
	lightDiffuse[0] = colour.x;
	lightDiffuse[1] = colour.y;
	lightDiffuse[2] = colour.z;
	lightDiffuse[3] = 1.0f;
	lightAmbient[0] = colour.x * ambient;
	lightAmbient[1] = colour.y * ambient;
	lightAmbient[2] = colour.z * ambient;
	glLightfv(lights[nextActive].id, GL_POSITION, lightDirection);
	glLightfv(lights[nextActive].id, GL_DIFFUSE, lightDiffuse);
	glLightfv(lights[nextActive].id, GL_AMBIENT, lightDiffuse);
	glLightf(lights[nextActive].id, GL_CONSTANT_ATTENUATION, attenuation.x);
	glLightf(lights[nextActive].id, GL_LINEAR_ATTENUATION, attenuation.y);
	glLightf(lights[nextActive].id, GL_QUADRATIC_ATTENUATION, attenuation.z);
	glEnable(lights[nextActive].id);

	//Add to activeLights.
	lights[nextActive].position = position;
	lights[nextActive].colour = colour;
	lights[nextActive].global = false;
	activeLights.insert(make_pair(currentState, &(lights[nextActive])));
	++nextActive;
}

//------------------------------------------------------------------------------
void Lighting::addSpotLight(const ThreeFloats& position,
						    const ThreeFloats& attenuation,
						    const ThreeFloats& colour,
						    float ambient,
							float beamWidth,
					  		float cutOffAngle,
					  		const ThreeFloats& direction)
{
	GLfloat lightPosition[4];
	GLfloat lightDiffuse[4];
	GLfloat lightAmbient[4];
	GLfloat lightDirection[3];

	if(nextActive == 7)
	{
		cout << "Lighting: Too many lights active, unable to add SpotLight." << endl;
		return;
	}

	//Enable light.
	lightPosition[0] = position.x;
	lightPosition[1] = position.y;
	lightPosition[2] = position.z;
	lightPosition[3] = 1.0f; //1 makes it a point light.
	lightDiffuse[0] = colour.x;
	lightDiffuse[1] = colour.y;
	lightDiffuse[2] = colour.z;
	lightDiffuse[3] = 1.0f;
	lightAmbient[0] = colour.x * ambient;
	lightAmbient[1] = colour.y * ambient;
	lightAmbient[2] = colour.z * ambient;
	lightDirection[0] = direction.x;
	lightDirection[1] = direction.y;
	lightDirection[2] = direction.z;
	glLightfv(lights[nextActive].id, GL_POSITION, lightPosition);
	glLightfv(lights[nextActive].id, GL_DIFFUSE, lightDiffuse);
	glLightfv(lights[nextActive].id, GL_AMBIENT, lightDiffuse);
	glLightfv(lights[nextActive].id, GL_SPOT_DIRECTION, lightDirection);
	glLightf(lights[nextActive].id, GL_CONSTANT_ATTENUATION, attenuation.x);
	glLightf(lights[nextActive].id, GL_LINEAR_ATTENUATION, attenuation.y);
	glLightf(lights[nextActive].id, GL_QUADRATIC_ATTENUATION, attenuation.z);
	glLightf(lights[nextActive].id, GL_SPOT_EXPONENT, beamWidth);
	glLightf(lights[nextActive].id, GL_SPOT_CUTOFF, cutOffAngle);
	glEnable(lights[nextActive].id);

	//Add to activeLights.
	lights[nextActive].position = position;
	lights[nextActive].colour = colour;
	lights[nextActive].global = false;
	activeLights.insert(make_pair(currentState, &(lights[nextActive])));
	++nextActive;
}

//------------------------------------------------------------------------------
void Lighting::addGlobalDirectionalLight(const ThreeFloats& direction,
							 	   		 const ThreeFloats& colour,
								   		 float ambient)
{
	GLfloat lightDirection[4];
	GLfloat lightDiffuse[4];
	GLfloat lightAmbient[4];

	if(nextActive == 7)
	{
		cout << "Lighting: Too many lights active, unable to add global DirectionalLight." << endl;
		return;
	}

	//Enable light.
	lightDirection[0] = direction.x;
	lightDirection[1] = direction.y;
	lightDirection[2] = direction.z;
	lightDirection[3] = 0.0f; //0 makes it a directional light rather than a point.
	lightDiffuse[0] = colour.x;
	lightDiffuse[1] = colour.y;
	lightDiffuse[2] = colour.z;
	lightDiffuse[3] = 1.0f;
	lightAmbient[0] = colour.x * ambient;
	lightAmbient[1] = colour.y * ambient;
	lightAmbient[2] = colour.z * ambient;
	glLightfv(lights[nextActive].id, GL_POSITION, lightDirection);
	glLightfv(lights[nextActive].id, GL_DIFFUSE, lightDiffuse);
	glLightfv(lights[nextActive].id, GL_AMBIENT, lightDiffuse);
	glEnable(lights[nextActive].id);

	//Add to activeLights.
	lights[nextActive].position = direction;
	lights[nextActive].colour = colour;
	lights[nextActive].global = true;
	activeLights.insert(make_pair(currentState, &(lights[nextActive])));
	++nextActive;
}

//------------------------------------------------------------------------------
void Lighting::addGlobalPointLight(const ThreeFloats& position,
								   const ThreeFloats& attenuation,
								   const ThreeFloats& colour,
								   float ambient)
{
	GLfloat lightDirection[4];
	GLfloat lightDiffuse[4];
	GLfloat lightAmbient[4];

	if(nextActive == 7)
	{
		cout << "Lighting: Too many lights active, unable to add global PointLight." << endl;
		return;
	}

	//Enable light.
	lightDirection[0] = position.x;
	lightDirection[1] = position.y;
	lightDirection[2] = position.z;
	lightDirection[3] = 1.0f; //1 makes it a point light.
	lightDiffuse[0] = colour.x;
	lightDiffuse[1] = colour.y;
	lightDiffuse[2] = colour.z;
	lightDiffuse[3] = 1.0f;
	lightAmbient[0] = colour.x * ambient;
	lightAmbient[1] = colour.y * ambient;
	lightAmbient[2] = colour.z * ambient;
	glLightfv(lights[nextActive].id, GL_POSITION, lightDirection);
	glLightfv(lights[nextActive].id, GL_DIFFUSE, lightDiffuse);
	glLightfv(lights[nextActive].id, GL_AMBIENT, lightDiffuse);
	glLightf(lights[nextActive].id, GL_CONSTANT_ATTENUATION, attenuation.x);
	glLightf(lights[nextActive].id, GL_LINEAR_ATTENUATION, attenuation.y);
	glLightf(lights[nextActive].id, GL_QUADRATIC_ATTENUATION, attenuation.z);
	glEnable(lights[nextActive].id);

	//Add to activeLights.
	lights[nextActive].position = position;
	lights[nextActive].colour = colour;
	lights[nextActive].global = true;
	activeLights.insert(make_pair(currentState, &(lights[nextActive])));
	++nextActive;
}

//------------------------------------------------------------------------------
void Lighting::addGlobalSpotLight(const ThreeFloats& position,
								  const ThreeFloats& attenuation,
								  const ThreeFloats& colour,
								  float ambient,
								  float beamWidth,
								  float cutOffAngle,
								  const ThreeFloats& direction)
{
	GLfloat lightPosition[4];
	GLfloat lightDiffuse[4];
	GLfloat lightAmbient[4];
	GLfloat lightDirection[3];

	if(nextActive == 7)
	{
		cout << "Lighting: Too many lights active, unable to add global SpotLight." << endl;
		return;
	}

	//Enable light.
	lightPosition[0] = position.x;
	lightPosition[1] = position.y;
	lightPosition[2] = position.z;
	lightPosition[3] = 1.0f; //1 makes it a point light.
	lightDiffuse[0] = colour.x;
	lightDiffuse[1] = colour.y;
	lightDiffuse[2] = colour.z;
	lightDiffuse[3] = 1.0f;
	lightAmbient[0] = colour.x * ambient;
	lightAmbient[1] = colour.y * ambient;
	lightAmbient[2] = colour.z * ambient;
	lightDirection[0] = direction.x;
	lightDirection[1] = direction.y;
	lightDirection[2] = direction.z;
	glLightfv(lights[nextActive].id, GL_POSITION, lightPosition);
	glLightfv(lights[nextActive].id, GL_DIFFUSE, lightDiffuse);
	glLightfv(lights[nextActive].id, GL_AMBIENT, lightDiffuse);
	glLightfv(lights[nextActive].id, GL_SPOT_DIRECTION, lightDirection);
	glLightf(lights[nextActive].id, GL_CONSTANT_ATTENUATION, attenuation.x);
	glLightf(lights[nextActive].id, GL_LINEAR_ATTENUATION, attenuation.y);
	glLightf(lights[nextActive].id, GL_QUADRATIC_ATTENUATION, attenuation.z);
	glLightf(lights[nextActive].id, GL_SPOT_EXPONENT, beamWidth);
	glLightf(lights[nextActive].id, GL_SPOT_CUTOFF, cutOffAngle);
	glEnable(lights[nextActive].id);

	//Add to activeLights.
	lights[nextActive].position = position;
	lights[nextActive].colour = colour;
	lights[nextActive].global = true;
	activeLights.insert(make_pair(currentState, &(lights[nextActive])));
	++nextActive;
}

//------------------------------------------------------------------------------
void Lighting::pushLighting()
{
	++currentState;
}

//------------------------------------------------------------------------------
void Lighting::popLighting()
{
	LightMap::iterator it;

	for(it=activeLights.lower_bound(currentState);
		it!=activeLights.upper_bound(currentState);
		++it)
	{
		if(!it->second->global)
		{
			//Disable light.
			glDisable(it->second->id);
			--nextActive;
		}
	}

	activeLights.erase(currentState);
	--currentState;
}

//------------------------------------------------------------------------------
void Lighting::switchHeadlight(bool val)
{
	headlight = val;
}

//------------------------------------------------------------------------------
void Lighting::drawHeadlight()
{
	if(headlight)
		glEnable(GL_LIGHT0);
	else
		glDisable(GL_LIGHT0);
}

//------------------------------------------------------------------------------
Lighting::Lighting():
currentState(0),
nextActive(0),
headlight(true)
{
	headlight = true;
	lights[0].position = ThreeFloats(0.0f, 0.0f, 0.0f);
	lights[0].colour = ThreeFloats(1.0f, 1.0f, 1.0f);
	lights[0].id = GL_LIGHT1;

	lights[1].position = ThreeFloats(0.0f, 0.0f, 0.0f);
	lights[1].colour = ThreeFloats(1.0f, 1.0f, 1.0f);
	lights[1].id = GL_LIGHT2;

	lights[2].position = ThreeFloats(0.0f, 0.0f, 0.0f);
	lights[2].colour = ThreeFloats(1.0f, 1.0f, 1.0f);
	lights[2].id = GL_LIGHT3;

	lights[3].position = ThreeFloats(0.0f, 0.0f, 0.0f);
	lights[3].colour = ThreeFloats(1.0f, 1.0f, 1.0f);
	lights[3].id = GL_LIGHT4;

	lights[4].position = ThreeFloats(0.0f, 0.0f, 0.0f);
	lights[4].colour = ThreeFloats(1.0f, 1.0f, 1.0f);
	lights[4].id = GL_LIGHT5;

	lights[5].position = ThreeFloats(0.0f, 0.0f, 0.0f);
	lights[5].colour = ThreeFloats(1.0f, 1.0f, 1.0f);
	lights[5].id = GL_LIGHT6;

	lights[6].position = ThreeFloats(0.0f, 0.0f, 0.0f);
	lights[6].colour = ThreeFloats(1.0f, 1.0f, 1.0f);
	lights[6].id = GL_LIGHT7;
}

//------------------------------------------------------------------------------
Lighting::~Lighting()
{
	
}

}
