//	OpenSoundController.cpp - Singleton class handling OSC input.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "OpenSoundController.h"
#include "AbstractNode.h"
#include "Camera.h"
#include "StringParser.h"
#include "AttributeHelper.h"
#include "OSCMessage.h"
#include "OSCBundle.h"
#include "OscRecorder.h"

#include <sstream>
#include <iostream>

using namespace OSC;
using namespace std;

namespace HeilanX3D
{

//------------------------------------------------------------------------------
OpenSoundController *OpenSoundController::getInstance()
{
	static OpenSoundController retval;

	return &retval;
}

//------------------------------------------------------------------------------
void OpenSoundController::setMulticastGroup(const std::string& address)
{
	sock.setMulticastGroup(address);
}

//------------------------------------------------------------------------------
void OpenSoundController::setPort(int newPort)
{
	port = newPort;
	sock.setPort(port);
}

//------------------------------------------------------------------------------
void OpenSoundController::setOutDetails(const string& address, int newPort)
{
	outSock.setAddress(address);
	outPort = newPort;
	outSock.setPort(outPort);
}

//------------------------------------------------------------------------------
void OpenSoundController::startListening()
{
	startThread(this);
}

//------------------------------------------------------------------------------
void OpenSoundController::stopListening()
{
	timeToStop = true;
}

//------------------------------------------------------------------------------
void OpenSoundController::registerNode(AbstractNode *node)
{
	nodes.insert(make_pair(node->getDefName(), node));
}

//------------------------------------------------------------------------------
bool OpenSoundController::run()
{
	int i;
	Bundle *recBundle;
	Message *recMessage;
	int recursiveSize;
	char *recursiveData;
	long receivedSize = -1;
	char *receivedData = 0;

	sock.bindSocket();

	while(!timeToStop)
	{
		//Listen for OSC packets.
		receivedData = sock.getData(receivedSize);

		if((receivedSize > -1) && (receivedData))
		{
			//If we've received an OSC message.
			if(Message::isMessage(receivedData, receivedSize))
			{
				recMessage = new Message(receivedData, receivedSize);
				handleOSCMessage(recMessage);
				delete recMessage;
			}
			//If we've received an OSC bundle.
			//Note: we don't do anything about timestamps...
			else if(Bundle::isBundle(receivedData, receivedSize))
			{
				recBundle = new OSC::Bundle(receivedData, receivedSize);

				for(i=0;i<recBundle->getNumMessages();++i)
				{
					recursiveSize = recBundle->getMessage(i)->getSize();
					recursiveData = recBundle->getMessage(i)->getData();

					recMessage = new Message(recursiveData, recursiveSize);
					handleOSCMessage(recMessage);
					delete recMessage;
				}
				delete recBundle;
			}
			else
			{
				cout << "Received non-OSC packet: " << endl;
				for(i=0;i<receivedSize;++i)
					cout << receivedData[i];
				cout << endl;
			}
		}
	}

	return true;
}

//------------------------------------------------------------------------------
void OpenSoundController::sendOSCMessage(OSC::Message& message)
{
	OSC::Int32 dataSize;

	dataSize = message.getSize();
	outSock.sendData(message.getData(), dataSize);
}

//------------------------------------------------------------------------------
OpenSoundController::OpenSoundController():
Thread(),
port(5678),
timeToStop(false)
{
	sock.setPort(port);
}

//------------------------------------------------------------------------------
OpenSoundController::~OpenSoundController()
{
	stopListening();
}

//------------------------------------------------------------------------------
void OpenSoundController::handleOSCMessage(OSC::Message *message,
										   bool fromCamera)
{
	unsigned int i;
	string tempstr;
	string typeTag;
	string tempParam;
	ostringstream tempConverter;
	int floatCount, intCount, stringCount;
	multimap<string, AbstractNode *>::iterator endIt;
	multimap<string, AbstractNode *>::iterator tempIt;
	multimap<string, AbstractNode *>::iterator startIt;

	tempstr = message->getAddress();
	if(isHeilanAddress(tempstr))
	{
		tempParam = getAttributeFromOSC(tempstr);
		tempstr = getAddressFromOSC(tempstr);

		//First check if it's a message for the camera.
		if(tempstr == "camera")
		{
			string realAtt = AttributeHelperBase::getAddress(tempParam);

			if(realAtt == "position")
			{
				int floatCount = 0;
				int intCount = 0;
				int stringCount = 0;
				ThreeFloats position = Camera::getInstance()->getGraphicPosition();
				AttributeHelper<ThreeFloats> helper(&position);

				typeTag = message->getTypeTag();
				for(i=0;i<typeTag.length();++i)
				{
					switch(typeTag[i])
					{
						case 'f':
							if(floatCount < message->getNumFloats())
							{
								tempConverter << message->getFloat(floatCount) << " ";
								++floatCount;
							}
							break;
						case 'i':
							if(floatCount < message->getNumInts())
							{
								tempConverter << message->getInt(intCount) << " ";
								++intCount;
							}
							break;
						case 's':
							if(floatCount < message->getNumStrings())
							{
								tempConverter << message->getString(stringCount) << " ";
								++intCount;
							}
							break;
					}
				}

				StringParser parser(tempConverter.str());

				if(helper.isValid(tempParam))
					helper.convert(tempParam, tempConverter.str());
				else if(parser.getNumThreeFloats() > 0)
					position = parser.getThreeFloats();

				if(!fromCamera)
					Camera::getInstance()->setGraphicPosition(position);
			}
			else if(realAtt == "rotation")
			{
				floatCount = 0;
				intCount = 0;
				stringCount = 0;
				Quaternion rotation = Camera::getInstance()->getGraphicRotation();
				AttributeHelper<Quaternion> helper(&rotation);

				typeTag = message->getTypeTag();
				for(i=0;i<typeTag.length();++i)
				{
					switch(typeTag[i])
					{
						case 'f':
							if(floatCount < message->getNumFloats())
							{
								tempConverter << message->getFloat(floatCount) << " ";
								++floatCount;
							}
							break;
						case 'i':
							if(floatCount < message->getNumInts())
							{
								tempConverter << message->getInt(intCount) << " ";
								++intCount;
							}
							break;
						case 's':
							if(floatCount < message->getNumStrings())
							{
								tempConverter << message->getString(stringCount) << " ";
								++intCount;
							}
							break;
					}
				}

				StringParser parser(tempConverter.str());

				if(helper.isValid(tempParam))
					helper.convert(tempParam, tempConverter.str());
				else if(parser.getNumQuaternions() > 0)
					rotation = parser.getQuaternion();

				if(!fromCamera)
					Camera::getInstance()->setGraphicRotation(rotation);
			}
			if(fromCamera)
			{
				//Write to OscRecorder.
				OscRecorder::getInstance()->writeMessage(tempstr,
														 tempParam,
														 tempConverter.str());
			}
			return;
		}

		floatCount = 0;
		intCount = 0;
		stringCount = 0;

		typeTag = message->getTypeTag();
		for(i=0;i<typeTag.length();++i)
		{
			switch(typeTag[i])
			{
				case 'f':
					if(floatCount < message->getNumFloats())
					{
						tempConverter << message->getFloat(floatCount) << " ";
						++floatCount;
					}
					break;
				case 'i':
					if(floatCount < message->getNumInts())
					{
						tempConverter << message->getInt(intCount) << " ";
						++intCount;
					}
					break;
				case 's':
					if(floatCount < message->getNumStrings())
					{
						tempConverter << message->getString(stringCount) << " ";
						++intCount;
					}
					break;
			}
		}

		//Look for nodes with the tempstr address.
		startIt = nodes.lower_bound(tempstr);
		endIt = nodes.upper_bound(tempstr);
		for(tempIt=startIt;tempIt!=endIt;++tempIt)
		{
			string attName(tempParam);
			EventValue attVal(tempConverter.str());
			tempIt->second->setAndSendAttribute(attName, attVal);
		}

		//Write to OscRecorder.
		OscRecorder::getInstance()->writeMessage(tempstr,
												 tempParam,
												 tempConverter.str());
	}
}

//------------------------------------------------------------------------------
bool OpenSoundController::isHeilanAddress(const std::string& addr)
{
	bool retval = false;

	if(addr.find("/heilan/", 0, 8) != string::npos)
		retval = true;

	return retval;
}

//------------------------------------------------------------------------------
string OpenSoundController::getAddressFromOSC(const std::string& addr)
{
	int tempint;
	string retval;
	string tempstr;

	//Skip the initial "/heilan/".
	tempstr = addr.substr(8, (addr.size()-8));

	tempint = tempstr.find_first_of('/', 0);
	retval = tempstr.substr(0, tempint);

	return retval;
}

//------------------------------------------------------------------------------
string OpenSoundController::getAttributeFromOSC(const std::string& addr)
{
	int tempint;
	string retval;
	string tempstr;

	//Skip the initial "/heilan/".
	tempstr = addr.substr(8, (addr.size()-8));

	tempint = tempstr.find_first_of('/', 0);
	retval = tempstr.substr((tempint+1), tempstr.size()-(tempint+1));

	return retval;
}

}
