//	OpenSoundController.h - Singleton class handling OSC input.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef OPENSOUNDCONTROLLER_H_
#define OPENSOUNDCONTROLLER_H_

//Stupid MSVC v6 warning about the STL...
#ifdef WIN32
#pragma warning(disable:4786)
#endif

#include "Thread.h"
#include "UDPSocket.h"

#include <map>
#include <string>

namespace OSC
{
	class Message;
}

namespace HeilanX3D
{

class AbstractNode;

///	Singleton class handling OSC input.
/*!
	Runs a separate thread listening on the relevant port for OSC messages.
	When a message is received, the class goes through it's multimap looking for
	a node with the relevant address, and then passes the message onto that node
	(or nodes) via AbstractNode::setAttribute().

	Heilan OSC addresses have a common root ("/heilan"), and do not get nested.
	i.e. the address to the pitch parameter of an AudioClip node with an OSC
	name of "bob" would be: "/heilan/bob/pitch".  There should never be more
	than 3 slashes.

	\todo Handle timestamps for bundles.
 */
class dllStuff OpenSoundController : public Thread
{
  public:
	///	Returns the sole instance of OpenSoundController.
	static OpenSoundController *getInstance();

	///	Called to set the multicast group address to listen on.
	void setMulticastGroup(const std::string& address);
	///	Called to set the port to listen on (5678 by default).
	void setPort(int newPort);
	///	Called to set the address and port to send messages to.
	void setOutDetails(const std::string& address, int newPort);

	///	Starts listening for OSC messages on the specified port.
	void startListening();
	///	Stops listening for OSC messages.
	void stopListening();

	///	Called by nodes when they are assigned an OSC name.
	/*!
		This adds the node to our multimap, so we can quickly find them when we
		receive an appropriate OSC message.
	 */
	void registerNode(AbstractNode *node);

	///	The thread loop.
	bool run();

	///	Sends an OSC message to another OSC server.
	void sendOSCMessage(OSC::Message& message);

	///	Passes on an OSC message to all the relevant nodes.
	/*!
		\param message The message.
		\param fromCamera True if Camera has called this method directly.  We
		use this to be able to record camera movements as OSC messages.

		This should be private, but we use it in the OscPlayback node as well
		as run().
	 */
	void handleOSCMessage(OSC::Message *message, bool fromCamera=false);

#ifdef WIN32
	///	Stupid fix for MSVC v6.
	~OpenSoundController();
#endif
  private:
	///	Constructor.
	OpenSoundController();
#ifndef WIN32
	///	Destructor.
	~OpenSoundController();
#endif

	///	Returns tru if addr is a heilan OSC address.
	/*!
		Just checks if the first section of the address is "/heilan/"
	 */
	bool isHeilanAddress(const std::string& addr);
	///	Returns the node address in an OSC address.
	/*!
		i.e. we'd receive /heilan/bill/location_x ; this method returns bill.
	 */
	std::string getAddressFromOSC(const std::string& addr);
	///	Returns the attribute address in an OSC address.
	/*!
		i.e. we'd receive /heilan/bill/location_x ; this method returns
		location_x.
	 */
	std::string getAttributeFromOSC(const std::string& addr);

	///	Our multimap of nodes and OSC addresses.
	std::multimap<std::string, AbstractNode *> nodes;

	///	The socket we listen on.
	UDPSocket sock;
	///	Port to listen on.
	int port;

	///	The socket we're sending to.
	UDPSocket outSock;
	///	The port we're sending to.
	int outPort;

	///	Whether it's time to stop yet.
	/*!
		I'm assuming this will always be accessed atomically, so we don't need
		to surround it with a critical section etc.
	 */
	bool timeToStop;
};

}

#endif
