//	ThreadManager.cpp - Singleton class handling additional threads.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "ThreadManager.h"

#include "SDL.h"
#include <iostream>
#include <algorithm>

using namespace std;

namespace HeilanX3D
{

//------------------------------------------------------------------------------
ThreadManager& ThreadManager::getInstance()
{
	static ThreadManager instance;

	return instance;
}

//------------------------------------------------------------------------------
void ThreadManager::initialiseThread(ThreadObject *node, int delay)
{
	bool found = false;
	vector<ManagedThread *>::iterator it;

	for(it=threads.begin();it!=threads.end();++it)
	{
		if((*it)->getType() == node->getNodeType())
		{
			found = true;

			(*it)->addThreadObject(node);

			break;
		}
	}

	if(!found)
	{
		ManagedThread *thread = new ManagedThread(node->getNodeType(), delay);
		thread->addThreadObject(node);
		threads.push_back(thread);
	}
}

//------------------------------------------------------------------------------
void ThreadManager::removeNode(ThreadObject *node)
{
	vector<ManagedThread *>::iterator it;

	for(it=threads.begin();it!=threads.end();++it)
	{
		if((*it)->getType() == node->getNodeType())
		{
			(*it)->removeThreadObject(node);
			break;
		}
	}
}

//------------------------------------------------------------------------------
void ThreadManager::startThreads()
{
	vector<ManagedThread *>::iterator it;

	for(it=threads.begin();it!=threads.end();++it)
		(*it)->startThread(*it);
}

//------------------------------------------------------------------------------
void ThreadManager::stopThreads()
{
	vector<ManagedThread *>::iterator it;

	for(it=threads.begin();it!=threads.end();++it)
		(*it)->stopThread();
}

//------------------------------------------------------------------------------
ThreadManager::ThreadManager()
{

}

//------------------------------------------------------------------------------
ThreadManager::~ThreadManager()
{
	vector<ManagedThread *>::iterator it;

	stopThreads();

	for(it=threads.begin();it!=threads.end();++it)
		delete (*it);

}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
ManagedThread::ManagedThread(const string &type, int delay):
Thread(),
nodeType(type),
loopDelay(delay),
timeToStop(false)
{
	
}

//------------------------------------------------------------------------------
ManagedThread::~ManagedThread()
{
	
}

//------------------------------------------------------------------------------
bool ManagedThread::run()
{
	vector<ThreadObject *>::iterator it;

	try
	{
		while(!timeToStop)
		{
			critSec.lock();
			{
				for(it=nodes.begin();it!=nodes.end();++it)
					(*it)->threadIteration();
			}
			critSec.unlock();

			if(loopDelay > 0)
				SDL_Delay(loopDelay);
		}
	}
	catch(...)
	{
		cout << "Thread for " << getType() << " crashed." << endl;
		return false;
	}

	return true;
}

//------------------------------------------------------------------------------
void ManagedThread::addThreadObject(ThreadObject *node)
{
	critSec.lock();
	{
		nodes.push_back(node);
	}
	critSec.unlock();
}

//------------------------------------------------------------------------------
void ManagedThread::removeThreadObject(ThreadObject *node)
{
	critSec.lock();
	{
		vector<ThreadObject *>::iterator it;

		it = find(nodes.begin(), nodes.end(), node);
		if(it != nodes.end())
			nodes.erase(it);
	}
	critSec.unlock();
}

//------------------------------------------------------------------------------
void ManagedThread::stopThread()
{
	timeToStop = true;

	if(loopDelay > 0)
		SDL_Delay(loopDelay);
}

}
