//	ThreadManager.h - Singleton class handling additional threads.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef THREADMANAGER_H_
#define THREADMANAGER_H_

#include "Thread.h"
#include "DllStuff.h"
#include "CriticalSection.h"

#include <vector>
#include <string>

namespace HeilanX3D
{

class ThreadObject;
class ManagedThread;

//------------------------------------------------------------------------------
///	Singleton class handling additional threads.
/*!
	Some node types (AudioClip, MovieTexture) require additional threads in
	order to stream data off the disk.  This class is used to manage these
	threads.

	For a node type to have access to a thread, it must derive from the
	ThreadObject abstract class, and implement the
	ThreadObject::threadIteration() method.  This method will be called
	repeatedly by the ThreadManager, with the user able to specify how often.
	Note that there is <em>one</em> thread per node type, not per instance of
	the node type.  As such, the threadIteration() method of each instance of a
	ThreadObject-derived node type will be called in succession, before the
	thread returns to the start of its loop and repeats the process.

	In order to setup a thread for a particular node type,
	ThreadManager::initialiseThread() must be called in the node's
	AbstractNode::initialise() method.  This will ensure a new thread is created
	for this node type if it does not already exist, and add this instance of
	the node to the list of nodes whose AbstractNode::threadIteration() method
	should be called.

	The ThreadObject class calls ThreadManager::removeNode() in its destructor
	to ensure a node does not get accessed once it has been destroyed.  There is
	no harm in calling this method more than once, if you would rather call it
	elsewhere.
 */
class dllStuff ThreadManager
{
  public:
	///	Returns the sole instance of ThreadManager.
	static ThreadManager& getInstance();

	///	Sets up a thread for a particular node type.
	/*!
		\param node Pointer to the ThreadObject to call for every iteration of
		the thread's loop.
		\param delay The number of ms to wait at the end of each loop iteration.
	 */
	void initialiseThread(ThreadObject *node, int delay);
	///	Removes a ThreadObject from the list of objects called for each iteration.
	void removeNode(ThreadObject *node);

	///	Starts all the threads running (should only be called once).
	void startThreads();
	///	Stops all the threads running.
	void stopThreads();

#ifdef WIN32
	///	Stupid fix for MSVC v6.
	~ThreadManager();
#endif

  private:
	///	Constructor.
	ThreadManager();
#ifndef WIN32
	///	Destructor.
	~ThreadManager();
#endif

	///	Contains all the threads.
	std::vector<ManagedThread *> threads;
};

//------------------------------------------------------------------------------
///	A single thread managed by ThreadManager.
class ManagedThread : public Thread
{
  public:
	///	Constructor.
	ManagedThread(const std::string &type, int delay);
	///	Destructor.
	~ManagedThread();

	///	The thread's loop.
	bool run();

	///	Adds a ThreadObject to our list of objects to be called.
	void addThreadObject(ThreadObject *node);
	///	Removes a ThreadObject from our list of objects to be called.
	void removeThreadObject(ThreadObject *node);

	///	Stops the thread.
	void stopThread();

	///	Returns which type of node this thread is for.
	const std::string& getType() const {return nodeType;};
  private:
	///	All the nodes whose threadIteration() method should be called.
	std::vector<ThreadObject *> nodes;
	///	Which type of node this thread is for.
	std::string nodeType;
	///	The length of time to wait at the end of the loop.
	int loopDelay;

	///	Signal used to stop the thread.
	bool timeToStop;

	///	To protect the thread from adding/removing ThreadObjects.
	CriticalSection critSec;
};

//------------------------------------------------------------------------------
///	Abstract base class for any node types which need an additional thread.
/*!
	\sa ThreadManager
 */
class dllStuff ThreadObject
{
  public:
	///	Constructor.
	/*!
		\param nodeType Which node type is deriving from this class (e.g. AudioClip,
		MovieTexture etc.).
	 */
	ThreadObject(const std::string& nodeType):
	type(nodeType)
	{
		
	};
	///	Destructor.
	virtual ~ThreadObject()
	{
		ThreadManager::getInstance().removeNode(this);
	};

	///	Called repeatedly in a separate thread by ThreadManager.
	virtual void threadIteration() = 0;

	///	Returns this ThreadObject's type.
	const std::string& getNodeType() const {return type;};
  private:
	///	This ThreadObject's type.
	std::string type;
};

}

#endif
