//	X3DFileLoader.cpp - Class to parse an X3D file and create a scene from it.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "X3DFileLoader.h"
#include "AbstractNode.h"
#include "ParentNode.h"
#include "Metadata.h"
#include "SceneConstructor.h"
#include "NodeFactory.h"
#include "FatalError.h"
#include "FileFinder.h"

#include "tinyxml.h"
#include <iostream>
#include <sstream>
#include <algorithm>

using namespace std;

namespace HeilanX3D
{

//------------------------------------------------------------------------------
X3DFileLoader::X3DFileLoader(const string& filepath):
file(filepath),
sceneRoot(0)
{
	sceneConstructor = new SceneConstructor;
}

//------------------------------------------------------------------------------
X3DFileLoader::~X3DFileLoader()
{
	delete sceneConstructor;
}

//------------------------------------------------------------------------------
bool X3DFileLoader::isValidX3DFile() const
{
	string tempPath;
	bool retval = false;

	tempPath = FileFinder::getInstance().findFile(file);
	if(tempPath != "")
	{
		TiXmlDocument doc(tempPath);
		TiXmlHandle handle(&doc);

		if(doc.LoadFile())
		{
			//Only need to check that these two nodes are present really.
			if(handle.FirstChild("X3D").FirstChild("Scene").Element())
				retval = true;
			else
				cout << "X3DFileLoader: Could not find standard X3D->Scene->... node heirarchy." << endl;
		}
	}
	else
		cout << "X3DFileLoader: " << file << " does not seem to exist?" << endl;

	return retval;
}

//------------------------------------------------------------------------------
AbstractNode *X3DFileLoader::loadScene(bool mainFile)
{
	string tempPath;

	AbstractNode *retval = 0;
	TiXmlElement *tempRoot = 0;

	if(mainFile)
		FileFinder::getInstance().clearStoredPaths();
	tempPath = FileFinder::getInstance().findFile(file, mainFile);
	if(tempPath != "")
	{
		TiXmlDocument doc(tempPath);
		TiXmlHandle handle(&doc);

		//Load file, pass to parseNode().
		if(doc.LoadFile())
		{
			//In case we haven't checked it's a valid file with isValidX3DFile()
			//first...
			if((tempRoot = handle.FirstChild("X3D").FirstChild("Scene").Element()))
				retval = parseNode(tempRoot);
		}
		else
		{
			ostringstream err;
			err << "Could not load file: " << tempPath;
			throw FatalError(err.str());
		}

		sceneRoot = retval;
	}

	return retval;
}

//------------------------------------------------------------------------------
RoutedNodes *X3DFileLoader::getRoutedNodes()
{
	string tempPath;
	TiXmlElement *tempRoot = 0;
	RoutedNodes *retval = new RoutedNodes;

	tempPath = FileFinder::getInstance().findFile(file);
	if(tempPath != "")
	{
		TiXmlDocument doc(tempPath);
		TiXmlHandle handle(&doc);

		//Load file, pass to parseRoute().
		if(doc.LoadFile())
		{
			//In case we haven't checked it's a valid file with isValidX3DFile()
			//first...
			if((tempRoot = handle.FirstChild("X3D").FirstChild("Scene").Element()))
				parseRoute(tempRoot, retval);
		}
		else
		{
			ostringstream err;
			err << "getRoutedNodes(): Could not load file: " << tempPath;
			throw FatalError(err.str());
		}
	}

	//In case there were no ROUTEs.
	if(retval->size () < 1)
	{
		delete retval;
		retval = 0;
	}

	return retval;
}

//------------------------------------------------------------------------------
void X3DFileLoader::setupEXPORTNames(const string& inlineName)
{
	string tempPath;
	TiXmlElement *tempRoot = 0;

	tempPath = FileFinder::getInstance().findFile(file);
	if(tempPath != "")
	{
		TiXmlDocument doc(tempPath);
		TiXmlHandle handle(&doc);

		//Load file, pass to parseRoute().
		if(doc.LoadFile())
		{
			//In case we haven't checked it's a valid file with isValidX3DFile()
			//first...
			if((tempRoot = handle.FirstChild("X3D").FirstChild("Scene").Element()))
			{
				vector<AbstractNode *> savedNodes;

				parseExport(tempRoot, savedNodes, inlineName);
				clearNames(sceneRoot, savedNodes);
			}
		}
		else
			cout << "setupEXPORTNames(): Could not find " << tempPath << endl;
	}
}

//------------------------------------------------------------------------------
void X3DFileLoader::handleIMPORTS()
{
	string tempPath;
	TiXmlElement *tempRoot = 0;

	tempPath = FileFinder::getInstance().findFile(file);
	if(tempPath != "")
	{
		TiXmlDocument doc(tempPath);
		TiXmlHandle handle(&doc);

		//Load file, pass to parseRoute().
		if(doc.LoadFile())
		{
			//In case we haven't checked it's a valid file with isValidX3DFile()
			//first...
			if((tempRoot = handle.FirstChild("X3D").FirstChild("Scene").Element()))
				parseImport(tempRoot);
		}
		else
			cout << "handleIMPORTS(): Could not find " << tempPath << endl;
	}
}

//------------------------------------------------------------------------------
AbstractNode *X3DFileLoader::parseNode(TiXmlElement *node, AbstractNode *parent)
{
	string tempstr;
	AbstractNode *newNode = 0;
	TiXmlAttribute *tempAtt = 0;
	TiXmlNode *tempChild = 0;
	TiXmlElement *tempChild2 = 0;

	//First check if this is a USE node.
	if((node->Attribute("USE")) && parent)
	{
		vector<ParentNode *>& parentNode = parent->getParent();
		AbstractNode *rootNode = 0;
		AbstractNode *tempNode = 0;

		//Get root node.
		rootNode = parent;
		parentNode = rootNode->getParent();
		while(parentNode.size() > 0)
		{
			rootNode = parentNode[0];
			parentNode = rootNode->getParent();

			//Because the root Scene node has a parent of 0.
			if(parentNode[0] == 0)
				break;
		}

		tempNode = rootNode->findNode(node->Attribute("USE"));

		if(tempNode)
		{
			dynamic_cast<ParentNode *>(parent)->addNode(tempNode, tempNode->shouldBeStart());
			tempNode->addParent(dynamic_cast<ParentNode *>(parent));
			tempNode->incRefCount();
		}

		return parent;
	}

	//Construct node from the X3D file.
	newNode = NodeFactory::getInstance().createNode(node->Value());
	if(!newNode)
	{
		tempstr = node->Value();
		if((tempstr != "ROUTE") &&
		   (tempstr != "IMPORT") &&
		   (tempstr != "EXPORT"))
		{
			cout << "Could not create node: " << tempstr << endl;
		}
		return 0;
	}
	newNode->addParent(static_cast<ParentNode *>(parent));

	if(newNode)
	{
		//If the node's not a metadata node (or a ROUTE node, which isn't really
		//a node anyway).
		if(newNode->getType() != "MetadataString")
		{
			//Add it to the scene.
			//If there's not a current parent, it means this is the root Scene
			//node, and the parent will be set in sceneConstructionEnter(),
			//below.
			if(sceneConstructor->getCurrentParent())
				sceneConstructor->getCurrentParent()->addNode(newNode, newNode->shouldBeStart());

			//Initialise it before we set any attributes.
			newNode->initialise();

			//Set the node's attributes.
			//We do this after the node's been added to the scene here, because
			//certain nodes (e.g. Sound) need to communicate with their parents
			//in setAttribute().
			tempAtt = node->FirstAttribute();
			while(tempAtt)
			{
				//If the attribute is DEF, we handle it differently.
				tempstr = tempAtt->Name();
				if(tempstr == "DEF")
					newNode->setDefName(tempAtt->Value());
				else
				{
					string attName(tempAtt->Name());
					EventValue attVal(tempAtt->Value());

					newNode->setAndSendAttribute(attName, attVal);
				}

				tempAtt = tempAtt->Next();
			}

			//Push node onto the stack, if it's a ParentNode.
			newNode->sceneConstructionEnter(sceneConstructor);

			//Go through all this node's children, if they exist.
			while((tempChild = node->IterateChildren(tempChild)))
			{
				tempChild2 = tempChild->ToElement();
				if(tempChild2)
					parseNode(tempChild2, newNode);
			}

			//Pop node off the stack, if it's a ParentNode.
			newNode->sceneConstructionLeave(sceneConstructor);
		}
		else //It's a metadata node, which we treat differently.
		{
			Metadata *meta = static_cast<Metadata *>(newNode);

			//Set the node's attributes.
			tempAtt = node->FirstAttribute();
			while(tempAtt)
			{
				string attName(tempAtt->Name());
				EventValue attVal(tempAtt->Value());

				newNode->setAndSendAttribute(attName, attVal);

				tempAtt = tempAtt->Next();
			}

			//If it's setting the node's OSC name.
			/*if(meta->getName() == "heilan_oscName")
			{
				if(parent)
					parent->setOscName(meta->getValue());
			}
			else //Otherwise, inform the parent about the metadata.*/
			{
				if(parent)
					parent->setMetadata(meta->getName(), meta->getValue());
			}

			//Delete newNode, since we won't use it in the actual scene.
			delete newNode;

			return 0;
		}
	}

	return newNode;
}

//------------------------------------------------------------------------------
void X3DFileLoader::parseRoute(TiXmlElement *node, RoutedNodes *routedNodes)
{
	string tempstr;
	string fromNode;
	string fromAttribute;
	string toNode;
	string toAttribute;
	TiXmlAttribute *tempAtt = 0;
	TiXmlNode *tempChild = 0;
	TiXmlElement *tempChild2 = 0;

	//If it's a ROUTE, inform the relevant nodes, and add it to routedNodes.
	tempstr = node->Value();
	if(tempstr == "ROUTE")
	{
		//Go through attributes, setting the appropriate values.
		tempAtt = node->FirstAttribute();
		while(tempAtt)
		{
			tempstr = tempAtt->Name();
			if(tempstr == "fromNode")
				fromNode = tempAtt->Value();
			else if(tempstr == "fromField")
				fromAttribute = tempAtt->Value();
			else if(tempstr == "toNode")
				toNode = tempAtt->Value();
			else if(tempstr == "toField")
				toAttribute = tempAtt->Value();

			tempAtt = tempAtt->Next();
		}

		//Check that none of the relevant strings are empty.
		if(!fromNode.empty() &&
		   !fromAttribute.empty() &&
		   !toNode.empty() &&
		   !toAttribute.empty())
		{
			AbstractNode *tempFromNode = sceneRoot->findNode(fromNode);
			AbstractNode *tempToNode = sceneRoot->findNode(toNode);

			//Make sure we have those nodes in our scene.
			//if(!((!tempFromNode) || (!tempToNode)))
			if(tempFromNode && tempToNode)
			{
				//Set up the route.
				tempFromNode->addRoute(fromAttribute, tempToNode, toAttribute);

				//Add the nodes to routedNodes.
				routedNodes->push_back(tempFromNode);
				routedNodes->push_back(tempToNode);
			}
			else if(!tempFromNode)
				cout << "ROUTE: Could not find \"" << fromNode << "\" fromNode." << endl;
			else if(!tempToNode)
				cout << "ROUTE: Could not find \"" << toNode << "\" toNode." << endl;
		}
	}

	//Go through all this node's children, if they exist.
	while((tempChild = node->IterateChildren(tempChild)))
	{
		tempChild2 = tempChild->ToElement();
		if(tempChild2)
			parseRoute(tempChild2, routedNodes);
	}
}

//------------------------------------------------------------------------------
void X3DFileLoader::parseExport(TiXmlElement *node,
					 			vector<AbstractNode *>& savedNodes,
					 			const string& inlineName)
{
	string tempstr;
	string localDEF;
	string as;
	TiXmlAttribute *tempAtt = 0;
	TiXmlNode *tempChild = 0;
	TiXmlElement *tempChild2 = 0;

	//If it's a ROUTE, inform the relevant nodes, and add it to routedNodes.
	tempstr = node->Value();
	if(tempstr == "EXPORT")
	{
		//Go through attributes, setting the appropriate values.
		tempAtt = node->FirstAttribute();
		while(tempAtt)
		{
			tempstr = tempAtt->Name();
			if(tempstr == "localDEF")
				localDEF = tempAtt->Value();
			else if(tempstr == "AS")
				as = tempAtt->Value();

			tempAtt = tempAtt->Next();
		}

		//Check that none of the relevant strings are empty.
		if(!localDEF.empty() && !as.empty())
		{
			AbstractNode *tempNode = sceneRoot->findNode(localDEF);

			//Make sure tempNode exists.
			if(tempNode)
			{
				stringstream tempConv;

				if(inlineName != "")
					tempConv << inlineName << "." << as;
				else
					tempConv << as; //Will things break if Inline has no name?
				tempNode->setDefName(tempConv.str());
				savedNodes.push_back(tempNode);
			}
		}
	}

	//Go through all this node's children, if they exist.
	while((tempChild = node->IterateChildren(tempChild)))
	{
		tempChild2 = tempChild->ToElement();
		if(tempChild2)
			parseExport(tempChild2, savedNodes, inlineName);
	}
}

//------------------------------------------------------------------------------
void X3DFileLoader::parseImport(TiXmlElement *node)
{
	string tempstr;
	string inlineDEF;
	string exportedDEF;
	string as;
	TiXmlAttribute *tempAtt = 0;
	TiXmlNode *tempChild = 0;
	TiXmlElement *tempChild2 = 0;

	//If it's an IMPORT, find the relevant node, update its DEF name.
	tempstr = node->Value();
	if(tempstr == "IMPORT")
	{
		//Go through attributes, setting the appropriate values.
		tempAtt = node->FirstAttribute();
		while(tempAtt)
		{
			tempstr = tempAtt->Name();
			if(tempstr == "InlineDEF")
				inlineDEF = tempAtt->Value();
			else if(tempstr == "exportedDEF")
				exportedDEF = tempAtt->Value();
			else if(tempstr == "AS")
				as = tempAtt->Value();

			tempAtt = tempAtt->Next();
		}

		//Check that none of the relevant strings are empty.
		if(!inlineDEF.empty() && !exportedDEF.empty() && !as.empty())
		{
			stringstream tempConv;
			AbstractNode *tempNode = 0;

			tempConv << inlineDEF << "." << exportedDEF;
			tempNode = sceneRoot->findNode(tempConv.str());

			//Make sure tempNode exists.
			if(tempNode)
				tempNode->setDefName(as);
			else
				cout << "handleIMPORTS(): Could not find " << tempConv.str() << " node." << endl;
		}
		else
			cout << "handleIMPORTS(): Missing an attribute?" << endl;
	}

	//Go through all this node's children, if they exist.
	while((tempChild = node->IterateChildren(tempChild)))
	{
		tempChild2 = tempChild->ToElement();
		if(tempChild2)
			parseImport(tempChild2);
	}
}

//------------------------------------------------------------------------------
void X3DFileLoader::clearNames(AbstractNode *node,
							   const vector<AbstractNode *> savedNodes)
{
	ParentNode *tempParent = 0;
	vector<AbstractNode *>::const_iterator it;

	//Clear node's DEF name if necessary.
	it = find(savedNodes.begin(), savedNodes.end(), node);
	if(it == savedNodes.end())
		node->setDefName("");

	//Iterate through its children if necessary.
	tempParent = dynamic_cast<ParentNode *>(node);
	if(tempParent)
	{
		int i;

		for(i=0;i<tempParent->getNumChildren();++i)
			clearNames(tempParent->getChild(i), savedNodes);
	}
}

}
