//	X3DFileLoader.h - Class to parse an X3D file and create a scene from it.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef X3DFILELOADER_H_
#define X3DFILELOADER_H_

//Stupid MSVC v6 warning about the STL...
#ifdef WIN32
#pragma warning(disable:4786)
#endif

#include <string>
#include <vector>

class TiXmlElement;

namespace HeilanX3D
{

class AbstractNode;
class SceneConstructor;

typedef std::vector<AbstractNode *> RoutedNodes;

///	Class to parse an X3D file, and create a scene from it.
/*!
	We use tinyxml for all the low-level xml stuff, because it's small and easy
	to understand (unlike libxml...).
 */
class X3DFileLoader
{
  public:
	///	Constructor.
	/*!
  		\param filepath Path to the file to try and load.

		The constructor will not do anything with filepath, you should call the
		isValidX3DFile() method to check it's a valid file, and loadScene() to
		then load the actual scene.
	 */
	X3DFileLoader(const std::string& filepath);
	///	Destructor.
	~X3DFileLoader();

	///	Returns true if file is valid, false otherwise.
	/*!
  		This will parse file, and check if it has the various relevant nodes.
	 */
	bool isValidX3DFile() const;

	///	Parses file, and returns the root node of the scene.
	/*!
		\param mainFile Whether or not to store the file's path so future file
		searches can be done relative to it's location.
		\return 0 if the file was invalid, or something else(?) went wrong.

  		The root node will be a ParentNode, with the rest of the scene as it's
		children.
	 */
	AbstractNode *loadScene(bool mainFile = false);
	///	Parses file, and returns a vector of any ROUTEd nodes in the scene.
	/*!
		Must not be called before loadScene()!!!  It will return 0 if you try
		this, or if there are no ROUTEd nodes in the scene.

		The caller is responsible for deleting the returned RoutedNodes vector.
	 */
	RoutedNodes *getRoutedNodes();
	///	Renames nodes (DEF names) according to any EXPORT statements.
	/*!
		This is/should only be used by Inline.
	 */
	void setupEXPORTNames(const std::string& inlineName);
	///	Handles any IMPORT statements.
	/*!
		This remaps any DEF names it finds that match the attributes of the
		IMPORT statements it finds.
	 */
	void handleIMPORTS();
  private:
	///	Helper method to load one node from the X3D file.
	/*!
		\param node A TiXmlElement containing the current node to load from the
		X3D file into our X3D scene.
		\param parent An AbstractNode pointer to the parent node.
		\return Pointer to the node constructed (0 if it failed for some
		reason).

		This method works recursively, so what it will do is:
		- Construct the node from NodeFactory, according to it's type.
		- Set any attributes it may have in the file.
		- Add it to the scene.
		- For each child node/xml element, call itself with that node as
		  the parameter.
		- Exit.
	 */
	AbstractNode *parseNode(TiXmlElement *node, AbstractNode *parent = 0);
	///	Helper method to gather up any ROUTEs.
	/*!
		\param node A TiXmlElement containing the current node to load from the
		X3D file into our X3D scene.
		\param routedNodes Pointer to the vector to add any discovered ROUTEs
		to.
	 */
	void parseRoute(TiXmlElement *node, RoutedNodes *routedNodes);
	///	Helper method to act on any EXPORT nodes in a scene.
	/*!
		\param node The XML node to parse.
		\param savedNodes A vector of any nodes whose name should not be cleared
		at the end of setupEXPORTNames().
		\param inlineName Prefixes the updated names, in case we have multiple
		instances of the same Inline file.
	 */
	void parseExport(TiXmlElement *node,
					 std::vector<AbstractNode *>& savedNodes,
					 const std::string& inlineName);
	///	Helper method to act on any IMPORT nodes in a scene.
	/*!
		\param node The XML node to parse.
	 */
	void parseImport(TiXmlElement *node);
	///	Helper method to clear any node's DEF name not defined in savedNodes.
	void clearNames(AbstractNode *node,
					const std::vector<AbstractNode *> savedNodes);
	///	Helper class to search through sceneRoot looking for the named node.
	AbstractNode *getNode(const std::string& name);

	///	The path to the file this object is supposed to load.
	std::string file;
	///	Helper class for constructing the scene.
	SceneConstructor *sceneConstructor;

	///	The scene root, set in loadScene().
	AbstractNode *sceneRoot;
};

}

#endif
