//	ExamineNavigation.cpp - 'Examine'-type navigation class.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "ExamineNavigation.h"
#include "Camera.h"
#include "ThreeFloats.h"
//#include "FourFloats.h"
#include "Viewpoint.h"

#include "SDL_mouse.h"
#include <iostream>
#include <cmath>

using namespace std;

namespace HeilanX3D
{

NavigationConstructor<ExamineNavigation> ExamineNavigation::constructor;

//------------------------------------------------------------------------------
ExamineNavigation::ExamineNavigation():
Navigation(),
leftIsDown(false),
middleIsDown(false),
rightIsDown(false),
lastX(0),
lastY(0),
xRot(0.0f),
yRot(0.0f),
distance(8.0f),
xPan(0.0f),
yPan(0.0f),
start(true)
{
	keyDown(SDLK_HOME);
}

//------------------------------------------------------------------------------
ExamineNavigation::~ExamineNavigation()
{
	
}

/*//------------------------------------------------------------------------------
void ExamineNavigation::draw()
{
	glTranslatef(position.x, position.y, position.z);
	glRotatef(xRotation, 0.0f, 1.0f, 0.0f);
	glRotatef(yRotation, 1.0f, 0.0f, 0.0f);
}*/

//------------------------------------------------------------------------------
void ExamineNavigation::mouseDown(int button, int x, int y)
{
	switch(button)
	{
		case SDL_BUTTON_LEFT:
			leftIsDown = true;
			break;
		case SDL_BUTTON_MIDDLE:
			middleIsDown = true;
			break;
		case SDL_BUTTON_RIGHT:
			rightIsDown = true;
			break;
	}
	lastX = x;
	lastY = y;
}

//------------------------------------------------------------------------------
void ExamineNavigation::mouseMove(int x, int y)
{
	if(leftIsDown)
	{
		yRot += static_cast<float>(x - lastX) * 0.25f;
		xRot += static_cast<float>(y - lastY) * 0.25f;

		updateCamera();
	}
	if(middleIsDown)
	{
		distance += static_cast<float>(y - lastY) * (1.0f/128.0f);
		updateCamera();
	}
	if(rightIsDown)
	{
		xPan += (x - lastX) * 0.005f;
		yPan += (y - lastY) * 0.005f;

		updateCamera();
	}

	lastX = x;
	lastY = y;
}

//------------------------------------------------------------------------------
void ExamineNavigation::mouseUp(int button, int x, int y)
{
	switch(button)
	{
		case SDL_BUTTON_LEFT:
			leftIsDown = false;
			break;
		case SDL_BUTTON_MIDDLE:
			middleIsDown = false;
			break;
		case SDL_BUTTON_RIGHT:
			rightIsDown = false;
			break;
	}
}

//------------------------------------------------------------------------------
void ExamineNavigation::mouseWheelUp()
{
	distance -= 0.25f;
	updateCamera();
}

//------------------------------------------------------------------------------
void ExamineNavigation::mouseWheelDown()
{
	distance += 0.25f;
	updateCamera();
}

//------------------------------------------------------------------------------
void ExamineNavigation::keyDown(SDLKey key)
{
	Viewpoint *activeView;

	if(key == SDLK_HOME)
	{
		BindableNodeSingleton::getInstance()->setIndexedNode("Viewpoint", 0);
		activeView = dynamic_cast<Viewpoint *>(BindableNodeSingleton::getInstance()->getActiveNode("Viewpoint"));
		if(activeView)
		{
			distance = 8.0f;
			xRot = 0.0f;
			yRot = 0.0f;
			if(!start)
				updateCamera();
			else
				start = false;
		}
	}
	else if(key == SDLK_END)
	{
		int tempint;

		tempint = BindableNodeSingleton::getInstance()->getNumNodes("Viewpoint")-1;
		BindableNodeSingleton::getInstance()->setIndexedNode("Viewpoint", tempint);
		activeView = dynamic_cast<Viewpoint *>(BindableNodeSingleton::getInstance()->getActiveNode("Viewpoint"));
		if(activeView)
		{
			distance = 8.0f;
			xRot = 0.0f;
			yRot = 0.0f;
			updateCamera();
		}
	}
	else if(key == SDLK_PAGEUP)
	{
		BindableNodeSingleton::getInstance()->nextNode("Viewpoint");
		activeView = dynamic_cast<Viewpoint *>(BindableNodeSingleton::getInstance()->getActiveNode("Viewpoint"));
		if(activeView)
		{
			distance = 8.0f;
			xRot = 0.0f;
			yRot = 0.0f;
			updateCamera();
		}
	}
	else if(key == SDLK_PAGEDOWN)
	{
		BindableNodeSingleton::getInstance()->previousNode("Viewpoint");
		activeView = dynamic_cast<Viewpoint *>(BindableNodeSingleton::getInstance()->getActiveNode("Viewpoint"));
		if(activeView)
		{
			distance = 8.0f;
			xRot = 0.0f;
			yRot = 0.0f;
			updateCamera();
		}
	}
}

//------------------------------------------------------------------------------
void ExamineNavigation::keyUp(SDLKey key)
{
	
}

//------------------------------------------------------------------------------
void ExamineNavigation::screenRefresh()
{
	ThreeFloats tempPos;
	Quaternion tempRot;
	Viewpoint *activeView;

	activeView = dynamic_cast<Viewpoint *>(BindableNodeSingleton::getInstance()->getActiveNode("Viewpoint"));
	if(activeView)
	{
		tempPos = activeView->getPosition();
		tempRot = activeView->getRotation();

		if(viewpointPos != tempPos)
		{
			updateCamera();
			viewpointPos = tempPos;
			viewpointRot = tempRot;
		}
		else if(viewpointRot != tempRot)
		{
			updateCamera();
			viewpointRot = tempRot;
		}
	}
}

//------------------------------------------------------------------------------
ThreeFloats ExamineNavigation::getPosition() const
{
	ThreeFloats tempPos;
	Quaternion tempRot;
	Quaternion tempRot2;
	Quaternion camRot(Camera::getInstance()->getGraphicRotation());
	Viewpoint *activeView;

	activeView = dynamic_cast<Viewpoint *>(BindableNodeSingleton::getInstance()->getActiveNode("Viewpoint"));
	if(activeView)
	{
		tempPos = activeView->getPosition();
		tempRot2 = activeView->getRotation();
	}

	tempPos.x += xPan;
	tempPos.y += yPan;
	tempPos.z += distance;

	tempRot.setFromEulerAngles(xRot, yRot, 0.0);
	tempRot = tempRot * tempRot2;
	camRot *= tempRot;

	return tempPos;
}

//------------------------------------------------------------------------------
Quaternion ExamineNavigation::getRotation() const
{
	ThreeFloats tempPos;
	Quaternion tempRot;
	Quaternion tempRot2;
	Quaternion camRot(Camera::getInstance()->getGraphicRotation());
	Viewpoint *activeView;

	activeView = dynamic_cast<Viewpoint *>(BindableNodeSingleton::getInstance()->getActiveNode("Viewpoint"));
	if(activeView)
	{
		tempPos = activeView->getPosition();
		tempRot2 = activeView->getRotation();
	}

	tempPos.x += xPan;
	tempPos.y += yPan;
	tempPos.z += distance;

	tempRot.setFromEulerAngles(xRot, yRot, 0.0);
	tempRot = tempRot * tempRot2;
	camRot *= tempRot;

	return tempRot;
}

//------------------------------------------------------------------------------
void ExamineNavigation::updateCamera()
{
	ThreeFloats tempPos;
	Quaternion tempRot;
	Quaternion tempRot2;
	Quaternion camRot(Camera::getInstance()->getGraphicRotation());
	Viewpoint *activeView;

	activeView = dynamic_cast<Viewpoint *>(BindableNodeSingleton::getInstance()->getActiveNode("Viewpoint"));
	if(activeView)
	{
		tempPos = activeView->getPosition();
		tempRot2 = activeView->getRotation();
	}

	tempPos.x += xPan;
	tempPos.y += yPan;
	tempPos.z += distance;

	tempRot.setFromEulerAngles(xRot, yRot, 0.0);
	tempRot = tempRot * tempRot2;
	camRot *= tempRot;

	Camera::getInstance()->setGraphicRotation(tempRot);
	Camera::getInstance()->setGraphicPosition(tempPos);
}

}
