//	Navigation.h - Base class for all classes controlling how the camera moves.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef NAVIGATION_H_
#define NAVIGATION_H_

//Stupid MSVC v6 warning about the STL...
#ifdef WIN32
#pragma warning(disable:4786)
#endif

#include "ThreeFloats.h"
//#include "FourFloats.h"
#include "Quaternion.h"
#include "DllStuff.h"

#include "SDL_keysym.h"
#include <map>
#include <string>

namespace HeilanX3D
{

//------------------------------------------------------------------------------
///	Base class for all classes controlling how the camera moves.
class dllStuff Navigation
{
  public:
	///	Constructor.
	Navigation();
	///	Destructor.
	virtual ~Navigation() {};

	///	Called on a mouse button down event.
	virtual void mouseDown(int button, int x, int y) {};
	///	Called on a mouse button move event.
	virtual void mouseMove(int x, int y) {};
	///	Called on a mouse button up event.
	virtual void mouseUp(int button, int x, int y) {};
	///	Called on a mouse wheel up event.
	virtual void mouseWheelUp() {};
	///	Called on a mouse wheel down event.
	virtual void mouseWheelDown() {};

	///	Called on a key down event.
	virtual void keyDown(SDLKey key) {};
	///	Called on a key up event.
	virtual void keyUp(SDLKey key) {};

	///	Called for every screen refresh.
	/*!
		This is used for navigation types which need to update the camera's
		position outside of any mouse/key events.

		(use SDL_GetTicks() to get ~accurate timing)
	 */
	virtual void screenRefresh() {};

	///	Used to set whether we're rendering offline or not.
	void setIsOffline(bool val) {offline = val;};
	///	Returns whether we're rendering offline or not.
	bool getIsOffline() const {return offline;};

	///	Called by Camera to get the initial position for a Navigation type.
	virtual ThreeFloats getPosition() const {return 0.0f;};
	///	Called by Camera to get the initial rotation for a Navigation type.
	virtual Quaternion getRotation() const {return Quaternion();};
  private:
	///	Whether we're rendering offline or not.
	bool offline;
};

//------------------------------------------------------------------------------
///	Base class for NavigationConstructor.
/*!
	\sa NavigationConstructor
 */
class dllStuff NavigationConstructorBase
{
  public:
	///	Constructor (does nothing).
	NavigationConstructorBase() {};
	///	Destructor (does nothing).
	virtual ~NavigationConstructorBase() {};

	///	Returns a new instance of an Navigation subclass.
	virtual Navigation *createInstance() const = 0;
};

//------------------------------------------------------------------------------
///	A simple template class used to construct navigations from a factory.
/*!
	All subclasses of Navigation that can occur in an X3D file should have one
	of these.  You would do something like the following, and then register
	thingConstructor with NavigationFactory...
	\code
	class thing : public Navigation
	{
	  public:
		...
		static NavigationConstructor<thing> thingConstructor;
	};
	\endcode
 */
template <class T>
class dllStuff NavigationConstructor : public NavigationConstructorBase
{
  public:
	///	Constructor (does nothing).
	NavigationConstructor() {};
	///	Destructor (does nothing).
	~NavigationConstructor() {};

	///	Returns a new instance of an AbstractNode subclass.
	Navigation *createInstance() const {return new T;};
};

//------------------------------------------------------------------------------
///	Singleton class used to construct Navigation classes according to NavigationInfo's type attribute.
/*!
	\todo Make this stuff more generic, so it can be merged with NodeFactory,
	NodeConstructor etc.?
 */
class dllStuff NavigationFactory
{
  public:
	///	Returns the sole instance of NavigationConstructor.
	static NavigationFactory *getInstance();

	///	Returns a newly-constructed Navigation instance according to text.
	/*!
		\param text The name given to the Navigation type in the X3D file
		(EXAMINE, ANY, etc.).
	 */
	Navigation *createNav(const std::string& text) const;

	///	Registers a NavigationConstructor for a Navigation subclass.
	/*!
		\param text The name given to the Navigation type in the X3D file
		(EXAMINE, ANY, etc.).
		\param constructor Pointer to the Navigation subclass' constructor
		object.
	 */
	void registerNavigation(const std::string text,
							NavigationConstructorBase *constructor);

//Stupid fix for MSVC v6
#ifdef WIN32
	///	Destructor.
	~NavigationFactory();
#endif
  private:
	///	Constructor.
	NavigationFactory();
#ifndef WIN32
	///	Destructor.
	~NavigationFactory();
#endif

	///	Map containing all the available Navigation types.
	std::map<std::string, NavigationConstructorBase *> navTypes;
};

}

#endif
