//	SoundOutput.cpp - Classes handling sound output.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "SoundOutput.h"
#include "FatalError.h"
#include "portaudio.h"
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

//------------------------------------------------------------------------------
SoundOutput::SoundOutput():
running(false),
samplerate(44100.0f),
callback(0),
blockSize(256),
audioApi(-1),
audioDevice(-1),
audioStream(0)
{
	PaError err;

	err = Pa_Initialize();
	if(err != paNoError)
	{
		ostringstream errst;
		errst << "Could not initialise PortAudio: " << Pa_GetErrorText(err);
		throw FatalError(errst.str());
	}
}

//------------------------------------------------------------------------------
SoundOutput::~SoundOutput()
{
	PaError err;

	if(running)
		stopAudio();

	err = Pa_Terminate();
	if(err != paNoError)
	{
		//This doesn't throw a FatalError, in the hopes that we can at least
		//clean up the rest of the stuff before the app exits.
		cout << "Uh Oh...  Could not terminate PortAudio: ";
		cout << Pa_GetErrorText(err) << endl;
	}
}

//------------------------------------------------------------------------------
bool SoundOutput::startAudio()
{
	bool retval = false;
	PaError err;
	PaStreamParameters streamParam;
	PaStreamParameters inputParam;
	PaStreamParameters *inputP;

	if(!running)
	{
		const PaHostApiInfo *tempInfo = Pa_GetHostApiInfo(static_cast<PaHostApiIndex>(audioApi));

		//Check if our (input) audio format is available.
		if(tempInfo)
		{
			inputDevice = tempInfo->defaultInputDevice;
			if(inputDevice != -1)
				inputNumChannels = Pa_GetDeviceInfo(inputDevice)->maxInputChannels;
		}
		inputParam.device = inputDevice;
		inputParam.channelCount = inputNumChannels;
		inputParam.sampleFormat = paFloat32;
		inputParam.suggestedLatency = 0.0;
		inputParam.hostApiSpecificStreamInfo = 0;

		if(inputDevice != -1)
			inputP = &inputParam;
		else
			inputP = 0;

		//Check if our (output) audio format is available.
		streamParam.device = audioDevice;
		streamParam.channelCount = speakers.getNumChannels();
		streamParam.sampleFormat = paFloat32;
		streamParam.suggestedLatency = 0.0;
		streamParam.hostApiSpecificStreamInfo = 0;
		//0 == success here...
		if(!(err = Pa_IsFormatSupported(inputP, &streamParam, samplerate)))
		{
			//Open audio stream.
			callbackStruct.speakers = &speakers;
			callbackStruct.callback = callback;
			err = Pa_OpenStream(&audioStream,
								inputP,
								&streamParam,
								samplerate,
								0,
								paNoFlag,
								&portAudioCallback,
								static_cast<void *>(&callbackStruct));
			if(err != paNoError)
				cout << "Could not open PortAudio stream." << endl;
			else
			{
				if(inputDevice != -1)
					callbackStruct.callback->setNumInputChannels(inputNumChannels);

				//Start audio stream.
				err = Pa_StartStream(audioStream);
				if(err != paNoError)
				{
					cout << "Could not start PortAudio stream" << endl;
					Pa_CloseStream(audioStream);
					audioStream = 0;
				}
				else
				{
					running = true;
					retval = true;
				}
			}
		}
		else
		{
			cout << "The audio stream format is not supported..." << endl;
			cout << "Device: " << static_cast<int>(audioDevice) << endl;
			cout << "Channel Count: " << speakers.getNumChannels() << endl;
			cout << "Sample Format: paFloat32" << endl;
			cout << "Samplerate: " << samplerate << endl;
			cout << "Input Device: " << static_cast<int>(inputDevice) << endl;
			cout << "Input Channel Count: " << inputNumChannels << endl;
			cout << "PortAudio error: " << Pa_GetErrorText(err) << endl;
		}
	}
	else
		cout << "Call to startAudio() while audio is already running." << endl;

	return retval;
}

//------------------------------------------------------------------------------
void SoundOutput::stopAudio()
{
	PaError err;

	if(running)
	{
		//Stop the audio stream.
		err = Pa_StopStream(audioStream);
		if(err != paNoError)
		{
			cout << "Runaway steam train!  Could not stop PortAudio stream!";
			cout << endl;
			return;
		}

		//Close the audio stream.
		err = Pa_CloseStream(audioStream);
		if(err != paNoError)
		{
			cout << "Could not close PortAudio stream." << endl;
			return;
		}

		audioStream = 0;
		running = false;
	}
}

//------------------------------------------------------------------------------
void SoundOutput::setSamplerate(float val)
{
	if(!running)
		samplerate = val;
	else
	{
		cout << "SoundOutput: Trying to set the samplerate while audio is ";
		cout << "running is not allowed." << endl;
	}
}

//------------------------------------------------------------------------------
void SoundOutput::setCallback(SoundOutputCallback *newCallback)
{
	if(!running)
		callback = newCallback;
	else
	{
		cout << "SoundOutput: Trying to set the callback while audio is ";
		cout << "running is not allowed." << endl;
	}
}

//------------------------------------------------------------------------------
void SoundOutput::setSpeakerConfig(const SpeakerConfig& newSpeakers)
{
	if(!running)
		speakers = newSpeakers;
	else
	{
		cout << "SoundOutput: Trying to set the speaker config while audio is ";
		cout << "running is not allowed." << endl;
	}

	//For offline rendering.
	callbackStruct.speakers = &speakers;
	callbackStruct.callback = callback;
}

//------------------------------------------------------------------------------
void SoundOutput::setBlockSize(int val)
{
	if(!running)
		blockSize = val;
	else
	{
		cout << "SoundOutput: Trying to set the block size while audio is ";
		cout << "running is not allowed." << endl;
	}
}

//------------------------------------------------------------------------------
void SoundOutput::setDevice(int index)
{
	PaDeviceIndex temp;

	if(index == -1)
	{
		const PaHostApiInfo *tempInfo = Pa_GetHostApiInfo(static_cast<PaHostApiIndex>(audioApi));
		if(tempInfo)
		{
			audioDevice = tempInfo->defaultOutputDevice;
			cout << "Using (default) audioDevice: " << audioDevice << endl;
		}
		else
			cout << "Could not get default output device..." << endl;
	}
	else
	{
		temp = Pa_HostApiDeviceIndexToDeviceIndex(static_cast<PaHostApiIndex>(audioApi),
									   			  index);
		if(temp == paInvalidHostApi)
		{
			cout << "SoundOutput: Could not set audio device: " << index << endl;
			cout << "PortAudio error: " << Pa_GetErrorText(temp) << endl;
		}
		else if(temp == paInvalidDevice)
		{
			cout << "SoundOutput: Could not set audio device: " << index << endl;
			cout << "PortAudio error: " << Pa_GetErrorText(temp) << endl;

			cout << "Trying default output device..." << endl;
			{
				const PaHostApiInfo *tempInfo = Pa_GetHostApiInfo(static_cast<PaHostApiIndex>(audioApi));
				if(tempInfo)
				{
					audioDevice = tempInfo->defaultOutputDevice;
					cout << "Using audioDevice: " << audioDevice << endl;
				}
				else
					cout << "Could not get default output device..." << endl;
			}
		}
		else
			audioDevice = temp;
	}
}

//------------------------------------------------------------------------------
void SoundOutput::setInputDevice(int index)
{
	PaDeviceIndex temp;

	if(index == -1)
	{
		const PaHostApiInfo *tempInfo = Pa_GetHostApiInfo(static_cast<PaHostApiIndex>(audioApi));
		if(tempInfo)
		{
			inputDevice = tempInfo->defaultInputDevice;
			cout << "Using (default) audioInputDevice: " << inputDevice << endl;
		}
		else
			cout << "Could not get default input device..." << endl;
	}
	else
	{
		temp = Pa_HostApiDeviceIndexToDeviceIndex(static_cast<PaHostApiIndex>(audioApi),
									   			  index);
		if(temp == paInvalidHostApi)
		{
			cout << "SoundOutput: Could not set audio input device: " << index << endl;
			cout << "PortAudio error: " << Pa_GetErrorText(temp) << endl;
		}
		else if(temp == paInvalidDevice)
		{
			cout << "SoundOutput: Could not set audio input device: " << index << endl;
			cout << "PortAudio error: " << Pa_GetErrorText(temp) << endl;

			cout << "Trying default input device..." << endl;
			{
				const PaHostApiInfo *tempInfo = Pa_GetHostApiInfo(static_cast<PaHostApiIndex>(audioApi));
				if(tempInfo)
				{
					inputDevice = tempInfo->defaultInputDevice;
					cout << "Using audioInputDevice: " << inputDevice << endl;
				}
				else
					cout << "Could not get default input device..." << endl;
			}
		}
		else
			inputDevice = temp;
	}
}

//------------------------------------------------------------------------------
void SoundOutput::setApi(int val)
{
	if(val != -1)
		audioApi = val;
	else
	{
#ifdef WIN32
		//For some reason, PortAudio's MME implementation takes up an unusable
		//number of clock cycles on my machine, so we try and use DirectSound as
		//the default to avoid that.
		for(int i=0;i<Pa_GetHostApiCount();++i)
		{
			const PaHostApiInfo *info = Pa_GetHostApiInfo(i);
			if(info->type == paDirectSound)
			{
				audioApi = i;
				break;
			}
		}
		//Fallback in case for some reason your system doesn't have DirectSound
		//support.
		if(audioApi == -1)
			audioApi = Pa_GetDefaultHostApi();
#else
		audioApi = Pa_GetDefaultHostApi();
#endif
	}
}

//------------------------------------------------------------------------------
int SoundOutput::portAudioCallback(const void *input,
								   void *output,
								   unsigned long frameCount,
								   const PaStreamCallbackTimeInfo* timeInfo,
								   PaStreamCallbackFlags statusFlags,
								   void *userData)
{
	AmbisonicData *audio = 0;
	CallbackStruct *thing = static_cast<CallbackStruct *>(userData);

	//Set the input audio.
	thing->callback->setInputAudio((float *)input);

	//Get audio for the scene.
	audio = thing->callback->processBlock(frameCount);

	//Convert to float to send to the soundcard.
	if(audio)
	{
		thing->speakers->convertAmbi2Float(audio,
										   static_cast<float *>(output),
										   frameCount);
	}
	else
		cout << "No audio block returned from scene." << endl;

	//Clear the input audio.
	thing->callback->setInputAudio(0);

	return paContinue;
}

//------------------------------------------------------------------------------
void SoundOutput::enumerateSoundcards()
{
	int i, j;
	int numAPIs;
	int defaultAPI;
	PaError ernie;

	ernie = Pa_Initialize();

	if(ernie != paNoError)
	{
		return;
	}

	numAPIs = Pa_GetHostApiCount();
	defaultAPI = Pa_GetDefaultHostApi();

	cout << endl;
	cout << "Available APIs:" << endl;
	cout << "---------------" << endl;
	for(i=0;i<numAPIs;++i)
	{
		const PaHostApiInfo *apiInfo = Pa_GetHostApiInfo(i);

		cout << i << ": " << apiInfo->name;
		if(i == defaultAPI)
			cout << " (default)" << endl;
		else
			cout << endl;
	}

	cout << endl;
	cout << "Available Devices:" << endl;
	cout << "------------------" << endl;
	for(i=0;i<numAPIs;++i)
	{
		const PaHostApiInfo *apiInfo = Pa_GetHostApiInfo(i);

		cout << apiInfo->name << " API:" << endl;

		for(j=0;j<apiInfo->deviceCount;++j)
		{
			const PaDeviceInfo *deviceInfo = Pa_GetDeviceInfo(Pa_HostApiDeviceIndexToDeviceIndex(i, j));

			cout << "\t" << j << ": " << deviceInfo->name;
			if(Pa_HostApiDeviceIndexToDeviceIndex(i, j) == apiInfo->defaultOutputDevice)
				cout << " (default)" << endl;
			else if(Pa_HostApiDeviceIndexToDeviceIndex(i, j) == apiInfo->defaultInputDevice)
				cout << " (default)" << endl;
			else
				cout << endl;
			if(deviceInfo->maxOutputChannels == 0)
				cout << "\t\t" << "Max input channels: " << deviceInfo->maxInputChannels << endl;
			else
				cout << "\t\t" << "Max output channels: " << deviceInfo->maxOutputChannels << endl;
		}
		cout << endl;
	}
}
