//	SoundOutput.h - Classes handling sound output.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SOUNDOUTPUT_H_
#define SOUNDOUTPUT_H_

#include "SpeakerConfig.h"

class SoundOutputCallback;

//This stuff is so we don't have to include portaudio.h here.  Is it really
//worth it?
typedef int PaHostApiIndex;
typedef int PaDeviceIndex;
typedef void PaStream;
struct PaStreamCallbackTimeInfo;
typedef unsigned long PaStreamCallbackFlags;

//------------------------------------------------------------------------------
///	Simple struct passed to SoundOutput::portAudioCallback().
typedef struct
{
	///	Pointer to SoundOutput's SpeakerConfig.
	SpeakerConfig *speakers;
	///	Pointer to SoundOutput's current callback instance.
	SoundOutputCallback *callback;
} CallbackStruct;

//------------------------------------------------------------------------------
///	Class to take care of all sound output, using PortAudio.
class SoundOutput
{
  public:
	///	Constructor.
	SoundOutput();
	///	Destructor.
	~SoundOutput();

	///	Starts the audio output.
	/*!
		\return True if successful, false otherwise.
	 */
	bool startAudio();
	///	Stops the audio output.
	void stopAudio();
	///	Whether or not audio is currently running.
	bool isRunning() const {return running;};

	///	Sets the samplerate.
	/*!
		Call this before you call startAudio().
	 */
	void setSamplerate(float val);
	///	Sets the callback to call for every audio block.
	/*!
		Call this before you call startAudio().

		SoundOutput will not delete newCallback - it is assumed you will have a
		mechanism of your own to do that.
	 */
	void setCallback(SoundOutputCallback *newCallback);
	///	Sets the speaker configuration to use.
	/*!
		This will also set the number of output channels used.
		Call this before you call startAudio().
	 */
	void setSpeakerConfig(const SpeakerConfig& newSpeakers);
	///	Sets the blocksize to use.
	/*!
		Call this before you call startAudio().
	 */
	void setBlockSize(int val);
	///	Sets the audio API to use.
	/*!
		If index == -1, the default api (according to PortAudio) is used.
	 */
	void setApi(int val);
	///	Sets the audio device to use.
	/*!
		If index == -1, the default output device is used.
	 */
	void setDevice(int index);
	///	Sets the input audio device to use.
	/*!
		If index == -1, the default output device is used.
	 */
	void setInputDevice(int index);

	///	Returns the current block size.
	int getBlockSize() const {return blockSize;};
	///	Returns the current samplerate.
	float getSamplerate() const {return samplerate;};

	///	Returns the object's CallbackStruct.
	CallbackStruct *getCallbackStruct() {return &callbackStruct;};

	///	The function to be called by PortAudio.
	/*!
		\todo The soundfield rotation uses a lot of cosf and sinf calls - can
		this be sped up?
		\todo Interpolate rotation to avoid zipper noise.
	 */
	static int portAudioCallback(const void *input,
								 void *output,
								 unsigned long frameCount,
								 const PaStreamCallbackTimeInfo* timeInfo,
								 PaStreamCallbackFlags statusFlags,
								 void *userData);

	///	Helper method enumerate the possible soundcards to the command line.
	static void enumerateSoundcards();
  private:
	///	Whether or not the audio is currently running.
	bool running;
	///	The samplerate we're using.
	float samplerate;
	///	The callback we're using.
	SoundOutputCallback *callback;
	///	The speaker configuration we're using.
	SpeakerConfig speakers;
	///	The block size we're using.
	int blockSize;
	///	Callback struct passed to portAudioCallback().
	CallbackStruct callbackStruct;

	//----PortAudio-specific Stuff----
	///	(PortAudio) The API to use.
	/*!
		At the moment this is hard-coded - jack on linux, asio on windows,
		core audio on osx.

		\todo Allow user to set.
	 */
	PaHostApiIndex audioApi;
	///	(PortAudio) The output device selected from setDevice().
	PaDeviceIndex audioDevice;
	///	(PortAudio) The audio stream.
	PaStream *audioStream;

	///	(PortAudio) Input audio device to use.
	PaDeviceIndex inputDevice;
	///	(PortAudio) The number of input channels our device has.
	int inputNumChannels;
};

//------------------------------------------------------------------------------
///	Callback class, called for every audio block from SoundOutput.
/*!
	The idea is that, to output sound, you will subclass this class, and
	override its processBlock() method, then pass your subclass to SoundOutput
	to take care of the all the setup etc.
 */
class SoundOutputCallback
{
  public:
	///	Constructor.
	SoundOutputCallback():
	inputAudio(0),
	numInputChannels(0)
	{
		
	};
	///	Destructor.
	virtual ~SoundOutputCallback() {};

	///	Method called for every audio block, to process/generate audio.
	/*!
		\param numSamples Number of samples to process.

  		\return Pointer to the block of samples you've just generated.
	 */
	virtual AmbisonicData *processBlock(unsigned int numSamples) = 0;

	///	Returns the input audio for this block.
	/*!
		This should return 0 if we're not in the middle of an audio processing
		block.

		Note that the returned buffer will be interleaved, depending on how
		many channels we're receiving as input.
	 */
	float *getInputAudio() {return inputAudio;};
	///	Returns the number of channels in the input audio.
	int getNumInputChannels() const {return numInputChannels;};

	///	Sets the input audio for this block (called from SoundOutput).
	void setInputAudio(float *audio)
	{
		inputAudio = audio;
	};
	///	Sets the number of channels in the input audio (called from SoundOutput).
	void setNumInputChannels(int val)
	{
		numInputChannels = val;
	};
  private:
	///	The (interleaved) input audio.
	float *inputAudio;
	///	The number of channels in the input audio.
	int numInputChannels;
};

#endif
