//	SpeakerConfig.h - Classes to handle the various speaker configurations.
//  ----------------------------------------------------------------------------
//	This file is part of 'Heilan Browser', an X3D browser specialised for audio
//	work.
//	Copyright (C) 2005-2009  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SPEAKERCONFIG_H_
#define SPEAKERCONFIG_H_

#include "AmbisonicData.h"

#include <vector>

//------------------------------------------------------------------------------
///	Speaker configuration base class.
class dllStuff SpeakerConfig
{
  public:
	///	Constructor.
	SpeakerConfig();

	///	Converts a block of Ambisonic audio into the correct float speaker feeds.
	/*!
		\param input Block of Ambisonic data to convert.
		\param output Float buffer to hold the interleaved output.
		\param numSamples The number of samples to process.
	 */
	void convertAmbi2Float(AmbisonicData *input,
						   float *output,
						   int numSamples) const;

	///	Returns the number of channels this SpeakerConfig has.
	int getNumChannels() const {return matrix.size();};
  protected:
	///	Called by subclasses to set the decoding matrix.
	void setDecodingMatrix(const std::vector<AmbisonicData>& newMatrix);
  private:
	///	Holds the decoding matrix for this speaker configuration.
	std::vector<AmbisonicData> matrix;
};

//------------------------------------------------------------------------------
///	Mono speaker configuration.
class MonoConfig : public SpeakerConfig
{
  public:
	///	Constructor.
	MonoConfig();
};

//------------------------------------------------------------------------------
///	Stereo speaker configuration.
class StereoConfig : public SpeakerConfig
{
  public:
	///	Constructor.
	StereoConfig();
};

//------------------------------------------------------------------------------
///	Square speaker configuration.
class SquareConfig : public SpeakerConfig
{
  public:
	///	Constructor.
	SquareConfig();
};

//------------------------------------------------------------------------------
///	Pentagon speaker configuration.
class PentagonConfig : public SpeakerConfig
{
  public:
	///	Constructor.
	PentagonConfig();
};

//------------------------------------------------------------------------------
///	Hexagon speaker configuration.
class HexagonConfig : public SpeakerConfig
{
  public:
	///	Constructor.
	HexagonConfig();
};

//------------------------------------------------------------------------------
///	Octagon1 speaker configuration.
class Octagon1Config : public SpeakerConfig
{
  public:
	///	Constructor.
	Octagon1Config();
};

//------------------------------------------------------------------------------
///	Octagon2 speaker configuration.
class Octagon2Config : public SpeakerConfig
{
  public:
	///	Constructor.
	Octagon2Config();
};

//------------------------------------------------------------------------------
///	Surround1 speaker configuration.
class Surround1Config : public SpeakerConfig
{
  public:
	///	Constructor.
	Surround1Config();
};

//------------------------------------------------------------------------------
///	Cube speaker configuration.
class CubeConfig : public SpeakerConfig
{
  public:
	///	Constructor.
	CubeConfig();
};

//------------------------------------------------------------------------------
///	Dodecahedron1 speaker configuration.
class Dodecahedron1Config : public SpeakerConfig
{
  public:
	///	Constructor.
	Dodecahedron1Config();
};

//------------------------------------------------------------------------------
///	Dodecahedron2 speaker configuration.
class Dodecahedron2Config : public SpeakerConfig
{
  public:
	///	Constructor.
	Dodecahedron2Config();
};

//------------------------------------------------------------------------------
///	Straight ambisonic components speaker configuration.
class ComponentsConfig : public SpeakerConfig
{
  public:
	///	Constructor.
	ComponentsConfig();
};

#endif
